Package: geoBayes
Type: Package
Title: Analysis of Geostatistical Data using Bayes and Empirical Bayes
        Methods
Description: Functions to fit geostatistical data. The data can be
        continuous, binary or count data and the models implemented are
        flexible. Conjugate priors are assumed on some parameters while
        inference on the other parameters can be done through a full
        Bayesian analysis of by empirical Bayes methods.
Version: 0.7.7
Date: 2026-06-17
Authors@R: c(person(given = "Evangelos", family = "Evangelou", role =
        c("aut", "cre"), email = "e.evangelou@maths.bath.ac.uk"),
        person(given = "Vivekananda", family = "Roy", role = "aut", email
        = "vroy@iastate.edu"))
Maintainer: Evangelos Evangelou <e.evangelou@maths.bath.ac.uk>
Imports: coda, graphics, sp, stats, optimx
Depends: R (>= 3.0.0)
Suggests:
LazyData: true
Encoding: UTF-8
License: GPL (>= 2)
Classification/ACM: 86A32, 62M30, 62F15
ByteCompile: yes
NeedsCompilation: yes
RoxygenNote: 7.3.1
Packaged: 2026-06-17 12:30:53 UTC; ee224
Author: Evangelos Evangelou [aut, cre],
  Vivekananda Roy [aut]
Repository: CRAN
Date/Publication: 2026-06-17 12:50:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-06-17 15:39:24 UTC; unix
Archs: geoBayes.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(spcovariance,dist)
S3method(spcovariance,formula)
S3method(spcovariance,numeric)
S3method(subset,geomcmc)
export(.geoBayes_corrfcn)
export(.geoBayes_models)
export(alik_cutoff)
export(alik_inla)
export(alik_optim)
export(alik_plot)
export(bf1skel)
export(bf2new)
export(bf2optim)
export(bf2se)
export(bmbfse)
export(linkfcn)
export(linkinv)
export(mcmcmake)
export(mcsglmm)
export(mcsglmm_mala)
export(mcstrga)
export(mcstrga_mala)
export(mkpredgrid2d)
export(plotbf2)
export(revlogreg)
export(rsglmm)
export(rsgrf)
export(rstrga)
export(select_proposals_ENT)
export(select_proposals_MNX)
export(select_proposals_SEQ)
export(spcovariance)
export(sploglik)
export(sploglik_cross)
export(stackdata)
importFrom(coda,mcmc)
importFrom(coda,mcmc.list)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,segments)
importFrom(graphics,title)
importFrom(optimx,optimr)
importFrom(sp,point.in.polygon)
importFrom(sp,spDists)
importFrom(stats,approx)
importFrom(stats,as.formula)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.offset)
importFrom(stats,model.response)
importFrom(stats,model.weights)
importFrom(stats,optim)
importFrom(stats,optimHess)
importFrom(stats,plogis)
importFrom(stats,pnorm)
importFrom(stats,pt)
importFrom(stats,qf)
importFrom(stats,qlogis)
importFrom(stats,qnorm)
importFrom(stats,qt)
importFrom(stats,rbeta)
importFrom(stats,rbinom)
importFrom(stats,rgamma)
importFrom(stats,rnorm)
importFrom(stats,rpois)
importFrom(stats,runif)
importFrom(stats,update)
importFrom(stats,var)
useDynLib(geoBayes, .registration = TRUE)
useDynLib(geoBayes,bfse_mu)
useDynLib(geoBayes,bfse_no)
useDynLib(geoBayes,bfse_tr)
useDynLib(geoBayes,bfse_wo)
useDynLib(geoBayes,bfsp_mu)
useDynLib(geoBayes,bfsp_no)
useDynLib(geoBayes,bfsp_tr)
useDynLib(geoBayes,bfsp_wo)
useDynLib(geoBayes,calcb_mu_cv)
useDynLib(geoBayes,calcb_mu_st)
useDynLib(geoBayes,calcb_no_cv)
useDynLib(geoBayes,calcb_no_st)
useDynLib(geoBayes,calcb_tr_cv)
useDynLib(geoBayes,calcb_tr_st)
useDynLib(geoBayes,calcb_wo_cv)
useDynLib(geoBayes,calcb_wo_st)
useDynLib(geoBayes,calcbd_mu)
useDynLib(geoBayes,calcbd_no)
useDynLib(geoBayes,calcbd_tr)
useDynLib(geoBayes,calcbd_wo)
useDynLib(geoBayes,flinkfcn)
useDynLib(geoBayes,flinkinv)
useDynLib(geoBayes,llikfcn_dh_tr)
useDynLib(geoBayes,llikfcn_mu)
useDynLib(geoBayes,llikfcn_no)
useDynLib(geoBayes,llikfcn_tr)
useDynLib(geoBayes,llikfcn_wo)
useDynLib(geoBayes,llikparscalc)
useDynLib(geoBayes,llikparsval)
useDynLib(geoBayes,mcspsample)
useDynLib(geoBayes,mcspsamtry)
useDynLib(geoBayes,spcorr)
useDynLib(geoBayes,trgasample)
useDynLib(geoBayes,trgasamtry)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Compute the Bayes factors using the z sample
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine bfsp_no (weights, zcv, logbf, lglk1, lglk2, &
   philist, omglist, nulist, &
   zsample1, Nout1, Ntot1, zsample2, Nout2, Ntot2, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   kappalist, icf, n, p, kg, ifam, imeth, itr)
  use modelfcns, jointyz_sp => jointyz
  use interfaces
  use linkfcns
  use flogsumexp
  use bmargin
  use covfun
  use jointyz, only: jointyz_gt
  use betaprior
  implicit none
  integer, intent(in) :: n, p, kg, ifam, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     zsample1(n,Ntot1), zsample2(n,Ntot2), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: logbf(kg), lglk1(Ntot1,kg), &
     lglk2(Ntot2,kg), weights(Ntot2), zcv(Ntot2,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, ssqdfsc, modeldfh, &
     tsqdfsc, respdfh, xi(n), eta(kg), mxlglk, lglketa(Ntot2, kg), nu
  integer i, j, m
  double precision zsam(n), logNN2(kg)

  call create_model (ifam)
  call create_spcor (icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  logNN2 = log(dble(Ntot2)) - log(dble(Nout2))

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  select case (ifam)
  case (0)
    do i = 1, kg
      nu = nulist(i)
      call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
         kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot1
        call rchkusr
        lglk1(j,i) = jointyz_gt(n, zsample1(:,j), y, l, Ups, ldh_Ups, &
           nu, xi, lmxi, ssqdfsc, tsqdfsc, modeldfh, respdfh)
      end do
      do j = 1, Ntot2
        call rchkusr
        lglk2(j,i) = jointyz_gt(n, zsample2(:,j), y, l, Ups, ldh_Ups, &
           nu, xi, lmxi, ssqdfsc, tsqdfsc, modeldfh, respdfh)
      end do
    end do
  case default
    do i = 1, kg
      nu = nulist(i)
      call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
         kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot1
        call rchkusr
        lglk1(j,i) = jointyz_sp(n, zsample1(:,j), &
           y, l, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
      end do
      do j = 1, Ntot2
        call rchkusr
        lglk2(j,i) = jointyz_sp(n, zsample2(:,j), &
           y, l, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
      end do
    end do
  end select
  mxlglk = maxval(lglk1)
  lglk1 = lglk1 - mxlglk
  ! lglk2 = lglk2 - mxlglk XXX Don't do this; may err SE calc different fcn

  ! Use the sample in reverse logistic regression
  eta = log(dble(Nout1))
  select case (imeth)
  case (1)
    call revlogistic (eta,lglk1,kg,Ntot1,Nout1)
  case (2)
    call revlogistic (eta,lglk1,kg,Ntot1,Nout1)
    call mengwong (eta,lglk1,kg,Ntot1,Nout1)
  end select
  ! eta_i = log(N_i) - log(bf_i)
  logbf = log(dble(Nout1)) - eta

  if (Ntot2 .eq. 0) return

  ! Compute weights
  eta = log(dble(Nout2)) - logbf
  do j = 1,kg
    do i = 1,Ntot2
      lglketa(i,j) = eta(j) + lglk2(i,j)
    end do
  end do
  weights = logrsumexp(lglketa,Ntot2,kg)

  ! Compute control variates
  do j = 1, kg
    do i = 1, Ntot2
      lglketa(i,j) = lglketa(i,j) - weights(i) + logNN2(j)
    end do
  end do
  do j = 2, kg
    do i = 1, Ntot2
      zcv(i,j) = lglketa(i,1) - lglketa(i,j)
    end do
  end do
  where (zcv(:,2:kg) > 0d0)
    zcv(:,2:kg) = -exp(lglketa(:,2:kg) + flogexpm1(zcv(:,2:kg)))
  elsewhere (zcv(:,2:kg) < 0d0)
    zcv(:,2:kg) = exp(lglketa(:,2:kg) + flog1mexp(zcv(:,2:kg)))
  end where
  zcv(:,1) = 1d0
end subroutine bfsp_no






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Compute the Bayes factors using the mu sample
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine bfsp_mu (weights, zcv, logbf, lglk1, lglk2, &
   philist, omglist, nulist, &
   musample1, Nout1, Ntot1, musample2, Nout2, Ntot2, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   kappalist, icf, n, p, kg, ifam, imeth, itr)

  use modelfcns
  use interfaces
  use flogsumexp
  use bmargin
  use covfun
  use pdfmu, only: logpdfmu_ga
  use betaprior
  implicit none
  integer, intent(in) :: n, p, kg, ifam, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     musample1(n,Ntot1), musample2(n,Ntot2), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: logbf(kg), lglk1(Ntot1, kg), &
     lglk2(Ntot2, kg), weights(Ntot2), zcv(Ntot2, kg)
  logical lmxi
  double precision T(n, n), TiF(n, p), FTF(p, p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, tsqdfsc, respdfh, xi(n), eta(kg), mxlglk, &
     lglketa(Ntot2, kg), nu, logNN2(kg)
  integer i, j

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  logNN2 = log(dble(Ntot2)) - log(dble(Nout2))

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  select case (ifam)
  case (0)
    do i = 1, kg
      nu = nulist(i)
      call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
         kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot1
        call rchkusr
        lglk1(j,i) = logpdfmu_ga(n, musample1(:, j), Ups, ldh_Ups, &
           nu, xi, lmxi, ssqdfsc, modeldfh)
      end do
      do j = 1, Ntot2
        call rchkusr
        lglk2(j,i) = logpdfmu_ga(n, musample2(:, j), Ups, ldh_Ups, &
           nu, xi, lmxi, ssqdfsc, modeldfh)
      end do
    end do
  case default
    do i = 1, kg
      nu = nulist(i)
      call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
         kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot1
        call rchkusr
        lglk1(j,i) = logpdfmu(n, musample1(:, j), Ups, ldh_Ups, &
           nu, xi, lmxi, ssqdfsc, modeldfh)
      end do
      do j = 1, Ntot2
        call rchkusr
        lglk2(j,i) = logpdfmu(n, musample2(:, j), Ups, ldh_Ups, &
           nu, xi, lmxi, ssqdfsc, modeldfh)
      end do
    end do
  end select
  mxlglk = maxval(lglk1)
  lglk1 = lglk1 - mxlglk
  ! lglk2 = lglk2 - mxlglk XXX Don't do this; may err SE calc different fcn

  ! Use the sample in reverse logistic regression
  eta = log(dble(Nout1))
  select case (imeth)
  case (1)
!     open(11,file='eta.txt'); write(11,*) eta; close(11)
!     open(11,file='lglk1.txt'); write(11,*) lglk1; close(11)
!     open(11,file='kg.txt'); write(11,*) kg; close(11)
!     open(11,file='Ntot1.txt'); write(11,*) Ntot1; close(11)
!     open(11,file='Nout1.txt'); write(11,*) Nout1; close(11)
    call revlogistic (eta,lglk1,kg,Ntot1,Nout1)
  case (2)
    call revlogistic (eta,lglk1,kg,Ntot1,Nout1)
    call mengwong (eta,lglk1,kg,Ntot1,Nout1)
  end select
  ! eta_i = log(N_i) - log(bf_i)
  logbf = log(dble(Nout1)) - eta

  if (Ntot2 .eq. 0) return

  ! Compute weights
  eta = log(dble(Nout2)) - logbf
  do j = 1,kg
    do i = 1,Ntot2
      lglketa(i,j) = eta(j) + lglk2(i,j)
    end do
  end do
  weights = logrsumexp(lglketa,Ntot2,kg)

  ! Compute control variates
  do j = 1, kg
    do i = 1, Ntot2
      lglketa(i,j) = lglketa(i,j) - weights(i) + logNN2(j)
    end do
  end do
  do j = 2, kg
    do i = 1, Ntot2
      zcv(i,j) = lglketa(i,1) - lglketa(i,j)
    end do
  end do
  where (zcv(:,2:kg) > 0d0)
    zcv(:,2:kg) = -exp(lglketa(:,2:kg) + flogexpm1(zcv(:,2:kg)))
  elsewhere (zcv(:,2:kg) < 0d0)
    zcv(:,2:kg) = exp(lglketa(:,2:kg) + flog1mexp(zcv(:,2:kg)))
  end where
  zcv(:,1) = 1d0
end subroutine bfsp_mu



subroutine bfsp_wo (weights, zcv, logbf, lglk1, lglk2, &
   philist, omglist, nulist, &
   sample1, Nout1, Ntot1, sample2, Nout2, Ntot2, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   kappalist, icf, n, p, kg, ifam, imeth, itr)
  use modelfcns, logpdfzf => logpdfz, condymuf => condymu
  use interfaces
  use linkfcns
  use flogsumexp
  use bmargin
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, kg, ifam, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     sample1(n,Ntot1), sample2(n,Ntot2), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: logbf(kg), lglk1(Ntot1,kg), &
     lglk2(Ntot2,kg), weights(Ntot2), zcv(Ntot2,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, ssqdfsc, modeldfh, &
     xi(n), eta(kg), lglketa(Ntot2, kg)
  integer i, j
  double precision zsam(n), msam(n), jsam(n), sam(n)
  double precision nu, phi, omg, kappa, logNN2(kg)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc

  logNN2 = log(dble(Ntot2)) - log(dble(Nout2))

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  do i = 1, kg
    nu = nulist(i)
    phi = philist(i)
    omg = omglist(i)
    kappa = kappalist(i)
    call calc_cov (phi,omg,dm,F,betQ0,&
       kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, Ntot1
      call rchkusr
      sam = sample1(:,j)
      zsam = transfw(sam,nu)
      msam = invlink(zsam,nu)
      jsam = loginvtrwdz(zsam,nu)
      lglk1(j,i) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
         + condymuf(n,y,l,msam,tsq) - sum(jsam)
    end do
    do j = 1, Ntot2
      call rchkusr
      sam = sample2(:,j)
      zsam = transfw(sam,nu)
      msam = invlink(zsam,nu)
      jsam = loginvtrwdz(zsam,nu)
      lglk2(j,i) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
         + condymuf(n,y,l,msam,tsq) - sum(jsam)
    end do
  end do

  ! Use the sample in reverse logistic regression
  eta = log(dble(Nout1))
  select case (imeth)
  case (1)
    call revlogistic (eta,lglk1,kg,Ntot1,Nout1)
  case (2)
    call revlogistic (eta,lglk1,kg,Ntot1,Nout1)
    call mengwong (eta,lglk1,kg,Ntot1,Nout1)
  end select
  ! eta_i = log(N_i) - log(bf_i)
  logbf = log(dble(Nout1)) - eta

  if (Ntot2 .eq. 0) return

  ! Compute weights
  eta = log(dble(Nout2)) - logbf
  do j = 1,kg
    do i = 1,Ntot2
      lglketa(i,j) = eta(j) + lglk2(i,j)
    end do
  end do
  weights = logrsumexp(lglketa,Ntot2,kg)

  ! Compute control variates
  do j = 1, kg
    do i = 1, Ntot2
      lglketa(i,j) = lglketa(i,j) - weights(i) + logNN2(j)
    end do
  end do
  do j = 2, kg
    do i = 1, Ntot2
      zcv(i,j) = lglketa(i,1) - lglketa(i,j)
    end do
  end do
  where (zcv(:,2:kg) > 0d0)
    zcv(:,2:kg) = -exp(lglketa(:,2:kg) + flogexpm1(zcv(:,2:kg)))
  elsewhere (zcv(:,2:kg) < 0d0)
    zcv(:,2:kg) = exp(lglketa(:,2:kg) + flog1mexp(zcv(:,2:kg)))
  end where
  zcv(:,1) = 1d0
end subroutine bfsp_wo


subroutine bfsp_tr (weights, zcv, logbf, lglk1, lglk2, &
   philist, omglist, nulist, &
   sample1, Nout1, Ntot1, sample2, Nout2, Ntot2, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   kappalist, icf, n, p, kg, ifam, imeth, itr)
  use modelfcns, logpdfzf => logpdfz, condymu_mf => condymu
  use interfaces
  use linkfcns
  use flogsumexp
  use bmargin
  use covfun
  use condymu, only: condymu_gt
  use betaprior
  implicit none
  integer, intent(in) :: n, p, kg, ifam, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     sample1(n,Ntot1), sample2(n,Ntot2), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: logbf(kg), lglk1(Ntot1,kg), &
     lglk2(Ntot2,kg), weights(Ntot2), zcv(Ntot2,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, ssqdfsc, modeldfh, &
     tsqval, respdfh, xi(n), eta(kg), lglketa(Ntot2, kg)
  integer i, j
  double precision zsam(n), msam(n), jsam(n), sam(n)
  double precision nu, phi, omg, kappa, logNN2(kg)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
    !!condymuf => condymu_gt
  case default
    tsqval = tsq
    !!condymuf => condymu_sp
  end select

  logNN2 = log(dble(Ntot2)) - log(dble(Nout2))

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  do i = 1, kg
    nu = nulist(i)
    phi = philist(i)
    omg = omglist(i)
    kappa = kappalist(i)
    call calc_cov (phi,omg,dm,F,betQ0,&
       kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, Ntot1
      call rchkusr
      sam = sample1(:,j)
      where (itr == 0)
        zsam = sam
        msam = invlink(zsam,nu)
        jsam = 0d0
      elsewhere (itr == 1)
        msam = sam
        zsam = flink(msam,nu)
        jsam = loginvlinkdz(zsam,nu)
      elsewhere (itr == 2)
        zsam = transfw(sam,nu)
        msam = invlink(zsam,nu)
        jsam = loginvtrwdz(zsam,nu)
      end where
      lglk1(j,i) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
         + condymuf(ifam, n,y,l,msam,tsqval,respdfh) - sum(jsam)
    end do
    do j = 1, Ntot2
      call rchkusr
      sam = sample2(:,j)
      where (itr == 0)
        zsam = sam
        msam = invlink(zsam,nu)
        jsam = 0d0
      elsewhere (itr == 1)
        msam = sam
        zsam = flink(msam,nu)
        jsam = loginvlinkdz(zsam,nu)
      elsewhere (itr == 2)
        zsam = transfw(sam,nu)
        msam = invlink(zsam,nu)
        jsam = loginvtrwdz(zsam,nu)
      end where
      lglk2(j,i) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
         + condymuf(ifam, n,y,l,msam,tsqval,respdfh) - sum(jsam)
    end do
  end do

  ! Use the sample in reverse logistic regression
  eta = log(dble(Nout1))
  select case (imeth)
  case (1)
    call revlogistic (eta,lglk1,kg,Ntot1,Nout1)
  case (2)
    call revlogistic (eta,lglk1,kg,Ntot1,Nout1)
    call mengwong (eta,lglk1,kg,Ntot1,Nout1)
  end select
  ! eta_i = log(N_i) - log(bf_i)
  logbf = log(dble(Nout1)) - eta

  if (Ntot2 .eq. 0) return

  ! Compute weights
  eta = log(dble(Nout2)) - logbf
  do j = 1,kg
    do i = 1,Ntot2
      lglketa(i,j) = eta(j) + lglk2(i,j)
    end do
  end do
  weights = logrsumexp(lglketa,Ntot2,kg)

  ! Compute control variates
  do j = 1, kg
    do i = 1, Ntot2
      lglketa(i,j) = lglketa(i,j) - weights(i) + logNN2(j)
    end do
  end do
  do j = 2, kg
    do i = 1, Ntot2
      zcv(i,j) = lglketa(i,1) - lglketa(i,j)
    end do
  end do
  where (zcv(:,2:kg) > 0d0)
    zcv(:,2:kg) = -exp(lglketa(:,2:kg) + flogexpm1(zcv(:,2:kg)))
  elsewhere (zcv(:,2:kg) < 0d0)
    zcv(:,2:kg) = exp(lglketa(:,2:kg) + flog1mexp(zcv(:,2:kg)))
  end where
  zcv(:,1) = 1d0

contains

  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: ifam
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf
end subroutine bfsp_tr

######################################################################
##
### Commentary: Some auxiliary functions.
##
######################################################################

getbetaprior <- function (betm0, betQ0, p) {
  if (length(betQ0) == 1 && betQ0[1] == 0) {
    ## Uniform prior
    betQ0 <- matrix(0, p, p)
    betm0 <- rep(0, p)
  } else if (length(betQ0) == 1 || length(betQ0) == p) {
    if (any(betQ0 <= 0)) stop ('betQ0 not > 0')
    betQ0 <- diag(as.vector(betQ0), p, p)
    betm0 <- rep(as.double(betm0), length.out = p)
  } else if (length(betQ0) == p*p) {
    betQ0 <- matrix(as.double(betQ0), p, p)
    betQ0[lower.tri(betQ0)] <- 0
    betQ0eig <- eigen(t(betQ0), 1, 1)$values
    if (any (betQ0eig < sqrt(.Machine$double.eps))) {
      stop ('betQ0 not > 0 within tolerance')
    }
    betm0 <- rep(as.double(betm0), length.out = p)
  } else stop ('Bad betQ0')
  if (!all(is.finite(betm0))) {
    stop ('Non-finite betm0')
  }
  if (!all(is.finite(betQ0[upper.tri(betQ0, diag = TRUE)]))) {
    stop ('Non-finite betQ0')
  }
  list(betm0 = as.double(betm0), betQ0 = as.double(betQ0))
}

## Checks if the paroptim argument is valid and returns a list with
## components pstart, lower, upper, estim.
getparoptim <- function (paroptim, ifam, icf) {
  paroptim <- as.list(paroptim)
  parnmall <- c("linkp", "phi", "omg", "kappa")
  if (!is.numeric(ifam)) ifam <- .geoBayes_family(ifam)
  if (!is.numeric(icf)) icf <- .geoBayes_correlation(icf)
  paroptim <- paroptim[parnmall]
  lower <- upper <- pstart <- rep.int(0, 4) # nu, phi, omg, kappa
  estim <- rep.int(FALSE, 4)

  ##
  needlinkp <- .geoBayes_models$needlinkp[ifam]
  needkappa <- .geoBayes_corrfcn$needkappa[icf]
  need <- c(needlinkp, TRUE, TRUE, needkappa)

  getnonneg <- function(x) {
    if (any(x < 0)) stop ("All values must be non-negative.")
    as.double(x)
  }
  getf <- list(linkp = function(x) .geoBayes_getlinkp(x, ifam),
               phi = getnonneg, omg = getnonneg,
               kappa = function(x) .geoBayes_getkappa(x, icf))

  ## Check parameters
  for (i in 1:4) {
    if (!need[i]) next        # Skip those not needed
    ppp <- paroptim[[parnmall[i]]]
    if (!is.numeric(ppp)) {
      stop(paste("The element", parnmall[i], "in paroptim must be numeric"))
    }
    ppp <- getf[[i]](ppp)
    lppp <- length(ppp)
    if (lppp < 1 || lppp > 3) {
      stop (paste("The element", parnmall[i],
                  "in paroptim must have 1, 2, or 3 components"))
    }
    if (lppp == 1) {
      pstart[i] <- ppp
      estim[i] <- FALSE
    } else if (lppp == 2) {
      pstart[i] <- .5*(ppp[1]+ppp[2])
      estim[i] <- TRUE
      lower[i] <- ppp[1]
      upper[i] <- ppp[2]
      if (lower[i] >= upper[i]) {
        stop (paste("The lower bound must be less than the upper bound for",
                     parnmall[i], "in paroptim"))
      }
    } else {
      pstart[i] <- ppp[2]
      estim[i] <- TRUE
      lower[i] <- ppp[1]
      upper[i] <- ppp[3]
      if (lower[i] > pstart[i] || pstart[i] > upper[i] || lower[i] == upper[i]) {
        stop (paste("The elements in the component", parnmall[i],
                     "in paroptim must be ordered"))
      }
    }
  }
  list(pstart = pstart, lower = lower, upper = upper, estim = estim)
}


## Take the sample sizes in N and split them according to fun
## using sz as a parameter.
## allow0 will allow sizes to be 0.
getsize <- function(sz, N, FUN = "*", allow0 = FALSE)
{
  nm <- deparse(substitute(sz))
  FUN <- match.fun(FUN)
  sz <- as.double(sz)
  nruns <- length(N)
  if (length(sz) > nruns) {
    warning(paste("The number of elements in", nm, "exceeds the
number of runs; the extra elements will be discarded."))
  }
  sz <- rep(sz, length.out = nruns)
  if (allow0) {
    if (any(sz < 0)) {
      stop(paste("Argument", nm, "must be non-negative"))
    }
  } else {
    if (any(sz <= 0)) {
      stop(paste("Argument", nm, "must be positive"))
    }
  }
  if (all(sz < 1)) {
    Nout <- as.integer(FUN(N, sz))
    if (!allow0 && any(Nout == 0)) {
      stop(paste("Calculated 0 sizes; give a larger", nm))
    }
  } else if (all(sz == 1)) {
    Nout <- as.integer(FUN(N, sz))
  } else if (all(sz >= 1)) {
    Nout <- as.integer(sz)
    if (any(Nout > N)) {
      stop(paste(nm, "is too big."))
    }
  } else {
    stop (paste("Argument", nm, "must be all < 1 or all >= 1."))
  }
  Nout
}

## Compute transformed sample.
##
## Computes the transformed sample according to transf.
## @title Compute transformed sample
## @param runs A list of outputs from \code{mcsglmm} or \code{mctrga}.
## @param model A list with elements response and family.
## @param transf The type of transformation to use.
## @return A list with elements sample, transf, itr, ifam.
## @useDynLib geoBayes transformz
transfsample <- function (runs, model, transf = c("no", "mu", "wo"), verb = TRUE)
{
  y <- model$response
  n <- as.integer(length(y))
  family <- model$family
  ifam <- .geoBayes_family(family)
  if (!is.logical(transf)) {
    transf <- transf0 <- match.arg(transf)
  } else if (transf) {
    transf <- transf0 <- "mu"
  } else {
    transf <- transf0 <- "no"
  }
  if (transf == "wo" && !.geoBayes_models$haswo[ifam]) {
    if (verb)
      warning("Can't use workaround for this family. Computing no transformation.")
    transf0 <- "no"
  }
  if (transf == "wo" && family == "poisson.boxcox") {
    ypo <- y > 0
    if (all(ypo)) {
      if (verb)
        message("For poisson.boxcox with all positive responses, the workaround is the same as the mean transformation.")
      transf0 <- "mu"
    } else if (any(ypo)) {
      transf0 <- "tr"
    }
  }
  if (transf0 == "mu") {
    itr <- rep.int(1L, n)
    ftrw <- function (r) {
      o <- r$MCMC[["whichobs"]]
      out <- r$MCMC[["mu"]][o, , drop = FALSE]
      if (is.null(out) && any(o)) {
        out <- r$MCMC[["z"]][o, , drop = FALSE]
        nu <- r$FIXED[["linkp_num"]][1]
        nn <- as.integer(length(out))
        out[] <- .Fortran("transformz", out, nu, nn, ifam,
                          PACKAGE = "geoBayes")[[1]]
      }
      out
    }
    sample <- lapply(runs, ftrw)
  } else if (transf0 == "tr") {
    ifam <- -ifam
    if (family == "poisson.boxcox") { # Poisson y = 0
      itr <- 2L - ypo
      ftrw <- function (r) {
        i2 <- itr == 2
        o <- r$MCMC[["whichobs"]]
        out <- r$MCMC[["mu"]][o, , drop = FALSE]
        z <- r$MCMC[["z"]][o, ][i2, ]
        nu <- r$FIXED[["linkp_num"]][1]
        nn <- as.integer(length(z))
        out[i2, ] <- .Fortran("transformz", z, nu, nn, ifam,
                              PACKAGE = "geoBayes")[[1]]
        out
      }
    } else { # All other cases
      stop ("BUG: Only the poisson.boxcox family is treated differently.")
    }
    sample <- lapply(runs, ftrw)
  } else if (transf0 == "wo") {
    ifam <- -ifam
    itr <- rep.int(2L, n)
    ftrw <- function (r) {
      o <- r$MCMC[["whichobs"]]
      out <- r$MCMC[["z"]][o, , drop = FALSE]
      nu <- r$FIXED[["linkp_num"]][1]
      nn <- as.integer(length(out))
      out[] <- .Fortran("transformz", out, nu, nn, ifam,
                        PACKAGE = "geoBayes")[[1]]
      out
    }
    sample <- lapply(runs, ftrw)
  } else {
    ftrw <- function (r) {
      o <- r$MCMC[["whichobs"]]
      out <- r$MCMC[["z"]][o, , drop = FALSE]
      out
    }
    sample <- lapply(runs, ftrw)
    itr <- rep.int(0L, n)
  }
  list(sample = sample, transf = transf,
       real_transf = transf0, itr = as.integer(itr),
       ifam = ifam)
}



## Check and return the parameter grid
check_pargrid <- function(pargrid, family, corrfcn)
{
  pargrid <- as.data.frame(pargrid, stringsAsFactors = FALSE)
  ##if(!is.list(pargrid)) stop ("Argument pargrid must be a list")
  parnm <- c("linkp", "phi", "omg", "kappa")
  ifam <- .geoBayes_family(family)
  needlinkp <- .geoBayes_models$needlinkp[ifam]
  if (!needlinkp) pargrid$linkp <- 0
  icf <- .geoBayes_correlation(corrfcn)
  needkappa <- .geoBayes_corrfcn$needkappa[icf]
  if (!needkappa) {
    pargrid$kappa <- 0
  } else {
    pargrid$kappa <- .geoBayes_getkappa(pargrid$kappa, icf)
  }
  if (!all(parnm %in% names(pargrid))) {
    stop (paste("Argument", deparse(substitute(pargrid)), "must have the names",
                paste(c(if (needlinkp) "linkp", "phi", "omg",
                        if (needkappa) "kappa"), collapse=", ")))
  } else {
    pargrid <- pargrid[parnm]
  }
  nruns <- nrow(pargrid)
  phi <- pargrid[["phi"]]
  omg <- pargrid[["omg"]]
  kappa <- pargrid[["kappa"]]
  linkp <- pargrid[["linkp"]]
  if (any (phi < 0)) stop ("Element phi given must be non-negative")
  if (any (omg < 0)) stop ("Element omg given must be non-negative")

  ## Check if linkp conforms with family
  nu <- .geoBayes_getlinkp(linkp, family)

  ## Output
  data.frame(linkp = linkp, phi = phi, omg = omg, kappa = kappa, nu = nu)
}

check_gengamma_prior <- function (pars)
  ## Checks if the generalised gamma prior parameters are correct.
{
  pars <- as.double(pars)
  if (4 != length(pars))
    stop ("Generalised gamma parameters must be a vector of length 4.")
  if (!all(is.finite(pars)))
    stop ("Non-finite generalised gamma parameters.")
  if (pars[1] <= 0 || pars[4] < 0 || pars[2]*pars[3] < 0) {
    stop ("Invalid values in pars: pars[1] <= 0 || pars[4] < 0 || pars[2]*pars[3] < 0")
  }
  pars
}

check_unif_prior <- function (pars)
  ## Checks if the uniform prior parameters are correct.
{
  pars <- as.double(pars)
  if (2 != length(pars))
    stop ("Uniform parameters must be a vector of length 2.")
  if (!all(is.finite(pars)))
    stop ("Non-finite uniform prior parameters.")
  if (pars[1] >= pars[2]) {
    stop ("Invalid values in pars: pars[1] >= pars[2]")
  }
  pars
}

subroutine llikfcn_dcov_tr (dlglk, phi, omg, nu, kappa, &
   sample, Ntot, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, ifam, itr)
!! Log-likelihood function derivative w.r.t covariance parameters.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: phi, omg, nu, &
     sample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappa
  double precision, intent(out) :: dlglk(3,Ntot)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, tsqval, respdfh, xi(n)
  integer i, j
  double precision zsam(n), zU(n), zUz, zUDTUz, DT(n,n,3), trUpsDT(3)

  call create_model (ifam)
  call create_spcor (icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)

  do i = 1, 3
    DT(:,:,i) = cor_dcov(n, dm, phi, omg, kappa, i)
    trUpsDT(i) = traceAB(Ups,DT(:,:,i),n)
  end do

  do j = 1, Ntot
    call rchkusr
    where (itr .eq. 0)
      zsam = sample(:,j)
    elsewhere (itr .eq. 1)
      zsam = flink(sample(:,j),nu)
    elsewhere (itr .eq. 2)
      zsam = transfw(sample(:,j),nu)
    end where
    if (lmxi) zsam = zsam - xi
    call dsymv ('u',n,1d0,Ups,n,zsam,1,0d0,zU,1)
    zUz = dot_product(zsam,zU) + ssqdfsc
    do i = 1, 3
      zUDTUz = qform(zU,DT(:,:,i),n)
      dlglk(i,j) = -.5d0*trUpsDT(i) + modeldfh*zUDTUz/zUz
    end do
  end do

contains

  double precision function qform(v,A,n)
    ! Compute v'*A*v where A is symmetric with only the upper triangle
    ! specified.
    integer, intent(in) :: n
    double precision, intent(in) :: v(n), A(n,n)
    double precision Av(n)
    call dsymv ('u',n,1d0,A,n,v,1,0d0,Av,1)
    qform = dot_product(Av,v)
  end function qform

  double precision function traceAB(A,B,n)
    ! Compute the trace(matmul(A,B)) where A,B are symmetric matrices with
    ! only their upper triangular elements specified.
    integer, intent(in) :: n
    double precision, intent(in) :: A(n,n), B(n,n)
    integer j
    double precision c
    traceAB = A(1,1)*B(1,1)
    do j = 2, n
      c = dot_product(A(:j-1,j),B(:j-1,j))
      traceAB = traceAB + c + c + A(j,j)*B(j,j)
    end do
  end function traceAB

  double precision function cor_dcov (n, dm, phi, omg, kappa, id) result (DT)
    ! id = which derivative do you want? phi, omg, kappa
    ! Return dT where T = omg*I + R(phi,kappa)
    use covfun, only: spcor_dh, spcor_dk, fill_symmetric_matrix, upper_tri
    implicit none
    integer, intent(in) :: n, id
    double precision, intent(in) :: dm(n,n), phi, omg, kappa
    !double precision, intent(out) :: DT(n,n)
    dimension :: DT(n,n)
    integer i
    double precision H(n,n)
    select case (id)
    case (1)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = -(H/phi)*spcor_dh(H,kappa)
        end where
      end if
    case (2)
      DT = 0d0
      do i = 1, n
        DT(i,i) = 1d0
      end do
    case (3)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = spcor_dk(H,kappa)
        end where
      end if
    end select
  end function cor_dcov
end subroutine llikfcn_dcov_tr

subroutine llikfcn_hcov_tr (hlglk, phi, omg, nu, kappa, &
   sample, Ntot, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, ifam, itr)
!! Log-likelihood function derivative w.r.t covariance parameters.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: phi, omg, nu, &
     sample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappa
  double precision, intent(out) :: hlglk(3,3,Ntot)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, tsqval, respdfh, xi(n)
  integer i, j, k
  double precision zsam(n), zU(n), zUz, zUDTUz(3), zUHTUz, zUDTUDTUz
  double precision DT(n,n,3), HT(n,n,3,3), trUpsHT(3,3), trUpsDTUpsDT(3,3)
  double precision UpsDT(n,n), DTUpsDT(n,n)

  call create_model (ifam)
  call create_spcor (icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  call fill_symmetric_matrix(Ups,n)

  do i = 1, 3
    DT(:,:,i) = cor_dcov(n, dm, phi, omg, kappa, i)
  end do

  do j = 1, 3
    do i = 1, j
      HT(:,:,i,j) = cor_hcov(n, dm, phi, omg, kappa, i, j)
      trUpsHT(i,j) = traceAB(Ups,HT(:,:,i,j),n)
      trUpsDTUpsDT(i,j) = traceABAC(Ups,DT(:,:,i),DT(:,:,j),n)
    end do
  end do

  do k = 1, Ntot
    call rchkusr
    where (itr .eq. 0)
      zsam = sample(:,k)
    elsewhere (itr .eq. 1)
      zsam = flink(sample(:,k),nu)
    elsewhere (itr .eq. 2)
      zsam = transfw(sample(:,k),nu)
    end where
    if (lmxi) zsam = zsam - xi
    call dsymv ('u',n,1d0,Ups,n,zsam,1,0d0,zU,1)
    zUz = dot_product(zsam,zU) + ssqdfsc
    do j = 1, 3
      zUDTUz(j) = qform(zU,DT(:,:,j),n)/zUz
    end do
    do j = 1, 3
      call dsymm ('r','u',n,n,1d0,DT(:,:,j),n,Ups,n,0d0,UpsDT,n)
      do i = 1, j
        zUHTUz = qform(zU,HT(:,:,i,j),n)
        call dsymm ('l','u',n,n,1d0,DT(:,:,i),n,UpsDT,n,0d0,DTUpsDT,n)
        DTUpsDT = .5d0*(DTUpsDT + transpose(DTUpsDT))
        zUDTUDTUz = qform(zU,DTUpsDT,n)
        hlglk(i,j,k) = -.5d0*trUpsHT(i,j) + .5d0*trUpsDTUpsDT(i,j) &
           + modeldfh*zUDTUz(i)*zUDTUz(j) &
           - modeldfh*(zUDTUDTUz + zUDTUDTUz - zUHTUz)/zUz
      end do
    end do
  end do

contains

  double precision function qform(v,A,n)
    ! Compute v'*A*v where A is symmetric with only the upper triangle
    ! specified.
    integer, intent(in) :: n
    double precision, intent(in) :: v(n), A(n,n)
    double precision Av(n)
    call dsymv ('u',n,1d0,A,n,v,1,0d0,Av,1)
    qform = dot_product(Av,v)
  end function qform

  double precision function traceAB(A,B,n)
    ! Compute the trace(matmul(A,B)) where A,B are symmetric matrices with
    ! only their upper triangular elements specified.
    integer, intent(in) :: n
    double precision, intent(in) :: A(n,n), B(n,n)
    integer j
    double precision c
    traceAB = A(1,1)*B(1,1)
    do j = 2, n
      c = dot_product(A(:j-1,j),B(:j-1,j))
      traceAB = traceAB + c + c + A(j,j)*B(j,j)
    end do
  end function traceAB

  double precision function traceABAC(A,B,C,n)
    ! Compute the trace(matmul(A,B,A,C)) where A is some matrix and B,C are
    ! symmetric matrices with only their upper triangular elements
    ! specified.
    integer, intent(in) :: n
    double precision, intent(in) :: A(n,n), B(n,n), C(n,n)
    !integer i,j
    double precision AB(n,n), AC(n,n)
    call dsymm('r','u',n,n,1d0,B,n,A,n,0d0,AB,n)
    call dsymm('r','u',n,n,1d0,C,n,A,n,0d0,AC,n)
    traceABAC = sum(AB*transpose(AC))
  end function traceABAC

  double precision function cor_dcov (n, dm, phi, omg, kappa, id) result (DT)
    ! id = which derivative do you want? phi, omg, kappa
    ! Return dT where T = omg*I + R(phi,kappa)
    use covfun, only: spcor_dh, spcor_dk, fill_symmetric_matrix, upper_tri
    implicit none
    integer, intent(in) :: n, id
    double precision, intent(in) :: dm(n,n), phi, omg, kappa
    !double precision, intent(out) :: DT(n,n)
    dimension :: DT(n,n)
    integer i
    double precision H(n,n)
    select case (id)
    case (1)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = -(H/phi)*spcor_dh(H,kappa)
        end where
      end if
    case (2)
      DT = 0d0
      do i = 1, n
        DT(i,i) = 1d0
      end do
    case (3)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = spcor_dk(H,kappa)
        end where
      end if
    end select
  end function cor_dcov

  double precision function cor_hcov (n, dm, phi, omg, kappa, id, jd) result (HT)
    ! id,jd = which derivative do you want? phi, omg, kappa
    ! Return d^2 T where T = omg*I + R(phi,kappa)
    use covfun, only: spcor_dh, spcor_dk, spcor_hh, spcor_hk, spcor_dhdk, &
       fill_symmetric_matrix, upper_tri
    implicit none
    integer, intent(in) :: n, id, jd
    double precision, intent(in) :: dm(n,n), phi, omg, kappa
    dimension :: HT(n,n)
    integer ij
    double precision H(n,n), H1(n,n)
    ij = id + 3*(jd-1)
    select case (ij)
    case (1) ! (phi,phi)
      if (phi .eq. 0d0) then
        HT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          H1 = -H/phi
          HT = H1*H1*spcor_hh(H,kappa) - (2/phi)*H1*spcor_dh(H,kappa)
        end where
      end if
    case (2,4) ! (phi,omg)
      HT = 0d0
    case (3,7) ! (phi,kappa)
      H = dm/phi
      H1 = -H/phi
      HT = H1*spcor_dhdk(H,kappa)
    case (5) ! (omg,omg)
      HT = 0d0
    case (6,8) ! (omg,kappa)
      HT = 0d0
    case (9) ! (kappa,kappa)
      if (phi .eq. 0d0) then
        HT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          HT = spcor_hk(H,kappa)
        end where
      end if
    end select
  end function cor_hcov
end subroutine llikfcn_hcov_tr



subroutine llikfcn_dlnk_tr (dlglk, phi, omg, nu, kappa, &
   sample, Ntot, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, ifam, itr)
!! Log-likelihood function derivative w.r.t link parameter.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns, logpdfydmu => logpdfydlnk
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: phi, omg, nu, &
     sample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappa
  double precision, intent(out) :: dlglk(Ntot)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, tsqval, respdfh, xi(n)
  integer j
  double precision zsam(n), msam(n), jsam(n), sam(n)
  double precision dlogpy(n), dmudnu(n), dy, dj, dz, zU(n), dzdnu(n)

  call create_model (ifam)
  call create_spcor (icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)
  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)

  do j = 1, Ntot
    call rchkusr
    sam = sample(:,j)
    where (itr == 0)
      zsam = sam
      msam = invlink(zsam,nu)
      dlogpy = logpdfydmu(y,l,msam)
      dmudnu = invlinkdn(zsam,nu)
      dzdnu = 0d0
      jsam = 0d0
    elsewhere (itr == 1)
      msam = sam
      zsam = flink(msam,nu)
      dlogpy = 0d0
      dmudnu = 0d0
      dzdnu = -invlinkdn(zsam,nu)/invlinkdz(zsam,nu)
      jsam = (invlinkdzdn(zsam,nu) + invlinkhz(zsam,nu)*dzdnu) &
         /invlinkdz(zsam,nu)
      !dzdnu = invlinkdn(zsam,nu)
    elsewhere (itr == 2)
      zsam = transfw(sam,nu)
      msam = invlink(zsam,nu)
      dzdnu = -invtrwdn(zsam,nu)/invtrwdz(zsam,nu)
      dlogpy = logpdfydmu(y,l,msam)
      dmudnu = invlinkdn(zsam,nu) + invlinkdz(zsam,nu)*dzdnu
      jsam = (invtrwdzdn(zsam,nu) + invtrwhz(zsam,nu)*dzdnu)/invtrwdz(zsam,nu)
      !dzdnu = invtrwdn(zsam,nu)
    end where
    dy = dot_product(dlogpy,dmudnu)
    zU = logpdfz_dz(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh)
    dz = dot_product(zU,dzdnu)
    dj = -sum(jsam)
    dlglk(j) = dy + dz + dj
  end do
end subroutine llikfcn_dlnk_tr



subroutine llikfcn_hlnk_tr (hlglk, phi, omg, nu, kappa, &
   sample, Ntot, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, ifam, itr)
!! Log-likelihood function.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns, logpdfydmu => logpdfydlnk, logpdfyhmu => logpdfyhlnk
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: phi, omg, nu, &
     sample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappa
  double precision, intent(out) :: hlglk(Ntot)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, tsqval, respdfh, xi(n)
  integer j
  double precision zsam(n), msam(n), djsam(n), sam(n), hjsam(n)
  double precision dlogpy(n), dmudnu(n), dy, dj, dz, dlogpz(n), dzdnu(n)
  double precision hlogpy(n), hmuhnu(n), hy, hj, hz, hzhnu(n), hlogpz(n,n)

  call create_model (ifam)
  call create_spcor (icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  do j = 1, Ntot
    call rchkusr
    sam = sample(:,j)
    where (itr == 0)
      zsam = sam
      msam = invlink(zsam,nu)
      dlogpy = logpdfydmu(y,l,msam)
      hlogpy = logpdfyhmu(y,l,msam)
      dmudnu = invlinkdn(zsam,nu)
      hmuhnu = invlinkhn(zsam,nu)
      djsam = 0d0
      dzdnu = 0d0
      hzhnu = 0d0
      hjsam = 0d0
    elsewhere (itr == 1)
      msam = sam
      zsam = flink(msam,nu)
      dlogpy = 0d0
      hlogpy = 0d0
      dzdnu = -invlinkdn(zsam,nu)/invlinkdz(zsam,nu)
      djsam = (invlinkdzdn(zsam,nu) + invlinkhz(zsam,nu)*dzdnu) &
         /invlinkdz(zsam,nu)
      hzhnu = -(invlinkhn(zsam,nu) + 2*invlinkdzdn(zsam,nu)*dzdnu &
         + invlinkhz(zsam,nu)*dzdnu*dzdnu)/invlinkdz(zsam,nu)
      hjsam = (invlink3z(zsam,nu)*dzdnu*dzdnu + invlinkhz(zsam,nu)*hzhnu &
         + 2*invlinkhzdn(zsam,nu)*dzdnu + invlinkdzhn(zsam,nu))&
         /invlinkdz(zsam,nu) - djsam*djsam
      dmudnu = 0d0
      hmuhnu = 0d0
    elsewhere (itr == 2)
      zsam = transfw(sam,nu)
      msam = invlink(zsam,nu)
      dlogpy = logpdfydmu(y,l,msam)
      hlogpy = logpdfyhmu(y,l,msam)
      dzdnu = -invtrwdn(zsam,nu)/invtrwdz(zsam,nu)
      hzhnu = -(invtrwhn(zsam,nu) + 2*invtrwdzdn(zsam,nu)*dzdnu &
         + invtrwhz(zsam,nu)*dzdnu*dzdnu)/invtrwdz(zsam,nu)
      dmudnu = invlinkdn(zsam,nu) + invlinkdz(zsam,nu)*dzdnu
      hmuhnu = invlinkhn(zsam,nu) + invlinkdz(zsam,nu)*hzhnu &
         + 2*invlinkdzdn(zsam,nu)*dzdnu + invlinkhz(zsam,nu)*dzdnu*dzdnu
      djsam = (invtrwdzdn(zsam,nu) + invtrwhz(zsam,nu)*dzdnu)/invtrwdz(zsam,nu)
      hjsam = (invtrw3z(zsam,nu)*dzdnu*dzdnu + invtrwhz(zsam,nu)*hzhnu &
         + 2*invtrwhzdn(zsam,nu)*dzdnu + invtrwdzhn(zsam,nu))&
         /invtrwdz(zsam,nu) - djsam*djsam
    end where
    dy = dot_product(dlogpy,dmudnu)
    dlogpz = logpdfz_dz(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh)
    hlogpz = logpdfz_hz(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh)
    dz = dot_product(dlogpz,dzdnu)
    dj = -sum(djsam)
    hy = dot_product(hlogpy,dmudnu*dmudnu) + dot_product(dlogpy,hmuhnu)
    hz = qform(dzdnu,hlogpz,n) + dot_product(dlogpz,hzhnu)
    hj = -sum(hjsam)
    hlglk(j) = hy + hz + hj
  end do

contains

  double precision function qform(v,A,n)
    ! Compute v'*A*v where A is symmetric with only the upper triangle
    ! specified.
    integer, intent(in) :: n
    double precision, intent(in) :: v(n), A(n,n)
    double precision Av(n)
    call dsymv ('u',n,1d0,A,n,v,1,0d0,Av,1)
    qform = dot_product(Av,v)
  end function qform
end subroutine llikfcn_hlnk_tr


subroutine llikfcn_dlnkdcov_tr (dlglk, phi, omg, nu, kappa, &
   sample, Ntot, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, ifam, itr)
!! Log-likelihood function derivative w.r.t link and covariance parameters.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: phi, omg, nu, &
     sample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappa
  double precision, intent(out) :: dlglk(3,Ntot)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, tsqval, respdfh, xi(n)
  integer i, j
  double precision zsam(n), zU(n), zUz, zUDTUz, DT(n,n,3), Dz(n), DzU(n), &
     DzUDTUz, DzUz, DTzU(n)

  call create_model (ifam)
  call create_spcor (icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)

  do i = 1, 3
    DT(:,:,i) = cor_dcov(n, dm, phi, omg, kappa, i)
  end do

  do j = 1, Ntot
    call rchkusr
    where (itr .eq. 0)
      zsam = sample(:,j)
      Dz = 0d0
    elsewhere (itr .eq. 1)
      zsam = flink(sample(:,j),nu)
      Dz = -invlinkdn(zsam,nu)/invlinkdz(zsam,nu)
    elsewhere (itr .eq. 2)
      zsam = transfw(sample(:,j),nu)
      Dz = -invtrwdn(zsam,nu)/invtrwdz(zsam,nu)
    end where
    if (lmxi) zsam = zsam - xi
    call dsymv ('u',n,1d0,Ups,n,zsam,1,0d0,zU,1)
    zUz = dot_product(zsam,zU) + ssqdfsc
    call dsymv ('u',n,1d0,Ups,n,Dz,1,0d0,DzU,1)
    DzUz = dot_product(Dz,zU)
    do i = 1, 3
      call dsymv ('u',n,1d0,DT(:,:,i),n,zU,1,0d0,DTzU,1)
      zUDTUz = dot_product(zU,DTzU)
      DzUDTUz = dot_product(DzU,DTzU)
      dlglk(i,j) = (2*modeldfh)*(DzUDTUz/zUz - zUDTUz*DzUz/(zUz*zUz))
    end do
  end do

contains

  double precision function qform(v,A,n)
    ! Compute v'*A*v where A is symmetric with only the upper triangle
    ! specified.
    integer, intent(in) :: n
    double precision, intent(in) :: v(n), A(n,n)
    double precision Av(n)
    call dsymv ('u',n,1d0,A,n,v,1,0d0,Av,1)
    qform = dot_product(Av,v)
  end function qform

  double precision function cor_dcov (n, dm, phi, omg, kappa, id) result (DT)
    ! id = which derivative do you want? phi, omg, kappa
    ! Return dT where T = omg*I + R(phi,kappa)
    use covfun, only: spcor_dh, spcor_dk, fill_symmetric_matrix, upper_tri
    implicit none
    integer, intent(in) :: n, id
    double precision, intent(in) :: dm(n,n), phi, omg, kappa
    !double precision, intent(out) :: DT(n,n)
    dimension :: DT(n,n)
    integer i
    double precision H(n,n)
    select case (id)
    case (1)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = -(H/phi)*spcor_dh(H,kappa)
        end where
      end if
    case (2)
      DT = 0d0
      do i = 1, n
        DT(i,i) = 1d0
      end do
    case (3)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = spcor_dk(H,kappa)
        end where
      end if
    end select
  end function cor_dcov
end subroutine llikfcn_dlnkdcov_tr

!!! flogsumexp.f90 --- 
!! 
!! Author: Evangelos Evangelou
!! Created: Thu, 19 Jun, 2014 13:34 (BST)
!! Last-Updated: Tue, 13 Jan, 2015 15:19 (GMT)
!!     Update #: 2
!! 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! 
!!! Commentary: 
!! 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


module flogsumexp
contains
  pure function logsumexpv (x,n)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: x(n)
    double precision logsumexpv
    double precision xmm(n), se
    logsumexpv = maxval(x)
    xmm = x - logsumexpv
    xmm = exp(xmm)
    se = sum(xmm)
    logsumexpv = logsumexpv + log(se)
  end function logsumexpv

  pure function logrsumexp (x,m,n)
    implicit none
    integer, intent(in) :: m, n
    double precision, intent(in) :: x(m,n)
    double precision logrsumexp(m)
    double precision xmm(m,n), se(m)
    integer i, j
    logrsumexp = maxval(x,2)
    do j = 1, n
      do i = 1, m
        xmm(i,j) = x(i,j) - logrsumexp(i)
      end do
    end do
    xmm = exp(xmm)
    se = sum(xmm,2)
    logrsumexp = logrsumexp + log(se)
  end function logrsumexp

  pure function logcsumexp (x,m,n)
    implicit none
    integer, intent(in) :: m, n
    double precision, intent(in) :: x(m,n)
    double precision logcsumexp(n)
    double precision xmm(m,n), se(n)
    integer i, j
    logcsumexp = maxval(x,1)
    do j = 1, n
      do i = 1, m
        xmm(i,j) = x(i,j) - logcsumexp(j)
      end do
    end do
    xmm = exp(xmm)
    se = sum(xmm,1)
    logcsumexp = logcsumexp + log(se)
  end function logcsumexp
end module flogsumexp

subroutine llikfcnmc_11 (lglk, nulist, kappalist, &
   sample, phisample, omgsample, phipars, omgpars, Ntot, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icflist, n, p, kg, ifamlist, itr)
!! Log-likelihood function.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns, logpdfzf => logpdfz, condymu_mf => condymu
  use covfun
  use betaprior
  use condymu, only: condymu_gt
  implicit none
  integer, intent(in) :: n, p, kg, ifamlist(kg), Ntot, icflist(kg), itr(n)
  double precision, intent(in) :: nulist(kg), phipars(4,kg), omgpars(4,kg), &
     sample(n,Ntot,kg), phisample(Ntot), omgsample(Ntot), &
     y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, nu, phi, omg, kappa, &
     tsqval, respdfh, xi(n)
  integer i, j, ifam, icf
  double precision zsam(n), msam(n), jsam(n), sam(n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifamlist(1))
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  do i = 1, kg
    ifam = ifamlist(i)
    icf = icflist(i)
    call create_model (ifam)
    call create_spcor (icf,n)
    nu = nulist(i)
    kappa = kappalist(i)
    do j = 1, Ntot
      call rchkusr
      phi = phisample(j)
      omg = omgsample(j)
      call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
      sam = sample(:,j,i)
      where (itr == 0)
        zsam = sam
        msam = invlink(zsam,nu)
        jsam = 0d0
      elsewhere (itr == 1)
        msam = sam
        zsam = flink(msam,nu)
        jsam = loginvlinkdz(zsam,nu)
      elsewhere (itr == 2)
        zsam = transfw(sam,nu)
        msam = invlink(zsam,nu)
        jsam = loginvtrwdz(zsam,nu)
      end where
      lglk(j,i) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
         + condymuf(ifam,n,y,l,msam,tsqval,respdfh) - sum(jsam) &
         + lgengam(phi,phipars(:,i)) + lgengam(omg,omgpars(:,i))
    end do
  end do

contains
  pure function lgengam (x, q)
    double precision, intent(in) :: x, q(4)
    double precision lgengam
    lgengam = (x - q(4))/q(1)
    lgengam = log(lgengam)
    lgengam = (q(2)-1)*lgengam - exp(lgengam*q(3))
  end function lgengam

  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: ifam
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf
end subroutine llikfcnmc_11

subroutine llikfcnmc_01 (lglk, nulist, philist, kappalist, &
   sample, omgsample, omgpars, Ntot, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icflist, n, p, kg, ifamlist, itr)
!! Log-likelihood function.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns, logpdfzf => logpdfz, condymu_mf => condymu
  use covfun
  use betaprior
  use condymu, only: condymu_gt
  implicit none
  integer, intent(in) :: n, p, kg, ifamlist(kg), Ntot, icflist(kg), itr(n)
  double precision, intent(in) :: nulist(kg), omgpars(4,kg), offset(n), &
     sample(n,Ntot,kg), philist(kg), omgsample(Ntot), y(n), l(n), F(n,p), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, nu, phi, omg, kappa, &
     tsqval, respdfh, xi(n)
  integer i, j, ifam, icf
  double precision zsam(n), msam(n), jsam(n), sam(n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifamlist(1))
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  do i = 1, kg
    ifam = ifamlist(i)
    icf = icflist(i)
    call create_model (ifam)
    call create_spcor(icf,n)
    nu = nulist(i)
    kappa = kappalist(i)
    phi = philist(i)
    do j = 1, Ntot
      call rchkusr
      omg = omgsample(j)
      call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
      sam = sample(:,j,i)
      where (itr == 0)
        zsam = sam
        msam = invlink(zsam,nu)
        jsam = 0d0
      elsewhere (itr == 1)
        msam = sam
        zsam = flink(msam,nu)
        jsam = loginvlinkdz(zsam,nu)
      elsewhere (itr == 2)
        zsam = transfw(sam,nu)
        msam = invlink(zsam,nu)
        jsam = loginvtrwdz(zsam,nu)
      end where
      lglk(j,i) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
         + condymuf(ifam,n,y,l,msam,tsqval,respdfh) - sum(jsam) &
         + lgengam(omg,omgpars(:,i))
    end do
  end do

contains
  pure function lgengam (x, q)
    double precision, intent(in) :: x, q(4)
    double precision lgengam
    lgengam = (x - q(4))/q(1)
    lgengam = log(lgengam)
    lgengam = (q(2)-1)*lgengam - exp(lgengam*q(3))
  end function lgengam

  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: ifam
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf
end subroutine llikfcnmc_01

subroutine llikfcnmc_10 (lglk, nulist, omglist, kappalist, &
   sample, phisample, phipars, Ntot, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icflist, n, p, kg, ifamlist, itr)
!! Log-likelihood function.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns, logpdfzf => logpdfz, condymu_mf => condymu
  use covfun
  use betaprior
  use condymu, only: condymu_gt
  implicit none
  integer, intent(in) :: n, p, kg, ifamlist(kg), Ntot, icflist(kg), itr(n)
  double precision, intent(in) :: nulist(kg), phipars(4,kg), offset(n), &
     sample(n,Ntot,kg), omglist(kg), phisample(Ntot), y(n), l(n), F(n,p), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, nu, phi, omg, kappa, &
     tsqval, respdfh, xi(n)
  integer i, j, ifam, icf
  double precision zsam(n), msam(n), jsam(n), sam(n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifamlist(1))
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  do i = 1, kg
    ifam = ifamlist(i)
    icf = icflist(i)
    call create_model (ifam)
    call create_spcor(icf,n)
    nu = nulist(i)
    kappa = kappalist(i)
    omg = omglist(i)
    do j = 1, Ntot
      call rchkusr
      phi = phisample(j)
      call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
      sam = sample(:,j,i)
      where (itr == 0)
        zsam = sam
        msam = invlink(zsam,nu)
        jsam = 0d0
      elsewhere (itr == 1)
        msam = sam
        zsam = flink(msam,nu)
        jsam = loginvlinkdz(zsam,nu)
      elsewhere (itr == 2)
        zsam = transfw(sam,nu)
        msam = invlink(zsam,nu)
        jsam = loginvtrwdz(zsam,nu)
      end where
      lglk(j,i) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
         + condymuf(ifam,n,y,l,msam,tsqval,respdfh) - sum(jsam) &
         + lgengam(phi,phipars(:,i))
    end do
  end do

contains
  pure function lgengam (x, q)
    double precision, intent(in) :: x, q(4)
    double precision lgengam
    lgengam = (x - q(4))/q(1)
    lgengam = log(lgengam)
    lgengam = (q(2)-1)*lgengam - exp(lgengam*q(3))
  end function lgengam

  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: ifam
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf
end subroutine llikfcnmc_10

subroutine llikfcnmc_00 (lglk, nulist, philist, omglist, kappalist, &
   sample, Ntot, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icflist, n, p, kg, ifamlist, itr)
!! Log-likelihood function.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns, logpdfzf => logpdfz, condymu_mf => condymu
  use covfun
  use betaprior
  use condymu, only: condymu_gt
  implicit none
  integer, intent(in) :: n, p, kg, ifamlist(kg), Ntot, icflist(kg), itr(n)
  double precision, intent(in) :: nulist(kg), philist(kg), &
     sample(n,Ntot,kg), omglist(kg), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, ssqdfsc, modeldfh, nu, phi, omg, kappa, &
     tsqval, respdfh, xi(n)
  integer i, j, ifam, icf
  double precision zsam(n), msam(n), jsam(n), sam(n)
  double precision ll1, ll2, ll3

  ssqdfsc = ssqdf*ssqsc
  select case (ifamlist(1))
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  do i = 1, kg
    ifam = ifamlist(i)
    icf = icflist(i)
    call create_model (ifam)
    call create_spcor(icf,n)
    nu = nulist(i)
    kappa = kappalist(i)
    phi = philist(i)
    omg = omglist(i)
    call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, Ntot
      call rchkusr
      sam = sample(:,j,i)
      where (itr == 0)
        zsam = sam
        msam = invlink(zsam,nu)
        jsam = 0d0
      elsewhere (itr == 1)
        msam = sam
        zsam = flink(msam,nu)
        jsam = loginvlinkdz(zsam,nu)
      elsewhere (itr == 2)
        zsam = transfw(sam,nu)
        msam = invlink(zsam,nu)
        jsam = loginvtrwdz(zsam,nu)
      end where
      ll1 = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh)
      ll2 = condymuf(ifam,n,y,l,msam,tsqval,respdfh)
      ll3 = sum(jsam)
      lglk(j,i) = ll1 + ll2 - ll3
    end do
  end do
  
contains
  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: ifam
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf
end subroutine llikfcnmc_00

######################################################################
## 
### Commentary: Estimate the effect of the rhizoctonia disease on
### yield using full Bayesian analysis.
## 
######################################################################

library(geoBayes)

### Load the data
data(rhizoctonia)
rhiz <- na.omit(rhizoctonia)
rhiz$IR <- rhiz$Infected/rhiz$Total # Incidence rate of the
                              # rhizoctonia disease

### Define the model
corrf <- "spherical"
ssqdf <- 4
ssqsc <- 1
tsqdf <- 4
tsqsc <- 4
betm0 <- 0
betQ0 <- diag(.01, 2, 2)
phiprior <- c(200, 1, 1000, 100) # U(100, 300)
phisc <- .7
omgprior <- c(3, 1, 1000, 0) # U(0, 3)
omgsc <- 1
linkp <- 1

## MCMC parameters
Nout <- 1000
Nbi <- 3000
Nthin <- 10

samplt <- mcstrga(Yield ~ IR, data = rhiz, 
                  atsample = ~ Xcoord + Ycoord, corrf = corrf, 
                  Nout = Nout, Nthin = Nthin,
                  Nbi = Nbi, betm0 = betm0, betQ0 = betQ0,
                  ssqdf = ssqdf, ssqsc = ssqsc,
                  tsqdf = tsqdf, tsqsc = tsqsc,
                  corrprior = list(phi = phiprior, omg = omgprior),
                  linkp = linkp,
                  corrtuning = list(phi = phisc, omg = omgsc, kappa = 0), 
                  test=100)

sample <- update(samplt, test = FALSE)

mcsamp <- mcmcmake(sample)

ipar <- grep(paste(c("phi", "tsq", "ssq", "beta(_[0-9]+)?", "omg"),
                   collapse = "|"), dimnames(mcsamp)[[2]])

plot(mcsamp[, ipar])
summary(mcsamp[, ipar])

##' Compute the Bayes factors.
##'
##' Computes the Bayes factors using the importance weights at the new
##' points. The new points are taken from the grid derived by
##' expanding the parameter values inputted. The arguments
##' \code{linkp} \code{phi} \code{omg} \code{kappa} correspond to the
##' link function, spatial range, relative nugget, and correlation
##' function parameters respectively.
##' @title Compute the Bayes factors at new points
##' @param bf1obj Output from the function \code{\link{bf1skel}} which
##' contains the Bayes factors and importance sampling weights.
##' @param linkp,phi,omg,kappa Optional scalar or vector or
##' \code{NULL}. If scalar or vector, the Bayes factors are calculated
##' at those values with respect to the reference model used in
##' \code{\link{bf1skel}}. If missing or \code{NULL} then the unique
##' values from the MCMC chains that were inputted in
##' \code{\link{bf1skel}} will be used.
##' @param useCV Whether to use control variates for finer
##' corrections.
##' @return An array of size \code{length(linkp) * length(phi) *
##' length(omg) * length(kappa)} containing the Bayes factors for each
##' combination of the parameters.
##' @examples \dontrun{
##' data(rhizoctonia)
##' ### Define the model
##' corrf <- "spherical"
##' kappa <- 0
##' ssqdf <- 1
##' ssqsc <- 1
##' betm0 <- 0
##' betQ0 <- .01
##' family <- "binomial.probit"
##' ### Skeleton points
##' philist <- c(100, 140, 180)
##' omglist <- c(.5, 1)
##' parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
##' ### MCMC sizes
##' Nout <- 100
##' Nthin <- 1
##' Nbi <- 0
##' ### Take MCMC samples
##' runs <- list()
##' for (i in 1:NROW(parlist)) {
##'   runs[[i]] <- mcsglmm(Infected ~ 1, family, rhizoctonia, weights = Total,
##'                        atsample = ~ Xcoord + Ycoord,
##'                        Nout = Nout, Nthin = Nthin, Nbi = Nbi,
##'                        betm0 = betm0, betQ0 = betQ0,
##'                        ssqdf = ssqdf, ssqsc = ssqsc,
##'                        phi = parlist$phi[i], omg = parlist$omg[i],
##'                        linkp = parlist$linkp[i], kappa = parlist$kappa[i],
##'                        corrfcn = corrf,
##'                        corrtuning=list(phi = 0, omg = 0, kappa = 0))
##' }
##' bf <- bf1skel(runs)
##' bfall <- bf2new(bf, phi = seq(100, 200, 10), omg = seq(0, 2, .2))
##' plotbf2(bfall, c("phi", "omg"))
##' }
##' @importFrom sp spDists
##' @references Doss, H. (2010). Estimation of large families of Bayes
##' factors from Markov chain output. \emph{Statistica Sinica}, 20(2),
##' 537.
##'
##' Roy, V., Evangelou, E., and Zhu, Z. (2015). Efficient estimation
##' and prediction for the Bayesian spatial generalized linear mixed
##' model with flexible link functions. \emph{Biometrics}, 72(1),
##'   289-298.
##' @useDynLib geoBayes calcb_no_st calcb_mu_st calcb_wo_st calcb_tr_st
##' @useDynLib geoBayes calcb_no_cv calcb_mu_cv calcb_wo_cv calcb_tr_cv
##' @export
bf2new <- function (bf1obj, linkp, phi, omg, kappa, useCV = TRUE)
{
  ## Logical input
  useCV <- as.logical(useCV)

  ## Extract model variables
  transf <- bf1obj$transf
  real_transf <- bf1obj$real_transf
  itr <- bf1obj$itr
  y <- bf1obj$response
  n <- length(y)
  l <- bf1obj$weights
  F <- bf1obj$modelmatrix
  offset <- bf1obj$offset
  p <- NCOL(F)
  family <- bf1obj$family
  ifam <- .geoBayes_family(family)
  corrfcn <- bf1obj$corrfcn
  icf <- .geoBayes_correlation(corrfcn)
  betm0 <- bf1obj$betm0
  betQ0 <- bf1obj$betQ0
  ssqdf <- bf1obj$ssqdf
  ssqsc <- bf1obj$ssqsc
  dispersion <- bf1obj$dispersion
  tsqdf <- bf1obj$tsqdf
  tsqsc <- bf1obj$tsqsc
  tsq <- if (ifam == 0) tsqsc else dispersion
  loc <- bf1obj$locations
  dm <- sp::spDists(loc)
  sample <- bf1obj$sample2
  Ntot <- NCOL(sample)
  isweights <- bf1obj$isweights
  zcv <- bf1obj$controlvar
  kg <- NCOL(zcv)

  ## Examine link function parameter
  if (missing(linkp) || is.null(linkp)) {
    nu <- unique(bf1obj$pnts$nu)
  } else {
    nu <- unique(.geoBayes_getlinkp(linkp, ifam))
  }
  n_nu <- length(nu)

  ## Examine covariance parameters
  if (missing(phi) || is.null(phi)) {
    phi <- unique(bf1obj$pnts$phi)
  } else if (!is.numeric(phi)) {
    stop ("Argument phi must be numeric or NULL")
  }
  phi <- as.double(phi)
  if (any(phi < 0)) stop ("Argument phi must be non-negative")
  n_phi <- length(phi)
  if (missing(omg) || is.null(omg)) {
    omg <- unique(bf1obj$pnts$omg)
  } else if (!is.numeric(omg)) {
    stop ("Argument omg must be numeric or NULL")
  }
  omg <- as.double(omg)
  if (any(omg < 0)) stop ("Argument omg must be non-negative")
  n_omg <- length(omg)
  if (missing(kappa) || is.null(kappa)) {
    kappa <- unique(bf1obj$pnts$kappa)
  } else {
    kappa <- unique(.geoBayes_getkappa(kappa, icf))
  }
  n_kappa <- length(kappa)
  covpars <- expand.grid(phi = phi, omg = omg, kappa = kappa)
  n_cov <- NROW(covpars)

  bfact <- numeric(n_nu*n_cov)

  ## Check for non-finite values in control variates
  if (useCV && any(!is.finite(zcv))) {
    warning ("Computed non-finite control variates, probably due to numerical
overflow. Control variates corrections will not be used.")
    useCV <- FALSE
  }

  froutine <- paste0("calcb_", real_transf)
  if (real_transf == "wo" || real_transf == "tr") {
    ifam <- -ifam
  }

  if (useCV) {
    froutine <- paste(froutine, '_cv', sep='')
    QRin <- QRintercept(zcv)
    RUN <- .Fortran(froutine, bfact,
                    as.double(covpars$phi), as.double(nu),
                    as.double(covpars$omg),
                    as.double(covpars$kappa), as.integer(icf),
                    as.integer(n_cov), as.integer(n_nu), as.integer(Ntot),
                    as.double(sample), as.double(isweights),
                    as.double(QRin), as.integer(n), as.integer(p),
                    as.double(betm0),
                    as.double(betQ0), as.double(ssqdf), as.double(ssqsc),
                    max(tsqdf, 0), as.double(tsq), as.double(y),
                    as.double(l), as.double(F), as.double(offset), as.double(dm),
                    as.integer(ifam), as.integer(itr),
                    PACKAGE = "geoBayes")
  } else {
    froutine <- paste(froutine, '_st', sep='')
    RUN <- .Fortran(froutine, bfact,
                    as.double(covpars$phi), as.double(nu),
                    as.double(covpars$omg),
                    as.double(covpars$kappa), as.integer(icf),
                    n_cov, n_nu, Ntot, as.double(sample), as.double(isweights),
                    as.integer(n), as.integer(p), as.double(betm0),
                    as.double(betQ0), as.double(ssqdf), as.double(ssqsc),
                    max(tsqdf, 0), as.double(tsq), as.double(y), as.double(l),
                    as.double(F), as.double(offset), as.double(dm),
                    as.integer(ifam), as.integer(itr),
                    PACKAGE = "geoBayes")
  }

  logbf <- array(RUN[[1]], c(n_nu, n_phi, n_omg, n_kappa))
  logbf <- logbf + bf1obj$logbf[1] # Constant to match bf1obj$logbf
  maxid <- arrayInd(which.max(logbf), c(n_nu, n_phi, n_omg, n_kappa))
  out <- list(logbf = logbf, linkp = nu, phi = phi,
              omg = omg, corrfcn = corrfcn, kappa = kappa, indmax = maxid)
  ## class(out) <- c("bfsp", "list")
  out
}


##' This function plots the estimated logarithm Bayes factors from the
##' function \code{\link{bf2new}}.
##'
##' Depending on whether \code{pars} has length 1 or 2, this function
##' creates a line or a contour plot of the estimated Bayes factors.
##' If its length is 3 or 4, then it produces multiple profile plots.
##' In this case the variable is fixed at different values and the
##' maximum Bayes factor corresponding to the fixed value is plotted
##' against that value.
##' @title Plot the estimated Bayes factors
##' @param bf2obj Output from the function \code{\link{bf2new}}.
##' @param pars A vector with the names of the parameters to plot.
##' @param profile Whether it should produce a profile plot or a
##' contour plot if the length of pars is 2.
##' @param ... Other input to be passed to either \code{plot} or
##' \code{contour}.
##' @return This function returns nothing.
##' @examples \dontrun{
##' data(rhizoctonia)
##' ### Define the model
##' corrf <- "spherical"
##' kappa <- 0
##' ssqdf <- 1
##' ssqsc <- 1
##' betm0 <- 0
##' betQ0 <- .01
##' family <- "binomial.probit"
##' ### Skeleton points
##' philist <- c(100, 140, 180)
##' omglist <- c(.5, 1)
##' parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
##' ### MCMC sizes
##' Nout <- 100
##' Nthin <- 1
##' Nbi <- 0
##' ### Take MCMC samples
##' runs <- list()
##' for (i in 1:NROW(parlist)) {
##'   runs[[i]] <- mcsglmm(Infected ~ 1, family, rhizoctonia, weights = Total,
##'                        atsample = ~ Xcoord + Ycoord,
##'                        Nout = Nout, Nthin = Nthin, Nbi = Nbi,
##'                        betm0 = betm0, betQ0 = betQ0,
##'                        ssqdf = ssqdf, ssqsc = ssqsc,
##'                        phi = parlist$phi[i], omg = parlist$omg[i],
##'                        linkp = parlist$linkp[i], kappa = parlist$kappa[i],
##'                        corrfcn = corrf,
##'                        corrtuning=list(phi = 0, omg = 0, kappa = 0))
##' }
##' bf <- bf1skel(runs)
##' bfall <- bf2new(bf, phi = seq(100, 200, 10), omg = seq(0, 2, .2))
##' plotbf2(bfall, c("phi", "omg"))
##' plotbf2(bfall, c("phi", "omg"), profile = TRUE, type = "b", ylab="log(BF)")
##' }
##' @importFrom graphics plot points par title
##' @export
plotbf2 <- function (bf2obj, pars = c("linkp", "phi", "omg", "kappa"),
                     profile = length(pars) > 2, ...) {
  N <- 4
  pars <- unique(pars)
  pars <- match.arg(pars, several.ok = TRUE)
  usepars <- sapply(bf2obj[pars], length) > 1
  pars <- pars[usepars]
  ipar <- match(pars, eval(formals()$pars))
  npars <- length(pars)
  if (npars == 0) stop ("No parameters to plot.")
  dots <- list(...)
  nmdots <- names(dots)

  ## Determine type of plot
  if (profile) {
    ptype <- "profile"
  } else if (npars == 1) {
    ptype <- "line"
  } else if (npars == 2) {
    ptype <- "contour"
  } else {
    warning ("Only a profile plot is available for more than 3 parameters")
    ptype <- "profile"
  }

  grknm <- list(nu = expression(nu), phi = expression(phi),
                omg = expression(omega), kappa = expression(kappa))

  if (ptype == "line") {
    maxid <- bf2obj$indmax
    ii <- as.list(maxid)
    ii[ipar] <- TRUE
    bf <- drop(do.call('[', c(bf2obj["logbf"], ii, list(drop = FALSE))))
    pdata <- data.frame(bf2obj[pars], logbf = bf)
    if ("type" %in% nmdots) {
      plot(pdata, ...)
    } else {
      plot(pdata, type = "l", ...)
    }
  } else if (ptype == "contour") {
    maxid <- bf2obj$indmax
    ii <- as.list(maxid)
    ii[ipar] <- TRUE
    bf <- do.call('[', c(list(bf2obj$logbf), ii, list(drop = FALSE)))
    jj <- seq(N); jj[sort(ipar)] <- ipar
    bf <- drop(aperm(bf, jj))
    CC <- call("contour", bf2obj[[pars[1]]], bf2obj[[pars[2]]], bf,
               quote(...))
    if (!("xlab" %in% nmdots)) CC["xlab"] = pars[1]
    if (!("ylab" %in% nmdots)) CC["ylab"] = pars[2]
    eval(CC)
    points((bf2obj[[pars[1]]])[maxid[ipar[1]]],
           (bf2obj[[pars[2]]])[maxid[ipar[2]]])
  } else if (ptype == "profile") {
    ii <- as.list(rep(TRUE, N))
    if (prod(par("mfrow")) < npars) {
      oldpar <- par(mfrow = c(1, npars))
      on.exit(par(oldpar), add = TRUE)
    }
    for (i in 1:npars) {
      bf <- apply(bf2obj[["logbf"]], ipar[i], max)
      pdata <- data.frame(bf2obj[pars[i]], logbf = bf)
      CC <- call("plot", pdata, quote(...))
      if (!("type" %in% nmdots)) CC["type"] = "l"
      if (!("xlab" %in% nmdots)) CC["xlab"] = ""
      eval(CC)
      if (!("xlab" %in% nmdots)) {
        title(xlab = switch(pars[i],
                linkp = expression(nu),
                phi = expression(phi),
                omg = expression(omega),
                kappa = expression(kappa)))
      }
    }
  }
  invisible()
}


##' Estimation by empirical Bayes.
##'
##' This function is a wrap around \code{\link{bf2new}} using the
##' "L-BFGS-B" method of the function \code{\link[stats]{optim}} to
##' estimate the parameters.
##' @title Empirical Bayes estimator
##' @param bf1obj Output from the function \code{\link{bf1skel}} which
##' contains the Bayes factors and importance sampling weights.
##' @param paroptim A named list with the components "linkp", "phi",
##' "omg", "kappa". Each component must be numeric with length 1, 2,
##' or 3 with elements in increasing order but for the binomial family
##' linkp is also allowed to be the character "logit" and "probit". If
##' the compontent's length is 1, then the corresponding parameter is
##' considered to be fixed at that value. If 2, then the two numbers
##' denote the lower and upper bounds for the optimisation of that
##' parameter (infinities are allowed). If 3, these correspond to
##' lower bound, starting value, upper bound for the estimation of
##' that parameter.
##' @param useCV Whether to use control variates for finer
##' corrections.
##' @param control A list of control parameters for the optimisation.
##' See \code{\link[stats]{optim}}.
##' @return The output from the function \code{\link[stats]{optim}}.
##'   The \code{"value"} element is the log-Bayes factor, not the
##'   negative log-Bayes factor.
##' @examples \dontrun{
##' data(rhizoctonia)
##' ### Define the model
##' corrf <- "spherical"
##' kappa <- 0
##' ssqdf <- 1
##' ssqsc <- 1
##' betm0 <- 0
##' betQ0 <- .01
##' family <- "binomial.probit"
##' ### Skeleton points
##' philist <- c(100, 140, 180)
##' omglist <- c(.5, 1)
##' parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
##' ### MCMC sizes
##' Nout <- 100
##' Nthin <- 1
##' Nbi <- 0
##' ### Take MCMC samples
##' runs <- list()
##' for (i in 1:NROW(parlist)) {
##'   runs[[i]] <- mcsglmm(Infected ~ 1, family, rhizoctonia, weights = Total,
##'                        atsample = ~ Xcoord + Ycoord,
##'                        Nout = Nout*c(.8,.2), Nthin = Nthin, Nbi = Nbi,
##'                        betm0 = betm0, betQ0 = betQ0,
##'                        ssqdf = ssqdf, ssqsc = ssqsc,
##'                        phi = parlist$phi[i], omg = parlist$omg[i],
##'                        linkp = parlist$linkp[i], kappa = parlist$kappa[i],
##'                        corrfcn = corrf,
##'                        corrtuning=list(phi = 0, omg = 0, kappa = 0))
##' }
##' bf <- bf1skel(runs)
##' est <- bf2optim(bf, list(phi = c(100, 200), omg = c(0, 2)))
##' est
##' }
##' @importFrom stats optim optimHess
##' @importFrom optimx optimr
##' @useDynLib geoBayes calcb_no_st calcb_mu_st calcb_wo_st calcb_tr_st
##' @useDynLib geoBayes calcb_no_cv calcb_mu_cv calcb_wo_cv calcb_tr_cv
##' @useDynLib geoBayes calcbd_no calcbd_mu calcbd_wo calcbd_tr
##' @export
bf2optim <- function (bf1obj, paroptim, useCV = TRUE,
                      control = list()) {

  log.phi <- FALSE
  log.omg <- FALSE

  ## Logical input
  useCV <- isTRUE(useCV)

  ## Extract model variables
  transf <- bf1obj$transf
  real_transf <- bf1obj$real_transf
  itr <- bf1obj$itr
  y <- bf1obj$response
  n <- length(y)
  l <- bf1obj$weights
  F <- bf1obj$modelmatrix
  p <- NCOL(F)
  offset <- bf1obj$offset
  family <- bf1obj$family
  ifam <- .geoBayes_family(family)
  corrfcn <- bf1obj$corrfcn
  icf <- .geoBayes_correlation(corrfcn)
  betm0 <- bf1obj$betm0
  betQ0 <- bf1obj$betQ0
  ssqdf <- bf1obj$ssqdf
  ssqsc <- bf1obj$ssqsc
  dispersion <- bf1obj$dispersion
  tsqdf <- bf1obj$tsqdf
  tsqsc <- bf1obj$tsqsc
  tsq <- if (ifam == 0) tsqsc else dispersion
  loc <- bf1obj$locations
  dm <- sp::spDists(loc)
  sample <- bf1obj$sample2
  Ntot <- NCOL(sample)
  isweights <- bf1obj$isweights
  zcv <- bf1obj$controlvar
  kg <- NCOL(zcv)


  ## Check paroptim
  ## parnmall <- c("linkp", "phi", "omg", "kappa")
  paroptim.orig <- paroptim
  paroptim <- getparoptim(paroptim, ifam, icf)
  pstart <- as.double(paroptim$pstart)
  lower <- paroptim$lower
  upper <- paroptim$upper
  estim <- paroptim$estim
  log.phi <- log.phi && estim[2]
  log.omg <- log.omg && estim[3]

  ## Check for non-finite values in control variates
  if (useCV && any(!is.finite(zcv))) {
    warning ("Computed non-finite control variates, probably due to numerical
overflow. Control variates corrections will not be used.")
    useCV <- FALSE
  }

  froutine <- paste0("calcb_", real_transf)
  groutine <- paste0("calcbd_", real_transf)
  if (real_transf == "wo" || real_transf == "tr") {
    ifam <- -ifam
  }

  ## Function to optimise
  if (useCV) {
    froutine <- paste(froutine, "_cv", sep = "")
    QRin <- QRintercept(zcv)
    fn <- function (par) {
      parin <- pstart
      parin[estim] <- par
      if (log.phi) parin[2] <- exp(parin[2])
      if (log.omg) parin[3] <- exp(parin[3])
      RUN <- try(
        .Fortran(froutine, 0,
                 parin[2], parin[1], parin[3], parin[4],
                 as.integer(icf), 1L, 1L, as.integer(Ntot),
                 as.double(sample), as.double(isweights),
                 as.double(QRin),
                 as.integer(n), as.integer(p),
                 as.double(betm0),
                 as.double(betQ0), as.double(ssqdf), as.double(ssqsc),
                 max(tsqdf, 0), as.double(tsq), as.double(y),
                 as.double(l), as.double(F), as.double(offset), as.double(dm),
                 as.integer(ifam), as.integer(itr),
                 PACKAGE = "geoBayes"), silent = TRUE)
      if (inherits(RUN, "try-error")) return (NA)
      -RUN[[1]][1]
    }
  } else {
    QRin <- numeric(Ntot)
    froutine <- paste(froutine, "_st", sep = "")
    fn <- function (par) {
      parin <- pstart
      parin[estim] <- par
      if (log.phi) parin[2] <- exp(parin[2])
      if (log.omg) parin[3] <- exp(parin[3])
      RUN <- try(
        .Fortran(froutine, 0,
                 parin[2], parin[1], parin[3], parin[4],
                 as.integer(icf), 1L, 1L,
                 as.integer(Ntot), as.double(sample),
                 as.double(isweights), as.integer(n), as.integer(p),
                 as.double(betm0),
                 as.double(betQ0), as.double(ssqdf), as.double(ssqsc),
                 max(tsqdf, 0), as.double(tsq), as.double(y), as.double(l),
                 as.double(F), as.double(offset), as.double(dm),
                 as.integer(ifam), as.integer(itr),
                 PACKAGE = "geoBayes"), silent = TRUE)
      if (inherits(RUN, "try-error")) return (NA)
      -RUN[[1]][1]
    }
  }

### Compute derivative
  gr <- function (par) {
    parin <- pstart
    parin[estim] <- par
    d1 <- c(1, 1, 1, 1)
    if (log.phi) d1[2] <- parin[2] <- exp(parin[2])
    if (log.omg) d1[3] <- parin[3] <- exp(parin[3])
    RUN <- try(
      .Fortran(groutine, 0, 0, 0, 0, 0,
               parin[2], parin[1], parin[3], parin[4],
               as.integer(icf), as.integer(Ntot), as.double(sample),
               as.double(isweights),
               as.double(QRin),
               as.integer(n), as.integer(p),
               as.double(betm0),
               as.double(betQ0), as.double(ssqdf), as.double(ssqsc),
               max(tsqdf, 0), as.double(tsq), as.double(y), as.double(l),
               as.double(F), as.double(offset), as.double(dm),
               as.integer(ifam), as.integer(itr), as.integer(useCV),
               PACKAGE = "geoBayes"), silent = TRUE)
    if (inherits(RUN, "try-error")) return (rep.int(NA, sum(estim)))
    gg <- -do.call("c", RUN[2:5])[estim] * d1[estim]
    gg
  }


  logbf <- bf1obj$logbf
  imaxlogbf <- which.max(logbf)
  wbf <- exp(logbf - logbf[imaxlogbf] - log(sum(exp(logbf - logbf[imaxlogbf]))))
  pstart.d <- colSums(data.frame(bf1obj$pnts[c("nu", "phi", "omg", "kappa")])*
                      wbf)
  i <- is.na(pstart) & estim
  pstart[i] <- pmax(pmin(upper[i], pstart.d[i]), lower[i])
  if (log.phi) {
    pstart[2] <- log(pstart[2])
    lower[2] <- log(lower[2])
    upper[2] <- log(upper[2])
  }
  if (log.omg) {
    pstart[3] <- log(pstart[3])
    lower[3] <- log(lower[3])
    upper[3] <- log(upper[3])
  }
  method <- "L-BFGS-B"
  method <- "nlminb"
  if (isTRUE(control$fnscale < 0)) control$fnscale <- -control$fnscale
  if (isTRUE(control$maximize)) control$maximize <- FALSE
###   op <- stats::nlminb(pstart[estim], fn, gr, scale = scale,
###                       control = control,
###                       lower = lower[estim], upper = upper[estim])
###   op$value <- -op$objective + bf1obj$logbf[1]
###   op$hessian <- optimHess(op$par, fn, gr, control = control)
###   op <- optimx::optimr(pstart[estim], fn, gr, method = method,
###                        lower = lower[estim], upper = upper[estim],
###                        control = control, hessian = TRUE)
###   op <- stats::optim(pstart[estim], fn, gr, method = method,
###                      lower = lower[estim], upper = upper[estim],
###                      control = control, hessian = TRUE)
  op <- optimx::optimr(pstart[estim], fn, gr, method = method,
                       lower = lower[estim], upper = upper[estim],
                       control = control, hessian = FALSE)
  op$gradient <- gr(op$par)
  op$hessian <- stats::optimHess(op$par, fn, gr, control = control)
  op$value <- -op$value + bf1obj$logbf[1]
  parout <- pstart
  parout[estim] <- op$par
  d1 <- c(1, 1, 1, 1)
  names(parout) <- names(d1) <- c("linkp", "phi", "omg", "kappa")
  if (log.phi) parout[2] <- d1[2] <- exp(parout[2])
  if (log.omg) parout[3] <- d1[3] <- exp(parout[3])
  op$par <- parout
  op$hessian <- op$hessian*tcrossprod(d1[estim])
  op$fixed <- !estim
  op
}


QRintercept <- function(zcv) {
  ## Retrun the column from Q-R that corresponds to the intercept.
  QR <- qr(zcv)
  Rcv <- qr.R(QR)
  Qcv <- qr.Q(QR)
###   tol <- max(abs(diag(Rcv))) * NROW(zcv) * .Machine$double.eps
###   ii <- abs(diag(Rcv)) > tol
###   zcvrnk <- sum(ii)
  zcvii <- QR$pivot == 1
##   zcvQR <- Qcv
##   zcvQR <- .Fortran("dtrsm", "r", "u", "t", "n", as.integer(Ntot),
##                     as.integer(zcvrnk), 1.0, as.double(Rcv),
##                     as.integer(kg), as.double(zcvQR), as.integer(Ntot))[[10]]
  ## zcvQR <- t(backsolve(Rcv,t(Qcv)))
  backsolve(Rcv,t(Qcv))[zcvii, ]
}

##' Combine \code{data.frame}s
##'
##' This function combines \code{data.frame}s by filling in missing
##' variables. This is useful for combining data from
##' sampled locations with prediction locations.
##'
##' If \code{fillwith} is a named object, its names must correspond to
##' the names of variables in the data frames. If a variable is
##' missing, then it is filled with the corresponding value in
##' \code{fillwith}. \code{fillwith} can contain only one unnamed
##' component which corresponds to the default filling.
##' @title Combine \code{data.frame}s
##' @param ... \code{data.frame}s or objects that can be coerced to
##' \code{data.frame}s
##' @param fillwith Which value to use for missing variables. This
##' could be a scalar, a named vector, or a named list with one value
##' in each component; see Details.
##' @param keepclass Whether to preserve the \code{\link[base]{class}}
##' of each variable. The elements in \code{fillwith} are coerced to
##' the corresponding variable's class.
##' @return A stacked \code{data.frame}.
##' @export
##' @examples
##' \dontrun{
##' d1 <- data.frame(w = 1:3, z = 4:6 + 0.1)
##' d2 <- data.frame(w = 3:7, x = 1:5, y = 6:10)
##' (d12a <- stackdata(d1, d2))
##' lapply(d12a, class)
##' (d12b <- stackdata(d1, d2, fillwith = c(x = NA, y = 0, z = -99)))
##' lapply(d12b, class)
##' (d12c <- stackdata(d1, d2, fillwith = c(x = NA, y = 0, z = -99),
##'                    keepclass = TRUE))
##' lapply(d12c, class)
##' (d12d <- stackdata(d1, d2, fillwith = c(x = NA, 0)))
##'
##' data(rhizoctonia)
##' predgrid <- mkpredgrid2d(rhizoctonia[c("Xcoord", "Ycoord")],
##'                          par.x = 100, chull = TRUE, exf = 1.2)
##' rhizdata <- stackdata(rhizoctonia, predgrid$grid)
##' }
stackdata <- function (..., fillwith = NA, keepclass = FALSE) {
  fillNA <- function (d, allnames, fillwith) {
    miss <- allnames[!(allnames %in% names(d))]
    d[miss] <- fillwith[miss]
    d
  }
  input <- lapply(list(...), data.frame)
  nmall <- unique(unlist(lapply(input, names)))
  if (all(is.na(fillwith))) {
    fillwith <- rep(NA, length(nmall))
    names(fillwith) <- nmall
    keepclass <- FALSE
  } else {
    if (is.numeric(fillwith) | is.logical(fillwith)) {
      if (length(fillwith) == 1) {
        fillwith <- rep(fillwith, length(nmall))
        fillwith <- as.list(fillwith)
        names(fillwith) <- nmall
      } else {
        fillwith <- lapply(nmall, function(nm)
          do.call("switch", c(list(nm), as.list(fillwith))))
        names(fillwith) <- nmall
      }
    } else if (!is.list(fillwith)) {
      stop ("Argument fillwith must be either numeric or list")
    } else {
      fillwith <- lapply(nmall, function(nm)
        do.call("switch", c(list(nm), fillwith)))
      names(fillwith) <- nmall
    }
  }
  fmiss <- nmall[!(nmall %in% names(fillwith))]
  fillwith[fmiss] <- NA
  fillwith <- fillwith[nmall]
  if(keepclass) {
    types <- lapply(stackdata(..., fillwith = NA), class)
    typef <- lapply(types, function (c) match.fun(paste("as.", c, sep = "")))
    fillwith <- lapply(nmall, function(nm) typef[[nm]](fillwith[[nm]]))
    names(fillwith) <- nmall
    # fillwith <- mapply("class<-", fillwith, types, SIMPLIFY = FALSE)
  }
  newdata <- lapply(input, fillNA, allnames = nmall, fillwith = fillwith)
  out <- do.call("rbind", newdata)
  out
}

##' This function creates a grid for prediction.
##'
##' If \code{chull} this function first calculates the convex hull of
##' the points. If \code{exf} is not 1 the borders are expanded. Then
##' the function calls \code{\link[sp]{point.in.polygon}} to select
##' points that fall inside the borders.
##' @title Make prediction grid
##' @param pnts.x x coordinate of the domain. Could also be a
##' two-column matrix containing the x and y coordinates
##' @param pnts.y y coordinate of the domain. Should be
##' omitted or set to \code{NULL} if the argument \code{pnts.x} is a
##' two-column matrix.
##' @param par.x A scalar parameter for the x component of the new
##' grid. This parameter corresponds to either the \code{by} or the
##' \code{length.out} arguments of the function
##' \code{\link[base]{seq}}. Could also be a vector of two elements
##' containing the parameter for x and y.
##' @param par.y As in \code{par.x} for the y component of the new
##' grid. Should be omitted or set to \code{NULL} if the argument
##' \code{par.x} is a two-dimensional vector.
##' @param isby If \code{TRUE}, the arguments \code{par.x} and
##' \code{par.y} correspond to the \code{by} argument of the function
##' \code{\link[base]{seq}}, otherwise they correspond to
##' \code{length.out}.
##' @param chull Whether to calculate the convex hull of the points.
##' Set this to \code{TRUE} if \code{pnts.x} and \code{pnts.y} denote
##' the sampled locations. If they correspond to the borders of the
##' domain, it is recommended to set this to \code{FALSE}.
##' @param exf An expansion factor of the convex hull of
##' \code{cbind(pnts.x, pnts.y)}. Must be positive. If larger or
##' smaller than 1, the convex hull is respectively expanded or
##' contracted.
##' @return A list with components
##' \itemize{
##' \item \code{grid} A two-column matrix with the prediction grid.
##' \item \code{xycoord} A list with components "x" and "y"
##' containing the sequence of points used to create the grid.
##' \item \code{xygrid} A matrix with the full square grid derived
##' from \code{xycoord}.
##' \item \code{borders} The (expanded) borders of the domain.
##' \item \code{inxygrid} A logical vector indicating which rows of
##'   \code{xycoord} fall inside \code{borders}, and therefore
##'   correspond to the \code{grid}.
##' }
##' @importFrom sp point.in.polygon
##' @export
##' @examples
##' \dontrun{
##' data(rhizoctonia)
##' predgrid <- mkpredgrid2d(rhizoctonia[c("Xcoord", "Ycoord")],
##'                          par.x = 100, chull = TRUE, exf = 1.2)
##' plot(predgrid$borders, type = "l")         # Domain for prediction
##' points(predgrid$grid, pch = 20, cex = .3)  # Prediction locations
##' points(rhizoctonia[c("Xcoord", "Ycoord")]) # Observed locations
##' }
mkpredgrid2d <- function (pnts.x, pnts.y, par.x, par.y, isby = FALSE,
                          chull = FALSE, exf = 1) {
  if (exf <= 0) stop ("Argument exf must be positive")
  if (missing(pnts.y)) pnts.y <- NULL
  if (missing(par.y)) par.y <- NULL
  if (!is.null(pnts.x)) pnts.x <- as.matrix(pnts.x)
  if (!is.null(pnts.y)) pnts.y <- as.matrix(pnts.y)
  ph <- cbind(pnts.x, pnts.y)
  nm <- colnames(ph)
  par <- c(par.x[1], par.y[1])
  d <- NCOL(ph)
  if (d != 2) stop ("Can only generate 2-dimensional grids")
  if (isTRUE(chull)) {
    ph <- ph[chull(ph), , drop = FALSE]
  }
  if (exf != 1) { ## Extend the covex hull by a factor exf
    centr <- colMeans(ph)     # Central coordinate of the polygon
    phu <- cbind(ph[, 1, drop = FALSE] - centr[1],
                 ph[, 2, drop = FALSE] - centr[2]) # Uncenter
    phu_r <- exf*sqrt(phu[, 2, drop = FALSE]^2 +
                        phu[, 1, drop = FALSE]^2) # Covert to polar
    phu_u <- atan2(phu[, 2, drop = FALSE], phu[, 1, drop = FALSE])
    phu_x <- phu_r*cos(phu_u) # Convert back to cartesian
    phu_y <- phu_r*sin(phu_u)
    ph[, 1] <- phu_x + centr[1]
    ph[, 2] <- phu_y + centr[2]
  }
  ft <- apply(ph, 2, range)
  if (isTRUE(!isby)) {
    par <- ceiling(par)
    par <- ((ft[2, , drop = FALSE] - ft[1, , drop = FALSE])/(par - 1))
  } else {
    par <- rep(par, length.out = d)
  }
  xycoord <- lapply(1:d, function (i) seq(ft[1, i], ft[2, i], par[i]))
  names(xycoord) <- nm
  eg <- as.matrix(expand.grid(xycoord, KEEP.OUT.ATTRS = FALSE))
  dimnames(eg) <- list(NULL, nm)
  iin <- sp::point.in.polygon(eg[, 1, drop = FALSE],
                              eg[, 2, drop = FALSE],
                              ph[, 1, drop = FALSE],
                              ph[, 2, drop = FALSE]) > 0
  grid <- eg[iin, , drop = FALSE]
  out <- list(grid = grid, xycoord = xycoord, xygrid  = eg, borders = ph,
              inxygrid = iin)
  out
}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! 
!!! Commentary: Functions related to the prior for beta
!! 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module betaprior
  implicit none
contains
  subroutine betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, &
     offset)
    implicit none
    integer, intent(in) :: n, p
    double precision, intent(in) :: betm0(p), betQ0(p,p), F(n,p), ssqdf, &
       offset(n) 
    logical, intent(out) :: lmxi
    double precision, intent(out) :: modeldfh, xi(n)
    lmxi = (betQ0(1,1) .gt. 0d0) ! Normal prior
    if (lmxi) then
      modeldfh = .5d0*(n + ssqdf)
      xi = matmul(F,betm0) + offset
      lmxi = any(xi .ne. 0d0)
    else ! Uniform prior
      modeldfh = .5d0*(n - p + ssqdf)
      xi = offset
      lmxi = any(xi .ne. 0d0)
    end if
  end subroutine betapriorz
end module betaprior

module cor_fcns
  double precision, parameter :: bigneg = -huge(1d0)
  private matern, logbesselk_dk, logbesselk_hk, bigneg

!!!!!!!!!!!!!!!!!!!!!!!!!! Correlation functions !!!!!!!!!!!!!!!!!!!!!!!!!!
contains

!!! Matern
  pure elemental double precision function matern (h,k) result (c)
    use interfaces, only: fgamma, fbesselk
    implicit none
    double precision, intent(in) :: h, k
    if (h .eq. 0d0) then
      c = 1d0
    else
      c = 2d0/fgamma(k) * (.5d0*h)**k * fbesselk(h,k)
    end if
  end function matern

  pure elemental double precision function logbesselk_dk (h,k) result (c)
    ! Numerical derivatives. See
    ! http://www.math.ubc.ca/~jfeng/CHBE553/Example7/Formulae.pdf
    use interfaces, only: fbesselkexp
    implicit none
    double precision, intent(in) :: h, k
    double precision eps
    double precision kk
    eps = sqrt(epsilon(1d0))
    kk = abs(k)
    if (kk .eq. 0d0) then
      c = 0d0
    else if (kk .gt. eps) then
      c = .5d0*(log(fbesselkexp(h,kk+eps)) - log(fbesselkexp(h,kk-eps)))
    else
      c = .5d0*(-log(fbesselkexp(h,kk+eps+eps)) &
         + 4d0*log(fbesselkexp(h,kk+eps)) - 3d0*log(fbesselkexp(h,kk)))
    end if
    c = c/eps
    if (k .lt. 0d0) c = -c
  end function logbesselk_dk

  pure elemental double precision function logbesselk_hk (h,k) result (c)
    ! Numerical derivatives. See
    ! http://www.math.ubc.ca/~jfeng/CHBE553/Example7/Formulae.pdf
    use interfaces, only: fbesselkexp
    implicit none
    double precision, intent(in) :: h, k
    double precision eps
    double precision kk
    eps = sqrt(epsilon(1d0))
    kk = abs(k)
    if (kk .eq. 0d0) then
      c = 0d0
    else if (kk .gt. eps) then
      c = log(fbesselkexp(h,kk+eps)) + log(fbesselkexp(h,kk-eps)) &
         - 2d0*log(fbesselkexp(h,kk))
    else
      c = 4d0*log(fbesselkexp(h,kk+eps+eps)) + 2d0*log(fbesselkexp(h,kk)) &
         - 5d0*log(fbesselkexp(h,kk+eps)) - log(fbesselkexp(h,kk+eps*3d0))
    end if
    c = c/(eps*eps)
  end function logbesselk_hk

  pure double precision function cor_matern (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 1d0
    else if (h .gt. 0d0) then
      if (kappa .eq. .5d0) then
        c = exp(-h)
      else if (kappa .eq. 1.5d0) then
        c = exp(-h)*(1d0 + h)
      else if (kappa .eq. 2.5d0) then
        c = exp(-h)*(1d0 + h + h*h/3d0)
      else if (kappa .gt. 0d0) then
        c = matern(h,kappa)
      end if
    else
      c = bigneg
    end if
  end function cor_matern

  pure double precision function cor_dh_matern (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      if (kappa .eq. .5d0) then
        c = -exp(-h)
      else if (kappa .eq. 1.5d0) then
        c = -h*exp(-h)
      else if (kappa .eq. 2.5d0) then
        c = -exp(-h)*(h + h*h)/3d0
      else if (kappa .gt. 0d0) then
        c = matern_dh(h,kappa)
      end if
    else
      c = bigneg
    end if
  contains
    pure elemental double precision function matern_dh (h,k) result (c)
      use interfaces, only: fgamma, fbesselk
      implicit none
      double precision, intent(in) :: h, k
      if (h .eq. 0d0) then
        c = 0d0
      else
        c = -2d0/fgamma(k) * (.5d0*h)**k * fbesselk(h,k-1d0)
      end if
    end function matern_dh
  end function cor_dh_matern

  pure double precision function cor_hh_matern (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      if (kappa .eq. .5d0) then
        c = exp(-h)
      else if (kappa .eq. 1.5d0) then
        c = (h-1d0)*exp(-h)
      else if (kappa .eq. 2.5d0) then
        c = exp(-h)*(-1d0 - h + h*h)/3d0
      else if (kappa .gt. 0d0) then
        c = matern_hh(h,kappa)
      end if
    else
      c = bigneg
    end if
  contains
    pure elemental double precision function matern_hh (h,k) result (c)
      use interfaces, only: fgamma, fbesselk
      implicit none
      double precision, intent(in) :: h, k
      if (h .eq. 0d0) then
        c = 0d0
      else
        c = 1d0/fgamma(k) * (.5d0*h)**(k-1) * &
           (h*fbesselk(h,k-2d0) - fbesselk(h,k-1d0))
      end if
    end function matern_hh
  end function cor_hh_matern

  pure double precision function cor_dk_matern (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = matern_dk(h,kappa)
    end if
  contains
    pure elemental double precision function matern_dk (h,k) result (c)
      use interfaces, only: fdigamma
      double precision, intent(in) :: h, k
      double precision d1K, t
      if (h .eq. 0d0) then
        c = 0d0
      else
        d1K = logbesselk_dk(h,k)
        t = fdigamma(k) - log(.5d0*h)
        c = matern(h,k)*(d1K - t)
      end if
    end function matern_dk
  end function cor_dk_matern

  pure double precision function cor_dhdk_matern (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = matern_dhdk(h,kappa)
    end if
  contains
    pure elemental double precision function matern_dhdk (h,k) result (c)
      use interfaces, only: fbesselk, fgamma, fdigamma
      double precision, intent(in) :: h, k
      double precision d1K, t
      if (h .eq. 0d0) then
        c = 0d0
      else
        d1K = logbesselk_dk(h,k-1d0)
        t = fdigamma(k) - log(.5d0*h)
        c = -2d0/fgamma(k) * (.5d0*h)**k * fbesselk(h,k-1d0) * (d1K - t)
      end if
    end function matern_dhdk
  end function cor_dhdk_matern

  pure double precision function cor_hk_matern (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = matern_hk(h,kappa)
    end if
  contains
    pure elemental double precision function matern_hk (h,k) result (c)
      use interfaces, only: fdigamma, ftrigamma
      double precision, intent(in) :: h, k
      double precision d1K, d2K, t
      if (h .eq. 0d0) then
        c = 0d0
      else
        d1K = logbesselk_dk(h,k)
        d2K = logbesselk_hk(h,k) + d1K*d1K
        t = fdigamma(k) - log(.5d0*h)
        c = matern(h,k)*(d2K - (t+t)*d1K + t*t - ftrigamma(k))
      end if
    end function matern_hk
  end function cor_hk_matern


!!! Spherical
  pure double precision function cor_spher (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 1d0
    else if (h .ge. 1d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = 1d0 - 1.5d0*h + .5d0*h*h*h
    else
      c = bigneg
    end if
  end function cor_spher

  pure double precision function cor_dh_spher (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = -1.5d0
    else if (h .gt. 1d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = -1.5d0 + 1.5d0*h*h
    else
      c = bigneg
    end if
  end function cor_dh_spher

  pure double precision function cor_dk_spher (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_dk_spher

  pure double precision function cor_hh_spher (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 1d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = 3d0*h
    else
      c = bigneg
    end if
  end function cor_hh_spher

  pure double precision function cor_dhdk_spher (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_dhdk_spher

  pure double precision function cor_hk_spher (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_hk_spher

!!! Power-exponential
  pure double precision function cor_powexp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 1d0
    else if (h .gt. 0d0) then
      if (kappa .eq. 1d0) then
        c = exp(-h)
      else if (kappa .eq. 2d0) then
        c = exp(-h*h)
      else if (kappa .gt. 0d0 .and. kappa .le. 2d0) then
        c = exp(-(h**kappa))
      else
        c = bigneg
      end if
    else
      c = bigneg
    end if
  end function cor_powexp

  pure double precision function cor_dh_powexp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      if (kappa .eq. 1d0) then
        c = -exp(-h)
      else if (kappa .eq. 2d0) then
        c = -2d0*h*exp(-h*h)
      else if (kappa .gt. 0d0 .and. kappa .le. 2d0) then
        c = -kappa*(h**(kappa-1d0))*exp(-(h**kappa))
      else
        c = bigneg
      end if
    else
      c = bigneg
    end if
  end function cor_dh_powexp

  pure double precision function cor_dk_powexp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    double precision hk
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      if (kappa .eq. 1d0) then
        c = -h*log(h)*exp(-h)
      else if (kappa .eq. 2d0) then
        c = -h*h*log(h)*exp(-h*h)
      else if (kappa .gt. 0d0 .and. kappa .le. 2d0) then
        hk = h**kappa
        c = -hk*log(h)*exp(-hk)
      else
        c = bigneg
      end if
    else
      c = bigneg
    end if
  end function cor_dk_powexp

  pure double precision function cor_dhdk_powexp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    double precision lh, hk1, hk0, hk1lh, h2k1lh
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      lh = log(h)
      if (kappa .eq. 1d0) then
        c = (h*lh - lh - 1d0)*exp(-h)
      else if (kappa .eq. 2d0) then
        hk0 = h*h
        hk1 = hk0*h
        c = (2d0*lh*hk1 - 2d0*lh*h - h)*exp(-hk0)
      else if (kappa .gt. 0d0 .and. kappa .le. 2d0) then
        hk1 = h**(kappa -1d0)
        hk0 = hk1*h
        hk1lh = kappa*lh*hk1
        h2k1lh = hk1lh*hk0
        c = (h2k1lh - hk1lh - hk1)*exp(-hk0)
      else
        c = bigneg
      end if
    else
      c = bigneg
    end if
  end function cor_dhdk_powexp

  pure double precision function cor_hh_powexp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    double precision hk2, hk1, hk0
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      if (kappa .eq. 1d0) then
        c = exp(-h)
      else if (kappa .eq. 2d0) then
        c = 2d0*(2d0*h*h-1d0)*exp(-h*h)
      else if (kappa .gt. 0d0 .and. kappa .le. 2d0) then
        hk2 = h**(kappa - 2d0)
        hk1 = hk2*h
        hk0 = hk1*h
        c = (kappa*kappa*hk1*hk1 - kappa*(kappa - 1d0)*hk2)*exp(-hk0)
      else
        c = bigneg
      end if
    else
      c = bigneg
    end if
  end function cor_hh_powexp

  pure double precision function cor_hk_powexp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    double precision hk, lh2
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      lh2 = log(h)
      lh2 = lh2*lh2
      if (kappa .eq. 1d0) then
        c = lh2*h*(h-1d0)*exp(-h)
      else if (kappa .eq. 2d0) then
        hk = h*h
        c = lh2*hk*(hk-1d0)*exp(-hk)
      else if (kappa .gt. 0d0 .and. kappa .le. 2d0) then
        hk = h**kappa
        c = lh2*hk*(hk-1d0)*exp(-hk)
      else
        c = bigneg
      end if
    else
      c = bigneg
    end if
  end function cor_hk_powexp

!!! Exponential
  pure double precision function cor_exp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 1d0
    else if (h .gt. 0d0) then
      c = exp(-h)
    else
      c = bigneg
    end if
  end function cor_exp

  pure double precision function cor_dh_exp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = -exp(-h)
    else
      c = bigneg
    end if
  end function cor_dh_exp

  pure double precision function cor_dk_exp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_dk_exp

  pure double precision function cor_dhdk_exp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_dhdk_exp

  pure double precision function cor_hh_exp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = exp(-h)
    else
      c = bigneg
    end if
  end function cor_hh_exp

  pure double precision function cor_hk_exp (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_hk_exp

!!! Gaussian
  pure double precision function cor_gaussian (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 1d0
    else if (h .gt. 0d0) then
      c = exp(-h*h)
    else
      c = bigneg
    end if
  end function cor_gaussian

  pure double precision function cor_dh_gaussian (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = -2d0*h*exp(-h*h)
    else
      c = bigneg
    end if
  end function cor_dh_gaussian

  pure double precision function cor_dk_gaussian (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_dk_gaussian

  pure double precision function cor_dhdk_gaussian (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_dhdk_gaussian

  pure double precision function cor_hh_gaussian (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .eq. 0d0) then
      c = 0d0
    else if (h .gt. 0d0) then
      c = 2d0*(2d0*h*h-1d0)*exp(-h*h)
    else
      c = bigneg
    end if
  end function cor_hh_gaussian

  pure double precision function cor_hk_gaussian (h,kappa) result (c)
    implicit none
    double precision, intent(in) :: h, kappa
    if (h .ge. 0d0) then
      c = 0d0
    else
      c = bigneg
    end if
  end function cor_hk_gaussian
end module cor_fcns

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! 
!!! Commentary: Derivatives of pdfs w.r.t. z.
!! 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


module pdfdz
  
contains

  subroutine logcondyzdz_gt (fc, gr, nu, y1, y2, z, n, tsq)
!! TODO respdf is not needed here. Can be computed outside the fcn. XXX
!! log-pdf of y|z and its derivative w.r.t. z.
!! Gaussian transformed
!! tsq is tsqdf*tsqsc on input
!! respdf is n + tsqdf on input
    use linkfcns, only: invlink_ga
    use linkdz, only: invlinkdz_ga
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz
    fc = tsq
    do i = 1, n
      par = invlink_ga(z(i),nu)
      pardz = invlinkdz_ga(z(i),nu)
      par = y1(i) - par
      gr(i) = y2(i)*par
      fc = fc + gr(i)*par
      gr(i) = gr(i)*pardz
    end do
    gr = gr/fc ! XXX Missing respdf factor. Must be computed outside the fcn
    fc = -.5d0*log(fc) ! XXX Also missing respdf = n + tsqdf
  end subroutine logcondyzdz_gt

  subroutine logcondyzhs_gt (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Gaussian transformed
!! tsq is tsqdf*tsqsc on input
    use linkfcns, only: invlink_ga
    use linkdz, only: invlinkdz_ga
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz, fc
    fc = tsq
    do i = 1, n
      par = invlink_ga(z(i),nu)
      pardz = invlinkdz_ga(z(i),nu)
      hs(i) = y2(i)*par*pardz*pardz
      par = y1(i) - par
      fc = fc + y2(i)*par*par
    end do
    hs = -hs/fc ! XXX Missing respdf factor. Must be computed outside the fcn
  end subroutine logcondyzhs_gt

  subroutine logcondyzdz_ga (fc, gr, nu, y1, y2, z, n, tsq)
!! log-pdf of y|z and its derivative w.r.t. z.
!! Gaussian 
    use linkfcns, only: invlink_ga
    use linkdz, only: invlinkdz_ga
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz, tmp
    fc = 0d0
    do i = 1, n
      par = invlink_ga(z(i),nu)
      pardz = invlinkdz_ga(z(i),nu)
      tmp = y2(i)*par
      fc = fc + y1(i)*par - .5d0*tmp*par
      gr(i) = (y1(i) - tmp)*pardz
    end do
    fc = fc/tsq
    gr = gr/tsq
  end subroutine logcondyzdz_ga

  subroutine logcondyzhs_ga (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Gaussian 
    use linkdz, only: invlinkdz_ga
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision pardz
    do i = 1, n
      pardz = invlinkdz_ga(z(i),nu)
      hs(i) = y2(i)*pardz*pardz
    end do
    hs = -hs/tsq
  end subroutine logcondyzhs_ga

  subroutine logcondyzdz_bi (fc, gr, nu, y1, y2, z, n, tsq)
!! log-pdf of y|z and its derivative w.r.t. z.
!! Binomial robit
    use interfaces, only: fexpm1, flog1mexp
    use linkfcns, only: invlink_bi
    use linkdz, only: invlinkdz_bi
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz
    fc = 0d0
    do i = 1, n
      par = invlink_bi(z(i),nu)
      pardz = invlinkdz_bi(z(i),nu)
      fc = fc + y1(i)*par + y2(i)*flog1mexp(par)
      gr(i) = (y1(i) - y2(i)/fexpm1(-par))*pardz
    end do
    fc = fc/tsq
    gr = gr/tsq
  end subroutine logcondyzdz_bi

  subroutine logcondyzhs_bi (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Binomial robit
    use interfaces, only: fexpm1, flog1mexp
    use linkfcns, only: invlink_bi
    use linkdz, only: invlinkdz_bi
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz, tmp
    do i = 1, n
      par = invlink_bi(z(i),nu)
      pardz = invlinkdz_bi(z(i),nu)
      tmp = 1d0/fexpm1(-par)
      hs(i) = y2(i)*tmp*(1d0+tmp)*pardz*pardz
    end do
    hs = -hs/tsq
  end subroutine logcondyzhs_bi

  subroutine logcondyzdz_po (fc, gr, nu, y1, y2, z, n, tsq)
!! log-pdf of y|z and its derivative w.r.t. z.
!! Poisson
    use linkfcns, only: invlink_po
    use linkdz, only: invlinkdz_po
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz, tmp
    fc = 0d0
    do i = 1, n
      par = invlink_po(z(i),nu)
      pardz = invlinkdz_po(z(i),nu)
      tmp = y2(i)*exp(par)
      fc = fc + y1(i)*par - tmp
      gr(i) = (y1(i) - tmp)*pardz
    end do
    fc = fc/tsq
    gr = gr/tsq
  end subroutine logcondyzdz_po

  subroutine logcondyzhs_po (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Poisson
    use linkfcns, only: invlink_po
    use linkdz, only: invlinkdz_po
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz, tmp
    do i = 1, n
      par = invlink_po(z(i),nu)
      pardz = invlinkdz_po(z(i),nu)
      hs(i) = y2(i)*exp(par)*pardz*pardz
    end do
    hs = -hs/tsq
  end subroutine logcondyzhs_po

  subroutine logcondyzdz_gm (fc, gr, nu, y1, y2, z, n, tsq)
!! log-pdf of y|z and its derivative w.r.t. z.
!! Gamma
    use linkfcns, only: invlink_gm
    use linkdz, only: invlinkdz_gm
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz, tmp
    fc = 0d0
    do i = 1, n
      par = invlink_gm(z(i),nu)
      pardz = invlinkdz_gm(z(i),nu)
      tmp = y1(i)*exp(-par)
      fc = fc - tmp - y2(i)*par
      gr(i) = (tmp - y2(i))*pardz
    end do
    fc = fc/tsq
    gr = gr/tsq
  end subroutine logcondyzdz_gm

  subroutine logcondyzhs_gm (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Gamma
    use linkfcns, only: invlink_gm
    use linkdz, only: invlinkdz_gm
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz
    do i = 1, n
      par = invlink_gm(z(i),nu)
      pardz = invlinkdz_gm(z(i),nu)
      hs(i) = y1(i)*exp(-par)*pardz*pardz
    end do
    hs = -hs/tsq
  end subroutine logcondyzhs_gm

  subroutine logcondyzdz_ba (fc, gr, nu, y1, y2, z, n, tsq)
!! log-pdf of y|z and its derivative w.r.t. z.
!! Binomial asymmetric
    use interfaces, only: fexpm1, flog1mexp
    use linkfcns, only: invlink_ba
    use linkdz, only: invlinkdz_ba
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz
    fc = 0d0
    do i = 1, n
      par = invlink_ba(z(i),nu)
      pardz = invlinkdz_ba(z(i),nu)
      fc = fc + y1(i)*par + y2(i)*flog1mexp(par)
      gr(i) = (y1(i) - y2(i)*fexpm1(-par))*pardz
    end do
    fc = fc/tsq
    gr = gr/tsq
  end subroutine logcondyzdz_ba

  subroutine logcondyzhs_ba (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Binomial asymmetric
    use interfaces, only: fexpm1, flog1mexp
    use linkfcns, only: invlink_ba
    use linkdz, only: invlinkdz_ba
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz, tmp
    do i = 1, n
      par = invlink_ba(z(i),nu)
      pardz = invlinkdz_ba(z(i),nu)
      tmp = 1d0/fexpm1(-par)
      hs(i) = y2(i)*tmp*(1d0+tmp)*pardz*pardz
    end do
    hs = -hs/tsq
  end subroutine logcondyzhs_ba

  subroutine logcondyzdz_bd (fc, gr, nu, y1, y2, z, n, tsq)
!! log-pdf of y|z and its derivative w.r.t. z.
!! Binomial asymmetric decreasing
    use interfaces, only: fexpm1, flog1mexp
    use linkfcns, only: invlink_bd
    use linkdz, only: invlinkdz_bd
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz
    fc = 0d0
    do i = 1, n
      par = invlink_bd(z(i),nu)
      pardz = invlinkdz_bd(z(i),nu)
      fc = fc + y2(i)*par + y1(i)*flog1mexp(par)
      gr(i) = (y2(i) - y1(i)*fexpm1(-par))*pardz
    end do
    fc = fc/tsq
    gr = gr/tsq
  end subroutine logcondyzdz_bd

  subroutine logcondyzhs_bd (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Binomial asymmetric decreasing
    use interfaces, only: fexpm1, flog1mexp
    use linkfcns, only: invlink_bd
    use linkdz, only: invlinkdz_bd
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz, tmp
    fc = 0d0
    do i = 1, n
      par = invlink_bd(z(i),nu)
      pardz = invlinkdz_bd(z(i),nu)
      tmp = 1d0/fexpm1(-par)
      hs(i) = y1(i)*tmp*(1d0+tmp)*pardz*pardz
    end do
    hs = -hs/tsq
  end subroutine logcondyzhs_bd

  subroutine logcondyzdz_bw (fc, gr, nu, y1, y2, z, n, tsq)
!! log-pdf of y|z and its derivative w.r.t. z.
!! Binomial Wallace
    use interfaces, only: fexpm1, flog1mexp
    use linkfcns, only: invlink_bw
    use linkdz, only: invlinkdz_bw
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz
    fc = 0d0
    do i = 1, n
      par = invlink_bw(z(i),nu)
      pardz = invlinkdz_bw(z(i),nu)
      fc = fc + y1(i)*par + y2(i)*flog1mexp(par)
      gr(i) = (y1(i) - y2(i)*fexpm1(-par))*pardz
    end do
    fc = fc/tsq
    gr = gr/tsq
  end subroutine logcondyzdz_bw

  subroutine logcondyzhs_bw (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Binomial Wallace
    use interfaces, only: fexpm1, flog1mexp
    use linkfcns, only: invlink_bw
    use linkdz, only: invlinkdz_bw
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz, tmp
    fc = 0d0
    do i = 1, n
      par = invlink_bw(z(i),nu)
      pardz = invlinkdz_bw(z(i),nu)
      tmp = 1d0/fexpm1(-par)
      hs(i) = y2(i)*tmp*(1d0+tmp)*pardz*pardz
    end do
    hs = -hs/tsq
  end subroutine logcondyzhs_bw
end module pdfdz

subroutine flinkfcn (z, n, mu, linkp, ifam)
  use modelfcns, only: create_model, invtruemu, flink
  implicit none
  integer, intent(in) :: n, ifam
  double precision, intent(in) :: mu(n), linkp
  double precision, intent(out) :: z(n)
  call create_model(ifam)
  z = invtruemu(mu)
  z = flink(z,linkp)
end subroutine flinkfcn

subroutine flinkinv (mu, n, z, linkp, ifam)
  use modelfcns, only: create_model, fcntruemu, invlink
  implicit none
  integer, intent(in) :: n, ifam
  double precision, intent(in) :: z(n), linkp
  double precision, intent(out) :: mu(n)
  call create_model(ifam)
  mu = invlink(z,linkp)
  mu = fcntruemu(mu)
end subroutine flinkinv

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary:
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


subroutine spcorr (dist,phi,kappa,n,icf)
  use covfun, only: create_spcor, covmat
  implicit none
  integer, intent(in) :: n, icf
  double precision, intent(in) :: phi, kappa
  double precision, intent(inout) :: dist(n)
  call create_spcor(icf,0)
  call covmat (dist,phi,kappa,n)
end subroutine spcorr

##' @param ... Further arguments. Not currently in use.
##' @export
##' @name spcovariance
spcovariance <- function (...) {
  UseMethod("spcovariance")
}

##' Calculates the spatial variance-covariance matrix for a selection
##' of correlation functions.
##'
##' @title Spatial variance-covariance matrix
##' @param formula A formula of the form \code{~ Xcoord + Ycoord}
##'   specifying the sampled locations.
##' @param data An optional data frame containing the variables in the
##'   model.
##' @param subset An optional set of indices. The covariance will be
##'   calculated for those coordinates only.
##' @param corrfcn The correlation function to use.
##' @param ssq The partial sill parameter.
##' @param phi The spatial range parameter.
##' @param omg The relative nugget parameter.
##' @param kappa The spatial smoothness parameter.
##' @param longlat How to compute the distance between locations. If
##'   \code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
##'   distance. See \code{\link[sp]{spDists}}.
##' @return For a formula input, a variance-covariance matrix. For a
##'   numeric input, an object of the the same dimensions as its first
##'   input. 
##' @export
##' @importFrom sp spDists
##' @useDynLib geoBayes spcorr
##' @name spcovariance
spcovariance.formula <- function (formula, data, subset, corrfcn, ssq, 
                                  phi, omg, kappa, longlat = FALSE, ...)
{
  ## Correlation function
  icf <- .geoBayes_correlation(corrfcn)
  corrfcn <- .geoBayes_corrfcn$corrfcn[icf]
  if (!.geoBayes_corrfcn$needkappa[icf]) kappa <- 0

  ## Design matrix and data
  if (missing(data)) data <- environment(formula)
  mfc <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data", "subset"), names(mfc), 0L)
  mfc <- mfc[c(1L, m)]
  atsample <- update(formula, NULL ~ . + 0) # No response and no intercept
  mfc$formula <- atsample
  mfc$drop.unused.levels <- TRUE
  mfc$na.action <- "na.pass"
  mfc[[1L]] <- quote(stats::model.frame)
  mf <- eval(mfc, parent.frame())
  loc <- as.matrix(mf)
  if (!all(is.finite(loc))) stop ("Non-finite values in the locations")
  if (corrfcn == "spherical" && NCOL(loc) > 3) {
    stop ("Cannot use the spherical correlation for dimensions
grater than 3.")
  }
  nloc <- dim(loc)[1]
  labels <- rownames(loc)
  dm <- sp::spDists(loc, longlat = longlat)
  dmup <- dm[upper.tri(dm)]
  spcorr <- spcovariance.numeric(dmup, corrfcn, ssq=ssq, phi=phi,
                                 omg=0, kappa=kappa)
  Sg <- diag(ssq*(1 + omg), nloc, nloc)
  Sg[upper.tri(Sg)] <- spcorr
  Sg[lower.tri(Sg)] <- t(Sg)[lower.tri(Sg)]
  if (is.null(labels)) {
    dimnames(Sg) <- list(seq_len(nloc), seq_len(nloc))
  } else {
      dimnames(Sg) <- list(labels, labels)
  }
  Sg
}

##' @param x A numerical object of distances.
##' @export
##' @useDynLib geoBayes spcorr
##' @name spcovariance
spcovariance.numeric <- function (x, corrfcn, ssq, 
                                  phi, omg, kappa, ...)
{
  ## Correlation function
  icf <- .geoBayes_correlation(corrfcn)
  if (!.geoBayes_corrfcn$needkappa[icf]) kappa <- 0

  xin <- as.vector(x)
  lenx <- length(xin)
  spcorr <- .Fortran("spcorr", as.double(xin), as.double(phi), as.double(kappa),
                     as.integer(lenx), icf, PACKAGE = "geoBayes")[[1]]
  out <- spcorr*ssq
  ii <- xin == 0
  if (omg > 0 && any(ii)) out[ii] <- out[ii] + ssq*omg
  x[] <- out
  x
}


##' @export
##' @useDynLib geoBayes spcorr
##' @name spcovariance
spcovariance.dist <- function (x, corrfcn, ssq, 
                               phi, omg, kappa, ...)
{
  ## Correlation function
  icf <- .geoBayes_correlation(corrfcn)
  if (!.geoBayes_corrfcn$needkappa[icf]) kappa <- 0

  nloc <- attr(x,"Size")
  labels <- attr(x, "Labels")
  x <- as.vector(x)
  spcorr <- spcovariance.numeric(x, corrfcn, ssq, phi, omg=0, kappa)
  Sg <- diag(ssq*(1 + omg), nloc, nloc)
  Sg[upper.tri(Sg)] <- spcorr
  Sg[lower.tri(Sg)] <- t(Sg)[lower.tri(Sg)]
  if (is.null(labels)) {
    dimnames(Sg) <- list(seq_len(nloc), seq_len(nloc))
  } else {
      dimnames(Sg) <- list(labels, labels)
  }
  Sg
}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! 
!!! Commentary: Compute the derivative of log pdf(y|mu) w.r.t. mu
!! 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module pdfydlnk
contains
!!!!!!!!!!!!!! Compute the log-pdf of y for given parameter !!!!!!!!!!!!!!!
  pure elemental double precision function logpdfydlnk_gt (y1, y2, par)
    ! Transformed Gaussian
    ! y1 :: The average across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none 
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_gt = 2d0*y2*(par-y1)
  end function logpdfydlnk_gt

  pure elemental double precision function logpdfydlnk_ga (y1, y2, par)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none 
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_ga = y1 - y2*par
  end function logpdfydlnk_ga

  pure elemental double precision function logpdfydlnk_bi (y1, y2, par)
    ! Binomial
    ! To get the asymmetric version, switch y1 and y2
    ! y1 :: The number of successes
    ! y2 :: The number of faillures
    ! par :: The logartithm of the probability of success
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_bi = y1 - y2/fexpm1(-par)
  end function logpdfydlnk_bi

  pure elemental double precision function logpdfydlnk_po (y1, y2, par)
    ! Poisson
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_po = y1 - y2*exp(par)
  end function logpdfydlnk_po

  pure elemental double precision function logpdfydlnk_gm (y1, y2, par)
    ! Gamma
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_gm = y1*exp(-par) - y2
  end function logpdfydlnk_gm
end module pdfydlnk

!! Sample z0 many times when the covariance parameters change.
subroutine mcspz0mc (z0, Ns, z, beta, ssq, phi, omg, F, F0, &
   betQ0, kappa, icf, dm, dmdm0, Nout, n, n0, p)

  use interfaces, only: rngini, rngend, randnorm
  use covfun, only: create_spcor, calc_cov, calc_cov_pred
  implicit none
  integer, intent(in) :: Nout, n, n0, p, icf, Ns
  double precision, intent(in) :: F(n,p), F0(n0,p), &
     dm(n,n), dmdm0(n,n0), betQ0(p,p), kappa, ssq(Nout), &
     z(n,Nout), phi(Nout), omg(Nout), beta(p, Nout)
  double precision, intent(out) :: z0(n0, Ns, Nout)
  integer i, j, k
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     TC(n,n0), FCTF(n0,p), z0_ups(n0), ldh_Ups, z0_mean(n0), z0_sd(n0)

  call create_spcor(icf,n)

  call rngini
  do i = 1, Nout
    call rchkusr
    ! Compute covariance
    call calc_cov (phi(i),omg(i),dm,F,betQ0, &
       kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
    call calc_cov_pred(z0_ups, TC, FCTF, phi(i), omg(i), dmdm0, F, &
       F0, kappa,  T, n, n0, p)
    call dgemv ('t',n,n0,1d0,TC,n,z,1,0d0,z0_mean,1) ! z0_mean = C'*T^{-1}*z
    call dgemv ('n',n0,p,1d0,FCTF,n0,beta(:,i),1,1d0,z0_mean,1) ! + (F0-C'T^-1F)*b
    z0_sd = sqrt(ssq(i))*z0_ups
    do j = 1, Ns
      do k = 1, n0
        z0(k,j,i) = z0_sd(k)*randnorm() + z0_mean(k)
      end do
    end do
  end do
  call rngend
end subroutine mcspz0mc

!! Sample z0 many times when the covariance parameters are the same.
subroutine mcspz0eb (z0, Ns, z, beta, ssq, phi, omg, F, F0, &
   betQ0, kappa, icf, dm, dmdm0, Nout, n, n0, p)

  use interfaces, only: rngini, rngend, randnorm
  use covfun, only: create_spcor, calc_cov, calc_cov_pred
  implicit none
  integer, intent(in) :: Nout, n, n0, p, icf, Ns
  double precision, intent(in) :: F(n,p), F0(n0,p), &
     dm(n,n), dmdm0(n,n0), betQ0(p,p), kappa, ssq(Nout), &
     z(n,Nout), phi, omg, beta(p, Nout)
  double precision, intent(out) :: z0(n0, Ns, Nout)
  integer i, j, k
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     TC(n,n0), FCTF(n0,p), z0_ups(n0), ldh_Ups, z0_mean(n0), z0_sd(n0), &
     z0_mfix(n0)

  call create_spcor(icf,n)

  call rngini
  ! Compute covariance
  call calc_cov (phi,omg,dm,F,betQ0, &
     kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  call calc_cov_pred(z0_ups, TC, FCTF, phi, omg, dmdm0, F, &
     F0, kappa,  T, n, n0, p)
  call dgemv ('t',n,n0,1d0,TC,n,z,1,0d0,z0_mfix,1) ! z0_mean = C'*T^{-1}*z
  do i = 1, Nout
    call rchkusr
    z0_mean = z0_mfix
    call dgemv ('n',n0,p,1d0,FCTF,n0,beta(:,i),1,1d0,z0_mean,1) ! + (F0-C'T^-1F)*b
    z0_sd = sqrt(ssq(i))*z0_ups
    do j = 1, Ns
      do k = 1, n0
        z0(k,j,i) = z0_sd(k)*randnorm() + z0_mean(k)
      end do
    end do
  end do
  call rngend
end subroutine mcspz0eb


subroutine DICmc (DIC, y, l, Ns, z, beta, ssq, phi, omg, nu, F, F0, &
   kappa, icf, tsq, respdfh, dm, dm0, dmdm0, ifam, Nout, n, n0, p)

  ! tsq is tsqdfsc for transformed gaussian

  use modelfcns, condyz_sp => condyz
  use interfaces, only: rngini, rngend, randnorm
  use covfun, only: create_spcor, covmat
  use condyz, only: condyz_gt

  implicit none
  integer, intent(in) :: Nout, n, n0, p, icf, Ns, ifam
  double precision, intent(in) :: y(n0), l(n0), F(n,p), F0(n0,p), &
     dm(n,n), dm0(n0,n0), dmdm0(n,n0), kappa, ssq(Nout), &
     z(n,Nout), phi(Nout), omg(Nout), beta(p,Nout), nu, tsq, respdfh
  double precision, intent(out) :: DIC
  integer i, j, k
  logical lup(n,n), lup0(n0,n0)
  double precision T(n,n), TC(n,n0), z0_mean(n0), u(n0), &
     z0(n0), Dy, T0(n0,n0), omgp1, cnt, ssqrt, zz(n), alphz, alphz0

  call create_spcor(icf,0)

  do i = 1, n
    lup(:i-1,i) = .true.
    lup(i:,i) = .false.
  end do

  do i = 1, n0
    lup0(:i-1,i) = .true.
    lup0(i:,i) = .false.
  end do

  DIC = 0d0
  cnt = 0d0

  if (all (F .eq. 0d0)) then
    alphz = 0d0
  else
    alphz = -1d0
  end if
  if (all (F0 .eq. 0d0)) then
    alphz0 = 0d0
  else
    alphz0 = 1d0
  end if

  call rngini

  do i = 1, Nout
    call rchkusr
    ! Compute covariance
    omgp1 = omg(i) + 1d0
    T = dm
    call covmat (T, phi(i), kappa, n, n, lup)
    do k = 1, n
      T(k, k) = omgp1
    end do
    call dpotrf ('u', n, T, n, j)
    T0 = dm0
    call covmat (T0, phi(i), kappa, n0, n0, lup0)
    do k = 1, n0
      T0(k, k) = omgp1
    end do
    TC = dmdm0
    call covmat (TC, phi(i), kappa, n, n0)
    call dtrsm ('l', 'u', 't', 'n', n, n0, 1d0, T, n, TC, n)
    call dsyrk ('u', 't', n0, n, -1d0, TC, n, 1d0, T0, n0)
    call dpotrf ('u', n0, T0, n0, j)

    ssqrt = sqrt(ssq(i))
    zz = z(:,i)
    call dgemv ('n', n, p, alphz, F, n, beta(:,i), 1, 1d0, zz, 1)
    call dtrmv ('u', 't', 'n', n, T, n, zz, 1)
    call dgemv ('n', n0, p, alphz0, F0, n0, beta(:,i), 1, 0d0, z0_mean, 1)
    call dgemv ('t', n, n0, 1d0, TC, n, zz, 1, 1d0, z0_mean, 1)

    call create_model (ifam)

    do j = 1, Ns
      cnt = cnt + 1d0
      do k = 1, n0
        u(k) = randnorm()
        z0(k) = u(k)*ssqrt
      end do
      call dtrmv ('u', 't', 'n', n0, T0, n0, z0, 1)
      z0 = z0 + z0_mean
      select case (ifam)
      case (0)
        Dy = -2d0*condyz_gt(n0, y, l, z0, nu, tsq, respdfh)
      case default
        Dy = -2d0*condyz_sp(n0, y, l, z0, nu, tsq)
      end select
      DIC = DIC + (Dy - DIC)/cnt
    end do
  end do

  call rngend
end subroutine DICmc


subroutine DICeb (DIC, y, l, Ns, z, beta, ssq, phi, omg, nu, F, F0, &
   kappa, icf, tsq, respdfh, dm, dm0, dmdm0, ifam, Nout, n, n0, p)

  ! tsq is tsqdfsc for transformed gaussian

  use modelfcns, condyz_sp => condyz
  use interfaces, only: rngini, rngend, randnorm
  use covfun, only: create_spcor, covmat
  use condyz, only: condyz_gt

  implicit none
  integer, intent(in) :: Nout, n, n0, p, icf, Ns, ifam
  double precision, intent(in) :: y(n0), l(n0), F(n,p), F0(n0,p), &
     dm(n,n), dm0(n0,n0), dmdm0(n,n0), kappa, ssq(Nout), &
     z(n,Nout), phi, omg, beta(p,Nout), nu, tsq, respdfh
  double precision, intent(out) :: DIC
  integer i, j, k
  logical lup(n,n), lup0(n0,n0)
  double precision T(n,n), TC(n,n0), z0_mean(n0), u(n0), &
     z0(n0), Dy, T0(n0,n0), omgp1, cnt, zz(n), ssqrt, alphz, alphz0

  call create_spcor(icf,0)

  do i = 1, n
    lup(:i-1,i) = .true.
    lup(i:,i) = .false.
  end do

  do i = 1, n0
    lup0(:i-1,i) = .true.
    lup0(i:,i) = .false.
  end do

  DIC = 0d0
  cnt = 0d0

  if (all (F .eq. 0d0)) then
    alphz = 0d0
  else
    alphz = -1d0
  end if
  if (all (F0 .eq. 0d0)) then
    alphz0 = 0d0
  else
    alphz0 = 1d0
  end if

  call rngini

  ! Compute covariance
  omgp1 = omg + 1d0
  T = dm
  call covmat (T, phi, kappa, n, n, lup)
  do k = 1, n
    T(k, k) = omgp1
  end do
  call dpotrf ('u', n, T, n, j)
  T0 = dm0
  call covmat (T0, phi, kappa, n0, n0, lup0)
  do k = 1, n0
    T0(k, k) = omgp1
  end do
  TC = dmdm0
  call covmat (TC, phi, kappa, n, n0)
  call dtrsm ('l', 'u', 't', 'n', n, n0, 1d0, T, n, TC, n)
  call dsyrk ('u', 't', n0, n, -1d0, TC, n, 1d0, T0, n0)
  call dpotrf ('u', n0, T0, n0, j)

  call create_model (ifam)

  do i = 1, Nout
    call rchkusr
    ssqrt = sqrt(ssq(i))
    zz = z(:,i)
    call dgemv ('n', n, p, alphz, F, n, beta(:,i), 1, 1d0, zz, 1)
    call dtrmv ('u', 't', 'n', n, T, n, zz, 1)
    call dgemv ('n', n0, p, alphz0, F0, n0, beta(:,i), 1, 0d0, z0_mean, 1)
    call dgemv ('t', n, n0, 1d0, TC, n, zz, 1, 1d0, z0_mean, 1)
    do j = 1, Ns
      cnt = cnt + 1d0
      do k = 1, n0
        u(k) = randnorm()
        z0(k) = u(k)*ssqrt
      end do
      call dtrmv ('u', 't', 'n', n0, T0, n0, z0, 1)
      z0 = z0 + z0_mean
      select case (ifam)
      case (0)
        Dy = -2d0*condyz_gt(n0, y, l, z0, nu, tsq, respdfh)
      case default
        Dy = -2d0*condyz_sp(n0, y, l, z0, nu, tsq)
      end select
      DIC = DIC + (Dy - DIC)/cnt
    end do
  end do

  call rngend
end subroutine DICeb

######################################################################
##
### Commentary: EB analysis of the rhizoctonia data. Uses the
### functions mcsglmm, bf1skel, bf2new, bf2optim.
##
######################################################################

library(geoBayes)


### Load data
data(rhizoctonia)


### Create prediction grid
predgrid <- mkpredgrid2d(rhizoctonia[c("Xcoord", "Ycoord")],
                         par.x = 100, chull = TRUE, exf = 1.2)
rhizdata <- stackdata(rhizoctonia, predgrid$grid)


### Define the model
corrf <- "spherical"
kappa <- 0
ssqdf <- 1
ssqsc <- 1
betm0 <- 0
betQ0 <- .01
linkp <- 20


### Skeleton points
philist <- c(100, 140, 180)
omglist <- c(0, .5, 1, 1.5)
parlist <- expand.grid(phi=philist, linkp=linkp, omg=omglist, kappa = kappa)
estimate <- list(linkp = linkp, phi = c(100, 200), omg = c(0, 2),
                 kappa = kappa)


### MCMC sizes
Nout <- 1000
Nthin <- 10
Nbi <- 300


### Take MCMC samples
runs <- list()
for (i in seq_len(NROW(parlist))) {
  runs[[i]] <- mcsglmm(Infected ~ 1, 'binomial', rhizdata, weights = Total,
                       atsample = ~ Xcoord + Ycoord,
                       Nout = Nout*c(.8, .2), Nthin = Nthin, Nbi = Nbi,
                       betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf, ssqsc = ssqsc,
                       phi = parlist$phi[i], omg = parlist$omg[i],
                       linkp = parlist$linkp[i], kappa = parlist$kappa[i],
                       corrfcn = corrf,
                       corrtuning = list(phi = 0, omg = 0, kappa = 0))
}

bf <- bf1skel(runs)

bfall <- bf2new(bf, phi = seq(100, 200, 10), omg = seq(0, 2, .2))


plotbf2(bfall, c("phi", "omg"))

plotbf2(bfall, c("phi", "omg"), profile = TRUE, type = "b", ylab="log(BF)")

bf2optim(bf, estimate)

######################################################################
##
### Commentary: MCMC analysis of the rhizoctonia data.
##
######################################################################

library(geoBayes)

data(rhizoctonia)

### Create prediction grid
predgrid <- mkpredgrid2d(rhizoctonia[c("Xcoord", "Ycoord")],
                         par.x = 100, chull = TRUE, exf = 1.2)
rhizdata <- stackdata(rhizoctonia, predgrid$grid)

### Define the model
corrf <- "spherical"
kappa <- 0
ssqdf <- 1
ssqsc <- 1
betm0 <- 0
betQ0 <- .01
phiprior <- c(100, 1, 1000, 100) # U(100, 200)
phisc <- 2.5
omgprior <- c(2, 1, 1, 0)        # Exp(mean = 2)
omgsc <- .3
linkp <- 20

### MCMC sizes
Nout <- 1000
Nthin <- 10
Nbi <- 300

emt <- mcsglmm(Infected ~ 1, 'binomial', rhizdata, weights = Total,
               atsample = ~ Xcoord + Ycoord,
               Nout = Nout, Nthin = Nthin, Nbi = Nbi,
               betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf, ssqsc = ssqsc,
               corrprior = list(phi = phiprior, omg = omgprior), linkp = linkp,
               corrfcn = corrf, kappa = kappa,
               corrtuning = list(phi = phisc, omg = omgsc, kappa = 0), 
               dispersion = 1, test = TRUE)

emc <- update(emt, test = FALSE)

emcmc <- mcmcmake(emc)

plot.ts(cbind(phi = emc$MCMC$phi, omg = emc$MCMC$omg,
              beta = c(emc$MCMC$beta), ssq = emc$MCMC$ssq),
        nc = 2)

summary(emcmc[, c("phi", "omg", "beta", "ssq")])

plot(emcmc[, c("phi", "omg", "beta", "ssq")])

z0img <- matrix(NA, 100, 100)
z0img[predgrid$inxygrid] <- rowMeans(emc$MCMC$z[!emc$MCMC$whichobs, ])
image(predgrid$xycoord$Xcoord, predgrid$xycoord$Ycoord, z0img,
      col = gray((64:32)/64), xlab = "X", ylab = "Y")
polygon(predgrid$borders)

!!! Compute the BF standard errors

!!!   Compute BF and SE
!!!
!!! @param logbf        The log(BF) (output)
!!! @param Sig  The covariance matrix of exp(logbf)
!!! @param SE   The SE value (output)
!!! @param VT1  The first term of variance
!!! @param VT2  The second term of variance
!!! @param iref Which model is the reference model
!!! @param phi  A vector of values to compute the SE
!!! @param omg  A vector of values to compute the SE
!!! @param nu   A vector of values to compute the SE
!!! @param kappa        A vector of values to compute the SE
!!! @param philist      A vector of values to compute the BF
!!! @param omglist      A vector of values to compute the BF
!!! @param nulist       A vector of values to compute the BF
!!! @param kappalist    A vector of values to compute the BF
!!! @param sample1      The sample to be used to compute the BF
!!! @param Nout1        The sample size from each model
!!! @param Ntot1        sum(Nout1)
!!! @param sample2      The sample to be used to compute the SE
!!! @param Nout2        The sample size from each model
!!! @param Ntot2        sum(Nout2)
!!! @param y    Vector of observations
!!! @param l    Vector of replications
!!! @param F    Design matrix
!!! @param dm   Distance matrix
!!! @param betm0        Prior mean for beta
!!! @param betQ0        Prior precision for beta
!!! @param ssqdf        Prior df for sigma^2
!!! @param ssqsc        Prior scale for sigma^2
!!! @param tsqdf        Prior df for tau^2 if transformed Gaussian model
!!! @param tsq  Prior scale for tau^2 for transf Ga model or dispersion ow
!!! @param icf	Correlation function
!!! @param n	Spatial dimension
!!! @param p	length of beta
!!! @param nnew	length of phi
!!! @param kg	number of models
!!! @param ifam	family
!!! @param imeth	Method to estimate the BF
!!! @param nb1	Number of batches to create when computing the 1st SE term
!!! @param nb2	Number of batches to create when computing the 2nd SE term
!!! @param ibvmeth	Which method to use for batch variance calculation
!!! @return SE, logbf, Sig, VT1, VT2

subroutine bfse_no (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
   phi, omg, nu, kappa, &
   philist, omglist, nulist, kappalist, &
   sample1, Nout1, Ntot1, sample2, Nout2, Ntot2, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, nnew, kg, ifam, imeth, nb1, nb2, ibvmeth, itr)
  implicit none
  integer, intent(in) :: n, p, kg, ifam, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, icf, iref, nb1(kg), nnew, nb2(kg), ibvmeth, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     sample1(n,Ntot1), sample2(n,Ntot2), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p, p), ssqdf, ssqsc, tsqdf, tsq, &
     kappalist(kg), phi(nnew), omg(nnew), nu(nnew), kappa(nnew)
  double precision, intent(out) :: bf(kg), SE(nnew), Sig(1:kg-1,1:kg-1), &
     VT1(nnew), VT2(nnew), logbfnew(nnew)
  double precision llik1(Ntot1,kg), llik2(Ntot2,kg), &
     llikn(Ntot2,nnew)
  double precision Bet(kg,kg), OOmg(kg,kg)

  ! Compute log-likelihood values
  call llikfcn_no (llik1, philist, omglist, nulist, kappalist, sample1, &
     Ntot1, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, kg, ifam, itr)

  call llikfcn_no (llik2, philist, omglist, nulist, kappalist, sample2, &
     Ntot2, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, kg, ifam, itr)

  call llikfcn_no (llikn, phi, omg, nu, kappa, sample2, &
     Ntot2, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, nnew, ifam, itr)

  call bfsecalc (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
   llik1, llik2, llikn, Nout1, Ntot1, Nout2, Ntot2, &
   nnew, kg, imeth, nb1, nb2, ibvmeth, Bet, OOmg)
end subroutine bfse_no

subroutine bfse_mu (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
   phi, omg, nu, kappa, &
   philist, omglist, nulist, kappalist, &
   sample1, Nout1, Ntot1, sample2, Nout2, Ntot2, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, nnew, kg, ifam, imeth, nb1, nb2, ibvmeth, itr)
  implicit none
  integer, intent(in) :: n, p, kg, ifam, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, icf, iref, nb1(kg), nnew, nb2(kg), ibvmeth, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     sample1(n,Ntot1), sample2(n,Ntot2), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, &
     kappalist(kg), phi(nnew), omg(nnew), nu(nnew), kappa(nnew)
  double precision, intent(out) :: bf(kg), SE(nnew), Sig(1:kg-1,1:kg-1), &
     VT1(nnew), VT2(nnew), logbfnew(nnew)
  double precision llik1(Ntot1,kg), llik2(Ntot2,kg), &
     llikn(Ntot2,nnew)
  double precision Bet(kg,kg), OOmg(kg,kg)

  ! Compute log-likelihood values
  call llikfcn_mu (llik1, philist, omglist, nulist, kappalist, sample1, &
     Ntot1, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, kg, ifam, itr)

  call llikfcn_mu (llik2, philist, omglist, nulist, kappalist, sample2, &
     Ntot2, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, kg, ifam, itr)

  call llikfcn_mu (llikn, phi, omg, nu, kappa, sample2, &
     Ntot2, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, nnew, ifam, itr)

  call bfsecalc (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
     llik1, llik2, llikn, Nout1, Ntot1, Nout2, Ntot2, &
     nnew, kg, imeth, nb1, nb2, ibvmeth, Bet, OOmg)
end subroutine bfse_mu

subroutine bfse_wo (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
   phi, omg, nu, kappa, &
   philist, omglist, nulist, kappalist, &
   sample1, Nout1, Ntot1, sample2, Nout2, Ntot2, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, nnew, kg, ifam, imeth, nb1, nb2, ibvmeth, itr)
  implicit none
  integer, intent(in) :: n, p, kg, ifam, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, icf, iref, nb1(kg), nnew, nb2(kg), ibvmeth, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     sample1(n,Ntot1), sample2(n,Ntot2), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, &
     kappalist(kg), phi(nnew), omg(nnew), nu(nnew), kappa(nnew)
  double precision, intent(out) :: bf(kg), SE(nnew), Sig(1:kg-1,1:kg-1), &
     VT1(nnew), VT2(nnew), logbfnew(nnew)
  double precision llik1(Ntot1,kg), llik2(Ntot2,kg), &
     llikn(Ntot2,nnew)
  double precision Bet(kg,kg), OOmg(kg,kg)

  ! Compute log-likelihood values
  call llikfcn_wo (llik1, philist, omglist, nulist, kappalist, sample1, &
     Ntot1, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, kg, ifam, itr)

  call llikfcn_wo (llik2, philist, omglist, nulist, kappalist, sample2, &
     Ntot2, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, kg, ifam, itr)

  call llikfcn_wo (llikn, phi, omg, nu, kappa, sample2, &
     Ntot2, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, nnew, ifam, itr)

  call bfsecalc (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
   llik1, llik2, llikn, Nout1, Ntot1, Nout2, Ntot2, &
   nnew, kg, imeth, nb1, nb2, ibvmeth, Bet, OOmg)
end subroutine bfse_wo

subroutine bfse_tr (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
   phi, omg, nu, kappa, &
   philist, omglist, nulist, kappalist, &
   sample1, Nout1, Ntot1, sample2, Nout2, Ntot2, &
   y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, nnew, kg, ifam, imeth, nb1, nb2, ibvmeth, itr)
  implicit none
  integer, intent(in) :: n, p, kg, ifam, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, icf, iref, nb1(kg), nnew, nb2(kg), ibvmeth, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     sample1(n,Ntot1), sample2(n,Ntot2), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, &
     kappalist(kg), phi(nnew), omg(nnew), nu(nnew), kappa(nnew)
  double precision, intent(out) :: bf(kg), SE(nnew), Sig(1:kg-1,1:kg-1), &
     VT1(nnew), VT2(nnew), logbfnew(nnew)
  double precision llik1(Ntot1,kg), llik2(Ntot2,kg), &
     llikn(Ntot2,nnew)
  double precision Bet(kg,kg), OOmg(kg,kg)

  ! Compute log-likelihood values
  call llikfcn_tr (llik1, philist, omglist, nulist, kappalist, sample1, &
     Ntot1, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, kg, ifam, itr)

  call llikfcn_tr (llik2, philist, omglist, nulist, kappalist, sample2, &
     Ntot2, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, kg, ifam, itr)

  call llikfcn_tr (llikn, phi, omg, nu, kappa, sample2, &
     Ntot2, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, icf, &
     n, p, nnew, ifam, itr)

  call bfsecalc (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
   llik1, llik2, llikn, Nout1, Ntot1, Nout2, Ntot2, &
   nnew, kg, imeth, nb1, nb2, ibvmeth, Bet, OOmg)
end subroutine bfse_tr



subroutine bfsecalc (bf, logbfnew, Sig, SE, VT1, VT2, iref, &
   llik1, llik2, llikn, Nout1, Ntot1, Nout2, Ntot2, &
   nnew, kg, imeth, nb1, nb2, ibvmeth, Bet, OOmg)
  use bmargin
  implicit none
  integer, intent(in) :: kg, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, iref, nb1(kg), nnew, nb2(kg), ibvmeth
  double precision, intent(in) :: &
     llik1(Ntot1,kg), llik2(Ntot2,kg), llikn(Ntot2,nnew)
  double precision, intent(out) :: bf(kg), SE(nnew), Sig(1:kg-1,1:kg-1), &
     VT1(nnew), VT2(nnew), OOmg(kg,kg), logbfnew(nnew)
  double precision eta(kg), logbf(kg)
  double precision logYY(Ntot1,kg), YY(Ntot1,kg), logVV(Ntot2,nnew), &
     VV(Ntot2,nnew), VVcol(Ntot2)
  double precision Bet(kg,kg), BMP(kg,kg), BD(kg,1:kg-1), &
     OOmgBD(kg,1:kg-1), cvec(1:kg-1,nnew), logcall(kg,nnew), &
     SigC(1:kg-1)
  double precision NNratio(kg), logN(kg)
  integer i, j, kgm1, ia, ie
  double precision tmp

  if (kg .lt. 2) then
    logbf = 0d0
    bf = 1d0
    go to 9
  end if

  ! Estimate BF
  logN = log(dble(Nout1))
  eta = logN
  select case (imeth)
  case (1)
    call revlogistic (eta,llik1,kg,Ntot1,Nout1)
  case (2)
    call revlogistic (eta,llik1,kg,Ntot1,Nout1)
    call mengwong (eta,llik1,kg,Ntot1,Nout1)
  end select
  ! eta_i = log(N_i) - log(bf_i)
  logbf = logN - eta ! = log(bf_i)
  logbf = logbf - logbf(iref)
  bf = exp(logbf)
  eta = eta - sum(eta)/dble(kg) ! zeta

  logYY = logp(llik1,eta,Ntot1,kg)
  YY = exp(logYY)

  OOmg = 0d0
  ie = 0
  do j = 1, kg
    ia = ie + 1
    ie = ie + Nout1(j)
    Bet = bmmcvrmat(ibvmeth,YY(ia:ie,:),Nout1(j),kg,nb1(j))
    OOmg = OOmg + Bet*dble(Nout1(j))/dble(Ntot1)
  end do

  do j = 1, kg
    do i = 1, j-1
      Bet(i,j) = -dot_product(YY(:,i),YY(:,j))/dble(Ntot1)
    end do
    Bet(j,j) = dot_product(YY(:,j),1d0-YY(:,j))/dble(Ntot1)
  end do

  BMP = Bet + 1d0/dble(kg)
  call dpotrf("u",kg,BMP,kg,i)
  call dpotri("u",kg,BMP,kg,i)
  BMP = BMP - 1d0/dble(kg)

  j = 0
  eta = symcol(BMP,kg,iref)
  do i = 1, kg
    if (i .eq. iref) cycle
    j = j + 1
    BD(:,j) = (eta - symcol(BMP,kg,i))*bf(i) ! B * D
  end do

9 kgm1 = kg - 1
  logN = log(dble(Nout2))
  eta = logN - logbf
  logbfnew = logbfnew_calc(llik2,llikn,eta,Ntot2,kg,nnew)
  OOmgBD = 0d0
  Sig = 0d0
  SigC = 0d0
  VT1 = 0d0

  if (kgm1 .gt. 0) then
    call dsymm("l","u",kg,kgm1,1d0,OOmg,kg,BD,kg,0d0,OOmgBD,kg)
    call dgemm("t","n",kgm1,kgm1,kg,1d0,BD,kg,OOmgBD,kg,0d0,Sig,kgm1)

    logcall = logc(llik2,llikn,eta,Ntot2,kg,nnew) + spread(eta+eta-logN,2,nnew)
    cvec(:iref-1,:) = exp(logcall(:iref-1,:))
    cvec(iref:,:) = exp(logcall(iref+1:,:))

    do i = 1, nnew
      call dsymv ("u",kgm1,1d0,Sig,kgm1,cvec(:,i),1,0d0,SigC,1)
      VT1(i) = dot_product(cvec(:,i),SigC)
    end do
    VT1 = dble(Ntot2)*VT1/dble(Ntot1)
  end if

  logVV = logb(llik2,llikn,eta,Ntot2,kg,nnew) + log(dble(Ntot2))
  VV = exp(logVV)

  NNratio = dble(Nout2)/dble(Ntot2)
  VT2 = 0d0
  do i = 1, nnew
    VVcol = VV(:,i)
    ie = 0
    do j = 1, kg
      ia = ie + 1
      ie = ie + Nout2(j)
      tmp = bmmcvr(ibvmeth,VVcol(ia:ie),Nout2(j),nb2(j))
      VT2(i) = VT2(i) + tmp*NNratio(j)
    end do
  end do

  SE = sqrt(VT1 + VT2)
contains
  function symcol (A,n,i)
    ! Gives the ith column of a symmetric matrix when only its upper triangle
    ! is filled
    implicit none
    integer, intent(in) :: n, i
    double precision, intent(in) :: A(n,n)
    double precision symcol(n)
    symcol(:i) = A(:i,i)
    symcol(i+1:) = A(i,i+1:)
  end function symcol

  function logp (llik,eta,Ntot,kg)
    use flogsumexp, only: logrsumexp
    implicit none
    integer, intent(in) :: Ntot, kg
    double precision, intent(in) :: llik(Ntot,kg), eta(kg)
    double precision logp(Ntot,kg)
    double precision lliketa(Ntot,kg), lgdenom(Ntot)
    integer i, j
    do j = 1, kg
      do i = 1, Ntot
        lliketa(i,j) = llik(i,j) + eta(j)
      end do
    end do
    lgdenom = logrsumexp(lliketa,Ntot,kg)
    logp = spread(lgdenom,2,kg)
    logp = lliketa - logp
  end function logp

  function logb (llik1,llik2,eta,Ntot,kg,nnew)
    use flogsumexp, only: logrsumexp
    implicit none
    integer, intent(in) :: Ntot, kg, nnew
    double precision, intent(in) :: llik1(Ntot,kg), llik2(Ntot,nnew), eta(kg)
    double precision logb(Ntot,nnew)
    double precision lliketa(Ntot,kg), lgdenom(Ntot)
    integer i, j
    do j = 1, kg
      do i = 1, Ntot
        lliketa(i,j) = llik1(i,j) + eta(j)
      end do
    end do
    lgdenom = logrsumexp(lliketa,Ntot,kg)
    logb = spread(lgdenom,2,nnew)
    logb = llik2 - logb
  end function logb

  function logc (llik1,llik2,eta,Ntot,kg,nnew)
    ! Computes log c(h)_j without the N_j/d_j^2 term
    ! eta(j) = log(Nout2(j)) - log(BF(j))
    use flogsumexp
    implicit none
    integer, intent(in) :: Ntot, kg, nnew
    double precision, intent(in) :: llik1(Ntot,kg), llik2(Ntot,nnew), eta(kg)
    double precision logc(kg,nnew)
    double precision lliketa(Ntot,kg), lgdenom(Ntot), llikrat(Ntot,kg)
    integer i
    lliketa = spread(eta,1,Ntot)
    lliketa = llik1 + lliketa
    lgdenom = logrsumexp(lliketa,Ntot,kg)
    lgdenom = lgdenom + lgdenom
    llikrat = spread(lgdenom,2,kg)
    llikrat = llik1 - llikrat
    do i = 1, nnew
      lliketa = spread(llik2(:,i),2,kg)
      lliketa = lliketa + llikrat
      logc(:,i) = logcsumexp(lliketa,Ntot,kg)
    end do
  end function logc

  function bmmcvrmat_st (x,n,k,b)
    ! Split matrix x(n,k) to b batches of roughly equal size and compute the
    ! MC variance. The output is a k*k matrix.
    implicit none
    integer, intent(in) :: n, k, b
    double precision, intent(in) :: x(n,k)
    double precision bmmcvrmat_st(k,k)
    integer bsz(b), m, r, i
    double precision bmean(b,k), xmean(k), dbsz(b)
    m = n/b ! Minimum size for each batch
    r = mod(n,b)
    bsz(:r) = m + 1
    bsz(r+1:) = m ! Size of each batch
    dbsz = dble(bsz)
    do i = 1, k
      bmean(:,i) = batchmeans(x(:,i),n,b)
    end do
    xmean = matmul(dbsz,bmean)/dble(n)
    do i = 1, k
      bmean(:,i) = bmean(:,i) - xmean(i)
    end do
    bmmcvrmat_st = 0d0
    do i = 1, b
      xmean = bmean(i,:)
      call dsyr ("u",k,dbsz(i),xmean,1,bmmcvrmat_st,k)
    end do
    bmmcvrmat_st = bmmcvrmat_st/dble(b-1)
  end function bmmcvrmat_st

  function bmmcvr_st(x,n,b)
    ! Split vector x(n) to b batches of roughly equal size and compute the
    ! MC variance. The output is a scalar.
    implicit none
    integer, intent(in) :: n, b
    double precision, intent(in) :: x(n)
    double precision bmmcvr_st
    integer bsz(b), m, r
    double precision bmean(b), xmean, dbsz(b)
    m = n/b
    r = mod(n,b)
    bsz(:r) = m + 1
    bsz(r+1:) = m
    dbsz = dble(bsz)
    bmean = batchmeans(x,n,b)
    xmean = dot_product(bmean,dbsz)/dble(n)
    bmean = bmean - xmean
    bmean = bmean*bmean
    bmmcvr_st = dot_product(bmean,dbsz)/dble(b-1)
    ! bmmcvr_st = bmmcvr_st/dble(n)
  end function bmmcvr_st

  function bmmcvrmat (i,x,n,k,b)
    ! Split matrix x(n,k) to b batches of roughly equal size and compute the
    ! MC variance. The output is a k*k matrix.
    implicit none
    integer, intent(in) :: i, n, k, b
    double precision, intent(in) :: x(n,k)
    double precision bmmcvrmat(k,k)
    select case (i)
    case (1) ! Standard
      bmmcvrmat = bmmcvrmat_st(x,n,k,b)
    case (2) ! Spectral Tukey-Hanning
      bmmcvrmat = bmmcvrmat_th(x,n,k,b)
    case (3) ! Spectral modified Bartlett
      bmmcvrmat = bmmcvrmat_mb(x,n,k,b)
    end select
  end function bmmcvrmat

  function batchmeans (x,n,b)
    ! Split vector x(n) to b batches of roughly equal size and compute the
    ! mean of each batch. The output is a vector of size b.
    implicit none
    integer, intent(in) :: n, b
    double precision, intent(in) :: x(n)
    double precision batchmeans(b)
    integer m, i, ia, ie, r, d

    d = n/b ! Minimum batch size
    if (d .eq. 0) then
      batchmeans(1:n) = x
      batchmeans(n+1:b) = (d*b)/d
      return
    end if
    r = n - d*b
    m = d + 1
    ie = 0
    do i = 1, r
      ia = ie + 1
      ie = ie + m
      batchmeans(i) = mean(x(ia:ie),m)
    end do
    m = d
    do i = r+1, b
      ia = ie + 1
      ie = ie + m
      batchmeans(i) = mean(x(ia:ie),m)
    end do
  end function batchmeans

  function bmmcvrmat_th (x,n,k,b)
    implicit none
    integer, intent(in) :: n, k, b
    double precision, intent(in) :: x(n,k)
    double precision bmmcvrmat_th(k,k)
    integer i, j
    double precision xmm(n,k), x1(k), x2(k), wj
    double precision, parameter :: pi = 3.1415926535897932384626433832795d0
    x1 = sum(x,1)/dble(n)
    do j = 1, k
      do i = 1, n
        xmm(i,j) = x(i,j) - x1(j) ! Y_i - W
      end do
    end do
    bmmcvrmat_th = 0d0
    do i = 1, n ! j = 0
      x1 = xmm(i,:)
      call dsyr ('u',k,1d0,x1,1,bmmcvrmat_th,k)
    end do
    do j = 1, b - 1
      wj = .5d0*(1d0 + cos(pi*dble(j)/dble(b)))
      do i = 1, n - j
        x1 = xmm(i,:)
        x2 = xmm(i+j,:)
        call dsyr2 ('u',k,wj,x1,1,x2,1,bmmcvrmat_th,k)
      end do
    end do
    bmmcvrmat_th = bmmcvrmat_th/dble(n)
  end function bmmcvrmat_th

  function bmmcvr_th(x,n,b)
    ! Compute the univariate spectral variance estimator using the
    ! Tukey-Hanning lag window.
    implicit none
    integer, intent(in) :: n, b
    double precision, intent(in) :: x(n)
    double precision bmmcvr_th
    integer i, j
    double precision xmm(n), wj
    double precision, parameter :: pi = 3.1415926535897932384626433832795d0
    wj = sum(x)/dble(n) ! mean(x,n)
    xmm = x - wj
    bmmcvr_th = dot_product(xmm,xmm) ! j = 0
    do j = 1, b - 1
      wj = 1d0 + cos(pi*dble(j)/dble(b)) ! Without the 0.5 factor because
      ! it cancels with the factor 2 below.
      do i = 1, n - j
        bmmcvr_th = bmmcvr_th + wj*xmm(i)*xmm(i+j)
      end do
    end do
    bmmcvr_th = bmmcvr_th/dble(n)
  end function bmmcvr_th

  function bmmcvr (i,x,n,b)
    ! Compute the univariate spectral variance estimator using the
    ! Tukey-Hanning lag window.
    implicit none
    integer, intent(in) :: i, n, b
    double precision, intent(in) :: x(n)
    double precision bmmcvr
    select case (i)
    case (1) ! Standard
      bmmcvr = bmmcvr_st(x,n,b)
    case (2) ! Spectral Tukey-Hanning
      bmmcvr = bmmcvr_th(x,n,b)
    case (3) ! Spectral modified Bartlett
      bmmcvr = bmmcvr_mb(x,n,b)
    case default
      bmmcvr = -huge(bmmcvr)
    end select
  end function bmmcvr

  function bmmcvrmat_mb (x,n,k,b)
    implicit none
    integer, intent(in) :: n, k, b
    double precision, intent(in) :: x(n,k)
    double precision bmmcvrmat_mb(k,k)
    integer i, j
    double precision xmm(n,k), x1(k), x2(k), wj
    x1 = sum(x,1)/dble(n)
    do j = 1, k
      do i = 1, n
        xmm(i,j) = x(i,j) - x1(j) ! Y_i - W
      end do
    end do
    bmmcvrmat_mb = 0d0
    do i = 1, n ! j = 0
      x1 = xmm(i,:)
      call dsyr ('u',k,1d0,x1,1,bmmcvrmat_mb,k)
    end do
    do j = 1, b - 1
      wj = 1d0 - dble(j)/dble(b)
      do i = 1, n - j
        x1 = xmm(i,:)
        x2 = xmm(i+j,:)
        call dsyr2 ('u',k,wj,x1,1,x2,1,bmmcvrmat_mb,k)
      end do
    end do
    bmmcvrmat_mb = bmmcvrmat_mb/dble(n)
  end function bmmcvrmat_mb

  function bmmcvr_mb(x,n,b)
    ! Compute the univariate spectral variance estimator using the
    ! modified Bartlett lag window.
    implicit none
    integer, intent(in) :: n, b
    double precision, intent(in) :: x(n)
    double precision bmmcvr_mb
    integer i, j
    double precision xmm(n), wj
    wj = mean(x,n)
    xmm = x - wj
    bmmcvr_mb = dot_product(xmm,xmm) ! j = 0
    do j = 1, b - 1
      wj = 2d0*(1d0 - dble(j)/dble(b)) ! With an extra factor 2 beacause
      ! of the sum below
      do i = 1, n - j
        bmmcvr_mb = bmmcvr_mb + wj*xmm(i)*xmm(i+j)
      end do
    end do
    bmmcvr_mb = bmmcvr_mb/dble(n)
  end function bmmcvr_mb

  double precision function mean(x, m)
    use interfaces, only: isfinite
    implicit none
    integer, intent(in) :: m
    double precision, intent(in) :: x(m)
    integer i
    double precision tot, tot0, mm
    mean = 0d0
    tot = 0d0
    mm = dble(m)
    do i = 1, m
      tot0 = tot
      tot = tot + x(i)
      if (isfinite(tot) .eq. 0) then
        mean = mean + tot0/mm
        tot = x(i)
      end if
    end do
    mean = mean + tot/mm
  end function mean

  function logbfnew_calc (llik2,llikn,eta,Ntot,kg,nnew)
    use flogsumexp, only: logrsumexp, logcsumexp
    implicit none
    integer, intent(in) :: Ntot, kg, nnew
    double precision, intent(in) :: llik2(Ntot,kg), llikn(Ntot,nnew), &
       eta(kg)
    double precision logbfnew_calc(nnew)
    double precision logb(Ntot,nnew)
    double precision lliketa(Ntot,kg), lgdenom(Ntot)
    integer i, j
    if (nnew .gt. 0) then
      do j = 1, kg
        do i = 1, Ntot
          lliketa(i,j) = llik2(i,j) + eta(j)
        end do
      end do
      lgdenom = logrsumexp(lliketa,Ntot,kg)
      logb = spread(lgdenom,2,nnew)
      logb = llikn - logb
      logbfnew_calc = logcsumexp(logb,Ntot,nnew)
    end if
  end function logbfnew_calc
end subroutine bfsecalc



subroutine logbfcalc (logbf, logbfnew, iref, &
   llik1, llik2, llikn, Nout1, Ntot1, Nout2, Ntot2, &
   nnew, kg, imeth)
  use bmargin
  implicit none
  integer, intent(in) :: kg, imeth, Nout1(kg), Ntot1, &
     Nout2(kg), Ntot2, nnew, iref
  double precision, intent(in) :: &
     llik1(Ntot1,kg), llik2(Ntot2,kg), llikn(Ntot2,nnew)
  double precision, intent(out) :: logbf(kg), logbfnew(nnew)
  double precision eta(kg), logN(kg)

  if (kg .lt. 2) then
    logbf = 0d0
    go to 9
  end if

  ! Estimate BF
  logN = log(dble(Nout1))
  eta = logN
  select case (imeth)
  case (1)
    call revlogistic (eta,llik1,kg,Ntot1,Nout1)
  case (2)
    call revlogistic (eta,llik1,kg,Ntot1,Nout1)
    call mengwong (eta,llik1,kg,Ntot1,Nout1)
  end select
  ! eta_i = log(N_i) - log(bf_i)
  logbf = logN - eta ! = log(bf_i)
  logbf = logbf - logbf(iref)

9 if (nnew > 0) then
    logN = log(dble(Nout2))
    eta = logN - logbf
    logbfnew = logbfnew_calc(llik2,llikn,eta,Ntot2,kg,nnew)
  end if

contains
  function logbfnew_calc (llik2,llikn,eta,Ntot,kg,nnew)
    use flogsumexp, only: logrsumexp, logcsumexp
    implicit none
    integer, intent(in) :: Ntot, kg, nnew
    double precision, intent(in) :: llik2(Ntot,kg), llikn(Ntot,nnew), &
       eta(kg)
    double precision logbfnew_calc(nnew)
    double precision logb(Ntot,nnew)
    double precision lliketa(Ntot,kg), lgdenom(Ntot)
    integer i, j
    do j = 1, kg
      do i = 1, Ntot
        lliketa(i,j) = llik2(i,j) + eta(j)
      end do
    end do
    lgdenom = logrsumexp(lliketa,Ntot,kg)
    logb = spread(lgdenom,2,nnew)
    logb = llikn - logb
    logbfnew_calc = logcsumexp(logb,Ntot,nnew)
  end function logbfnew_calc
end subroutine logbfcalc


subroutine bfsecalcef (Ef, VT2, logbfnew, evec, iref, &
   logbf,llik2,llikn,fval,kg,nnew,nf,Nout2,Ntot2,nb,ibvmeth)
  use flogsumexp, only: logrsumexp, logcsumexp, logsumexpv
  implicit none
  integer, intent(in) :: kg, nnew, Nout2(kg), Ntot2, nf, ibvmeth, nb(kg), &
     iref 
  double precision, intent(in) :: logbf(kg), llik2(Ntot2,kg), &
     llikn(Ntot2,nnew), fval(Ntot2,nf)
  double precision, intent(out) :: Ef(nnew,nf), VT2(nnew,nf), &
     logbfnew(nnew), evec(kg-1,nnew,nf)
  integer i, ia, ie, j
  double precision zeta(kg), logVV(Ntot2,nnew), lliketa(Ntot2,kg), &
     logdenom(Ntot2), VV(Ntot2,nnew), WW(Ntot2,nnew,nf), NNratio(kg), &
     logweight(Ntot2,nnew), weight(Ntot2,nnew), WWwght(Ntot2,nnew,nf), &
     logbfnorm(kg), GamWW(nnew,nf), GamVW(nnew,nf), GamVV(nnew), &
     invbfnew(nnew), logV1(Ntot2,kg-1), V1(Ntot2,kg-1), cvec(kg-1,nnew), &
     bf(kg), bbn(kg-1), invbfnewsq(nnew)
  NNratio = dble(Nout2)/dble(Ntot2)
  logbfnorm = logbf - logbf(iref)
  bf = exp(logbfnorm)
  do i = 1, iref - 1
    bbn(i) = NNratio(i)/(bf(i)*bf(i)*Ntot2)
  end do
  do i = iref+1, kg
    bbn(i-1) = NNratio(i)/(bf(i)*bf(i)*Ntot2)
  end do
  zeta = -logbfnorm + log(NNratio)
  do i = 1, kg
    lliketa(:,i) = llik2(:,i) + zeta(i)
  end do
  logdenom = logrsumexp(lliketa,Ntot2,kg)
  do i = 1, iref-1
    logV1(:,i) = llik2(:,i) - logdenom
  end do
  do i = iref+1, kg
    logV1(:,i-1) = llik2(:,i) - logdenom
  end do
  V1 = exp(logV1)
  do i = 1, nnew
    logVV(:,i) = llikn(:,i) - logdenom
  end do
  VV = exp(logVV)
  logbfnew = logcsumexp(logVV,Ntot2,nnew)
  j = kg-1
  call dgemm ("t","n",j,nnew,Ntot2,1d0,V1,Ntot2,VV,Ntot2,0d0,cvec,j)
  do i = 1, nnew
    cvec(:,i) = cvec(:,i)*bbn
  end do
  do i = 1, nf
    do j = 1, nnew
      WW(:,j,i) = fval(:,i)*VV(:,j)
    end do
  end do
  do i = 1, nnew
    logweight(:,i) = logVV(:,i) - logbfnew(i)
  end do
  weight = exp(logweight) ! Importance weights
  do i = 1, nf
    do j = 1, nnew
      WWwght(:,j,i) = fval(:,i)*weight(:,j)
    end do
  end do
  Ef = sum(WWwght,1) ! Estimation of expectation
  logbfnew = logbfnew - log(dble(Ntot2))
  invbfnew = exp(-logbfnew)
  j = kg-1
  i = nnew*nf
  call dgemm ("t","n",j,i,Ntot2,1d0,V1,Ntot2,WW,Ntot2,0d0,evec,j)
  do j = 1, nf
    do i = 1, nnew
      evec(:,i,j) = evec(:,i,j)*bbn - cvec(:,i)*Ef(i,j)
      evec(:,i,j) = evec(:,i,j)*invbfnew(i)
    end do
  end do
  GamWW = 0d0
  GamVW = 0d0
  GamVV = 0d0
  ie = 0
  do i = 1, kg
    ia = ie + 1
    ie = ie + Nout2(i)
    select case (ibvmeth)
    case (1)
      call covmat_accum_bm (GamWW, GamVW, GamVV,WW(ia:ie,:,:),VV(ia:ie,:),&
         NNratio(i), &
         Nout2(i),nnew,nf,nb(i))
    case (2)
      call covmat_accum_th (GamWW, GamVW, GamVV,WW(ia:ie,:,:),VV(ia:ie,:),&
         NNratio(i), &
         Nout2(i),nnew,nf,nb(i))
    case (3)
      call covmat_accum_mb (GamWW, GamVW, GamVV,WW(ia:ie,:,:),VV(ia:ie,:),&
         NNratio(i), &
         Nout2(i),nnew,nf,nb(i))
    end select
  end do
  invbfnewsq = invbfnew*invbfnew
  do i = 1, nf
    VT2(:,i) = GamWW(:,i) - (2*Ef(:,i))*GamVW(:,i) + (Ef(:,i)*Ef(:,i))*GamVV
    VT2(:,i) = VT2(:,i)*invbfnewsq
  end do
  
contains

  subroutine covmat_accum_bm (GamWW, GamVW, GamVV, WW, VV, NNratio, &
     Nout,nnew,nf,nb)
    implicit none
    integer, intent(in) :: Nout, nnew, nf, nb
    double precision, intent(in) :: WW(Nout,nnew,nf), VV(Nout,nnew), &
       NNratio
    double precision, intent(inout) :: GamWW(nnew,nf), GamVW(nnew,nf), &
       GamVV(nnew)
    integer i, j, k, l, batch_size(nb), ia, ie
    double precision Wb(nb,nnew,nf), Vb(nb,nnew), Wm(nnew,nf), Vm(nnew), &
       Wbm(nb,nnew,nf), Vbm(nb,nnew), s, wgt
    batch_size = Nout/nb
    i = mod(Nout,nb)
    batch_size(:i) = batch_size(:i) + 1
    do l = 1, nf
      do k = 1, nnew
        Wm(k,l) = 0d0
        ie = 0
        do i = 1, nb
          ia = ie + 1
          ie = ie + batch_size(i)
          s = 0d0
          do j = ia, ie
            s = s + WW(j,k,l)
          end do
          Wm(k,l) = Wm(k,l) + s
          Wb(i,k,l) = s/dble(batch_size(i))
        end do
        Wm(k,l) = Wm(k,l)/dble(Nout)
        Wbm(:,k,l) = Wb(:,k,l) - Wm(k,l)
      end do
    end do
    do k = 1, nnew
      Vm(k) = 0d0
      ie = 0
      do i = 1, nb
        ia = ie + 1
        ie = ie + batch_size(i)
        s = 0d0
        do j = ia, ie
          s = s + VV(j,k)
        end do
        Vm(k) = Vm(k) + s
        Vb(i,k) = s/dble(batch_size(i))
      end do
      Vm(k) = Vm(k)/dble(Nout)
      Vbm(:,k) = Vb(:,k) - Vm(k)
    end do
    do l = 1, nf
      do k = 1, nnew
        do i = 1, nb
          wgt = NNratio*dble(batch_size(i))/dble(nb-1)
          GamWW(k,l) = GamWW(k,l) + wgt*Wbm(i,k,l)*Wbm(i,k,l)
        end do
      end do
    end do
    do k = 1, nnew
      do i = 1, nb
        wgt = NNratio*dble(batch_size(i))/dble(nb-1)
        GamVV(k) = GamVV(k) + wgt*Vbm(i,k)*Vbm(i,k)
      end do
    end do
    do l = 1, nf
      do k = 1, nnew
        do i = 1, nb
          wgt = NNratio*dble(batch_size(i))/dble(nb-1)
          GamVW(k,l) = GamVW(k,l) + wgt*Vbm(i,k)*Wbm(i,k,l)
        end do
      end do
    end do
  end subroutine covmat_accum_bm
  
  subroutine covmat_accum_th (GamWW, GamVW, GamVV, WW, VV, NNratio, &
     Nout,nnew,nf,nb)
    implicit none
    integer, intent(in) :: Nout, nnew, nf, nb
    double precision, intent(in) :: WW(Nout,nnew,nf), VV(Nout,nnew), &
       NNratio
    double precision, intent(inout) :: GamWW(nnew,nf), GamVW(nnew,nf), &
       GamVV(nnew)
    integer i, j, k, l
    double precision wgt, Wm(nnew,nf), Vm(nnew), Wmm(Nout,nnew,nf), &
       Vmm(Nout,nnew)
    double precision, parameter :: pi = 3.1415926535897932384626433832795d0 
    Vm = sum(VV,1)/dble(Nout)
    Wm = sum(WW,1)/dble(Nout)
    do k = 1, nf
      do j = 1, nnew
        do i = 1, Nout
          Wmm(i,j,k) = WW(i,j,k) - Wm(j,k)
        end do
      end do
    end do
    do j = 1, nnew
      do i = 1, Nout
        Vmm(i,j) = VV(i,j) - Vm(j)
      end do
    end do
    ! j = 0
    wgt = NNratio/dble(Nout)
    do l = 1, nf
      do k = 1, nnew
        do i = 1, Nout
          GamWW(k,l) = GamWW(k,l) + wgt*Wmm(i,k,l)*Wmm(i,k,l)
        end do
      end do
    end do
    do k = 1, nnew
      do i = 1, Nout
        GamVV(k) = GamVV(k) + wgt*Vmm(i,k)*Vmm(i,k)
      end do
    end do
    do l = 1, nf
      do k = 1, nnew
        do i = 1, Nout
          GamVW(k,l) = GamVW(k,l) + wgt*Vmm(i,k)*Wmm(i,k,l)
        end do
      end do
    end do
    do j = 1, nb - 1
      wgt = NNratio*(1d0 + cos((pi*j)/nb))/Nout ! w/o the .5 factor because it
      ! cancels
      do l = 1, nf
        do k = 1, nnew
          do i = 1, Nout-j
            GamWW(k,l) = GamWW(k,l) + wgt*Wmm(i,k,l)*Wmm(i+j,k,l)
          end do
        end do
      end do
      do k = 1, nnew
        do i = 1, Nout-j
          GamVV(k) = GamVV(k) + wgt*Vmm(i,k)*Vmm(i+j,k)
        end do
      end do
      wgt = .5d0*wgt
      do l = 1, nf
        do k = 1, nnew
          do i = 1, Nout-j
            GamVW(k,l) = GamVW(k,l) + wgt*(Vmm(i,k)*Wmm(i+j,k,l) &
               + Vmm(i+j,k)*Wmm(i,k,l))
          end do
        end do
      end do
    end do
  end subroutine covmat_accum_th

  subroutine covmat_accum_mb (GamWW, GamVW, GamVV, WW, VV, NNratio, &
     Nout,nnew,nf,nb)
    implicit none
    integer, intent(in) :: Nout, nnew, nf, nb
    double precision, intent(in) :: WW(Nout,nnew,nf), VV(Nout,nnew), &
       NNratio
    double precision, intent(inout) :: GamWW(nnew,nf), GamVW(nnew,nf), &
       GamVV(nnew)
    integer i, j, k, l
    double precision wgt, Wm(nnew,nf), Vm(nnew), Wmm(Nout,nnew,nf), &
       Vmm(Nout,nnew)
    Vm = sum(VV,1)/dble(Nout)
    Wm = sum(WW,1)/dble(Nout)
    do k = 1, nf
      do j = 1, nnew
        do i = 1, Nout
          Wmm(i,j,k) = WW(i,j,k) - Wm(j,k)
        end do
      end do
    end do
    do j = 1, nnew
      do i = 1, Nout
        Vmm(i,j) = VV(i,j) - Vm(j)
      end do
    end do
    ! j = 0
    wgt = NNratio/dble(Nout)
    do l = 1, nf
      do k = 1, nnew
        do i = 1, Nout
          GamWW(k,l) = GamWW(k,l) + wgt*Wmm(i,k,l)*Wmm(i,k,l)
        end do
      end do
    end do
    do k = 1, nnew
      do i = 1, Nout
        GamVV(k) = GamVV(k) + wgt*Vmm(i,k)*Vmm(i,k)
      end do
    end do
    do l = 1, nf
      do k = 1, nnew
        do i = 1, Nout
          GamVW(k,l) = GamVW(k,l) + wgt*Vmm(i,k)*Wmm(i,k,l)
        end do
      end do
    end do
    do j = 1, nb - 1
      wgt = 2*NNratio*(1d0 - dble(j)/nb)/Nout ! w/ a factor because only
      ! one term is computed
      do l = 1, nf
        do k = 1, nnew
          do i = 1, Nout-j
            GamWW(k,l) = GamWW(k,l) + wgt*Wmm(i,k,l)*Wmm(i+j,k,l)
          end do
        end do
      end do
      do k = 1, nnew
        do i = 1, Nout-j
          GamVV(k) = GamVV(k) + wgt*Vmm(i,k)*Vmm(i+j,k)
        end do
      end do
      wgt = .5d0*wgt
      do l = 1, nf
        do k = 1, nnew
          do i = 1, Nout-j
            GamVW(k,l) = GamVW(k,l) + wgt*(Vmm(i,k)*Wmm(i+j,k,l) &
               + Vmm(i+j,k)*Wmm(i,k,l))
          end do
        end do
      end do
    end do
  end subroutine covmat_accum_mb
  
  
  double precision function mean(x, m)
    use interfaces, only: isfinite
    implicit none
    integer, intent(in) :: m
    double precision, intent(in) :: x(m)
    integer i
    double precision tot, tot0, mm
    mean = 0d0
    tot = 0d0
    mm = dble(m)
    do i = 1, m
      tot0 = tot
      tot = tot + x(i)
      if (isfinite(tot) .eq. 0) then
        mean = mean + tot0/mm
        tot = x(i)
      end if
    end do
    mean = mean + tot/mm
  end function mean
end subroutine bfsecalcef

##' Rhizoctonia root rot infections.
##'
##' A dataset containing the number of infected roots and the sample
##' coordinate. The data were collected by Dr Jim Cook at Washington
##' State University. 
##'
##' \itemize{
##' \item Xcoord Longitude of the sampling location.
##' \item Ycoord Latitude of the sampling location.
##' \item Total Total number of roots sampled at that location.
##' \item Infected Number of infected roots found at that location.
##' \item Yield Barley yield at that location. These data were
##' obtained by hand-harvesting a 4-square-meter area in the sampling
##' location. One observation is missing.
##' }
##' @title Rhizoctonia root rot infections
##' @docType data
##' @format A data frame with 100 rows and 5 variables.
##' @source See reference below.
##' @name rhizoctonia
##' @usage data(rhizoctonia)
##' @references Zhang, H. (2002). On estimation and prediction for
##' spatial generalized linear mixed models. \emph{Biometrics}, 58(1),
##' 129-136.
##' @note We acknowledge Hao Zhang for providing these data.
NULL

###############################################################################
#### Code for the paper "Selection of Proposal Distributions for
#### Multiple Importance Sampling" by V Roy and E Evangelou
###############################################################################

##' Selection of multiple importance sampling distributions
##'
##' \describe{
##' \item{SEQ}{is a sequential method starting with \code{istart} and
##' additng to it until \code{K} proposals have been selected. At each
##' iteration, the point with the highest (relative?) standard error
##' is added} 
##' \item{MNX}{is the minimax method. The chosen proposal corresponds
##' to the lowest maximum (relative?) standard error.}
##' \item{ENT}{is the entropy method. The chosen proposal corresponds
##' to the highest determinant of the (relative?) covariance matrix at
##' the first stage.}
##' }
##' @title Selection of multiple importance sampling distributions
##' @param pargrid A data frame with components "linkp", "phi", "omg",
##'   "kappa". Each row gives a combination of the parameters to
##'   compute the new standard errors.
##' @param K How many proposal densities in total to choose among the
##'   rows of \code{pargrid}. Needed for SEQ only. For MNX and ENT
##'   this is determined by the length of \code{istart}. 
##' @param istart Start with these rows of \code{pargrid}. A vector of
##'   indices.
##' @param relativeSE Logical. Whether the choice is based on the
##'   standard error (FALSE), or relative standard error (TRUE).
##' @param N1 The sample size for stage 1.
##' @param N2 The sample sie for stage 2.
##' @param Nthin Thinning
##' @param Nbi Burn-in
##' @param formula A representation of the model in the form
##'   \code{response ~ terms}. The response must be set to \code{NA}'s
##'   at the prediction locations (see the examples on how to do this
##'   using the function \code{\link{stackdata}}). At the observed
##'   locations the response is assumed to be a total of replicated
##'   measurements. The number of replications is inputted using the
##'   argument \code{weights}.
##' @param family The distribution of the data. The
##'   \code{"GEVbinomial"} family is the binomial family with link the
##'   GEV link (see Details).
##' @param data An optional data frame containing the variables in the
##'   model.
##' @param weights An optional vector of weights. Number of replicated
##'   samples for Gaussian and gamma, number of trials for binomial,
##'   time length for Poisson.
##' @param subset An optional vector specifying a subset of
##'   observations to be used in the fitting process.
##' @param offset See \code{\link[stats]{lm}}.
##' @param atsample A formula in the form \code{~ x1 + x2 + ... + xd}
##'   with the coordinates of the sampled locations.
##' @param corrfcn Spatial correlation function. See
##'   \code{\link{geoBayes_correlation}} for details.
##' @param betm0 Prior mean for beta (a vector or scalar).
##' @param betQ0 Prior standardised precision (inverse variance)
##'   matrix. Can be a scalar, vector or matrix. The first two imply a
##'   diagonal with those elements. Set this to 0 to indicate a flat
##'   improper prior.
##' @param ssqdf Degrees of freedom for the scaled inverse chi-square
##'   prior for the partial sill parameter.
##' @param ssqsc Scale for the scaled inverse chi-square prior for the
##'   partial sill parameter.
##' @param dispersion The fixed dispersion parameter.
##' @param longlat How to compute the distance between locations. If
##'   \code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
##'   distance. See \code{\link[sp]{spDists}}.
##' @param nbatch1 A scalar or vector of the same length as
##'   \code{runs}. All values must be integers or less than 1. This is
##'   used for calculating how many batches to split each of the
##'   sample in runs for the calculation of the Bayes factors standard
##'   errors for the parameters corresponding to \code{runs}.
##' @param nbatch2 A scalar or vector of the same length as
##'   \code{runs}. All values must be integers or less than 1. This is
##'   used for calculating how many batches to split each of the
##'   sample in runs for the calculation of the Bayes factors standard
##'   errors for the parameters corresponding to \code{pargrid}.
##' @param S1method Which method to use to calculate the Bayes
##'   factors: Reverse logistic or Meng-Wong.
##' @param bvmethod Which method to use for the calculation of the
##'   batch variance. The standard method splits to disjoint batches.
##'   The second and third method use the spectral variance method
##'   with different lag windows.
##' @param transf Whether to use a transformed sample for the
##'   computations. If \code{"no"} or \code{FALSE}, it doesn't. If
##'   \code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
##'   \code{"wo"} it uses an alternative transformation. The latter
##'   can be used only for the families indicated by
##'   \code{.geoBayes_models$haswo}.
##' @return A list with components
##' \describe{
##' \item{selected}{The rows of \code{pargrid} selected.}
##' \item{isel}{The indices of the rows of \code{pargrid} selected.}
##' \item{se}{The standard error corresponding to the selected
##'   parameters.}
##' \item{samples}{A list containing the samples from the selected
##'   parameters.} 
##' }
##' @name select_proposals
##' @references Roy, V., & Evangelou, E. (2018). Selection of proposal
##'   distributions for generalized importance sampling estimators.
##'   arXiv preprint arXiv:1805.00829. 
##' @export
##' @examples
##' \dontrun{
##' data(rhizoctonia)
##' ### Define the model
##' corrf <- "spherical"
##' kappa <- 0
##' ssqdf <- 1
##' ssqsc <- 1
##' betm0 <- 0
##' betQ0 <- .01
##' family <- "binomial.probit"
##' formula <- Infected ~ 1
##' atsample <- ~ Xcoord + Ycoord
##' ### Skeleton points
##' philist <- seq(100, 200, 10)
##' omglist <- 0
##' parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
##' ### MCMC sizes
##' Nout <- 100
##' Nthin <- 1
##' Nbi <- 10
##' ## Select proposals
##' K <- 3                        # Choose 3 proposals
##' istart_SEQ <- 6               # Start with middle
##' istart_MNX <- istart_ENT <- c(6, 2, 10)
##' cooling_MNX <- .05/log((0:24 %/% 5)*5 + exp(1))
##' cooling_ENT <- .3/log((0:49 %/% 10)*10 + exp(1))
##' prop_SEQ <- select_proposals_SEQ(pargrid = parlist, K = K,
##'                                  istart = istart_SEQ,
##'                                  relativeSE = TRUE, 
##'                                  N1 = Nout, N2 = Nout,
##'                                  Nthin = Nthin, Nbi = Nbi,
##'                                  formula = formula, family = family,
##'                                  data = rhizoctonia, weights = Total, 
##'                                  atsample = atsample, corrfcn = corrf,
##'                                  betm0 = betm0, betQ0 = betQ0,
##'                                  ssqdf = ssqdf, ssqsc = ssqsc,
##'                                  dispersion = 1, longlat = FALSE,
##'                                  nbatch1 = 0.5, nbatch2 = 0.5,
##'                                  bvmethod = "TukeyHanning",
##'                                  transf = "mu")
##' prop_MNX <- select_proposals_MNX(pargrid = parlist,
##'                                  istart = istart_MNX, nfix = 1L,
##'                                  cooling = cooling_MNX, 
##'                                  relativeSE = TRUE, 
##'                                  N1 = Nout, N2 = Nout,
##'                                  Nthin = Nthin, Nbi = Nbi,
##'                                  formula = formula, family = family,
##'                                  data = rhizoctonia, weights = Total, 
##'                                  atsample = atsample, corrfcn = corrf,
##'                                  betm0 = betm0, betQ0 = betQ0,
##'                                  ssqdf = ssqdf, ssqsc = ssqsc,
##'                                  dispersion = 1, longlat = FALSE,
##'                                  nbatch1 = 0.5, nbatch2 = 0.5,
##'                                  bvmethod = "TukeyHanning",
##'                                  transf = "mu",
##'                                  verbose = TRUE)
##' prop_ENT <- select_proposals_ENT(pargrid = parlist,
##'                                  istart = istart_ENT, nfix = 1L,
##'                                  cooling = cooling_ENT, 
##'                                  relativeSE = TRUE, 
##'                                  N1 = Nout, 
##'                                  Nthin = Nthin, Nbi = Nbi,
##'                                  formula = formula, family = family,
##'                                  data = rhizoctonia, weights = Total, 
##'                                  atsample = atsample, corrfcn = corrf,
##'                                  betm0 = betm0, betQ0 = betQ0,
##'                                  ssqdf = ssqdf, ssqsc = ssqsc,
##'                                  dispersion = 1, longlat = FALSE,
##'                                  nbatch1 = 0.5, nbatch2 = 0.5,
##'                                  bvmethod = "TukeyHanning",
##'                                  transf = "mu",
##'                                  verbose = TRUE)
##' }
select_proposals_SEQ <- function (pargrid, K, istart,
                                  relativeSE = FALSE, 
                                  N1, N2, Nthin, Nbi,
                                  formula, family = "gaussian",
                                  data, weights, subset, offset, 
                                  atsample, corrfcn = "matern",
                                  betm0, betQ0, ssqdf, ssqsc,
                                  dispersion = 1, longlat = FALSE,
                                  nbatch1 = 0.5, nbatch2 = 0.5,
                                  S1method = c("RL", "MW"),
                                  bvmethod = c("Standard", "TukeyHanning",
                                               "Bartlett"),
                                  transf = c("no", "mu", "wo"))
{
  cl <- match.call()
  pargrid <- check_pargrid(pargrid, family, corrfcn)
  partry <- pargrid[, c("nu", "phi", "omg", "kappa")]
  ngrid <- nrow(pargrid)
  istart <- unique(as.integer(istart))
  if (any(istart < 1L) || any(istart) > ngrid) {
    stop ("istart elements must be between 1 and nrow(pargrid).")
  }
  K <- as.integer(K)
  if (K <= 0L) stop("K must be a positive integer.")
  if (K > ngrid) stop("K cannot be more than the number of elements in pargrid.")
  N1 <- as.integer(N1)
  N2 <- as.integer(N2)
  Nout <- c(N1, N2)
  samples <- list()
  length(samples) <- K
  args_mcsglmm <- names(formals(mcsglmm))
  args_mcsglmm <- args_mcsglmm[args_mcsglmm %in% names(cl)[-1]]
  clsample <- cl[c("", args_mcsglmm)]
  clsample[[1]] <- quote(mcsglmm)
  clsample$corrtuning <- list(phi = 0, omg = 0, kappa = 0)
  clsample$Nout <- Nout
  clsample$test <- FALSE
  samplefun <- function(call, pars) {
    call$linkp <- pars[[1]]
    call$phi <- pars[[2]]
    call$omg <- pars[[3]]
    call$kappa <- pars[[4]]
    eval(call)
  }
  nstart <- length(istart)
  if (nstart > K) stop("Need at most K indices in istart.")
  iin <- integer(K)
  iin[1:nstart] <- istart
  for (i in seq_len(nstart)) {
    samples[[i]] <- samplefun(clsample, partry[iin[i], ])
  }
  se <- bmbfse(pargrid = pargrid,
               runs = samples[1L:nstart],
               bfsize1 = Nout[1L],
               nbatch1 = nbatch1,
               nbatch2 = nbatch2,
               S1method = S1method,
               bvmethod = bvmethod,
               transf = transf)
  relativeSE <- as.logical(relativeSE)
  if (isTRUE(relativeSE)) {
    objfun <- function(s) s$pargrid$SE*exp(-s$pargrid$logbf)
  } else {
    objfun <- function(s) s$pargrid$SE
  }
  if (K > nstart) {
    for (i in (nstart+1L):K) {
      sevals <- objfun(se)
      sevals[iin[1L:(i-1L)]] <- NA
      iin[i] <- which.max(sevals)
      samples[[i]] <- samplefun(clsample, partry[iin[i], ])
      se <- bmbfse(pargrid = pargrid,
                   runs = samples[1L:i],
                   bfsize1 = Nout[1L],
                   nbatch1 = nbatch1,
                   nbatch2 = nbatch2,
                   S1method = S1method,
                   bvmethod = bvmethod,
                   transf = transf)
    }
  }
  list(selected = pargrid[iin, ], isel = iin, se = se,
       samples = samples)
}

##' @param cooling A decreasing sequence of temperature values for the
##'   simulated annealing. All elements must be positive. A suggested
##'   value is \code{Tinit / log(((0:N) \%/\% Tstp)*Tstp + exp(1))} for
##'   \code{N+1} iterations, where \code{Tinit} is the initial
##'   temperature and \code{Tstp} is the number of iterations before
##'   the temperature is reduced.
##' @param nfix In the case of MNX and ENT, the first \code{nfix}
##'   elements of \code{istart} will always be included.
##' @param verbose Logical. Prints information about the simulated
##'   annealing. 
##' @name select_proposals
##' @export
##' @importFrom stats runif
select_proposals_MNX <- function (pargrid, istart, nfix, 
                                  relativeSE = FALSE, 
                                  N1, N2, Nthin, Nbi,
                                  cooling, 
                                  formula, family = "gaussian",
                                  data, weights, subset, offset, 
                                  atsample, corrfcn = "matern",
                                  betm0, betQ0, ssqdf, ssqsc,
                                  dispersion = 1, longlat = FALSE,
                                  nbatch1 = 0.5, nbatch2 = 0.5,
                                  S1method = c("RL", "MW"),
                                  bvmethod = c("Standard", "TukeyHanning",
                                               "Bartlett"),
                                  transf = c("no", "mu", "wo"),
                                  verbose = FALSE)
{
  cl <- match.call()
  verbose <- isTRUE(as.logical(verbose))
  pargrid <- check_pargrid(pargrid, family, corrfcn)
  partry <- pargrid[, c("nu", "phi", "omg", "kappa")]
  ngrid <- nrow(pargrid)
  istart <- unique(as.integer(istart))
  if (any(istart < 1L) || any(istart) > ngrid) {
    stop ("istart elements must be between 1 and nrow(pargrid).")
  }
  K <- length(istart)
  if (K <= 0L) stop("K must be a positive integer.")
  if (K > ngrid) stop("K cannot be more than the number of elements in pargrid.")
  nfix <- as.integer(nfix)
  N1 <- as.integer(N1)
  N2 <- as.integer(N2)
  Nout <- c(N1, N2)
  samples <- list()
  length(samples) <- ngrid
  cooling <- as.numeric(cooling)
  niter <- length(cooling)
  if (niter && !all(cooling > 0)) {
    stop ("Input cooling must contain only positive values.")
  }
  if (niter > 1L && any(cooling[2L:niter] > cooling[1L:(niter-1L)])) {
    stop ("Input cooling must be a decreasing sequence.")
  }
  args_mcsglmm <- names(formals(mcsglmm))
  args_mcsglmm <- args_mcsglmm[args_mcsglmm %in% names(cl)[-1]]
  clsample <- cl[c("", args_mcsglmm)]
  clsample[[1]] <- quote(mcsglmm)
  clsample$corrtuning <- list(phi = 0, omg = 0, kappa = 0)
  clsample$Nout <- Nout
  clsample$test <- FALSE
  samplefun <- function(call, pars) {
    call$linkp <- pars[[1]]
    call$phi <- pars[[2]]
    call$omg <- pars[[3]]
    call$kappa <- pars[[4]]
    eval(call)
  }
  for (i in seq_len(K)) {
    samples[[istart[i]]] <- samplefun(clsample, partry[istart[i], ])
  }
  se <- bmbfse(pargrid = pargrid,
               runs = samples[istart],
               bfsize1 = Nout[1L],
               nbatch1 = nbatch1,
               nbatch2 = nbatch2,
               S1method = S1method,
               bvmethod = bvmethod,
               transf = transf)
  relativeSE <- as.logical(relativeSE)
  if (isTRUE(relativeSE)) {
    objfun <- function(s) -max(s$pargrid$SE*exp(-s$pargrid$logbf))
  } else {
    objfun <- function(s) -max(s$pargrid$SE)
  }
  id_bst <- id_cur <- istart
  se_bst <- se
  s_bst <- s_cur <- objfun(se)
  lsam <- K - nfix
  ltry <- ngrid - K
  set <- seq_len(ngrid)
  ## Simulated annealing
  if (verbose) {
    cat("     Iter   Objective", "\n")
    cat("---------   ---------", "\n")
    cat(format(0L, width = 9), "   ", format(s_cur, width = 9), "\n")
  }
  for (i in seq_len(niter)) {
    set_new <- set[-id_cur]
    j <- nfix+sample.int(lsam, 1L)
    k <- set_new[sample.int(ltry, 1L)]
    id_new <- id_cur
    id_new[j] <- k
    if (is.null(samples[[k]])) {
      samples[[k]] <- samplefun(clsample, partry[k, ])
    }
    se <- bmbfse(pargrid = pargrid,
                 runs = samples[id_new],
                 bfsize1 = Nout[1L],
                 nbatch1 = nbatch1,
                 nbatch2 = nbatch2,
                 S1method = S1method,
                 bvmethod = bvmethod,
                 transf = transf)
    s_new <- objfun(se)
    accept <- runif(1L) < exp((s_new - s_cur)/cooling[i])
    if (accept) {
      id_cur <- id_new
      s_cur <- s_new
      if (s_new > s_bst) {
        id_bst <- id_new
        s_bst <- s_new
        se_bst <- se
      }
    }
    if(verbose) {
      cat(format(i, width = 9), "   ", format(s_cur, width = 9), "\n")
    }
  }
  list(selected = pargrid[id_bst, ], isel = id_bst, bmbfse = se_bst,
       samples = samples[id_bst])
}


##' @name select_proposals
##' @export 
select_proposals_ENT <- function (pargrid, istart, nfix, 
                                  relativeSE = FALSE, 
                                  N1, Nthin, Nbi,
                                  cooling, 
                                  formula, family = "gaussian",
                                  data, weights, subset, offset, 
                                  atsample, corrfcn = "matern",
                                  betm0, betQ0, ssqdf, ssqsc,
                                  dispersion = 1, longlat = FALSE,
                                  nbatch1 = 0.5, nbatch2 = 0.5,
                                  S1method = c("RL", "MW"),
                                  bvmethod = c("Standard", "TukeyHanning",
                                               "Bartlett"),
                                  transf = c("no", "mu", "wo"),
                                  verbose = FALSE)
{
  cl <- match.call()
  verbose <- isTRUE(as.logical(verbose))
  pargrid <- check_pargrid(pargrid, family, corrfcn)
  partry <- pargrid[, c("nu", "phi", "omg", "kappa")]
  ngrid <- nrow(pargrid)
  istart <- unique(as.integer(istart))
  if (any(istart < 1L) || any(istart) > ngrid) {
    stop ("istart elements must be between 1 and nrow(pargrid).")
  }
  K <- length(istart)
  if (K <= 0L) stop("K must be a positive integer.")
  if (K > ngrid) stop("K cannot be more than the number of elements in pargrid.")
  nfix <- as.integer(nfix)
  N1 <- as.integer(N1)
  N2 <- 0L
  Nout <- c(N1, N2)
  samples <- list()
  length(samples) <- ngrid
  cooling <- as.numeric(cooling)
  niter <- length(cooling)
  if (niter && !all(cooling > 0)) {
    stop ("Input cooling must contain only positive values.")
  }
  if (niter > 1L && any(cooling[2L:niter] > cooling[1L:(niter-1L)])) {
    stop ("Input cooling must be a decreasing sequence.")
  }
  args_mcsglmm <- names(formals(mcsglmm))
  args_mcsglmm <- args_mcsglmm[args_mcsglmm %in% names(cl)[-1]]
  clsample <- cl[c("", args_mcsglmm)]
  clsample[[1]] <- quote(mcsglmm)
  clsample$corrtuning <- list(phi = 0, omg = 0, kappa = 0)
  clsample$Nout <- Nout
  clsample$test <- FALSE
  samplefun <- function(call, pars) {
    call$linkp <- pars[[1]]
    call$phi <- pars[[2]]
    call$omg <- pars[[3]]
    call$kappa <- pars[[4]]
    eval(call)
  }
  for (i in seq_len(K)) {
    samples[[istart[i]]] <- samplefun(clsample, partry[istart[i], ])
  }
  se <- bmbfse(pargrid = pargrid,
               runs = samples[istart],
               bfsize1 = Nout[1L],
               nbatch1 = nbatch1,
               nbatch2 = nbatch2,
               S1method = S1method,
               bvmethod = bvmethod,
               transf = transf)
  relativeSE <- as.logical(relativeSE)
  if (isTRUE(relativeSE)) {
    objfun <- function(s)
      sum(log(diag(chol(s$bfSigma/tcrossprod(s$bfEstimate[-1])))))
  } else {
    objfun <- function(s) sum(log(diag(chol(s$bfSigma))))
  }
  id_bst <- id_cur <- istart
  se_bst <- se
  s_bst <- s_cur <- objfun(se)
  lsam <- K - nfix
  ltry <- ngrid - K
  set <- seq_len(ngrid)
  ## Simulated annealing
  if (verbose) {
    cat("     Iter   Objective", "\n")
    cat("---------   ---------", "\n")
    cat(format(0L, width = 9), "   ", format(s_cur, width = 9), "\n")
  }
  for (i in seq_len(niter)) {
    set_new <- set[-id_cur]
    j <- nfix+sample.int(lsam, 1L)
    k <- set_new[sample.int(ltry, 1L)]
    id_new <- id_cur
    id_new[j] <- k
    if (is.null(samples[[k]])) {
      samples[[k]] <- samplefun(clsample, partry[k, ])
    }
    se <- bmbfse(pargrid = pargrid,
                 runs = samples[id_new],
                 bfsize1 = Nout[1L],
                 nbatch1 = nbatch1,
                 nbatch2 = nbatch2,
                 S1method = S1method,
                 bvmethod = bvmethod,
                 transf = transf)
    s_new <- objfun(se)
    accept <- runif(1L) < exp((s_new - s_cur)/cooling[i])
    if (accept) {
      id_cur <- id_new
      s_cur <- s_new
      if (s_new > s_bst) {
        id_bst <- id_new
        s_bst <- s_new
        se_bst <- se
      }
    }
    if(verbose) {
      cat(format(i, width = 9), "   ", format(s_cur, width = 9), "\n")
    }
  }
  list(selected = pargrid[id_bst, ], isel = id_bst, bmbfse = se_bst,
       samples = samples[id_bst])
}

module msgmc
  interface
    subroutine msgmci (i, a)
      integer i, a
    end subroutine msgmci
  end interface

  interface
    subroutine msgmca
    end subroutine msgmca
  end interface

  interface
    subroutine msgmce (a)
      integer a
    end subroutine msgmce
  end interface

  interface
    subroutine msgmcl
    end subroutine msgmcl
  end interface

  interface
    subroutine msgmci2 (i, a, b)
      integer i, a, b
    end subroutine msgmci2
  end interface

  interface
    subroutine msgmca2
    end subroutine msgmca2
  end interface

  interface
    subroutine msgmce2 (a, b)
      integer a, b
    end subroutine msgmce2
  end interface

  interface
    subroutine msgmcl2
    end subroutine msgmcl2
  end interface
end module msgmc

### skelpnts.R ---
##
## Created: Wed, 1 Jun, 2016 14:05 (BST)
## Last-Updated: Sat, 4 Jun, 2016 15:29 (BST)
##
######################################################################
##
### Commentary: This functions are used for deriving appropriate
### skeleton points.
##
######################################################################


##' Log-likelihood approximation.
##'
##' Computes and approximation to the log-likelihood for the given
##' parameters using integrated nested Laplace approximations.
##' @title Log-likelihood approximation
##' @param par_vals A data frame with the components "linkp", "phi",
##'   "omg", "kappa". The approximation will be computed at each row
##'   of the data frame.
##' @param formula A representation of the model in the form
##'   \code{response ~ terms}.
##' @param family The distribution of the response. Can be one of the
##'   options in \code{\link{.geoBayes_models}} or
##'   \code{"transformed.gaussian"}.
##' @param data An optional data frame containing the variables in the
##'   model.
##' @param weights An optional vector of weights. Number of replicated
##'   samples for Gaussian and gamma, number of trials for binomial,
##'   time length for Poisson.
##' @param subset An optional vector specifying a subset of
##'   observations to be used in the fitting process.
##' @param offset See \code{\link[stats]{lm}}.
##' @param atsample A formula in the form \code{~ x1 + x2 + ... + xd}
##'   with the coordinates of the sampled locations.
##' @param corrfcn Spatial correlation function. Can be one of the
##'   choices in \code{\link{.geoBayes_corrfcn}}.
##' @param np The number of integration points for the spatial
##'   variance parameter sigma^2. The total number of points will be
##'   \code{2*np + 1}.
##' @param betm0 Prior mean for beta (a vector or scalar).
##' @param betQ0 Prior standardised precision (inverse variance)
##'   matrix. Can be a scalar, vector or matrix. The first two imply a
##'   diagonal with those elements. Set this to 0 to indicate a flat
##'   improper prior.
##' @param ssqdf Degrees of freedom for the scaled inverse chi-square
##'   prior for the partial sill parameter.
##' @param ssqsc Scale for the scaled inverse chi-square prior for the
##'   partial sill parameter.
##' @param tsqdf Degrees of freedom for the scaled inverse chi-square
##'   prior for the measurement error parameter.
##' @param tsqsc Scale for the scaled inverse chi-square prior for the
##'   measurement error parameter.
##' @param dispersion The fixed dispersion parameter.
##' @param longlat How to compute the distance between locations. If
##'   \code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
##'   distance. See \code{\link[sp]{spDists}}.
##' @return A list with components
##' \itemize{
##' \item \code{par_vals} A data frame of the parameter values.
##' \item \code{aloglik} The approximate log-likelihood at thos
##'   parameter values.
##' }
##' @useDynLib geoBayes llikparscalc
##' @importFrom stats model.offset
##' @export
##' @references Evangelou, E., & Roy, V. (2019). Estimation and prediction for
##'   spatial generalized linear mixed models with parametric links
##'   via reparameterized importance sampling. Spatial Statistics, 29,
##'   289-315.
alik_inla <- function (par_vals, formula,
                       family = "gaussian",
                       data, weights, subset, offset, atsample,
                       corrfcn = "matern",
                       np, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsqsc,
                       dispersion = 1, longlat = FALSE)
{
  ## Family
  ifam <- .geoBayes_family(family)
  if (ifam) {
    family <- .geoBayes_models$family[ifam]
  } else {
    stop ("This family has not been implemented.")
  }

  ## Logical input
  longlat <- as.logical(longlat)

  ## Correlation function and parameters
  icf <- .geoBayes_correlation(corrfcn)
  corrfcn <- .geoBayes_corrfcn$corrfcn[icf]
  if (!.geoBayes_corrfcn$needkappa[icf]) par_vals$kappa <- 0
  kappa <- .geoBayes_getkappa(par_vals$kappa, icf)
  phi <- as.double(par_vals$phi)
  if (any(phi < 0)) stop ("Input phi must >= 0")
  omg <- as.double(par_vals$omg)
  if (any(omg < 0)) stop ("Input omg must >= 0")

  ## Check the link parameter
  if (!.geoBayes_models$needlinkp[ifam]) par_vals$linkp <- 0
  nu <- .geoBayes_getlinkp(par_vals$linkp, ifam)

  ## All parameters
  allpars <- data.frame(nu = nu, phi = phi, omg = omg, kappa = kappa)
  nu <- as.double(allpars[["nu"]])
  phi <- as.double(allpars[["phi"]])
  omg <- as.double(allpars[["omg"]])
  kappa <- as.double(allpars[["kappa"]])
  npars <- nrow(allpars)

  ## Check dispersion input
  tsq <- as.double(dispersion)
  if (tsq <= 0) stop ("Invalid argument dispersion")
  tsqdf <- 0

  ## Design matrix and data
  if (missing(data)) data <- environment(formula)
  if (length(formula) != 3) stop ("The formula input is incomplete.")
  if ("|" == all.names(formula[[2]], TRUE, 1)) formula[[2]] <- formula[[2]][[2]]
  mfc <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data", "subset", "weights", "offset"),
             names(mfc), 0L)
  mfc <- mfc[c(1L, m)]
  mfc$formula <- formula
  mfc$drop.unused.levels <- TRUE
  mfc$na.action <- "na.pass"
  mfc[[1L]] <- quote(stats::model.frame)
  mf <- eval(mfc, parent.frame())
  mt <- attr(mf, "terms")
  FF <- model.matrix(mt,mf)
  if (!all(is.finite(FF))) stop ("Non-finite values in the design matrix")
  p <- NCOL(FF)
  yy <- unclass(model.response(mf))
  if (!is.vector(yy)) {
    stop ("The response must be a vector")
  }
  yy <- as.double(yy)
  ll <- as.double(model.weights(mf))
  oofset <- as.vector(model.offset(mf))
  if (!is.null(oofset)) {
    if (length(oofset) != NROW(yy)) {
      stop(gettextf("number of offsets is %d, should equal %d (number of observations)", 
                    length(oofset), NROW(yy)), domain = NA)
    } else {
      oofset <- as.double(oofset)
    }
  } else {
    oofset <- double(NROW(yy))
  }

  ## All locations
  atsample <- update(atsample, NULL ~ .)
  locvars <- all.vars(atsample)
  formula1 <- as.formula(paste('~', paste(c(locvars, all.vars(formula)),
                                          collapse = ' + ')))
  mfc1 <- mfc
  mfc1$formula <- formula1
  mf1 <- eval(mfc1, parent.frame())
  m <- match(locvars, names(mf1))
  loc <- as.matrix(mf1[, m])
  if (!all(is.finite(loc))) stop ("Non-finite values in the locations")

  ## Check corrfcn with loc
  if (corrfcn == "spherical" & NCOL(loc) > 3) {
    stop ("Cannot use the spherical correlation for dimensions
grater than 3.")
  }

  ## Split sample, prediction
  ii <- is.finite(yy)
  y <- yy[ii]
  n <- sum(ii)
  l <- ll[ii]
  l <- if (is.null(l)) rep.int(1.0, n) else as.double(l)
  if (any(!is.finite(l))) stop ("Non-finite values in the weights")
  if (any(l <= 0)) stop ("Non-positive weights not allowed")
  if (grepl("^binomial(\\..+)?$", family)) {
    l <- l - y # Number of failures
  }
  F <- FF[ii, , drop = FALSE]
  dm <- sp::spDists(loc[ii, , drop = FALSE], longlat = longlat)
  offset <- oofset[ii]

  ## Prior for ssq
  ssqdf <- as.double(ssqdf)
  if (ssqdf <= 0) stop ("Argument ssqdf must > 0")
  ssqsc <- as.double(ssqsc)
  if (ssqsc <= 0) stop ("Argument ssqsc must > 0")

  ## Prior for beta
  betaprior <- getbetaprior(betm0, betQ0, p)
  betm0 <- betaprior$betm0
  betQ0 <- betaprior$betQ0

  np <- as.integer(np)
  if (np < 1) stop ("Input np must be at least 1.")
  fval <- double(npars)
  ssq <- 0                    # Not used

  f90 <- .Fortran("llikparscalc", fval, nu, phi, omg, kappa, npars,
                  as.double(y), as.double(l),
                  F, as.double(offset), betm0, betQ0, ssqdf, ssqsc,
                  dm, tsq, tsqdf, as.integer(n), as.integer(p),
                  as.integer(np), ssq, ifam, icf, PACKAGE = "geoBayes")
  list(par_vals = allpars, aloglik = f90[[1]])
}


##' Approximate log-likelihood maximisation
##'
##' Uses the "L-BFGS-B" method of the function
##' \code{\link[stats]{optim}} to maximise the log-likelihood for the
##' parameters \code{linkp}, \code{phi}, \code{omg}, \code{kappa}.
##' @title Log-likelihood maximisation
##' @param paroptim A named list with the components "linkp", "phi",
##'   "omg", "kappa". Each component must be numeric with length 1, 2,
##'   or 3 with elements in increasing order. If the compontent's
##'   length is 1, then the corresponding parameter is considered to
##'   be fixed at that value. If 2, then the two numbers denote the
##'   lower and upper bounds for the optimisation of that parameter
##'   (infinities are allowed). If 3, these correspond to lower bound,
##'   starting value, upper bound for the estimation of that
##'   parameter.
##' @param formula A representation of the model in the form
##'   \code{response ~ terms}.
##' @param family The distribution of the response.
##' @param data An optional data frame containing the variables in the
##'   model.
##' @param weights An optional vector of weights. Number of replicated
##'   samples for Gaussian and gamma, number of trials for binomial,
##'   time length for Poisson.
##' @param subset An optional vector specifying a subset of
##'   observations to be used in the fitting process.
##' @param offset See \code{\link[stats]{lm}}.
##' @param atsample A formula in the form \code{~ x1 + x2 + ... + xd}
##'   with the coordinates of the sampled locations.
##' @param corrfcn Spatial correlation function. See
##'   \code{\link{geoBayes_correlation}} for details.
##' @param np The number of integration points for the spatial
##'   variance parameter sigma^2. The total number of points will be
##'   \code{2*np + 1}.
##' @param betm0 Prior mean for beta (a vector or scalar).
##' @param betQ0 Prior standardised precision (inverse variance)
##'   matrix. Can be a scalar, vector or matrix. The first two imply a
##'   diagonal with those elements. Set this to 0 to indicate a flat
##'   improper prior.
##' @param ssqdf Degrees of freedom for the scaled inverse chi-square
##'   prior for the partial sill parameter.
##' @param ssqsc Scale for the scaled inverse chi-square prior for the
##'   partial sill parameter.
##' @param dispersion The fixed dispersion parameter.
##' @param longlat How to compute the distance between locations. If
##'   \code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
##'   distance. See \code{\link[sp]{spDists}}.
##' @param control A list of control parameters for the optimisation.
##'   See \code{\link[stats]{optim}}.
##' @return The output from the function \code{\link[stats]{optim}}.
##'   The \code{"value"} element is the log-likelihood, not the
##'   negative log-likelihood.
##' @importFrom stats optim model.frame optimHess model.offset
##' @importFrom optimx optimr
##' @useDynLib geoBayes llikparsval
##' @export
##' @references Evangelou, E., & Roy, V. (2019). Estimation and prediction for
##'   spatial generalized linear mixed models with parametric links
##'   via reparameterized importance sampling. Spatial Statistics, 29,
##'   289-315.
alik_optim <- function (paroptim,
                      formula,
                      family = "gaussian",
                      data, weights, subset, offset, atsample,
                      corrfcn = "matern",
                      np, betm0, betQ0, ssqdf, ssqsc, dispersion = 1,
                      longlat = FALSE, control = list())
{
  INPUT <- environment()
  log.phi <- FALSE
  log.omg <- FALSE
  ## Family
  ifam <- .geoBayes_family(family)
  if (ifam) {
    family <- .geoBayes_models$family[ifam]
  } else {
    stop ("This family has not been implemented.")
  }

  ## Logical input
  longlat <- as.logical(longlat)

  ## Correlation function
  icf <- .geoBayes_correlation(corrfcn)
  corrfcn <- .geoBayes_corrfcn$corrfcn[icf]

  ## Check dispersion input
  tsq <- as.double(dispersion)
  if (tsq <= 0) stop ("Invalid argument dispersion")
  tsqdf <- 0

  ## Design matrix and data
  if (missing(data)) data <- environment(formula)
  if (length(formula) != 3) stop ("The formula input is incomplete.")
  if ("|" == all.names(formula[[2]], TRUE, 1)) formula[[2]] <- formula[[2]][[2]]
  mfc <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data", "subset", "weights", "offset"),
             names(mfc), 0L)
  mfc <- mfc[c(1L, m)]
  mfc$formula <- formula
  mfc$drop.unused.levels <- TRUE
  mfc$na.action <- "na.pass"
  mfc[[1L]] <- quote(stats::model.frame)
  mf <- eval(mfc, parent.frame())
  mt <- attr(mf, "terms")
  FF <- model.matrix(mt,mf)
  if (!all(is.finite(FF))) stop ("Non-finite values in the design matrix")
  p <- NCOL(FF)
  yy <- unclass(model.response(mf))
  if (!is.vector(yy)) {
    stop ("The response must be a vector")
  }
  yy <- as.double(yy)
  ll <- model.weights(mf)
  oofset <- as.vector(model.offset(mf))
  if (!is.null(oofset)) {
    if (length(oofset) != NROW(yy)) {
      stop(gettextf("number of offsets is %d, should equal %d (number of observations)", 
                    length(oofset), NROW(yy)), domain = NA)
    } else {
      oofset <- as.double(oofset)
    }
  } else {
    oofset <- double(NROW(yy))
  }

  ## All locations
  atsample <- update(atsample, NULL ~ .)
  locvars <- all.vars(atsample)
  formula1 <- as.formula(paste('~', paste(c(locvars, all.vars(formula)),
                                          collapse = ' + ')))
  mfc1 <- mfc
  mfc1$formula <- formula1
  mf1 <- eval(mfc1, parent.frame())
  m <- match(locvars, names(mf1))
  loc <- as.matrix(mf1[, m])
  if (!all(is.finite(loc))) stop ("Non-finite values in the locations")

  ## Check corrfcn with loc
  if (corrfcn == "spherical" && NCOL(loc) > 3) {
    stop ("Cannot use the spherical correlation for dimensions
grater than 3.")
  }

  ## Split sample, prediction
  ii <- is.finite(yy)
  y <- yy[ii]
  n <- sum(ii)
  l <- ll[ii]
  l <- if (is.null(l)) rep.int(1.0, n) else as.double(l)
  if (any(!is.finite(l))) stop ("Non-finite values in the weights")
  if (any(l <= 0)) stop ("Non-positive weights not allowed")
  if (grepl("^binomial(\\..+)?$", family)) {
    l <- l - y # Number of failures
  }
  F <- FF[ii, , drop = FALSE]
  offset <- oofset[ii]
  dm <- sp::spDists(loc[ii, , drop = FALSE], longlat = longlat)

  ## Prior for ssq
  ssqdf <- as.double(ssqdf)
  if (ssqdf <= 0) stop ("Argument ssqdf must > 0")
  ssqsc <- as.double(ssqsc)
  if (ssqsc <= 0) stop ("Argument ssqsc must > 0")

  ## Prior for beta
  betaprior <- getbetaprior(betm0, betQ0, p)
  betm0 <- betaprior$betm0
  betQ0 <- betaprior$betQ0

  np <- as.integer(np)
  if (np < 1) stop ("Input np must be at least 1.")
  ssq <- 0 # Not used

  ## Read and check paroptim
  paroptim <- getparoptim(paroptim, ifam, icf)
  pstart <- as.double(paroptim$pstart)
  lower <- paroptim$lower
  upper <- paroptim$upper
  estim <- paroptim$estim
  log.phi <- log.phi && estim[2]
  log.omg <- log.omg && estim[3]
  if (log.phi) {             # Estimating phi so transform
    pstart[2] <- log(pstart[2])
    lower[2] <- log(lower[2])
    upper[2] <- log(upper[2])
  }
  if (log.omg) {             # Estimating omg so transform
    pstart[3] <- log(pstart[3])
    lower[3] <- log(lower[3])
    upper[3] <- log(upper[3])
  }

  fn <- function (par)
  {
    parin <- pstart
    parin[estim] <- par
    nu <- parin[1]
    phi <- parin[2]; if(log.phi) phi <- exp(phi)
    omg <- parin[3]; if(log.omg) omg <- exp(omg)
    kappa <- parin[4]
    fval <- 0
    gval <- rep.int(0, 4)
    ideriv <- rep.int(0L, 4)
    f90 <- try(
      .Fortran("llikparsval", fval, gval, ideriv, nu, phi, omg, kappa,
               as.double(y), as.double(l),
               as.double(F), as.double(offset),
               as.double(betm0), as.double(betQ0),
               as.double(ssqdf), as.double(ssqsc),
               as.double(dm), as.double(tsq), as.double(tsqdf),
               as.integer(n), as.integer(p),
               as.integer(np), as.double(ssq), ifam, icf,
               PACKAGE = "geoBayes"),
      silent = TRUE)
    if (inherits(f90, "try-error")) return (NA)
    -f90[[1]]
  }
  gr <- function (par)
  {
    d1 <- c(1, 1, 1, 1)
    parin <- pstart
    parin[estim] <- par
    nu <- parin[1]
    phi <- parin[2]; if(log.phi) phi <- d1[2] <- exp(phi)
    omg <- parin[3]; if(log.omg) omg <- d1[3] <- exp(omg)
    kappa <- parin[4]
    fval <- 0
    gval <- rep.int(0, 4)
    ideriv <- as.integer(estim)
    f90 <- try(
      .Fortran("llikparsval", fval, gval, ideriv, nu, phi, omg, kappa,
               as.double(y), as.double(l),
               F, as.double(offset), betm0, betQ0, ssqdf, ssqsc,
               dm, tsq, tsqdf, as.integer(n), as.integer(p),
               as.integer(np), ssq, ifam, icf, PACKAGE = "geoBayes"),
      silent = TRUE)
    if (inherits(f90, "try-error")) return (rep(NA, sum(estim)))
    -f90[[2]][estim]*d1[estim]
  }
  method <- "L-BFGS-B" # if (sum(estim) == 1) "Brent" else "L-BFGS-B"
  method <- "nlminb"
  if (isTRUE(control$fnscale < 0)) control$fnscale <- -control$fnscale
###   op <- stats::optim(pstart[estim], fn, gr, method = method,
###                      lower = lower[estim], upper = upper[estim],
###                      hessian = TRUE, control = control)
  if (isTRUE(control$maximize)) control$maximize <- FALSE
  op <- optimx::optimr(pstart[estim], fn, gr, method = method,
                       lower = lower[estim], upper = upper[estim],
                       hessian = FALSE, control = control)
  op$gradient <- gr(op$par)
  op$hessian <- stats::optimHess(op$par, fn, gr, control = control)
  parout <- pstart
  parout[estim] <- op$par
  d1 <- c(1, 1, 1, 1)
  if (log.phi) d1[2] <- parout[2] <- exp(parout[2])
  if (log.omg) d1[3] <- parout[3] <- exp(parout[3])
  names(parout) <- names(d1) <- c("linkp", "phi", "omg", "kappa")
  op$par <- parout
  op$value <- -op$value
  op$hessian <- op$hessian*tcrossprod(d1[estim])
  op$INPUT <- INPUT
###   opscale <- if(!is.null(control$parscale)) control$parscale else 1
###   control$parscale <- NULL
###   op <- stats::nlminb(pstart[estim], fn, gr, scale = opscale,
###                       control = control,
###                       lower = lower[estim], upper = upper[estim])
  op
}


##' Calculate the likelihood approximation at different parameter
##' values. This function is useful for choosing the skeleton set.
##'
##' The input \code{par_vals} is meant to contain vector of parameter
##' values for each parameter. For each element in \code{par_vals},
##' the other parameters are set equal to the maximisers given in
##' \code{likopt} and the approximate likelihood is computed. The
##' cuttoff is calculated using linear interpolation provided by
##' \code{\link[stats]{approx}}.
##' @title Approximate log-likelihood calculation
##' @param likopt Output from the function \code{\link{alik_optim}}.
##' @param par_vals A named list with some of the components "linkp",
##'   "phi", "omg", "kappa".
##' @param likthreshold A threshold value proportion to calculate the
##'   cutoff. The cutoff will be calculated as that proportion
##'   relative to the maximum value of the log-likelihood.
##' @return A list with the log-likelihood approximation and cutoff values.
##' @export
##' @references Evangelou, E., & Roy, V. (2019). Estimation and prediction for
##'   spatial generalized linear mixed models with parametric links
##'   via reparameterized importance sampling. Spatial Statistics, 29,
##'   289-315.
alik_cutoff <- function (likopt, par_vals, likthreshold) {
  ## For each parameter value in par_vals, fix the other parameters at
  ## likopt and compute the approximate log-likelihood.
  if (likthreshold >= 1) stop ("Input likthreshold must < 1.")
  if (likthreshold < 0) stop ("Input likthreshold must >= 0.")
  parnm <- c("linkp", "phi", "omg", "kappa")
  par_vals <- par_vals[parnm]
  names(par_vals) <- parnm
  family <- get("family", likopt$INPUT)
  ifam <- .geoBayes_family(family)
  needlnk <- .geoBayes_models$needlinkp[ifam]
  corrfcn <- get("corrfcn", likopt$INPUT)
  icf <- .geoBayes_correlation(corrfcn)
  needkap <- .geoBayes_corrfcn$needkappa[icf]
  calcpar <- c(needlnk, TRUE, TRUE, needkap) & !sapply(par_vals, is.null)
  parmx <- as.list(likopt$par)
  likmx <- likopt$value
  lthr <- log(likthreshold)
  alk <- as.list(rep(likmx, 4)); names(alk) <- parnm
  cff <- parmx
  input <- substitute(
    alik_inla(par_vals = NULL, formula = formula, family = family,
              data = data, weights = weights, subset = subset,
              atsample = atsample, corrfcn = corrfcn, np = np,
              betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf,
              ssqsc = ssqsc, tsqdf = tsqdf, tsqsc = tsqsc,
              dispersion = dispersion, longlat = longlat), env = likopt$INPUT)
  for (i in 1:4) {
    if (calcpar[i]) {
      tmp <- parmx
      tmp[[i]] <- par_vals[[i]]
      input[[2]] <- as.data.frame(tmp)
      alk[[i]] <- eval(input, likopt$INPUT)$aloglik
    }
  }
  if (is.finite(lthr)) {
    for (i in 1:4) {
      if (calcpar[i]) {
        if (is.finite(lthr)) {
          cff[[i]] <- calc_cutoff(par_vals[[i]], alk[[i]] - likmx, lthr)
        } else {
          cff[[i]] <- c(min(par_vals[[i]]), max(par_vals[[i]]))
        }
      }
    }
  }
  list(par_vals = par_vals, aloglik = alk, cutoff = cff,
       threshold = likmx + lthr, likmax = likmx)
}

##' @importFrom stats approx
calc_cutoff <- function (x, y, y0) {
  i <- which.max(y)
  xmx <- x[i]
  i1 <- x <= xmx
  i2 <- x >= xmx
  x1 <- if (sum(i1) > 1) stats::approx(y[i1], x[i1], y0, rule = 2)$y else x[i1]
  x2 <- if (sum(i2) > 1) stats::approx(y[i2], x[i2], y0, rule = 2)$y else x[i2]
  c(x1, x2)
}

##' Plot likelihood approximation.
##'
##' The plot can be used to visualise the Laplace approximation to the
##' likelihood provided by the function \code{\link{alik_cutoff}}.
##' @title Plot likelihood approximation
##' @param alikobj Output from \code{\link{alik_cutoff}}.
##' @return Draws a plot.
##' @importFrom graphics segments
##' @rdname alik_cutoff
##' @export
alik_plot <- function (alikobj) {
  pvl <- alikobj$par_vals
  alk <- alikobj$aloglik
  cff <- alikobj$cutoff
  lmx <- alikobj$likmax
  thr <- exp(alikobj$threshold - lmx)
  lplt <- sapply(alk, function(x) length(x) > 1)
  npar <- sum(lplt)
  oldpar <- par(mfrow = c(1, npar)); on.exit(par(oldpar))
  parlb <- expression(linkp = nu, phi = phi, omg = omega, kappa = kappa)
  for (i in 1:4) {
    if (lplt[i]) {
      plot(pvl[[i]], exp(alk[[i]]-lmx), type = 'l', xlab = parlb[i], ylab = "")
      segments(c(cff[[i]][1], cff[[i]][1], cff[[i]][2]), c(0, thr, thr),
               c(cff[[i]][1], cff[[i]][2], cff[[i]][2]), c(thr, thr, 0),
               lty = 2)
    }
  }
}


## .. content for \description{} (no empty lines) .. TODO
##
## .. content for \details{} .. TODO
## @title Approximate log-likelihood calculation
## @param likopt Output from the function \code{\link{alik_optim}}.
## @param par_grid A \code{data.frame} with the components "linkp",
##   "phi", "omg", "kappa". The likelihood will be evalutated on each
##   row of this data frame.
## @param likthreshold A threshold value proportion to discard
##   points. The cutoff will be calculated as that proportion
##   relative to the maximum value of the log-likelihood.
## @param formula A representation of the model in the form
##   \code{response ~ terms}.
## @param family The distribution of the response. Can be one of the
##   options in \code{\link{.geoBayes_models}} or
##   \code{"transformed.gaussian"}.
## @param data An optional data frame containing the variables in the
##   model.
## @param weights An optional vector of weights. Number of replicated
##   samples for Gaussian and gamma, number of trials for binomial,
##   time length for Poisson.
## @param subset An optional vector specifying a subset of
##   observations to be used in the fitting process.
## @param atsample A formula in the form \code{~ x1 + x2 + ... + xd}
##   with the coordinates of the sampled locations.
## @param corrfcn Spatial correlation function. Can be one of the
##   choices in \code{\link{.geoBayes_corrfcn}}.
## @param np The number of integration points for the spatial
##   variance parameter sigma^2. The total number of points will be
##   \code{2*np + 1}.
## @param betm0 Prior mean for beta (a vector or scalar).
## @param betQ0 Prior standardised precision (inverse variance)
##   matrix. Can be a scalar, vector or matrix. The first two imply a
##   diagonal with those elements. Set this to 0 to indicate a flat
##   improper prior.
## @param ssqdf Degrees of freedom for the scaled inverse chi-square
##   prior for the partial sill parameter.
## @param ssqsc Scale for the scaled inverse chi-square prior for the
##   partial sill parameter.
## @param tsqdf Degrees of freedom for the scaled inverse chi-square
##   prior for the measurement error parameter.
## @param tsqsc Scale for the scaled inverse chi-square prior for the
##   measurement error parameter.
## @param dispersion The fixed dispersion parameter.
## @param longlat How to compute the distance between locations. If
##   \code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
##   distance. See \code{\link[sp]{spDists}}.
## @return A list with the log-likelihood approximation and cutoff values.
## calc_alik2 <- function (likopt, par_grid, likthreshold, formula,
##                         family = "gaussian",
##                         data, weights, subset, atsample,
##                         corrfcn = "matern",
##                         np, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsqsc,
##                         dispersion = 1, longlat = FALSE) {
##   ## For each parameter value in par_grid, fix the other parameters at
##   ## likopt and compute the approximate log-likelihood.
##   if (likthreshold >= 1) stop ("Input likthreshold must < 1.")
##   par_grid <- as.data.frame(par_grid, stringsAsFactors = FALSE)
##   parnm <- c("linkp", "phi", "omg", "kappa")
##   ifam <- .geoBayes_family(family)
##   needlnk <- .geoBayes_models$needlinkp[ifam]
##   if (!needlnk) par_grid$linkp <- 0
##   icf <- .geoBayes_correlation(corrfcn)
##   needkap <- .geoBayes_corrfcn$needkappa[icf]
##   if (!needkap) par_grid$kappa <- 0
##   par_grid <- par_grid[, parnm]
##   parmx <- likopt$par
##   likmx <- likopt$value
##   lthr <- log(likthreshold)
##   input <- substitute(
##     alik_inla(linkp = pargrid[, 1], phi = pargrid[, 2], omg = pargrid[, 3],
##              kappa = pargrid[, 4], formula = formula, family = family,
##              data = data, weights = weights, subset = subset,
##              atsample = atsample, corrfcn = corrfcn, np = np,
##              betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf,
##              ssqsc = ssqsc, tsqdf = tsqdf, tsqsc = tsqsc,
##              dispersion = dispersion, longlat = longlat))
##   alk <- eval(input)
##
## }

!!! skelpnts.f90 ---
!!
!! Author: Evangelos Evangelou
!! Created: Thu, 25 Sep, 2014 16:19 (BST)
!! Last-Updated: Wed, 1 Jun, 2016 13:26 (BST)
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: These routines are used for finding appropriate skeleton
!!! points.
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine gaussaprx (mean, prec, fcyz, y1, y2, Ups, ldh_Ups, &
   nu, xi, lmxi, ssq, tsq, tsqdf, n, ifam)
  ! Gaussian approximation to f(z|y;phi,ssq)
  ! mean is the Gaussian mean of the approximation
  ! prec is the precision matrix
  ! fcyz is the value at the maximum
  ! Rest are input to the joint pdf
  use lbfgsbmod
  use linkfcns, only: flink_ga
  use modelfcns, only: mustart, flink_sp => flink
  implicit none
  integer, intent(in) :: n, ifam
  logical, intent(in) :: lmxi
  double precision, intent(in) :: y1(n), y2(n), Ups(n, n), ldh_Ups, &
     nu, xi(n), ssq, tsq, tsqdf
  double precision, intent(out) :: mean(n), prec(n, n), fcyz
  double precision gr(n), gr1(n), gr2(n), fc1, fc2, hs(n)
  double precision tsqdfsc, respdf
  integer opnb(n), iflag, maxiter, i
  double precision oplb(n), opub(n), opfac, oppgt
  type(lbfgsb_store) opstore
  parameter ( maxiter = 1500, opfac = 1d7, oppgt = 0d0 )

  opnb = 0
  iflag = 0
  select case (ifam)
  case (0) ! Gaussian transformed
    tsqdfsc = tsq*tsqdf
    respdf = n + tsqdf
    mean = flink_ga(y1,nu) ! Starting value
    do i = 1, maxiter
      call logpdfzdz (fc1, gr1, mean, Ups, ldh_Ups, xi, lmxi, ssq, n)
      call logcondyzdz_gt (fc2, gr2, nu, y1, y2, mean, n, tsqdfsc)
      fcyz = -fc1 - respdf*fc2
      gr = -gr1 - respdf*gr2
      call lbfgsb (n, mean, oplb, opub, opnb, fcyz, gr, &
         opfac, oppgt, iflag, opstore)
      if (iflag .eq. 0) then
        exit
      else if (iflag .lt. 0) then
        call rwarn ('The optimisation for the Gaussian approximation didn''&
           &t converge')
        exit
      end if
    end do
    if (iflag .gt. 0) then
      call rwarn ('The optimisation for the Gaussian approximation needs more&
         & iterations')
    end if
    call logcondyzhs_gt (hs, nu, y1, y2, mean, n, tsqdfsc)
    hs = respdf*hs
  case default
    gr = mustart(y1,y2)
    mean = flink_sp(gr,nu) ! Starting value
    do i = 1, maxiter
      call logpdfzdz (fc1, gr1, mean, Ups, ldh_Ups, xi, lmxi, ssq, n)
      call logcondyzdz_sp (fc2, gr2, nu, y1, y2, mean, n, tsq)
      fcyz = -fc1 - fc2
      gr = -gr1 - gr2
      call lbfgsb (n, mean, oplb, opub, opnb, fcyz, gr, &
         opfac, oppgt, iflag, opstore)
      if (iflag .eq. 0) then
        exit
      else if (iflag .lt. 0) then
        call rwarn ('The optimisation for the Gaussian approximation didn''&
           &t converge')
        exit
      end if
    end do
    if (iflag .gt. 0) then
      call rwarn ('The optimisation for the Gaussian approximation needs more&
         & iterations')
    end if
    call logcondyzhs_sp (hs, nu, y1, y2, mean, n, tsq)
  end select
  fcyz = -fcyz ! Because it was minus before
  do i = 1, n
    prec(:i,i) = Ups(:i,i)/ssq ! Upper triangular elements
    prec(i,i) = prec(i,i) - hs(i)
  end do

contains

  subroutine logpdfzdz (fc, gr, z, Ups, ldh_Ups, xi, lmxi, ssq, n)
!! log-pdf of z and its derivative after integrating out beta. The 2*pi
!! constant is removed.
    implicit none
    integer, intent(in) :: n
    logical, intent(in) :: lmxi
    double precision, intent(in) :: z(n), Ups(n,n), ldh_Ups, xi(n), ssq
    double precision, intent(out) :: fc, gr(n)
    double precision zmxi(n)
    if (lmxi) then
      zmxi = z - xi
    else
      zmxi = z
    end if
    call dsymv ('u',n,1d0,Ups,n,zmxi,1,0d0,gr,1) ! gr = Ups*(z-xi)
    gr = -gr/ssq ! gr = -Ups*(z-x)/ssq
    fc = -.5d0*n*log(ssq) + ldh_Ups + .5d0*dot_product(zmxi,gr)
  end subroutine logpdfzdz

  subroutine logcondyzdz_gt (fc, gr, nu, y1, y2, z, n, tsq)
!! TODO respdf is not needed here. Can be computed outside the fcn. XXX
!! log-pdf of y|z and its derivative w.r.t. z.
!! Gaussian transformed
!! tsq is tsqdf*tsqsc on input
!! respdf is n + tsqdf on input
    use linkfcns, only: invlink_ga
    use linkdz, only: invlinkdz_ga
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz
    fc = tsq
    do i = 1, n
      par = invlink_ga(z(i),nu)
      pardz = invlinkdz_ga(z(i),nu)
      par = y1(i) - par
      gr(i) = y2(i)*par
      fc = fc + gr(i)*par
      gr(i) = gr(i)*pardz
    end do
    gr = gr/fc ! XXX Missing respdf factor. Must be computed outside the fcn
    fc = -.5d0*log(fc) ! XXX Also missing respdf = n + tsqdf
  end subroutine logcondyzdz_gt

  subroutine logcondyzhs_gt (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
!! Gaussian transformed
!! tsq is tsqdf*tsqsc on input
    use linkfcns, only: invlink_ga
    use linkdz, only: invlinkdz_ga
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz, fc
    fc = tsq
    do i = 1, n
      par = invlink_ga(z(i),nu)
      pardz = invlinkdz_ga(z(i),nu)
      hs(i) = y2(i)*par*pardz*pardz
      par = y1(i) - par
      fc = fc + y2(i)*par*par
    end do
    hs = -hs/fc ! XXX Missing respdf factor. Must be computed outside the fcn
  end subroutine logcondyzhs_gt

  subroutine logcondyzdz_sp (fc, gr, nu, y1, y2, z, n, tsq)
!! log-pdf of y|z and its derivative w.r.t. z.
    use modelfcns, only: invlink, invlinkdz, logpdfy, logpdfydlnk
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: fc, gr(n)
    integer i
    double precision par, pardz
    fc = 0d0
    do i = 1, n
      par = invlink(z(i),nu)
      pardz = invlinkdz(z(i),nu)
      fc = fc + logpdfy(y1(i),y2(i),par)
      gr(i) = logpdfydlnk(y1(i),y2(i),par) * pardz
    end do
    fc = fc/tsq
    gr = gr/tsq
  end subroutine logcondyzdz_sp

  subroutine logcondyzhs_sp (hs, nu, y1, y2, z, n, tsq)
!! Component used in the calculation of the Hessian
    use modelfcns, only: invlink, invlinkdz, logpdfyhlnk, logpdfydlnk, invlinkhz
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision, intent(out) :: hs(n)
    integer i
    double precision par, pardz
    do i = 1, n
      par = invlink(z(i),nu)
      pardz = invlinkdz(z(i),nu)
      hs(i) = logpdfyhlnk(y1(i),y2(i),par) * pardz*pardz
      hs(i) = hs(i) + logpdfydlnk(y1(i),y2(i),par) * invlinkhz(z(i),nu)
    end do
    hs = hs/tsq
  end subroutine logcondyzhs_sp
end subroutine gaussaprx


subroutine poster (fssq, meang, prechg, ssq, ssqdfh, ssqdfsc, &
   y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  ! Gives the approximation to the log posterior of the nuissance
  ! parameters f(psi|y,theta) propto f(y,z,psi,theta)/f(z|y,psi,theta);
  ! psi = ssq here
  ! meang is the mean of the Gaussian approximation for that ssq
  ! prechg is the Chol of the precision matrix of the Gaussian
  ! approximation for that ssq.
  implicit none
  integer, intent(in) :: n, ifam
  logical, intent(in) :: lmxi
  double precision, intent(in) :: y1(n), y2(n), Ups(n, n), ldh_Ups, nu, &
     xi(n), ssq, tsq, tsqdf, ssqdfh, ssqdfsc
  double precision, intent(out) :: fssq, meang(n), prechg(n,n)
  double precision :: ldh, fpr
  logical, external :: disnan ! LAPACK auxiliary routine.
  if (disnan(ssq)) call rexit("poster - ssq entered is NaN.")
  fpr = -(1d0+ssqdfh)*log(ssq) - .5d0*ssqdfsc/ssq ! Prior
  call gaussaprx (meang, prechg, fssq, y1, y2, Ups, ldh_Ups, &
     nu, xi, lmxi, ssq, tsq, tsqdf, n, ifam)
  call factorpd(n,prechg,ldh)
  fssq = fssq - ldh + fpr
contains
  subroutine factorpd(n,x,ldh)
    ! The returned matrix x is upper triangular s.t. U'*U = X
    integer, intent(in) :: n
    double precision, intent(inout) :: x(n,n)
    double precision, intent(out) :: ldh
    integer i
!!     open(11,file='postermat.txt')
!!     write(11,*) x
!!     close(11)
    call dpotrf ('u',n,x,n,i)
    if (i .ne. 0) then
      call rexit("poster - Non positive definite matrix")
    end if
    ldh = 0d0
    do i = 1,n
      ldh = ldh + log(x(i,i))
    end do
  end subroutine factorpd
end subroutine poster

subroutine aprxposterssq (fssq, meang, prechg, dz_dnu, dz_dphi, &
   ssq, nu, phi, omg, kappa, y1, y2, F, offset, betm0, betQ0, &
   ssqdf, ssqsc, dm, tsq, tsqdf, n, p, ifam, icf)
  use covfun
  use betaprior
  use modelfcns
  use calcbd_fcns, only: qform, traceAB, cor_dcov
  implicit none
  integer, intent(in) :: n, p, ifam, icf
  double precision, intent(in) :: ssq, ssqdf, ssqsc, &
     y1(n), y2(n), phi, nu, omg, kappa, dm(n,n), F(n,p), &
     betm0(p), betQ0(p,p), tsq, tsqdf, offset(n)
  double precision, intent(out) :: fssq, meang(n), prechg(n,n), &
     dz_dnu(n), dz_dphi(n)
  double precision Ups(n,n), ldh_Ups, xi(n), modeldfh, ssqdfh, ssqdfsc
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), par(n), varh(n,n)
  integer j
  call create_model (ifam)
  call create_spcor (icf,n)
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)
  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  ssqdfh = .5d0*ssqdf
  ssqdfsc = ssqdf*ssqsc
  call poster (fssq, meang, prechg, ssq, ssqdfh, ssqdfsc, &
     y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  par = invlink(meang,nu)
  varh = prechg
  call dtrtri ('u','n',n,varh,n,j)
  dz_dnu = fdz_dnu (meang, par, ssq, varh, y1, y2, Ups, &
     nu, xi, lmxi, tsq, tsqdf, n)
  dz_dphi = fdz_dphi (meang, par, ssq, varh, y1, y2, dm, Ups, &
     nu, xi, lmxi, tsq, tsqdf, n, icf)

contains
!   function fdz_dnu (z, par, ssq, prechg, y1, y2, Ups, &
!      nu, xi, lmxi, tsq, tsqdf, n)
!     implicit none
!     integer, intent(in) :: n
!     logical, intent(in) :: lmxi
!     double precision, intent(in) :: y1(n), y2(n), Ups(n,n), nu, &
!        xi(n), tsq, tsqdf, ssq, z(n), prechg(n,n), par(n)
!     double precision, dimension(n) :: fdz_dnu
!     fdz_dnu = logpdfyhlnk(y1,y2,par) * invlinkdz(z,nu) * invlinkdn(z,nu) &
!        + logpdfydlnk(y1,y2,par) * invlinkdzdn(z,nu)
!     call dtrsv ('u','t','n',n,prechg,n,fdz_dnu,1)
!     call dtrsv ('u','n','n',n,prechg,n,fdz_dnu,1)
!   end function fdz_dnu

  function fdz_dnu (z, par, ssq, varh, y1, y2, Ups, &
     nu, xi, lmxi, tsq, tsqdf, n)
    implicit none
    integer, intent(in) :: n
    logical, intent(in) :: lmxi
    double precision, intent(in) :: y1(n), y2(n), Ups(n,n), nu, &
       xi(n), tsq, tsqdf, ssq, z(n), varh(n,n), par(n)
    double precision, dimension(n) :: fdz_dnu
    fdz_dnu = logpdfyhlnk(y1,y2,par) * invlinkdz(z,nu) * invlinkdn(z,nu) &
       + logpdfydlnk(y1,y2,par) * invlinkdzdn(z,nu)
    fdz_dnu = fdz_dnu/tsq
    call dtrmv ('u','t','n',n,varh,n,fdz_dnu,1)
    call dtrmv ('u','n','n',n,varh,n,fdz_dnu,1)
  end function fdz_dnu

  function fdz_dphi (z, par, ssq, varh, y1, y2, dm, Ups, &
     nu, xi, lmxi, tsq, tsqdf, n, icf)
    implicit none
    integer, intent(in) :: n, icf
    logical, intent(in) :: lmxi
    double precision, intent(in) :: y1(n), y2(n), nu, dm(n,n), &
       xi(n), tsq, tsqdf, ssq, z(n), varh(n,n), par(n), Ups(n,n)
    double precision, dimension(n) :: fdz_dphi
    double precision zmxi(n), UpsDTUps(n,n), DT(n,n), &
       DTUps(n,n)
    DT = cor_dcov(n,dm,phi,omg,kappa,1)
    call fill_symmetric_matrix(DT,n)
    call dsymm ('r','u',n,n,1d0,Ups,n,DT,n,0d0,DTUps,n)
    call dsymm ('l','u',n,n,1d0,Ups,n,DTUps,n,0d0,UpsDTUps,n)
    UpsDTUps = UpsDTUps/ssq
    if (lmxi) then
      zmxi = z - xi
    else
      zmxi = z
    end if
    call dsymv ('u',n,1d0,UpsDTUps,n,zmxi,1,0d0,fdz_dphi,1) ! Ups*(z-xi)
    call dtrmv ('u','t','n',n,varh,n,fdz_dphi,1)
    call dtrmv ('u','n','n',n,varh,n,fdz_dphi,1)
  end function fdz_dphi
end subroutine aprxposterssq


subroutine posterlog (fval, meang, prechg, logssq, ssqdfh, ssqdfsc, &
   y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  ! Gives the approximation to the log posterior of the nuissance
  ! parameters f(psi|y,theta) propto f(y,z,psi,theta)/f(z|y,psi,theta);
  ! psi = log(ssq) here
  ! meang is the mean of the Gaussian approximation for that ssq
  ! prechg is the Chol of the precision matrix of the Gaussian
  ! approximation for that psi.
  implicit none
  integer, intent(in) :: n, ifam
  logical, intent(in) :: lmxi
  double precision, intent(in) :: y1(n), y2(n), Ups(n, n), ldh_Ups, nu, &
     xi(n), logssq, tsq, tsqdf, ssqdfh, ssqdfsc
  double precision, intent(out) :: fval, meang(n), prechg(n,n)
  double precision ssq
  logical, external :: disnan
  if (disnan(logssq)) call rexit("posterlog - logssq entered is NaN.")
  ssq = exp(logssq)
  call poster (fval, meang, prechg, ssq, ssqdfh, ssqdfsc, &
     y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  fval = fval + logssq
end subroutine posterlog

subroutine optlogssq (tval, tprc, pdfval, meang, prechg, ssqdfh, ssqdfsc, &
   y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  ! Maximise the posterior for t = log(ssq). Uses 2nd degree polynomial
  ! interpolation to the posterior of log(ssq).
  implicit none
  integer, intent(in) :: n, ifam
  logical, intent(in) :: lmxi
  double precision, intent(in) :: y1(n), y2(n), Ups(n, n), ldh_Ups, nu, &
     xi(n), tsq, tsqdf, ssqdfh, ssqdfsc
  double precision, intent(inout) :: tval
  double precision, intent(out) :: tprc, pdfval, meang(n), prechg(n,n)
  double precision pdfpnts(3), tnew, tpnts(3), mtmp(n), ptmp(n,n), pdfnew
  double precision, parameter :: eps = 1d-4
  integer, parameter :: maxiter = 100
  integer i, ii
  logical lconv, ok
  tpnts(1) = .8*tval - 15d-1
  tpnts(2) = tval
  tpnts(3) = 1.2*tval + 15d-1
  do i = 1, 3
    call posterlog (pdfpnts(i), mtmp, ptmp, tpnts(i), ssqdfh, ssqdfsc, &
       y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
    if (i .eq. 1 .or. pdfpnts(i) .gt. pdfval) then
      tval = tpnts(i)
      pdfval = pdfpnts(i)
      meang = mtmp
      prechg = ptmp
    end if
  end do
  ! print*, 0, real(tpnts), real(pdfpnts), real(pdfval)
  do i = 1, maxiter
    call polmax(tnew, ii, ok, tpnts, pdfpnts) ! Find new point
    call posterlog (pdfnew, mtmp, ptmp, tnew, ssqdfh, ssqdfsc, &
       y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
    lconv = .false.
    if (ok) lconv = abs(pdfnew - pdfval) .lt. eps*(1d0+sqrt(abs(pdfval)))
    if (ok .and. .not. lconv) lconv = minval(abs(tnew-tpnts)) .lt. &
       eps*1e2*(1d0+sqrt(abs(tnew)))
    ! print*, real(tnew), real(pdfnew), lconv
    if (lconv) then
      exit
    end if
    tpnts(ii) = tnew        ! Replace with new value
    tval = tnew             ! Current best guess
    pdfpnts(ii) = pdfnew
    if (pdfnew .gt. pdfval) then
      pdfval = pdfnew
      meang = mtmp
      prechg = ptmp
    end if
  ! print*, i, real(tpnts), real(pdfpnts), real(pdfval)
  end do
  if (.not. lconv) then
    call rwarn("optlogssq - Was not able to find the maximum posterior.")
  end if
  tprc = -.5d0*polcf2(tpnts, pdfpnts)
  if (tprc .le. 0d0) then
    call rexit("optlogssq - Computed non-positive precision.")
  end if
contains
  subroutine polmax(xmax, irem, ok, x, y)
    ! Find the argmax of a 2nd degree interpolated polynomial.
    implicit none
    double precision, intent(in), dimension(3) :: x, y
    double precision, intent(out) :: xmax
    integer, intent(out) :: irem
    logical, intent(out) :: ok
    double precision :: xm(3), xp(3), a, b
    xm(1) = x(2) - x(3)
    xm(2) = x(3) - x(1)
    xm(3) = x(1) - x(2)
    xp(1) = x(2) + x(3)
    xp(2) = x(3) + x(1)
    xp(3) = x(1) + x(2)
    b = xm(1)*xm(2)*xm(3)
    xm = y*xm
    a = xm(1) + xm(2) + xm(3)
    b = a*b
    ok = b .gt. 0d0 ! Check if the polynomial has a max
    if (ok) then
      xp = xm*xp
      b = xp(1) + xp(2) + xp(3)
      xmax = .5d0*b/a
      irem = minloc(y,1)
    else ! polynomial is the otherway around
      xmax = (x(1) + x(2) + x(3))/3d0
      xm = abs(x - xmax)
      irem = minloc(xm,1)
      if (xm(irem) .eq. 0d0) then
        irem = maxloc(xm,1)
        if (xm(irem) .eq. 0d0) then
          call rexit ("polmax - Cannot locate max")
        end if
      end if
    end if
  end subroutine polmax

  double precision function polcf2(x, y)
    ! Find the coefficient of x^2 on the interolated polynomial.
    implicit none
    double precision, intent(in), dimension(3) :: x, y
    double precision :: xm(3), xp(3)
    xm(1) = x(2) - x(3)
    xm(2) = x(3) - x(1)
    xm(3) = x(1) - x(2)
    xp(1) = xm(2)*xm(3)
    xp(2) = xm(3)*xm(1)
    xp(3) = xm(1)*xm(2)
    xm = y/xp
    polcf2 = -sum(xm)
  end function polcf2

  double precision function variance(x,n)
    integer, intent(in) :: n
    double precision, intent(in) :: x(1:n)
    integer i
    double precision mn, df
    mn = x(1)
    variance = 0d0
    do i = 2, n
      df = x(i) - mn
      mn = mn + df/i
      variance = variance + df*(x(i) - mn)
    end do
    variance = variance/(n-1)
  end function variance
end subroutine optlogssq

subroutine gridposter (np, tg, twght, meang, prechg, ssqdfh, ssqdfsc, &
   ssqin, y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  ! Create an appropriate grid of size np for the posterior of ssq. The
  ! grid size is 2*np+1 and the midpoint corresponds to the maximiser of
  ! the approximate posterior.
  implicit none
  integer, intent(in) :: np, n, ifam
  logical, intent(in) :: lmxi
  double precision, intent(in) :: y1(n), y2(n), Ups(n, n), ldh_Ups, nu, &
     xi(n), tsq, tsqdf, ssqdfh, ssqdfsc, ssqin
  double precision, intent(out) :: tg(np+np+1), twght(np+np+1), &
     meang(n,np+np+1), prechg(n,n,np+np+1)
  double precision tmx, tsd, tlo, thi, step, sfctr, tst
  integer i, m, e
  double precision, parameter :: zq = 4d0, eps = -6.5
  m = np + 1
  e = m + np
  sfctr = 1d0 - 1d0/np
  ! Find the middle point
  tmx = log(ssqin)
  call optlogssq (tmx, tsd, twght(m), meang(:,m), prechg(:,:,m), ssqdfh, &
     ssqdfsc, y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  tg(m) = tmx
  tsd = zq/sqrt(tsd)
  ! Lo
  tlo = tmx - tsd
  step = tsd/np
  tg(1) = tlo
  do i = 1, 20
    call posterlog (twght(1), meang(:,1), prechg(:,:,1), tg(1), ssqdfh, &
       ssqdfsc, y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
    tst = twght(1) - twght(m)
    if (tst .gt. eps) exit
    tlo = tlo + step
    tg(1) = tlo
    step = step*sfctr
  end do
  do i = 2, np
    tg(i) = tg(i-1) + step
    call posterlog (twght(i), meang(:,i), prechg(:,:,i), tg(i), ssqdfh, &
       ssqdfsc, y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
!     twght(i) = twght(i) - twght(m)
  end do
  ! Hi
  thi = tmx + tsd
  step = tsd/np
  tg(e) = thi
  do i = 1, 20
    call posterlog (twght(e), meang(:,e), prechg(:,:,e), tg(e), ssqdfh, &
       ssqdfsc, y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
    tst = twght(e) - twght(m)
    if (tst .gt. eps) exit
    thi = thi - step
    tg(e) = thi
    step = step*sfctr
  end do
  do i = e-1, np+2, -1
    tg(i) = tg(i+1) - step
    call posterlog (twght(i), meang(:,i), prechg(:,:,i), tg(i), ssqdfh, &
       ssqdfsc, y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
!     twght(i) = twght(i) - twght(m)
  end do
!!   open(11, file = "gridposter_data.txt")
!!   write(11,*) tg
!!   write(11,*) twght
!!   close(11)
!   twght(m) = 0d0
  !! twght = twght - log(sum(exp(twght))) ! XXX Not normalised.
end subroutine gridposter

subroutine aloglik (np, logssqg, out, meang, prechg, ssqdfh, ssqdfsc, &
   y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  implicit none
  integer, intent(in) :: np, n, ifam
  logical, intent(in) :: lmxi
  double precision, intent(in) :: y1(n), y2(n), Ups(n,n), nu, &
     xi(n), tsq, tsqdf, ssqdfh, ssqdfsc, ldh_Ups
  double precision, intent(in) :: logssqg(np+np+1)
  double precision, intent(out) :: out(np+np+1), meang(n,np+np+1), &
     prechg(n,n,np+np+1)
  integer i
  do i = 1, np+np+1
    call posterlog (out(i),meang(:,i),prechg(:,:,i),logssqg(i),ssqdfh,ssqdfsc,&
       y1,y2,Ups,ldh_Ups,nu,xi,lmxi,tsq,tsqdf,n,ifam)
  end do
end subroutine aloglik


subroutine aloglik_dnu (np, logssqg, dnu, meang, prechg, ssqdfh, ssqdfsc, &
   y1, y2, Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  use modelfcns
  implicit none
  integer, intent(in) :: np, n, ifam
  logical, intent(in) :: lmxi
  double precision, intent(in) :: y1(n), y2(n), Ups(n, n), nu, &
     xi(n), tsq, tsqdf, ssqdfh, ssqdfsc
  double precision, intent(in) :: logssqg(np+np+1), &
     meang(n,np+np+1), prechg(n,n,np+np+1)
  double precision, intent(out) :: dnu(np+np+1)
  double precision par(n), z(n), dz_dnu(n), dpar_dnu(n), &
     dpz_dz(n), ssq
  double precision dpym_dpar(n), dpym_hpar(n), dpym_3par(n)
  double precision invlink_dz(n), invlink_dn(n), invlink_hz(n), &
     invlink_dzdn(n), invlink_3z(n), invlink_hzdn(n)
  double precision dH(n)
  double precision varh(n,n)
  integer i, j
  do i = 1, np+np+1
    z = meang(:,i)
    ssq = exp(logssqg(i))
    varh = prechg(:,:,i)
    call dtrtri ('u','n',n,varh,n,j)
    if (j .ne. 0) call rexit ("aloglik_dnu - Non-invertible precision.")
    par = invlink(z,nu)
    invlink_dz = invlinkdz(z,nu)
    invlink_dn = invlinkdn(z,nu)
    invlink_hz = invlinkhz(z,nu)
    invlink_dzdn = invlinkdzdn(z,nu)
    invlink_3z = invlink3z(z,nu)
    invlink_hzdn = invlinkhzdn(z,nu)
    dpym_dpar = logpdfydlnk(y1,y2,par)
    dpym_hpar = logpdfyhlnk(y1,y2,par)
    dpym_3par = logpdfy3lnk(y1,y2,par)
    dpz_dz = logpdfzdz(z, Ups, xi, lmxi, ssq, n)
    dz_dnu = fdz_dnu(z, par, ssq, varh, &
       y1, y2, Ups, nu, xi, lmxi, tsq, tsqdf, n)
    dpar_dnu = invlink_dn + invlink_dz*dz_dnu
    dH = -dpym_3par*invlink_dz*invlink_dz*dpar_dnu &
       - 3d0*dpym_hpar*invlink_dz*invlink_hz*dz_dnu &
       - 2d0*dpym_hpar*invlink_dz*invlink_dzdn &
       - dpym_hpar*invlink_hz*invlink_dn &
       - dpym_dpar*invlink_3z*dz_dnu &
       - dpym_dpar*invlink_hzdn
    dH = dH/tsq
    dnu(i) = dot_product(dpym_dpar, dpar_dnu)/tsq &
       + dot_product(dpz_dz, dz_dnu) - .5d0*traceH(varh,dH,n)
  end do

contains
  function logpdfzdz (z, Ups, xi, lmxi, ssq, n) result (gr)
!! log-pdf of z and its derivative after integrating out beta. The 2*pi
!! constant is removed.
    implicit none
    integer, intent(in) :: n
    logical, intent(in) :: lmxi
    double precision, intent(in) :: z(n), Ups(n,n), xi(n), ssq
    double precision :: gr(n)
    double precision zmxi(n)
    if (lmxi) then
      zmxi = z - xi
    else
      zmxi = z
    end if
    call dsymv ('u',n,1d0,Ups,n,zmxi,1,0d0,gr,1) ! gr = Ups*(z-xi)
    gr = -gr/ssq ! gr = -Ups*(z-x)/ssq
  end function logpdfzdz

  function fdz_dnu (z, par, ssq, varh, y1, y2, Ups, &
     nu, xi, lmxi, tsq, tsqdf, n)
    implicit none
    integer, intent(in) :: n
    logical, intent(in) :: lmxi
    double precision, intent(in) :: y1(n), y2(n), Ups(n,n), nu, &
       xi(n), tsq, tsqdf, ssq, z(n), varh(n,n), par(n)
    double precision, dimension(n) :: fdz_dnu
    fdz_dnu = logpdfyhlnk(y1,y2,par) * invlinkdz(z,nu) * invlinkdn(z,nu) &
       + logpdfydlnk(y1,y2,par) * invlinkdzdn(z,nu)
    fdz_dnu = fdz_dnu/tsq
    call dtrmv ('u','t','n',n,varh,n,fdz_dnu,1)
    call dtrmv ('u','n','n',n,varh,n,fdz_dnu,1)
  end function fdz_dnu

  function traceH (varh, dh, n)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: varh(n,n), dh(n)
    double precision traceH
    integer i
    double precision v(n)
    traceH = 0d0
    do i = 1, n
      v(1:i) = dh(1:i)*varh(1:i,i)
      traceH = traceH + dot_product(varh(1:i,i),v(1:i))
    end do
  end function traceH
end subroutine aloglik_dnu


subroutine aloglik_dcov (np, logssqg, dcov, ideriv, &
   meang, prechg, ssqdfh, ssqdfsc, &
   y1, y2, dm, phi, omg, kappa, Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  use modelfcns
  use covfun
  use calcbd_fcns, only: qform, traceAB, cor_dcov
  implicit none
  integer, intent(in) :: np, n, ifam, ideriv
  logical, intent(in) :: lmxi
  double precision, intent(in) :: y1(n), y2(n), Ups(n, n), nu, &
     xi(n), tsq, tsqdf, ssqdfh, ssqdfsc
  double precision, intent(in) :: dm(n,n), phi, omg, kappa
  double precision, intent(in) :: logssqg(np+np+1), &
     meang(n,np+np+1), prechg(n,n,np+np+1)
  double precision, intent(out) :: dcov(np+np+1)
  double precision par(n), z(n), dz_dcov(n), dpar_dcov(n), &
     dpz_dz(n), ssq
  double precision dpym_dpar(n), dpym_hpar(n), dpym_3par(n)
  double precision invlink_dz(n), invlink_hz(n), invlink_3z(n)
  double precision dH1(n), dH(n,n)
  double precision varh(n,n), DT(n,n), trUpsDTh, UpsDTUps(n,n), DTUps(n,n)
  double precision dpz_dcov
  integer i, j
  DT = cor_dcov(n,dm,phi,omg,kappa,ideriv)
  trUpsDTh = .5d0*traceAB(Ups,DT,n)
  call fill_symmetric_matrix(DT,n)
  call dsymm ('r','u',n,n,1d0,Ups,n,DT,n,0d0,DTUps,n)
  call dsymm ('l','u',n,n,1d0,Ups,n,DTUps,n,0d0,UpsDTUps,n)
  do i = 1, np+np+1
    z = meang(:,i)
    ssq = exp(logssqg(i))
    dH = UpsDTUps/ssq
    varh = prechg(:,:,i)
    call dtrtri ('u','n',n,varh,n,j)
    if (j .ne. 0) call rexit ("aloglik_dcov - Non-invertible precision.")
    par = invlink(z,nu)
    invlink_dz = invlinkdz(z,nu)
    invlink_hz = invlinkhz(z,nu)
    invlink_3z = invlink3z(z,nu)
    dpym_dpar = logpdfydlnk(y1,y2,par)
    dpym_hpar = logpdfyhlnk(y1,y2,par)
    dpym_3par = logpdfy3lnk(y1,y2,par)
    dpz_dz = logpdfzdz(z, Ups, xi, lmxi, ssq, n)
    dpz_dcov = logpdfzdcov(z, xi, lmxi, dH, trUpsDth, n)
    dz_dcov = fdz_dcov(z, par, ssq, varh, &
       y1, y2, dH, nu, xi, lmxi, tsq, tsqdf, n)
    dpar_dcov = invlink_dz*dz_dcov
    dH1 = -dpym_3par*invlink_dz*invlink_dz*dpar_dcov &
       - 3d0*dpym_hpar*invlink_dz*invlink_hz*dz_dcov &
       - dpym_dpar*invlink_3z*dz_dcov
    dH1 = dH1/tsq
    dH = -dH
    do j = 1, n
      dH(j,j) = dH(j,j) + dH1(j)
    end do
    dcov(i) = dpz_dcov &
       + dot_product(dpym_dpar, dpar_dcov)/tsq &
       + dot_product(dpz_dz, dz_dcov) &
       - .5d0*traceH(varh,dH,n)
  end do

contains
  function logpdfzdz (z, Ups, xi, lmxi, ssq, n) result (gr)
!! log-pdf of z and its derivative after integrating out beta. The 2*pi
!! constant is removed.
    implicit none
    integer, intent(in) :: n
    logical, intent(in) :: lmxi
    double precision, intent(in) :: z(n), Ups(n,n), xi(n), ssq
    double precision :: gr(n)
    double precision zmxi(n)
    if (lmxi) then
      zmxi = z - xi
    else
      zmxi = z
    end if
    call dsymv ('u',n,1d0,Ups,n,zmxi,1,0d0,gr,1) ! gr = Ups*(z-xi)
    gr = -gr/ssq ! gr = -Ups*(z-x)/ssq
  end function logpdfzdz

  function logpdfzdcov (z, xi, lmxi, UpsDTUps, trUpsDTh, n)
!! log-pdf of z and its derivative after integrating out beta. The 2*pi
!! constant is removed.
    implicit none
    integer, intent(in) :: n
    logical, intent(in) :: lmxi
    double precision, intent(in) :: z(n), xi(n), UpsDTUps(n,n), trUpsDTh
    double precision :: logpdfzdcov
    double precision zmxi(n), zUDTUz
    if (lmxi) then
      zmxi = z - xi
    else
      zmxi = z
    end if
    zUDTUz = .5d0*qform(zmxi,UpsDTUps,n)
    logpdfzdcov = -trUpsDTh + zUDTUz
  end function logpdfzdcov

  function fdz_dcov (z, par, ssq, varh, y1, y2, UpsDTUps, &
     nu, xi, lmxi, tsq, tsqdf, n)
    implicit none
    integer, intent(in) :: n
    logical, intent(in) :: lmxi
    double precision, intent(in) :: y1(n), y2(n), UpsDTUps(n,n), nu, &
       xi(n), tsq, tsqdf, ssq, z(n), varh(n,n), par(n)
    double precision, dimension(n) :: fdz_dcov
    double precision zmxi(n)
    if (lmxi) then
      zmxi = z - xi
    else
      zmxi = z
    end if
    call dsymv ('u',n,1d0,UpsDTUps,n,zmxi,1,0d0,fdz_dcov,1) ! Ups*(z-xi)
    call dtrmv ('u','t','n',n,varh,n,fdz_dcov,1)
    call dtrmv ('u','n','n',n,varh,n,fdz_dcov,1)
  end function fdz_dcov

  function traceH (varh, dH, n)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: varh(n,n), dH(n,n)
    double precision traceH
    integer i
    double precision v(n,n)
    v = dH
    call dtrmm ('r','u','n','n',n,n,1d0,varh,n,v,n)
    traceH = 0d0
    do i = 1, n
      traceH = traceH + dot_product(varh(1:i,i),v(1:i,i))
    end do
  end function traceH
end subroutine aloglik_dcov


subroutine llikpars2 (fval, gval, lderiv, &
   nu, phi, omg, kappa, y1, y2, F, offset, betm0, betQ0, &
   ssqdf, ssqsc, dm, tsq, tsqdf, n, p, np, ssqin, ifam, icf)
  use covfun, only: create_spcor, calc_cov
  use betaprior
  implicit none
  integer, intent(in) :: np, n, p, ifam, icf
  logical, intent(in) :: lderiv(4) ! Which derivatives to calculate?
  double precision, intent(in) :: ssqin, ssqdf, ssqsc, &
     y1(n), y2(n), phi, nu, omg, kappa, dm(n,n), F(n,p), &
     betm0(p), betQ0(p,p), tsq, tsqdf, offset(n)
  double precision, intent(out) :: fval, gval(4)
  double precision Ups(n,n), ldh_Ups, xi(n), modeldfh, ssqdfh, ssqdfsc
  integer i
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p)
  double precision tg(np+np+1), logw(np+np+1), &
     meang(n,np+np+1), prechg(n,n,np+np+1)
  double precision, dimension(np+np+1) :: w_dnu, w_dphi, w_dnsq, w_dkap
  double precision ssqst
  call create_spcor(icf,n)
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)
  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  ssqdfh = .5d0*ssqdf
  ssqdfsc = ssqdf*ssqsc
  ssqst = ssqstart(y1,y2,nu,Ups,n,ifam)
  ! TODO ssqin or ssqst?
  ! ssqst = ssqin
  i = np+np+1
  call gridposter (np, tg, logw, meang, prechg, ssqdfh, ssqdfsc, &
     ssqst, y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  fval = trapezoid_lf(i, tg, logw)
  gval = 0d0
  if (lderiv(1)) then
    call aloglik_dnu (np, tg, w_dnu, meang, prechg, ssqdfh, ssqdfsc, &
       y1, y2, Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
    w_dnu = w_dnu*exp(logw - fval)
    gval(1) = trapezoid_f(i, tg, w_dnu)
  end if
  if (lderiv(2)) then
    call aloglik_dcov (np, tg, w_dphi, 1, &
       meang, prechg, ssqdfh, ssqdfsc, &
       y1, y2, dm, phi, omg, kappa, Ups, nu, &
       xi, lmxi, tsq, tsqdf, n, ifam)
    w_dphi = w_dphi*exp(logw - fval)
    gval(2) = trapezoid_f(i, tg, w_dphi)
  end if
  if (lderiv(3)) then
    call aloglik_dcov (np, tg, w_dnsq, 2, &
       meang, prechg, ssqdfh, ssqdfsc, &
       y1, y2, dm, phi, omg, kappa, Ups, nu, &
       xi, lmxi, tsq, tsqdf, n, ifam)
    w_dnsq = w_dnsq*exp(logw - fval)
    gval(3) = trapezoid_f(i, tg, w_dnsq)
  end if
  if (lderiv(4)) then
    call aloglik_dcov (np, tg, w_dkap, 3, &
       meang, prechg, ssqdfh, ssqdfsc, &
       y1, y2, dm, phi, omg, kappa, Ups, nu, &
       xi, lmxi, tsq, tsqdf, n, ifam)
    w_dkap = w_dkap*exp(logw - fval)
    gval(4) = trapezoid_f(i, tg, w_dkap)
  end if
  ! print*, real(fval), real(gval)

contains
  function trapezoid_lf (n, x, f)
    ! Compute the log integral of exp(f) along x
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: x(1:n), f(1:n)
    double precision trapezoid_lf
    double precision fpf(1:n-1), xmx(1:n-1), fmx, fmm(1:n), ln2
    parameter (ln2 = 0.6931471805599453d0)
    fmx = maxval(f)
    fmm = f - fmx
    fmm = exp(fmm)
    fpf = fmm(2:n) + fmm(1:n-1)
    xmx = x(2:n) - x(1:n-1)
    trapezoid_lf = dot_product(xmx,fpf)
    trapezoid_lf = log(trapezoid_lf) + fmx - ln2
  end function trapezoid_lf

  function trapezoid_f (n, x, f)
    ! Compute the integral of f along x
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: x(1:n), f(1:n)
    double precision trapezoid_f
    double precision fpf(1:n-1), xmx(1:n-1)
    fpf = f(2:n) + f(1:n-1)
    xmx = x(2:n) - x(1:n-1)
    trapezoid_f = dot_product(xmx,fpf)*.5d0
  end function trapezoid_f

  function ssqstart (y1,y2,nu,Ups,n,ifam)
    use linkfcns, only: flink_ga
    use modelfcns, only: mustart, flink_sp => flink
    implicit none
    integer, intent(in) :: n, ifam
    double precision, intent(in) :: y1(n), y2(n), nu, Ups(n,n)
    double precision ssqstart
    double precision, dimension(n) :: z, tmp
    double precision zbar
    if (ifam .eq. 0) then ! Gaussian transformed
      z = flink_ga(y1,nu) ! Starting value
    else
      tmp = mustart(y1,y2)
      z = flink_sp(tmp,nu) ! Starting value
    end if
    zbar = sum(z)/n
    z = z - zbar
    call dsymv("u",n,1d0,Ups,n,z,1,0d0,tmp,1)
    ssqstart = dot_product(z,tmp)
    ssqstart = ssqstart/n
  end function ssqstart
end subroutine llikpars2

subroutine llikparsval (fval, gval, ideriv, &
   nu, phi, omg, kappa, y1, y2, F, offset, betm0, betQ0, &
   ssqdf, ssqsc, dm, tsq, tsqdf, n, p, np, ssqin, ifam, icf)
  use modelfcns, only: create_model
  implicit none
  integer, intent(in) :: np, n, p, ifam, icf, ideriv(4)
  double precision, intent(in) :: ssqin, ssqdf, ssqsc, &
     y1(n), y2(n), phi, nu, omg, kappa, dm(n,n), F(n,p), &
     betm0(p), betQ0(p,p), tsq, tsqdf, offset(n)
  double precision, intent(out) :: fval, gval(4)
  logical lderiv(4)
  call create_model (ifam)
  lderiv = ideriv .ne. 0
  call llikpars2 (fval, gval, lderiv, &
     nu, phi, omg, kappa, y1, y2, F, offset, betm0, betQ0, &
   ssqdf, ssqsc, dm, tsq, tsqdf, n, p, np, ssqin, ifam, icf)
end subroutine llikparsval

subroutine aloglikval (fval, gval, &
   nu, phi, omg, kappa, y1, y2, F, offset, betm0, betQ0, &
   ssqdf, ssqsc, dm, tsq, tsqdf, n, p, np, logssqg, ifam, icf)
  use modelfcns, only: create_model
  use covfun, only: create_spcor, calc_cov
  use betaprior
  implicit none
  integer, intent(in) :: np, n, p, ifam, icf
  double precision, intent(in) :: logssqg(np+np+1), ssqdf, ssqsc, &
     y1(n), y2(n), phi, nu, omg, kappa, dm(n,n), F(n,p), &
     betm0(p), betQ0(p,p), tsq, tsqdf, offset(n)
  double precision, intent(out) :: fval(np+np+1), gval(np+np+1)
  double precision Ups(n,n), ldh_Ups, xi(n), modeldfh, ssqdfh, ssqdfsc
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p)
  double precision meang(n,np+np+1), prechg(n,n,np+np+1)
  call create_model (ifam)
  call create_spcor(icf,n)
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)
  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  ssqdfh = .5d0*ssqdf
  ssqdfsc = ssqdf*ssqsc
  call aloglik (np, logssqg, fval, meang, prechg, ssqdfh, ssqdfsc, &
     y1, y2, Ups, ldh_Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
  call aloglik_dnu (np, logssqg, gval, meang, prechg, ssqdfh, ssqdfsc, &
     y1, y2, Ups, nu, xi, lmxi, tsq, tsqdf, n, ifam)
end subroutine aloglikval

subroutine llikparscalc (fval, nu, phi, omg, kappa, npars, &
   y1, y2, F, offset, betm0, betQ0, &
   ssqdf, ssqsc, dm, tsq, tsqdf, n, p, np, ssqin, ifam, icf)
  ! Calls subroutine llikpars2 repeatedly for each parameter input.
  use modelfcns
  implicit none
  integer, intent(in) :: np, n, p, ifam, icf, npars
  double precision, intent(in) :: ssqin, ssqdf, ssqsc, &
     y1(n), y2(n), phi(npars), nu(npars), omg(npars), kappa(npars), &
     dm(n,n), F(n,p), betm0(p), betQ0(p,p), tsq, tsqdf, offset(n)
  double precision, intent(out) :: fval(npars)
  integer i
  logical lderiv(4)
  double precision gval(4)
  call create_model (ifam)
  lderiv = .false.
  do i = 1, npars
    call llikpars2 (fval(i), gval, lderiv, &
       nu(i), phi(i), omg(i), kappa(i), y1, y2, F, offset, betm0, betQ0, &
       ssqdf, ssqsc, dm, tsq, tsqdf, n, p, np, ssqin, ifam, icf)
  end do
end subroutine llikparscalc


! Local Variables:
! compile-command: "gfortran -c -fpic -Wunused-parameter -Wall \
!   -pedantic -o skelpnts.o skelpnts.f90"
! End:

######################################################################
##
### Commentary: Functions to simulate from the model.
##
######################################################################

##' Simulate from a variety of spatial models.
##'
##' The spatial Gaussian random field is simulated using the Cholesky
##' decomposition of the covariance matrix.
##'
##' The sample from a quasi distribution uses a hack which matches the
##' mean and the variance of the distribution. See the source code for
##' details.
##' @title Simulation from a spatial model
##' @param n The number of instances to simulate
##' @param formula A representation of the model in the form
##'   \code{response ~ terms}. The LHS can be omitted. If the LHS
##'   exists, it can be of the form \code{y}, \code{y|z}, or sums of
##'   terms at either side of the \code{|} to specify the names of the
##'   variables to include in the data frame.
##' @param family The distribution of the data to simulate from.
##' @param data An optional data frame containing the variables in the
##'   model.
##' @param weights An optional vector of weights. Number of replicated
##'   samples for Gaussian and gamma, number of trials for binomial,
##'   time length for Poisson.
##' @param subset An optional set of indices. Simulations will be
##'   provided for those locations only.
##' @param offset See \code{\link[stats]{lm}}.
##' @param atsample A formula of the form \code{~ Xcoord + Ycoord}
##'   specifying the sampled locations.
##' @param beta A vector of the regressor coefficents to use.
##' @param linkp The link function parameter.
##' @param phi The spatial range parameter.
##' @param omg The relative nugget parameter.
##' @param kappa The spatial smoothness parameter.
##' @param ssq The partial sill parameter.
##' @param corrfcn The correlation function to use.
##' @param longlat How to compute the distance between locations. If
##'   \code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
##'   distance. See \code{\link[sp]{spDists}}.
##' @param dispersion The fixed dispersion parameter. When this is not
##'   1 and the sample is from a binomial or a Poisson distribution,
##'   no such distribution exists so an approximate sample is
##'   returned. Use with caution.
##' @param returnGRF Whether to return the simulate Gaussian random
##'   field as well.
##' @param warndisp Whether to warn when sampling from a quasi
##'   distribution. This is the case for binomial and Poisson when the
##'   dispersion is not one.
##' @return A data frame containing the predictors, sampling
##'   locations, optional weights, and samples.
##' @name rsglmm
##' @examples \dontrun{
##' n <- 100
##' beta <- c(-2, 1)
##' phi <- .2
##' omg <- .3
##' linkp <- 0
##' ssq <- 1
##' l <- rep(10, n)
##' corrf <- "matern"
##' kappa <- .5
##' family <- "poisson"
##' Xcoord <- runif(n)
##' Ycoord <- runif(n)
##' f <- Xcoord + Ycoord
##' formula <- y|z ~ f
##' mydata <- rsglmm(1, formula, family, weights = l,
##'                  atsample = ~ Xcoord + Ycoord, beta = beta, linkp = linkp,
##'                  phi = phi, omg = omg, kappa = kappa, ssq = ssq,
##'                  corrfcn = corrf, returnGRF = TRUE)
##' }
##' @export
##' @importFrom stats rnorm rbinom rpois rgamma rbeta model.frame
##'   update as.formula model.offset
##' @importFrom sp spDists
rsglmm <- function(n, formula,
                   family = "gaussian",
                   data, weights, subset, offset, atsample,
                   beta, linkp, phi, omg, kappa, ssq,
                   corrfcn = "matern",
                   longlat = FALSE, dispersion = 1, returnGRF = FALSE,
                   warndisp = TRUE) {

  ## Family
  ifam <- .geoBayes_family(family)
  if (ifam) {
    family <- .geoBayes_models$family[ifam]
  } else {
    stop ("You cannot simulate transformed Gaussian using this function.")
  }

  ## Logical input
  returnGRF <- as.logical(returnGRF)
  if (length(returnGRF) != 1 || is.na(returnGRF))
    stop ("Input returnGRF must be a single logical value.")
  warndisp <- as.logical(warndisp)
  if (length(warndisp) != 1 || is.na(warndisp))
    stop ("Input warndisp must be a single logical value.")

  ## Check dispersion input
  dispersion <- as.double(dispersion)
  if (dispersion <= 0) stop ("Invalid argument dispersion")

  ## Check the link parameter
  nu <- .geoBayes_getlinkp(linkp, ifam)

  ## kappa
  if (missing(kappa)) kappa <- NA

  ## Response name
  formula <- as.formula(formula)
  if (length(formula) == 2) { # No LHS
    ## RHS <- formula[[2]]
    nmy <- "y"
    nmz <- "z"
  } else { # LHS names provided
    LHS <- formula[[2]]
    ## RHS <- formula[[3]]
    if ("|" %in% all.names(LHS)) { # LHS contains bar
      nmy <- all.vars(LHS[[2]])    # Before the bar
      nmz <- all.vars(LHS[[3]])    # After the bar
    } else {
      nmy <- all.vars(LHS)
      nmz <- "z"
    }
  }
  formula0 <- update(formula, NULL ~ .)

  ## Design matrix and data
  if (missing(data)) data <- environment(formula)
  mfc <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data", "subset", "weights", "offset"),
             names(mfc), 0L)
  mfc <- mfc[c(1L, m)]
  mfc$formula <- formula0
  mfc$drop.unused.levels <- TRUE
  mfc$na.action <- "na.fail"
  mfc[[1L]] <- quote(stats::model.frame)
  mf <- eval(mfc, parent.frame())
  mt <- attr(mf, "terms")
  FF <- model.matrix(mt,mf)
  nloc <- nrow(FF)
  offset <- as.vector(model.offset(mf))
  if (!is.null(offset)) {
    if (length(offset) != nloc) 
      stop(gettextf("number of offsets is %d, should equal %d (number of observations)", 
                    length(offset), nloc), domain = NA)
  }

  ## Weights
  weightsc <- mfc$weights
  noweights <- is.null(weightsc)
  if (noweights) {
    ll <- rep.int(1, nloc)
    nmweights <- weights <- NULL
  } else {
    ll <- model.weights(mf)
    if (any(!is.finite(ll))) {
      stop ("Non-finite values in the weights")
    } else if (any(ll <= 0)) {
      stop ("Non-positive weights not allowed")
    }
    weights <- eval(weightsc, envir = data)
    if (length(weightsc) > 1) {
      nmweights <- "(weights)"
    } else {
      nmweights <- as.character(weightsc)
    }
    weights <- list(weights)
    names(weights) <- nmweights
  }

  ## All locations
  atsample <- update(atsample, NULL ~ . + 0) # No response and no intercept
  mfatc <- mfc
  mfatc$weights = NULL
  mfatc$formula = atsample
  mfat <- eval(mfatc, parent.frame())
  loc <- as.matrix(mfat)
  if (!all(is.finite(loc))) stop ("Non-finite values in the locations")

  ## Simulate
  z <- randgrf(n, loc, beta, phi, omg, kappa, ssq, FF, offset, corrfcn, longlat)
  pary <- linkinv(z, linkp, family)
  if (warndisp && (dispersion != 1) &&
      grepl("^(binomial|poisson)(\\..+)?$", family))
    warning ("Sampling from a quasi distribution: results are approximate.")
  distny <- gsub("\\.[[:alnum:]]+", "", family)
  y <- if (n > 0)
         switch(distny,
                gaussian = randgauss(n, nloc, ll, pary, dispersion),
                binomial = randqbinom(n, nloc, ll, pary, dispersion),
                poisson = randqpois(n, nloc, ll, pary, dispersion),
                Gamma = randgamma(n, nloc, ll, pary, dispersion))

  ## Return data.frame
  varsnm <- unique(c(all.vars(formula0), all.vars(atsample)))
  varsfm <- as.formula(paste("~", paste(varsnm, collapse = "+"), "+0"))
  varscall <- list(quote(stats::model.frame), formula = varsfm, data = data)
  outdf <- eval(as.call(varscall), parent.frame())
  varsnm <- names(outdf)
  if (!noweights && !(nmweights %in% varsnm)) {
    outdf <- cbind(outdf, weights)
    varsnm <- c(varsnm, nmweights)
  }
  NN <- nrow(outdf)
  nmall <- make.names(c(varsnm, rep(nmy, length = n),
                        if(returnGRF) rep(nmz, length = n)), unique = TRUE)
  nmyz <- nmall[(ncol(outdf) + 1):(length(nmall))]
  if (missing(subset) || is.null(subset)) subset <- 1:NN
  dfy <- matrix(NA, NN, n)
  dfy[subset, ] <- y
  if (returnGRF) {
    dfz <- matrix(NA, NN, n)
    dfz[subset, ] <- z
  } else dfz <- NULL
  dfyz <- cbind(dfy, dfz)
  colnames(dfyz) <- nmyz
  outdf <- cbind(outdf, dfyz)
  outdf
}

##' @name rsglmm
##' @export
rstrga <- function(n, formula, data, weights, subset, offset, atsample,
                   beta, linkp, phi, omg, kappa, ssq,
                   corrfcn = "matern",
                   longlat = FALSE, dispersion = 1, returnGRF = FALSE) {

  call <- match.call()
  call[[1]] <- quote(rsglmm)
  call$family <- "gaussian"
  call$warndisp <- FALSE
  eval(call, parent.frame())
}


##' @name rsglmm
##' @export
rsgrf <- function (n, formula, data, subset, offset, atsample,
                   beta, phi, omg, kappa, ssq,
                   corrfcn = "matern",
                   longlat = FALSE) {
  
  ## kappa
  if (missing(kappa)) kappa <- NA

  ## Response name
  formula <- as.formula(formula)
  if (length(formula) == 2) { # No LHS
    ## RHS <- formula[[2]]
    nmz <- "z"
  } else { # LHS names provided
    LHS <- formula[[2]]
    ## RHS <- formula[[3]]
    if ("|" %in% all.names(LHS)) { # LHS contains bar
      nmz <- all.vars(LHS[[3]])    # After the bar
    } else {
      nmz <- all.vars(LHS)
    }
  }
  formula0 <- update(formula, NULL ~ .)

  ## Design matrix and data
  if (missing(data)) data <- environment(formula)
  mfc <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data", "subset", "weights", "offset"),
             names(mfc), 0L)
  mfc <- mfc[c(1L, m)]
  mfc$formula <- formula0
  mfc$drop.unused.levels <- TRUE
  mfc$na.action <- "na.fail"
  mfc[[1L]] <- quote(stats::model.frame)
  mf <- eval(mfc, parent.frame())
  mt <- attr(mf, "terms")
  FF <- model.matrix(mt,mf)
  nloc <- nrow(FF)
  offset <- as.vector(model.offset(mf))
  if (!is.null(offset)) {
    if (length(offset) != nloc) 
      stop(gettextf("number of offsets is %d, should equal %d (number of observations)", 
                    length(offset), nloc), domain = NA)
  }

  ## All locations
  atsample <- update(atsample, NULL ~ . + 0) # No response and no intercept
  mfatc <- mfc
  mfatc$weights = NULL
  mfatc$formula = atsample
  mfat <- eval(mfatc, parent.frame())
  loc <- as.matrix(mfat)
  if (!all(is.finite(loc))) stop ("Non-finite values in the locations.")

  z <- randgrf(n, loc, beta, phi, omg, kappa, ssq, FF, offset, corrfcn, longlat)

  ## Return data.frame
  varsnm <- unique(c(all.vars(formula0), all.vars(atsample)))
  formulavars <- as.formula(paste("~", paste(varsnm, collapse = "+"), "+0"))
  varscall <- list(quote(stats::model.frame),
                   formula = formulavars, data = data)
  outdf <- eval(as.call(varscall),parent.frame())
  varsnm <- names(outdf)
  NN <- nrow(outdf)
  nmall <- make.names(c(varsnm, rep(nmz, length = n)), unique = TRUE)
  nmz <- nmall[(ncol(outdf) + 1):(length(nmall))]
  if (missing(subset) || is.null(subset)) subset <- 1:NN
  dfz <- matrix(NA, NN, n)
  dfz[subset, ] <- z
  colnames(dfz) <- nmz
  outdf <- cbind(outdf, dfz)
  outdf
}


##' @useDynLib geoBayes spcorr
randgrf <- function (n, loc, beta, phi, omg, kappa, ssq, FF, offset = NULL, 
                     corrfcn = "matern",
                     longlat = FALSE) {

  ## Check sample size
  n <- as.integer(n)
  if (n < 0) stop ("Negative sample size n entered.")

  ## Logical input
  longlat <- as.logical(longlat)
  if (length(longlat) != 1 || is.na(longlat))
    stop ("Input longlat must be a single logical value.")

  ## Correlation function
  icf <- .geoBayes_correlation(corrfcn)
  corrfcn <- .geoBayes_corrfcn$corrfcn[icf]
  kappa <- .geoBayes_getkappa(kappa, icf)
  phi <- as.double(phi)
  if (phi < 0) stop ("Input phi must >= 0")
  omg <- as.double(omg)
  if (omg < 0) stop ("Input omg must >= 0")

  ## Check beta, ssq
  beta <- as.vector(beta)
  ssq <- as.double(ssq)
  if (ssq <= 0) stop ("Non positive ssq entered.")

  ## Check loc
  loc <- as.matrix(loc)
  if (!all(is.finite(loc))) stop ("Non-finite values in the locations.")
  if (corrfcn == "spherical" && ncol(loc) > 3) {
    stop ("Cannot use the spherical correlation for dimensions
grater than 3.")
  }

  ## Check model matrix
  FF <- as.matrix(FF)
  if (!all(is.finite(FF))) stop ("Non-finite values in the design matrix.")
  nloc <- nrow(FF)
  p <- ncol(FF)
  if (length(beta) != p)
    stop ("The length of beta does not match the number of predictors.")

  ## Covariance matrix
  dm <- sp::spDists(loc, longlat = longlat)
  dmup <- dm[upper.tri(dm)]
  ndm <- as.integer(length(dmup))
  spcorr <- .Fortran("spcorr", dmup, as.double(phi), as.double(kappa),
                     ndm, icf, PACKAGE = "geoBayes")[[1]]
  Sg <- diag(ssq*(1 + omg), nloc, nloc)
  Sg[upper.tri(Sg)] <- ssq*spcorr
  Sgch <- chol(Sg)

  ## Simulate
  meanz <- as.vector(FF %*% beta)
  z <- matrix(rnorm(n*nloc), nloc, n)
  z <- meanz + crossprod(Sgch, z)
  if (!is.null(offset)) z <- z + offset
  z
}

randgauss <- function (n, ns, weight, par, dispersion) {
  mean <- weight*par
  sd <- sqrt(weight*dispersion)
  y <- matrix(rnorm(n*ns, mean, sd), ns, n)
  y
}

randqbinom <- function (n, ns, weight, par, dispersion) {
  if (dispersion == 1) {
    y <- matrix(rbinom(n*ns, weight, par), ns, n)
  } else {
    ## y = dispersion*Bin(m/dispersion,p) +
    ## dispersion*eps*Beta(1+p,2-p)
    mt <- weight/dispersion
    mtm <- floor(mt)
    y <- matrix(rbinom(ns*n, mtm, par), ns, n)*dispersion
    ii <- mt > mtm
    ni <- sum(ii)
    if (ni > 0) {
      eps <- weight[ii] %% dispersion
      bet1 <- 1 + par[ii]
      bet2 <- 2 - par[ii]
      betr <- matrix(rbeta(ni*n, bet1, bet2), ni, n)*eps
      y[ii, ] <- y[ii, ] + betr
    }
  }
  y
}

randqpois <- function (n, ns, weight, par, dispersion) {
  if (dispersion == 1) {
    mean <- weight*par
    y <- matrix(rpois(n*ns, mean), ns, n)
  } else {
    mean <- weight*par/dispersion
    y <- matrix(rpois(n*ns, mean), ns, n)*dispersion
  }
  y
}

randgamma <- function (n, ns, weight, par, dispersion) {
  scale <- par*dispersion
  shape <- weight/dispersion
  y <- matrix(rgamma(n*ns, shape = shape, scale = scale), n, ns)
  y
}

##' Perform the reverse logistic regression estimation
##'
##' Estimation is done by maximising the reverse logistic log likelihood.
##' @title Reverse logistic regression estimation
##' @param lglk The value of the loglikelihood at different samples and
##'   different parameters. This should be entered as a matrix where the
##'   rows are the values of the samples and the columns correspond to
##'   the parameters. The
##'   [i,j] element of the matrix is the value of the loglikelihood at
##'   the ith sample when all samples are put together evaluated
##'   at the jth parameter value.
##' @param N A vector of length ncol(lglk) or a scalar corresponding to
##'   the sample sizes from each model. Must sum(N) == nrow(lglk). The
##'   first N[1] samples come from model corresponding to the first set
##'   of parameters, then (N[1]+1):N[2] are from the model corresponding
##'   to the second set of parameters, and so on.
##' @return A vector containing the reverse logistic regression estimates
##'   of the logarithm of the Bayes factors.
##'   The first set of parameters is taken as the reference model so its
##'   estimate is always 0.
##' @references Geyer, C. J. (1994). Estimating normalizing constants
##'   and reweighting mixtures in Markov chain Monte Carlo. Technical
##'   Report 568, School of Statistics, University of Minnesota.
##' @importFrom stats optim
##' @export
revlogreg <- function(lglk, N)
{
  ## Purpose: Reverse logistic regression
  ## ----------------------------------------------------------------------
  ## Arguments:
  ## lglk   The log-likelihood. Each column corresponds to
  ## the value of the log-likelihood evaluated at different parameters.
  ## N   A vector of length ncol(lglk) or a scalar corresponding to
  ## the sample sizes from each model. sum(N) == nrow(lglk)

  lglk <- as.matrix(lglk)
  nruns <- ncol(lglk)
  Ntot <- nrow(lglk)
  N <- rep(as.integer(N), length = nruns)
  if (sum(N) != Ntot) stop ("Inconsistent sample sizes implied by lglk and N")
  ## if (any(!is.finite(lglk))) stop ("Log-likelihood contains non-finite values")

  if (nruns == 1) return (0)

  isum <- matrix(rep(c(TRUE, rep(FALSE, nruns)), length.out = nruns*nruns),
                 nruns)[rep(1:nruns, N), ]

  fun <- function (eta) {
    lliketa <- lglk + matrix(eta, Ntot, nruns, TRUE)
    mxlliketa <- apply(lliketa, 1, max)
    lliketammx <- lliketa - mxlliketa
    lgdenom <- mxlliketa + log(rowSums(exp(lliketammx)))
    lgnum <- lliketa[isum]
    out <- sum(lgnum - lgdenom)
    -out
  }

  gun <- function (eta) {
    lliketa <- lglk + matrix(eta, Ntot, nruns, TRUE)
    mxlliketa <- apply(lliketa, 1, max)
    lliketammx <- lliketa - mxlliketa
    lgp <- lliketammx - log(rowSums(exp(lliketammx)))
    out <- N - colSums(exp(lgp))
    -out
  }

  op <- stats::optim(log(N/Ntot), fun, gun, method = "BFGS")
  out <- op$par - log(N/Ntot)
  out[1] - out
}

subroutine llikfcn_no (lglk, philist, omglist, nulist, kappalist, &
   zsample, Ntot, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, kg, ifam, itr)

  use modelfcns, jointyz_sp => jointyz
  use covfun
  use jointyz, only: jointyz_gt
  use betaprior
  implicit none
  integer, intent(in) :: n, p, kg, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     zsample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot, kg)
  logical lmxi
  double precision T(n, n), TiF(n, p), FTF(p, p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, &
     tsqdfsc, respdfh, xi(n)
  integer i, j

  call create_model (ifam)
  call create_spcor (icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  select case (ifam)
  case (0)
    do i = 1, kg
      call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
         kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        call rchkusr
        lglk(j,i) = jointyz_gt(n, zsample(:,j), y, l, Ups, ldh_Ups, &
           nulist(i), xi, lmxi, ssqdfsc, tsqdfsc, modeldfh, respdfh)
      end do
    end do
  case default
    do i = 1, kg
      call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
         kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        call rchkusr
        lglk(j,i) = jointyz_sp(n, zsample(:,j), y, l, Ups, ldh_Ups, &
           nulist(i), xi, lmxi, ssqdfsc, tsq, modeldfh)
      end do
    end do
  end select
end subroutine llikfcn_no




subroutine llikfcn_mu (lglk, philist, omglist, nulist, kappalist, &
   musample, Ntot, y, l, F, offset, &
   dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, kg, ifam, itr)

  use modelfcns, jointymu_sp => jointymu, logpdfmu_sp => logpdfmu
  use covfun
  use jointymu, only: jointymu_gt
  use betaprior
  implicit none
  integer, intent(in) :: n, p, kg, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     musample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot, kg)
  logical lmxi
  double precision T(n, n), TiF(n, p), FTF(p, p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, &
     tsqdfsc, respdfh, xi(n)
  integer i, j

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  select case (ifam)
  case (0)
    do i = 1, kg
      call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
         kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        call rchkusr
        lglk(j,i) = jointymu_gt(n, musample(:,j), y, l, Ups, ldh_Ups, &
           nulist(i), xi, lmxi, ssqdfsc, tsqdfsc, modeldfh, respdfh)
      end do
    end do
  case default
    do i = 1, kg
      call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
         kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        call rchkusr
        lglk(j,i) = jointymu_sp(n, musample(:,j), y, l, Ups, ldh_Ups, &
           nulist(i), xi, lmxi, ssqdfsc, tsq, modeldfh)
      end do
    end do
  end select
end subroutine llikfcn_mu

subroutine lpdffcn_mu (lglk, philist, omglist, nulist, kappalist, &
   musample, Ntot, y, l, F, offset, &
   dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, kg, ifam, itr)
  ! Like llikfcn_mu but not multiplying by p(y|mu)

  use modelfcns, jointymu_sp => jointymu, logpdfmu_sp => logpdfmu
  use covfun
  use jointymu, only: jointymu_gt
  use betaprior
  implicit none
  integer, intent(in) :: n, p, kg, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     musample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, ssqdfsc, modeldfh, &
     tsqdfsc, respdfh, xi(n)
  integer i, j

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  do i = 1, kg
    call calc_cov (philist(i),omglist(i),dm,F,betQ0,&
       kappalist(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, Ntot
      call rchkusr
      lglk(j,i) = logpdfmu_sp(n, musample(:,j), Ups, ldh_Ups, &
         nulist(i), xi, lmxi, ssqdfsc, modeldfh)
    end do
  end do
end subroutine lpdffcn_mu


subroutine llikfcn_wo (lglk, philist, omglist, nulist, kappalist, &
   wsample, Ntot, y, l, F, offset, &
   dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, kg, ifam, itr)
  !! Work-around for the binomial family.

  use modelfcns, jointyz_sp => jointyz
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, kg, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     wsample(n,Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n,n), betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot, kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, ssqdfsc, modeldfh, nu, phi, omg, kappa, &
     tsqdfsc, respdfh, xi(n)
  integer i, j, m
  double precision zsam(n)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  select case (ifam)
  case (0)
    call rexit ("This method has not been implemented.")
  case default
    do i = 1, kg
      nu = nulist(i)
      phi = philist(i)
      omg = omglist(i)
      kappa = kappalist(i)
      call calc_cov (phi,omg,dm,F,betQ0,&
         kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        call rchkusr
        zsam = transfw(wsample(:,j),nu)
        lglk(j,i) = jointyz_sp(n, zsam, y, l, Ups, ldh_Ups, &
           nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
        do m = 1, n
          lglk(j,i) = lglk(j,i) - loginvtrwdz(zsam(m),nu)
        end do
      end do
    end do
  end select
end subroutine llikfcn_wo

subroutine llikfcn_tr (lglk, philist, omglist, nulist, kappalist, &
   sample, Ntot, y, l, F, offset, &
   dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, kg, ifam, itr)
!! Log-likelihood function.
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns, logpdfzf => logpdfz, condymu_mf => condymu
  use covfun
  use betaprior
  use condymu, only: condymu_gt
  implicit none
  integer, intent(in) :: n, p, kg, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: philist(kg), omglist(kg), nulist(kg), &
     sample(n, Ntot), y(n), l(n), F(n, p), offset(n), &
     dm(n, n), betm0(p), betQ0(p, p), ssqdf, ssqsc, tsqdf, tsq, kappalist(kg)
  double precision, intent(out) :: lglk(Ntot,kg)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n, n), &
     ldh_Ups, ssqdfsc, modeldfh, nu, phi, omg, kappa, &
     tsqval, respdfh, xi(n)
  integer i, j
  double precision zsam(n), msam(n), jsam(n), sam(n)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  do i = 1, kg
    nu = nulist(i)
    phi = philist(i)
    omg = omglist(i)
    kappa = kappalist(i)
    call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, Ntot
      call rchkusr
      sam = sample(:,j)
      where (itr == 0)
        zsam = sam
        msam = invlink(zsam,nu)
        jsam = 0d0
      elsewhere (itr == 1)
        msam = sam
        zsam = flink(msam,nu)
        jsam = loginvlinkdz(zsam,nu)
      elsewhere (itr == 2)
        zsam = transfw(sam,nu)
        msam = invlink(zsam,nu)
        jsam = loginvtrwdz(zsam,nu)
      end where
      lglk(j,i) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
         + condymuf(ifam,n,y,l,msam,tsqval,respdfh) - sum(jsam)
    end do
  end do

contains
  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: ifam
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf
end subroutine llikfcn_tr



! Local Variables:
! compile-command: "gfortran -c -fpic -Wunused-parameter -Wall \
!   -pedantic -o llik.o llik.f90"
! End:

##' Link function for the exponential family.
##'
##' \code{linkfcn} maps the mean of the response variable \code{mu} to
##' the linear predictor \code{z}. \code{linkinv} is its inverse.
##'
##' For the Gaussian family, if the link parameter is positive, then
##' the extended link is used, defined by \deqn{z =
##' \frac{sign(\mu)|\mu|^\nu - 1}{\nu}}{z = (sign(mu)*abs(mu)^nu -
##' 1)/nu} In the other case, the usual Box-Cox link is used.
##'
## For the Poisson and gamma families, the Box-Cox transformation is
## used, defined by \deqn{z = \frac{\mu^\nu - 1}{\nu}}{z = (mu^nu -
## 1)/nu}
##'
##' For the Poisson and gamma families, if the link parameter is
##' positive, then the link is defined by \deqn{z = \frac{sign(w)
##' (e^{\nu |w|}-1)}{\nu}}{z = sign(w)*expm1(nu*w)/nu} where
##' \eqn{w = \log(\mu)}{w = log(mu)}. In the other case, the usual
##' Box-Cox link is used.
##'
##' For the GEV binomial family, the link function is defined by
##' \deqn{\mu = 1 - \exp\{-\max(0, 1 + \nu z)^{\frac{1}{\nu}}\}}{mu =
##' 1 - exp[-max(0, 1 + nu z)^(1/nu)]} for any real \eqn{\nu}{nu}. At
##' \eqn{\nu = 0}{nu = 0} it reduces to the complementary log-log
##' link.
##'
##' The Wallace binomial family is a fast approximation to the robit
##' family. It is defined as \deqn{\mu =
##' \Phi(\mbox{sign}(z) c(\nu) \sqrt{\nu \log(1 + z^2/\nu)})}{mu =
##' Phi(sign(z) c(nu) sqrt{nu log(1 + z^2/nu)})}
##' where \eqn{c(\nu) = (8\nu+1)/(8\nu+3)}{c(nu) = (8*nu+1)/(8*nu+3)}
##'
##' @title Calculate the link function for exponential families
##' @param mu Numeric. The mean of the response variable.
##' @param z Numeric. The linear predictor.
##' @param linkp The link function parameter. A scalar.
##' @param family The distribution of the response variable from
##'   \code{\link{.geoBayes_models}}. Either an integer or the
##'   family name.
##' @return A numeric array of the same dimension as the function's
##' first argument.
## ##' @seealso \code{\link{comparebinlinks}}
##' @examples \dontrun{
##' mu <- seq(0.1, 0.9, 0.1)
##' linkfcn(mu, 7, "binomial")       # robit(7) link function
##' linkfcn(mu, , "binomial.logit")  # logit link function
##'
##' mu <- seq(-3, 3, 1)
##' linkfcn(mu, 0.5, "gaussian")     # sqrt transformation
##' linkinv(linkfcn(mu, 0.5, "gaussian"), 0.5, "gaussian")
##' curve(linkfcn(x, 0.5, "gaussian"), -3, 3)
##' }
##' @family linkfcn
##' @name linkfcn
##' @rdname linkfcn
##' @importFrom stats qlogis qnorm qt qf plogis pnorm pt
##' @useDynLib geoBayes flinkfcn
##' @references Evangelou, E., & Roy, V. (2019). Estimation and prediction for
##'   spatial generalized linear mixed models with parametric links
##'   via reparameterized importance sampling. Spatial Statistics, 29,
##'   289-315.
##' @export
linkfcn <- function (mu, linkp, family = "gaussian") {
  if (is.numeric(family)) {
    ifam <- as.integer(family)
    if (ifam < 0 || ifam > NROW(.geoBayes_models)) {
      stop ("Cannot deduce family")
    }
  } else {
    ifam <- .geoBayes_family(family)
  }
  if (!.geoBayes_models$needlinkp[ifam]) linkp <- 0
  if (length(linkp) != 1) {
    stop ("The linkp argument must be scalar.")
  }
  linkp <- .geoBayes_getlinkp(linkp, ifam)
  x <- y <- as.double(mu)
  ii <- is.finite(x)
  n <- as.integer(sum(ii))
  mu[ii] <- .Fortran("flinkfcn", x[ii], n, y[ii], linkp, ifam,
                     PACKAGE = "geoBayes")[[1]]
  mu
}

##' @family linkfcn
##' @rdname linkfcn
##' @useDynLib geoBayes flinkinv
##' @export
linkinv <- function (z, linkp, family = "gaussian") {
  if (is.numeric(family)) {
    ifam <- as.integer(family)
    if (ifam < 0 || ifam > NROW(.geoBayes_models)) {
      stop ("Cannot deduce family")
    }
  } else {
    ifam <- .geoBayes_family(family)
  }
  if (!.geoBayes_models$needlinkp[ifam]) linkp <- 0
  if (length(linkp) != 1) {
    stop ("The linkp argument must be scalar.")
  }
  linkp <- .geoBayes_getlinkp(linkp, ifam)
  x <- y <- as.double(z)
  ii <- is.finite(x)
  n <- as.integer(sum(ii))
  z[ii] <- .Fortran("flinkinv", x[ii], n, y[ii], linkp, ifam,
                    PACKAGE = "geoBayes")[[1]]
  z
}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! 
!!! Commentary: Derivatives of inverse link fcn w.r.t. first arg.
!! 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module linkdz

contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Binomial !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! 1st derivative w.r.t. z
  pure function invlinkdz_bi (z, d) result (y)
    use interfaces, only: logprobt, logprobnorm, logpdft, logpdfnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision f, ff
    if (d .gt. 0d0) then
      ff = logprobt(z, d)
      f = logpdft(z, d)
      y = exp(f - ff)
    else if (d .lt. 0d0) then
      y = 1d0/(1d0 + exp(z))
    else
      ff = logprobnorm(z)
      f = logpdfnorm(z)
      y = exp(f - ff)      
    end if
  end function invlinkdz_bi

!! 2nd derivative w.r.t. z
  pure function invlinkddz_bi (z, d) result (y)
    use interfaces, only: logprobt, logprobnorm, logpdft, logpdfnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision f, ff
    if (d .gt. 0d0) then
      ff = logprobt(z, d)
      f = logpdft(z, d)
      y = f/ff
      y = -y*(z+y)
    else if (d .lt. 0d0) then
      y = z*.5d0
      y = .5d0/cosh(y)
      y = -y*y
    else
      ff = logprobnorm(z)
      f = logpdfnorm(z)
      y = f/ff
      y = -y*((d+1d0)*z/(d+z*z) + y)
    end if
  end function invlinkddz_bi

!! 3rd derivative w.r.t. z
  pure function invlinkdddz_bi (z, d) result (y)
    use interfaces, only: logprobt, logprobnorm, logpdft, logpdfnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision f, ff
    if (d .gt. 0d0) then
      f = invlinkdz_bi(z, d)
      ff = invlinkddz_bi(z, d)
      y = d + z*z
      y = (d+1d0)*(d + d - y)/(y*y)*f + (d+1d0)*z/y*ff + 2d0*f*ff
      y = -y
    else if (d .lt. 0d0) then
      y = z
      ff = .5d0*y
      y = 1d0/(sinh(y))
      ff = sinh(ff)
      ff = ff*ff
      y = y*y*y*ff*ff
      y = y + y
    else
      f = invlinkdz_bi(z, d)
      ff = invlinkddz_bi(z, d)
      y = f + z*ff + 2d0*f*ff
      y = -y
    end if
  end function invlinkdddz_bi
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! GEV Binomial !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  pure function invlinkdz_ba (z,d) result(w)
    ! Binomial asymmetric (GEV) link fcn
    ! mu = 1 - exp{-max(0,1+d*z)**(1/d)} , if d /= 0
    !    = 1 - exp(-exp(z))              , if d == 0
    ! w = log(1-mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = -exp(z)
    else if (d .eq. 0.5d0) then
      w = 1d0 + 0.5d0*z
      if (w .gt. 0d0) then
        w = -w
      else
        w = 0d0
      end if
    else if (d .eq. -0.5d0) then
      w = 1d0 - 0.5d0*z
      if (w .gt. 0d0) then
        w = -1d0/(w*w*w)
      else
        w = 0d0
      end if
    else if (d .eq. 1d0) then
      w = 1d0 + z
      if (w .gt. 0d0) then
        w = -1d0
      else
        w = 0d0
      end if
    else if (d .eq. -1d0) then
      w = 1d0 - z
      if (w .gt. 0d0) then
        w = -1d0/(w*w)
      else
        w = 0d0
      end if
    else if (d .eq. 2d0) then
      w = 1d0 + z + z
      if (w .gt. 0d0) then
        w = -1d0/sqrt(w)
      else
        w = 0d0
      end if
    else if (d .eq. -2d0) then
      w = 1d0 - z - z
      if (w .gt. 0d0) then
        w = -1d0/(w*sqrt(w))
      else
        w = 0d0
      end if
    else
      w = 1d0 + d*z
      if (w .gt. 0d0) then
        w = -w**(1d0/d - 1d0)
      else
        w = 0d0
      end if
    end if
  end function invlinkdz_ba

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! GEVD Binomial !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  pure elemental function invlinkdz_bd (z, d) result (w)
    ! Binomial asymmetric (GEV) decreasing link fcn
    ! p = 1 - exp{-max(0, 1+d*z)**(-1/d)}
    ! w = log(1-p) = -max(0, 1+d*z)**(-1/d)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = exp(-z)
    else
      w = 1d0 + d*z
      if (w .gt. 0d0) then
        if (d .eq. -1d0) then
          w = 1d0
        else
          w = w**(-1d0/d - 1d0)
        end if
      else
        w = 0d0
      end if
    end if
  end function invlinkdz_bd

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Box-Cox !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  pure elemental function invlinkdz_boxcox (z,d) result (w)
    ! w = log(mu)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 1d0
    else 
      w = d*z
      if (w .gt. -1d0) then
        w = 1d0/(1d0+w)
      else
        w = 0d0
      end if
    end if
  end function invlinkdz_boxcox

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Poisson !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  pure elemental function invlinkdz_po (z,d) result (w)
    ! w = log(mu)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 1d0
    else if (d .gt. 0d0) then
      w = 1d0 + d*abs(z)
      w = 1d0/w
    else ! Use regular Box-Cox
      w = d*z
      if (w .gt. -1d0) then
        w = 1d0/(1d0+w)
      else
        w = 0d0
      end if
    end if
  end function invlinkdz_po

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Gamma !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  pure elemental function invlinkdz_gm (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = invlinkdz_po(z, d)
  end function invlinkdz_gm

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Gaussian !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  pure elemental function invlinkdz_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = exp(z)
    else if (d .eq. 1d0) then
      w = 1d0
    else if (d .gt. 0d0) then
      w = abs(d*z + 1d0)
      if (d .eq. 2d0) then
        w = 1d0/sqrt(w)
      else if (d .ne. .5d0) then
        w = w**(1d0/d - 1d0)
      end if
    else ! d .lt. 0d0
      w = d*z + 1d0
      if (w .gt. 0d0) then
        if (d .eq. -1d0) then
          w = 1d0/(w*w)
        else if (d .eq. -2d0) then
          w = 1d0/(w*sqrt(w))
        else if (d .eq. -.5d0) then
          w = 1d0/(w*w*w)
        else 
          w = w**(1d0/d - 1d0)
        end if
      else
        w = 0d0
      end if
    end if
  end function invlinkdz_ga

!!!!!!!!!!!!!!!!!!! Wallace transformation for binomial !!!!!!!!!!!!!!!!!!!
  pure elemental function invlinkdz_bw (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision x, t, c, f, u
    c = 8d0*d
    c = (c + 1d0)/(c + 3d0)
    t = z*z/d
    u = sqrt(d*flog1p(t))
    u = sign(u,z)
    x = c*u
    w = logprobnorm(x)
    f = logpdfnorm(x)
    w = exp(f-w)*c*z/((1d0+t)*u)
  end function invlinkdz_bw
end module linkdz

##' Analysis of geostatistical data using Bayes and Empirical Bayes
##' methods.
##'
##' This package provides functions to fit geostatistical data. The
##' data can be continuous, binary or count data and the models
##' implemented are flexible. Conjugate priors are assumed on some
##' parameters while inference on the other parameters can be done
##' through a full Bayesian analysis of by empirical Bayes methods.
##'
##' Some demonstration examples are provided. Type \code{demo(package
##' = "geoBayes")} to examine them.
##' @title The \code{geoBayes} package
##' @name geoBayes
##' @aliases geoBayes-package
## @docType package
##' @useDynLib geoBayes, .registration = TRUE
##' @author Evangelos Evangelou <e.evangelou@@maths.bath.ac.uk> and
##' Vivekananda Roy <vroy@@iastate.edu>
##' @examples \dontrun{
##' demo(package = "geoBayes")
##' demo(rhizoctonia1, package = "geoBayes")
##' demo(rhizoctonia1, package = "geoBayes")
##' }
##' @references Roy, V., Evangelou, E. and Zhu, Z. (2014). Empirical
##' Bayes methods for the transformed Gaussian random fields model
##' with additive measurement errors. In Upadhyay, S. K., Singh, U.,
##' Dey, D. K., and Loganathan, A., editors, \emph{Current Trends in
##' Bayesian Methodology with Applications}, Boca Raton, FL, USA, CRC
##' Press.
##'
##' Roy, V., Evangelou, E., and Zhu, Z. (2015). Efficient estimation
##' and prediction for the Bayesian spatial generalized linear mixed
##' model with flexible link functions. \emph{Biometrics}, 72(1),
##'   289-298.
##'
##' Evangelou, E., & Roy, V. (2019). Estimation and prediction for
##'   spatial generalized linear mixed models with parametric links
##'   via reparameterized importance sampling. \emph{Spatial Statistics}, 29,
##'   289-315.
##'
##' Roy, V., & Evangelou, E. (2024). Selection of proposal
##'   distributions for multiple importance sampling.
##'   \emph{Statistica Sinica}, 34, 27-46.
##' @keywords package
"_PACKAGE"


.onUnload <- function (libpath) {
  library.dynam.unload("geoBayes", libpath)
}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! 
!!! Commentary: 
!! 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module linkfcns
  implicit none
  double precision, parameter :: bigpos = huge(1d0), bigneg = -bigpos, &
     smallpos = epsilon(1d0), smallneg = -smallpos
  private bigpos, bigneg, smallpos, smallneg
contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! binomial !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link function
  pure elemental function flink_bi (w, d) result (z)
    use interfaces, only: quantt, quantnorm, quantlogis
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    if (d .gt. 0d0) then
      z = quantt(w, d)
    else if (d .lt. 0d0) then
      z = quantlogis(w)
    else
      z = quantnorm(w)
    end if
  end function flink_bi

!! Inverse link function (log scale)
  pure elemental function invlink_bi (z, d) result (w)
    ! Binomial symmetric (robit) link fcn
    use interfaces, only: logprobt, logprobnorm, logproblogis
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .gt. 0d0) then
      w = logprobt(z, d)
    else if (d .lt. 0d0) then
      w = logproblogis(z)
    else
      w = logprobnorm(z)
    end if
  end function invlink_bi

!! Derivative of inverse link function w.r.t. parameter
  function invlinkdnu_bi (z, d) result (y)
    ! Binomial symmetric (robit) link fcn
    use interfaces, only: logprobt
    use tcdfder
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    if (d .gt. 0d0) then
      y = tcdfdnu(z, d)
    else 
      y = 0d0
    end if
  end function invlinkdnu_bi

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! GEV binomial !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link function
!! mu = 1 - exp{-max(0,1+d*z)**(1/d)} , if d /= 0
!!    = 1 - exp(-exp(z))              , if d == 0
  pure elemental function flink_ba (w, d) result (z)
    ! w is log(1-mu)
    implicit none
    double precision, intent(in) :: w, d
    double precision z
      z = -w
      if (d .eq. 0d0) then
        z = log(z)
      else if (d .eq. .5d0) then
        z = -1d0 + sqrt(z)
        z = z + z
      else if (d .eq. 1d0) then
        z = z - 1d0
      else if (d .eq. 2d0) then
        z = .5d0*(z*z - 1d0)
      else if (d .eq. -.5d0) then
        z = -1d0 + 1d0/sqrt(z)
        z = z + z
      else if (d .eq. -1d0) then
        z = 1d0 - 1d0/z
      else if (d .eq. -2d0) then
        z = .5d0 - 1d0/(z*z)
      else
        z = (-1d0 + z**d)/d
      end if
  end function flink_ba

!! Inverse link fcn. Returns log(1-mu)
  pure elemental function invlink_ba (z, d) result (w)
    ! Binomial asymmetric (GEV) link fcn
    ! mu = 1 - exp{-max(0,1+d*z)**(1/d)} , if d /= 0
    !    = 1 - exp(-exp(z))              , if d == 0
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = -exp(z)
    else if (d .eq. 0.5d0) then
      w = 1d0 + 0.5d0*z
      if (w .gt. 0d0) then
        w = -w*w
      else
        w = smallneg
      end if
    else if (d .eq. -0.5d0) then
      w = 1d0 - 0.5d0*z
      if (w .gt. 0d0) then
        w = -1d0/(w*w)
      else
        w = bigneg
      end if
    else if (d .eq. 1d0) then
      w = 1d0 + z
      if (w .gt. 0d0) then
        w = -w
      else
        w = smallneg
      end if
    else if (d .eq. -1d0) then
      w = 1d0 - z
      if (w .gt. 0d0) then
        w = -1d0/w
      else
        w = bigneg
      end if
    else if (d .eq. 2d0) then
      w = 1d0 + z + z
      if (w .gt. 0d0) then
        w = -sqrt(w)
      else
        w = smallneg
      end if
    else if (d .eq. -2d0) then
      w = 1d0 - z - z
      if (w .gt. 0d0) then
        w = -1d0/sqrt(w)
      else
        w = bigneg
      end if
    else if (d .gt. 0d0) then 
      w = 1d0 + d*z
      if (w .gt. 0d0) then
        w = -w**(1d0/d)
      else
        w = smallneg
      end if
    else
      w = 1d0 + d*z
      if (w .gt. 0d0) then
        w = -w**(1d0/d)
      else
        w = bigneg
      end if
    end if
  end function invlink_ba

!! Inverse link fcn derivative w.r.t. nu
  pure elemental function invlinkdnu_ba (z, d) result (y)
    ! Binomial asymmetric (GEV) link fcn
    ! mu = 1 - exp{-max(0,1+d*z)**(1/d)} , if d /= 0
    !    = 1 - exp(-exp(z))              , if d == 0
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    if (d .eq. 0d0) then
      y = .5d0*exp(z)*z*z
    else if (d .eq. 0.5d0) then
      y = 2d0 + z
      if (y .gt. 0d0) then
        y = y*(y*log(.5d0*y) - z)
      else
        y = 0d0
      end if
    else if (d .eq. -0.5d0) then
      y = 2d0 - z
      if (y .gt. 0d0) then
        y = 16d0 * (y*log(.5d0*y) + z)/(y*y*y)
      else
        y = bigpos
      end if
    else if (d .eq. 1d0) then
      y = 1d0 + z
      if (y .gt. 0d0) then
        y = y*log(y) - z
      else
        y = -z
      end if
    else if (d .eq. -1d0) then
      y = 1d0 - z
      if (y .gt. 0d0) then
        y = (y*log(y) + z)/(y*y)
      else
        y = bigpos
      end if
    else if (d .eq. 2d0) then
      y = 1d0 + z + z
      if (y .gt. 0d0) then
        y = (y*log(y) - z - z)/(4d0*sqrt(y))
      else
        y = bigpos
      end if
    else if (d .eq. -2d0) then
      y = 1d0 - z - z
      if (y .gt. 0d0) then
        y = (y*log(y) + z + z)/(4d0*y*sqrt(y))
      else
        y = bigpos
      end if
    else
      y = 1d0 + d*z
      if (y .gt. 0d0) then
        y = (y*log(y) - d*z)/(d*d*y**(1d0-1d0/d))
      else
        y = bigpos
      end if
    end if
  end function invlinkdnu_ba

 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! GEVD binomial !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link function
!! mu = 1 - exp{-max(0, 1+d*z)**(-1/d)} , if d /= 0
!!    = 1 - exp(-exp(-z))               , if d == 0
  pure elemental function flink_bd (w, d) result (z)
    ! w is log (1-mu)
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    z = -w
    if (d .eq. 0d0) then
      z = -log(z)
    else ! d .ne. 0d0
      z = (-1d0 + z**(-d))/d
    end if
  end function flink_bd
  
!! Inverse link fcn. Returns log(1-mu)
  pure elemental function invlink_bd (z, d) result (w)
    ! Binomial asymmetric (GEV) decreasing link fcn
    ! p = 1 - exp{-max(0, 1+d*z)**(-1/d)}
    ! w = log(1-p) = -max(0, 1+d*z)**(-1/d)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = -exp(-z)
    else
      w = 1d0 + d*z
      if (w .gt. 0d0) then
        w = -w**(-1d0/d)
      else if (d .gt. 0d0) then
        w = bigneg
      else
        w = smallneg
      end if
    end if
  end function invlink_bd

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Box-Cox !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link fcn
  pure elemental function flink_boxcox (w, d) result (z)
    ! z = (mu^d - 1)/d
    ! w is log(mu)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    if (d .eq. 0d0) then
      z = w
    else if (d .eq. 1d0) then
      z = fexpm1(w)
    else if (d .eq. -1d0) then
      z = -fexpm1(-w)
    else
      z = fexpm1(w*d)/d
    end if
  end function flink_boxcox
  
!! Inverse link fcn, returns log(mu)
  pure elemental function invlink_boxcox (z,d) result (w)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = z
    else 
      w = d*z
      if (w .gt. -1d0) then
        w = flog1p(w)/d
      else
        w = bigneg
      end if
    end if
  end function invlink_boxcox

!!   elemental function invlinkdnu_boxcox (z,d) result (y)
!!     implicit none
!!     double precision, intent(in) :: z, d
!!     double precision y
!!     if (d .eq. 0d0) then
!!       y = -.5d0*z*z
!!     else 
!!       y = 1d0 + d*z
!!       if (y .gt. 0d0) then
!!         y = z/(d*y) - log(y)/(d*d)
!!       else
!!         y = bigneg
!!       end if
!!     end if
!!   end function invlinkdnu_boxcox

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Poisson !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link fcn
  pure elemental function flink_po (w, d) result (z)
    ! w is log(mu)
    ! z = sign(w)*expm1(d*|w|)/d for d > 0
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    if (d .eq. 0d0) then
      z = w
    else if (d .eq. 1d0) then
      z = fexpm1(abs(w))
      z = sign(z,w)
    else if (d .gt. 0d0) then
      z = fexpm1(d*abs(w))/d
      z = sign(z,w)
    else ! Use regular Box-Cox
      z = fexpm1(w*d)/d
    end if
  end function flink_po

  pure elemental function invlink_po (z,d) result (w)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = z
    else if (d .eq. 1d0) then
      w = flog1p(abs(z))
      w = sign(w,z)
    else if (d .gt. 0d0) then
      w = flog1p(d*abs(z))/d
      w = sign(w,z)
    else ! Use regular Box-Cox
      w = d*z
      if (w .gt. -1d0) then
        w = flog1p(w)/d
      else
        w = bigneg
      end if
    end if
  end function invlink_po

 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Gamma !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link fcn
  pure elemental function flink_gm (w, d) result (z)
    ! z = (mu^d - 1)/d
    ! w is log(mu)
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    z = flink_po(w, d)
  end function flink_gm
  
!! Inverse link fcn, returns log(mu)
  pure elemental function invlink_gm (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = invlink_po(z, d)
  end function invlink_gm

!!   elemental function invlinkdnu_gm (z,d) result (y)
!!     implicit none
!!     double precision, intent(in) :: z, d
!!     double precision y
!!     y = invlinkdnu_po(z, d)
!!   end function invlinkdnu_gm


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Gaussian !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Linf fcn
  pure elemental function flink_ga (w, d) result (z)
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    if (d .eq. 0d0) then
      if (w > 0d0) then
        z = log(w)
      else
        z = bigneg
      end if
    else if (d .gt. 0d0) then
      if (d .eq. 2d0) then
        z = .5d0*(w*w - 1d0)
      else if (d .eq. 1d0) then
        z = w - 1d0
      else if (d .eq. .5d0) then
        z = sign(sqrt(abs(w)), w) - 1d0
        z = z + z
      else
        z = (sign(abs(w)**d, w) - 1d0)/d
      end if
    else ! d .lt. 0d0
      if (w .gt. 0d0) then
        if (d .eq. -0.5d0) then
          z = -(1d0/sqrt(w) - 1d0)
          z = z + z
        else if (d .eq. -1d0) then
          z = -(1d0/w - 1d0)
        else if (d .eq. -2d0) then
          z = -.5d0*(1d0/(w*w) - 1d0)
        else
          z = (w**d - 1d0)/d
        end if
      else ! w .le. 0d0
        z = bigneg
      end if
    end if
  end function flink_ga
  
!! Inverse link fcn 
  pure elemental function invlink_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = exp(z)
    else if (d .gt. 0d0) then
      w = d*z + 1d0
      if (d .eq. 2d0) then
        w = sign(sqrt(abs(w)),w)
      else if (d .eq. .5d0) then
        w = sign(w*w,w)
      else if (d .ne. 1d0) then
        w = sign(abs(w)**(1d0/d),w)
      end if
    else ! d .lt. 0d0
      w = d*z + 1d0
      if (w .gt. 0d0) then
        w = 1d0/w
        if (d .eq. -2d0) then
          w = sqrt(w)
        else if (d .eq. -.5d0) then
          w = w*w
        else if (d .ne. -1d0) then
          w = w**(1d0/(-d))
        end if
      else
        w = bigpos
      end if
    end if
  end function invlink_ga
  
  
!!!!!!!!!!!!!!!!!!! Wallace transformation for binomial !!!!!!!!!!!!!!!!!!!
!!! See function u4 in
!!!  Wallace, D. L. (1959). Bounds on normal approximations to Student's
!!!  and the chi-square distributions. The Annals of Mathematical
!!!  Statistics, 1121-1130.
!! Link fcn
!! mu = PHI[sign(z) * c(nu) * sqrt(nu*log(1+z*z/nu))]
!! where c(nu) = (8*nu+1)/(8*nu+3)
!! w = log(mu)
  pure elemental function flink_bw (w, d) result(z)
    use interfaces, only: quantnorm, fexpm1
    implicit none
    double precision, intent(in) :: w, d
    double precision :: z
    double precision t, e
    t = 8d0*d
    e = quantnorm(w)*(t + 3d0)/(t + 1d0)
    t = sqrt(d*fexpm1(e*e/d))
    z = sign(t, e)
  end function flink_bw

!! Inverse link fcn (log scale)
  pure elemental function invlink_bw (z,d) result (w)
    use interfaces, only: logprobnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision x, t
    t = 8d0*d
    x = sqrt(d*flog1p(z*z/d))*(t + 1d0)/(t + 3d0)
    t = sign(x, z)
    w = logprobnorm(t)
  end function invlink_bw
end module linkfcns


! Local Variables:
! compile-command: "R CMD SHLIB"
! End:

##' Spatial joint log likelihood
##'
##' Computes the joint log likelihood log f(y,T(z)|parameters) where
##' T(z) is the transformation, for each (y,z) in runs and for parameters
##' in pargrid up to a constant which does not depend on the
##' parameters. The parameters beta and sigma^2 are integrated out.
##' @title Spatial log likelihood
##' @param pargrid A data frame with components "linkp", "phi", "omg",
##' "kappa". Each row gives a combination of the parameters to compute
##' the log-likelihood.
##' @param runs A list with outputs from the function
##' \code{\link{mcsglmm}} or \code{\link{mcstrga}}.
##' @param transf Whether to use a transformed sample for the
##'   computations. If \code{"no"} or \code{FALSE}, it doesn't. If
##'   \code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
##'   \code{"wo"} it uses an alternative transformation. The latter
##'   can be used only for the families indicated by
##'   \code{.geoBayes_models$haswo}.
##' @return A matrix with number of rows the total number of samples
##'   in runs and number of columns the number of rows in pargrid. The
##'   [i,j] element of the matrix is the value of the loglikelihood at
##'   the ith sample when all samples in runs are put together evaluated
##'   at the jth parameter value.
##' @useDynLib geoBayes llikfcn_no llikfcn_mu llikfcn_wo llikfcn_tr
##' @export
sploglik <- function(pargrid, runs, transf = c("no", "mu", "wo"))
{
  ## Check input
  if (!all(sapply(runs, inherits, what = "geomcmc"))) {
    stop ("Input runs is not a list with elements of class geomcmc")
  }
  nruns <- length(runs)
  if (nruns == 0) stop ("No runs specified")
  Nout <- sapply(runs, function(x) x$MCMC$Nout)
  Ntot <- sum(Nout)

  ## Extract data and model
  nm_DATA <- c("response", "weights", "modelmatrix", "offset", "locations",
               "longlat")
  nm_MODEL <- c("family", "corrfcn", "betm0", "betQ0", "ssqdf", "ssqsc",
                "tsqdf", "tsqsc", "dispersion")
  DATA <- runs[[1]]$DATA[nm_DATA]
  MODEL <- runs[[1]]$MODEL[nm_MODEL]
  if (nruns > 1) {
    for (i in 2:nruns) {
      if (!identical(runs[[i]]$DATA[nm_DATA], DATA)) {
        stop("MCMC chains don't all correspond to the same data.")
      }
      if (!identical(runs[[i]]$MODEL[nm_MODEL], MODEL)) {
        stop("MCMC chains don't all correspond to the same model.")
      }
    }
  }
  y <- DATA$response
  n <- as.integer(length(y))
  l <- DATA$weights
  F <- DATA$modelmatrix
  offset <- DATA$offset
  p <- NCOL(F)
  loc <- DATA$locations
  dm <- sp::spDists(loc, longlat = DATA$longlat)
  family <- MODEL$family
  ## ifam <- .geoBayes_family(family)
  corrfcn <- MODEL$corrfcn
  icf <- .geoBayes_correlation(corrfcn)
  betm0 <- MODEL$betm0
  betQ0 <- MODEL$betQ0
  ssqdf <- MODEL$ssqdf
  ssqsc <- MODEL$ssqsc
  tsqdf <- MODEL$tsqdf
  tsqsc <- MODEL$tsqsc
  dispersion <- MODEL$dispersion

  ## Extract the grid for the new parameters
  pargrid <- check_pargrid(pargrid, family, corrfcn)
  phi <- pargrid$phi
  omg <- pargrid$omg
  kappa <- pargrid$kappa
  nu <- pargrid$nu
  kg <- NROW(pargrid)

  ## Transformed sample
  getsample <-
    transfsample(runs,
                 list(response = y, family = family), transf)
  sample <- matrix(unlist(getsample$sample), n)
  itr <- getsample$itr
  transf <- getsample$transf
  real_transf <- getsample$real_transf
  ifam <- getsample$ifam
  froutine <- paste0("llikfcn_", real_transf)

  if (ifam == 0) {
    tsq <- tsqsc
  } else {
    tsq <- dispersion
    tsqdf <- 0
  }

  lglk <- matrix(0, Ntot, kg)
  fcall <- .Fortran(froutine,
                    lglk = lglk,
                    as.double(phi), as.double(omg), as.double(nu),
                    as.double(kappa),
                    as.double(sample), as.integer(Ntot), as.double(y),
                    as.double(l), as.double(F), as.double(offset), as.double(dm),
                    as.double(betm0), as.double(betQ0), as.double(ssqdf),
                    as.double(ssqsc), as.double(tsqdf), as.double(tsq),
                    as.integer(icf), as.integer(n), as.integer(p),
                    as.integer(kg), as.integer(ifam), as.integer(itr),
                    PACKAGE = "geoBayes")
  lglk <- fcall$lglk
  lglk
}

##' Spatial joint log likelihood
##'
##' Computes the joint log likelihood log f(y,T(z)|parameters) where
##' T(z) is the transformation, for each (y,z) in runs and for parameters
##' in runs up to a constant which does not depend on the parameters.
##' The parameters beta and sigma^2 are integrated out. 
##' @title Spatial log likelihood
##' @param runs A list with outputs from the function
##' \code{\link{mcsglmm}} or \code{\link{mcstrga}}.
##' @param transf Whether to use a transformed sample for the
##'   computations. If \code{"no"} or \code{FALSE}, it doesn't. If
##'   \code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
##'   \code{"wo"} it uses an alternative transformation. The latter
##'   can be used only for the families indicated by
##'   \code{.geoBayes_models$haswo}. The input can also be a vector
##'   (of the same length as \code{runs} to allow for different
##'   transformation to be used when evaluating each likelihood.
##' @return A matrix with number of rows the total number of samples
##'   in runs and number of columns the length of \code{runs}. The
##'   [i,j] element of the matrix is the value of the loglikelihood at
##'   the ith sample when all samples in \code{runs} are put together evaluated
##'   at the jth parameter value.
##' @useDynLib geoBayes llikfcn_no llikfcn_mu llikfcn_wo llikfcn_tr
##' @export
sploglik_cross <- function(runs, transf = c("no", "mu", "wo"))
{
  ## Check input
  if (!all(sapply(runs, inherits, what = "geomcmc"))) {
    stop ("Input runs is not a list with elements of class geomcmc")
  }
  nruns <- length(runs)
  if (nruns == 0) stop ("No runs specified")
  Nout <- sapply(runs, function(x) x$MCMC$Nout)
  Ntot <- sum(Nout)

  ## Extract data and model
  nm_DATA <- c("response", "weights", "modelmatrix", "offset", "locations",
               "longlat")
  nm_MODEL <- c("betm0", "betQ0", "ssqdf", "ssqsc",
                "tsqdf", "tsqsc", "dispersion")
  DATA <- runs[[1]]$DATA[nm_DATA]
  MODEL <- runs[[1]]$MODEL[nm_MODEL]
  if (nruns > 1) {
    for (i in 2:nruns) {
      if (!identical(runs[[i]]$DATA[nm_DATA], DATA)) {
        stop("MCMC chains don't all correspond to the same data.")
      }
      if (!identical(runs[[i]]$MODEL[nm_MODEL], MODEL)) {
        stop("MCMC chains don't all correspond to the same model.")
      }
    }
  }
  y <- DATA$response
  n <- as.integer(length(y))
  l <- DATA$weights
  F <- DATA$modelmatrix
  p <- NCOL(F)
  offset <- DATA$offset
  loc <- DATA$locations
  dm <- sp::spDists(loc, longlat = DATA$longlat)
  corrfcn <- MODEL$corrfcn
  betm0 <- MODEL$betm0
  betQ0 <- MODEL$betQ0
  ssqdf <- MODEL$ssqdf
  ssqsc <- MODEL$ssqsc
  tsqdf <- MODEL$tsqdf
  tsqsc <- MODEL$tsqsc
  dispersion <- MODEL$dispersion

  ifam_list <- sapply(runs, function(r) .geoBayes_family(r$MODEL$family))
  if (any(ifam_list == 0) && any(ifam_list > 0))
    stop ("Cannot mix transformed Gaussian with other families.")
  icf_list <- sapply(runs, function(r) .geoBayes_correlation(r$MODEL$corrfcn))
  nu_list <- sapply(runs, function(r) r$FIXED$linkp_num)
  kappa_list <- sapply(runs, function(r) r$FIXED$kappa)
  phi_list <- sapply(runs, function(r) r$FIXED$phi)
  omg_list <- sapply(runs, function(r) r$FIXED$omg)
  if (length(phi_list) == 0) {
    phi_mcmc <- TRUE
    phi_sample <- do.call("c", lapply(runs, function(r) r$MCMC$phi))
    phi_pars <- sapply(runs, function(r) r$MODEL$phipars)
  } else if (length(phi_list) == nruns) {
    phi_mcmc <- FALSE
  } else {
    stop ("Models are not consistent w.r.t fixed/random phi.")
  }
  if (length(omg_list) == 0) {
    omg_mcmc <- TRUE
    omg_sample <- do.call("c", lapply(runs, function(r) r$MCMC$omg))
    omg_pars <- sapply(runs, function(r) r$MODEL$omgpars)
  } else if (length(omg_list) == nruns) {
    omg_mcmc <- FALSE
  } else {
    stop ("Models are not consistent w.r.t fixed/random omg.")
  }

  ## Transformed sample
  sample <- array(, c(n, Ntot, nruns))
  itr <- array(, c(n, nruns))
  transf <- rep(transf, length = nruns)
  for (i in 1:nruns) {
    getsample <-
      transfsample(runs, list(response = y, family = ifam_list[i]), transf[[i]],
                   verb = TRUE)
    sample[, , i] <- do.call("c", getsample$sample)
    itr[, i] <- getsample$itr
  }

  if (ifam_list[1] == 0) {
    tsq <- tsqsc
  } else {
    tsq <- dispersion
    tsqdf <- 0
  }

  lglk <- matrix(0, Ntot, nruns)

  if (phi_mcmc && omg_mcmc) {
    fcall <- .Fortran("llikfcnmc_11",
                      lglk = lglk,
                      as.double(nu_list),
                      as.double(kappa_list),
                      as.double(sample),
                      as.double(phi_sample),
                      as.double(omg_sample),
                      as.double(phi_pars),
                      as.double(omg_pars),
                      as.integer(Ntot), as.double(y),
                      as.double(l), as.double(F), as.double(offset), 
                      as.double(dm),
                      as.double(betm0), as.double(betQ0), as.double(ssqdf),
                      as.double(ssqsc), as.double(tsqdf), as.double(tsq),
                      as.integer(icf_list), as.integer(n), as.integer(p),
                      as.integer(nruns), as.integer(ifam_list), as.integer(itr),
                      PACKAGE = "geoBayes")
  } else if (!phi_mcmc && omg_mcmc) {
    fcall <- .Fortran("llikfcnmc_01",
                      lglk = lglk,
                      as.double(nu_list),
                      as.double(phi_list),
                      as.double(kappa_list),
                      as.double(sample),
                      as.double(omg_sample),
                      as.double(omg_pars),
                      as.integer(Ntot), as.double(y),
                      as.double(l), as.double(F), as.double(offset),
                      as.double(dm),
                      as.double(betm0), as.double(betQ0), as.double(ssqdf),
                      as.double(ssqsc), as.double(tsqdf), as.double(tsq),
                      as.integer(icf_list), as.integer(n), as.integer(p),
                      as.integer(nruns), as.integer(ifam_list), as.integer(itr),
                      PACKAGE = "geoBayes")
  } else if (phi_mcmc && !omg_mcmc) {
    fcall <- .Fortran("llikfcnmc_10",
                      lglk = lglk,
                      as.double(nu_list),
                      as.double(omg_list),
                      as.double(kappa_list),
                      as.double(sample),
                      as.double(phi_sample),
                      as.double(phi_pars),
                      as.integer(Ntot), as.double(y),
                      as.double(l), as.double(F), as.double(offset),
                      as.double(dm),
                      as.double(betm0), as.double(betQ0), as.double(ssqdf),
                      as.double(ssqsc), as.double(tsqdf), as.double(tsq),
                      as.integer(icf_list), as.integer(n), as.integer(p),
                      as.integer(nruns), as.integer(ifam_list), as.integer(itr),
                      PACKAGE = "geoBayes")
  } else if (!phi_mcmc && !omg_mcmc) {
    fcall <- .Fortran("llikfcnmc_00",
                      lglk = lglk,
                      as.double(nu_list),
                      as.double(phi_list),
                      as.double(omg_list),
                      as.double(kappa_list),
                      as.double(sample),
                      as.integer(Ntot), as.double(y),
                      as.double(l), as.double(F), as.double(offset),
                      as.double(dm),
                      as.double(betm0), as.double(betQ0), as.double(ssqdf),
                      as.double(ssqsc), as.double(tsqdf), as.double(tsq),
                      as.integer(icf_list), as.integer(n), as.integer(p),
                      as.integer(nruns), as.integer(ifam_list), as.integer(itr),
                      PACKAGE = "geoBayes")
  }
  lglk <- fcall$lglk
  lglk
}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Derivatives of BF
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine calcbd_no (bfact, bfdnu, bfdphi, bfdnsq, bfdkappa, &
   phi, nu, omg, kappa, icf, Ntot, sample, weights, QRin, &
   n, p, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, y, l, F, offset, &
   dm, ifam, itr, icv)

  use modelfcns, logpdfzf => logpdfz, logpdfydmu => logpdfydlnk
  use covfun
  use betaprior
  use calcbd_fcns
  implicit none
  integer, intent(in) :: n, p, Ntot, ifam, icf, itr(n), icv
  double precision, intent(in) :: phi, omg, &
     kappa, nu, sample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(in) :: QRin(Ntot)
  double precision, intent(out) :: bfact, bfdphi, bfdnu, bfdnsq, bfdkappa
  logical lmxi
  double precision ssqdfsc, respdfh, modeldfh, tsqval
  integer j
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, lglk(Ntot), xi(n)
  double precision zsam(n), msam(n), jsam(n), sam(n), dlogpy(n), dzdnu(n), &
     dmudnu(n), djsam(n), lpdfz, lpdfy
  double precision dy, dz, dj, dlglkdnu(Ntot), dlglkdphi(Ntot), &
     dlglkdnsq(Ntot), dlglkdkappa(Ntot)
  double precision zUz, zUDTUz, DT(n,n,3), trUpsDTh(3), Upsz(n), dfozUz
  double precision ymm(n), ymml(n)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  do j = 1, 3
    DT(:,:,j) = cor_dcov(n, dm, phi, omg, kappa, j)
    trUpsDTh(j) = .5d0*traceAB(Ups,DT(:,:,j),n)
  end do

  do j = 1, Ntot
    call rchkusr
    sam = sample(:,j)
    zsam = sam
    msam = invlink(zsam,nu)
    dlogpy = logpdfydmu(y,l,msam)/tsq
    dmudnu = invlinkdn(zsam,nu)
    if (lmxi) zsam = zsam - xi
    call dsymv ('u',n,1d0,Ups,n,zsam,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
    zUz = dot_product(zsam,Upsz) + ssqdfsc
    lpdfz = ldh_Ups - modeldfh*log(zUz)
    dfozUz = modeldfh/zUz
    if (ifam .eq. 0) then
      ymm = y - msam
      ymml = l*ymm
      lpdfy = condymuf(ifam,n,y,l,msam,tsqval,respdfh)
      dy = dot_product(ymm,ymml) + tsqval
      dlogpy = (2*respdfh/dy)*ymml
      dy = dot_product(dlogpy,dmudnu)
    else
      lpdfy = condymuf(ifam,n,y,l,msam,tsqval,respdfh)
      dy = dot_product(dlogpy,dmudnu)
    end if
    lglk(j) = lpdfz + lpdfy
    dlglkdnu(j) = dy
    zUDTUz = qform(Upsz,DT(:,:,1),n)
    dlglkdphi(j) = -trUpsDTh(1) + dfozUz*zUDTUz
    zUDTUz = qform(Upsz,DT(:,:,2),n)
    dlglkdnsq(j) = -trUpsDTh(2) + dfozUz*zUDTUz
    zUDTUz = qform(Upsz,DT(:,:,3),n)
    dlglkdkappa(j) = -trUpsDTh(3) + dfozUz*zUDTUz
  end do
  bfact = weigh_llik(icv, lglk, weights, QRin, Ntot)
  bfdnu = weigh_llik_deriv(icv, dlglkdnu, lglk, weights, QRin, Ntot)
  bfdphi = weigh_llik_deriv(icv, dlglkdphi, lglk, weights, QRin, Ntot)
  bfdnsq = weigh_llik_deriv(icv, dlglkdnsq, lglk, weights, QRin, Ntot)
  bfdkappa = weigh_llik_deriv(icv, dlglkdkappa, lglk, weights, QRin, Ntot)
end subroutine calcbd_no

subroutine calcbd_mu (bfact, bfdnu, bfdphi, bfdnsq, bfdkappa, &
   phi, nu, omg, kappa, icf, Ntot, sample, weights, QRin, &
   n, p, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, y, l, F, offset, &
   dm, ifam, itr, icv)

  use modelfcns, logpdfzf => logpdfz, logpdfydmu => logpdfydlnk
  use covfun
  use betaprior
  use calcbd_fcns
  implicit none
  integer, intent(in) :: n, p, Ntot, ifam, icf, itr(n), icv
  double precision, intent(in) :: phi, omg, &
     kappa, nu, sample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(in) :: QRin(Ntot)
  double precision, intent(out) :: bfact, bfdphi, bfdnu, bfdnsq, bfdkappa
  logical lmxi
  double precision ssqdfsc, respdfh, modeldfh, tsqval
  integer j
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, lglk(Ntot), xi(n)
  double precision zsam(n), msam(n), jsam(n), sam(n), dlogpy(n), dzdnu(n), &
     dmudnu(n), djsam(n), lpdfz, lpdfy
  double precision dy, dz, dj, dlglkdnu(Ntot), dlglkdphi(Ntot), &
     dlglkdnsq(Ntot), dlglkdkappa(Ntot)
  double precision zUz, zUDTUz, DT(n,n,3), trUpsDTh(3), Upsz(n), dfozUz
  double precision ymm(n), ymml(n)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  do j = 1, 3
    DT(:,:,j) = cor_dcov(n, dm, phi, omg, kappa, j)
    trUpsDTh(j) = .5d0*traceAB(Ups,DT(:,:,j),n)
  end do

  do j = 1, Ntot
    call rchkusr
    sam = sample(:,j)
    msam = sam
    zsam = flink(msam,nu)
    jsam = loginvlinkdz(zsam,nu)
    dlogpy = 0d0
    dzdnu = -invlinkdn(zsam,nu)/invlinkdz(zsam,nu)
    dmudnu = 0d0
    djsam = (invlinkdzdn(zsam,nu) + invlinkhz(zsam,nu)*dzdnu) &
       /invlinkdz(zsam,nu)
    if (lmxi) zsam = zsam - xi
    call dsymv ('u',n,1d0,Ups,n,zsam,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
    zUz = dot_product(zsam,Upsz) + ssqdfsc
    lpdfz = ldh_Ups - modeldfh*log(zUz)
    dfozUz = modeldfh/zUz
    dz = -(2*dfozUz)*dot_product(Upsz,dzdnu)
    dj = -sum(djsam)
    lpdfy = 0d0
    dy = 0d0
    lglk(j) = lpdfz + lpdfy - sum(jsam)
    dlglkdnu(j) = dy + dz + dj
    zUDTUz = qform(Upsz,DT(:,:,1),n)
    dlglkdphi(j) = -trUpsDTh(1) + dfozUz*zUDTUz
    zUDTUz = qform(Upsz,DT(:,:,2),n)
    dlglkdnsq(j) = -trUpsDTh(2) + dfozUz*zUDTUz
    zUDTUz = qform(Upsz,DT(:,:,3),n)
    dlglkdkappa(j) = -trUpsDTh(3) + dfozUz*zUDTUz
  end do
  bfact = weigh_llik(icv, lglk, weights, QRin, Ntot)
  bfdnu = weigh_llik_deriv(icv, dlglkdnu, lglk, weights, QRin, Ntot)
  bfdphi = weigh_llik_deriv(icv, dlglkdphi, lglk, weights, QRin, Ntot)
  bfdnsq = weigh_llik_deriv(icv, dlglkdnsq, lglk, weights, QRin, Ntot)
  bfdkappa = weigh_llik_deriv(icv, dlglkdkappa, lglk, weights, QRin, Ntot)
end subroutine calcbd_mu


subroutine calcbd_wo (bfact, bfdnu, bfdphi, bfdnsq, bfdkappa, &
   phi, nu, omg, kappa, icf, Ntot, sample, weights, QRin, &
   n, p, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, y, l, F, offset, &
   dm, ifam, itr, icv)

  use modelfcns, logpdfzf => logpdfz, logpdfydmu => logpdfydlnk
  use covfun
  use betaprior
  use calcbd_fcns
  implicit none
  integer, intent(in) :: n, p, Ntot, ifam, icf, itr(n), icv
  double precision, intent(in) :: phi, omg, &
     kappa, nu, sample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(in) :: QRin(Ntot)
  double precision, intent(out) :: bfact, bfdphi, bfdnu, bfdnsq, bfdkappa
  logical lmxi
  double precision ssqdfsc, respdfh, modeldfh, tsqval
  integer j
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, lglk(Ntot), xi(n)
  double precision zsam(n), msam(n), jsam(n), sam(n), dlogpy(n), dzdnu(n), &
     dmudnu(n), djsam(n), lpdfz, lpdfy
  double precision dy, dz, dj, dlglkdnu(Ntot), dlglkdphi(Ntot), &
     dlglkdnsq(Ntot), dlglkdkappa(Ntot)
  double precision zUz, zUDTUz, DT(n,n,3), trUpsDTh(3), Upsz(n), dfozUz
  double precision ymm(n), ymml(n)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  do j = 1, 3
    DT(:,:,j) = cor_dcov(n, dm, phi, omg, kappa, j)
    trUpsDTh(j) = .5d0*traceAB(Ups,DT(:,:,j),n)
  end do

  do j = 1, Ntot
    call rchkusr
    sam = sample(:,j)
    zsam = transfw(sam,nu)
    msam = invlink(zsam,nu)
    jsam = loginvtrwdz(zsam,nu)
    dlogpy = logpdfydmu(y,l,msam)/tsq
    dzdnu = -invtrwdn(zsam,nu)/invtrwdz(zsam,nu)
    dmudnu = invlinkdn(zsam,nu) + invlinkdz(zsam,nu)*dzdnu
    djsam = (invtrwdzdn(zsam,nu) + invtrwhz(zsam,nu)*dzdnu)/invtrwdz(zsam,nu)
    if (lmxi) zsam = zsam - xi
    call dsymv ('u',n,1d0,Ups,n,zsam,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
    zUz = dot_product(zsam,Upsz) + ssqdfsc
    lpdfz = ldh_Ups - modeldfh*log(zUz)
    dfozUz = modeldfh/zUz
    dz = -(2*dfozUz)*dot_product(Upsz,dzdnu)
    dj = -sum(djsam)
    if (ifam .eq. 0) then
      ymm = y - msam
      ymml = l*ymm
      lpdfy = condymuf(ifam,n,y,l,msam,tsqval,respdfh)
      dy = dot_product(ymm,ymml) + tsqval
      dlogpy = (2*respdfh/dy)*ymml
      dy = dot_product(dlogpy,dmudnu)
    else
      lpdfy = condymuf(ifam,n,y,l,msam,tsqval,respdfh)
      dy = dot_product(dlogpy,dmudnu)
    end if
    lglk(j) = lpdfz + lpdfy - sum(jsam)
    dlglkdnu(j) = dy + dz + dj
    zUDTUz = qform(Upsz,DT(:,:,1),n)
    dlglkdphi(j) = -trUpsDTh(1) + dfozUz*zUDTUz
    zUDTUz = qform(Upsz,DT(:,:,2),n)
    dlglkdnsq(j) = -trUpsDTh(2) + dfozUz*zUDTUz
    zUDTUz = qform(Upsz,DT(:,:,3),n)
    dlglkdkappa(j) = -trUpsDTh(3) + dfozUz*zUDTUz
  end do
  bfact = weigh_llik(icv, lglk, weights, QRin, Ntot)
  bfdnu = weigh_llik_deriv(icv, dlglkdnu, lglk, weights, QRin, Ntot)
  bfdphi = weigh_llik_deriv(icv, dlglkdphi, lglk, weights, QRin, Ntot)
  bfdnsq = weigh_llik_deriv(icv, dlglkdnsq, lglk, weights, QRin, Ntot)
  bfdkappa = weigh_llik_deriv(icv, dlglkdkappa, lglk, weights, QRin, Ntot)
end subroutine calcbd_wo


subroutine calcbd_tr (bfact, bfdnu, bfdphi, bfdnsq, bfdkappa, &
   phi, nu, omg, kappa, icf, Ntot, sample, weights, QRin, &
   n, p, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, y, l, F, offset, &
   dm, ifam, itr, icv)

  use modelfcns, logpdfzf => logpdfz, logpdfydmu => logpdfydlnk
  use covfun
  use betaprior
  use calcbd_fcns
  implicit none
  integer, intent(in) :: n, p, Ntot, ifam, icf, itr(n), icv
  double precision, intent(in) :: phi, omg, &
     kappa, nu, sample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(in) :: QRin(Ntot)
  double precision, intent(out) :: bfact, bfdphi, bfdnu, bfdnsq, bfdkappa
  logical lmxi
  double precision ssqdfsc, respdfh, modeldfh, tsqval
  integer j
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, lglk(Ntot), xi(n)
  double precision zsam(n), msam(n), jsam(n), sam(n), dlogpy(n), dzdnu(n), &
     dmudnu(n), djsam(n), lpdfz, lpdfy
  double precision dy, dz, dj, dlglkdnu(Ntot), dlglkdphi(Ntot), &
     dlglkdnsq(Ntot), dlglkdkappa(Ntot)
  double precision zUz, zUDTUz, DT(n,n,3), trUpsDTh(3), Upsz(n), dfozUz
  double precision ymm(n), ymml(n)
  logical ltr0(n), ltr1(n), ltr2(n), ltrmu
  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  do j = 1, 3
    DT(:,:,j) = cor_dcov(n, dm, phi, omg, kappa, j)
    trUpsDTh(j) = .5d0*traceAB(Ups,DT(:,:,j),n)
  end do

  ltr0 = itr .eq. 0
  ltr1 = itr .eq. 1
  ltr2 = itr .eq. 2
  ltrmu = all(ltr1)

  do j = 1, Ntot
    call rchkusr
    sam = sample(:,j)
    where (ltr0)
      zsam = sam
      msam = invlink(zsam,nu)
      jsam = 0d0
      dlogpy = logpdfydmu(y,l,msam)/tsq
      dzdnu = 0d0
      dmudnu = invlinkdn(zsam,nu)
      djsam = 0d0
    elsewhere (ltr1)
      msam = sam
      zsam = flink(msam,nu)
      jsam = loginvlinkdz(zsam,nu)
      dlogpy = 0d0
      dzdnu = -invlinkdn(zsam,nu)/invlinkdz(zsam,nu)
      dmudnu = 0d0
      djsam = (invlinkdzdn(zsam,nu) + invlinkhz(zsam,nu)*dzdnu) &
         /invlinkdz(zsam,nu)
    elsewhere (ltr2)
      zsam = transfw(sam,nu)
      msam = invlink(zsam,nu)
      jsam = loginvtrwdz(zsam,nu)
      dlogpy = logpdfydmu(y,l,msam)/tsq
      dzdnu = -invtrwdn(zsam,nu)/invtrwdz(zsam,nu)
      dmudnu = invlinkdn(zsam,nu) + invlinkdz(zsam,nu)*dzdnu
      djsam = (invtrwdzdn(zsam,nu) + invtrwhz(zsam,nu)*dzdnu)/invtrwdz(zsam,nu)
    end where
    if (lmxi) zsam = zsam - xi
    call dsymv ('u',n,1d0,Ups,n,zsam,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
    zUz = dot_product(zsam,Upsz) + ssqdfsc
    lpdfz = ldh_Ups - modeldfh*log(zUz)
    dfozUz = modeldfh/zUz
    dz = -(2*dfozUz)*dot_product(Upsz,dzdnu)
    dj = -sum(djsam)
    if (ltrmu) then
      lpdfy = 0d0
      dy = 0d0
    else
      if (ifam .eq. 0) then
        ymm = y - msam
        ymml = l*ymm
        lpdfy = condymuf(ifam,n,y,l,msam,tsqval,respdfh)
        dy = dot_product(ymm,ymml) + tsqval
        dlogpy = (2*respdfh/dy)*ymml
        dy = dot_product(dlogpy,dmudnu)
      else
        lpdfy = condymuf(ifam,n,y,l,msam,tsqval,respdfh)
        dy = dot_product(dlogpy,dmudnu)
      end if
    end if
    lglk(j) = lpdfz + lpdfy - sum(jsam)
    dlglkdnu(j) = dy + dz + dj
    zUDTUz = qform(Upsz,DT(:,:,1),n)
    dlglkdphi(j) = -trUpsDTh(1) + dfozUz*zUDTUz
    zUDTUz = qform(Upsz,DT(:,:,2),n)
    dlglkdnsq(j) = -trUpsDTh(2) + dfozUz*zUDTUz
    zUDTUz = qform(Upsz,DT(:,:,3),n)
    dlglkdkappa(j) = -trUpsDTh(3) + dfozUz*zUDTUz
  end do
  bfact = weigh_llik(icv, lglk, weights, QRin, Ntot)
  bfdnu = weigh_llik_deriv(icv, dlglkdnu, lglk, weights, QRin, Ntot)
  bfdphi = weigh_llik_deriv(icv, dlglkdphi, lglk, weights, QRin, Ntot)
  bfdnsq = weigh_llik_deriv(icv, dlglkdnsq, lglk, weights, QRin, Ntot)
  bfdkappa = weigh_llik_deriv(icv, dlglkdkappa, lglk, weights, QRin, Ntot)
end subroutine calcbd_tr

module mcmcfcns
  implicit none
  private
  double precision, parameter :: bigpos = huge(1d0), bigneg = -bigpos
  public ini_mcmc, end_mcmc, sample_ssq, sample_tsq, sample_beta, &
     sample_cov, sample_z, sample_z0, samplez_gt, sample_z_mala, samplez_gt_mala

contains
  subroutine ini_mcmc (lglk, z, p0, phi, omg, kappa, y1, y2, F, offset, &
     icf, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, dft, n, p, ifam, &
     betQm0, zmxi, T, TiF, FTF, Ups, Upsz, zUz, ldh_Ups, &
     modeldfh, ssqdfsc, respdf, tsqdfsc, tsqyy, lnewcov)
    ! tsq is either tsq or tsqsc
    use interfaces
    use covfun
    use betaprior
    use modelfcns
    use linkfcns, only: invlink_ga
    use pdfy, only: logpdfy_gt
    implicit none
    integer, intent(in) :: n, p, ifam, icf
    double precision, intent(in) :: y1(n), y2(n), dm(n,n), &
       F(n,p), kappa, betm0(p), betQ0(p,p), ssqdf, ssqsc, &
       tsqdf, tsq, phi, omg, dft, offset(n)
    double precision, intent(out) :: zUz, ldh_Ups, modeldfh, ssqdfsc, &
       respdf, tsqdfsc, tsqyy
    double precision, intent(inout) :: z(n)
    logical, intent(out) :: lnewcov
    double precision, intent(out) :: lglk, p0(n), betQm0(p), zmxi(n), &
       T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), Upsz(n)
    logical lmxi
    integer i

    call create_spcor(icf,n)

    ! Starting value for z
    z = mustart(y1,y2)
    z = flink(z,dft)

    ssqdfsc = ssqdf*ssqsc

    ! Determine flat or normal prior
    call betapriorz (modeldfh, zmxi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)
    if (lmxi) then
      zmxi = z - zmxi
    else
      zmxi = z
    end if

    if (betQ0(1,1) .gt. 0d0) then
      call dsymv ('u',p,1d0,betQ0,p,betm0,1,0d0,betQm0,1) ! betQm0 = Q0*m0
    else
      betQm0 = 0d0
    end if

    call calc_cov (phi,omg,dm,F,betQ0, &
       kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)

    call dsymv ('u',n,1d0,Ups,n,zmxi,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
    zUz = dot_product(zmxi,Upsz) + ssqdfsc

    lnewcov = .true. ! Indicates whether to compute prediction matrix

    lglk = ldh_Ups - modeldfh*log(zUz)

    select case (ifam)
    case (0)
      tsqdfsc = tsqdf*tsq
      respdf = n + tsqdf
      tsqyy = tsqdfsc
      do i = 1, n
        p0(i) = invlink_ga(z(i),dft)
        tsqyy = tsqyy + logpdfy_gt(y1(i), y2(i), p0(i))
      end do
      lglk = lglk - .5d0*respdf*log(tsqyy)
    case default
      tsqyy = 0d0
      do i = 1, n
        p0(i) = invlink(z(i),dft)
        tsqyy = tsqyy + logpdfy(y1(i), y2(i), p0(i))
      end do
      lglk = lglk + tsqyy/tsq
    end select

    call rngini
  end subroutine ini_mcmc

  subroutine end_mcmc
    use interfaces, only: rngend
    call rngend
  end subroutine end_mcmc


  subroutine sample_ssq (ssq, modeldfh, zUz)
    use interfaces, only: randgamma
    implicit none
    double precision, intent(in) :: zUz, modeldfh
    double precision, intent(out) :: ssq
    ssq = randgamma(modeldfh)
    ssq = zUz/(ssq+ssq)
  end subroutine sample_ssq

  subroutine sample_tsq (tsq, respdf, tsqyy)
    use interfaces, only: randchisq
    implicit none
    double precision, intent(in) :: respdf, tsqyy
    double precision, intent(out) :: tsq
    tsq = randchisq(respdf)
    tsq = tsqyy/tsq
  end subroutine sample_tsq

  subroutine sample_beta (beta, z, ssq, n, p, betQm0, TiF, FTF)
    use interfaces, only: randnorm
    implicit none
    integer, intent(in) :: n, p
    double precision, intent(in) :: z(n), betQm0(p), TiF(n,p), FTF(p,p)
    double precision, intent(inout) :: beta(p), ssq
    integer j
    beta = betQm0
    call dgemv ('t',n,p,1d0,TiF,n,z,1,1d0,beta,1)
    call dtrmv ('u','t','n',p,FTF,p,beta,1)
    do j = 1, p
      beta(j) = sqrt(ssq)*randnorm() + beta(j)
    end do
    call dtrmv ('u','n','n',p,FTF,p,beta,1)
  end subroutine sample_beta

  subroutine sample_cov (lglk, phi, omg, kappa, &
     phipars, omgpars, kappapars, phisc, omgsc, kappasc, dm, &
     F, betQ0, n, p, acc, zmxi, T, TiF, FTF, Ups, &
     Upsz, lnewcov, zUz, ldh_Ups, modeldfh, ssqdfsc)
    use interfaces
    use covfun
    implicit none
    integer, intent(in) :: n, p
    logical, intent(inout) :: lnewcov
    integer, intent(inout):: acc
    double precision, intent(in) :: modeldfh, dm(n,n), F(n,p), &
       betQ0(p,p), zmxi(n), ssqdfsc
    double precision, intent(in) :: phipars(4), omgpars(4), kappapars(2), &
       phisc, omgsc, kappasc
    double precision, intent(inout) :: phi, omg, lglk, T(n,n), TiF(n,p), &
       kappa, FTF(p,p), Ups(n,n), Upsz(n), zUz, ldh_Ups
    double precision tr, ll, u, phi2, omg2, up3, kappa2, phi_kap2, phi_kap
    double precision FTF2(p,p), Ups2(n,n), Upsz2(n), &
       zUz2, ldh_Ups2, T2(n,n), TiF2(n,p)

    if (phisc .le. 0d0 .and. omgsc .le. 0d0 .and. kappasc .le. 0d0) return

    tr = 0d0

!! Sample phi
    if (phisc .gt. 0d0) then
      ll = phi - phipars(4)
      u = randnorm()*phisc ! u ~ N(0,phisc^2)
      phi2 = ll*exp(u) + phipars(4) ! log(phi2-p4) ~ N(log(phi-p4), phisc^2)
      ! Add logratio of priors
      up3 = u*phipars(3)
      if (up3 .lt. 0d0) then
        tr = tr + phipars(2)*u + exp(phipars(3)*log(ll/phipars(1)) &
           + flog1mexp(up3))
      else
        tr = tr + phipars(2)*u - exp(phipars(3)*log(ll/phipars(1)) &
           + flogexpm1(up3))
      end if
    else
      phi2 = phi
    end if

!! Sample omg
    if (omgsc .gt. 0d0) then
      ll = omg - omgpars(4)
      u = randnorm()*omgsc ! u ~ N(0,omgsc^2)
      omg2 = ll*exp(u) + omgpars(4) ! log(omg2-p4) ~ N(log(omg-p4), omgsc^2)
      ! Add logratio of priors
      up3 = u*omgpars(3)
      if (up3 .lt. 0d0) then
        tr = tr + omgpars(2)*u + exp(omgpars(3)*log(ll/omgpars(1)) &
           + flog1mexp(up3))
      else
        tr = tr + omgpars(2)*u - exp(omgpars(3)*log(ll/omgpars(1)) &
           + flogexpm1(up3))
      end if
    else
      omg2 = omg
    end if

!! Sample kappa
    if (kappasc .gt. 0d0) then
      phi_kap = logitfcn_ab(kappa,kappapars)
      u = randnorm()*kappasc
      phi_kap2 = phi_kap + u
      kappa2 = logitinv_ab(phi_kap2,kappapars)
      tr = tr + log_logitder_ab(kappa2,kappapars) &
         - log_logitder_ab(kappa,kappapars)
    else
      kappa2 = kappa
    end if

    if (tr .le. bigneg .or. tr .ne. tr) return

!! Compute covariance matrices for the new phi, omg, kappa
    call calc_cov (phi2,omg2,dm,F,betQ0, &
       kappa2,n,p,T2,TiF2,FTF2,Ups2,ldh_Ups2)

!! Compute log likelihood ratio
    call dsymv ('u',n,1d0,Ups2,n,zmxi,1,0d0,Upsz2,1) ! Upsz = Ups*(z-xi)
    zUz2 = dot_product(zmxi,Upsz2) + ssqdfsc
    ll = ldh_Ups2 - ldh_Ups - modeldfh*(log(zUz2) - log(zUz))
!! Add the logratio and transition probabilities
    tr = ll + tr

    u = randunif()
    u = log(u)
    if (u .lt. tr) then ! Update phi, omg, kappa
      lnewcov = .true.
      lglk = lglk + ll
      phi = phi2
      omg = omg2
      kappa = kappa2
      FTF = FTF2
      T = T2
      TiF = TiF2
      Ups = Ups2
      Upsz = Upsz2
      zUz = zUz2
      ldh_Ups = ldh_Ups2
      acc = acc + 1
    end if

  contains
    pure double precision function logitfcn_ab (x,ab)
      double precision, intent(in) :: x, ab(2)
      logitfcn_ab = log(x-ab(1)) - log(ab(2)-x)
    end function logitfcn_ab
    pure double precision function logitinv_ab (y,ab)
      double precision, intent(in) :: y, ab(2)
      logitinv_ab = ab(2) - (ab(2) - ab(1))/(1d0+exp(y))
    end function logitinv_ab
    pure double precision function log_logitder_ab (x,ab)
      double precision, intent(in) :: x, ab(2)
      log_logitder_ab = log(ab(2)-x) + log(x-ab(1))
    end function log_logitder_ab
  end subroutine sample_cov

  subroutine sample_z0 (z0,z,beta,ssq,phi,omg,n0,n,p,dmdm0,F,offset,F0,&
     offset0,kappa,icf, T, z0_ups, TC, FCTF, lnewcov)
    use interfaces
    use covfun
    implicit none
    integer, intent(in) :: n, n0, p, icf
    double precision, intent(in) :: z(n), beta(p), ssq, phi, omg, &
       dmdm0(n,n0), F(n,p), F0(n0,p), kappa, T(n,n), offset(n), offset0(n0)
    double precision, intent(out) :: z0(n0)
    double precision, intent(inout) :: TC(n,n0), FCTF(n0,p), &
       z0_ups(n0)
    logical, intent(inout) :: lnewcov
    integer j
    double precision z0_sd(n0), z0_mean(n0), zmo(n)
    call create_spcor(icf,0)
    if (lnewcov) then
      call calc_cov_pred(z0_ups, TC, FCTF, phi, omg, dmdm0, F, &
         F0, kappa, T, n, n0, p)
      lnewcov = .false.
    end if
    zmo = z - offset
    z0_mean = offset0 ! o0
    call dgemv ('t',n,n0,1d0,TC,n,zmo,1,1d0,z0_mean,1) ! + C'*T^{-1}*(z-o)
    call dgemv ('n',n0,p,1d0,FCTF,n0,beta,1,1d0,z0_mean,1) ! + (F0-C'T^-1F)*beta
    z0_sd = sqrt(ssq)*z0_ups
    do j = 1, n0
      z0(j) = z0_sd(j)*randnorm() + z0_mean(j)
    end do
  end subroutine sample_z0



  subroutine sample_z (lglk,z,p0,y1,y2,dft,ssq,tsq,zmxi,Ups,Upsz,zUz, &
     modeldfh,n)
    use interfaces
    use modelfcns
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), dft, ssq, tsq, Ups(n,n), &
       modeldfh
    double precision, intent(inout) :: z(n), lglk, p0(n), zmxi(n), Upsz(n),&
       zUz
    integer j
    double precision uj(n), z_mean, u, zz, pp, ll

    do j = 1, n
      uj = (/Ups(1:j,j),Ups(j,j+1:n)/) ! jth column of Ups
      z_mean = z(j) - Upsz(j)/uj(j)
      u = randnorm()
      zz = u*sqrt(ssq/uj(j)) + z_mean
      pp = invlink(zz,dft)
      ll = logdffy(y1(j), y2(j), pp, p0(j))
      ll = ll/tsq
      if (ll .le. bigneg .or. ll .ne. ll) return
      u = randunif()
      u = log(u)
      if (u .lt. ll) then ! Update z(j)
        u = zz - z(j)
        z(j) = zz
        p0(j) = pp
        zmxi(j) = zmxi(j) + u
        Upsz = Upsz + uj*u
        zz = zUz
        zUz = zUz + 2d0*u*Upsz(j) - uj(j)*u*u
        lglk = lglk + ll - modeldfh*(log(zUz) - log(zz))
      end if
    end do
  end subroutine sample_z

  subroutine sample_z_mala (lglk,z,p0,y1,y2,dft,ssq,tsq,zmxi,Ups,Upsz,zUz, &
     modeldfh,n,z_eps,iupdt)
    use interfaces
    use modelfcns
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), dft, ssq, tsq, Ups(n,n), &
       modeldfh, z_eps
    integer, intent(inout) :: iupdt
    double precision, intent(inout) :: z(n), lglk, p0(n), zmxi(n), Upsz(n),&
       zUz
    integer j
    double precision u(n), logfz_grad(n), logfy_grad(n), dmudz(n), &
       gradz(n), gradznew(n), znew(n), zmxinew(n), &
       logacc, r, Upsznew(n), p0new(n), zUznew, ll, zz, zznew, zzdiff
    logfz_grad = -Upsz/ssq
    dmudz = invlinkdz(z,dft)
    logfy_grad = logpdfydlnk(y1,y2,p0)/tsq
    gradz = logfy_grad*dmudz + logfz_grad
    do j = 1, n
      u(j) = randnorm()
    end do
    u = z_eps*gradz + sqrt(z_eps+z_eps)*u
    znew = z + u
    zmxinew = zmxi + u
    call dsymv('u', n, 1d0, Ups, n, zmxinew, 1, 0d0, Upsznew, 1)
    logfz_grad = -Upsznew/ssq
    dmudz = invlinkdz(znew,dft)
    p0new = invlink(znew,dft)
    logfy_grad = logpdfydlnk(y1,y2,p0new)/tsq
    gradznew = logfy_grad*dmudz + logfz_grad
    zz = dot_product(z, Upsz)
    zznew = dot_product(zmxinew, Upsznew)
    zzdiff = zznew - zz
    gradz = u - z_eps*gradz
    gradznew = u + z_eps*gradznew
    ll = sum(logdffy(y1,y2,p0new,p0))/tsq
    logacc =  ll -.5d0*zzdiff/ssq &
       + .25d0/z_eps*dot_product(gradz,gradz) &
       - .25d0/z_eps*dot_product(gradznew,gradznew)
    if (logacc .le. bigneg .or. logacc .ne. logacc) return
    r = randunif()
    r = log(r)
    if (r .lt. logacc) then
      iupdt = iupdt + 1
      zUznew = zUz + zzdiff
      lglk = lglk + ll - modeldfh*(log(zUznew) - log(zUz))
      z = znew
      p0 = p0new
      zmxi = zmxinew
      Upsz = Upsznew
      zUz = zUznew
    end if
  end subroutine sample_z_mala
  

!!! Sample GRF with Gaussian response from the transformed model, i.e.
!!! also estimate tsq
!!!
!!! @param z    The GRF sample
!!! @param ym   Average at each location
!!! @param l    Sample size at each location
!!! @param dft  Link parameter
!!! @param ssq  Variance parameter
!!! @param tsq  Dispersion parameter
!!! @param n	Number of locations
  subroutine samplez_gt (lglk,z,p0,ym,l,dft,ssq,zmxi,Ups,Upsz,zUz, &
     modeldfh,respdf,tsqyy,n)
    use interfaces
    use modelfcns
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: ym(n), l(n), dft, ssq, Ups(n,n), &
       modeldfh, respdf
    double precision, intent(inout) :: z(n), lglk, p0(n), zmxi(n), Upsz(n),&
       zUz, tsqyy
    integer j
    double precision uj(n), z_mean, u, zz, pp, ll, tsqyy2

    do j = 1, n
      uj = (/Ups(1:j,j),Ups(j,j+1:n)/) ! jth column of Ups
      z_mean = z(j) - Upsz(j)/uj(j)
      u = randnorm()
      zz = u*sqrt(ssq/uj(j)) + z_mean
      pp = invlink(zz,dft) ! Mean of y
      tsqyy2 = tsqyy + logdffy(ym(j), l(j), pp, p0(j))
      ll = -.5d0*respdf*(log(tsqyy2) - log(tsqyy))
      if (ll .le. bigneg .or. ll .ne. ll) return
      u = randunif()
      u = log(u)
      if (u .lt. ll) then ! Update z(j)
        tsqyy = tsqyy2
        u = zz - z(j)
        z(j) = zz
        p0(j) = pp
        zmxi(j) = zmxi(j) + u
        Upsz = Upsz + uj*u
        zz = zUz
        zUz = zUz + 2d0*u*Upsz(j) - uj(j)*u*u
        lglk = lglk + ll - modeldfh*(log(zUz) - log(zz))
      end if
    end do
  end subroutine samplez_gt


  subroutine samplez_gt_mala (lglk,z,p0,ym,l,dft,ssq,zmxi,Ups,Upsz,zUz, &
     modeldfh,respdf,tsqyy,n,z_eps,iupdt)
    use interfaces
    use modelfcns
    implicit none
    integer, intent(in) :: n
    integer, intent(inout) :: iupdt
    double precision, intent(in) :: ym(n), l(n), dft, ssq, Ups(n,n), &
       modeldfh, respdf, z_eps
    double precision, intent(inout) :: z(n), lglk, p0(n), zmxi(n), Upsz(n),&
       zUz, tsqyy
    integer j
    double precision u(n), logfz_grad(n), logfy_grad(n), dmudz(n), &
       gradz(n), gradznew(n), znew(n), zmxinew(n), &
       logacc, r, Upsznew(n), p0new(n), zUznew, ll, zz, zznew, zzdiff, &
       tsqyynew
    logfz_grad = -Upsz/ssq
    dmudz = invlinkdz(z,dft)
    logfy_grad = logpdfydlnk(ym,l,p0)
    gradz = logfy_grad*dmudz + logfz_grad
    do j = 1, n
      u(j) = randnorm()
    end do
    u = z_eps*gradz + sqrt(z_eps+z_eps)*u
    znew = z + u
    zmxinew = zmxi + u
    call dsymv('u', n, 1d0, Ups, n, zmxinew, 1, 0d0, Upsznew, 1)
    logfz_grad = -Upsznew/ssq
    dmudz = invlinkdz(znew,dft)
    p0new = invlink(znew,dft)
    logfy_grad = logpdfydlnk(ym,l,p0new)
    gradznew = logfy_grad*dmudz + logfz_grad
    zz = dot_product(z, Upsz)
    zznew = dot_product(zmxinew, Upsznew)
    zzdiff = zznew - zz
    gradz = u - z_eps*gradz
    gradznew = u + z_eps*gradznew
    tsqyynew = tsqyy + sum(logdffy(ym, l, p0new, p0))
    ll = -.5d0*respdf*(log(tsqyynew) - log(tsqyy))
    logacc =  ll -.5d0*zzdiff/ssq &
       + .25d0/z_eps*dot_product(gradz,gradz) &
       - .25d0/z_eps*dot_product(gradznew,gradznew)
    if (logacc .le. bigneg .or. logacc .ne. logacc) return
    r = randunif()
    r = log(r)
    if (r .lt. logacc) then
      iupdt = iupdt + 1
      zUznew = zUz + zzdiff
      lglk = lglk + ll - modeldfh*(log(zUznew) - log(zUz))
      z = znew
      p0 = p0new
      zmxi = zmxinew
      Upsz = Upsznew
      zUz = zUznew
      tsqyy = tsqyynew
    end if
  end subroutine samplez_gt_mala
end module mcmcfcns

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Computes the conditional log-pdf of y
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module pdfy
contains
!!!!!!!!!!!!!! Compute the log-pdf of y for given parameter !!!!!!!!!!!!!!!
  pure elemental double precision function logpdfy_gt (y1, y2, par)
    ! Transformed Gaussian
    ! y1 :: The average across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    double precision d
    d = y1 - par
    logpdfy_gt = y2*d*d
  end function logpdfy_gt

  pure elemental double precision function logpdfy_ga (y1, y2, par)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy_ga = y1*par - .5d0*y2*par*par
  end function logpdfy_ga

  pure elemental double precision function logpdfy_bi (y1, y2, par)
    ! Binomial
    ! To get the asymmetric version, switch y1 and y2
    ! y1 :: The number of successes
    ! y2 :: The number of faillures
    ! par :: The logartithm of the probability of success
    use interfaces, only: flog1mexp
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy_bi = y1*par + y2*flog1mexp(par)
  end function logpdfy_bi

  pure elemental double precision function logpdfy_po (y1, y2, par)
    ! Poisson
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy_po = y1*par - y2*exp(par)
  end function logpdfy_po

  pure elemental double precision function logpdfy_gm (y1, y2, par)
    ! Gamma
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy_gm = -y1*exp(-par) - y2*par
  end function logpdfy_gm


!!!!!!!!! Compute the difference log pdf of y wrt two parameters !!!!!!!!!!
  pure elemental double precision function logdffy_gt (y1, y2, p1, p2)
    ! Transformed Gaussian
    ! y1 :: The average across all realisations
    ! y2 :: The number of replications
    ! p1, p2 :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    double precision d1, d2
    d1 = y1 - p1
    d2 = y1 - p2
    logdffy_gt = y2*(d1*d1 - d2*d2)
  end function logdffy_gt

  pure elemental double precision function logdffy_ga (y1, y2, p1, p2)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! p1, p2 :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    logdffy_ga = y1*(p1 - p2) - .5d0*y2*(p1*p1 - p2*p2)
  end function logdffy_ga

  pure elemental double precision function logdffy_bi (y1, y2, p1, p2)
    ! Binomial
    ! To get the asymmetric version, switch y1 and y2
    ! y1 :: The number of successes
    ! y2 :: The number of faillures
    ! p1, p2 :: The logartithm of the probability of success
    use interfaces, only: flog1mexp
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    logdffy_bi = y1*(p1 - p2) + y2*(flog1mexp(p1) - flog1mexp(p2))
  end function logdffy_bi

  pure elemental double precision function logdffy_po (y1, y2, p1, p2)
    ! Poisson
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! p1, p2 :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    logdffy_po = y1*(p1 - p2) - y2*(exp(p1) - exp(p2))
  end function logdffy_po

  pure elemental double precision function logdffy_gm (y1, y2, p1, p2)
    ! Gamma
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! p1, p2 :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    logdffy_gm = -y1*(exp(-p1) - exp(-p2)) - y2*(p1 - p2)
  end function logdffy_gm
end module pdfy



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Computes the log-pdf of z
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module pdfz
contains
  function logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, modeldfh, xi(n)
    double precision logpdfz
    double precision Upsz(n), zUz, zmxi(n)
    if (lmxi) then
      zmxi = z - xi
      call dsymv ('u',n,1d0,Ups,n,zmxi,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
      zUz = dot_product(zmxi,Upsz) + ssqdfsc
    else
      call dsymv ('u',n,1d0,Ups,n,z,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
      zUz = dot_product(z,Upsz) + ssqdfsc
    end if
    logpdfz = ldh_Ups - modeldfh*log(zUz)
  end function logpdfz
end module pdfz





!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Computes the conditional log-pdf of y|z
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module condyz
contains
  function condyz_gt (n, y1, y2, z, nu, tsqdfsc, respdfh)
    use pdfy, only: logpdfy_gt
    use linkfcns, only: invlink_ga
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsqdfsc, respdfh
    double precision condyz_gt
    integer i
    double precision mu, lfy
    lfy = tsqdfsc
    do i = 1, n
      mu = invlink_ga(z(i),nu)
      lfy = lfy + logpdfy_gt(y1(i), y2(i), mu)
    end do
    condyz_gt = -respdfh*log(lfy)
  end function condyz_gt

  function condyz_ga (n, y1, y2, z, nu, tsq)
    use pdfy, only: logpdfy_ga
    use linkfcns, only: invlink_ga
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision condyz_ga
    integer i
    double precision mu, lfy
    lfy = 0d0
    do i = 1, n
      mu = invlink_ga(z(i),nu)
      lfy = lfy + logpdfy_ga(y1(i), y2(i), mu)
    end do
    condyz_ga = lfy/tsq
  end function condyz_ga

  function condyz_bi (n, y1, y2, z, nu, tsq)
    use pdfy, only: logpdfy_bi
    use linkfcns, only: invlink_bi
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision condyz_bi
    integer i
    double precision mu, lfy
    lfy = 0d0
    do i = 1, n
      mu = invlink_bi(z(i),nu)
      lfy = lfy + logpdfy_bi(y1(i), y2(i), mu)
    end do
    condyz_bi = lfy/tsq
  end function condyz_bi

  function condyz_bw (n, y1, y2, z, nu, tsq) ! Bin Wallace
    use pdfy, only: logpdfy_bi
    use linkfcns, only: invlink_bw
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision condyz_bw
    integer i
    double precision mu, lfy
    lfy = 0d0
    do i = 1, n
      mu = invlink_bw(z(i),nu)
      lfy = lfy + logpdfy_bi(y1(i), y2(i), mu)
    end do
    condyz_bw = lfy/tsq
  end function condyz_bw

  function condyz_po (n, y1, y2, z, nu, tsq)
    use pdfy, only: logpdfy_po
    use linkfcns, only: invlink_po
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision condyz_po
    integer i
    double precision mu, lfy
    lfy = 0d0
    do i = 1, n
      mu = invlink_po(z(i),nu)
      lfy = lfy + logpdfy_po(y1(i), y2(i), mu)
    end do
    condyz_po = lfy/tsq
  end function condyz_po

  function condyz_gm (n, y1, y2, z, nu, tsq)
    use pdfy, only: logpdfy_gm
    use linkfcns, only: invlink_gm
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision condyz_gm
    integer i
    double precision mu, lfy
    lfy = 0d0
    do i = 1, n
      mu = invlink_gm(z(i),nu)
      lfy = lfy + logpdfy_gm(y1(i), y2(i), mu)
    end do
    condyz_gm = lfy/tsq
  end function condyz_gm

  function condyz_ba (n, y1, y2, z, nu, tsq)
    use pdfy, only: logpdfy_bi
    use linkfcns, only: invlink_ba
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision condyz_ba
    integer i
    double precision mu, lfy
    lfy = 0d0
    do i = 1, n
      mu = invlink_ba(z(i),nu)
      lfy = lfy + logpdfy_bi(y2(i), y1(i), mu)
    end do
    condyz_ba = lfy/tsq
  end function condyz_ba

  function condyz_bd (n, y1, y2, z, nu, tsq)
    use pdfy, only: logpdfy_bi
    use linkfcns, only: invlink_bd
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision condyz_bd
    integer i
    double precision mu, lfy
    lfy = 0d0
    do i = 1, n
      mu = invlink_bd(z(i),nu)
      lfy = lfy + logpdfy_bi(y2(i), y1(i), mu)
    end do
    condyz_bd = lfy/tsq
  end function condyz_bd
end module condyz





!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Computes the joint log-pdf of (y,z)
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module jointyz
contains
  function jointyz_gt(n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsqdfsc, modeldfh, respdfh)
    use condyz, only: condyz_gt
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsqdfsc, modeldfh, respdfh, nu, xi(n)
    double precision jointyz_gt
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz_gt(n, y, l, z, nu, tsqdfsc, respdfh)
    jointyz_gt = lfz + lfy
  end function jointyz_gt

  function jointyz_ga (n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condyz, only: condyz_ga
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsq, nu, modeldfh, xi(n)
    double precision jointyz_ga
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz_ga(n, y, l, z, nu, tsq)
    jointyz_ga = lfz + lfy
  end function jointyz_ga

  function jointyz_bi (n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condyz, only: condyz_bi
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsq, nu, modeldfh, xi(n)
    double precision jointyz_bi
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz_bi(n, y, l, z, nu, tsq)
    jointyz_bi = lfz + lfy
  end function jointyz_bi

  function jointyz_bw (n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condyz, only: condyz_bw
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsq, nu, modeldfh, xi(n)
    double precision jointyz_bw
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz_bw(n, y, l, z, nu, tsq)
    jointyz_bw = lfz + lfy
  end function jointyz_bw

  function jointyz_po (n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condyz, only: condyz_po
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsq, nu, modeldfh, xi(n)
    double precision jointyz_po
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz_po(n, y, l, z, nu, tsq)
    jointyz_po = lfz + lfy
  end function jointyz_po

  function jointyz_gm (n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condyz, only: condyz_gm
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsq, nu, modeldfh, xi(n)
    double precision jointyz_gm
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz_gm(n, y, l, z, nu, tsq)
    jointyz_gm = lfz + lfy
  end function jointyz_gm

  function jointyz_ba (n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condyz, only: condyz_ba
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsq, nu, modeldfh, xi(n)
    double precision jointyz_ba
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz_ba(n, y, l, z, nu, tsq)
    jointyz_ba = lfz + lfy
  end function jointyz_ba

  function jointyz_bd (n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condyz, only: condyz_bd
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsq, nu, modeldfh, xi(n)
    double precision jointyz_bd
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz_bd(n, y, l, z, nu, tsq)
    jointyz_bd = lfz + lfy
  end function jointyz_bd
end module jointyz






!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Computes the log-pdf of mu
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module pdfmu
  implicit none
  double precision, parameter :: bigneg = -huge(1d0)
  private bigneg
contains
  function logpdfmu_ga (n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    use linkfcns, only: flink_ga
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, modeldfh
    double precision logpdfmu_ga
    integer i
    double precision z(n), logjac, lfz
    ! Linear predictor
    do i = 1, n
      z(i) = flink_ga(mu(i), nu)
    end do
    ! Jacobian
    if (nu .gt. 0d0) then
      logjac = 0d0
      do i = 1, n
        logjac = logjac + log(abs(mu(i)))
      end do
      logjac = (nu - 1d0)*logjac
    else if (all(mu .gt. 0d0)) then
      logjac = 0d0
      do i = 1, n
        logjac = logjac + log(mu(i))
      end do
      logjac = (nu - 1d0)*logjac
    else
      logjac = bigneg
    end if
    ! log-likelihood for z
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    ! Put all together
    logpdfmu_ga = lfz + logjac
  end function logpdfmu_ga

  function logpdfmu_bi (n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    use linkfcns, only: flink_bi
    use interfaces, only: flog1pexp, flog1p, logpdft!, logpdfnorm, logpdflogis
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, modeldfh
    double precision logpdfmu_bi
    integer i
    double precision z(n), logjac, lfz, tmp
    ! Linear predictor
    do i = 1, n
      z(i) = flink_bi(mu(i), nu)
    end do
    ! Jacobian
    logjac = 0d0
    if (nu .gt. 0d0) then ! pdf t
      do i = 1, n
        logjac = logjac - logpdft(z(i), nu)
      end do
    else if (nu .lt. 0d0) then ! pdf logistic
      do i = 1, n
        tmp = -z(i)
        logjac = logjac - tmp + 2d0*flog1pexp(tmp)
        ! - logpdflogis(z(i))
      end do
    else ! pdf normal
      do i = 1, n
        logjac = logjac + .5*z(i)*z(i) ! - logpdfnorm(z(i))
      end do
    end if
    ! log-likelihood for z
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    ! Put all together
    logpdfmu_bi = lfz + logjac
  end function logpdfmu_bi

  function logpdfmu_po (n, tht, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, &
     modeldfh)
    ! tht is log(mean)
    use linkfcns, only: flink_po
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: tht(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, modeldfh
    double precision logpdfmu_po
    integer i
    double precision z(n), logjac, lfz
    ! Linear predictor
    do i = 1, n
      z(i) = flink_po(tht(i), nu)
    end do
    ! Jacobian
    logjac = 0d0
    if (nu .gt. 0d0) then ! Using extended link
      do i = 1, n
        logjac = logjac + abs(tht(i))
      end do
      logjac = nu*logjac
    else if (nu .lt. 0d0) then
      do i = 1, n
        logjac = logjac + tht(i)
      end do
      logjac = nu*logjac
    end if
    ! log-likelihood for z
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    ! Put all together
    logpdfmu_po = lfz + logjac
  end function logpdfmu_po

  function logpdfmu_gm (n, tht, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, &
     modeldfh)
    ! tht is log(mean)
    use linkfcns, only: flink_gm
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: tht(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, modeldfh
    double precision logpdfmu_gm
    integer i
    double precision z(n), logjac, lfz
    ! Linear predictor
    do i = 1, n
      z(i) = flink_gm(tht(i), nu)
    end do
    ! Jacobian
    logjac = 0d0
    if (nu .gt. 0d0) then ! Using extended link
      do i = 1, n
        logjac = logjac + abs(tht(i))
      end do
      logjac = nu*logjac
    else if (nu .lt. 0d0) then
      do i = 1, n
        logjac = logjac + tht(i)
      end do
      logjac = nu*logjac
    end if
    ! log-likelihood for z
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    ! Put all together
    logpdfmu_gm = lfz + logjac
  end function logpdfmu_gm

  function logpdfmu_ba (n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    use linkfcns, only: flink_ba
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, modeldfh
    double precision logpdfmu_ba
    integer i
    double precision z(n), logjac, lfz, logjac1, logjac2
    ! Linear predictor
    do i = 1, n
      z(i) = flink_ba(mu(i), nu)
    end do
    ! Jacobian
    logjac1 = 0d0
    logjac2 = 0d0
    do i = 1, n
      logjac1 = logjac1 + log(-mu(i))
      logjac2 = logjac2 - mu(i)
    end do
    logjac = (nu - 1d0)*logjac1 + logjac2
    ! log-likelihood for z
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    ! Put all together
    logpdfmu_ba = lfz + logjac
  end function logpdfmu_ba

  function logpdfmu_bd (n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    use linkfcns, only: flink_bd
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, modeldfh
    double precision logpdfmu_bd
    integer i
    double precision z(n), logjac, lfz, logjac1, logjac2
    ! Linear predictor
    do i = 1, n
      z(i) = flink_bd(mu(i), nu)
    end do
    ! Jacobian
    logjac1 = 0d0
    logjac2 = 0d0
    do i = 1, n
      logjac1 = logjac1 + log(-mu(i))
      logjac2 = logjac2 - mu(i)
    end do
    logjac = (nu - 1d0)*logjac1 + logjac2
    ! log-likelihood for z
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    ! Put all together
    logpdfmu_bd = lfz + logjac
  end function logpdfmu_bd

  function logpdfmu_bw (n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    use linkfcns, only: flink_bw
    use interfaces, only: flog1p
    use pdfz
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, modeldfh
    double precision logpdfmu_bw
    integer i
    double precision z(n), logjac, logjac1, logjac2, lfz, cnu, log1pzsq
    ! Linear predictor
    do i = 1, n
      z(i) = flink_bw(mu(i), nu)
    end do
    ! Jacobian
    cnu = (8d0*nu + 1d0)/(8d0*nu + 3d0)
    logjac1 = 0d0
    logjac2 = 0d0
    do i = 1, n
      if (z(i) .ne. 0d0) then
        log1pzsq = flog1p(z(i)*z(i)/nu)
        logjac1 = logjac1 + log(abs(z(i))) - .5d0*log(log1pzsq) - log1pzsq
        logjac2 = logjac2 - log1pzsq
      end if
    end do
    logjac = .5d0*n*log(nu) - n*log(cnu) - logjac1 - .5d0*cnu*cnu*nu*logjac2
    ! log-likelihood for z
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    ! Put all together
    logpdfmu_bw = lfz + logjac
  end function logpdfmu_bw
end module pdfmu







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Computes the conditional log-pdf of y|mu
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module condymu
contains
  pure function condymu_gt (n, y1, y2, mu, tsqdfsc, respdfh)
    use pdfy, only: logpdfy_gt
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqdfsc, &
       respdfh
    double precision condymu_gt
    integer i
    double precision lfy
    lfy = tsqdfsc
    do i = 1, n
      lfy = lfy + logpdfy_gt(y1(i), y2(i), mu(i))
    end do
    condymu_gt = -respdfh*log(lfy)
  end function condymu_gt

  function condymu_ga (n, y1, y2, mu, tsq)
    use pdfy, only: logpdfy_ga
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsq
    double precision condymu_ga
    integer i
    double precision lfy
    lfy = 0d0
    do i = 1, n
      lfy = lfy + logpdfy_ga(y1(i), y2(i), mu(i))
    end do
    condymu_ga = lfy/tsq
  end function condymu_ga

  function condymu_bi (n, y1, y2, mu, tsq)
    use pdfy, only: logpdfy_bi
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsq
    double precision condymu_bi
    integer i
    double precision lfy
    lfy = 0d0
    do i = 1, n
      lfy = lfy + logpdfy_bi(y1(i), y2(i), mu(i))
    end do
    condymu_bi = lfy/tsq
  end function condymu_bi

  function condymu_po (n, y1, y2, mu, tsq)
    use pdfy, only: logpdfy_po
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsq
    double precision condymu_po
    integer i
    double precision lfy
    lfy = 0d0
    do i = 1, n
      lfy = lfy + logpdfy_po(y1(i), y2(i), mu(i))
    end do
    condymu_po = lfy/tsq
  end function condymu_po

  function condymu_gm (n, y1, y2, mu, tsq)
    use pdfy, only: logpdfy_gm
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsq
    double precision condymu_gm
    integer i
    double precision lfy
    lfy = 0d0
    do i = 1, n
      lfy = lfy + logpdfy_gm(y1(i), y2(i), mu(i))
    end do
    condymu_gm = lfy/tsq
  end function condymu_gm

  function condymu_ba (n, y1, y2, mu, tsq)
    use pdfy, only: logpdfy_bi
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsq
    double precision condymu_ba
    integer i
    double precision lfy
    lfy = 0d0
    do i = 1, n
      lfy = lfy + logpdfy_bi(y2(i), y1(i), mu(i))
    end do
    condymu_ba = lfy/tsq
  end function condymu_ba

  function condymu_bd (n, y1, y2, mu, tsq)
    use pdfy, only: logpdfy_bi
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsq
    double precision condymu_bd
    integer i
    double precision lfy
    lfy = 0d0
    do i = 1, n
      lfy = lfy + logpdfy_bi(y2(i), y1(i), mu(i))
    end do
    condymu_bd = lfy/tsq
  end function condymu_bd

  function condymu_bw (n, y1, y2, mu, tsq)
    use pdfy, only: logpdfy_bi
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsq
    double precision condymu_bw
    integer i
    double precision lfy
    lfy = 0d0
    do i = 1, n
      lfy = lfy + logpdfy_bi(y1(i), y2(i), mu(i))
    end do
    condymu_bw = lfy/tsq
  end function condymu_bw
end module condymu







!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Computes the joint log-pdf of (y,mu)
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module jointymu
  implicit none
  double precision, parameter :: bigneg = -huge(1d0)
  private bigneg
contains
  function jointymu_gt (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsqdfsc, modeldfh, respdfh)
    use condymu, only: condymu_gt
    use pdfmu, only: logpdfmu_ga
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsqdfsc, modeldfh, respdfh
    double precision jointymu_gt
    double precision lfmu, lfy
    lfmu = logpdfmu_ga(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condymu_gt(n, y, l, mu, tsqdfsc, respdfh)
    jointymu_gt = lfmu + lfy
  end function jointymu_gt

  function jointymu_ga (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condymu, only: condymu_ga
    use pdfmu, only: logpdfmu_ga
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsq, modeldfh
    double precision jointymu_ga
    double precision lfmu, lfy
    lfmu = logpdfmu_ga(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condymu_ga(n, y, l, mu, tsq)
    jointymu_ga = lfmu + lfy
  end function jointymu_ga

  function jointymu_bi (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condymu, only: condymu_bi
    use pdfmu, only: logpdfmu_bi
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsq, modeldfh
    double precision jointymu_bi
    double precision lfmu, lfy
    lfmu = logpdfmu_bi(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condymu_bi(n, y, l, mu, tsq)
    jointymu_bi = lfmu + lfy
  end function jointymu_bi

  function jointymu_po (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condymu, only: condymu_po
    use pdfmu, only: logpdfmu_po
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsq, modeldfh
    double precision jointymu_po
    double precision lfmu, lfy
    lfmu = logpdfmu_po(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condymu_po(n, y, l, mu, tsq)
    jointymu_po = lfmu + lfy
  end function jointymu_po

  function jointymu_gm (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condymu, only: condymu_gm
    use pdfmu, only: logpdfmu_gm
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsq, modeldfh
    double precision jointymu_gm
    double precision lfmu, lfy
    lfmu = logpdfmu_gm(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condymu_gm(n, y, l, mu, tsq)
    jointymu_gm = lfmu + lfy
  end function jointymu_gm

  function jointymu_ba (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condymu, only: condymu_ba
    use pdfmu, only: logpdfmu_ba
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsq, modeldfh
    double precision jointymu_ba
    double precision lfmu, lfy
    lfmu = logpdfmu_ba(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condymu_ba(n, y, l, mu, tsq)
    jointymu_ba = lfmu + lfy
  end function jointymu_ba

  function jointymu_bd (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condymu, only: condymu_bd
    use pdfmu, only: logpdfmu_bd
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsq, modeldfh
    double precision jointymu_bd
    double precision lfmu, lfy
    lfmu = logpdfmu_bd(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, &
       modeldfh)
    lfy = condymu_bd(n, y, l, mu, tsq)
    jointymu_bd = lfmu + lfy
  end function jointymu_bd

  function jointymu_bw (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    use condymu, only: condymu_bw
    use pdfmu, only: logpdfmu_bw
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsq, modeldfh
    double precision jointymu_bw
    double precision lfmu, lfy
    lfmu = logpdfmu_bw(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condymu_bw(n, y, l, mu, tsq)
    jointymu_bw = lfmu + lfy
  end function jointymu_bw
end module jointymu

module calcbd_fcns

  private :: condymu_sp, weigh_llik_st, weigh_llik_cv, weigh_llik_deriv_st,&
     weigh_llik_deriv_cv

contains

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    use condymu, only: condymu_gt
    implicit none
    integer, intent(in) :: n, ifam
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf

  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    use modelfcns, only: condymu_mf => condymu
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure double precision function weigh_llik (icv, llik, lw, Qcv, n)
    implicit none
    integer, intent(in) :: n, icv
    double precision, intent(in) :: llik(n), lw(n), Qcv(n)
    select case (icv)
    case (0) ! No control variates
      weigh_llik = weigh_llik_st(llik,lw,Qcv,n)
    case (1) ! With control variates
      weigh_llik = weigh_llik_cv(llik,lw,Qcv,n)
    end select
  end function weigh_llik

  pure double precision function weigh_llik_st (llik, lw, Qcv, n) result (out)
    ! out = log(sum(exp(llik - lw)))
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: llik(n), lw(n), Qcv(n)
    double precision llikw(n)
    llikw = llik - lw
    out = maxval(llikw)
    llikw = exp(llikw - out)
    out = out + log(sum(llikw))
  end function weigh_llik_st

  pure double precision function weigh_llik_cv (llik, lw, Qcv, n) result (out)
    ! out = log(Qcv'*(n*exp(llik - lw)))
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: llik(n), lw(n), Qcv(n)
    double precision llikw(n), ql
    llikw = llik - lw
    out = maxval(llikw)
    llikw = exp(llikw - out)
    ql = dot_product(Qcv,llikw)
    if (ql .gt. 0d0) then
      out = out + log(n*ql)
    else
      out = -huge(1d0)
    end if
  end function weigh_llik_cv

  pure double precision function weigh_llik_deriv (icv, dllik, llik, lw, Qcv, n)
    implicit none
    integer, intent(in) :: n, icv
    double precision, intent(in) :: dllik(n), llik(n), lw(n), Qcv(n)
    select case (icv)
    case (0) ! No control variates
      weigh_llik_deriv = weigh_llik_deriv_st(dllik,llik,lw,Qcv,n)
    case (1) ! With control variates
      weigh_llik_deriv = weigh_llik_deriv_cv(dllik,llik,lw,Qcv,n)
    end select
  end function weigh_llik_deriv

  pure double precision function weigh_llik_deriv_st (dllik, llik, lw, Qcv, n) &
     result (out)
    ! out = log(sum(exp(llik - lw)))
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: dllik(n), llik(n), lw(n), Qcv(n)
    double precision llikw(n)
    llikw = llik - lw
    out = maxval(llikw)
    llikw = exp(llikw - out)
    llikw = llikw/sum(llikw)
    out = dot_product(dllik,llikw)
  end function weigh_llik_deriv_st

  pure double precision function weigh_llik_deriv_cv (dllik, llik, lw, Qcv, n) &
     result (out)
    ! out = log(Qcv'*(n*exp(llik - lw)))
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: dllik(n), llik(n), lw(n), Qcv(n)
    double precision llikw(n)
    llikw = llik - lw
    out = maxval(llikw)
    llikw = exp(llikw - out)
    llikw = Qcv*llikw
    llikw = llikw/sum(llikw)
    out = dot_product(dllik,llikw)
  end function weigh_llik_deriv_cv

  double precision function qform(v,A,n)
    ! Compute v'*A*v where A is symmetric with only the upper triangle
    ! specified.
    integer, intent(in) :: n
    double precision, intent(in) :: v(n), A(n,n)
    double precision Av(n)
    call dsymv ('u',n,1d0,A,n,v,1,0d0,Av,1)
    qform = dot_product(Av,v)
  end function qform

  double precision function traceAB(A,B,n)
    ! Compute the trace(matmul(A,B)) where A,B are symmetric matrices with
    ! only their upper triangular elements specified.
    integer, intent(in) :: n
    double precision, intent(in) :: A(n,n), B(n,n)
    integer j
    double precision c
    traceAB = A(1,1)*B(1,1)
    do j = 2, n
      c = dot_product(A(:j-1,j),B(:j-1,j))
      traceAB = traceAB + c + c + A(j,j)*B(j,j)
    end do
  end function traceAB

  double precision function cor_dcov (n, dm, phi, omg, kappa, id) result (DT)
    ! id = which derivative do you want? phi, omg, kappa
    ! Return dT where T = omg*I + R(phi,kappa)
    use covfun, only: spcor_dh, spcor_dk, fill_symmetric_matrix, upper_tri
    implicit none
    integer, intent(in) :: n, id
    double precision, intent(in) :: dm(n,n), phi, omg, kappa
    !double precision, intent(out) :: DT(n,n)
    dimension :: DT(n,n)
    integer i
    double precision H(n,n)
    select case (id)
    case (1)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = -(H/phi)*spcor_dh(H,kappa)
        end where
      end if
    case (2)
      DT = 0d0
      do i = 1, n
        DT(i,i) = 1d0
      end do
    case (3)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = spcor_dk(H,kappa)
        end where
      end if
    end select
  end function cor_dcov
end module calcbd_fcns

##' Standard errors for the empirical Bayes estimates of the parameters.
##'
##' @title Empirical Bayes standard errors
##' @param mcrun The output from the function \code{mcsglmm} where the
##'   parameters linkp, phi, omg, kappa are set at their empirical
##'   Bayes estimates (output of \code{\link{bf2optim}}).
##' @param transf The type of transformation to use.
##' @return The precision (inverse covariance) matrix.
##' @importFrom stats var
##' @useDynLib geoBayes llikfcn_dh_tr
##' @references Casella, G. (2001). Empirical Bayes Gibbs sampling.
##'   Biostatistics, 2(4), 485-500.
##'
##' Evangelou, E., & Roy, V. (2019). Estimation and prediction for
##'   spatial generalized linear mixed models with parametric links
##'   via reparameterized importance sampling. Spatial Statistics, 29,
##'   289-315.
##' @export
bf2se <- function(mcrun, transf = c("no", "mu", "wo"))
{
  family <- mcrun$MODEL$family
  ifam <- .geoBayes_family(family)
  corrfcn <- mcrun$MODEL$corrfcn
  icf <- .geoBayes_correlation(corrfcn)
  F <- mcrun$DATA$modelmatrix
  offset <- mcrun$DATA$offset
  n <- NROW(F)
  p <- NCOL(F)
  loc <- mcrun$DATA$locations
  longlat <- mcrun$DATA$longlat
  dm <- spDists(loc, longlat = longlat)
  y <- mcrun$DATA$response
  l <- mcrun$DATA$weights
  linkp <- mcrun$FIXED$linkp_num[1]
  if (length(linkp) < 1) stop ("This function requires fixed linkp.")
  nu <- .geoBayes_getlinkp(linkp, ifam)
  phi <- mcrun$FIXED$phi[1]
  if (length(phi) < 1) stop ("This function requires fixed phi.")
  omg <- mcrun$FIXED$omg[1]
  if (length(omg) < 1) stop ("This function requires fixed omg.")
  kappa <- mcrun$FIXED$kappa[1]
  if (length(kappa) < 1) stop ("This function requires fixed kappa.")
  betm0 <- mcrun$MODEL$betm0
  betQ0 <- mcrun$MODEL$betQ0
  ssqdf <- mcrun$MODEL$ssqdf
  ssqsc <- mcrun$MODEL$ssqsc
  tsqdf <- mcrun$MODEL$tsqdf
  tsqsc <- mcrun$MODEL$tsqsc
  dispersion <- mcrun$MODEL$dispersion
  if (ifam == 0) {
    tsq <- tsqsc
  } else {
    tsq <- dispersion
    tsqdf <- 0
  }
  getsample <- transfsample(list(mcrun),
                            list(response = y, family = family), transf)
  sample <- matrix(unlist(getsample$sample), n)
  itr <- getsample$itr
  transf <- getsample$transf
  real_transf <- getsample$real_transf
  ifam <- getsample$ifam
  froutine <- paste0("llikfcn_dh_", real_transf)
  froutine <- "llikfcn_dh_tr" # XXX
  nsim <- NCOL(sample)
  dlglk <- array(0, c(4, nsim))
  hlglk <- array(0, c(16, nsim))
  fff <- .Fortran(froutine, dlglk, hlglk, phi, omg, linkp, kappa,
                  sample, nsim, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc,
                  tsqdf, tsq, icf, n, p, ifam, itr, PACKAGE = "geoBayes")
  dlglk <- fff[[1]]
  hlglk <- fff[[2]]
  out <- matrix(0, 4, 4)
  ll <- upper.tri(out, diag = TRUE)
  out[ll] <- rowMeans(hlglk[ll, ])
  out[lower.tri(out)] <- t(out)[lower.tri(out)]
  out <- out + var(t(dlglk))
  out <- -out
  out
}

##' Function to compute the Bayes factors from MCMC samples.
##'
##' Computes the Bayes factors using \code{method} with respect to
##' \code{reference}.
##' @title Computation of Bayes factors at the skeleton points
##' @param runs A list with outputs from the function
##'   \code{\link{mcsglmm}} or \code{\link{mcstrga}}.
##' @param bfsize1 A scalar or vector of the same length as
##'   \code{runs} with all integer values or all values in (0, 1]. How
##'   many samples (or what proportion of the sample) to use for
##'   estimating the Bayes factors at the first stage. The remaining
##'   sample will be used for estimating the Bayes factors in the
##'   second stage. Setting it to 1 will perform only the first stage.
##' @param method Which method to use to calculate the Bayes factors:
##' Reverse logistic or Meng-Wong.
##' @param reference Which model goes in the denominator.
##' @param transf Whether to use a transformed sample for the
##'   computations. If \code{"no"} or \code{FALSE}, it doesn't. If
##'   \code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
##'   \code{"wo"} it uses an alternative transformation. The latter
##'   can be used only for the families indicated by
##'   \code{.geoBayes_models$haswo}.
##' @return A list with components
##' \itemize{
##' \item \code{logbf} A vector containing logarithm of the Bayes factors.
##' \item \code{logLik1} \code{logLik2} Matrices with the values of
##' the log-likelihood computed from the samples for each model at the
##' first and second stages.
##' \item \code{isweights} A vector with the importance sampling
##' weights for computing the Bayes factors at new points that will be
##' used at the second stage. Used internally in
##' \code{\link{bf2new}} and \code{\link{bf2optim}}.
##' \item \code{controlvar} A matrix with the control variates
##' computed at the samples that will be used in the second stage.
##' \item \code{sample2} The MCMC sample for mu or z that will be
##' used in the second stage. Used internally in
##' \code{\link{bf2new}} and \code{\link{bf2optim}}.
##' \item \code{N1}, \code{N2} Vectors containing the sample sizes
##' used in the first and second stages.
##' \item \code{distmat} Matrix of distances between locations.
##' \item \code{betm0}, \code{betQ0}, \code{ssqdf}, \code{ssqsc},
##' \code{tsqdf}, \code{tsqsc}, \code{dispersion}, \code{response},
##' \code{weights}, \code{modelmatrix}, \code{locations},
##' \code{family}, \code{corrfcn}, \code{transf} Model parameters used
##' internally in.
##' \code{\link{bf2new}} and \code{\link{bf2optim}}.
##' \item \code{pnts} A list containing the skeleton points. Used
##' internally in \code{\link{bf2new}} and \code{\link{bf2optim}}.
##' }
##' @references Geyer, C. J. (1994). Estimating normalizing constants
##' and reweighting mixtures. Technical report, University of
##' Minnesota.
##'
##' Meng, X. L., & Wong, W. H. (1996). Simulating ratios of
##' normalizing constants via a simple identity: A theoretical
##' exploration. \emph{Statistica Sinica}, 6, 831-860.
##'
##' Roy, V., Evangelou, E., and Zhu, Z. (2015). Efficient estimation
##' and prediction for the Bayesian spatial generalized linear mixed
##' model with flexible link functions. \emph{Biometrics}, 72(1), 289-298.
##' @examples \dontrun{
##' data(rhizoctonia)
##' ### Define the model
##' corrf <- "spherical"
##' kappa <- 0
##' ssqdf <- 1
##' ssqsc <- 1
##' betm0 <- 0
##' betQ0 <- .01
##' family <- "binomial.probit"
##' ### Skeleton points
##' philist <- c(100, 140, 180)
##' omglist <- c(.5, 1)
##' parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
##' ### MCMC sizes
##' Nout <- 100
##' Nthin <- 1
##' Nbi <- 0
##' ### Take MCMC samples
##' runs <- list()
##' for (i in 1:NROW(parlist)) {
##'   runs[[i]] <- mcsglmm(Infected ~ 1, family, rhizoctonia, weights = Total,
##'                        atsample = ~ Xcoord + Ycoord,
##'                        Nout = Nout, Nthin = Nthin, Nbi = Nbi,
##'                        betm0 = betm0, betQ0 = betQ0,
##'                        ssqdf = ssqdf, ssqsc = ssqsc,
##'                        phi = parlist$phi[i], omg = parlist$omg[i],
##'                        linkp = parlist$linkp[i], kappa = parlist$kappa[i],
##'                        corrfcn = corrf,
##'                        corrtuning=list(phi = 0, omg = 0, kappa = 0))
##' }
##' bf <- bf1skel(runs)
##' bf$logbf
##' }
##' @importFrom sp spDists
##' @useDynLib geoBayes bfsp_no bfsp_mu bfsp_wo bfsp_tr
##' @export
bf1skel <- function(runs, bfsize1 = 0.80, method = c("RL", "MW"),
                    reference = 1, transf = c("no", "mu", "wo"))
{
  method <- match.arg(method)
  imeth <- match(method, eval(formals()$method))
  if (!all(sapply(runs, inherits, what = "geomcmc"))) {
    stop ("Input runs is not a list with elements of class geomcmc.")
  }
  nruns <- length(runs)
  if (nruns == 0) stop ("No runs specified")
  reference <- as.integer(reference)
  if (isTRUE(reference < 1L | reference > nruns)) {
    stop("Argument reference does not correspond to a run in runs.")
  }
  Nout <- sapply(runs, function(x) x$MCMC$Nout)
  Nout1 <- getsize(bfsize1, Nout, "*")
  Ntot1 <- sum(Nout1)
  Nout2 <- Nout - Nout1
  Ntot2 <- sum(Nout2)

  ## Check if fixed phi and omg
  if (!all(sapply(runs, function(x) length(x$FIXED$phi) == 1))) {
    stop("Each input runs must have a fixed value phi.")
  }
  if (!all(sapply(runs, function(x) length(x$FIXED$omg) == 1))) {
    stop("Each input runs must have a fixed value omg.")
  }

  ## Extract data and model
  nm_DATA <- c("response", "weights", "modelmatrix", "offset", "locations",
               "longlat")
  nm_MODEL <- c("family", "corrfcn", "betm0", "betQ0", "ssqdf", "ssqsc",
                "tsqdf", "tsqsc", "dispersion")
  DATA <- runs[[1]]$DATA[nm_DATA]
  MODEL <- runs[[1]]$MODEL[nm_MODEL]
  if (nruns > 1) {
    for (i in 2:nruns) {
      if (!identical(runs[[i]]$DATA[nm_DATA], DATA)) {
        stop("MCMC chains don't all correspond to the same data.")
      }
      if (!identical(runs[[i]]$MODEL[nm_MODEL], MODEL)) {
        stop("MCMC chains don't all correspond to the same model.")
      }
    }
  }
  y <- DATA$response
  n <- as.integer(length(y))
  l <- DATA$weights
  F <- DATA$modelmatrix
  offset <- DATA$offset
  p <- NCOL(F)
  loc <- DATA$locations
  dm <- sp::spDists(loc, longlat = DATA$longlat)
  family <- MODEL$family
  ## ifam <- .geoBayes_family(family)
  corrfcn <- MODEL$corrfcn
  icf <- .geoBayes_correlation(corrfcn)
  betm0 <- MODEL$betm0
  betQ0 <- MODEL$betQ0
  ssqdf <- MODEL$ssqdf
  ssqsc <- MODEL$ssqsc
  tsqdf <- MODEL$tsqdf
  tsqsc <- MODEL$tsqsc
  dispersion <- MODEL$dispersion

  ## Choose sample
  getsample <-
    transfsample(runs,
                 list(response = y, family = family), transf)
  sample <- matrix(unlist(getsample$sample), n)
  itr <- getsample$itr
  transf <- getsample$transf
  real_transf <- getsample$real_transf
  ifam <- getsample$ifam

  ## Skeleton points
  phi_pnts <- as.double(sapply(runs, function(r) r$FIXED$phi))
  omg_pnts <- as.double(sapply(runs, function(r) r$FIXED$omg))
  nu_pnts <- as.double(sapply(runs, function(r) r$FIXED$linkp_num))
  if (.geoBayes_corrfcn$needkappa[icf]) {
    kappa_pnts <- sapply(runs, function(r) r$FIXED$kappa)
    kappa_pnts <- .geoBayes_getkappa(kappa_pnts, icf)
  } else {
    kappa_pnts <- rep(0, nruns)
  }

  bfroutine <- paste0("bfsp_", real_transf)

  if (nruns == 1) {
    MCMC <- runs[[1]]$MCMC
    out <- list(logbf = 1, logLik1 = MCMC$logLik[1:Ntot1],
                logLik2 = MCMC$logLik[-(1:Ntot1)],
                isweights = rep.int(0, Ntot2),
                controlvar = matrix(1, Ntot2, 1),
                z = sample[[1]][, -(1:Ntot1), drop = FALSE],
                N1 = Nout1, N2 = Nout2,
                betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf,
                ssqsc = ssqsc, tsqdf = tsqdf, tsqsc = tsqsc,
                dispersion = dispersion, response = y,
                weights = l, modelmatrix = F, offset = offset, 
                locations = loc, longlat = DATA$longlat,
                distmat = dm,
                family = family,
                referencebf = 0, corrfcn = corrfcn, transf = transf,
                real_transf = real_transf, itr = itr,
                pnts = list(nu = nu_pnts, phi = phi_pnts, omg = omg_pnts,
                  kappa = kappa_pnts))
    return(out)
  }

  ## Split the sample
  sel1 <- rep(rep(c(TRUE, FALSE), nruns), rbind(Nout1, Nout2))
  z1 <- sample[, sel1, drop = FALSE]
  z2 <- sample[, !sel1, drop = FALSE]

  logbf <- numeric(nruns)
  lglk1 <- matrix(0., Ntot1, nruns)
  lglk2 <- matrix(0., Ntot2, nruns)
  zcv <- matrix(0., Ntot2, nruns)
  weights <- numeric(Ntot2)
  if (ifam == 0) {
    tsq <- tsqsc
  } else {
    tsq <- dispersion
  }
  RUN <- .Fortran(bfroutine,
                  weights = weights,
                  zcv = zcv,
                  logbf = logbf,
                  lglk1 = lglk1,
                  lglk2 = lglk2,
                  as.double(phi_pnts), as.double(omg_pnts),
                  as.double(nu_pnts), as.double(z1),
                  as.integer(Nout1), as.integer(Ntot1),
                  as.double(z2), as.integer(Nout2), as.integer(Ntot2),
                  as.double(y), as.double(l), as.double(F), as.double(offset), 
                  as.double(dm), as.double(betm0), as.double(betQ0),
                  as.double(ssqdf), as.double(ssqsc), max(tsqdf, 0),
                  as.double(tsq), as.double(kappa_pnts), as.integer(icf),
                  as.integer(n), as.integer(p), as.integer(nruns),
                  as.integer(ifam), as.integer(imeth), as.integer(itr),
                  PACKAGE = "geoBayes")
  refbf <- RUN$logbf[reference]
  logbf <- RUN$logbf - refbf
  if (Ntot2 > 0) {
    weights <- RUN$weights
    lglk2 <- RUN$lglk2
    zcv <- RUN$zcv
  } else {
    weights <- lglk2 <- zcv <- NULL
  }
  out <- list(logbf = logbf, logLik1 = RUN$lglk1, logLik2 = lglk2,
              isweights = weights, controlvar = zcv, sample2 = z2,
              N1 = Nout1, N2 = Nout2,
              betm0 = betm0,
              betQ0 = betQ0, ssqdf = ssqdf, ssqsc = ssqsc, tsqdf = tsqdf,
              tsqsc = tsqsc, dispersion = dispersion, response = y, weights = l,
              modelmatrix = F, offset = offset, 
              locations = loc, distmat = dm, family = family,
              corrfcn = corrfcn, transf = transf,
              real_transf = real_transf, itr = itr,
              pnts = list(nu = nu_pnts, phi = phi_pnts, omg = omg_pnts,
                kappa = kappa_pnts))
  out
}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary:
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine calcb_no_st (bfact, phi, nu, omg, kappa, icf, n_cov, n_nu, &
   Ntot, zsample, weights, n, p, betm0, betQ0, ssqdf, &
   ssqsc, tsqdf, tsq, y, l, F, offset, dm, ifam, itr)

  use modelfcns, condyz_sp => condyz
  use interfaces
  use flogsumexp
  use covfun
  use condyz, only: condyz_gt
  use betaprior
  implicit none
  integer, intent(in) :: n, p, Ntot, n_cov, n_nu, ifam, icf, itr(n)
  double precision, intent(in) :: phi(n_cov), omg(n_cov), kappa(n_cov), &
     nu(n_nu), zsample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(out) :: bfact(n_nu, n_cov)
  logical lmxi
  double precision ssqdfsc, tsqdfsc, respdfh, modeldfh
  integer i, j, k
  double precision logfy(n_nu, Ntot), lfz
  double precision T(n, n), TiF(n, p), FTF(p, p), Ups(n, n), &
     ldh_Ups, llikw(n_nu, Ntot), xi(n)

  call create_model(ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  ! Calculate log f(y|z,nu)
  select case (ifam)
  case (0)
    do j = 1, Ntot
      do k = 1, n_nu
        logfy(k, j) = condyz_gt(n, y, l, zsample(:, j), nu(k), tsqdfsc, &
           respdfh)
      end do
      call rchkusr
    end do
  case default
    do j = 1, Ntot
      do k = 1, n_nu
        logfy(k, j) = condyz_sp(n, y, l, zsample(:, j), nu(k), tsq)
      end do
      call rchkusr
    end do
  end select

  do k = 1, n_cov
    call calc_cov (phi(k),omg(k),dm,F,betQ0,&
       kappa(k),n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, Ntot
      ! Calculate unnormalised log-likelihood at sampled points
      lfz = logpdfz(n, zsample(:, j), Ups, ldh_Ups, xi, lmxi, &
         ssqdfsc, modeldfh)
      do i = 1, n_nu
        llikw(i, j) = logfy(i, j) + lfz - weights(j)
      end do
    end do
    bfact(:, k) = logrsumexp(llikw, n_nu, Ntot)
  end do
end subroutine calcb_no_st


subroutine calcb_wo_st (bfact, phi, nu, omg, kappa, icf, n_cov, n_nu, &
   Ntot, wsample, weights, n, p, betm0, betQ0, ssqdf, &
   ssqsc, tsqdf, tsq, y, l, F, offset, dm, ifam, itr)

  use modelfcns
  use flogsumexp
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, Ntot, n_cov, n_nu, ifam, icf, itr(n)
  double precision, intent(in) :: phi(n_cov), omg(n_cov), kappa(n_cov), &
     nu(n_nu), wsample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(out) :: bfact(n_nu,n_cov)
  logical lmxi
  double precision ssqdfsc, tsqdfsc, respdfh, modeldfh, zsample(n)
  integer i, j, k, m
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, llikw(n_nu,Ntot), xi(n), lfw

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  select case (ifam)
  case (0)
    call rexit ("This method has not been implemented.")
  case default
    do i = 1, n_cov
      call rchkusr
      call calc_cov (phi(i),omg(i),dm,F,betQ0,&
         kappa(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        do k = 1, n_nu
          zsample = transfw(wsample(:,j), nu(k))
          lfw = jointyz(n, zsample, y, l, Ups, ldh_Ups, &
             nu(k), xi, lmxi, ssqdfsc, tsq, modeldfh)
          do m = 1, n
            lfw = lfw - loginvtrwdz(zsample(m),nu(k))
          end do
          llikw(k,j) = lfw - weights(j)
        end do
      end do
      bfact(:, i) = logrsumexp(llikw, n_nu, Ntot)
    end do
  end select
end subroutine calcb_wo_st


subroutine calcb_mu_st (bfact, phi, nu, omg, kappa, icf, n_cov, n_nu, &
   Ntot, musample, weights, n, p, betm0, betQ0, ssqdf, &
   ssqsc, tsqdf, tsq, y, l, F, offset, dm, ifam, itr)

  use modelfcns
  use interfaces
  use flogsumexp
  use covfun
  use pdfmu, only: logpdfmu_ga
  use betaprior
  implicit none
  integer, intent(in) :: n, p, Ntot, n_cov, n_nu, ifam, icf, itr(n)
  double precision, intent(in) :: phi(n_cov), omg(n_cov), kappa(n_cov), &
     nu(n_nu), musample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(out) :: bfact(n_nu, n_cov)
  logical lmxi
  double precision ssqdfsc, tsqdfsc, respdfh, modeldfh
  integer i, j, k
  double precision lfmu
  double precision T(n, n), TiF(n, p), FTF(p, p), Ups(n, n), &
     ldh_Ups, llikw(n_nu, Ntot), xi(n)

  call create_model(ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  select case (ifam)
  case (0)
    do i = 1, n_cov
      call rchkusr
      call calc_cov (phi(i),omg(i),dm,F,betQ0,&
         kappa(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        do k = 1, n_nu
          lfmu = logpdfmu_ga(n, musample(:, j), Ups, ldh_Ups, &
             nu(k), xi, lmxi, ssqdfsc, modeldfh)
          llikw(k,j) = lfmu - weights(j)
        end do
      end do
      bfact(:, i) = logrsumexp(llikw, n_nu, Ntot)
    end do
  case default
    do i = 1, n_cov
      call rchkusr
      call calc_cov (phi(i),omg(i),dm,F,betQ0,&
         kappa(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      call fill_symmetric_matrix(Ups,n)
      do j = 1, Ntot
        do k = 1, n_nu
          lfmu = logpdfmu(n, musample(:,j), Ups, ldh_Ups, &
             nu(k), xi, lmxi, ssqdfsc, modeldfh)
          llikw(k,j) = lfmu - weights(j)
        end do
      end do
      bfact(:,i) = logrsumexp(llikw, n_nu, Ntot)
    end do
  end select
end subroutine calcb_mu_st


subroutine calcb_tr_st (bfact, philist, nulist, omglist, kappalist, &
   icf, n_cov, n_nu, Ntot, sample, weights, n, p, betm0, betQ0, ssqdf, &
   ssqsc, tsqdf, tsq, y, l, F, offset, dm, ifam, itr)

  use modelfcns, logpdfzf => logpdfz, condymu_mf => condymu
  use flogsumexp
  use covfun
  use betaprior
  use condymu, only: condymu_gt
  implicit none
  integer, intent(in) :: n, p, Ntot, n_cov, n_nu, ifam, icf, itr(n)
  double precision, intent(in) :: philist(n_cov), omglist(n_cov), &
     kappalist(n_cov), nulist(n_nu), sample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(out) :: bfact(n_nu,n_cov)
  logical lmxi
  double precision ssqdfsc, respdfh, modeldfh, tsqval
  integer i, j, k
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, lglk(Ntot), xi(n)
  double precision nu, phi, omg, kappa
  double precision zsam(n), msam(n), jsam(n), sam(n)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  do i = 1, n_cov
    phi = philist(i)
    omg = omglist(i)
    kappa = kappalist(i)
    call calc_cov (phi,omg,dm,F,betQ0,&
       kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, n_nu
      nu = nulist(j)
      do k = 1, Ntot
        call rchkusr
        sam = sample(:,k)
        where (itr == 0)
          zsam = sam
          msam = invlink(zsam,nu)
          jsam = 0d0
        elsewhere (itr == 1)
          msam = sam
          zsam = flink(msam,nu)
          jsam = loginvlinkdz(zsam,nu)
        elsewhere (itr == 2)
          zsam = transfw(sam,nu)
          msam = invlink(zsam,nu)
          jsam = loginvtrwdz(zsam,nu)
        end where
        lglk(k) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
           + condymuf(ifam,n,y,l,msam,tsqval,respdfh) - sum(jsam) - weights(k)
      end do
      bfact(j,i) = logsumexpv(lglk,Ntot)
    end do
  end do

contains

  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n, ifam
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf
end subroutine calcb_tr_st



!!!!!!!!!!!!!!!!!!!!!!!!! Control variates method !!!!!!!!!!!!!!!!!!!!!!!!!

subroutine calcb_no_cv (bfact, phi, nu, omg, kappa, icf, n_cov, n_nu, &
   Ntot, zsample, weights, QRin, n, p, betm0, betQ0, ssqdf, &
   ssqsc, tsqdf, tsq, y, l, F, offset, dm, ifam, itr)

  use modelfcns, condyz_sp => condyz
  use interfaces
  use flogsumexp
  use covfun
  use condyz, only: condyz_gt
  use betaprior
  implicit none
  integer, intent(in) :: n, p, Ntot, n_cov, n_nu, ifam, icf, itr(n)
  double precision, intent(in) :: phi(n_cov), omg(n_cov), kappa(n_cov), &
     nu(n_nu), zsample(n,Ntot), weights(Ntot), QRin(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(out) :: bfact(n_nu, n_cov)
  logical lmxi
  double precision ssqdfsc, tsqdfsc, respdfh, modeldfh
  integer i, j, k
  double precision logfy(n_nu, Ntot), lfz, dNtot
  double precision T(n, n), TiF(n, p), FTF(p, p), Ups(n, n), &
     ldh_Ups, ycv(n_nu, Ntot), llikw, xi(n), betareg(n_nu)

  call create_model(ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)
  dNtot = log(dble(Ntot))

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr
  ! Calculate log f(y|z,nu)
  select case (ifam)
  case (0)
    do j = 1, Ntot
      do k = 1, n_nu
        logfy(k, j) = condyz_gt(n, y, l, zsample(:, j), nu(k), tsqdfsc, &
           respdfh)
      end do
      call rchkusr
    end do
  case default
    do j = 1, Ntot
      do k = 1, n_nu
        logfy(k, j) = condyz_sp(n, y, l, zsample(:, j), nu(k), tsq)
      end do
      call rchkusr
    end do
  end select

  do k = 1, n_cov
    call rchkusr
    call calc_cov (phi(k),omg(k),dm,F,betQ0,&
       kappa(k),n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, Ntot
      ! Calculate unnormalised log-likelihood at sampled points
      lfz = logpdfz(n, zsample(:,j), Ups, ldh_Ups, xi, lmxi, &
         ssqdfsc, modeldfh)
      do i = 1, n_nu
        llikw = logfy(i,j) + lfz - weights(j)
        ycv(i,j) = exp(llikw + dNtot)
      end do
    end do
    betareg = matmul(ycv,QRin)
    where (betareg .gt. 0d0)
      bfact(:,k) = log(betareg)
    elsewhere
      bfact(:,k) = -huge(1d0)
    end where
  end do
end subroutine calcb_no_cv




subroutine calcb_wo_cv (bfact, phi, nu, omg, kappa, icf, n_cov, n_nu, &
   Ntot, wsample, weights, QRin, n, p, betm0, betQ0, ssqdf, &
   ssqsc, tsqdf, tsq, y, l, F, offset, dm, ifam, itr)

  use modelfcns
  use flogsumexp
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, Ntot, n_cov, n_nu, ifam, icf, itr(n)
  double precision, intent(in) :: phi(n_cov), omg(n_cov), kappa(n_cov), &
     nu(n_nu), wsample(n,Ntot), weights(Ntot), QRin(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(out) :: bfact(n_nu,n_cov)
  logical lmxi
  double precision ssqdfsc, tsqdfsc, respdfh, modeldfh
  integer i, j, k, m
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, llikw, ycv(n_nu,Ntot), xi(n), lfw, zsample(n), betareg(n_nu)
  double precision dNtot

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)
  dNtot = log(dble(Ntot))

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  select case (ifam)
  case (0)
    call rexit ("This method has not been implemented.")
  case default
    do i = 1, n_cov
      call rchkusr
      call calc_cov (phi(i),omg(i),dm,F,betQ0,&
         kappa(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        do k = 1, n_nu
          zsample = transfw(wsample(:,j),nu(k))
          lfw = jointyz(n, zsample, y, l, Ups, ldh_Ups, &
             nu(k), xi, lmxi, ssqdfsc, tsq, modeldfh)
          do m = 1, n
            lfw = lfw - loginvtrwdz(zsample(m),nu(k))
          end do
          llikw = lfw - weights(j)
          ycv(k,j) = exp(llikw + dNtot)
        end do
      end do
      betareg = matmul(ycv,QRin)
      where (betareg .gt. 0d0)
        bfact(:,i) = log(betareg)
      elsewhere
        bfact(:,i) = -huge(1d0)
      end where
    end do
  end select
end subroutine calcb_wo_cv



subroutine calcb_mu_cv (bfact, phi, nu, omg, kappa, icf, n_cov, n_nu, &
   Ntot, musample, weights, QRin, n, p, betm0, betQ0, ssqdf, &
   ssqsc, tsqdf, tsq, y, l, F, offset, dm, ifam, itr)

  use modelfcns
  use interfaces
  use flogsumexp
  use covfun
  use pdfmu, only: logpdfmu_ga
  use betaprior
  implicit none
  integer, intent(in) :: n, p, Ntot, n_cov, n_nu, ifam, icf, itr(n)
  double precision, intent(in) :: phi(n_cov), omg(n_cov), kappa(n_cov), &
     nu(n_nu), musample(n,Ntot), weights(Ntot), QRin(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(out) :: bfact(n_nu, n_cov)
  logical lmxi
  double precision ssqdfsc, tsqdfsc, respdfh, modeldfh
  integer i, j, k
  double precision lfmu, dNtot
  double precision T(n, n), TiF(n, p), FTF(p, p), Ups(n, n), &
     ldh_Ups, ycv(n_nu, Ntot), llikw, xi(n), betareg(n_nu)

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  tsqdfsc = tsqdf*tsq
  respdfh = .5d0*(n + tsqdf)
  dNtot = log(dble(Ntot))

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  select case (ifam)
  case (0)
    do i = 1, n_cov
      call rchkusr
      call calc_cov (phi(i),omg(i),dm,F,betQ0,&
         kappa(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        do k = 1, n_nu
          lfmu = logpdfmu_ga(n, musample(:,j), Ups, ldh_Ups, &
             nu(k), xi, lmxi, ssqdfsc, modeldfh)
          llikw = lfmu - weights(j)
          ycv(k, j) = exp(llikw + dNtot)
        end do
      end do
      betareg = matmul(ycv,QRin)
      where (betareg .gt. 0d0)
        bfact(:,i) = log(betareg)
      elsewhere
        bfact(:,i) = -huge(1d0)
      end where
    end do
  case default
    do i = 1, n_cov
      call rchkusr
      call calc_cov (phi(i),omg(i),dm,F,betQ0,&
         kappa(i),n,p,T,TiF,FTF,Ups,ldh_Ups)
      do j = 1, Ntot
        do k = 1, n_nu
          lfmu = logpdfmu(n, musample(:,j), Ups, ldh_Ups, &
             nu(k), xi, lmxi, ssqdfsc, modeldfh)
          llikw = lfmu - weights(j)
          ycv(k, j) = exp(llikw + dNtot)
        end do
      end do
      betareg = matmul(ycv,QRin)
      where (betareg .gt. 0d0)
        bfact(:,i) = log(betareg)
      elsewhere
        bfact(:,i) = -huge(1d0)
      end where
    end do
  end select
end subroutine calcb_mu_cv



subroutine calcb_tr_cv (bfact, philist, nulist, omglist, kappalist, &
   icf, n_cov, n_nu, Ntot, sample, weights, QRin, &
   n, p, betm0, betQ0, ssqdf, &
   ssqsc, tsqdf, tsq, y, l, F, offset, dm, ifam, itr)

  use modelfcns, logpdfzf => logpdfz, condymu_mf => condymu
  use flogsumexp
  use covfun
  use betaprior
  use condymu, only: condymu_gt
  implicit none
  integer, intent(in) :: n, p, Ntot, n_cov, n_nu, ifam, icf, itr(n)
  double precision, intent(in) :: philist(n_cov), omglist(n_cov), &
     kappalist(n_cov), nulist(n_nu), sample(n,Ntot), weights(Ntot), &
     betm0(p), betQ0(p,p), ssqdf, ssqsc, tsqdf, tsq, y(n), l(n), &
     F(n,p), dm(n,n), offset(n)
  double precision, intent(in) :: QRin(Ntot)
  double precision, intent(out) :: bfact(n_nu,n_cov)
  logical lmxi
  double precision ssqdfsc, respdfh, modeldfh, tsqval
  integer i, j, k
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, lglk(Ntot), xi(n)
  double precision nu, phi, omg, kappa
  double precision zsam(n), msam(n), jsam(n), sam(n)
  double precision dNtot, ycv(Ntot), betareg

  call create_model (ifam)
  call create_spcor(icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select
  dNtot = log(dble(Ntot))

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call rchkusr

  do i = 1, n_cov
    phi = philist(i)
    omg = omglist(i)
    kappa = kappalist(i)
    call calc_cov (phi,omg,dm,F,betQ0,&
       kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
    do j = 1, n_nu
      nu = nulist(j)
      do k = 1, Ntot
        call rchkusr
        sam = sample(:,k)
        where (itr == 0)
          zsam = sam
          msam = invlink(zsam,nu)
          jsam = 0d0
        elsewhere (itr == 1)
          msam = sam
          zsam = flink(msam,nu)
          jsam = loginvlinkdz(zsam,nu)
        elsewhere (itr == 2)
          zsam = transfw(sam,nu)
          msam = invlink(zsam,nu)
          jsam = loginvtrwdz(zsam,nu)
        end where
        lglk(k) = logpdfzf(n,zsam,Ups,ldh_Ups,xi,lmxi,ssqdfsc,modeldfh) &
           + condymuf(ifam,n,y,l,msam,tsqval,respdfh) - sum(jsam) - weights(k)
        ycv(k) = exp(lglk(k) + dNtot)
      end do
      betareg = dot_product(ycv,QRin)
      if (betareg .gt. 0d0) then
        bfact(j,i) = log(betareg)
      else
        bfact(j,i) = -huge(1d0)
      end if
    end do
  end do

contains

  pure function condymu_sp (n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymu_sp
    condymu_sp = condymu_mf(n,y1,y2,mu,tsqval)
  end function condymu_sp

  pure function condymuf (ifam, n, y1, y2, mu, tsqval, respdfh)
    implicit none
    integer, intent(in) :: n, ifam
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsqval, &
       respdfh
    double precision condymuf
    select case (ifam)
    case (0)
      condymuf = condymu_gt(n, y1, y2, mu, tsqval, respdfh)
    case default
      condymuf = condymu_sp(n, y1, y2, mu, tsqval, respdfh)
    end select
  end function condymuf
end subroutine calcb_tr_cv

module bmargin
contains
  subroutine revlogistic (eta,llik,kg,Ntot,Nout)

    ! use lbfgsqck
    use lbfgsbmod
    use flogsumexp, only: logrsumexp

    implicit none

    integer, intent(in) :: kg, Ntot, Nout(kg)
    double precision, intent(in) :: llik(Ntot,kg)
    double precision, intent(inout) :: eta(kg)
    double precision lgdenom(Ntot), logp(Ntot,kg), lliketa(Ntot,kg)
    integer, parameter :: maxit=500
    type(lbfgsb_store) opstore
    ! double precision, parameter :: epstol = 5d-5
    double precision, parameter :: pgtol = 0d-5, factr = 1d7
    double precision neglogfun, neglogfun_grad(kg)
    integer ia, ie, i, j, k
    integer iflag
    logical lidx(Ntot,kg)
    double precision etalo(kg), etaup(kg)
    integer etabd(kg)


    lidx = .false.
    ia = 1
    ie = 0
    do i = 1, kg
      ie = ie + Nout(i)
      lidx(ia:ie,i) = .true.
      ia = ia + Nout(i)
    end do

    etalo = eta
    etaup = eta
    etabd = 0
    ! etabd(1) = 2
    iflag = 0
    do i = 1, maxit
      do k = 1, kg
        do j = 1, Ntot
          lliketa(j,k) = llik(j,k) + eta(k)
        end do
      end do
      lgdenom = logrsumexp(lliketa,Ntot,kg)
      do k = 1, kg
        do j = 1, Ntot
          logp(j,k) = lliketa(j,k) - lgdenom(j)
        end do
      end do
      neglogfun = -sum(logp,lidx)
      neglogfun_grad = -dble(Nout) + sum(exp(logp),1)
      !       call lbfgs (kg-1,eta(2:kg),neglogfun,neglogfun_grad(2:kg),&
      !          iprint,epstol,iflag)
      call lbfgsb (kg,eta,etalo,etaup,etabd,neglogfun,neglogfun_grad,&
         factr,pgtol,iflag,opstore)
      if (iflag == 0) then
        !       print '(" Convergence in ",I0," iterations")', i
        exit
      else if (iflag < 0) then
        call rwarn ('The reverse logistic agorithm didn''t converge')
        exit
      end if
    end do
    if (iflag > 0) then
      call rwarn ('The reverse logistic algorithm needs more iterations')
    end if
    eta = eta - eta(1) + log(dble(Nout)) ! -log(BF_i) + log(N_i)
  end subroutine revlogistic


  subroutine mengwong (eta,llik,kg,Ntot,Nout)

    use flogsumexp, only: logrsumexp

    implicit none

    integer, intent(in) :: kg, Ntot, Nout(kg)
    double precision, intent(in) :: llik(Ntot,kg)
    double precision, intent(inout) :: eta(kg)
    integer, parameter :: maxit=500
    double precision, parameter :: pgtol = 5d-5
    integer ia, ie, i, j, it, ipiv(1:kg-1)
    double precision llik_mix_all(Ntot), a(kg), Bmat(2:kg,2:kg), bvec(2:kg), &
       r(kg), likw(Ntot,kg), logar(kg)
    logical convergence
    double precision bnorm, work(4*kg-4), rcond
    integer iwork(1:kg-1)
    double precision, external :: dlange

    a = dble(Nout)
    r = exp(eta)/a
    a = a/dble(Ntot)
    logar = log(a*r)

    convergence = .false.
    iterations: do it = 1, maxit
      call rchkusr ! Check if user has requested interrupt
      ia = 1
      ie = 0
      do j = 1, kg
        do i = 1, Ntot
          likw(i,j) = llik(i,j) + logar(j)
        end do
      end do
      llik_mix_all = -logrsumexp(likw,Ntot,kg)
      likw = spread(llik_mix_all,2,kg)
      likw = llik + likw - log(dble(Ntot))
      likw = exp(likw)
      ia = 1 + Nout(1)
      ie = Nout(1)
      do i = 2, kg
        ie = ie + Nout(i)
        forall (j=2:kg, j .ne. i)
          Bmat(i,j) = -a(j)*sum(likw(ia:ie,j))
        end forall
        bvec(i) = a(1)*sum(likw(ia:ie,1))
        ia = ia + Nout(i)
      end do
      do i = 2, kg
        ia = 1
        ie = 0
        Bmat(i,i) = 0d0
        do j = 1, kg
          ie = ie + Nout(j)
          if (j .ne. i) then
            Bmat(i,i) = Bmat(i,i) + a(i)*sum(likw(ia:ie,i))
          end if
          ia = ia + Nout(j)
        end do
      end do
      j = kg - 1
      bnorm = dlange('1',j,j,Bmat,j,work) ! Matrix norm used to check singularity
      call dgesv (j,1,Bmat,j,ipiv,bvec,j,i)
      if (i .ne. 0) then
        call rwarn ('mengwong - Cannot find solution to linear system. &
           &Previous solution restored.')
        exit iterations
      end if
      if (any(bvec .le. 0d0)) then
        call rwarn ('mengwong - Non-positive solution to the linear system. &
           &Previous solution restored.')
        exit iterations
      end if
      call dgecon ('1',j-1,Bmat,j-1,bnorm,rcond,work,iwork,i)
      if (rcond .lt. 1d-17) then
        call rwarn ('mengwong - System is computationally singular. &
           &Previous solution restored.')
        exit iterations
      end if
      convergence = sum(abs(bvec - r(2:kg))) < pgtol
      r(2:kg) = bvec
      if (convergence) exit iterations
    end do iterations
    if (.not. convergence) then
      call rwarn ('mengwong - Algorithm did''t converge.')
    end if
    eta = log(dble(Nout)) + log(r)
  end subroutine mengwong
end module bmargin

module modelfcns_pdfy
!! logpdfy is the log-pdf of y for given parameter value
!! logpdfydlnk is the derivative of the above w.r.t the parameter
!! logdffy is the difference between two such logpdfs at two different
!! parameter values
  implicit none 
  double precision, parameter :: bigpos = huge(1d0), bigneg = -bigpos, &
     smallpos = epsilon(1d0), smallneg = -smallpos
  private bigpos, bigneg, smallpos, smallneg

contains

!!!!!!!!!!!!!!!!!!!!!!!!!! Transformed Gaussian !!!!!!!!!!!!!!!!!!!!!!!!!!!
  pure elemental double precision function logpdfy_gt (y1, y2, par)
    ! Transformed Gaussian
    ! y1 :: The average across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    double precision d
    d = y1 - par
    logpdfy_gt = y2*d*d
  end function logpdfy_gt

  pure elemental double precision function logdffy_gt (y1, y2, p1, p2)
    ! Transformed Gaussian
    ! y1 :: The average across all realisations
    ! y2 :: The number of replications
    ! p1, p2 :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    double precision d1, d2
    d1 = y1 - p1
    d2 = y1 - p2
    logdffy_gt = y2*(d1*d1 - d2*d2)
  end function logdffy_gt

  pure elemental double precision function logpdfydlnk_gt (y1, y2, par)
    ! Transformed Gaussian
    ! y1 :: The average across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_gt = 2d0*y2*(par-y1)
  end function logpdfydlnk_gt

  pure elemental double precision function logpdfyhlnk_gt (y1, y2, par)
    ! Transformed Gaussian
    ! y1 :: The average across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfyhlnk_gt = 2d0*y2
  end function logpdfyhlnk_gt

  pure elemental double precision function logpdfy3lnk_gt (y1, y2, par)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy3lnk_gt = 0
  end function logpdfy3lnk_gt

  pure elemental double precision function mustart_gt (y1,y2)
    implicit none
    double precision, intent(in) :: y1, y2
    mustart_gt = y1
  end function mustart_gt

  pure elemental double precision function fcncum_gt (mu) result (w)
    ! Cumulant fcn as a fcn of mu
    implicit none
    double precision, intent(in) :: mu
    w = .5d0*mu*mu
  end function fcncum_gt

  pure elemental double precision function fcncumd2_gt (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 2nd derivative
    implicit none
    double precision, intent(in) :: mu
    w = 1d0
  end function fcncumd2_gt

  pure elemental double precision function fcncumd3_gt (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 3rd derivative
    implicit none
    double precision, intent(in) :: mu
    w = 0d0
  end function fcncumd3_gt


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Gaussian !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  pure elemental double precision function logpdfy_ga (y1, y2, par)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy_ga = y1*par - .5d0*y2*par*par
  end function logpdfy_ga

  pure elemental double precision function logdffy_ga (y1, y2, p1, p2)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! p1, p2 :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    logdffy_ga = y1*(p1 - p2) - .5d0*y2*(p1*p1 - p2*p2)
  end function logdffy_ga

  pure elemental double precision function logpdfydlnk_ga (y1, y2, par)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_ga = y1 - y2*par
  end function logpdfydlnk_ga

  pure elemental double precision function logpdfyhlnk_ga (y1, y2, par)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfyhlnk_ga = -y2
  end function logpdfyhlnk_ga

  pure elemental double precision function logpdfy3lnk_ga (y1, y2, par)
    ! Gaussian
    ! y1 :: The total across all realisations
    ! y2 :: The number of replications
    ! par :: The mean of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy3lnk_ga = 0
  end function logpdfy3lnk_ga

  pure elemental double precision function mustart_ga (y1,y2)
    implicit none
    double precision, intent(in) :: y1, y2
    mustart_ga = y1/y2
  end function mustart_ga

  pure elemental double precision function fcncum_ga (mu) result (w)
    ! Cumulant fcn as a fcn of mu
    implicit none
    double precision, intent(in) :: mu
    w = .5d0*mu*mu
  end function fcncum_ga

  pure elemental double precision function fcncumd2_ga (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 2nd derivative
    implicit none
    double precision, intent(in) :: mu
    w = 1d0
  end function fcncumd2_ga

  pure elemental double precision function fcncumd3_ga (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 3rd derivative
    implicit none
    double precision, intent(in) :: mu
    w = 0d0
  end function fcncumd3_ga


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Binomial !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  pure elemental double precision function logpdfy_bi (y1, y2, par)
    ! Binomial
    ! To get the asymmetric version, switch y1 and y2
    ! y1 :: The number of successes
    ! y2 :: The number of faillures
    ! par :: The logartithm of the probability of success
    use interfaces, only: flog1mexp
    implicit none
    double precision, intent(in) :: y1, y2, par
    if (y1 .eq. 0d0) then
      logpdfy_bi = y2*flog1mexp(par)
    else if (y2 .eq. 0d0) then
      logpdfy_bi = y1*par
    else 
      logpdfy_bi = y1*par + y2*flog1mexp(par)
    end if
  end function logpdfy_bi

  pure elemental double precision function logdffy_bi (y1, y2, p1, p2)
    ! Binomial
    ! To get the asymmetric version, switch y1 and y2
    ! y1 :: The number of successes
    ! y2 :: The number of faillures
    ! p1, p2 :: The logartithm of the probability of success
    use interfaces, only: flog1mexp
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    if (y1 .eq. 0d0) then
      logdffy_bi = y2*(flog1mexp(p1) - flog1mexp(p2))
    else if (y2 .eq. 0d0) then
      logdffy_bi = y1*(p1 - p2)
    else
      logdffy_bi = y1*(p1 - p2) + y2*(flog1mexp(p1) - flog1mexp(p2))
    end if
  end function logdffy_bi

  pure elemental double precision function logpdfydlnk_bi (y1, y2, par)
    ! Binomial
    ! To get the asymmetric version, switch y1 and y2
    ! y1 :: The number of successes
    ! y2 :: The number of faillures
    ! par :: The logartithm of the probability of success
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: y1, y2, par
    if (y2 .eq. 0d0) then
      logpdfydlnk_bi = y1
    else
      logpdfydlnk_bi = y1 - y2/fexpm1(-par)
    end if
  end function logpdfydlnk_bi

  pure elemental double precision function logpdfyhlnk_bi (y1, y2, par)
    ! Binomial
    ! To get the asymmetric version, switch y1 and y2
    ! y1 :: The number of successes
    ! y2 :: The number of faillures
    ! par :: The logartithm of the probability of success
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: y1, y2, par
    if (y2 .eq. 0d0) then
      logpdfyhlnk_bi = 0d0
    else
      logpdfyhlnk_bi = 1d0/fexpm1(-par)
      logpdfyhlnk_bi = -y2*logpdfyhlnk_bi * (1d0+logpdfyhlnk_bi)
    end if
  end function logpdfyhlnk_bi

  pure elemental double precision function logpdfy3lnk_bi (y1, y2, par)
    ! Binomial
    ! To get the asymmetric version, switch y1 and y2
    ! y1 :: The number of successes
    ! y2 :: The number of faillures
    ! par :: The logartithm of the probability of success
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: y1, y2, par
    if (y2 .eq. 0d0) then
      logpdfy3lnk_bi = 0d0
    else
      logpdfy3lnk_bi = 1d0/fexpm1(-par)
      logpdfy3lnk_bi = -y2*logpdfy3lnk_bi * (1d0+logpdfy3lnk_bi) &
         * (1d0+logpdfy3lnk_bi+logpdfy3lnk_bi)
    end if
  end function logpdfy3lnk_bi

  pure elemental double precision function mustart_bi (y1,y2)
    implicit none
    double precision, intent(in) :: y1, y2
    double precision t1, t2
    t1 = y1 + .5d0
    t2 = y1 + y2 + 1d0
    mustart_bi = log(t1) - log(t2)
  end function mustart_bi

  pure elemental double precision function fcncum_bi (mu) result (w)
    ! Cumulant fcn as a fcn of mu
    implicit none
    double precision, intent(in) :: mu
    w = -log(1-mu)
  end function fcncum_bi

  pure elemental double precision function fcncumd2_bi (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 2nd derivative
    implicit none
    double precision, intent(in) :: mu
    w = mu*(1d0-mu)
  end function fcncumd2_bi

  pure elemental double precision function fcncumd3_bi (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 3rd derivative
    implicit none
    double precision, intent(in) :: mu
    w = mu*(1d0-mu)*(1d0-mu-mu)
  end function fcncumd3_bi


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Poisson !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  pure elemental double precision function logpdfy_po (y1, y2, par)
    ! Poisson
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    if (y1 .eq. 0d0) then
      logpdfy_po = -y2*exp(par)
    else
      logpdfy_po = y1*par - y2*exp(par)
    end if
  end function logpdfy_po

  pure elemental double precision function logdffy_po (y1, y2, p1, p2)
    ! Poisson
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! p1, p2 :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    if (y1 .eq. 0d0) then
      logdffy_po = -y2*(exp(p1) - exp(p2))
    else
      logdffy_po = y1*(p1 - p2) - y2*(exp(p1) - exp(p2))
    end if
  end function logdffy_po

  pure elemental double precision function logpdfydlnk_po (y1, y2, par)
    ! Poisson
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_po = y1 - y2*exp(par)
  end function logpdfydlnk_po

  pure elemental double precision function logpdfyhlnk_po (y1, y2, par)
    ! Poisson
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfyhlnk_po = -y2*exp(par)
  end function logpdfyhlnk_po

  pure elemental double precision function logpdfy3lnk_po (y1, y2, par)
    ! Poisson
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy3lnk_po = -y2*exp(par)
  end function logpdfy3lnk_po

  pure elemental double precision function mustart_po (y1,y2)
    implicit none
    double precision, intent(in) :: y1, y2
    double precision t1, t2
    t1 = y1 + .5d0
    t2 = y2 + 1d0
    mustart_po = log(t1) - log(t2)
  end function mustart_po

  pure elemental double precision function fcncum_po (mu) result (w)
    ! Cumulant fcn as a fcn of mu
    implicit none
    double precision, intent(in) :: mu
    w = mu
  end function fcncum_po

  pure elemental double precision function fcncumd2_po (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 2nd derivative
    implicit none
    double precision, intent(in) :: mu
    w = mu
  end function fcncumd2_po

  pure elemental double precision function fcncumd3_po (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 3rd derivative
    implicit none
    double precision, intent(in) :: mu
    w = mu
  end function fcncumd3_po


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Gamma !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  pure elemental double precision function logpdfy_gm (y1, y2, par)
    ! Gamma
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy_gm = -y1*exp(-par) - y2*par
  end function logpdfy_gm

  pure elemental double precision function logdffy_gm (y1, y2, p1, p2)
    ! Gamma
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! p1, p2 :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    logdffy_gm = -y1*(exp(-p1) - exp(-p2)) - y2*(p1 - p2)
  end function logdffy_gm

  pure elemental double precision function logpdfydlnk_gm (y1, y2, par)
    ! Gamma
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfydlnk_gm = y1*exp(-par) - y2
  end function logpdfydlnk_gm

  pure elemental double precision function logpdfyhlnk_gm (y1, y2, par)
    ! Gamma
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfyhlnk_gm = -y1*exp(-par)
  end function logpdfyhlnk_gm

  pure elemental double precision function logpdfy3lnk_gm (y1, y2, par)
    ! Gamma
    ! y1 :: The total number of observations
    ! y2 :: The number of realisations
    ! par :: The log(mean) of a single realisation
    implicit none
    double precision, intent(in) :: y1, y2, par
    logpdfy3lnk_gm = y1*exp(-par)
  end function logpdfy3lnk_gm

  pure elemental double precision function mustart_gm (y1,y2)
    implicit none
    double precision, intent(in) :: y1, y2
    mustart_gm = log(y1) - log(y2)
  end function mustart_gm

  pure elemental double precision function fcncum_gm (mu) result (w)
    ! Cumulant fcn as a fcn of mu
    implicit none
    double precision, intent(in) :: mu
    w = log(mu)
  end function fcncum_gm

  pure elemental double precision function fcncumd2_gm (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 2nd derivative
    implicit none
    double precision, intent(in) :: mu
    w = mu*mu
  end function fcncumd2_gm

  pure elemental double precision function fcncumd3_gm (mu) result (w)
    ! Cumulant fcn as a fcn of mu, 3rd derivative
    implicit none
    double precision, intent(in) :: mu
    w = mu*mu*mu
    w = w + w
  end function fcncumd3_gm

end module modelfcns_pdfy

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary:
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module modelfcns_link
  implicit none
  double precision, parameter :: bigpos = huge(1d0), bigneg = -bigpos, &
     smallpos = epsilon(1d0), smallneg = -smallpos
  private bigpos, bigneg, smallpos, smallneg
contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!! binomial robit !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link function
  pure elemental function flink_robit (w,d) result (z)
    use interfaces, only: quantt
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    z = quantt(w, d)
  end function flink_robit

!! Inverse link function (log scale)
  pure elemental function invlink_robit (z,d) result (w)
    ! Binomial symmetric (robit) link fcn
    use interfaces, only: logprobt
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = logprobt(z, d)
  end function invlink_robit

  pure elemental function invlinkdz_robit (z,d) result (y)
    use interfaces, only: logprobt, logpdft
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision f, ff
    ff = logprobt(z, d)
    f = logpdft(z, d)
    y = exp(f - ff)
  end function invlinkdz_robit

  pure elemental function loginvlinkdz_robit (z,d) result (y)
    use interfaces, only: logprobt, logpdft
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision f, ff
    ff = logprobt(z, d)
    f = logpdft(z, d)
    y = f - ff
  end function loginvlinkdz_robit

  pure elemental function invlinkhz_robit (z,d) result (y)
    ! If g_d(z) is the pdf of the t_d distribution, then we need
    ! d2w = d(log g_d(z)) * d1w - (d1w)^2
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision dwdz, dlogtdz
    dwdz = invlinkdz_robit(z,d)
    dlogtdz = -(d+1d0)*z/(d+z*z)
    y = dlogtdz*dwdz - dwdz*dwdz
  end function invlinkhz_robit

  pure elemental function invlink3z_robit (z,d) result (y)
    ! If g_d(z) is the pdf of the t_d distribution, then we need
    ! d3w = d2(log g_d(z)) * d1w - (d1w)*(d2w) + (d2w)^2/(d1w)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision dw, hw, hlogt, zsq, dpzsq
    dw = invlinkdz_robit(z,d)
    hw = invlinkhz_robit(z,d)
    zsq = z*z
    dpzsq = d + zsq
    hlogt = -(d+1d0)/dpzsq*(1d0 - 2*zsq/dpzsq)
    y = hlogt*dw - dw*hw + hw*hw/dw
  end function invlink3z_robit

  pure elemental function invlinkdn_robit (z,d) result (y)
    ! First derivative w.r.t. d
    double precision, intent(in) :: z, d
    double precision y
    double precision eps
    eps = sqrt(epsilon(1d0))
    y = 8d0*invlink_robit(z,d+eps) - 8d0*invlink_robit(z,d-eps) &
       - invlink_robit(z,d+eps*2d0) + invlink_robit(z,d-eps*2d0)
    y = y/(12d0*eps)
  end function invlinkdn_robit

  pure elemental function invlinkhn_robit (z,d) result (y)
    ! First derivative w.r.t. d
    double precision, intent(in) :: z, d
    double precision y
    double precision eps
    eps = sqrt(sqrt(epsilon(1d0)))
    y = 16d0*invlink_robit(z,d+eps) + 16d0*invlink_robit(z,d-eps) &
       - invlink_robit(z,d+eps*2d0) - invlink_robit(z,d-eps*2d0) &
       - 30d0*invlink_robit(z,d)
    y = y/(12d0*eps*eps)
  end function invlinkhn_robit

  pure elemental function invlinkdzdn_robit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision dz, dlogF, dlogt
    dz = invlinkdz_robit(z,d)
    dlogF = invlinkdn_robit(z,d)
    dlogt = dnu_logpdft(z,d)
    y = (dlogt - dlogF)*dz
  contains
    pure elemental function dnu_logpdft (z,d)
      use interfaces, only: fdigamma, flog1p
      implicit none
      double precision, intent(in) :: z, d
      double precision dnu_logpdft
      double precision zzd
      zzd = z*z/d
      dnu_logpdft = -.5d0*flog1p(zzd) + .5d0*(d+1)/d*zzd/(zzd+1) &
         - .5d0/d -.5d0*fdigamma(.5d0*d) + .5d0*fdigamma(.5d0*(d+1))
    end function dnu_logpdft
  end function invlinkdzdn_robit

  pure elemental function invlinkhzdn_robit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision dz, hz, dzdn, ddlogt
    dz = invlinkdz_robit(z,d)
    hz = invlinkhz_robit(z,d)
    dzdn = invlinkdzdn_robit(z,d)
    ddlogt = dzdn_logpdft(z,d)
    y = ddlogt*dz + hz*dzdn/dz - dzdn*dz
  contains
    pure elemental function dzdn_logpdft (z,d)
      implicit none
      double precision, intent(in) :: z, d
      double precision dzdn_logpdft
      double precision zzd
      zzd = z*z+d
      !!dzdn_logpdft = (d+1)*z/(d*zzd) - z/zzd - (d+1)*z*z*z/(d*zzd*zzd)
      dzdn_logpdft = -z/zzd*(1d0 - (d+1)/d*(1d0 - z*z/zzd))
    end function dzdn_logpdft
  end function invlinkhzdn_robit

  pure elemental function invlinkdzhn_robit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision dz, hlogt, hlogF, dlogg
    dz = invlinkdz_robit(z,d)
    hlogt = hnu_logpdft(z,d)
    hlogF = invlinkhn_robit(z,d)
    dlogg = invlinkdzdn_robit(z,d)
    dlogg = dlogg/dz
    y = dz*(dlogg*dlogg + hlogt - hlogF)
  contains
    pure elemental function hnu_logpdft (z,d)
      use interfaces, only: ftrigamma, flog1p
      implicit none
      double precision, intent(in) :: z, d
      double precision hnu_logpdft
      double precision zzd
      zzd = z*z/d
      zzd = zzd/(1d0+zzd)
      hnu_logpdft = zzd/d - zzd*(d+1)/(d*d)*(1d0 - .5d0*zzd) + .5/(d*d) &
         - .25d0*ftrigamma(.5d0*d) + .25d0*ftrigamma(.5d0*(d+1))
    end function hnu_logpdft
  end function invlinkdzhn_robit

!!!!!!!!!!!!!!!!!!!!!!!!!!!!! binomial logit !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link function
  pure elemental function flink_logit (w,d) result (z)
    use interfaces, only: quantlogis
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    z = quantlogis(w)
  end function flink_logit

!! Inverse link function (log scale)
  pure elemental function invlink_logit (z,d) result (w)
    ! Binomial symmetric (robit) link fcn
    use interfaces, only: logproblogis
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = logproblogis(z)
  end function invlink_logit

  pure elemental function invlinkdz_logit (z,d) result (y)
    use interfaces, only: logproblogis
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    ! y = 1d0/(1d0 + exp(z))
    y = logproblogis(z)
    y = exp(y)
    y = 1d0 - y
  end function invlinkdz_logit

  pure elemental function loginvlinkdz_logit (z,d) result (y)
    use interfaces, only: logproblogis, flog1mexp
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    ! y = 1d0/(1d0 + exp(z))
    y = logproblogis(z)
    y = flog1mexp(y)
  end function loginvlinkdz_logit

  pure elemental function invlinkhz_logit (z,d) result (y)
    use interfaces, only: logproblogis
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = logproblogis(z)
    y = exp(y)
    y = y*(y-1d0)
  end function invlinkhz_logit

  pure elemental function invlink3z_logit (z,d) result (y)
    use interfaces, only: logproblogis
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = logproblogis(z)
    y = exp(y)
    y = y*(1d0-y)*(y+y-1d0)
  end function invlink3z_logit

  pure elemental function invlinkdn_logit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkdn_logit

  pure elemental function invlinkhn_logit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkhn_logit

  pure elemental function invlinkdzdn_logit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkdzdn_logit

  pure elemental function invlinkhzdn_logit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkhzdn_logit

  pure elemental function invlinkdzhn_logit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkdzhn_logit

!!!!!!!!!!!!!!!!!!!!!!!!!!!!! binomial probit !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link function
  pure elemental function flink_probit (w,d) result (z)
    use interfaces, only: quantnorm
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    z = quantnorm(w)
  end function flink_probit

!! Inverse link function (log scale)
  pure elemental function invlink_probit (z,d) result (w)
    ! Binomial symmetric (robit) link fcn
    use interfaces, only: logprobnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = logprobnorm(z)
  end function invlink_probit

  pure elemental function invlinkdz_probit (z,d) result (y)
    use interfaces, only: logprobnorm, logpdfnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision f, ff
    ff = logprobnorm(z)
    f = logpdfnorm(z)
    y = exp(f - ff)
  end function invlinkdz_probit

  pure elemental function loginvlinkdz_probit (z,d) result (y)
    use interfaces, only: logprobnorm, logpdfnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision f, ff
    ff = logprobnorm(z)
    f = logpdfnorm(z)
    y = f - ff
  end function loginvlinkdz_probit

  pure elemental function invlinkhz_probit (z,d) result (y)
    ! If g(z) is the pdf of the N(0,1) distribution, then we need
    ! d2w = d(log g(z)) * d1w - (d1w)^2
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision dwdz
    dwdz = invlinkdz_probit(z,d)
    y = -z*dwdz - dwdz*dwdz
  end function invlinkhz_probit

  pure elemental function invlink3z_probit (z,d) result (y)
    ! If g(z) is the pdf of the N(0,1) distribution, then we need
    ! d3w = h(log g(z)) * dw - dw * hw + hw*hw/dw
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    double precision dw, hw
    dw = invlinkdz_probit(z,d)
    hw = invlinkhz_probit(z,d)
    y = -dw - dw*hw + hw*hw/dw
  end function invlink3z_probit

  pure elemental function invlinkdn_probit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkdn_probit

  pure elemental function invlinkhn_probit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkhn_probit

  pure elemental function invlinkdzdn_probit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkdzdn_probit

  pure elemental function invlinkhzdn_probit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkhzdn_probit

  pure elemental function invlinkdzhn_probit (z,d) result (y)
    implicit none
    double precision, intent(in) :: z, d
    double precision y
    y = 0d0
  end function invlinkdzhn_probit


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Box-Cox !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link fcn
  pure elemental function flink_boxcox (w,d) result (z)
    ! z = (mu^d - 1)/d
    ! w is log(mu)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    if (d .eq. 0d0) then
      z = w
    else if (d .eq. 1d0) then
      z = fexpm1(w)
    else if (d .eq. -1d0) then
      z = -fexpm1(-w)
    else
      z = fexpm1(w*d)/d
    end if
  end function flink_boxcox

!! Inverse link fcn, returns log(mu)
  pure elemental function invlink_boxcox (z,d) result (w)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = z
    else
      w = d*z
      if (w .gt. -1d0) then
        w = flog1p(w)/d
      else if (d .gt. 0d0) then
        w = bigneg
      else
        w = bigpos
      end if
    end if
  end function invlink_boxcox

  pure elemental function invlinkdz_boxcox (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 1d0
    else
      w = d*z
      if (w .gt. -1d0) then
        w = 1d0/(1d0+w)
      else
        w = 0d0
      end if
    end if
  end function invlinkdz_boxcox

  pure elemental function loginvlinkdz_boxcox (z,d) result (w)
    ! w = log(mu)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 0d0
    else
      w = d*z
      if (w .gt. -1d0) then
        w = -flog1p(w)
      else
        w = bigneg
      end if
    end if
  end function loginvlinkdz_boxcox

  pure elemental function invlinkhz_boxcox (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 0d0
    else
      w = d*z
      if (w .gt. -1d0) then
        w = 1d0+w
        w = -d/(w*w)
      else
        w = 0d0
      end if
    end if
  end function invlinkhz_boxcox

  pure elemental function invlink3z_boxcox (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 0d0
    else
      w = d*z
      if (w .gt. -1d0) then
        w = 1d0+w
        w = 2*d*d/(w*w*w)
      else
        w = 0d0
      end if
    end if
  end function invlink3z_boxcox

  pure elemental function invlinkdn_boxcox (z,d) result (w)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = -.5d0*z*z
    else
      w = d*z
      if (w .gt. -1d0) then
        w = (w/(1d0+w) - flog1p(w))/(d*d)
      else
        w = 0d0
      end if
    end if
  end function invlinkdn_boxcox

  pure elemental function invlinkhn_boxcox (z,d) result (w)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zdp1
    if (d .eq. 0d0) then
      w = 2d0*z*z*z/3d0
    else
      w = d*z
      if (w .gt. -1d0) then
        zdp1 = w + 1d0
        w = -w*(3d0*zdp1 - 1d0) + 2d0*zdp1*zdp1*flog1p(w)
        w = w/(d*d*d*zdp1*zdp1)
      else
        w = 0d0
      end if
    end if
  end function invlinkhn_boxcox

  pure elemental function invlinkdzdn_boxcox (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = -z
    else
      w = d*z
      if (w .gt. -1d0) then
        w = 1d0 + w
        w = -z/(w*w)
      else
        w = 0d0
      end if
    end if
  end function invlinkdzdn_boxcox

  pure elemental function invlinkhzdn_boxcox (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w, dz
    if (d .eq. 0d0) then
      w = -1
    else
      dz = d*z
      if (dz .gt. -1d0) then
        w = 1d0 + dz
        w = (dz-1)/(w*w*w)
      else
        w = 0d0
      end if
    end if
  end function invlinkhzdn_boxcox

  pure elemental function invlinkdzhn_boxcox (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 2*z*z
    else
      w = d*z
      if (w .gt. -1d0) then
        w = 1d0 + w
        w = 2*z*z/(w*w*w)
      else
        w = 0d0
      end if
    end if
  end function invlinkdzhn_boxcox


!!!!!!!!!!!!!!!!!!!!!!!!!!!! Modified Box-Cox !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Link fcn
  pure elemental function flink_modbc (w, d) result (z)
    ! w is log(mu)
    ! z = sign(w)*expm1(d*|w|)/d for d > 0
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    if (d .eq. 0d0) then
      z = w
    else
      z = fexpm1(abs(d*w))/abs(d)
      z = sign(z,w)
!     else if (d .eq. 1d0) then
!       z = fexpm1(abs(w))
!       z = sign(z,w)
!     else if (d .gt. 0d0) then
!       z = fexpm1(d*abs(w))/d
!       z = sign(z,w)
!     else ! Use regular Box-Cox
!       z = fexpm1(w*d)/d
    end if
  end function flink_modbc

  pure elemental function invlink_modbc (z,d) result (w)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = z
    else
      w = flog1p(abs(d*z))/abs(d)
      w = sign(w,z)
!     else if (d .eq. 1d0) then
!       w = flog1p(abs(z))
!       w = sign(w,z)
!     else if (d .gt. 0d0) then
!       w = flog1p(d*abs(z))/d
!       w = sign(w,z)
!     else ! Use regular Box-Cox
!       w = d*z
!       if (w .gt. -1d0) then
!         w = flog1p(w)/d
!       else
!         w = bigneg
!       end if
    end if
  end function invlink_modbc

  pure elemental function invlinkdz_modbc (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 1d0
    else
      w = 1d0/(1d0 + abs(d*z))
!     else if (d .gt. 0d0) then
!       w = 1d0 + d*abs(z)
!       w = 1d0/w
!     else ! Use regular Box-Cox
!       w = d*z
!       if (w .gt. -1d0) then
!         w = 1d0/(1d0+w)
!       else
!         w = 0d0
!       end if
    end if
  end function invlinkdz_modbc

  pure elemental function loginvlinkdz_modbc (z,d) result (w)
    ! w = log(mu)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 0d0
    else
      w = -flog1p(abs(d*z))
    end if
  end function loginvlinkdz_modbc

  pure elemental function invlinkhz_modbc (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w, y
    if (d .eq. 0d0) then
      w = 0d0
    else
      w = d*z
      y = 1d0 + abs(w)
      if (w .lt. 0d0) then
        w = d/(y*y)
      else
        w = -d/(y*y)
      end if
    end if
  end function invlinkhz_modbc

  pure elemental function invlink3z_modbc (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w, y
    if (d .eq. 0d0) then
      w = 0d0
    else
      w = d*z
      y = 1d0 + abs(w)
      w = 2*d*d/(y*y*y)
    end if
  end function invlink3z_modbc

  pure elemental function invlinkdn_modbc (z,d) result (w)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      ! The derivative at 0 is undefined
      w = 0d0
      ! w = -.5d0*z*z
    else
      w = abs(z*d)
      w = (w/(1d0+w) - flog1p(w))/(d*d)
      if (d .lt. 0d0) w = -w
    end if
    if (z .lt. 0d0) w = -w
  end function invlinkdn_modbc

  pure elemental function invlinkhn_modbc (z,d) result (w)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zdp1
    if (d .eq. 0d0) then
      w = 2d0*z*z*z/3d0
    else
      w = abs(d*z)
      zdp1 = w + 1d0
      w = -w*(3d0*zdp1 - 1d0) + 2d0*zdp1*zdp1*flog1p(w)
      w = w/(d*d*d*zdp1*zdp1)
      if (d .lt. 0d0) w = -w
      if (z .lt. 0d0) w = -w
    end if
  end function invlinkhn_modbc

  pure elemental function invlinkdzdn_modbc (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      ! The derivative at 0 is undefined
      w = 0d0
      !w = -abs(z)
    else
      w = 1d0 + abs(d*z)
      w = -abs(z)/(w*w)
      if (d .lt. 0d0) w = -w
    end if
  end function invlinkdzdn_modbc

  pure elemental function invlinkhzdn_modbc (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w, adz
    if (d .eq. 0d0) then
      ! The derivative at 0 is undefined
      w = 0d0
      !w = -1
    else
      adz = abs(d*z)
      w = 1d0 + adz
      w = (adz - 1)/(w*w*w)
      if (d .lt. 0d0) w = -w
    end if
    if (z .lt. 0d0) w = -w
  end function invlinkhzdn_modbc

  pure elemental function invlinkdzhn_modbc (z,d) result (w)
    ! w = log(mu)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = 2*z*z
    else
      w = 1d0 + abs(d*z)
      w = 2*z*z/(w*w*w)
    end if
  end function invlinkdzhn_modbc


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Gaussian !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Linf fcn
  pure elemental function flink_ga (w, d) result (z)
    implicit none
    double precision, intent(in) :: w, d
    double precision z
    if (d .eq. 0d0) then
      if (w > 0d0) then
        z = log(w)
      else
        z = bigneg
      end if
    else if (d .gt. 0d0) then
      if (d .eq. 2d0) then
        z = .5d0*(w*w - 1d0)
      else if (d .eq. 1d0) then
        z = w - 1d0
      else if (d .eq. .5d0) then
        z = sign(sqrt(abs(w)), w) - 1d0
        z = z + z
      else
        z = (sign(abs(w)**d, w) - 1d0)/d
      end if
    else ! d .lt. 0d0
      if (w .gt. 0d0) then
        if (d .eq. -0.5d0) then
          z = -(1d0/sqrt(w) - 1d0)
          z = z + z
        else if (d .eq. -1d0) then
          z = -(1d0/w - 1d0)
        else if (d .eq. -2d0) then
          z = -.5d0*(1d0/(w*w) - 1d0)
        else
          z = (w**d - 1d0)/d
        end if
      else ! w .le. 0d0
        z = bigneg
      end if
    end if
  end function flink_ga

!! Inverse link fcn
  pure elemental function invlink_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = exp(z)
    else if (d .gt. 0d0) then
      w = d*z + 1d0
      if (d .eq. 2d0) then
        w = sign(sqrt(abs(w)),w)
      else if (d .eq. .5d0) then
        w = sign(w*w,w)
      else if (d .ne. 1d0) then
        w = sign(abs(w)**(1d0/d),w)
      end if
    else ! d .lt. 0d0
      w = d*z + 1d0
      if (w .gt. 0d0) then
        w = 1d0/w
        if (d .eq. -2d0) then
          w = sqrt(w)
        else if (d .eq. -.5d0) then
          w = w*w
        else if (d .ne. -1d0) then
          w = w**(1d0/(-d))
        end if
      else
        w = bigpos
      end if
    end if
  end function invlink_ga

  pure elemental function invlinkdz_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = exp(z)
    else if (d .eq. 1d0) then
      w = 1d0
    else if (d .gt. 0d0) then
      w = abs(d*z + 1d0)
      if (d .eq. 2d0) then
        w = 1d0/sqrt(w)
      else if (d .ne. .5d0) then
        w = w**(1d0/d - 1d0)
      end if
    else ! d .lt. 0d0
      w = d*z + 1d0
      if (w .gt. 0d0) then
        if (d .eq. -1d0) then
          w = 1d0/(w*w)
        else if (d .eq. -2d0) then
          w = 1d0/(w*sqrt(w))
        else if (d .eq. -.5d0) then
          w = 1d0/(w*w*w)
        else
          w = w**(1d0/d - 1d0)
        end if
      else
        w = 0d0
      end if
    end if
  end function invlinkdz_ga

  pure elemental function loginvlinkdz_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = z
    else if (d .eq. 1d0) then
      w = 0d0
    else if (d .gt. 0d0) then
      w = abs(d*z + 1d0)
      w = log(w)
      if (d .eq. 2d0) then
        w = -.5d0*w
      else if (d .ne. .5d0) then
        w = (1d0/d - 1d0)*w
      end if
    else ! d .lt. 0d0
      ! w = d*z + 1d0
      w = flog1p(d*z)
      if (w .gt. 0d0) then
        if (d .eq. -1d0) then
          w = -(w + w)
        else if (d .eq. -2d0) then
          w = -1.5d0*w
        else if (d .eq. -.5d0) then
          w = -(w+w+w)
        else
          w = (1d0/d - 1d0)*w
        end if
      else
        w = bigneg
      end if
    end if
  end function loginvlinkdz_ga

  pure elemental function invlinkhz_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = exp(z)
    else if (d .eq. 1d0) then
      w = 0d0
    else if (d .eq. .5d0) then
      w = .5d0
    else if (d .gt. 0d0) then
      w = d*z + 1d0
      if (w .ge. 0d0) then
        w = (1-d) * w**(1d0/d - 2d0)
      else
        w = (d-1) * (-w)**(1d0/d - 2d0)
      end if
    else ! d .lt. 0d0
      w = d*z + 1d0
      if (w .gt. 0d0) then
        w = (1-d) * w**(1d0/d - 2d0)
      else
        w = 0d0
      end if
    end if
  end function invlinkhz_ga

  pure elemental function invlink3z_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = exp(z)
    else if (d .eq. 1d0) then
      w = 0d0
    else if (d .eq. .5d0) then
      w = 0d0
    else if (d .gt. 0d0) then
      w = d*z + 1d0
      w = (1-d)*(1-2*d) * abs(w)**(1d0/d - 3d0)
    else ! d .lt. 0d0
      w = d*z + 1d0
      if (w .gt. 0d0) then
        w = (1-d)*(1-2*d) * (w)**(1d0/d - 3d0)
      else
        w = 0d0
      end if
    end if
  end function invlink3z_ga

  pure elemental function invlinkdn_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zdp1
    if (d .eq. 0d0) then
      w = -.5d0*z*z*exp(z)
    else if (d .gt. 0d0) then
      w = d*z
      zdp1 = w + 1d0
      w = abs(zdp1)**(1d0/d-1d0)*(w - zdp1*log(abs(zdp1)))/(d*d)
    else ! d .lt. 0d0
      w = d*z
      zdp1 = w + 1d0
      if (zdp1 .gt. 0d0) then
        w = zdp1**(1d0/d-1d0)*(w - zdp1*log(zdp1))/(d*d)
      else
        w = 0d0
      end if
    end if
  end function invlinkdn_ga

  pure elemental function invlinkhn_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zdp1, zoz, t
    if (d .eq. 0d0) then
      w = z*z*z*(8d0 + 3d0*z)*exp(z)/12d0
    else if (d .gt. 0d0) then
      w = d*z
      zdp1 = w + 1d0
      zoz = w/zdp1
      t = zoz - log(abs(zdp1))
      w = abs(zdp1)**(1d0/d)*((-zoz*zoz - t - t) + t*t/d)/(d*d*d)
      if (zdp1 .lt. 0d0) w = -w
    else ! d .lt. 0d0
      w = d*z
      zdp1 = w + 1d0
      if (zdp1 .gt. 0d0) then
      zdp1 = w + 1d0
      zoz = w/zdp1
      t = zoz - log(zdp1)
      w = zdp1**(1d0/d)*((-zoz*zoz - t - t) + t*t/d)/(d*d*d)
      else
        w = 0d0
      end if
    end if
  end function invlinkhn_ga

  pure elemental function invlinkdzdn_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w, omod
    if (d .eq. 0d0) then
      w = -.5d0*exp(z)*z*(z+2)
    else if (d .eq. 1d0) then
      w = -log(abs(z+1d0))
    else if (d .gt. 0d0) then
      w = d*z + 1d0
      omod = 1d0 - 1d0/d
      if (w .lt. 0d0) then
        w = -w
        w = omod*z*(w**(-1d0-omod)) - (w**(-omod))*log(w)/(d*d)
      else
        w = -omod*z*(w**(-1d0-omod)) - (w**(-omod))*log(w)/(d*d)
      end if
    else ! d .lt. 0d0
      w = d*z + 1d0
      omod = 1d0 - 1d0/d
      if (w .gt. 0d0) then
        w = -omod*z*(w**(-1d0-omod)) - (w**(-omod))*log(w)/(d*d)
      else
        w = 0d0
      end if
    end if
  end function invlinkdzdn_ga

  pure elemental function invlinkhzdn_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w, omod
    if (d .eq. 0d0) then
      w = -.5d0*exp(z)*(2+4*z+z*z)
    else if (d .eq. 1d0) then
      w = -1d0/(1d0+z)
    else if (d .gt. 0d0) then
      w = d*z + 1d0
      omod = 1d0 - 1d0/d
      if (w .lt. 0d0) then
        w = -w
        w = (w**(-1d0-omod))*(1d0 + omod*d*((1+omod)*z/w - log(w)/(d*d)))
      else
        w = -(w**(-1d0-omod))*(1d0 - omod*d*((1+omod)*z/w + log(w)/(d*d)))
      end if
    else ! d .lt. 0d0
      w = d*z + 1d0
      omod = 1d0 - 1d0/d
      if (w .gt. 0d0) then
        w = -(w**(-1d0-omod))*(1d0 - omod*d*((1+omod)*z/w + log(w)/(d*d)))
      else
        w = 0d0
      end if
    end if
  end function invlinkhzdn_ga

  pure elemental function invlinkdzhn_ga (z,d) result(w)
    ! Inverse Box-Cox transformation. Using extended if d > 0.
    implicit none
    double precision, intent(in) :: z, d
    double precision w, omod, zo1pzd
    if (d .eq. 0d0) then
      w = exp(z)*z*z*(24+20*z + 3*z*z)/12d0
    else if (d .eq. 1d0) then
      w = log(abs(z+1d0))
      w = w*w + 2*w - 2*z/(1+z)
    else if (d .gt. 0d0) then
      w = abs(d*z + 1d0)
      omod = 1d0 - 1d0/d
      zo1pzd = z/w
      w = w**(-omod)*(omod*zo1pzd*zo1pzd - 2*zo1pzd/(d*d) &
         + 2*log(w)/(d*d*d) + (omod*zo1pzd + log(w)/(d*d))**2)
    else ! d .lt. 0d0
      w = d*z + 1d0
      omod = 1d0 - 1d0/d
      zo1pzd = z/w
      if (w .gt. 0d0) then
      w = w**(-omod)*(omod*zo1pzd*zo1pzd - 2*zo1pzd/(d*d) &
         + 2*log(w)/(d*d*d) + (omod*zo1pzd + log(w)/(d*d))**2)
      else
        w = 0d0
      end if
    end if
  end function invlinkdzhn_ga


!!!!!!!!!!!!!!!!!!! Wallace transformation for binomial !!!!!!!!!!!!!!!!!!!
!!! See function u4 in
!!!  Wallace, D. L. (1959). Bounds on normal approximations to Student's
!!!  and the chi-square distributions. The Annals of Mathematical
!!!  Statistics, 1121-1130.
!! Link fcn
!! mu = PHI[sign(z) * c(nu) * sqrt(nu*log(1+z*z/nu))]
!! where c(nu) = (8*nu+1)/(8*nu+3)
!! w = log(mu)
  pure elemental function flink_wallace (w, d) result(z)
    use interfaces, only: quantnorm, fexpm1
    implicit none
    double precision, intent(in) :: w, d
    double precision :: z
    double precision t, e
    t = 8d0*d
    e = quantnorm(w)*(t + 3d0)/(t + 1d0)
    t = sqrt(d*fexpm1(e*e/d))
    z = sign(t, e)
  end function flink_wallace

!! Inverse link fcn (log scale)
  pure elemental function invlink_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision x, t
    t = 8d0*d
    x = sqrt(d*flog1p(z*z/d))*(t + 1d0)/(t + 3d0)
    t = sign(x, z)
    w = logprobnorm(t)
  end function invlink_wallace

  pure elemental function invlinkdz_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision t, c, f, u, dzt, zt
    c = 8d0*d
    c = (c + 1d0)/(c + 3d0)
    t = z*z/d
    u = sqrt(d*flog1p(t))
    if (u .eq. 0d0) then
      dzt = c
      zt = 0d0
    else
      dzt = c*abs(z)/u/(1+t)
      zt = c*u
      if (z .lt. 0d0) zt = -zt
    end if
    w = logprobnorm(zt)
    f = logpdfnorm(zt)
    w = exp(f-w)*dzt
  end function invlinkdz_wallace

  pure elemental function loginvlinkdz_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision t, c, f, u, dzt, zt
    c = 8d0*d
    c = (c + 1d0)/(c + 3d0)
    t = z*z/d
    u = sqrt(d*flog1p(t))
    if (u .eq. 0d0) then
      dzt = c
      zt = 0d0
    else
      dzt = c*abs(z)/u/(1+t)
      zt = c*u
      if (z .lt. 0d0) zt = -zt
    end if
    w = logprobnorm(zt)
    f = logpdfnorm(zt)
    w = f - w + log(dzt)
  end function loginvlinkdz_wallace

  pure elemental function invlinkhz_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision t, c, f, u, dzt, zt, hzt
    c = 8d0*d
    c = (c + 1d0)/(c + 3d0)
    t = z*z/d
    u = sqrt(d*flog1p(t))
    t = t+1d0
    if (u .eq. 0d0) then
      dzt = c
      zt = 0d0
      hzt = 0d0
    else
      zt = c*u
      dzt = c*abs(z)/u/t
      hzt = t*u
      u = u*u
      hzt = c*((1-z*z/d)*u-z*z)/(t*u*hzt)
      if (z .lt. 0d0) then
        zt = -zt
        hzt = -hzt
      end if
    end if
    w = logprobnorm(zt)
    f = logpdfnorm(zt)
    w = exp(f-w)
    dzt = dzt*dzt
    w = w*(hzt - zt*dzt) - w*w*dzt
  end function invlinkhz_wallace

  pure elemental function invlink3z_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision f, w, dw, hw, dzt, hzt, szt, zt
    zt = wallace_zeta(z,d)
    dzt = wallace_zeta_dz(z,d)
    hzt = wallace_zeta_hz(z,d)
    szt = wallace_zeta_3z(z,d)
    dw = invlinkdz_wallace(z,d)
    hw = invlinkhz_wallace(z,d)
    w = logprobnorm(zt)
    f = logpdfnorm(zt)
    w = exp(f-w)
    w = (zt*zt-1)*w*dzt*dzt*dzt -3*zt*w*dzt*hzt + w*szt - dw*dw*dw - 3*dw*hw
  end function invlink3z_wallace

  pure elemental function invlinkdn_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zzd, c, cd, lzzd, u, zt, dzt, phip, phid
    zzd = z*z/d
    c = 8*d
    cd = c + 3
    c = (c+1)/cd
    lzzd = flog1p(zzd)
    u = sqrt(d*lzzd)
    if (u .eq. 0d0) then
      zt = 0d0
      dzt = 0d0
    else
      zt = c*u
      dzt = 16*u/(cd*cd) + .5d0*c*(lzzd - zzd/(1+zzd))/u
      if (z .lt. 0d0) then
        zt = -zt
        dzt = -dzt
      end if
    end if
    phip = logprobnorm(zt)
    phid = logpdfnorm(zt)
    w = exp(phid - phip)*dzt
  end function invlinkdn_wallace

  pure elemental function invlinkhn_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm, flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zzd, c, cd, lzzd, u, zt, dzt, hzt, phip, phid, zo
    zzd = z*z/d
    zo = zzd/(1+zzd)
    c = 8*d
    cd = c + 3
    c = (c+1)/cd
    lzzd = flog1p(zzd)
    u = sqrt(d*lzzd)
    if (u .eq. 0d0) then
      zt = 0d0
      dzt = 0d0
      hzt = 0d0
    else
      zt = c*u
      dzt = 16*u/(cd*cd) + .5d0*c*(lzzd - zo)/u
      hzt = -.25d0*c*zo*zo/(u*u*u) &
         + .5d0*((3/d+64*d)*zo/(1+zzd)-32*zo*zo)/(u*cd*cd) &
         - .25d0*(9-72*d+960*d*d+512*d*d*d)*u/(d*d*cd*cd*cd)
      if (z .lt. 0d0) then
        zt = -zt
        dzt = -dzt
        hzt = -hzt
      end if
    end if
    phip = logprobnorm(zt)
    phid = logpdfnorm(zt)
    w = exp(phid-phip)
    dzt = dzt*dzt
    w = w*(hzt - zt*dzt) - w*w*dzt
  end function invlinkhn_wallace

  pure elemental function invlinkdzdn_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zt, ztdz, ztdn, ztdzdn, phid, phip
    zt = wallace_zeta(z,d)
    ztdz = wallace_zeta_dz(z,d)
    ztdn = wallace_zeta_dn(z,d)
    ztdzdn = wallace_zeta_dzdn(z,d)
    phid = logpdfnorm(zt)
    phip = logprobnorm(zt)
    w = exp(phid - phip)
    w = w*ztdzdn - (zt+w)*w*ztdz*ztdn
  end function invlinkdzdn_wallace

  pure elemental function invlinkhzdn_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zt, ztdz, ztdn, zthz, ztdzdn, zthzdn, phid, phip
    double precision wdz, wdn, whz, wdzdn
    zt = wallace_zeta(z,d)
    ztdz = wallace_zeta_dz(z,d)
    zthz = wallace_zeta_hz(z,d)
    ztdn = wallace_zeta_dn(z,d)
    ztdzdn = wallace_zeta_dzdn(z,d)
    zthzdn = wallace_zeta_hzdn(z,d)
    wdz = invlinkdz_wallace(z,d)
    whz = invlinkhz_wallace(z,d)
    wdn = invlinkdn_wallace(z,d)
    wdzdn = invlinkdzdn_wallace(z,d)
    phid = logpdfnorm(zt)
    phip = logprobnorm(zt)
    w = exp(phid - phip)
    w = (zt*zt-1)*w*ztdz*ztdz*ztdn - 2*zt*w*ztdz*ztdzdn - zt*w*zthz*ztdn &
       + w*zthzdn - wdz*wdz*wdn - 2*wdz*wdzdn - whz*wdn
  end function invlinkhzdn_wallace

  pure elemental function invlinkdzhn_wallace (z,d) result (w)
    use interfaces, only: logprobnorm, logpdfnorm
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision zt, ztdz, ztdn, ztdzdn, zthn, ztdzhn, phid, phip
    double precision dw1, dw2, dw3
    zt = wallace_zeta(z,d)
    ztdz = wallace_zeta_dz(z,d)
    ztdn = wallace_zeta_dn(z,d)
    ztdzdn = wallace_zeta_dzdn(z,d)
    zthn = wallace_zeta_hn(z,d)
    ztdzhn = wallace_zeta_dzhn(z,d)
    phid = logpdfnorm(zt)
    phip = logprobnorm(zt)
    dw1 = exp(phid - phip)
    dw2 = -zt*dw1 - dw1*dw1
    dw3 = -dw1 - zt*dw2 - 2*dw1*dw2
    w = dw3*ztdz*ztdn*ztdn &
       + dw2*(ztdzdn*ztdn*2 + ztdz*zthn) +dw1*ztdzhn
  end function invlinkdzhn_wallace

  pure elemental double precision function wallace_zeta (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zzd, c, c1, c3, lgz
    c = 8*d
    c1 = 1 + c
    c3 = 3 + c
    zzd = z*z/d
    lgz = flog1p(zzd)
    lgz = d*lgz
    zt = c1/c3*sqrt(lgz)
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta

  pure elemental double precision function wallace_zeta_dz (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zzd, c, c1, c3, lgz
    c = 8*d
    c1 = 1 + c
    c3 = 3 + c
    zzd = z*z/d
    lgz = flog1p(zzd)
    lgz = d*lgz
    zzd = 1 + zzd
    zt = z*c1/(c3*zzd*sqrt(lgz))
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta_dz

  pure elemental double precision function wallace_zeta_hz (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zzd, c, c1, c3, lgz, num, dum
    c = 8*d
    c1 = 1 + c
    c3 = 3 + c
    zzd = z*z/d
    lgz = flog1p(zzd)
    num = c1*(d*lgz - d*zzd*(1+lgz))
    lgz = d*lgz
    zzd = 1 + zzd
    dum = c3*zzd*zzd*lgz*sqrt(lgz)
    zt = num/dum
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta_hz

  pure elemental double precision function wallace_zeta_3z (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zzd, c, lgz, zlgz, num, dum, zsq
    c = 8*d
    zsq = z*z
    zzd = zsq/d
    lgz = flog1p(zzd)
    zlgz = z*lgz
    num = -d*(1+c)*(-3*z*z*z + 3*(d-zsq)*zlgz + 2*(3*d - zsq)*zlgz*lgz)
    zzd = zzd+1
    lgz = d*lgz
    dum = (3+c)*d*zzd*zzd*zzd*lgz*lgz*sqrt(lgz)
    zt = num/dum
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta_3z

  pure elemental double precision function wallace_zeta_dn (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zzd, c, c1, c3, lgz, num, dum
    c = 8*d
    c1 = 1 + c
    c3 = 3 + c
    zzd = z*z/d
    lgz = flog1p(zzd)
    num = -c1*c3*zzd + (3+8*c+c*c)*(1+zzd)*lgz
    lgz = d*lgz
    zzd = 1 + zzd
    dum = 2*c3*c3*zzd*sqrt(lgz)
    zt = num/dum
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta_dn

  pure elemental double precision function wallace_zeta_hn (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zzd, c, c1, c3, cc, lgz, lgzd, lgzr
    c = 8*d
    c1 = 1 + c
    c3 = 3 + c
    cc = c1/c3
    zzd = z*z/d
    lgz = flog1p(zzd)
    zzd = zzd/(1+zzd)
    lgzd = sqrt(d*lgz)
    lgzr = lgz/lgzd
    zt = -.25d0*cc*lgzr*lgzr/lgzd + (8/c3)*(1 - cc)*lgzr &
       - 128/(c3*c3)*(1 - cc)*lgzd &
       - .5d0*zzd*zzd*cc/lgzd*(1d0/d + .5d0/(lgzd*lgzd)) &
       + zzd/lgzd*(.5d0*cc*lgzr/lgzd - 8*(1-cc)/c3)
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta_hn

  pure elemental double precision function wallace_zeta_dzdn (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zzd, c, c1, c3, lgz, num, dum
    c = 8*d
    c1 = 1 + c
    c3 = 3 + c
    zzd = z*z/d
    lgz = flog1p(zzd)
    num = c1*c3*z*zzd - z*(3+c*c)*lgz + z*(3 + 8*c + c*c)*zzd*lgz
    lgz = d*lgz
    zzd = 1 + zzd
    dum = 2*c3*c3*zzd*zzd*lgz*sqrt(lgz)
    zt = num/dum
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta_dzdn

  pure elemental double precision function wallace_zeta_hzdn (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zsq, dsq, zzd, c3, lgz, num, dum
    c3 = 3 + 8*d
    dsq = d*d
    zsq = z*z
    zzd = z*z/d
    lgz = flog1p(zzd)
    num = -3*(3 + 32*d + 64*dsq)*zsq*zsq &
       + 2*zsq*(128*dsq*d - 3*zsq + dsq*(48 - 64*zsq) + d*(6 - 48*zsq))*lgz &
       - (64*dsq*dsq - 384*dsq*d*zsq + 3*zsq*zsq + 2*d*zsq*(-9 + 32*zsq) &
       + dsq*(3 - 192*zsq + 64*zsq*zsq))*lgz*lgz
    lgz = d*lgz
    zzd = 1+zzd
    dum = 2*c3*c3*zzd*zzd*zzd*d*lgz*lgz*sqrt(lgz)
    zt = num/dum
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta_hzdn

  pure elemental double precision function wallace_zeta_dzhn (z,d) result (zt)
    use interfaces, only: flog1p
    implicit none
    double precision, intent(in) :: z, d
    double precision zzd, c, c1, c3, cc, lgz, lgzd, lgzd2
    double precision t1, t2, t3
    c = 8*d
    c1 = 1 + c
    c3 = 3 + c
    cc = c1/c3
    zzd = z*z/d
    lgz = flog1p(zzd)
    zzd = zzd/(1+zzd)
    lgzd2 = d*lgz
    lgzd = sqrt(lgzd2)
    t1 = zzd/(z*lgzd)*(.75d0*cc/d - 16/(c3*c3) - 256/(c3*c3*c3)*d)
    t2 = zzd*zzd/(z*lgzd)*(-1.5d0*cc/lgzd2 + 16*d/(c3*c3*lgzd2) &
       - 3*cc/d + 32/(c3*c3))
    t3 = zzd*zzd*zzd*cc/(z*lgzd)*(1.5d0/lgzd2 + 2/d + .75d0*d/(lgzd2*lgzd2))
    zt =  t1 + t2 + t3
    if (z .lt. 0d0) then
      zt = -zt
    end if
  end function wallace_zeta_dzhn


!!!!!!!!!!!!!!!!!!!!!!!!!! modified GEV binomial !!!!!!!!!!!!!!!!!!!!!!!!!!
!! mu = exp{-(1+nu*|z|)^{-sign(z)/nu}} if nu neq 0
!!    = exp{-exp(z)}                   if nu == 0
  pure elemental function flink_modgev (w,d) result(z)
    implicit none
    double precision, intent(in) :: w, d
    double precision :: z
    z = log(-w)
    if (d .eq. 0d0) then
      z = -z
    else
      z = -flink_modbc(z,-d)
    end if
  end function flink_modgev

  pure elemental function invlink_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = -exp(-z)
    else
      w = -exp(invlink_modbc(-z,-d))
    end if
  end function invlink_modgev

  pure elemental function invlinkdz_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = exp(-z)
    else
      w = exp(invlink_modbc(-z,-d))*invlinkdz_modbc(-z,-d)
    end if
  end function invlinkdz_modgev

  pure elemental function loginvlinkdz_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    if (d .eq. 0d0) then
      w = -z
    else
      w = invlink_modbc(-z,-d) + loginvlinkdz_modbc(-z,-d)
    end if
  end function loginvlinkdz_modgev

  pure elemental function invlinkhz_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w, y, hz
    if (d .eq. 0d0) then
      w = -exp(-z)
    else
      y = exp(invlink_modbc(-z,-d))
      w = invlinkdz_modbc(-z,-d)
      hz = invlinkhz_modbc(-z,-d)
      w = -(w*w + hz)*y
    end if
  end function invlinkhz_modgev

  pure elemental function invlink3z_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision dz, y, hz, w
    if (d .eq. 0d0) then
      w = exp(-z)
    else
      y = exp(invlink_modbc(-z,-d))
      dz = invlinkdz_modbc(-z,-d)
      hz = invlinkhz_modbc(-z,-d)
      w = invlink3z_modbc(-z,-d)
      w = (dz*dz*dz + 3*dz*hz + w)*y
    end if
  end function invlink3z_modgev

  pure elemental function invlinkdn_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dg
    w = exp(invlink_modbc(-z,-d))
    dg = invlinkdn_modbc(-z,-d)
    w = dg*w
  end function invlinkdn_modgev

  pure elemental function invlinkhn_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dg, d2g
    w = exp(invlink_modbc(-z,-d))
    dg = invlinkdn_modbc(-z,-d)
    d2g = invlinkhn_modbc(-z,-d)
    w = -(dg*dg + d2g)*w
  end function invlinkhn_modgev

  pure elemental function invlinkdzdn_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dz, dn, dzdn
    w = exp(invlink_modbc(-z,-d))
    dz = invlinkdz_modbc(-z,-d)
    dn = invlinkdn_modbc(-z,-d)
    dzdn = invlinkdzdn_modbc(-z,-d)
    w = -(dz*dn + dzdn)*w
  end function invlinkdzdn_modgev

  pure elemental function invlinkhzdn_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dz, dn, hz, dzdn, hzdn
    w = exp(invlink_modbc(-z,-d))
    dz = invlinkdz_modbc(-z,-d)
    hz = invlinkhz_modbc(-z,-d)
    dn = invlinkdn_modbc(-z,-d)
    dzdn = invlinkdzdn_modbc(-z,-d)
    hzdn = invlinkhzdn_modbc(-z,-d)
    w = (dz*dz*dn + 2*dz*dzdn + hz*dn +hzdn)*w
  end function invlinkhzdn_modgev

  pure elemental function invlinkdzhn_modgev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dz, dn, hn, dzdn, dzhn
    w = exp(invlink_modbc(-z,-d))
    dz = invlinkdz_modbc(-z,-d)
    dn = invlinkdn_modbc(-z,-d)
    hn = invlinkhn_modbc(-z,-d)
    dzdn = invlinkdzdn_modbc(-z,-d)
    dzhn = invlinkdzhn_modbc(-z,-d)
    w = (dz*dn*dn + dz*hn + 2*dzdn*dn + dzhn)*w
  end function invlinkdzhn_modgev

!!!!!!!!!!!!!!!! modified GEV negative skeweness binomial !!!!!!!!!!!!!!!!!
!! mu = 1 - exp{-(1+nu*|z|)^{sign(z)/nu}} if nu neq 0
!!    = 1 - exp{-exp(-z)}                 if nu == 0
  pure elemental function flink_modgevns (w,d) result(z)
    use interfaces, only: flog1mexp
    implicit none
    double precision, intent(in) :: w, d
    double precision :: z
    z = -flink_modgev(flog1mexp(w),d)
  end function flink_modgevns

  pure elemental function invlink_modgevns (z,d) result (w)
    use interfaces, only: flog1mexp
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = flog1mexp(invlink_modgev(-z,d))
  end function invlink_modgevns

  pure elemental function invlinkdz_modgevns (z,d) result (w)
    use interfaces, only: fexpm1, flogexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision w, e, f1
    if (d .eq. 0d0) then
      f1 = exp(z)
!!       e = fexpm1(f1)
!!       w = f1/e
      w = z - flogexpm1(f1)
      w = exp(w)
    else
      e = fexpm1(-invlink_modgev(-z,d))
      f1 = invlinkdz_modgev(-z,d)
      w = f1/e
    end if
  end function invlinkdz_modgevns

  pure elemental function loginvlinkdz_modgevns (z,d) result (w)
    use interfaces, only: fexpm1, flogexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision w, e, f1
    if (d .eq. 0d0) then
      f1 = exp(z)
!!       e = fexpm1(f1)
!!       w = f1/e
      w = z - flogexpm1(f1)
    else
      e = flogexpm1(-invlink_modgev(-z,d))
      f1 = invlinkdz_modgev(-z,d)
      w = log(f1) - e
    end if
  end function loginvlinkdz_modgevns

  pure elemental function invlinkhz_modgevns (z,d) result (w)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision w, w1, f1, f2, e
    e = fexpm1(-invlink_modgev(-z,d))
    f1 = invlinkdz_modgev(-z,d)
    f2 = invlinkhz_modgev(-z,d)
    w1 = f1/e
    w = -w1*w1 - w1*f1 - f2/e
  end function invlinkhz_modgevns

  pure elemental function invlink3z_modgevns (z,d) result (w)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision w, e, f1, f2, f3, w1, w2, fr
    e = fexpm1(-invlink_modgev(-z,d))
    f1 = invlinkdz_modgev(-z,d)
    f2 = invlinkhz_modgev(-z,d)
    f3 = invlink3z_modgev(-z,d)
    fr = f2/f1
    w1 = f1/e
    w2 = -w1*w1 - w1*f1 - f2/e
    w = -2d0*w1*w2 - w2*f1 + w1*f2 - w2*fr + w1*fr*fr + f3/e
  end function invlink3z_modgevns

  pure elemental function invlinkdn_modgevns (z,d) result (w)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision w, e, f1
    e = fexpm1(-invlink_modgev(-z,d))
    f1 = invlinkdn_modgev(-z,d)
    w = -f1/e
  end function invlinkdn_modgevns

  pure elemental function invlinkhn_modgevns (z,d) result (w)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision w, w1, f1, f2, e
    e = fexpm1(-invlink_modgev(-z,d))
    f1 = invlinkdn_modgev(-z,d)
    f2 = invlinkhn_modgev(-z,d)
    w1 = -f1/e
    w = -w1*w1 + w1*f1 - f2/e
  end function invlinkhn_modgevns

  pure elemental function invlinkdzdn_modgevns (z,d) result (w)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision wz, e, f10, f01, f11, w
    e = fexpm1(-invlink_modgev(-z,d))
    f10 = invlinkdz_modgev(-z,d)
    f01 = invlinkdn_modgev(-z,d)
    f11 = invlinkdzdn_modgev(-z,d)
    wz = f10/e
    w = wz*f01 + wz*wz*f01/f10 + f11/e
  end function invlinkdzdn_modgevns

  pure elemental function invlinkhzdn_modgevns (z,d) result (w)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision wz, wzd, e, f10, f01, f11, f20, f21, w
    e = fexpm1(-invlink_modgev(-z,d))
    f10 = invlinkdz_modgev(-z,d)
    f01 = invlinkdn_modgev(-z,d)
    f11 = invlinkdzdn_modgev(-z,d)
    f20 = invlinkhz_modgev(-z,d)
    f21 = invlinkhzdn_modgev(-z,d)
    wz = f10/e
    wzd = wz*f01 + wz*wz*f01/f10 + f11/e
    w = -2d0*wz*wzd - wzd*f10 - wz*f11 - (wzd - f11/e)*f20/f10 - f21/e
  end function invlinkhzdn_modgevns

  pure elemental function invlinkdzhn_modgevns (z,d) result (w)
    use interfaces, only: fexpm1
    implicit none
    double precision, intent(in) :: z, d
    double precision wz, wzd, e, f10, f01, f11, f02, f12, w
    e = fexpm1(-invlink_modgev(-z,d))
    f10 = invlinkdz_modgev(-z,d)
    f01 = invlinkdn_modgev(-z,d)
    f11 = invlinkdzdn_modgev(-z,d)
    f02 = invlinkhn_modgev(-z,d)
    f12 = invlinkdzhn_modgev(-z,d)
    wz = f10/e
    wzd = wz*f01 + wz*wz*f01/f10 + f11/e
    w = wzd*f01 + wz*f02 + 2d0*wzd*f01/e + wz*f02/e - f01*f11/(e*e) &
       + wzd*f11/f10 + f12/e - f11*f11/(e*f10)
  end function invlinkdzhn_modgevns


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! GEV binomial !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! mu = exp{-max(0,1+nu*z)^{-1/nu}} if nu neq 0
!!    = exp{-exp(z)}                if nu == 0
  pure elemental function flink_gev (w,d) result(z)
    implicit none
    double precision, intent(in) :: w, d
    double precision :: z
    z = -flink_boxcox(log(-w),-d)
  end function flink_gev

  pure elemental function invlink_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = -exp(invlink_boxcox(-z,-d))
  end function invlink_gev

  pure elemental function invlinkdz_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = exp(invlink_boxcox(-z,-d))*invlinkdz_boxcox(-z,-d)
  end function invlinkdz_gev

  pure elemental function loginvlinkdz_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    w = invlink_boxcox(-z,-d) + loginvlinkdz_boxcox(-z,-d)
  end function loginvlinkdz_gev

  pure elemental function invlinkhz_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w, y, hz
    y = exp(invlink_boxcox(-z,-d))
    w = invlinkdz_boxcox(-z,-d)
    hz = invlinkhz_boxcox(-z,-d)
    w = -(w*w + hz)*y
  end function invlinkhz_gev

  pure elemental function invlink3z_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision dz, y, hz, w
    y = exp(invlink_boxcox(-z,-d))
    dz = invlinkdz_boxcox(-z,-d)
    hz = invlinkhz_boxcox(-z,-d)
    w = invlink3z_boxcox(-z,-d)
    w = (dz*dz*dz + 3*dz*hz + w)*y
  end function invlink3z_gev

  pure elemental function invlinkdn_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dg
    w = exp(invlink_boxcox(-z,-d))
    dg = invlinkdn_boxcox(-z,-d)
    w = dg*w
  end function invlinkdn_gev

  pure elemental function invlinkhn_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dg, d2g
    w = exp(invlink_boxcox(-z,-d))
    dg = invlinkdn_boxcox(-z,-d)
    d2g = invlinkhn_boxcox(-z,-d)
    w = -(dg*dg + d2g)*w
  end function invlinkhn_gev

  pure elemental function invlinkdzdn_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dz, dn, dzdn
    w = exp(invlink_boxcox(-z,-d))
    dz = invlinkdz_boxcox(-z,-d)
    dn = invlinkdn_boxcox(-z,-d)
    dzdn = invlinkdzdn_boxcox(-z,-d)
    w = -(dz*dn + dzdn)*w
  end function invlinkdzdn_gev

  pure elemental function invlinkhzdn_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dz, dn, hz, dzdn, hzdn
    w = exp(invlink_boxcox(-z,-d))
    dz = invlinkdz_boxcox(-z,-d)
    hz = invlinkhz_boxcox(-z,-d)
    dn = invlinkdn_boxcox(-z,-d)
    dzdn = invlinkdzdn_boxcox(-z,-d)
    hzdn = invlinkhzdn_boxcox(-z,-d)
    w = (dz*dz*dn + 2*dz*dzdn + hz*dn +hzdn)*w
  end function invlinkhzdn_gev

  pure elemental function invlinkdzhn_gev (z,d) result (w)
    implicit none
    double precision, intent(in) :: z, d
    double precision w
    double precision dz, dn, hn, dzdn, dzhn
    w = exp(invlink_boxcox(-z,-d))
    dz = invlinkdz_boxcox(-z,-d)
    dn = invlinkdn_boxcox(-z,-d)
    hn = invlinkhn_boxcox(-z,-d)
    dzdn = invlinkdzdn_boxcox(-z,-d)
    dzhn = invlinkdzhn_boxcox(-z,-d)
    w = (dz*dn*dn + dz*hn + 2*dzdn*dn + dzhn)*w
  end function invlinkdzhn_gev


!!!!!!!!!!!!!!!!!!!!!!!!!!!!! Other functions !!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!   elemental function identity (z,d) result (w)
!!     implicit none
!!     double precision, intent(in) :: z, d
!!     double precision w
!!     w = z
!!   end function identity
!!
!!   elemental function constant (z,d) result (w)
!!     implicit none
!!     double precision, intent(in) :: z, d
!!     double precision w
!!     w = 1d0
!!   end function constant
!!
!!   elemental function zero (z,d) result (w)
!!     implicit none
!!     double precision, intent(in) :: z, d
!!     double precision w
!!     w = 0d0
!!   end function zero

end module modelfcns_link

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary:
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


module covfun
  use cor_fcns
  implicit none
  public
  private :: covmat_a, covmat_l, covmat_v
  public :: spcor, spcor_dh, spcor_dk, spcor_hh, spcor_hk, &
     spcor_dhdk, upper_tri
  integer, parameter, private :: CORCODES(5) = (/1,2,3,4,5/)
  logical, private :: CORDEF = .false.
  integer, private :: CORIS = 0, DIMSP = 0
  logical, private, pointer :: lup(:,:) => null()


  interface covmat
    module procedure covmat_a, covmat_l, covmat_v
  end interface covmat

contains

  subroutine create_spcor (icf,n)
    use cor_fcns
    implicit none
    integer, intent(in) :: icf
    integer, intent(in) :: n
    integer i

    if (.not.(CORDEF .and. (CORIS .eq. icf))) then
      if (.not. (any(icf .eq. CORCODES))) then
        call rexit ("Unrecognised correlation.")
      end if
      CORIS = icf
      CORDEF = .true.
    end if

    if (n .gt. 0) then
      if (DIMSP .ne. n) then
        nullify(lup)
        allocate(lup(n,n))
        DIMSP = n
        do i = 1, n
          lup(1:i,i) = .true.
          lup(i+1:n,i) = .false.
        end do
      end if
    end if
  end subroutine create_spcor

  function upper_tri () result (lup_)
    logical, pointer :: lup_(:,:)
    allocate(lup_(DIMSP,DIMSP))
    lup_ = lup
  end function upper_tri

  pure elemental double precision function spcor (h,k)
    implicit none
    double precision, intent(in) :: h,k
    select case (CORIS)
    case (1) ! Matern
      spcor = cor_matern(h,k)
    case (2) ! Spherical
      spcor = cor_spher(h,k)
    case (3) ! Power exponential
      spcor = cor_powexp(h,k)
    case (4) ! Exponential
      spcor = cor_exp(h,k)
    case (5) ! Gaussian
      spcor = cor_gaussian(h,k)
    end select
  end function spcor

  pure elemental double precision function spcor_dh (h,k)
    implicit none
    double precision, intent(in) :: h,k
    select case (CORIS)
    case (1) ! Matern
      spcor_dh = cor_dh_matern(h,k)
    case (2) ! Spherical
      spcor_dh = cor_dh_spher(h,k)
    case (3) ! Power exponential
      spcor_dh = cor_dh_powexp(h,k)
    case (4) ! Exponential
      spcor_dh = cor_dh_exp(h,k)
    case (5) ! Gaussian
      spcor_dh = cor_dh_gaussian(h,k)
    end select
  end function spcor_dh

  pure elemental double precision function spcor_dk (h,k)
    implicit none
    double precision, intent(in) :: h,k
    select case (CORIS)
    case (1) ! Matern
      spcor_dk = cor_dk_matern(h,k)
    case (2) ! Spherical
      spcor_dk = cor_dk_spher(h,k)
    case (3) ! Power exponential
      spcor_dk = cor_dk_powexp(h,k)
    case (4) ! Exponential
      spcor_dk = cor_dk_exp(h,k)
    case (5) ! Gaussian
      spcor_dk = cor_dk_gaussian(h,k)
    end select
  end function spcor_dk

  pure elemental double precision function spcor_hh (h,k)
    implicit none
    double precision, intent(in) :: h,k
    select case (CORIS)
    case (1) ! Matern
      spcor_hh = cor_hh_matern(h,k)
    case (2) ! Spherical
      spcor_hh = cor_hh_spher(h,k)
    case (3) ! Power exponential
      spcor_hh = cor_hh_powexp(h,k)
    case (4) ! Exponential
      spcor_hh = cor_hh_exp(h,k)
    case (5) ! Gaussian
      spcor_hh = cor_hh_gaussian(h,k)
    end select
  end function spcor_hh

  pure elemental double precision function spcor_hk (h,k)
    implicit none
    double precision, intent(in) :: h,k
    select case (CORIS)
    case (1) ! Matern
      spcor_hk = cor_hk_matern(h,k)
    case (2) ! Spherical
      spcor_hk = cor_hk_spher(h,k)
    case (3) ! Power exponential
      spcor_hk = cor_hk_powexp(h,k)
    case (4) ! Exponential
      spcor_hk = cor_hk_exp(h,k)
    case (5) ! Gaussian
      spcor_hk = cor_hk_gaussian(h,k)
    end select
  end function spcor_hk

  pure elemental double precision function spcor_dhdk (h,k)
    implicit none
    double precision, intent(in) :: h,k
    select case (CORIS)
    case (1) ! Matern
      spcor_dhdk = cor_dhdk_matern(h,k)
    case (2) ! Spherical
      spcor_dhdk = cor_dhdk_spher(h,k)
    case (3) ! Power exponential
      spcor_dhdk = cor_dhdk_powexp(h,k)
    case (4) ! Exponential
      spcor_dhdk = cor_dhdk_exp(h,k)
    case (5) ! Gaussian
      spcor_dhdk = cor_dhdk_gaussian(h,k)
    end select
  end function spcor_dhdk



  subroutine calc_cov (phi,omg,dm,F,betQ0, &
     kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
    implicit none
    integer, intent(in) :: n, p
    double precision, intent(in) :: phi, omg, dm(n,n), &
       F(n,p), betQ0(p,p), kappa
    double precision, intent(out) :: T(n,n), FTF(p,p), Ups(n,n), &
       ldh_Ups, TiF(n,p)
    integer i, j
    double precision omgp1, Tih(n,n), TFFT(n,p), ldh_T, ldh_FTF

    omgp1 = omg + 1d0
    T = dm
    call covmat_l(T,phi,kappa,n,n,lup)
    do i = 1, n
      T(i,i) = omgp1
    end do
    Tih = T
    call oppdf(n,Tih,ldh_T) ! Tih is uppertri s.t. Tih*Tih' = T^{-1}
    TiF = F
    call dtrmm ('l','u','t','n',n,p,1d0,Tih,n,TiF,n) ! TiF = T^{-1/2}*F
    FTF = betQ0
    call dsyrk ('u','t',p,n,1d0,TiF,n,1d0,FTF,p) ! FTF = F'*T^{-1}*F + Q0
    call dtrmm ('l','u','n','n',n,p,1d0,Tih,n,TiF,n) ! TiF = T^{-1}*F
    call oppdf(p,FTF,ldh_FTF) ! FTF is uptri s.t. FTF*FTF' = (F'*T^{-1}*F + Q0)^{-1}
    TFFT = TiF
    call dtrmm ('r','u','n','n',n,p,1d0,FTF,p,TFFT,n) ! TFFT =
    ! T^{-1}*F*(F'*T^{-1}*F + Q0)^{-1/2}
    do i = 1, n
      Ups(1:i,i) = 0d0
      do j = i, 1, -1
        Ups(1:j,j) = Ups(1:j,j) + Tih(1:j,i)*Tih(j,i) ! Ups = T^{-1}
      end do
    end do
    T = Ups ! T is now T^{-1}
    call dsyrk ('u','n',n,p,-1d0,TFFT,n,1d0,Ups,n) ! Ups = T^{-1} -
    ! T^{-1}*F*(F'*T^{-1}*F + Q0)^{-1}*F'*T^{-1}
    ldh_Ups = -ldh_T - ldh_FTF
  end subroutine calc_cov

  subroutine calc_cov_pred (z0_ups, TC, FCTF, &
     phi, omg, dmdm0, F, F0, kappa,  T, n, n0, p)
    implicit none
    integer, intent(in) :: n, n0, p
    double precision, intent(in) :: phi, omg, dmdm0(n,n0), F(n,p), &
       F0(n0,p), kappa, T(n,n)
    double precision, intent(out) :: z0_ups(n0), TC(n,n0), FCTF(n0,p)
    double precision C(n,n0), omgp1

    omgp1 = omg + 1d0
    C = dmdm0
    call covmat_a (C,phi,kappa,n,n0)
    call dsymm ('l','u',n,n0,1d0,T,n,C,n,0d0,TC,n)
    z0_ups = sqrt(omgp1 - sum(TC*C,1))
    FCTF = F0
    call dgemm ('t','n',n0,p,n,-1d0,TC,n,F,n,1d0,FCTF,n0)
  end subroutine calc_cov_pred


  subroutine covlist (kg,philist,omglist,n,p,betQ0,F,dm,kappa, &
     Ulist,ldh_Ulist)
!!! Compute list of relevant covariance matrices

    implicit none

    integer, intent(in) :: kg, n, p
    double precision, intent(in) :: philist(kg), betQ0(p,p), omglist(kg), &
       F(n,p), dm(n,n), kappa
    double precision, intent(out) :: ldh_Ulist(kg), Ulist(n,n,kg)
    integer i
    double precision FTF(p,p), T(n,n), TiF(n,p)

    do i = 1, kg
      call calc_cov(philist(i),omglist(i),dm,F,betQ0,&
         kappa,n,p,T,TiF,FTF,Ulist(:,:,i),ldh_Ulist(i))
    end do
  end subroutine covlist

  subroutine covmat_v (dm,phi,kappa,n)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: phi, kappa
    double precision, intent(inout) :: dm(n)
    call covmat_a (dm,phi,kappa,n,1)
  end subroutine covmat_v

  subroutine covmat_a (dm,phi,kappa,n1,n2)
    implicit none
    integer, intent(in) :: n1, n2
    double precision, intent(in) :: phi, kappa
    double precision, intent(inout) :: dm(n1,n2)
    if (phi .eq. 0d0) then
      where (dm .eq. 0d0)
        dm = 1d0
      elsewhere
        dm = 0d0
      end where
    else if (phi .gt. 0d0) then
      dm = dm/phi
      dm = spcor(dm,kappa)
    else
      call rexit ('covmat - Negative phi')
    end if
  end subroutine covmat_a


  subroutine covmat_l (dm,phi,kappa,n1,n2,ldm)
    implicit none
    integer, intent(in) :: n1, n2
    double precision, intent(in) :: phi, kappa
    double precision, intent(inout) :: dm(n1,n2)
    logical, intent(in) :: ldm(n1,n2)
    if (phi .eq. 0d0) then
      where (ldm .and. (dm .eq. 0d0))
        dm = 1d0
      elsewhere (ldm)
        dm = 0d0
      end where
    else if (phi .gt. 0d0) then
      where (ldm) dm = dm/phi
      where (ldm) dm = spcor(dm,kappa)
    else
      call rexit ('covmat - Negative phi')
    end if
  end subroutine covmat_l


  subroutine oppdf (n,A,ldh)
    ! INPUT:
    ! A - A positive definite matrix
    ! OUTPUT:
    ! A - An upper triangular matrix U s.t. U*U' = A^{-1}
    ! ldh - .5* log det (A) where A is the original (input) matrix
    implicit none
    integer, intent(in) :: n
    double precision, intent(inout) :: A(n,n)
    double precision, intent(out) :: ldh
    integer i
    ! First call cholesky decomposition
    call dpotrf ('u',n,A,n,i)
    if (i .ne. 0) then
      call rexit ("oppdf - Matrix not positive definite")
    end if
    ldh = 0d0
    do i = 1, n
      ldh = ldh + log(A(i,i))
    end do
    ! Next invert
    call dtrtri ('u','n',n,A,n,i)
    if (i .ne. 0) then
      call rexit ("oppdf - Matrix not invertible")
    end if
  end subroutine oppdf

  subroutine fill_symmetric_matrix (A,n)
    ! If A is symmetric, but only the upper triangle is specified, this
    ! routine returns the complete matrix A.
    integer, intent(in) :: n
    double precision, intent(inout) :: A(n,n)
    integer i
    do i = 1, n-1
      A(i+1:,i) = A(i,i+1:)
    end do
  end subroutine fill_symmetric_matrix
end module covfun

!!! mcsp2.f90 ---
!!
!! Author: Evangelos Evangelou
!! Created: Tue, 15 Jul, 2014 16:59 (BST)
!! Last-Updated: Sun, 15 Nov, 2015 15:56 (GMT)
!!     Update #: 250
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine mcspsample (lglk, z, z0, gmu, gmu0, &
   beta, ssq, phi, omg, kappa, acc, y, l, F, offset, F0, offset0, &
   betm0, betQ0, ssqdf, ssqsc, phipars, omgpars, kappapars, &
   phisc, omgsc, kappasc, icf, &
   dft, tsq, dm, dmdm0, nch, Nmc, Nout, Nbi, Nthin, n, n0, p, ifam)

  use mcmcfcns
  use modelfcns
  use covfun, only: create_spcor
  implicit none
  integer, intent(in) :: nch, Nmc(nch), Nout, Nbi, Nthin, n, n0, p, ifam, icf
  double precision, intent(in) :: y(n), l(n), F(n,p), F0(n0,p), offset(n), &
     dft, tsq, dm(n,n), dmdm0(n,n0), betm0(p), betQ0(p,p), ssqdf, offset0(n0), &
     ssqsc, phipars(4), omgpars(4), kappapars(2), phisc, omgsc, kappasc
  double precision, intent(inout) :: z(n,Nout), phi(Nout), omg(Nout), &
     kappa(Nout)
  double precision, intent(out) :: z0(n0,Nout), beta(p,Nout), ssq(Nout), &
     lglk(Nout), gmu(n, Nout), gmu0(n0, Nout)
  integer, intent(out) :: acc(nch)
  integer i, j, k, ii
  double precision, parameter :: tsqdf = 0d0
  logical lnewcov
  double precision zUz, ldh_Ups, modeldfh, ssqdfsc, &
     respdf, tsqdfsc, tsqyy, betQm0(p), zmxi(n), z0_ups(n0), &
     T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), Upsz(n), TC(n,n0), FCTF(n0,p)

  acc = 0
  i = 0

  call create_model(ifam)
  call create_spcor(icf,n)

  do k = 1, nch
    i = i + 1
    
    call ini_mcmc(lglk(i),z(:,i),gmu(:,i),phi(i),omg(i),kappa(i),y,l,F,&
       offset,icf,dm,&
       betm0,betQ0,ssqdf,ssqsc,tsqdf,tsq,dft,n,p,ifam,betQm0,zmxi,T,&
       TiF,FTF,Ups,Upsz,zUz,ldh_Ups,modeldfh,ssqdfsc,respdf,tsqdfsc,tsqyy,&
       lnewcov)
    call rchkusr

    do j = 0, Nbi
      call sample_cov(lglk(i),phi(i),omg(i),kappa(i), &
         phipars,omgpars,kappapars,phisc,omgsc,kappasc,&
         dm,F,betQ0,n,p,acc(k),zmxi,T,TiF,FTF,Ups, &
         Upsz,lnewcov,zUz,ldh_Ups,modeldfh,ssqdfsc)
      call sample_ssq(ssq(i),modeldfh,zUz)
      call sample_z(lglk(i),z(:,i),gmu(:,i),y,l,dft,ssq(i),tsq,zmxi,Ups,&
         Upsz,zUz,modeldfh,n)
    end do
    call sample_beta(beta(:,i),z(:,i),ssq(i),n,p,betQm0,TiF,FTF)
    if (n0 .gt. 0) then
      call sample_z0(z0(:,i),z(:,i),beta(:,i),ssq(i),phi(i),omg(i),&
         n0,n,p,dmdm0,F,offset,F0,offset0,kappa(i),icf,T,z0_ups,TC,FCTF,lnewcov)
      gmu0(:,i) = invlink(z0(:,i), dft)
    end if
    call rchkusr
    do ii = 2, Nmc(k)
      i = i + 1
      lglk(i) = lglk(i-1)
      z(:,i) = z(:,i-1)
      gmu(:,i) = gmu(:,i-1)
      phi(i) = phi(i-1)
      omg(i) = omg(i-1)
      kappa(i) = kappa(i-1)
      do j = 1, Nthin
        call sample_cov(lglk(i),phi(i),omg(i),kappa(i),&
           phipars,omgpars,kappapars,phisc,omgsc,kappasc,&
           dm,F,betQ0,n,p,acc(k),zmxi,T,TiF,FTF,Ups, &
           Upsz,lnewcov,zUz,ldh_Ups,modeldfh,ssqdfsc)
        call sample_ssq(ssq(i),modeldfh,zUz)
        call sample_z(lglk(i),z(:,i),gmu(:,i),y,l,dft,ssq(i),tsq,zmxi,&
           Ups,Upsz,zUz,modeldfh,n)
      end do
      call sample_beta(beta(:,i),z(:,i),ssq(i),n,p,betQm0,TiF,FTF)
      if (n0 .gt. 0) then
        call sample_z0(z0(:,i),z(:,i),beta(:,i),ssq(i),phi(i),omg(i),&
           n0,n,p,dmdm0,F,offset,F0,offset0,kappa(i),icf,T,&
           z0_ups,TC,FCTF,lnewcov)
        gmu0(:,i) = invlink(z0(:,i), dft)
      end if
      call rchkusr
    end do
    call end_mcmc
  end do
end subroutine mcspsample


subroutine samplemulti (lglk, z, gmu, phi, omg, y, l, F, offset, &
   betm0, betQ0, ssqdf, ssqsc, kappa, icf, &
   nu, tsqdf, tsqin, dm, Ntot, Nout, Nbi, Nthin, n, p, kg, ifam)

  use mcmcfcns
  use modelfcns
  use covfun, only: create_spcor
  implicit none
  integer, intent(in) :: n, p, ifam, kg, icf, &
     Nbi(kg), Nthin(kg), Nout(kg), Ntot
  double precision, intent(in) :: y(n), l(n), F(n,p), offset(n), &
     nu(kg), tsqdf, tsqin, dm(n,n), betm0(p), betQ0(p,p), ssqdf, &
     ssqsc, phi(kg), omg(kg), kappa(kg)
  double precision, intent(inout) :: z(n, Ntot)
  double precision, intent(out) :: lglk(Ntot), gmu(n, Ntot)
  double precision :: ssq, tsq
  integer i, ii, j, k
  logical lnewcov
  double precision zUz, ldh_Ups, modeldfh, ssqdfsc, &
     respdf, tsqdfsc, tsqyy, betQm0(p), zmxi(n), &
     T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), Upsz(n)

  i = 0

  call create_model (ifam)
  call create_spcor (icf,n)

  select case (ifam)
  case (0)
    do k = 1, kg
      i = i + 1
      call ini_mcmc(lglk(i),z(:,i),gmu(:,i),phi(k),omg(k),kappa(k),y,l,F,&
         offset,icf,&
         dm,betm0,betQ0,ssqdf,ssqsc,tsqdf,tsqin,nu(k),n,p,ifam,&
         betQm0,zmxi,T,TiF,FTF,Ups,Upsz,zUz,ldh_Ups,modeldfh,ssqdfsc,respdf,&
         tsqdfsc,tsqyy,lnewcov)
      call rchkusr
      do j = 0, Nbi(k)
        call sample_ssq(ssq,modeldfh,zUz)
        call sample_tsq(tsq,respdf,tsqyy)
        call samplez_gt(lglk(i),z(:,i),gmu(:,i),y,l,nu(k),ssq,zmxi,Ups,&
           Upsz,zUz,modeldfh,respdf,tsqyy,n)
      end do
      call rchkusr
      do ii = 2, Nout(k)
        i = i + 1
        lglk(i) = lglk(i-1)
        z(:,i) = z(:,i-1)
        gmu(:,i) = gmu(:,i-1)
        do j = 1, Nthin(k)
          call sample_ssq(ssq,modeldfh,zUz)
          call sample_tsq(tsq,respdf,tsqyy)
          call samplez_gt(lglk(i),z(:,i),gmu(:,i),y,l,nu(k),ssq,zmxi,&
             Ups,Upsz,zUz,modeldfh,respdf,tsqyy,n)
        end do
        call rchkusr
      end do
      call end_mcmc
    end do
  case default
    tsq = tsqin
    do k = 1, kg
      i = i + 1
      call ini_mcmc(lglk(i),z(:,i),gmu(:,i),phi(k),omg(k),kappa(k),y,l,F,&
         offset,icf,&
         dm,betm0,betQ0,ssqdf,ssqsc,tsqdf,tsqin,nu(k),n,p,ifam,&
         betQm0,zmxi,T,TiF,FTF,Ups,Upsz,zUz,ldh_Ups,modeldfh,ssqdfsc,respdf,&
         tsqdfsc,tsqyy,lnewcov)
      call rchkusr
      call create_model(ifam)
      do j = 0, Nbi(k)
        call sample_ssq(ssq,modeldfh,zUz)
        call sample_z(lglk(i),z(:,i),gmu(:,i),y,l,nu(k),ssq,tsq,zmxi,Ups,&
           Upsz,zUz,modeldfh,n)
      end do
      call rchkusr
      do ii = 2, Nout(k)
        i = i + 1
        lglk(i) = lglk(i-1)
        z(:,i) = z(:,i-1)
        gmu(:,i) = gmu(:,i-1)
        do j = 1, Nthin(k)
          call sample_ssq(ssq,modeldfh,zUz)
          call sample_z(lglk(i),z(:,i),gmu(:,i),y,l,nu(k),ssq,tsq,zmxi,&
             Ups,Upsz,zUz,modeldfh,n)
        end do
        call rchkusr
      end do
      call end_mcmc
    end do
  end select
end subroutine samplemulti


subroutine mcspsamtry (lglk, z, phi, omg, kappa, acc, y, l, F, offset, &
   betm0, betQ0, ssqdf, ssqsc, phipars, omgpars, kappapars, &
   phisc, omgsc, kappasc, &
   icf, dft, tsq, dm, Nout, Npr, n, p, ifam)

  use mcmcfcns
  use modelfcns
  use msgmc
  use covfun, only: create_spcor
  implicit none
  integer, intent(in) :: Nout, n, p, ifam, Npr, icf
  double precision, intent(in) :: y(n), l(n), F(n,p), offset(n), &
     dft, tsq, dm(n,n), betm0(p), betQ0(p,p), ssqdf, &
     ssqsc, phipars(4), omgpars(4), kappapars(2), phisc, omgsc, kappasc
  double precision, intent(inout) :: z(n,Nout), phi(Nout), omg(Nout), &
     kappa(Nout)
  double precision, intent(out) :: lglk(Nout)
  integer, intent(out) :: acc
  integer i, ia
  double precision lglk1, ssq, phi1, omg1, kappa1, z1(n), gmu(n)
!   integer, parameter :: nl = 19
!   character(len=nl) :: label
  integer iap
  double precision :: tsqdf
  logical lnewcov
  double precision zUz, ldh_Ups, modeldfh, ssqdfsc, &
     respdf, tsqdfsc, tsqyy, betQm0(p), zmxi(n), &
     T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), Upsz(n)

  tsqdf = 0d0

!   write (label,'("Iteration",2x,"Acc% cov")')
!   call intpr (label,nl,0,0)
  call msgmca
!   call intpr ("-----------------------------------------------",nl,0,0)
  call msgmcl
  acc = 0
  ia = 0
  i = 0
  z1 = z(:,1)
  phi1 = phi(1)
  omg1 = omg(1)
  kappa1 = kappa(1)

  call create_model (ifam)
  call create_spcor (icf,n)

  call ini_mcmc(lglk1,z1,gmu,phi1,omg1,kappa1,y,l,F,offset,icf,&
     dm,betm0,betQ0,ssqdf,ssqsc,tsqdf,tsq,dft,n,p,ifam,&
     betQm0,zmxi,T,TiF,FTF,Ups,Upsz,zUz,ldh_Ups,modeldfh,ssqdfsc,respdf,&
     tsqdfsc,tsqyy,lnewcov)
  call rchkusr
  do i = 1, Nout
    call sample_cov(lglk1,phi1,omg1,kappa1,&
       phipars,omgpars,kappapars,phisc,omgsc,kappasc,dm,&
       F,betQ0,n,p,ia,zmxi,T,TiF,FTF,Ups, &
       Upsz,lnewcov,zUz,ldh_Ups,modeldfh,ssqdfsc)
    call sample_ssq(ssq,modeldfh,zUz)
    call sample_z(lglk1,z1,gmu,y,l,dft,ssq,tsq,zmxi,Ups,Upsz,zUz,modeldfh,n)
    lglk(i) = lglk1
    z(:,i) = z1
    phi(i) = phi1
    omg(i) = omg1
    kappa(i) = kappa1
    if (Npr .gt. 0 .and. mod(i,Npr) .eq. 0) then
      iap = 100*ia/Npr
      !         write (label,'(i9,2x,i8)') i, iap
      !         call intpr (label,nl,0,0)
      call msgmci (i, iap)
      acc = acc + ia
      ia = 0
      call rchkusr
    end if
  end do

  acc = acc + ia
  call end_mcmc
!   call intpr ("-----------------------------------------------",nl,0,0)
  call msgmcl
  iap = 100*acc/Nout
!   write (label,'("Avg acc %",2x,i8)') iap
!   call intpr (label,nl,0,0)
  call msgmce (iap)
!   call intpr ("-----------------------------------------------",nl,0,0)
  call msgmcl
end subroutine mcspsamtry


subroutine trgasample (lglk, z, z0, gmu, gmu0, &
   beta, ssq, tsq, phi, omg, kappa, acc, y, l, F, offset,&
   F0, offset0, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsqsc, phipars, omgpars, &
   kappapars, phisc, omgsc, kappasc, icf, dft, dm, dmdm0, nch, &
   Nmc, Nout, Nbi, Nthin, n, n0, p)

  use modelfcns
  use mcmcfcns
  use linkfcns, only: invlink_ga
  use covfun, only: create_spcor
  implicit none
  integer, intent(in) :: nch, Nmc(nch), Nout, Nbi, Nthin, n, n0, p, icf
  double precision, intent(in) :: y(n), l(n), F(n,p), F0(n0,p), offset(n), &
     dft, dm(n,n), dmdm0(n,n0), betm0(p), betQ0(p,p), ssqdf, offset0(n0), &
     ssqsc, tsqdf, tsqsc, phipars(4), omgpars(4), kappapars(2), &
     phisc, omgsc, kappasc
  double precision, intent(inout) :: z(n,Nout), phi(Nout), omg(Nout), &
     kappa(Nout)
  double precision, intent(out) :: z0(n0,Nout), beta(p,Nout), ssq(Nout), &
     tsq(Nout), lglk(Nout), gmu(n,Nout), gmu0(n0,Nout)
  integer, intent(out) :: acc(nch)
  integer i, j, k, ii
  integer, parameter :: ifam = 0
  logical lnewcov
  double precision zUz, ldh_Ups, modeldfh, ssqdfsc, &
     respdf, tsqdfsc, tsqyy, betQm0(p), zmxi(n), z0_ups(n0), &
     T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), Upsz(n), TC(n,n0), FCTF(n0,p)

  acc = 0
  i = 0

  call create_model(ifam)
  call create_spcor(icf,n)

  do k = 1, Nch
    i = i + 1
    
    call ini_mcmc(lglk(i),z(:,i),gmu(:,i),phi(i),omg(i),kappa(i),y,l,F, &
       offset,icf,&
       dm,betm0,betQ0,ssqdf,ssqsc,tsqdf,tsqsc,dft,n,p,ifam,&
       betQm0,zmxi,T,TiF,FTF,Ups,Upsz,zUz,ldh_Ups,modeldfh,ssqdfsc,respdf,&
       tsqdfsc,tsqyy,lnewcov)
    call rchkusr

    do j = 0, Nbi
      call sample_cov(lglk(i),phi(i),omg(i),kappa(i),&
         phipars,omgpars,kappapars,phisc,omgsc,kappasc,&
         dm,F,betQ0,n,p,acc(k),zmxi,T,TiF,FTF,Ups, &
         Upsz,lnewcov,zUz,ldh_Ups,modeldfh,ssqdfsc)
      call sample_ssq(ssq(i),modeldfh,zUz)
      call sample_tsq(tsq(i),respdf,tsqyy)
      call samplez_gt(lglk(i),z(:,i),gmu(:,i),y,l,dft,ssq(i),zmxi,Ups,&
         Upsz,zUz,modeldfh,respdf,tsqyy,n)
    end do
    call sample_beta(beta(:,i),z(:,i),ssq(i),n,p,betQm0,TiF,FTF)
    if (n0 .gt. 0) then
      call sample_z0(z0(:,i),z(:,i),beta(:,i),ssq(i),phi(i),omg(i),&
         n0,n,p,dmdm0,F,offset,F0,offset0,kappa(i),icf,T,z0_ups,TC,FCTF,lnewcov)
      gmu0(:,i) = invlink_ga(z0(:,i), dft)
    end if
    call rchkusr
    do ii = 2, Nmc(k)
      i = i + 1
      lglk(i) = lglk(i-1)
      z(:,i) = z(:,i-1)
      gmu(:,i) = gmu(:,i-1)
      phi(i) = phi(i-1)
      omg(i) = omg(i-1)
      kappa(i) = kappa(i-1)
      do j = 1, Nthin
        call sample_cov(lglk(i),phi(i),omg(i),kappa(i),&
           phipars,omgpars,kappapars,phisc,omgsc,kappasc,&
           dm,F,betQ0,n,p,acc(k),zmxi,T,TiF,FTF,Ups, &
           Upsz,lnewcov,zUz,ldh_Ups,modeldfh,ssqdfsc)
        call sample_ssq(ssq(i),modeldfh,zUz)
        call sample_tsq(tsq(i),respdf,tsqyy)
        call samplez_gt(lglk(i),z(:,i),gmu(:,i),y,l,dft,ssq(i),zmxi,&
           Ups,Upsz,zUz,modeldfh,respdf,tsqyy,n)
      end do
      call sample_beta(beta(:,i),z(:,i),ssq(i),n,p,betQm0,TiF,FTF)
      if (n0 .gt. 0) then
        call sample_z0(z0(:,i),z(:,i),beta(:,i),ssq(i),phi(i),omg(i),&
           n0,n,p,dmdm0,F,offset,F0,offset0,kappa(i),icf,T,&
           z0_ups,TC,FCTF,lnewcov)
        gmu0(:,i) = invlink_ga(z0(:,i), dft)
      end if
      call rchkusr
    end do

    call end_mcmc
  end do
end subroutine trgasample


subroutine trgasamtry (lglk, z, phi, omg, kappa, acc, y, l, F, offset, &
   betm0, betQ0, ssqdf, ssqsc, tsqdf, tsqsc, phipars, omgpars, kappapars, &
   phisc, omgsc, kappasc, icf, dft, dm, Nout, Npr, n, p)

  use modelfcns
  use mcmcfcns
  use msgmc
  use covfun, only: create_spcor
  implicit none
  integer, intent(in) :: Nout, n, p, Npr, icf
  double precision, intent(in) :: y(n), l(n), F(n,p), offset(n), &
     dft, dm(n,n), betm0(p), betQ0(p,p), ssqdf, &
     ssqsc, tsqdf, tsqsc, phipars(4), omgpars(4), kappapars(2), &
     phisc, omgsc, kappasc
  double precision, intent(inout) :: z(n,Nout), phi(Nout), omg(Nout), &
     kappa(Nout)
  double precision, intent(out) :: lglk(Nout)
  integer, intent(out) :: acc
  integer i, ia
  double precision ssq, tsq, phi1, omg1, z1(n), lglk1, gmu(n), kappa1
!   integer, parameter :: nl = 19
!   character(len=nl) :: label
  integer iap
  integer, parameter :: ifam = 0
  logical lnewcov
  double precision zUz, ldh_Ups, modeldfh, ssqdfsc, &
     respdf, tsqdfsc, tsqyy, betQm0(p), zmxi(n), &
     T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), Upsz(n)

!   write (label,'("Iteration",2x,"Acc% cov")')
!   call intpr (label,nl,0,0)
  call msgmca
!   call intpr ("-----------------------------------------------",nl,0,0)
  call msgmcl
  acc = 0
  ia = 0
  i = 0
  z1 = z(:,1)
  phi1 = phi(1)
  omg1 = omg(1)
  kappa1 = kappa(1)

  call create_model(ifam)
  call create_spcor(icf,n)

  call ini_mcmc(lglk1,z1,gmu,phi1,omg1,kappa1,y,l,F,offset,icf,dm,&
     betm0,betQ0,ssqdf,ssqsc,tsqdf,tsqsc,dft,n,p,ifam,betQm0,zmxi,T,&
     TiF,FTF,Ups,Upsz,zUz,ldh_Ups,modeldfh,ssqdfsc,respdf,tsqdfsc,tsqyy,&
     lnewcov)
  call rchkusr

  do i = 1, Nout
    call sample_cov(lglk1,phi1,omg1,kappa1,&
       phipars,omgpars,kappapars,phisc,omgsc,kappasc,dm,&
       F,betQ0,n,p,ia,zmxi,T,TiF,FTF,Ups, &
         Upsz,lnewcov,zUz,ldh_Ups,modeldfh,ssqdfsc)
    call sample_ssq(ssq,modeldfh,zUz)
    call sample_tsq(tsq,respdf,tsqyy)
    call samplez_gt(lglk1,z1,gmu,y,l,dft,ssq,zmxi,Ups,Upsz,zUz,modeldfh,&
       respdf,tsqyy,n)
    lglk(i) = lglk1
    z(:,i) = z1
    phi(i) = phi1
    omg(i) = omg1
    kappa(i) = kappa1
    if (Npr .gt. 0 .and. mod(i,Npr) .eq. 0) then
      iap = 100*ia/Npr
!       write (label,'(i9,2x,i8)') i, iap
!       call intpr (label,nl,0,0)
      call msgmci (i, iap)
      acc = acc + ia
      ia = 0
      call rchkusr
    end if
  end do
  acc = acc + ia
  call end_mcmc
!   call intpr ("-----------------------------------------------",nl,0,0)
  call msgmcl
  iap = 100*acc/Nout
!   write (label,'("Avg acc %",2x,i8)') iap
!   call intpr (label,nl,0,0)
  call msgmce (iap)
!   call intpr ("-----------------------------------------------",nl,0,0)
  call msgmcl
end subroutine trgasamtry

#' Models used in the geoBayes package
#'
#' This hidden variable contains a choice of models that can be fit
#' with this package. The model can be chosen in the \code{family}
#' input of the relevant function. This variable cannot be
#' overwritten.
#' @name geoBayes_models
#' @export
.geoBayes_models <-
  data.frame(
    family = c("gaussian",            # 1
               "binomial.robit",      # 2
               "binomial.logit",      # 3
               "binomial.probit",     # 4
               "binomial.Wallace",    # 5
               "poisson.modifiedbc",  # 6
               "poisson.boxcox",      # 7
               "Gamma.modifiedbc",    # 8
               "Gamma.boxcox",        # 9
               "binomial.modifiedGEV",# 10
               "binomial.modifiedGEVns",# 11
               "binomial.GEV"         # 12
               ),
    needlinkp = c(TRUE,               #  1 "gaussian",
                  TRUE,               #  2 "binomial.robit",
                  FALSE,              #  3 "binomial.logit",
                  FALSE,              #  4 "binomial.probit",
                  TRUE,               #  5 "binomial.Wallace",
                  TRUE,               #  6 "poisson.modifiedbc",
                  TRUE,               #  7 "poisson.boxcox",
                  TRUE,               #  8 "Gamma.modifiedbc",
                  TRUE,               #  9 "Gamma.boxcox"
                  TRUE,               # 10 "binomial.modifiedGEV"
                  TRUE,               # 11 "binomial.modifiedGEVns"
                  TRUE                # 12 "binomial.GEV"
                  ),
    linkplb = c(-Inf,                 #  1 "gaussian",
                0,                    #  2 "binomial.robit",
                NA,                   #  3 "binomial.logit",
                NA,                   #  4 "binomial.probit",
                0,                    #  5 "binomial.Wallace",
                0,                    #  6 "poisson.modifiedbc",
                -Inf,                 #  7 "poisson.boxcox",
                0,                    #  8 "Gamma.modifiedbc",
                -Inf,                 #  9 "Gamma.boxcox"
                0,                    # 10 "binomial.modifiedGEV"
                0,                    # 11 "binomial.modifiedGEVns"
                -Inf                  # 12 "binomial.GEV"
                ),
    incllb = c(NA,                    #  1 "gaussian",
               FALSE,                 #  2 "binomial.robit",
               NA,                    #  3 "binomial.logit",
               NA,                    #  4 "binomial.probit",
               FALSE,                 #  5 "binomial.Wallace",
               TRUE,                  #  6 "poisson.modifiedbc",
               NA,                    #  7 "poisson.boxcox",
               TRUE,                  #  8 "Gamma.modifiedbc",
               NA,                    #  9 "Gamma.boxcox"
               TRUE,                  # 10 "binomial.modifiedGEV"
               TRUE,                  # 11 "binomial.modifiedGEVns"
               NA                     # 12 "binomial.GEV"
               ),
    linkpub = c(Inf,                  #  1 "gaussian",
                Inf,                  #  2 "binomial.robit",
                NA,                   #  3 "binomial.logit",
                NA,                   #  4 "binomial.probit",
                Inf,                  #  5 "binomial.Wallace",
                Inf,                  #  6 "poisson.modifiedbc",
                Inf,                  #  7 "poisson.boxcox",
                Inf,                  #  8 "Gamma.modifiedbc",
                Inf,                  #  9 "Gamma.boxcox"
                Inf,                  # 10 "binomial.modifiedGEV"
                Inf,                  # 11 "binomial.modifiedGEVns"
                Inf                   # 12 "binomial.GEV"
                ),
    inclub = c(NA,                    #  1 "gaussian",
               NA,                    #  2 "binomial.robit",
               NA,                    #  3 "binomial.logit",
               NA,                    #  4 "binomial.probit",
               NA,                    #  5 "binomial.Wallace",
               NA,                    #  6 "poisson.modifiedbc",
               NA,                    #  7 "poisson.boxcox",
               NA,                    #  8 "Gamma.modifiedbc",
               NA,                    #  9 "Gamma.boxcox"
               NA,                    # 10 "binomial.modifiedGEV"
               NA,                    # 11 "binomial.modifiedGEVns"
               NA                     # 12 "binomial.GEV"
               ),
    haswo = c(FALSE,                  #  1 "gaussian",
              TRUE,                   #  2 "binomial.robit",
              FALSE,                  #  3 "binomial.logit",
              FALSE,                  #  4 "binomial.probit",
              FALSE,                  #  5 "binomial.Wallace",
              FALSE,                  #  6 "poisson.modifiedbc",
              TRUE,                   #  7 "poisson.boxcox",
              FALSE,                  #  8 "Gamma.modifiedbc",
              FALSE,                  #  9 "Gamma.boxcox"
              FALSE,                  # 10 "binomial.modifiedGEV"
              FALSE,                  # 11 "binomial.modifiedGEVns"
              TRUE                    # 12 "binomial.GEV"
              ),
    stringsAsFactors = FALSE)

lockBinding(".geoBayes_models", environment())

.geoBayes_family <- function (family) {
  ## family is a character, partially matching .geoBayes_models before
  ## and after the dot.
  nfamily <- length(family)
  if (nfamily > 1) {
    stop ("You cannot use more than one family.")
  } else if (nfamily == 0) {
    as.integer(1L)
  } else if (pmatch(family, "transformed.gaussian", 0)) {
    as.integer(0L)
  } else if (is.numeric(family) && family >= 0 &&
             family <= length(.geoBayes_models$family)) {
    as.integer(family)
  } else if (is.character(family)) {
    fl <- strsplit(family, split = "[[:punct:][:space:]]+")[[1]]
    pattern <- if(length(fl) > 1)
                 paste0("\\.", fl[-1], "[[:alnum:]]*", collapse = "")
    pattern <- paste0("^", fl[1], "[[:alnum:]]*", pattern)
    out <- grep(pattern, .geoBayes_models$family, value = FALSE)
    if (length(out) == 0) stop ("Cannot deduce family from input.")
    as.integer(out[1])
  } else {
    stop ("Cannot deduce family.")
  }
}



## Checks whether the link parameter is consistent with the family and
## returns a real representation of the link parameter.
.geoBayes_getlinkp <- function (linkp, ifam) {
  lbt <- `>`; ubt <- `<`
  if (ifam == 0 || pmatch(ifam, "transformed.gaussian", 0)) {
    family <- "transformed.gaussian"
    lb <- 0; ub <- Inf; needlinkp <- TRUE
  } else {
    if (!is.numeric(ifam)) ifam <- .geoBayes_family(ifam)
    family <- .geoBayes_models$family[ifam]
    lb <- .geoBayes_models$linkplb[ifam]
    ub <- .geoBayes_models$linkpub[ifam]
    needlinkp <- .geoBayes_models$needlinkp[ifam]
    if (needlinkp && is.finite(lb) && .geoBayes_models$incllb[ifam])
      lbt <- `>=`
    if (needlinkp && is.finite(ub) && .geoBayes_models$inclub[ifam])
      ubt <- `<=`
  }
  if (!needlinkp) return(0.)
  if (is.finite(lb) && !all(lbt(linkp, lb))) {
    stop (paste("Link parameter must be", gsub("[^><=]","",deparse(lbt)), lb,
                "for the", family, "family"))
  }
  if (is.finite(ub) && !all(ubt(linkp, ub))) {
    stop (paste("Link parameter must be", gsub("[^><=]","",deparse(ubt)), ub,
                "for the", family, "family"))
  }
  as.double(linkp)
}

###   valid <- .geoBayes_models_opts$linkp[[family]]
###   for (i in seq_along(valid)) {
###     test <- valid[[i]]
###     nmtest <- names(valid[i])
###     if (length(nmtest) && nmtest != "") { # Character value
###       if (all(as.character(linkp) == nmtest))
###         return (rep(as.double(test), k))
###     } else if (length(test) == 2) { # Interval value
###       if (all(linkp > test[1]) && all(linkp < test[2]))
###         return (as.double(linkp))
###     } else if (length(test) == 1) { # Scalar value
###       return (rep(as.double(test), k))
###     }
###   }
###   stop ("Cannot deduce link parameter from family.")

!! This module contains functions specific to each model. The functions
!! are pointers to model-specific functions.

module modelfcns

  integer, private, parameter :: &                 ! Models
     MODELCODES(15) = (/1,2,-2,3,4,5,6,7,-7,8,9,10,11,12,-12/)
  logical, private :: MODELDEF = .false.
  integer, private :: MODELIS = 0

contains

  subroutine create_model (ifam)
    integer, intent(in) :: ifam
    if (MODELDEF .and. (MODELIS .eq. ifam)) return ! Model is already defined
    if (.not. ((ifam .eq. 0) .or. any(ifam .eq. MODELCODES))) &
       call rexit ("Unrecognised family.")
    MODELIS = ifam
    MODELDEF = .true.
  end subroutine create_model

  pure elemental double precision function logitrwhz (z,d)
    double precision, intent(in) :: z, d
    double precision, parameter :: bigneg = -huge(1d0)
    logitrwhz = invtrwhz(z, d)
    if (logitrwhz .gt. 0d0) then
      logitrwhz = log(logitrwhz)
    else
      logitrwhz = bigneg
    end if
  end function logitrwhz

  function logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
!! log-pdf of z after integrating out beta and ssq.
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, modeldfh, xi(n)
    double precision logpdfz
    double precision Upsz(n), zUz, zmxi(n)
    if (lmxi) then
      zmxi = z - xi
      call dsymv ('u',n,1d0,Ups,n,zmxi,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
      zUz = dot_product(zmxi,Upsz) + ssqdfsc
    else
      call dsymv ('u',n,1d0,Ups,n,z,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
      zUz = dot_product(z,Upsz) + ssqdfsc
    end if
    logpdfz = ldh_Ups - modeldfh*log(zUz)
  end function logpdfz

  function logpdfz_dz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, modeldfh, xi(n)
    double precision logpdfz_dz(n)
    double precision Upsz(n), zUz, zmxi(n)
    if (lmxi) then
      zmxi = z - xi
      call dsymv ('u',n,1d0,Ups,n,zmxi,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
      zUz = dot_product(zmxi,Upsz) + ssqdfsc
    else
      call dsymv ('u',n,1d0,Ups,n,z,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
      zUz = dot_product(z,Upsz) + ssqdfsc
    end if
    logpdfz_dz = -(2*modeldfh/zUz)*Upsz
  end function logpdfz_dz

  function logpdfz_hz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), Ups(n,n), &
       ldh_Ups, ssqdfsc, modeldfh, xi(n)
    double precision logpdfz_hz(n,n)
    double precision Upsz(n), zUz, zmxi(n), UzzU(n,n), alpha
    if (lmxi) then
      zmxi = z - xi
    else
      zmxi = z
    end if
    call dsymv ('u',n,1d0,Ups,n,zmxi,1,0d0,Upsz,1) ! Upsz = Ups*(z-xi)
    zUz = dot_product(zmxi,Upsz) + ssqdfsc
    alpha = -2d0/zUz
    UzzU = Ups
    call dsyr('u',n,alpha,Upsz,1,UzzU,n)
    logpdfz_hz = alpha*modeldfh*UzzU
  end function logpdfz_hz

  function logpdfmu (n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, modeldfh
    double precision logpdfmu
    integer i
    double precision z(n)
    ! Linear predictor
    do i = 1, n
      z(i) = flink(mu(i), nu)
    end do
    ! log-likelihood for z
    logpdfmu = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    ! Jacobian
    do i = 1, n
      logpdfmu = logpdfmu - loginvlinkdz(z(i),nu)
    end do
  end function logpdfmu


  function condyz (n, y1, y2, z, nu, tsq)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), z(n), nu, tsq
    double precision condyz
    integer i
    double precision w, lfy
    lfy = 0d0
    do i = 1, n
      w = invlink(z(i),nu)
      lfy = lfy + logpdfy(y1(i), y2(i), w)
    end do
    condyz = lfy/tsq
  end function condyz

  function jointyz (n, z, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: z(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, ssqdfsc, tsq, nu, modeldfh, xi(n)
    double precision jointyz
    double precision lfz, lfy
    lfz = logpdfz(n, z, Ups, ldh_Ups, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condyz(n, y, l, z, nu, tsq)
    jointyz = lfz + lfy
  end function jointyz

  pure function condymu (n, y1, y2, mu, tsq)
    implicit none
    integer, intent(in) :: n
    double precision, intent(in) :: y1(n), y2(n), mu(n), tsq
    double precision condymu
    integer i
    double precision lfy
    lfy = 0d0
    do i = 1, n
      lfy = lfy + logpdfy(y1(i), y2(i), mu(i))
    end do
    condymu = lfy/tsq
  end function condymu

  function jointymu (n, mu, y, l, Ups, ldh_Ups, &
     nu, xi, lmxi, ssqdfsc, tsq, modeldfh)
    implicit none
    logical, intent(in) :: lmxi
    integer, intent(in) :: n
    double precision, intent(in) :: mu(n), y(n), l(n), Ups(n, n), &
       ldh_Ups, nu, xi(n), ssqdfsc, tsq, modeldfh
    double precision jointymu
    double precision lfmu, lfy
    lfmu = logpdfmu(n, mu, Ups, ldh_Ups, nu, xi, lmxi, ssqdfsc, modeldfh)
    lfy = condymu(n, y, l, mu, tsq)
    jointymu = lfmu + lfy
  end function jointymu


  ! Distribution functions

  pure elemental double precision function logpdfy (y1, y2, par)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: y1, y2, par
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      logpdfy = logpdfy_gt(y1,y2,par)
    case (1) ! Gaussian response
      logpdfy = logpdfy_ga(y1,y2,par)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      logpdfy = logpdfy_bi(y1,y2,par)
    case (6,7,-7) ! Poisson response
      logpdfy = logpdfy_po(y1,y2,par)
    case (8,9) ! Gamma repsponse
      logpdfy = logpdfy_gm(y1,y2,par)
    end select
  end function logpdfy

  pure elemental double precision function logdffy (y1, y2, p1, p2)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: y1, y2, p1, p2
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      logdffy = logdffy_gt(y1,y2,p1,p2)
    case (1) ! Gaussian response
      logdffy = logdffy_ga(y1,y2,p1,p2)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      logdffy = logdffy_bi(y1,y2,p1,p2)
    case (6,7,-7) ! Poisson response
      logdffy = logdffy_po(y1,y2,p1,p2)
    case (8,9) ! Gamma repsponse
      logdffy = logdffy_gm(y1,y2,p1,p2)
    end select
  end function logdffy

  pure elemental double precision function logpdfydlnk (y1, y2, par)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: y1, y2, par
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      logpdfydlnk = logpdfydlnk_gt(y1,y2,par)
    case (1) ! Gaussian response
      logpdfydlnk = logpdfydlnk_ga(y1,y2,par)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      logpdfydlnk = logpdfydlnk_bi(y1,y2,par)
    case (6,7,-7) ! Poisson response
      logpdfydlnk = logpdfydlnk_po(y1,y2,par)
    case (8,9) ! Gamma repsponse
      logpdfydlnk = logpdfydlnk_gm(y1,y2,par)
    end select
  end function logpdfydlnk

  pure elemental double precision function logpdfyhlnk (y1, y2, par)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: y1, y2, par
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      logpdfyhlnk = logpdfyhlnk_gt(y1,y2,par)
    case (1) ! Gaussian response
      logpdfyhlnk = logpdfyhlnk_ga(y1,y2,par)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      logpdfyhlnk = logpdfyhlnk_bi(y1,y2,par)
    case (6,7,-7) ! Poisson response
      logpdfyhlnk = logpdfyhlnk_po(y1,y2,par)
    case (8,9) ! Gamma repsponse
      logpdfyhlnk = logpdfyhlnk_gm(y1,y2,par)
    end select
  end function logpdfyhlnk

  pure elemental double precision function logpdfy3lnk (y1, y2, par)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: y1, y2, par
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      logpdfy3lnk = logpdfy3lnk_gt(y1,y2,par)
    case (1) ! Gaussian response
      logpdfy3lnk = logpdfy3lnk_ga(y1,y2,par)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      logpdfy3lnk = logpdfy3lnk_bi(y1,y2,par)
    case (6,7,-7) ! Poisson response
      logpdfy3lnk = logpdfy3lnk_po(y1,y2,par)
    case (8,9) ! Gamma repsponse
      logpdfy3lnk = logpdfy3lnk_gm(y1,y2,par)
    end select
  end function logpdfy3lnk

  pure elemental double precision function mustart (y1, y2)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: y1, y2
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      mustart = mustart_gt(y1,y2)
    case (1) ! Gaussian response
      mustart = mustart_ga(y1,y2)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      mustart = mustart_bi(y1,y2)
    case (6,7,-7) ! Poisson response
      mustart = mustart_po(y1,y2)
    case (8,9) ! Gamma repsponse
      mustart = mustart_gm(y1,y2)
    end select
  end function mustart

  pure elemental double precision function fcntruemu (x)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: x
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      fcntruemu = x
    case (1) ! Gaussian response
      fcntruemu = x
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      fcntruemu = exp(x)
    case (6,7,-7) ! Poisson response
      fcntruemu = exp(x)
    case (8,9) ! Gamma repsponse
      fcntruemu = exp(x)
    end select
  end function fcntruemu

  pure elemental double precision function invtruemu (x)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: x
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtruemu = x
    case (1) ! Gaussian response
      invtruemu = x
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      invtruemu = log(x)
    case (6,7,-7) ! Poisson response
      invtruemu = log(x)
    case (8,9) ! Gamma repsponse
      invtruemu = log(x)
    end select
  end function invtruemu

  pure elemental double precision function fcncum (x)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: x
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      fcncum = fcncum_gt(x)
    case (1) ! Gaussian response
      fcncum = fcncum_ga(x)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      fcncum = fcncum_bi(x)
    case (6,7,-7) ! Poisson response
      fcncum = fcncum_po(x)
    case (8,9) ! Gamma repsponse
      fcncum = fcncum_gm(x)
    end select
  end function fcncum

  pure elemental double precision function fcncumd2 (x)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: x
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      fcncumd2 = fcncumd2_gt(x)
    case (1) ! Gaussian response
      fcncumd2 = fcncumd2_ga(x)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      fcncumd2 = fcncumd2_bi(x)
    case (6,7,-7) ! Poisson response
      fcncumd2 = fcncumd2_po(x)
    case (8,9) ! Gamma repsponse
      fcncumd2 = fcncumd2_gm(x)
    end select
  end function fcncumd2

  pure elemental double precision function fcncumd3 (x)
    use modelfcns_pdfy
    implicit none
    double precision, intent(in) :: x
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      fcncumd3 = fcncumd3_gt(x)
    case (1) ! Gaussian response
      fcncumd3 = fcncumd3_ga(x)
    case (2,-2,3,4,5,10,11,-12,12) ! Binomial response
      fcncumd3 = fcncumd3_bi(x)
    case (6,7,-7) ! Poisson response
      fcncumd3 = fcncumd3_po(x)
    case (8,9) ! Gamma repsponse
      fcncumd3 = fcncumd3_gm(x)
    end select
  end function fcncumd3


  ! Link functions

  pure elemental double precision function flink (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      flink = flink_ga(w,d)
    case (1) ! Gaussian boxcox
      flink = flink_ga(w,d)
    case (2,-2) ! Robit
      flink = flink_robit(w,d)
    case (3) ! Logit
      flink = flink_logit(w,d)
    case (4) ! Probit
      flink = flink_probit(w,d)
    case (5) ! Wallace
      flink = flink_wallace(w,d)
    case (6,8) ! Modified boxcox
      flink = flink_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      flink = flink_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      flink = flink_modgev(w,d)
    case (12,-12) ! Binomial GEV
      flink = flink_gev(w,d)
    case (11) ! Binomial modifiedGEV
      flink = flink_modgevns(w,d)
    end select
  end function flink

  pure elemental double precision function invlink (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlink = invlink_ga(w,d)
    case (1) ! Gaussian boxcox
      invlink = invlink_ga(w,d)
    case (2,-2) ! Robit
      invlink = invlink_robit(w,d)
    case (3) ! Logit
      invlink = invlink_logit(w,d)
    case (4) ! Probit
      invlink = invlink_probit(w,d)
    case (5) ! Wallace
      invlink = invlink_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlink = invlink_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlink = invlink_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlink = invlink_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlink = invlink_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlink = invlink_modgevns(w,d)
    end select
  end function invlink

  pure elemental double precision function invlinkdz (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlinkdz = invlinkdz_ga(w,d)
    case (1) ! Gaussian boxcox
      invlinkdz = invlinkdz_ga(w,d)
    case (2,-2) ! Robit
      invlinkdz = invlinkdz_robit(w,d)
    case (3) ! Logit
      invlinkdz = invlinkdz_logit(w,d)
    case (4) ! Probit
      invlinkdz = invlinkdz_probit(w,d)
    case (5) ! Wallace
      invlinkdz = invlinkdz_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlinkdz = invlinkdz_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlinkdz = invlinkdz_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlinkdz = invlinkdz_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlinkdz = invlinkdz_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlinkdz = invlinkdz_modgevns(w,d)
    end select
  end function invlinkdz

  pure elemental double precision function loginvlinkdz (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      loginvlinkdz = loginvlinkdz_ga(w,d)
    case (1) ! Gaussian boxcox
      loginvlinkdz = loginvlinkdz_ga(w,d)
    case (2,-2) ! Robit
      loginvlinkdz = loginvlinkdz_robit(w,d)
    case (3) ! Logit
      loginvlinkdz = loginvlinkdz_logit(w,d)
    case (4) ! Probit
      loginvlinkdz = loginvlinkdz_probit(w,d)
    case (5) ! Wallace
      loginvlinkdz = loginvlinkdz_wallace(w,d)
    case (6,8) ! Modified boxcox
      loginvlinkdz = loginvlinkdz_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      loginvlinkdz = loginvlinkdz_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      loginvlinkdz = loginvlinkdz_modgev(w,d)
    case (12,-12) ! Binomial GEV
      loginvlinkdz = loginvlinkdz_gev(w,d)
    case (11) ! Binomial modifiedGEV
      loginvlinkdz = loginvlinkdz_modgevns(w,d)
    end select
  end function loginvlinkdz

  pure elemental double precision function invlinkhz (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlinkhz = invlinkhz_ga(w,d)
    case (1) ! Gaussian boxcox
      invlinkhz = invlinkhz_ga(w,d)
    case (2,-2) ! Robit
      invlinkhz = invlinkhz_robit(w,d)
    case (3) ! Logit
      invlinkhz = invlinkhz_logit(w,d)
    case (4) ! Probit
      invlinkhz = invlinkhz_probit(w,d)
    case (5) ! Wallace
      invlinkhz = invlinkhz_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlinkhz = invlinkhz_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlinkhz = invlinkhz_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlinkhz = invlinkhz_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlinkhz = invlinkhz_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlinkhz = invlinkhz_modgevns(w,d)
    end select
  end function invlinkhz

  pure elemental double precision function invlink3z (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlink3z = invlink3z_ga(w,d)
    case (1) ! Gaussian boxcox
      invlink3z = invlink3z_ga(w,d)
    case (2,-2) ! Robit
      invlink3z = invlink3z_robit(w,d)
    case (3) ! Logit
      invlink3z = invlink3z_logit(w,d)
    case (4) ! Probit
      invlink3z = invlink3z_probit(w,d)
    case (5) ! Wallace
      invlink3z = invlink3z_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlink3z = invlink3z_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlink3z = invlink3z_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlink3z = invlink3z_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlink3z = invlink3z_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlink3z = invlink3z_modgevns(w,d)
    end select
  end function invlink3z

  pure elemental double precision function invlinkdn (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlinkdn = invlinkdn_ga(w,d)
    case (1) ! Gaussian boxcox
      invlinkdn = invlinkdn_ga(w,d)
    case (2,-2) ! Robit
      invlinkdn = invlinkdn_robit(w,d)
    case (3) ! Logit
      invlinkdn = invlinkdn_logit(w,d)
    case (4) ! Probit
      invlinkdn = invlinkdn_probit(w,d)
    case (5) ! Wallace
      invlinkdn = invlinkdn_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlinkdn = invlinkdn_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlinkdn = invlinkdn_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlinkdn = invlinkdn_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlinkdn = invlinkdn_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlinkdn = invlinkdn_modgevns(w,d)
    end select
  end function invlinkdn

  pure elemental double precision function invlinkhn (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlinkhn = invlinkhn_ga(w,d)
    case (1) ! Gaussian boxcox
      invlinkhn = invlinkhn_ga(w,d)
    case (2,-2) ! Robit
      invlinkhn = invlinkhn_robit(w,d)
    case (3) ! Logit
      invlinkhn = invlinkhn_logit(w,d)
    case (4) ! Probit
      invlinkhn = invlinkhn_probit(w,d)
    case (5) ! Wallace
      invlinkhn = invlinkhn_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlinkhn = invlinkhn_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlinkhn = invlinkhn_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlinkhn = invlinkhn_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlinkhn = invlinkhn_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlinkhn = invlinkhn_modgevns(w,d)
    end select
  end function invlinkhn

  pure elemental double precision function invlinkdzdn (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlinkdzdn = invlinkdzdn_ga(w,d)
    case (1) ! Gaussian boxcox
      invlinkdzdn = invlinkdzdn_ga(w,d)
    case (2,-2) ! Robit
      invlinkdzdn = invlinkdzdn_robit(w,d)
    case (3) ! Logit
      invlinkdzdn = invlinkdzdn_logit(w,d)
    case (4) ! Probit
      invlinkdzdn = invlinkdzdn_probit(w,d)
    case (5) ! Wallace
      invlinkdzdn = invlinkdzdn_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlinkdzdn = invlinkdzdn_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlinkdzdn = invlinkdzdn_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlinkdzdn = invlinkdzdn_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlinkdzdn = invlinkdzdn_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlinkdzdn = invlinkdzdn_modgevns(w,d)
    end select
  end function invlinkdzdn

  pure elemental double precision function invlinkhzdn (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlinkhzdn = invlinkhzdn_ga(w,d)
    case (1) ! Gaussian boxcox
      invlinkhzdn = invlinkhzdn_ga(w,d)
    case (2,-2) ! Robit
      invlinkhzdn = invlinkhzdn_robit(w,d)
    case (3) ! Logit
      invlinkhzdn = invlinkhzdn_logit(w,d)
    case (4) ! Probit
      invlinkhzdn = invlinkhzdn_probit(w,d)
    case (5) ! Wallace
      invlinkhzdn = invlinkhzdn_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlinkhzdn = invlinkhzdn_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlinkhzdn = invlinkhzdn_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlinkhzdn = invlinkhzdn_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlinkhzdn = invlinkhzdn_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlinkhzdn = invlinkhzdn_modgevns(w,d)
    end select
  end function invlinkhzdn

  pure elemental double precision function invlinkdzhn (w, d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invlinkdzhn = invlinkdzhn_ga(w,d)
    case (1) ! Gaussian boxcox
      invlinkdzhn = invlinkdzhn_ga(w,d)
    case (2,-2) ! Robit
      invlinkdzhn = invlinkdzhn_robit(w,d)
    case (3) ! Logit
      invlinkdzhn = invlinkdzhn_logit(w,d)
    case (4) ! Probit
      invlinkdzhn = invlinkdzhn_probit(w,d)
    case (5) ! Wallace
      invlinkdzhn = invlinkdzhn_wallace(w,d)
    case (6,8) ! Modified boxcox
      invlinkdzhn = invlinkdzhn_modbc(w,d)
    case (7,-7,9) ! Standard boxcox
      invlinkdzhn = invlinkdzhn_boxcox(w,d)
    case (10) ! Binomial modifiedGEV
      invlinkdzhn = invlinkdzhn_modgev(w,d)
    case (12,-12) ! Binomial GEV
      invlinkdzhn = invlinkdzhn_gev(w,d)
    case (11) ! Binomial modifiedGEV
      invlinkdzhn = invlinkdzhn_modgevns(w,d)
    end select
  end function invlinkdzhn


  ! Transformation

  pure elemental double precision function transfw (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      transfw = w
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      transfw = w
    case (-2) ! Binomial robit workaround
      transfw = flink_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      transfw = flink_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      transfw = flink_modgev(w,d)
    end select
  end function transfw

  pure elemental double precision function invtrw (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrw = w
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrw = w
    case (-2) ! Binomial robit workaround
      invtrw = invlink_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrw = invlink_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrw = invlink_modgev(w,d)
    end select
  end function invtrw

  pure elemental double precision function invtrwdz (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrwdz = 1d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrwdz = 1d0
    case (-2) ! Binomial robit workaround
      invtrwdz = invlinkdz_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrwdz = invlinkdz_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrwdz = invlinkdz_modgev(w,d)
    end select
  end function invtrwdz

  pure elemental double precision function loginvtrwdz (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      loginvtrwdz = 0d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      loginvtrwdz = 0d0
    case (-2) ! Binomial robit workaround
      loginvtrwdz = loginvlinkdz_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      loginvtrwdz = loginvlinkdz_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      loginvtrwdz = loginvlinkdz_modgev(w,d)
    end select
  end function loginvtrwdz

  pure elemental double precision function invtrwhz (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrwhz = 0d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrwhz = 0d0
    case (-2) ! Binomial robit workaround
      invtrwhz = invlinkhz_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrwhz = invlinkhz_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrwhz = invlinkhz_modgev(w,d)
    end select
  end function invtrwhz

  pure elemental double precision function invtrw3z (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrw3z = 0d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrw3z = 0d0
    case (-2) ! Binomial robit workaround
      invtrw3z = invlink3z_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrw3z = invlink3z_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrw3z = invlink3z_modgev(w,d)
    end select
  end function invtrw3z

  pure elemental double precision function invtrwdn (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrwdn = 0d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrwdn = 0d0
    case (-2) ! Binomial robit workaround
      invtrwdn = invlinkdn_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrwdn = invlinkdn_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrwdn = invlinkdn_modgev(w,d)
    end select
  end function invtrwdn

  pure elemental double precision function invtrwhn (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrwhn = 0d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrwhn = 0d0
    case (-2) ! Binomial robit workaround
      invtrwhn = invlinkhn_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrwhn = invlinkhn_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrwhn = invlinkhn_modgev(w,d)
    end select
  end function invtrwhn

  pure elemental double precision function invtrwdzdn (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrwdzdn = 0d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrwdzdn = 0d0
    case (-2) ! Binomial robit workaround
      invtrwdzdn = invlinkdzdn_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrwdzdn = invlinkdzdn_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrwdzdn = invlinkdzdn_modgev(w,d)
    end select
  end function invtrwdzdn

  pure elemental double precision function invtrwhzdn (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrwhzdn = 0d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrwhzdn = 0d0
    case (-2) ! Binomial robit workaround
      invtrwhzdn = invlinkhzdn_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrwhzdn = invlinkhzdn_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrwhzdn = invlinkhzdn_modgev(w,d)
    end select
  end function invtrwhzdn

  pure elemental double precision function invtrwdzhn (w,d)
    use modelfcns_link
    implicit none
    double precision, intent(in) :: w, d
    select case (MODELIS)
    case (0) ! Transformed Gaussian
      invtrwdzhn = 0d0
    case (1,2,3,4,5,6,7,8,9,10,11,12) ! No transformation
      invtrwdzhn = 0d0
    case (-2) ! Binomial robit workaround
      invtrwdzhn = invlinkdzhn_wallace(w,d)
    case (-7) ! Poisson boxcox workaround
      invtrwdzhn = invlinkdzhn_modbc(w,d)
    case (-12) ! Binomial GEV workaround
      invtrwdzhn = invlinkdzhn_modgev(w,d)
    end select
  end function invtrwdzhn
end module modelfcns

!!! interfaces.f90 ---
!!
!! Author: Evangelos Evangelou
!! Created: Thu, 19 Jun, 2014 11:34 (BST)
!! Last-Updated: Sat, 28 Nov, 2015 16:13 (GMT)
!!     Update #: 15
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!
!!! Commentary: Contains general-purpose functions
!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


module interfaces
  implicit none

! !!!!!!!!!!!!!!!!!!!!!!!!!!! Correlation functions !!!!!!!!!!!!!!!!!!!!!!!!!!!
!   interface
!     elemental double precision function matern (d, kappa)
!       double precision, intent(in) :: d, kappa
!     end function matern
!   end interface


!!!!!!!!!!!!!!!!!!!!!!!!!!!! Auxliary functions !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  interface
    pure elemental function isfinite (x)
      double precision, intent(in) :: x
      integer isfinite
    end function isfinite
  end interface

  interface
    pure elemental double precision function flog1pexp (x)
      double precision, intent(in) :: x
    end function flog1pexp
  end interface

  interface
    pure elemental double precision function flog1mexp (x)
      double precision, intent(in) :: x
    end function flog1mexp
  end interface

  interface
    pure elemental double precision function flogexpm1 (x)
      double precision, intent(in) :: x
    end function flogexpm1
  end interface

  interface
    pure elemental double precision function flog1p (x)
      double precision, intent(in) :: x
    end function flog1p
  end interface

  interface
    pure elemental double precision function fexpm1 (x)
      double precision, intent(in) :: x
    end function fexpm1
  end interface

  interface
    pure elemental double precision function fgamma (x)
      double precision, intent(in) :: x
    end function fgamma
  end interface

  interface
    pure elemental double precision function fdigamma (x)
      double precision, intent(in) :: x
    end function fdigamma
  end interface

  interface
    pure elemental double precision function ftrigamma (x)
      double precision, intent(in) :: x
    end function ftrigamma
  end interface

  interface
    pure elemental double precision function fbesselk (x, k)
      double precision, intent(in) :: x, k
    end function fbesselk
  end interface

  interface
    pure elemental double precision function fbesselkexp (x, k)
      double precision, intent(in) :: x, k
    end function fbesselkexp
  end interface

  interface
    pure elemental double precision function fbesselkratio (x, ktop, kbot)
      double precision, intent(in) :: x, ktop, kbot
    end function fbesselkratio
  end interface

  interface
    pure elemental double precision function flogbesselkdk (x, kappa)
      double precision, intent(in) :: x, kappa
    end function flogbesselkdk
  end interface


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! RNG !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  interface
    subroutine rngini ()
    end subroutine rngini
  end interface

  interface
    subroutine rngend ()
    end subroutine rngend
  end interface


!!!!!!!!!!!!!!!!!!!!!!!!!! Random sample functions !!!!!!!!!!!!!!!!!!!!!!!!!!

  interface
    double precision function randnorm ()
    end function randnorm
  end interface

  interface
    double precision function randunif ()
    end function randunif
  end interface

  interface
    pure double precision function randt (d)
      double precision, intent(in) :: d
    end function randt
  end interface

  interface
    pure double precision function randgamma (shp)
      double precision, intent(in) :: shp
    end function randgamma
  end interface

  interface
    pure double precision function randchisq (df)
      double precision, intent(in) :: df
    end function randchisq
  end interface


!!!!!!!!!!!!!!!!!!!!!!!!!!! Probability functions !!!!!!!!!!!!!!!!!!!!!!!!!!!
  interface
    pure elemental double precision function logprobt (q, d)
      double precision, intent(in) :: q, d
    end function logprobt
  end interface

  interface
    pure elemental double precision function logborpt (q, d)
      double precision, intent(in) :: q, d
    end function logborpt
  end interface

  interface
    pure elemental double precision function quantt (p, d)
      double precision, intent(in) :: p, d
    end function quantt
  end interface

  interface
    pure elemental double precision function logpdft (x, d)
      double precision, intent(in) :: x, d
    end function logpdft
  end interface

  interface
    pure elemental double precision function logprobnorm (q)
      double precision, intent(in) :: q
    end function logprobnorm
  end interface

  interface
    pure elemental double precision function logborpnorm (q)
      double precision, intent(in) :: q
    end function logborpnorm
  end interface

  interface
    pure elemental double precision function quantnorm (p)
      double precision, intent(in) :: p
    end function quantnorm
  end interface

  interface
    pure elemental double precision function logpdfnorm (x)
      double precision, intent(in) :: x
    end function logpdfnorm
  end interface

  interface
    pure elemental double precision function logproblogis (q)
      double precision, intent(in) :: q
    end function logproblogis
  end interface

  interface
    pure elemental double precision function logborplogis (q)
      double precision, intent(in) :: q
    end function logborplogis
  end interface

  interface
    pure elemental double precision function quantlogis (p)
      double precision, intent(in) :: p
    end function quantlogis
  end interface

  interface
    pure elemental double precision function logpdflogis (x)
      double precision, intent(in) :: x
    end function logpdflogis
  end interface
end module interfaces



! Local Variables:
! compile-command: "R CMD SHLIB"
! End:

#' Spatial correlation used in the geoBayes package
#'
#' This hidden variable contains a choice of correlation functions
#' that can be fit with this package. The function can be chosen in
#' the \code{corrfcn} input of the relevant function. This variable
#' cannot be overwritten.
#' @name geoBayes_correlation
#' @export
.geoBayes_corrfcn <- data.frame(
  corrfcn   = c("matern", "spherical", "powerexponential", "exponential", "gaussian"),
  needkappa = c(    TRUE,       FALSE,               TRUE,         FALSE,      FALSE),
  validklb  = c(       0,          NA,                  0,            NA, NA        ),
  validkub  = c(     Inf,          NA,                  2,            NA, NA        ),
  stringsAsFactors = FALSE)

lockBinding(".geoBayes_corrfcn", environment())

.geoBayes_correlation <- function (corrfcn)
{
  if (is.numeric(corrfcn)) {
    i <- as.integer(corrfcn)
    if (any(!is.finite(i)) || any(i < 1) || any(i > dim(.geoBayes_corrfcn)[1])) {
      stop ("Cannot deduce correlation.")
    }
  } else {
    i <- pmatch(corrfcn, .geoBayes_corrfcn$corrfcn)
    if (is.na(i)) stop ("Cannot deduce correlation.")
  }
  i
}



## Checks if the value of kappa is valid for the given correlation
## function. Returns the value of kappa.
.geoBayes_getkappa <- function (kappa, icf) {
  if (length(icf) != 1) stop ("Multiple correlations inputted.")
  if (!is.numeric(icf)) icf <- .geoBayes_correlation(icf)
  if (.geoBayes_corrfcn$needkappa[icf]) {
    if (!is.finite(kappa)) stop ("Need finite kappa.")
    kappa <- as.double(kappa)
    if (length(kappa) < 1) stop ("No kappa value inputted.")
    if (is.finite(.geoBayes_corrfcn$validklb[icf]) &&
        any(kappa <= .geoBayes_corrfcn$validklb[icf])) {
      stop (paste0("Input kappa cannot be less than or equal to ",
                   .geoBayes_corrfcn$validklb[icf], "."))
    }
    if (is.finite(.geoBayes_corrfcn$validkub[icf]) &&
        any(kappa > .geoBayes_corrfcn$validkub[icf])) {
      stop (paste0("Input kappa cannot be greater than ",
                   .geoBayes_corrfcn$validkub[icf], "."))
    }
    kappa
  } else {
    rep(0, length(kappa))
  }
}

##' Compute the standard errors for the Bayes factors estimates.
##'
##' Uses the batch means method to compute the standard errors for
##' Bayes factors.
##' @title Batch means, Bayes factors standard errors
##' @param pargrid A data frame with components "linkp", "phi", "omg",
##'   "kappa". Each row gives a combination of the parameters to
##'   compute the new standard errors.
##' @param runs A list with outputs from the function
##'   \code{\link{mcsglmm}} or \code{\link{mcstrga}}.
##' @param bfsize1 A scalar or vector of the same length as
##'   \code{runs} with all integer values or all values in (0, 1]. How
##'   many samples (or what proportion of the sample) to use for
##'   estimating the Bayes factors at the first stage. The remaining
##'   sample will be used for estimating the standard errors in the
##'   second stage. Setting it to 1 will perform only the first stage.
##' @param nbatch1 A scalar or vector of the same length as
##'   \code{runs}. All values must be integers or less than 1. This is
##'   used for calculating how many batches to split each of the
##'   sample in runs for the calculation of the Bayes factors standard
##'   errors for the parameters corresponding to \code{runs}.
##' @param nbatch2 A scalar or vector of the same length as
##'   \code{runs}. All values must be integers or less than 1. This is
##'   used for calculating how many batches to split each of the
##'   sample in runs for the calculation of the Bayes factors standard
##'   errors for the parameters corresponding to \code{pargrid}.
##' @param S1method Which method to use to calculate the Bayes factors
##'    in stage 1: Reverse logistic or Meng-Wong.
##' @param bvmethod Which method to use for the calculation of the
##'   batch variance. The standard method splits to disjoint batches.
##'   The second and third method use the spectral variance method
##'   with different lag windows.
##' @param reference Which model goes in the denominator.
##' @param transf Whether to use a transformed sample for the
##'   computations. If \code{"no"} or \code{FALSE}, it doesn't. If
##'   \code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
##'   \code{"wo"} it uses an alternative transformation. The latter
##'   can be used only for the families indicated by
##'   \code{.geoBayes_models$haswo}.
##' @return A list with components
##' \itemize{
##' \item \code{pargrid} The inputted pargrid augmented with the
##'   computed log Bayes factors and standard errors.
##' \item \code{bfEstimate} The estimates of the Bayes factors
##' \item \code{bfSigma} The covariance matrix for the Bayes factors
##' estimates.
##' }
##' @references Roy, V., Tan, A. and Flegal, J. (2018). Estimating
##'   standard errors for importance sampling estimators with multiple
##'   Markov chains, Statistica Sinica, 28 1079-1101.
##'
##' Roy, V., & Evangelou, E. (2018). Selection of proposal
##'   distributions for generalized importance sampling estimators.
##'   arXiv preprint arXiv:1805.00829. 
##' @useDynLib geoBayes bfse_no bfse_mu bfse_wo bfse_tr
##' @export
bmbfse <- function(pargrid, runs, bfsize1 = 0.80, nbatch1 = 0.5, nbatch2 = 0.5,
                   S1method = c("RL", "MW"),
                   bvmethod = c("Standard", "TukeyHanning", "Bartlett"),
                   reference = 1, transf = c("no", "mu", "wo"))
{
  ## Stage 1 method
  S1method <- match.arg(S1method)
  imeth <- match(S1method, eval(formals()$S1method))

  ## Batch variance method
  bvmethod <- match.arg(bvmethod)
  ibvmeth <- match(bvmethod, eval(formals()$bvmethod))

  ## Check input
  if (!all(sapply(runs, inherits, what = "geomcmc"))) {
    stop ("Input runs is not a list with elements of class geomcmc")
  }
  nruns <- length(runs)
  if (nruns == 0) stop ("No runs specified")

  reference <- as.integer(reference)
  if (isTRUE(reference < 1L | reference > nruns)) {
    stop("Argument reference does not correspond to a run in runs")
  }

  ## Check if fixed phi and omg
  if (!all(sapply(runs, function(x) length(x$FIXED$phi) == 1))) {
    stop("Each input runs must have a fixed value phi.")
  }
  if (!all(sapply(runs, function(x) length(x$FIXED$omg) == 1))) {
    stop("Each input runs must have a fixed value omg.")
  }

  ## Extract data and model
  nm_DATA <- c("response", "weights", "modelmatrix", "offset", "locations",
               "longlat")
  nm_MODEL <- c("family", "corrfcn", "betm0", "betQ0", "ssqdf", "ssqsc",
                "tsqdf", "tsqsc", "dispersion")
  DATA <- runs[[1]]$DATA[nm_DATA]
  MODEL <- runs[[1]]$MODEL[nm_MODEL]
  if (nruns > 1) {
    for (i in 2:nruns) {
      if (!identical(runs[[i]]$DATA[nm_DATA], DATA)) {
        stop("MCMC chains don't all correspond to the same data.")
      }
      if (!identical(runs[[i]]$MODEL[nm_MODEL], MODEL)) {
        stop("MCMC chains don't all correspond to the same model.")
      }
    }
  }
  y <- DATA$response
  n <- as.integer(length(y))
  l <- DATA$weights
  F <- DATA$modelmatrix
  offset <- DATA$offset
  p <- NCOL(F)
  loc <- DATA$locations
  dm <- sp::spDists(loc, longlat = DATA$longlat)
  family <- MODEL$family
  ## ifam <- .geoBayes_family(family)
  corrfcn <- MODEL$corrfcn
  icf <- .geoBayes_correlation(corrfcn)
  betm0 <- MODEL$betm0
  betQ0 <- MODEL$betQ0
  ssqdf <- MODEL$ssqdf
  ssqsc <- MODEL$ssqsc
  tsqdf <- MODEL$tsqdf
  tsqsc <- MODEL$tsqsc
  dispersion <- MODEL$dispersion

  ## Skeleton points
  phi_pnts <- as.double(sapply(runs, function(r) r$FIXED$phi))
  omg_pnts <- as.double(sapply(runs, function(r) r$FIXED$omg))
  nu_pnts <- as.double(sapply(runs, function(r) r$FIXED$linkp_num))
  if (.geoBayes_corrfcn$needkappa[icf]) {
    kappa_pnts <- sapply(runs, function(r) r$FIXED$kappa)
    kappa_pnts <- .geoBayes_getkappa(kappa_pnts, icf)
  } else {
    kappa_pnts <- rep(0, nruns)
  }

  ## MCMC Sizes
  Nout <- sapply(runs, function(x) x$MCMC$Nout)
  Nout1 <- getsize(bfsize1, Nout, "*")
  Ntot1 <- sum(Nout1)
  Nout2 <- Nout - Nout1
  Ntot2 <- sum(Nout2)

  ## Sizes for SE
  nb1 <- getsize(nbatch1, Nout1, "^")
  nb2 <- if(Ntot2 > 0) getsize(nbatch2, Nout2, "^") else rep(0L, nruns)

  ## Transformed sample
  getsample <-
    transfsample(runs,
                 list(response = y, family = family), transf)
  sample <- matrix(unlist(getsample$sample), n)
  itr <- getsample$itr
  transf <- getsample$transf
  real_transf <- getsample$real_transf
  ifam <- getsample$ifam

  froutine <- paste0("bfse_", real_transf)
  if (real_transf == "wo" || real_transf == "tr") {
    ifam <- -ifam
  }

  ## New parameter values
  pargrid <- check_pargrid(pargrid, family, corrfcn)
  nnew <- nrow(pargrid)
  phi <- pargrid$phi
  omg <- pargrid$omg
  kappa <- pargrid$kappa
  nu <- pargrid$nu

  ## Split the sample
  sel1 <- rep(rep(c(TRUE, FALSE), nruns), rbind(Nout1, Nout2))
  sample <- matrix(unlist(sample), n)
  z1 <- sample[, sel1, drop = FALSE]
  z2 <- sample[, !sel1, drop = FALSE]

  ## dispersion or tsqsc
  if (ifam == 0) {
    tsq <- tsqsc
  } else {
    tsq <- dispersion
    tsqdf <- 0
  }

  ## Output
  SE <- numeric(nnew)
  bf <- numeric(nruns)
  Sig <- numeric((nruns - 1)^2)
  VT1 <- VT2 <- logbfnew <- numeric(nnew)

  RUN <- .Fortran(froutine, bf = bf, logbfnew = logbfnew, 
                  Sig = Sig, SE = SE, VT1 = VT1,
                  VT2 = VT2, as.integer(reference),
                  as.double(phi), as.double(omg), as.double(nu),
                  as.double(kappa),
                  as.double(phi_pnts), as.double(omg_pnts),
                  as.double(nu_pnts), as.double(kappa_pnts),
                  as.double(z1), as.integer(Nout1), as.integer(Ntot1),
                  as.double(z2), as.integer(Nout2), as.integer(Ntot2),
                  as.double(y), as.double(l), as.double(F),
                  as.double(offset), as.double(dm),
                  as.double(betm0), as.double(betQ0), as.double(ssqdf),
                  as.double(ssqsc), as.double(tsqdf), as.double(tsq),
                  as.integer(icf), as.integer(n), as.integer(p),
                  as.integer(nnew),
                  as.integer(nruns), as.integer(ifam),
                  as.integer(imeth), as.integer(nb1), as.integer(nb2),
                  as.integer(ibvmeth), as.integer(itr), PACKAGE = "geoBayes")

  ## Return
  out <- list()
  pargrid$logbf <- RUN$logbfnew
  pargrid$SE <- RUN$SE/sqrt(Ntot2)
  out$pargrid <- pargrid
  out$bfEstimate <- RUN$bf
  out$bfSigma <- matrix(RUN$Sig/Ntot1, nruns - 1)
  out$VT1 <- RUN$VT1/Ntot2
  out$VT2 <- RUN$VT2/Ntot2
  return (out)
}


groupColMeans <- function (x, i) {
  ## Compute the mean at each column of x by subsetting its rows
  ## according to the factor i. Returns a matrix with as many columns
  ## as x and as many rows as the number of levels in i.
  x <- as.matrix(x)
  i <- as.factor(i)
  xi <- split.data.frame(x, i, drop = FALSE)
  m <- lapply(xi, colMeans)
  do.call(rbind, m)
}

bmmcse <- function(x, size)
{
  ## Purpose: Compute batch-means standard error of univariate MC samples
  ## ----------------------------------------------------------------------
  ## Arguments:
  ## x   A matrix of MC samples. The SE corresponding to each
  ## column of x will be returned.
  ## size   The size of each batch.
  ## ----------------------------------------------------------------------
  x <- as.matrix(x)
  n <- nrow(x)
  if (missing(size)) {
    size <- "sqroot"
  } else if (length(size) != 1) stop ("Argument size must have length 1.")
  if (is.character(size)) {
    i <- pmatch(size, c("sqroot", "sqrt", "cuberoot", "cubert"), nomatch = 0L,
                duplicates.ok = FALSE)
    if (i == 1 | i == 2) {
      size <- as.integer(sqrt(n))
    } else if (i == 3 | i == 4) {
      size <- as.integer(n^(1/3))
    } else stop ("Character size must be one of sqroot or cuberoot.")
  } else {
    size <- as.numeric(size)
    if (size <= 0) {
      stop ("Argument size must be positive.")
    } else if (size <= 1) {
      size <- as.integer(size*n)
    } else if (size > n) stop ("Argument size is larger than sample size.")
  }
  ## Determine the number of batches and the size of each batch
  nbatches <- as.integer(n/size)
  if (nbatches <= 1L) stop ("Number of batches too small. Reduce size.")
  nrem <- n%%nbatches
  batchl <- rep(c(size + 1L, size), c(nrem, nbatches - nrem))
  batchf <- rep(seq_len(nbatches), batchl)
  ## Compute batch mean and overall mean
  bm <- groupColMeans(x, batchf)
  m <- colSums(bm*batchl)/n
  ## Compute the batch variance
  bmmm <- sweep(bm, 2, m, check.margin = FALSE)
  bv <- colSums(batchl*bmmm^2)/(nbatches - 1)
  sqrt(bv)
}


## .. content for \description{} (no empty lines) ..
##
## .. content for \details{} ..
## @title
## @param sample A list containing the samples from each run. Each
##   element of the list must be a matrix or a vector. If a matrix,
##   each column represents one sample.
## @param theta0 A matrix or vector containing the parameter values
##   corresponding to each run in \code{sample}. Each column of
##   \code{theta0} is a parameter.
## @param theta1 A matrix or vector containing the new parameter
##   values for which to compute the standard errors. Each column of
##   \code{theta1} is a parameter.
## @param llikfun A function that computes the log-likelihood. The
##   function is called as \code{llikfun(x,theta,...)} where x is a
##   sample and theta is a parameter vector. The function must return
##   a scalar for each \code{x,theta} pair.
## @param MoreArgs Further arguments to \code{llikfun}.
## @param bfsize1 A scalar or vector of the same length as
##   \code{sample} with all integer values or all values in (0, 1]. How
##   many samples (or what proportion of the sample) to use for
##   estimating the Bayes factors at the first stage. The remaining
##   sample will be used for estimating the standard errors in the
##   second stage. Setting it to 1 will perform only the first stage.
## @param nbatch1 A scalar or vector of the same length as \code{sample}. All
##   values must be integers or less than 1. This is used for
##   calculating how many batches to split each of the sample in runs
##   for the calculation of the Bayes factors standard errors for the
##   parameters corresponding to \code{theta0}.
## @param nbatch2 A scalar or vector of the same length as \code{sample}. All
##   values must be integers or less than 1. This is used for
##   calculating how many batches to split each of the sample in runs
##   for the calculation of the Bayes factors standard errors for the
##   parameters corresponding to \code{theta1}.
## @param S1method Which method to use to calculate the Bayes factors
## in stage 1: Reverse logistic or Meng-Wong.
## @param bvmethod Which method to use for the calculation of the
##   batch variance. The standard method splits to disjoint batches.
##   The second and third method use the spectral variance method
##   with different lag windows.
## @param reference Which model goes in the denominator.
## @return A list with components
## \itemize{
## \item \code{logbf} The estimated Bayes factors at \code{theta0}.
## \item \code{bfSigma} The covariance matrix for the Bayes factors
## estimates in \code{logbf}.
## \item \code{SE} The standard error of the Bayes factors estimates
##   at \code{theta1}
## \item \code{VT1, VT2} The first and second variance terms for
##   \code{SE} above.
## }
bfsecalc <- function(sample, theta0, theta1, llikfun, MoreArgs = NULL,
                     bfsize1 = 0.8, nbatch1 = 0.5, nbatch2 = 0.5,
                     S1method = c("RL", "MW"),
                     bvmethod = c("Standard", "TukeyHanning", "Bartlett"),
                     reference = 1)
{
  cl <- match.call()
  
  ## Stage 1 method
  S1method <- match.arg(S1method)
  imeth <- match(S1method, eval(formals()$S1method))

  ## Batch variance method
  bvmethod <- match.arg(bvmethod)
  ibvmeth <- match(bvmethod, eval(formals()$bvmethod))

  kg <- length(sample)

  reference <- as.integer(reference)
  if (reference < 1 || reference > kg)
    stop ("Argument reference must be in between 1 and number of parameters.")

  nCOL <- function(x) if (length(d <- dim(x)) > 1L) d[2L] else length(x)
  Nout <- sapply(sample, nCOL)
  Ntot <- sum(Nout)
  Nout1 <- getsize(bfsize1, Nout, "*")
  Ntot1 <- sum(Nout1)
  Nout2 <- Nout - Nout1
  Ntot2 <- sum(Nout2)

  ## Size of x
  nROW <- function(x) if (length(d <- dim(x)) > 1L) d[1L] else 1L
  n <- unique(sapply(sample, nROW))
  if (length(n) != 1L) stop ("The dimension of the sampled variable changes.")

  ## Check input theta
  ltht <- nROW(theta0)
  if (nCOL(theta0) != kg)
    stop ("Columns in theta0 must be the length of samples.")
  if (nROW(theta1) != ltht)
    stop ("theta0 and theta1 must have the same number of rows.")
  nnew <- nCOL(theta1)
  dim(theta0) <- c(ltht, kg)
  dim(theta1) <- c(ltht, nnew)

  ## Sizes for SE
  nb1 <- getsize(nbatch1, Nout1, "^")
  nb2 <- getsize(nbatch2, Nout2, "^")

  ## Split the sample
  sel1 <- rep(rep(c(TRUE, FALSE), kg), rbind(Nout1, Nout2))
  sample <- matrix(unlist(sample), n)
  z1 <- sample[, sel1, drop = FALSE]
  z2 <- sample[, !sel1, drop = FALSE]

  ## Compute log-likelihoods
  llik1 <- mapply(llikfun, do.call("c", apply(z1, 2, list)),
                  rep(do.call("c", apply(theta0, 2, list)), each = Ntot1),
                  MoreArgs = MoreArgs)
  llik2 <- mapply(llikfun, do.call("c", apply(z2, 2, list)),
                  rep(do.call("c", apply(theta0, 2, list)), each = Ntot2),
                  MoreArgs = MoreArgs)
  llikn <- mapply(llikfun, do.call("c", apply(z2, 2, list)),
                  rep(do.call("c", apply(theta1, 2, list)), each = Ntot2),
                  MoreArgs = MoreArgs)

  ## Compute SE
  bf <- numeric(kg)
  SE <- VT1 <- VT2 <- logbfnew <- numeric(nnew)
  Sig <- matrix(0, kg-1, kg-1)
  Bet <- Omg <- matrix(0, kg, kg)
  fcall <- .Fortran("bfsecalc", bf, logbfnew, Sig, SE, VT1, VT2, reference,
                    llik1, llik2, llikn, Nout1, Ntot1, Nout2, Ntot2,
                    nnew, kg, imeth, nb1, nb2, ibvmeth, Bet, Omg,
                    PACKAGE = "geoBayes")

  Bet <- fcall[[21]]; Bet[lower.tri(Bet)] <- t(Bet)[lower.tri(Bet)]
  Omg <- fcall[[22]]; Omg[lower.tri(Omg)] <- t(Omg)[lower.tri(Omg)]
  BMP <- chol2inv(chol(Bet + 1/kg)) - 1/kg
  BOB <- t(BMP) %*% Omg %*% BMP

  ## Return
  out <- fcall[1:6]
  names(out) <- c("bf", "logbfnew", "bfSigma", "SE", "VT1", "VT2")
  out$Varlogbf <- BOB
  out$call <- cl
  out
}

##' Draw MCMC samples from the Spatial GLMM with known link function
##'
##' The four-parameter prior for \code{phi} is defined by
##' \deqn{\propto (\phi - \theta_4)^{\theta_2 -1} \exp\{-(\frac{\phi -
##' \theta_4}{\theta_1})^{\theta_3}\}}{propto (phi -
##' phiprior[4])^(phiprior[2]-1) *
##' exp(-((phi-phiprior[4])/phiprior[1])^phiprior[3])} for \eqn{\phi >
##' \theta_4}{phi > phiprior[4]}. The prior for \code{omg} is similar.
##' The prior parameters correspond to scale, shape, exponent, and
##' location. See \code{arXiv:1005.3274} for details of this
##' distribution.
##'
##' The GEV (Generalised Extreme Value) link is defined by \deqn{\mu =
##' 1 - \exp\{-\max(0, 1 + \nu x)^{\frac{1}{\nu}}\}}{mu = 1 -
##' \exp[-max(0, 1 + nu x)^(1/nu)]} for any real \eqn{\nu}{nu}. At
##' \eqn{\nu = 0}{nu = 0} it reduces to the complementary log-log
##' link.
##' @title MCMC samples from the Spatial GLMM
##' @param formula A representation of the model in the form
##'   \code{response ~ terms}. The response must be set to \code{NA}'s
##'   at the prediction locations (see the examples on how to do this
##'   using the function \code{\link{stackdata}}). At the observed
##'   locations the response is assumed to be a total of replicated
##'   measurements. The number of replications is inputted using the
##'   argument \code{weights}.
##' @param family The distribution of the data. The
##'   \code{"GEVbinomial"} family is the binomial family with link the
##'   GEV link (see Details).
##' @param data An optional data frame containing the variables in the
##'   model.
##' @param weights An optional vector of weights. Number of replicated
##'   samples for Gaussian and gamma, number of trials for binomial,
##'   time length for Poisson.
##' @param subset An optional vector specifying a subset of
##'   observations to be used in the fitting process.
##' @param offset See \code{\link[stats]{lm}}.
##' @param atsample A formula in the form \code{~ x1 + x2 + ... + xd}
##'   with the coordinates of the sampled locations.
##' @param corrfcn Spatial correlation function. See
##'   \code{\link{geoBayes_correlation}} for details.
##' @param linkp Parameter of the link function. A scalar value.
##' @param phi Optional starting value for the MCMC for the
##'   spatial range parameter \code{phi}. Defaults to the mean of its
##'   prior. If \code{corrtuning[["phi"]]} is 0, then this argument is required and
##'   it corresponds to the fixed value of \code{phi}. This can be a
##'   vector of the same length as Nout.
##' @param omg Optional starting value for the MCMC for the
##'   relative nugget parameter \code{omg}. Defaults to the mean of
##'   its prior. If \code{corrtuning[["omg"]]} is 0, then this argument is required
##'   and it corresponds to the fixed value of \code{omg}. This can be
##'   a vector of the same length as Nout.
##' @param kappa Optional starting value for the MCMC for the
##'   spatial correlation parameter \code{kappa} (Matern smoothness or
##'   exponential power). Defaults to the mean of
##'   its prior. If \code{corrtuning[["kappa"]]} is 0 and it is needed for
##'   the chosen correlation function, then this argument is required
##'   and it corresponds to the fixed value of \code{kappa}. This can be
##'   a vector of the same length as Nout.
##' @param Nout Number of MCMC samples to return. This can be a vector
##'   for running independent chains.
##' @param Nthin The thinning of the MCMC algorithm.
##' @param Nbi The burn-in of the MCMC algorithm.
##' @param betm0 Prior mean for beta (a vector or scalar).
##' @param betQ0 Prior standardised precision (inverse variance)
##'   matrix. Can be a scalar, vector or matrix. The first two imply a
##'   diagonal with those elements. Set this to 0 to indicate a flat
##'   improper prior.
##' @param ssqdf Degrees of freedom for the scaled inverse chi-square
##'   prior for the partial sill parameter.
##' @param ssqsc Scale for the scaled inverse chi-square prior for the
##'   partial sill parameter.
##' @param corrpriors A list with the components \code{phi},
##'   \code{omg} and \code{kappa} as needed. These correspond to the
##'   prior distribution parameters. For \code{phi} and \code{omg} it
##'   must be a vector of length 4. The generalized inverse gamma
##'   prior is assumed and the input corresponds to the parameters
##'   scale, shape, exponent, location in that order (see Details).
##'   For \code{kappa} it must be a vector of length 2. A uniform
##'   prior is assumed and the input corresponds to the lower and
##'   upper bounds in that order.
##' @param corrtuning A vector or list with the components \code{phi},
##'   \code{omg} and \code{kappa} as needed. These correspond to the
##'   random walk parameter for the Metropolis-Hastings step. Smaller values
##'   increase the acceptance ratio. Set this to 0 for fixed
##'   parameter value.
##' @param malatuning Tuning parameter for the MALA updates.
##' @param dispersion The fixed dispersion parameter.
##' @param longlat How to compute the distance between locations. If
##'   \code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
##'   distance. See \code{\link[sp]{spDists}}.
##' @param test Whether this is a trial run to monitor the acceptance
##'   ratio of the random walk for \code{phi} and \code{omg}. If set
##'   to \code{TRUE}, the acceptance ratio will be printed on the
##'   screen every 100 iterations of the MCMC. Tune the \code{phisc}
##'   and \code{omgsc} parameters in order to achive 20 to 30\%
##'   acceptance. Set this to a positive number to change the default
##'   100. No thinning or burn-in are done when testing.
##' @return A list containing the objects \code{MODEL}, \code{DATA},
##'   \code{FIXED}, \code{MCMC} and \code{call}. The MCMC samples are
##'   stored in the object \code{MCMC} as follows:
##' \itemize{
##'  \item \code{z} A matrix containing the MCMC samples for the
##' spatial random field. Each column is one sample.
##'  \item \code{mu} A matrix containing the MCMC samples for the
##' mean response (a transformation of z). Each column is one sample.
##'  \item \code{beta} A matrix containing the MCMC samples for the
##' regressor coefficients. Each column is one sample.
##'  \item \code{ssq} A vector with the MCMC samples for the partial
## sill parameter.
##'  \item \code{phi} A vector with the MCMC samples for the spatial
##' range parameter, if sampled.
##'  \item \code{omg} A vector with the MCMC samples for the relative
##' nugget parameter, if sampled.
##'  \item \code{logLik} A vector containing the value of the
##' log-likelihood evaluated at each sample.
##'  \item \code{acc_ratio} The acceptance ratio for the joint update
##' of the parameters \code{phi} and \code{omg}, if sampled.
##'  \item \code{sys_time} The total computing time for the MCMC sampling.
##'  \item \code{Nout}, \code{Nbi},  \code{Nthin} As in input. Used
##' internally in other functions.
##' }
##' The other objects contain input variables. The object \code{call}
##'   contains the function call.
##' @examples \dontrun{
##' data(rhizoctonia)
##'
##' ### Create prediction grid
##' predgrid <- mkpredgrid2d(rhizoctonia[c("Xcoord", "Ycoord")],
##'                          par.x = 100, chull = TRUE, exf = 1.2)
##'
##' ### Combine observed and prediction locations
##' rhizdata <- stackdata(rhizoctonia, predgrid$grid)
##' ##'
##' ### Define the model
##' corrf <- "spherical"
##' family <- "binomial.probit"
##' kappa <- 0
##' ssqdf <- 1
##' ssqsc <- 1
##' betm0 <- 0
##' betQ0 <- .01
##' phiprior <- c(100, 1, 1000, 100) # U(100, 200)
##' phisc <- 3
##' omgprior <- c(2, 1, 1, 0)        # Exp(mean = 2)
##' omgsc <- .1
##' ##'
##' ### MCMC sizes
##' Nout <- 100
##' Nthin <- 1
##' Nbi <- 0
##'
##' ### Trial run
##' emt <- mcsglmm_mala(Infected ~ 1, family, rhizdata, weights = Total,
##'                atsample = ~ Xcoord + Ycoord,
##'                Nout = Nout, Nthin = Nthin, Nbi = Nbi,
##'                betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf, ssqsc = ssqsc,
##'                corrpriors = list(phi = phiprior, omg = omgprior), 
##'                corrfcn = corrf, kappa = kappa,
##'                corrtuning = list(phi = phisc, omg = omgsc, kappa = 0),
##'                malatuning = .003, dispersion = 1, test = 10)
##'
##' ### Full run
##' emc <- update(emt, test = FALSE)
##'
##' emcmc <- mcmcmake(emc)
##' summary(emcmc[, c("phi", "omg", "beta", "ssq")])
##' plot(emcmc[, c("phi", "omg", "beta", "ssq")])
##' }
##' @importFrom sp spDists
##' @importFrom stats model.matrix model.response model.weights
##'   as.formula update model.offset
##' @useDynLib geoBayes mcspsamtry mcspsample
##' @export
mcsglmm_mala <- function (formula, family = "gaussian",
                          data, weights, subset, offset,
                          atsample, corrfcn = "matern",
                          linkp, phi, omg, kappa,
                          Nout, Nthin = 1, Nbi = 0, betm0, betQ0, ssqdf, ssqsc,
                          corrpriors, corrtuning, malatuning, 
                          dispersion = 1, longlat = FALSE, test = FALSE) {
  cl <- match.call()
  ## Family
  ifam <- .geoBayes_family(family)
  if (ifam) {
    family <- .geoBayes_models$family[ifam]
  } else {
    stop ("This family has not been implemented.")
  }
  if (.geoBayes_models$needlinkp[ifam]) {
    if (missing(linkp))
      stop ("Missing input linkp.")
  } else {
    linkp <- 0
  }

  ## Correlation function
  icf <- .geoBayes_correlation(corrfcn)
  corrfcn <- .geoBayes_corrfcn$corrfcn[icf]
  needkappa <- .geoBayes_corrfcn$needkappa[icf]

  ## Design matrix and data
  if (missing(data)) data <- environment(formula)
  if (length(formula) != 3) stop ("The formula input is incomplete.")
  if ("|" == all.names(formula[[2]], TRUE, 1)) formula[[2]] <- formula[[2]][[2]]
  mfc <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data", "subset", "weights", "offset"),
             names(mfc), 0L)
  mfc <- mfc[c(1L, m)]
  mfc$formula <- formula
  mfc$drop.unused.levels <- TRUE
  mfc$na.action <- "na.pass"
  mfc[[1L]] <- quote(stats::model.frame)
  mf <- eval(mfc, parent.frame())
  mt <- attr(mf, "terms")
  FF <- model.matrix(mt,mf)
  if (!all(is.finite(FF))) stop ("Non-finite values in the design matrix")
  p <- NCOL(FF)
  yy <- unclass(model.response(mf))
  if (!is.vector(yy)) {
    stop ("The response must be a vector")
  }
  yy <- as.double(yy)
  ll <- model.weights(mf)
  oofset <- as.vector(model.offset(mf))
  if (!is.null(oofset)) {
    if (length(oofset) != NROW(yy)) {
      stop(gettextf("number of offsets is %d, should equal %d (number of observations)", 
                    length(oofset), NROW(yy)), domain = NA)
    } else {
      oofset <- as.double(oofset)
    }
  } else {
    oofset <- double(NROW(yy))
  }

  ## All locations
  atsample <- update(atsample, NULL ~ . + 0) # No response and no intercept
  mfatc <- mfc
  mfatc$weights = NULL
  mfatc$formula = atsample
  mfat <- eval(mfatc, parent.frame())
  loc <- as.matrix(mfat)
  if (!all(is.finite(loc))) stop ("Non-finite values in the locations")
  if (corrfcn == "spherical" && NCOL(loc) > 3) {
    stop ("Cannot use the spherical correlation for dimensions
grater than 3.")
  }

  ## Split sample, prediction
  ii <- is.finite(yy)
  y <- yy[ii]
  k <- sum(ii)
  l <- ll[ii]
  l <- if (is.null(l)) rep.int(1.0, k) else as.double(l)
  if (any(!is.finite(l))) stop ("Non-finite values in the weights")
  if (any(l <= 0)) stop ("Non-positive weights not allowed")
  if (grepl("^binomial(\\..+)?$", family)) {
    l <- l - y # Number of failures
  }
  F <- FF[ii, , drop = FALSE]
  offset <- oofset[ii]
  dm <- sp::spDists(loc[ii, , drop = FALSE], longlat = longlat)
  k0 <- sum(!ii)
  if (k0 > 0) {
    F0 <- FF[!ii, , drop = FALSE]
    dmdm0 <- sp::spDists(loc[ii, , drop = FALSE], loc[!ii, , drop = FALSE],
                         longlat = longlat)
    offset0 <- oofset[!ii]
  } else {
    F0 <- dmdm0 <- offset0 <- numeric(0)
    dim(F0) <- c(0, p)
    dim(dmdm0) <- c(k, 0)
  }

  ## Prior for ssq
  ssqdf <- as.double(ssqdf)
  if (ssqdf <= 0) stop ("Argument ssqdf must > 0")
  ssqsc <- as.double(ssqsc)
  if (ssqsc <= 0) stop ("Argument ssqsc must > 0")

  ## Prior for beta
  betaprior <- getbetaprior(betm0, betQ0, p)
  betm0 <- betaprior$betm0
  betQ0 <- betaprior$betQ0

  ## Other fixed parameters
  dispersion <- as.double(dispersion)
  if (dispersion <= 0) stop ("Invalid argument dispersion")

  nu <- .geoBayes_getlinkp(linkp, ifam)

  ## MCMC samples
  Nout <- as.integer(Nout)
  if (any(Nout < 0)) stop ("Negative MCMC sample size entered.")
  nch <- length(Nout)         # Number of chains
  Nmc <- Nout                 # Size of each chain
  Nout <- sum(Nout)           # Total MCMC size
  Nbi <- as.integer(Nbi)
  Nthin <- as.integer(Nthin)
  lglk <- numeric(Nout)
  z <- matrix(0, k, Nout)
  z0 <- matrix(0, k0, Nout)
  beta <- matrix(0, p, Nout)
  ssq <- numeric(Nout)
  if (malatuning <= 0) stop ("Input malatuning must > 0.")

  ## Starting values for correlation parameters
  phisc <- corrtuning[["phi"]]
  if (is.null(phisc) || !is.numeric(phisc) || phisc < 0)
    stop ("Invalid tuning parameter for phi.")
  if (phisc > 0) {
    phipars <- check_gengamma_prior(corrpriors[["phi"]])
  } else phipars <- rep.int(0, 4)
  if (missing(phi)) {
    if (phisc == 0) {
      stop ("Argument phi needed for fixed phi")
    } else {
      if(phipars[2] == -1) {
        tmp <- .1/abs(phipars[3])
      } else {
        tmp <- abs((phipars[2]+1)/phipars[3])
      }
      phistart <- phipars[4] + phipars[1]*gamma(tmp)/
        gamma(phipars[2]/phipars[3])
    }
  } else {
    phistart <- as.double(phi)
    if (phisc > 0 && phistart <= phipars[4]) {
      stop ("Starting value for phi not in the support of its prior")
    }
  }
  phi <- numeric(Nout)
  phi[cumsum(c(1, Nmc[-nch]))] <- phistart

  omgsc <- corrtuning[["omg"]]
  if (is.null(omgsc) || !is.numeric(omgsc) || omgsc < 0)
    stop ("Invalid tuning parameter for omg.")
  if (omgsc > 0) {
    omgpars <- check_gengamma_prior(corrpriors[["omg"]])
  } else omgpars <- rep.int(0, 4)
  if (missing(omg)) {
    if (omgsc == 0) {
      stop ("Argument omg needed for fixed omg")
    } else {
      if(omgpars[2] == -1) {
        tmp <- .1/abs(omgpars[3])
      } else {
        tmp <- abs((omgpars[2]+1)/omgpars[3])
      }
      omgstart <- omgpars[4] + omgpars[1]*gamma(tmp)/
        gamma(omgpars[2]/omgpars[3])
    }
  } else {
    omgstart <- as.double(omg)
    if (omgsc > 0 && omgstart <= omgpars[4]) {
      stop ("Starting value for omg not in the support of its prior")
    }
  }
  omg <- numeric(Nout)
  omg[cumsum(c(1, Nmc[-nch]))] <- omgstart

  if (needkappa) {
    kappasc <- corrtuning[["kappa"]]
  } else {
    kappasc <- 0
    kappa <- 0
  }
  if (is.null(kappasc) || !is.numeric(kappasc) || kappasc < 0)
    stop ("Invalid tuning parameter for kappa.")
  if (kappasc > 0) {
    kappapars <- check_unif_prior(corrpriors[["kappa"]])
  } else kappapars <- c(0, 0)
  if (missing(kappa)) {
    if (kappasc == 0) {
      stop ("Argument kappa needed for fixed kappa")
    } else {
      kappastart <- (kappapars[1] + kappapars[2])*.5
    }
  } else {
    kappastart <- as.double(kappa)
  }
  if (kappasc > 0) {
    kappastart <- .geoBayes_getkappa(kappastart, icf)
    kappapars <- .geoBayes_getkappa(kappapars, icf)
    if (kappastart >= kappapars[2] || kappastart <= kappapars[1]) {
      stop ("Starting value for kappa not in the support of its prior")
    }
  }
  kappa <- numeric(Nout)
  kappa[cumsum(c(1, Nmc[-nch]))] <- kappastart

  ## Run code
  if (test > 0) { # Running a test
    if (is.logical(test)) test <- 100
    test <- as.integer(test)
    acc <- acc_z <- 0L
    tm <- system.time({
      RUN <- .Fortran("mcspsamtry_mala", ll = lglk, z = z, phi = phi, omg = omg,
                      kappa = kappa,
                      acc = acc,
                      as.double(y), as.double(l), as.double(F),
                      as.double(offset), 
                      as.double(betm0), as.double(betQ0), as.double(ssqdf),
                      as.double(ssqsc), as.double(phipars),
                      as.double(omgpars), as.double(kappapars),
                      as.double(phisc), as.double(omgsc), as.double(kappasc),
                      as.integer(icf),
                      as.double(nu), as.double(dispersion), as.double(dm),
                      as.integer(Nout), as.integer(test), as.integer(k),
                      as.integer(p), as.integer(ifam), as.double(malatuning), 
                      acc_z = acc_z, PACKAGE = "geoBayes")
    })
    ## Store samples
    ll <- RUN$ll
    zz0 <- matrix(NA, NROW(yy), Nout)
    zz0[ii, ] <- RUN$z
    mm0 <- NULL
    beta <- NULL
    ssq <- NULL
    phi <- RUN$phi
###     attr(phi, 'fixed') <- phisc == 0
    omg <- RUN$omg
###     attr(omg, 'fixed') <- omgsc == 0
###     attr(nu, 'fixed') <- TRUE
    kappa <- RUN$kappa
    acc_ratio <- RUN$acc/Nout
    acc_ratio_z <- RUN$acc_z/Nout
    Nthin <- 1
    Nbi <- 0
###     out <- list(z = zz0, beta = beta, ssq = ssq, phi = phi, omg = omg, nu = nu,
###                 logLik = ll, acc_ratio = acc_ratio, sys_time = tm,
###                 Nout = Nout, Nbi = Nbi, Nthin = Nthin,
###                 response = y, weights = l, modelmatrix = F, family = family,
###                 betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf, ssqsc = ssqsc,
###                 corrfcn = corrfcn, kappa = kappa,
###                 dispersion = dispersion, locations = loc[ii, , drop = FALSE],
###                 longlat = longlat, whichobs = ii)
  } else {
    acc <- acc_z <- integer(nch)
    tm <- system.time({
      RUN <- .Fortran("mcspsample_mala", ll = lglk, z = z, z0 = z0,
                      mu = z, mu0 = z0,
                      beta = beta, ssq = ssq,
                      phi = phi, omg = omg, kappa = kappa, acc = acc,
                      as.double(y), as.double(l), as.double(F),
                      as.double(offset), as.double(F0), as.double(offset0), 
                      as.double(betm0), as.double(betQ0), as.double(ssqdf),
                      as.double(ssqsc), as.double(phipars), as.double(omgpars),
                      as.double(kappapars),
                      as.double(phisc), as.double(omgsc), as.double(kappasc),
                      as.integer(icf),
                      as.double(nu), as.double(dispersion), as.double(dm),
                      as.double(dmdm0), as.integer(nch), as.integer(Nmc),
                      as.integer(Nout), as.integer(Nbi),
                      as.integer(Nthin), as.integer(k), as.integer(k0),
                      as.integer(p), as.integer(ifam), as.double(malatuning),
                      acc_z = acc_z, 
                      PACKAGE = "geoBayes")
    })
    ## Store samples
    ll <- RUN$ll
    zz0 <- mm0 <- matrix(NA, NROW(yy), Nout)
    zz0[ii, ] <- RUN$z
    zz0[!ii, ] <- RUN$z0
    mm0[ii, ] <- RUN$mu
    mm0[!ii, ] <- RUN$mu0
    beta <- RUN$beta
    ssq <- RUN$ssq
    phi <- RUN$phi
###    attr(phi, 'fixed') <- phisc == 0
    omg <- RUN$omg
###    attr(omg, 'fixed') <- omgsc == 0
###    attr(nu, 'fixed') <- TRUE
    kappa <- RUN$kappa
    acc_ratio <- RUN$acc/(Nmc*Nthin + max(Nthin, Nbi))
    acc_ratio_z <- RUN$acc_z/(Nmc*Nthin + max(Nthin, Nbi))
###     out <- list(z = zz0, mu = mm0,
###                 beta = beta, ssq = ssq, phi = phi, omg = omg, nu = nu,
###                 logLik = ll, acc_ratio = acc_ratio, sys_time = tm,
###                 Nout = Nout, Nbi = Nbi, Nthin = Nthin,
###                 response = y, weights = l, modelmatrix = F, family = family,
###                 betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf, ssqsc = ssqsc,
###                 corrfcn = corrfcn, kappa = kappa,
###                 dispersion = dispersion, locations = loc[ii, , drop = FALSE],
###                 longlat = longlat, whichobs = ii)
  }
  MCMC <- FIXED <- MODEL <- DATA <- list()
  MCMC$z <- zz0
  MCMC$mu <- mm0
  MCMC$beta <- beta
  MCMC$ssq <- ssq
  FIXED$linkp <- as.vector(linkp)
  FIXED$linkp_num <- nu
  if (phisc == 0) {
    FIXED$phi <- phi[1]
  } else {
    MCMC$phi <- phi
  }
  if (omgsc == 0) {
    FIXED$omg <- omg[1]
  } else {
    MCMC$omg <- omg
  }
  if (kappasc == 0) {
    FIXED$kappa <- kappa[1]
  } else {
    MCMC$kappa <- kappa
  }
  MCMC$logLik <- ll
  MCMC$acc_ratio <- acc_ratio
  MCMC$acc_ratio_z <- acc_ratio_z
  MCMC$sys_time <- tm
  MCMC$Nout <- Nout
  MCMC$Nbi <- Nbi
  MCMC$Nthin <- Nthin
  MCMC$whichobs <- ii
  DATA$response <- y
  DATA$weights <- l
  DATA$modelmatrix <- F
  DATA$offset <- offset
  DATA$locations <- loc[ii, , drop = FALSE]
  DATA$longlat <- longlat
  MODEL$family <- family
  MODEL$corrfcn <- corrfcn
  MODEL$betm0 <- betm0
  MODEL$betQ0 <- betQ0
  MODEL$ssqdf <- ssqdf
  MODEL$ssqsc <- ssqsc
  MODEL$phipars <- phipars
  MODEL$omgpars <- omgpars
  MODEL$dispersion <- dispersion
  out <- list(MODEL = MODEL, DATA = DATA, FIXED = FIXED, MCMC = MCMC, call = cl)
  class(out) <- "geomcmc"
  out
}

##' Draw MCMC samples from the transformed Gaussian model with known
##' link function
##'
##' Simulates from the posterior distribution of this model.
##' @title MCMC samples from the transformed Gaussian model
##' @param formula A representation of the model in the form
##' \code{response ~ terms}. The response must be set to \code{NA}'s
##' at the prediction locations (see the example in
##' \code{\link{mcsglmm}} for how to do this using
##' \code{\link{stackdata}}). At the observed locations the response
##' is assumed to be a total of replicated measurements. The number of
##' replications is inputted using the argument \code{weights}.
##' @param data An optional data frame containing the variables in the
##' model.
##' @param weights An optional vector of weights. Number of replicated
##' samples.
##' @param subset An optional vector specifying a subset of
##' observations to be used in the fitting process.
##' @param offset See \code{\link[stats]{lm}}.
##' @param atsample A formula in the form \code{~ x1 + x2 + ... + xd}
##' with the coordinates of the sampled locations.
##' @param corrfcn Spatial correlation function. See
##' \code{\link{geoBayes_correlation}} for details.
##' @param linkp Parameter of the link function. A scalar value.
##' @param phi Optional starting value for the MCMC for the
##'   spatial range parameter \code{phi}. Defaults to the mean of its
##'   prior. If \code{corrtuning[["phi"]]} is 0, then this argument is required and
##'   it corresponds to the fixed value of \code{phi}. This can be a
##'   vector of the same length as Nout.
##' @param omg Optional starting value for the MCMC for the
##'   relative nugget parameter \code{omg}. Defaults to the mean of
##'   its prior. If \code{corrtuning[["omg"]]} is 0, then this argument is required
##'   and it corresponds to the fixed value of \code{omg}. This can be
##'   a vector of the same length as Nout.
##' @param kappa Optional starting value for the MCMC for the
##'   spatial correlation parameter \code{kappa} (Matern smoothness or
##'   exponential power). Defaults to the mean of
##'   its prior. If \code{corrtuning[["kappa"]]} is 0 and it is needed for
##'   the chosen correlation function, then this argument is required
##'   and it corresponds to the fixed value of \code{kappa}. This can be
##'   a vector of the same length as Nout.
##' @param Nout Number of MCMC samples to return. This can be a vector
##'   for running independent chains.
##' @param Nthin The thinning of the MCMC algorithm.
##' @param Nbi The burn-in of the MCMC algorithm.
##' @param betm0 Prior mean for beta (a vector or scalar).
##' @param betQ0 Prior standardised precision (inverse variance)
##' matrix. Can be a scalar, vector or matrix. The first two imply a
##' diagonal with those elements. Set this to 0 to indicate a flat
##' improper prior.
##' @param ssqdf Degrees of freedom for the scaled inverse chi-square
##' prior for the partial sill parameter.
##' @param ssqsc Scale for the scaled inverse chi-square prior for the
##' partial sill parameter.
##' @param tsqdf Degrees of freedom for the scaled inverse chi-square
##' prior for the measurement error parameter.
##' @param tsqsc Scale for the scaled inverse chi-square prior for the
##' measurement error parameter.
##' @param corrpriors A list with the components \code{phi},
##'   \code{omg} and \code{kappa} as needed. These correspond to the
##'   prior distribution parameters. For \code{phi} and \code{omg} it
##'   must be a vector of length 4. The generalized inverse gamma
##'   prior is assumed and the input corresponds to the parameters
##'   scale, shape, exponent, location in that order (see Details).
##'   For \code{kappa} it must be a vector of length 2. A uniform
##'   prior is assumed and the input corresponds to the lower and
##'   upper bounds in that order.
##' @param malatuning Tuning parameter for the MALA updates.
##' @param corrtuning A vector or list with the components \code{phi},
##'   \code{omg} and \code{kappa} as needed. These correspond to the
##'   random walk parameter for the Metropolis-Hastings step. Smaller values
##'   increase the acceptance ratio. Set this to 0 for fixed
##'   parameter value.
##' @param longlat How to compute the distance between locations. If
##' \code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
##' distance. See \code{\link[sp]{spDists}}.
##' @param test Whether this is a trial run to monitor the acceptance
##' ratio of the random walk for \code{phi} and \code{omg}. If set to
##' \code{TRUE}, the acceptance ratio will be printed on the screen
##' every 100 iterations of the MCMC. Tune the \code{phisc} and
##' \code{omgsc} parameters in order to achive 20 to 30\% acceptance.
##' Set this to a positive number to change the default 100. No
##' thinning or burn-in are done when testing.
##' @return A list containing the objects \code{MODEL}, \code{DATA},
##'   \code{FIXED}, \code{MCMC} and \code{call}. The MCMC samples are
##'   stored in the object \code{MCMC} as follows:
##' \itemize{
##'  \item \code{z} A matrix containing the MCMC samples for the
##' spatial random field. Each column is one sample.
##'  \item \code{mu} A matrix containing the MCMC samples for the
##' mean response (a transformation of z). Each column is one sample.
##'  \item \code{beta} A matrix containing the MCMC samples for the
##' regressor coefficients. Each column is one sample.
##'  \item \code{ssq} A vector with the MCMC samples for the partial
## sill parameter.
##'  \item \code{tsq} A vector with the MCMC samples for the
##' measurement error variance.
##'  \item \code{phi} A vector with the MCMC samples for the spatial
##' range parameter, if sampled.
##'  \item \code{omg} A vector with the MCMC samples for the relative
##' nugget parameter, if sampled.
##'  \item \code{logLik} A vector containing the value of the
##' log-likelihood evaluated at each sample.
##'  \item \code{acc_ratio} The acceptance ratio for the joint update
##' of the parameters \code{phi} and \code{omg}, if sampled.
##'  \item \code{sys_time} The total computing time for the MCMC sampling.
##'  \item \code{Nout}, \code{Nbi},  \code{Nthin} As in input. Used
##' internally in other functions.
##' }
##' The other objects contain input variables. The object \code{call}
##'   contains the function call.
##' @examples \dontrun{
##' ### Load the data
##' data(rhizoctonia)
##' rhiz <- na.omit(rhizoctonia)
##' rhiz$IR <- rhiz$Infected/rhiz$Total # Incidence rate of the
##'                               # rhizoctonia disease
##'
##' ### Define the model
##' corrf <- "spherical"
##' ssqdf <- 1
##' ssqsc <- 1
##' tsqdf <- 1
##' tsqsc <- 1
##' betm0 <- 0
##' betQ0 <- diag(.01, 2, 2)
##' phiprior <- c(200, 1, 1000, 100) # U(100, 300)
##' phisc <- 1
##' omgprior <- c(3, 1, 1000, 0) # U(0, 3)
##' omgsc <- 1
##' linkp <- 1
##'
##' ## MCMC parameters
##' Nout <- 100
##' Nbi <- 0
##' Nthin <- 1
##'
##' samplt <- mcstrga_mala(Yield ~ IR, data = rhiz,
##'                   atsample = ~ Xcoord + Ycoord, corrf = corrf,
##'                   Nout = Nout, Nthin = Nthin,
##'                   Nbi = Nbi, betm0 = betm0, betQ0 = betQ0,
##'                   ssqdf = ssqdf, ssqsc = ssqsc,
##'                   tsqdf = tsqdf, tsqsc = tsqsc,
##'                   corrprior = list(phi = phiprior, omg = omgprior),
##'                   linkp = linkp,
##'                   corrtuning = list(phi = phisc, omg = omgsc, kappa = 0),
##'                   malatuning = .0002, test=10)
##'
##' sample <- update(samplt, test = FALSE)
##' }
##' @importFrom sp spDists
##' @importFrom stats model.matrix model.response model.weights
##'   as.formula update model.offset
##' @useDynLib geoBayes trgasamtry trgasample
##' @export
mcstrga_mala <- function (formula,
                          data, weights, subset, offset,
                          atsample, corrfcn = "matern",
                          linkp, phi, omg, kappa,
                          Nout, Nthin = 1, Nbi = 0, betm0, betQ0, ssqdf, ssqsc,
                          tsqdf, tsqsc,
                          corrpriors, corrtuning, malatuning, 
                          longlat = FALSE,
                          test = FALSE) {
  cl <- match.call()

  family <- "transformed.gaussian"

  ## Correlation function
  icf <- .geoBayes_correlation(corrfcn)
  corrfcn <- .geoBayes_corrfcn$corrfcn[icf]
  needkappa <- .geoBayes_corrfcn$needkappa[icf]

  ## Design matrix and data
  if (missing(data)) data <- environment(formula)
  if (length(formula) != 3) stop ("The formula input is incomplete.")
  if ("|" == all.names(formula[[2]], TRUE, 1)) formula[[2]] <- formula[[2]][[2]]
  mfc <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data", "subset", "weights", "offset"),
             names(mfc), 0L)
  mfc <- mfc[c(1L, m)]
  mfc$formula <- formula
  mfc$drop.unused.levels <- TRUE
  mfc$na.action <- "na.pass"
  mfc[[1L]] <- quote(stats::model.frame)
  mf <- eval(mfc, parent.frame())
  mt <- attr(mf, "terms")
  FF <- model.matrix(mt,mf)
  if (!all(is.finite(FF))) stop ("Non-finite values in the design matrix")
  p <- NCOL(FF)
  yy <- unclass(model.response(mf))
  if (!is.vector(yy)) {
    stop ("The response must be a vector")
  }
  yy <- as.double(yy)
  ll <- model.weights(mf)
  oofset <- as.vector(model.offset(mf))
  if (!is.null(oofset)) {
    if (length(oofset) != NROW(yy)) {
      stop(gettextf("number of offsets is %d, should equal %d (number of observations)", 
                    length(oofset), NROW(yy)), domain = NA)
    } else {
      oofset <- as.double(oofset)
    }
  } else {
    oofset <- double(NROW(yy))
  }

  ## All locations
  atsample <- update(atsample, NULL ~ . + 0) # No response and no intercept
  mfatc <- mfc
  mfatc$weights = NULL
  mfatc$formula = atsample
  mfat <- eval(mfatc, parent.frame())
  loc <- as.matrix(mfat)
  if (!all(is.finite(loc))) stop ("Non-finite values in the locations")
  if (corrfcn == "spherical" && NCOL(loc) > 3) {
    stop ("Cannot use the spherical correlation for dimensions
grater than 3.")
  }

  ## Split sample, prediction
  ii <- is.finite(yy)
  y <- yy[ii]
  k <- sum(ii)
  l <- ll[ii]
  l <- if (is.null(l)) rep.int(1.0, k) else as.double(l)
  if (any(!is.finite(l))) stop ("Non-finite values in the weights")
  if (any(l <= 0)) stop ("Non-positive weights not allowed")
  ybar <- y/l
  F <- FF[ii, , drop = FALSE]
  offset <- oofset[ii]
  dm <- sp::spDists(loc[ii, , drop = FALSE], longlat = longlat)
  k0 <- sum(!ii)
  if (k0 > 0) {
    F0 <- FF[!ii, , drop = FALSE]
    dmdm0 <- sp::spDists(loc[ii, , drop = FALSE], loc[!ii, , drop = FALSE],
                         longlat = longlat)
    offset0 <- oofset[!ii]
  } else {
    F0 <- dmdm0 <- offset0 <- numeric(0)
    dim(F0) <- c(0, p)
    dim(dmdm0) <- c(k, 0)
  }

  ## Prior for ssq
  ssqdf <- as.double(ssqdf)
  if (ssqdf <= 0) stop ("Argument ssqdf must > 0")
  ssqsc <- as.double(ssqsc)
  if (ssqsc <= 0) stop ("Argument ssqsc must > 0")

  ## Prior for beta
  betaprior <- getbetaprior(betm0, betQ0, p)
  betm0 <- betaprior$betm0
  betQ0 <- betaprior$betQ0

  ## Prior for tsq
  tsqdf <- as.double(tsqdf)
  if (tsqdf <= 0) stop ("Argument tsqdf must > 0")
  tsqsc <- as.double(tsqsc)
  if (tsqsc <= 0) stop ("Argument tsqsc must > 0")

  if (missing(linkp))
    stop ("Missing input linkp.")
  nu <- .geoBayes_getlinkp(linkp, family)

  ## MCMC samples
  Nout <- as.integer(Nout)
  if (any(Nout < 0)) stop ("Negative MCMC sample size entered.")
  nch <- length(Nout)         # Number of chains
  Nmc <- Nout                 # Size of each chain
  Nout <- sum(Nout)           # Total MCMC size
  Nbi <- as.integer(Nbi)
  Nthin <- as.integer(Nthin)
  lglk <- numeric(Nout)
  z <- matrix(0, k, Nout)
  z0 <- matrix(0, k0, Nout)
  beta <- matrix(0, p, Nout)
  ssq <- tsq <- numeric(Nout)
  if (malatuning <= 0) stop ("Input malatuning must > 0.")

  ## Starting values for correlation parameters
  phisc <- corrtuning[["phi"]]
  if (is.null(phisc) || !is.numeric(phisc) || phisc < 0)
    stop ("Invalid tuning parameter for phi.")
  if (phisc > 0) {
    phipars <- check_gengamma_prior(corrpriors[["phi"]])
  } else phipars <- rep.int(0, 4)
  if (missing(phi)) {
    if (phisc == 0) {
      stop ("Argument phi needed for fixed phi")
    } else {
      if(phipars[2] == -1) {
        tmp <- .1/abs(phipars[3])
      } else {
        tmp <- abs((phipars[2]+1)/phipars[3])
      }
      phistart <- phipars[4] + phipars[1]*gamma(tmp)/
        gamma(phipars[2]/phipars[3])
    }
  } else {
    phistart <- as.double(phi)
    if (phisc > 0 && phistart <= phipars[4]) {
      stop ("Starting value for phi not in the support of its prior")
    }
  }
  phi <- numeric(Nout)
  phi[cumsum(c(1, Nmc[-nch]))] <- phistart

  omgsc <- corrtuning[["omg"]]
  if (is.null(omgsc) || !is.numeric(omgsc) || omgsc < 0)
    stop ("Invalid tuning parameter for omg.")
  if (omgsc > 0) {
    omgpars <- check_gengamma_prior(corrpriors[["omg"]])
  } else omgpars <- rep.int(0, 4)
  if (missing(omg)) {
    if (omgsc == 0) {
      stop ("Argument omg needed for fixed omg")
    } else {
      if(omgpars[2] == -1) {
        tmp <- .1/abs(omgpars[3])
      } else {
        tmp <- abs((omgpars[2]+1)/omgpars[3])
      }
      omgstart <- omgpars[4] + omgpars[1]*gamma(tmp)/
        gamma(omgpars[2]/omgpars[3])
    }
  } else {
    omgstart <- as.double(omg)
    if (omgsc > 0 && omgstart <= omgpars[4]) {
      stop ("Starting value for omg not in the support of its prior")
    }
  }
  omg <- numeric(Nout)
  omg[cumsum(c(1, Nmc[-nch]))] <- omgstart

  if (needkappa) {
    kappasc <- corrtuning[["kappa"]]
  } else {
    kappasc <- 0
    kappa <- 0
  }
  if (is.null(kappasc) || !is.numeric(kappasc) || kappasc < 0)
    stop ("Invalid tuning parameter for kappa.")
  if (kappasc > 0) {
    kappapars <- check_unif_prior(corrpriors[["kappa"]])
  } else kappapars <- c(0, 0)
  if (missing(kappa)) {
    if (kappasc == 0) {
      stop ("Argument kappa needed for fixed kappa")
    } else {
      kappastart <- (kappapars[1] + kappapars[2])*.5
    }
  } else {
    kappastart <- as.double(kappa)
  }
  if (kappasc > 0) {
    kappastart <- .geoBayes_getkappa(kappastart, icf)
    kappapars <- .geoBayes_getkappa(kappapars, icf)
    if (kappastart >= kappapars[2] || kappastart <= kappapars[1]) {
      stop ("Starting value for kappa not in the support of its prior")
    }
  }
  kappa <- numeric(Nout)
  kappa[cumsum(c(1, Nmc[-nch]))] <- kappastart

  ## Run code
  if (test > 0) { # Running a test
    if (is.logical(test)) test <- 100
    test <- as.integer(test)
    acc <- acc_z <- 0L
    tm <- system.time({
      RUN <- .Fortran("trgasamtry_mala", ll = lglk, z = z, phi = phi, omg = omg,
                      kappa = kappa, acc = acc,
                      as.double(ybar), as.double(l), as.double(F),
                      as.double(offset), 
                      as.double(betm0), as.double(betQ0), as.double(ssqdf),
                      as.double(ssqsc), as.double(tsqdf), as.double(tsqsc),
                      as.double(phipars), as.double(omgpars),
                      as.double(kappapars),
                      as.double(phisc), as.double(omgsc),
                      as.double(kappasc), as.integer(icf),
                      as.double(nu), as.double(dm), as.integer(Nout),
                      as.integer(test), as.integer(k), as.integer(p),
                      as.double(malatuning), acc_z = acc_z,
                      PACKAGE = "geoBayes")
    })
    ## Store samples
    ll <- RUN$ll
    zz0 <- matrix(NA, NROW(yy), Nout)
    zz0[ii, ] <- RUN$z
    mm0 <- NULL
    beta <- NULL
    ssq <- NULL
    phi <- RUN$phi
###     attr(phi, 'fixed') <- phisc == 0
    omg <- RUN$omg
###     attr(omg, 'fixed') <- omgsc == 0
###     attr(nu, 'fixed') <- TRUE
    kappa <- RUN$kappa
    acc_ratio <- RUN$acc/Nout
    acc_ratio_z <- RUN$acc_z/Nout
    Nthin <- 1
    Nbi <- 0
###     out <- list(z = zz0, beta = beta, ssq = ssq, phi = phi, omg = omg, nu = nu,
###                 logLik = ll, acc_ratio = acc_ratio, sys_time = tm,
###                 Nout = Nout, Nbi = Nbi, Nthin = Nthin,
###                 response = y, weights = l, modelmatrix = F, family = family,
###                 betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf, ssqsc = ssqsc,
###                 corrfcn = corrfcn, kappa = kappa,
###                 tsqdf = tsqdf, tsqsc = tsqsc,
###                 locations = loc[ii, , drop = FALSE],
###                 longlat = longlat, whichobs = ii)
  } else {
    acc <- acc_z <- integer(nch)
    tm <- system.time({
      RUN <- .Fortran("trgasample_mala", ll = lglk, z = z, z0 = z0,
                      mu = z, mu0 = z0, beta = beta, ssq = ssq,
                      tsq = tsq, phi = phi, omg = omg,
                      kappa = kappa, acc = acc,
                      as.double(ybar),
                      as.double(l), as.double(F),
                      as.double(offset), as.double(F0), as.double(offset0), 
                      as.double(betm0), as.double(betQ0), as.double(ssqdf),
                      as.double(ssqsc), as.double(tsqdf), as.double(tsqsc),
                      as.double(phipars), as.double(omgpars),
                      as.double(kappapars),
                      as.double(phisc), as.double(omgsc), as.double(kappasc),
                      as.integer(icf),
                      as.double(nu), as.double(dm), as.double(dmdm0),
                      as.integer(nch), as.integer(Nmc),
                      as.integer(Nout), as.integer(Nbi), as.integer(Nthin),
                      as.integer(k), as.integer(k0), as.integer(p),
                      as.double(malatuning), acc_z = acc_z,
                      PACKAGE = "geoBayes")
    })
    ## Store samples
    ll <- RUN$ll
    zz0 <- mm0 <- matrix(NA, NROW(yy), Nout)
    zz0[ii, ] <- RUN$z
    zz0[!ii, ] <- RUN$z0
    mm0[ii, ] <- RUN$mu
    mm0[!ii, ] <- RUN$mu0
    beta <- RUN$beta
    ssq <- RUN$ssq
    tsq <- RUN$tsq
    phi <- RUN$phi
###     attr(phi, 'fixed') <- phisc == 0
    omg <- RUN$omg
###     attr(omg, 'fixed') <- omgsc == 0
###     attr(nu, 'fixed') <- TRUE
    kappa <- RUN$kappa
    acc_ratio <- RUN$acc/(Nmc*Nthin + max(Nthin, Nbi))
    acc_ratio_z <- RUN$acc_z/(Nmc*Nthin + max(Nthin, Nbi))
###     out <- list(z = zz0, mu = mm0, beta = beta, ssq = ssq, tsq = tsq,
###                 phi = phi, omg = omg, nu = nu,
###                 logLik = ll, acc_ratio = acc_ratio, sys_time = tm,
###                 Nout = Nout, Nbi = Nbi, Nthin = Nthin,
###                 response = ybar, weights = l, modelmatrix = F, family = family,
###                 betm0 = betm0, betQ0 = betQ0, ssqdf = ssqdf, ssqsc = ssqsc,
###                 corrfcn = corrfcn, kappa = kappa,
###                 tsqdf = tsqdf, tsqsc = tsqsc,
###                 locations = loc[ii, , drop = FALSE],
###                 longlat = longlat, whichobs = ii)
  }
  MCMC <- FIXED <- MODEL <- DATA <- list()
  MCMC$z <- zz0
  MCMC$mu <- mm0
  MCMC$beta <- beta
  MCMC$ssq <- ssq
  MCMC$tsq <- tsq
  FIXED$linkp <- linkp
  FIXED$linkp_num <- nu
  if (phisc == 0) {
    FIXED$phi <- phi[1]
  } else {
    MCMC$phi <- phi
  }
  if (omgsc == 0) {
    FIXED$omg <- omg[1]
  } else {
    MCMC$omg <- omg
  }
  if (kappasc == 0) {
    FIXED$kappa <- kappa[1]
  } else {
    MCMC$kappa <- kappa
  }
  MCMC$logLik <- ll
  MCMC$acc_ratio <- acc_ratio
  MCMC$acc_ratio_z <- acc_ratio_z
  MCMC$sys_time <- tm
  MCMC$Nout <- Nout
  MCMC$Nbi <- Nbi
  MCMC$Nthin <- Nthin
  MCMC$whichobs <- ii
  DATA$response <- ybar
  DATA$weights <- l
  DATA$modelmatrix <- F
  DATA$offset <- offset
  DATA$locations <- loc[ii, , drop = FALSE]
  DATA$longlat <- longlat
  MODEL$family <- family
  MODEL$corrfcn <- corrfcn
  MODEL$betm0 <- betm0
  MODEL$betQ0 <- betQ0
  MODEL$ssqdf <- ssqdf
  MODEL$ssqsc <- ssqsc
  MODEL$tsqdf <- tsqdf
  MODEL$tsqsc <- tsqsc
  MODEL$phipars <- phipars
  MODEL$omgpars <- omgpars
  out <- list(MODEL = MODEL, DATA = DATA, FIXED = FIXED, MCMC = MCMC, call = cl)
  class(out) <- "geomcmc"
  out
}

!! L-BFGS-B is released under the "New BSD License" (aka "Modified BSD
!! License" or "3-clause license")


!! This version allows unlimited redistribution for any purpose as long as
!! its copyright notices and the license's disclaimers of warranty are
!! maintained. The license also contains a clause restricting use of the
!! names of contributors for endorsement of a derived work without specific
!! permission.
!!  
!! Copyright (c) 2011, C. Zhu, R. Byrd, J. Nocedal and J.L Morales
!! All rights reserved.
!!  
!! Redistribution and use in source and binary forms, with or without
!! modification, are permitted provided that the following conditions are met:
!!     * Redistributions of source code must retain the above copyright
!!       notice, this list of conditions and the following disclaimer.
!!     * Redistributions in binary form must reproduce the above copyright
!!       notice, this list of conditions and the following disclaimer in the
!!       documentation and/or other materials provided with the distribution.
!!     * Neither the name of the <organization> nor the
!!       names of its contributors may be used to endorse or promote products
!!       derived from this software without specific prior written permission.
!!  
!! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
!! ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
!! WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
!! DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
!! DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
!! (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
!! LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
!! ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
!! (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
!! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


module lbfgsbmod

  implicit none

  private

  type lbfgsb_store
    logical :: lsv(4)
    integer :: isv(44)
    double precision :: dsv(29)
    character(60) :: csv
    character(60) :: ctask
    integer, pointer :: wrki(:)
    double precision, pointer :: wrka(:)
  end type lbfgsb_store

  public lbfgsb, lbfgsb_store

contains

  subroutine lbfgsb (n,x,xl,xu,xnb,f,g,factr,pgtol,iflag,store)

    integer, intent(in) :: n, xnb(n)
    integer, intent(inout) :: iflag
    double precision, intent(inout) :: x(n)
    double precision, intent(in) :: f, g(n), factr, pgtol, xl(n), xu(n)
    type(lbfgsb_store), intent(inout) :: store
    integer, parameter :: iprint=-1, msv = 10

    integer ::   lws,lr,lz,lt,ld,lsg,lwa,lyg, &
       lsgo,lwy,lsy,lss,lyy,lwt,lwn,lsnd,lygo
    integer, pointer :: iwa(:)
    double precision, pointer :: wa(:)

    if (iflag == 0) then
      store%ctask = 'START'
      store%isv(1)  = msv*n
      store%isv(2)  = msv**2
      store%isv(3)  = 4*msv**2
      store%isv(4)  = 1
      store%isv(5)  = store%isv(4)  + store%isv(1)
      store%isv(6)  = store%isv(5)  + store%isv(1)
      store%isv(7)  = store%isv(6)  + store%isv(2)
      store%isv(8)  = store%isv(7)  + store%isv(2)
      store%isv(9)  = store%isv(8)  + store%isv(2)
      store%isv(10) = store%isv(9)  + store%isv(2)
      store%isv(11) = store%isv(10) + store%isv(3)
      store%isv(12) = store%isv(11) + store%isv(3)
      store%isv(13) = store%isv(12) + n
      store%isv(14) = store%isv(13) + n
      store%isv(15) = store%isv(14) + n
      store%isv(16) = store%isv(15) + n
      store%isv(17) = store%isv(16) + 8*msv
      store%isv(18) = store%isv(17) + msv
      store%isv(19) = store%isv(18) + msv
      store%isv(20) = store%isv(19) + msv
      ! Allocate arrays:
      allocate (store%wrki(3*n))
      allocate (store%wrka(2*msv*n+4*n+12*msv*msv+12*msv))
    end if
    iwa => store%wrki
    wa => store%wrka
    lws  = store%isv(4)
    lwy  = store%isv(5)
    lsy  = store%isv(6)
    lss  = store%isv(7)
    lyy  = store%isv(8)
    lwt  = store%isv(9)
    lwn  = store%isv(10)
    lsnd = store%isv(11)
    lz   = store%isv(12)
    lr   = store%isv(13)
    ld   = store%isv(14)
    lt   = store%isv(15)
    lwa  = store%isv(16)
    lsg  = store%isv(17)
    lsgo = store%isv(18)
    lyg  = store%isv(19)
    lygo = store%isv(20)

    call mainlb(n,msv,x,xl,xu,xnb,f,g,factr,pgtol, &
       wa(lws:lws+n*msv-1),wa(lwy:lwy+n*msv-1),wa(lsy:lsy+msv*msv-1),&
       wa(lss:lss+msv*msv-1),wa(lyy:lyy+msv*msv-1),wa(lwt:lwt+msv*msv-1), &
       wa(lwn:lwn+4*msv*msv-1),wa(lsnd:lsnd+4*msv*msv-1),wa(lz:lz+n-1),&
       wa(lr:lr+n-1), wa(ld:ld+n-1),wa(lt:lt+n-1),wa(lwa:lwa+8*msv-1),&
       wa(lsg:lsg+msv-1),wa(lsgo:lsgo+msv-1),wa(lyg:lyg+ msv-1),&
       wa(lygo:lygo+msv-1), &
       iwa(1:n),iwa(n+1:2*n),iwa(2*n+1:3*n),store%ctask,iprint, &
       store%csv,store%lsv,store%isv(22:44),store%dsv)

    if (store%ctask(1:4) .eq. 'CONV') then
      iflag = 0
    else if (store%ctask(1:4) .eq. 'ABNO') then
      iflag = -1
    else if (store%ctask(1:5) .eq. 'ERROR') then
      iflag = -2
    else if (store%ctask(1:2) .eq. 'FG') then
      iflag = 1
    else if (store%ctask(1:5) .eq. 'NEW_X') then
      iflag = 2
    else
      iflag = -3
    end if
  end subroutine lbfgsb


  subroutine mainlb(n, m, x, l, u, nbd, f, g, factr, pgtol, ws, wy, &
     sy, ss, yy, wt, wn, snd, z, r, d, t, wa, sg, &
     sgo, yg, ygo, index, iwhere, indx2, task, &
     iprint, csave, lsave, isave, dsave)

    character(60) ::     task, csave
    logical ::          lsave(4)
    integer ::          n, m, iprint, nbd(n), index(n), &
       iwhere(n), indx2(n), isave(23)
    double precision :: f, factr, pgtol, &
       x(n), l(n), u(n), g(n), z(n), r(n), d(n), t(n), &
       wa(8*m), sg(m), sgo(m), yg(m), ygo(m), &
       ws(n, m), wy(n, m), sy(m, m), ss(m, m), yy(m, m), &
       wt(m, m), wn(2*m, 2*m), snd(2*m, 2*m), dsave(29)

    ! ************

    ! Subroutine mainlb

    ! This subroutine solves bound constrained optimization problems by
    ! using the compact formula of the limited memory BFGS updates.

    ! n is an integer variable.
    ! On entry n is the number of variables.
    ! On exit n is unchanged.

    ! m is an integer variable.
    ! On entry m is the maximum number of variable metric
    ! corrections allowed in the limited memory matrix.
    ! On exit m is unchanged.

    ! x is a double precision array of dimension n.
    ! On entry x is an approximation to the solution.
    ! On exit x is the current approximation.

    ! l is a double precision array of dimension n.
    ! On entry l is the lower bound of x.
    ! On exit l is unchanged.

    ! u is a double precision array of dimension n.
    ! On entry u is the upper bound of x.
    ! On exit u is unchanged.

    ! nbd is an integer array of dimension n.
    ! On entry nbd represents the type of bounds imposed on the
    ! variables, and must be specified as follows:
    ! nbd(i)=0 if x(i) is unbounded,
    ! 1 if x(i) has only a lower bound,
    ! 2 if x(i) has both lower and upper bounds,
    ! 3 if x(i) has only an upper bound.
    ! On exit nbd is unchanged.

    ! f is a double precision variable.
    ! On first entry f is unspecified.
    ! On final exit f is the value of the function at x.

    ! g is a double precision array of dimension n.
    ! On first entry g is unspecified.
    ! On final exit g is the value of the gradient at x.

    ! factr is a double precision variable.
    ! On entry factr >= 0 is specified by the user.  The iteration
    ! will stop when

    ! (f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr*epsmch

    ! where epsmch is the machine precision, which is automatically
    ! generated by the code.
    ! On exit factr is unchanged.

    ! pgtol is a double precision variable.
    ! On entry pgtol >= 0 is specified by the user.  The iteration
    ! will stop when

    ! max{|proj g_i | i = 1, ..., n} <= pgtol

    ! where pg_i is the ith component of the projected gradient.
    ! On exit pgtol is unchanged.

    ! ws, wy, sy, and wt are double precision working arrays used to
    ! store the following information defining the limited memory
    ! BFGS matrix:
    ! ws, of dimension n x m, stores S, the matrix of s-vectors;
    ! wy, of dimension n x m, stores Y, the matrix of y-vectors;
    ! sy, of dimension m x m, stores S'Y;
    ! ss, of dimension m x m, stores S'S;
    ! yy, of dimension m x m, stores Y'Y;
    ! wt, of dimension m x m, stores the Cholesky factorization
    ! of (theta*S'S+LD^(-1)L'); see eq.
    ! (2.26) in [3].

    ! wn is a double precision working array of dimension 2m x 2m
    ! used to store the LEL^T factorization of the indefinite matrix
    ! K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
    ! [L_a -R_z           theta*S'AA'S ]

    ! where     E = [-I  0]
    ! [ 0  I]

    ! snd is a double precision working array of dimension 2m x 2m
    ! used to store the lower triangular part of
    ! N = [Y' ZZ'Y   L_a'+R_z']
    ! [L_a +R_z  S'AA'S   ]

    ! z(n),r(n),d(n),t(n),wa(8*m) are double precision working arrays.
    ! z is used at different times to store the Cauchy point and
    ! the Newton point.

    ! sg(m),sgo(m),yg(m),ygo(m) are double precision working arrays.

    ! index is an integer working array of dimension n.
    ! In subroutine freev, index is used to store the free and fixed
    ! variables at the Generalized Cauchy Point (GCP).

    ! iwhere is an integer working array of dimension n used to record
    ! the status of the vector x for GCP computation.
    ! iwhere(i)=0 or -3 if x(i) is free and has bounds,
    ! 1       if x(i) is fixed at l(i), and l(i) .ne. u(i)
    ! 2       if x(i) is fixed at u(i), and u(i) .ne. l(i)
    ! 3       if x(i) is always fixed, i.e.,  u(i)=x(i)=l(i)
    ! -1       if x(i) is always free, i.e., no bounds on it.

    ! indx2 is an integer working array of dimension n.
    ! Within subroutine cauchy, indx2 corresponds to the array iorder.
    ! In subroutine freev, a list of variables entering and leaving
    ! the free set is stored in indx2, and it is passed on to
    ! subroutine formk with this information.

    ! task is a working string of characters of length 60 indicating
    ! the current job when entering and leaving this subroutine.

    ! iprint is an INTEGER variable that must be set by the user.
    ! It controls the frequency and type of output generated:
    ! iprint<0    no output is generated;
    ! iprint=0    print only one line at the last iteration;
    ! 0<iprint<99 print also f and |proj g| every iprint iterations;
    ! iprint=99   print details of every iteration except n-vectors;
    ! iprint=100  print also the changes of active set and final x;
    ! iprint>100  print details of every iteration including x and g;
    ! When iprint > 0, the file iterate.dat will be created to
    ! summarize the iteration.

    ! csave is a working string of characters of length 60.

    ! lsave is a logical working array of dimension 4.

    ! isave is an integer working array of dimension 23.

    ! dsave is a double precision working array of dimension 29.


    ! Subprograms called

    ! L-BFGS-B Library ... cauchy, subsm, lnsrlb, formk,

    ! errclb, prn1lb, prn2lb, prn3lb, active, projgr,

    ! freev, cmprlb, matupd, formt.

    ! Minpack2 Library ... timer.

    ! Linpack Library ... dcopy, ddot.


    ! References:

    ! [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
    ! memory algorithm for bound constrained optimization'',
    ! SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.

    ! [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: FORTRAN
    ! Subroutines for Large Scale Bound Constrained Optimization''
    ! Tech. Report, NAM-11, EECS Department, Northwestern University,
    ! 1994.

    ! [3] R. Byrd, J. Nocedal and R. Schnabel "Representations of
    ! Quasi-Newton Matrices and their use in Limited Memory Methods'',
    ! Mathematical Programming 63 (1994), no. 4, pp. 129-156.

    ! (Postscript files of these papers are available via anonymous
    ! ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)

    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    logical ::          prjctd,cnstnd,boxed,updatd,wrk
    integer ::          i,k,nintol,itfile,iback,nskip, &
       head,col,iter,itail,iupdat, &
       nint,nfgv,info,ifun, &
       iword,nfree,nact,ileave,nenter
    double precision :: theta,fold,ddot,dr,rr,tol, &
       xstep,sbgnrm,ddum,dnorm,dtd,epsmch, &
       gd,gdold,stp,stpmx
    double precision :: one,zero
    parameter        (one=1.0d0,zero=0.0d0)

    if (task == 'START') then

      ! Generate the current machine precision.

      epsmch = epsilon(1d0)

      ! Initialize counters and scalars when task='START'.

      ! for the limited memory BFGS matrices:
      col    = 0
      head   = 1
      theta  = one
      iupdat = 0
      updatd = .false.

      ! for operation counts:
      iter   = 0
      nfgv   = 0
      nint   = 0
      nintol = 0
      nskip  = 0
      nfree  = n

      ! for stopping tolerance:
      tol = factr*epsmch

      ! 'info' records the termination information.
      info = 0

      !        if (iprint >= 1) then
      !        ! open a summary file 'iterate.dat'
      !            open (8, file = 'iterate.dat', status = 'unknown')
      !            itfile = 8
      !        endif

      ! Check the input arguments for errors.

      call errclb(n,m,factr,l,u,nbd,task,info,k)
      if (task(1:5) == 'ERROR') then
        return
      endif

      ! Initialize iwhere & project x onto the feasible set.

      call active(n,l,u,nbd,x,iwhere,iprint,prjctd,cnstnd,boxed)

      ! The end of the initialization.

    else
      ! restore local variables.

      prjctd = lsave(1)
      cnstnd = lsave(2)
      boxed  = lsave(3)
      updatd = lsave(4)

      nintol = isave(1)
      itfile = isave(3)
      iback  = isave(4)
      nskip  = isave(5)
      head   = isave(6)
      col    = isave(7)
      itail  = isave(8)
      iter   = isave(9)
      iupdat = isave(10)
      nint   = isave(12)
      nfgv   = isave(13)
      info   = isave(14)
      ifun   = isave(15)
      iword  = isave(16)
      nfree  = isave(17)
      nact   = isave(18)
      ileave = isave(19)
      nenter = isave(20)

      theta  = dsave(1)
      fold   = dsave(2)
      tol    = dsave(3)
      dnorm  = dsave(4)
      epsmch = dsave(5)
      gd     = dsave(11)
      stpmx  = dsave(12)
      sbgnrm = dsave(13)
      stp    = dsave(14)
      gdold  = dsave(15)
      dtd    = dsave(16)

      ! After returning from the driver go to the point where execution
      ! is to resume.

      if (task(1:5) == 'FG_LN') goto 666
      if (task(1:5) == 'NEW_X') goto 777
      if (task(1:5) == 'FG_ST') goto 111
      if (task(1:4) == 'STOP') then
        if (task(7:9) == 'CPU') then
          ! restore the previous iterate.
          call dcopy(n,t,1,x,1)
          call dcopy(n,r,1,g,1)
          f = fold
        endif
        goto 999
      endif
    endif

    ! Compute f0 and g0.

    task = 'FG_START'
    ! return to the driver to calculate f and g; reenter at 111.
    goto 1000
111 continue
    nfgv = 1

    ! Compute the infinity norm of the (-) projected gradient.

    call projgr(n,l,u,nbd,x,g,sbgnrm)

    if (sbgnrm <= pgtol) then
      ! terminate the algorithm.
      task = 'CONVERGENCE: NORM OF PROJECTED GRADIENT <= PGTOL'
      goto 999
    endif

    ! ----------------- the beginning of the loop --------------------------

222 continue
    iword = -1

    if ( .NOT. cnstnd .AND. col > 0) then
      ! skip the search for GCP.
      call dcopy(n,x,1,z,1)
      wrk = updatd
      nint = 0
      goto 333
    endif

    ! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

    ! Compute the Generalized Cauchy Point (GCP).

    ! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

    call cauchy(n,x,l,u,nbd,g,indx2,iwhere,t,d,z, &
       m,wy,ws,sy,wt,theta,col,head, &
       wa(1),wa(2*m+1),wa(4*m+1),wa(6*m+1),nint, &
       sg,yg,iprint,sbgnrm,info,epsmch)
    if (info /= 0) then
      ! singular triangular system detected; refresh the lbfgs memory.
      info   = 0
      col    = 0
      head   = 1
      theta  = one
      iupdat = 0
      updatd = .false.
      goto 222
    endif
    nintol = nintol + nint

    ! Count the entering and leaving variables for iter > 0;
    ! find the index set of free and active variables at the GCP.

    call freev(n,nfree,index,nenter,ileave,indx2, &
       iwhere,wrk,updatd,cnstnd,iprint,iter)

    nact = n - nfree

333 continue

    ! If there are no free variables or B=theta*I, then
    ! skip the subspace minimization.

    if (nfree == 0 .OR. col == 0) goto 555

    ! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

    ! Subspace minimization.

    ! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

    ! Form  the LEL^T factorization of the indefinite
    ! matrix    K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
    ! [L_a -R_z           theta*S'AA'S ]
    ! where     E = [-I  0]
    ! [ 0  I]

    if (wrk) call formk(n,nfree,index,nenter,ileave,indx2,iupdat, &
       updatd,wn,snd,m,ws,wy,sy,theta,col,head,info)
    if (info /= 0) then
      ! nonpositive definiteness in Cholesky factorization;
      ! refresh the lbfgs memory and restart the iteration.
      info   = 0
      col    = 0
      head   = 1
      theta  = one
      iupdat = 0
      updatd = .false.
      goto 222
    endif

    ! compute r=-Z'B(xcp-xk)-Z'g (using wa(2m+1)=W'(xcp-x)
    ! from 'cauchy').
    call cmprlb(n,m,x,g,ws,wy,sy,wt,z,r,wa,index, &
       theta,col,head,nfree,cnstnd,info)
    if (info /= 0) goto 444
    ! call the direct method.
    call subsm(n,m,nfree,index,l,u,nbd,z,r,ws,wy,theta, &
       col,head,iword,wa,wn,iprint,info)
444 continue
    if (info /= 0) then
      ! singular triangular system detected;
      ! refresh the lbfgs memory and restart the iteration.
      info   = 0
      col    = 0
      head   = 1
      theta  = one
      iupdat = 0
      updatd = .false.
      goto 222
    endif

555 continue

    ! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

    ! Line search and optimality tests.

    ! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

    ! Generate the search direction d:=z-x.

    do 40 i = 1, n
      d(i) = z(i) - x(i)
40  ENDDO
666 continue
    call lnsrlb(n,l,u,nbd,x,f,fold,gd,gdold,g,d,r,t,z,stp,dnorm, &
       dtd,xstep,stpmx,iter,ifun,iback,nfgv,info,task, &
       boxed,cnstnd,csave,isave(22),dsave(17))
    if (info /= 0 .OR. iback >= 20) then
      ! restore the previous iterate.
      call dcopy(n,t,1,x,1)
      call dcopy(n,r,1,g,1)
      f = fold
      if (col == 0) then
        ! abnormal termination.
        if (info == 0) then
          info = -9
          ! restore the actual number of f and g evaluations etc.
          nfgv = nfgv - 1
          ifun = ifun - 1
          iback = iback - 1
        endif
        task = 'ABNORMAL_TERMINATION_IN_LNSRCH'
        iter = iter + 1
        goto 999
      else
        ! refresh the lbfgs memory and restart the iteration.
        if (info == 0) nfgv = nfgv - 1
        info   = 0
        col    = 0
        head   = 1
        theta  = one
        iupdat = 0
        updatd = .false.
        task   = 'RESTART_FROM_LNSRCH'
        goto 222
      endif
    else if (task(1:5) == 'FG_LN') then
      ! return to the driver for calculating f and g; reenter at 666.
      goto 1000
    else
      ! calculate and print out the quantities related to the new X.
      iter = iter + 1

      ! Compute the infinity norm of the projected (-)gradient.

      call projgr(n,l,u,nbd,x,g,sbgnrm)

      goto 1000
    endif
777 continue

    ! Test for termination.

    if (sbgnrm <= pgtol) then
      ! terminate the algorithm.
      task = 'CONVERGENCE: NORM OF PROJECTED GRADIENT <= PGTOL'
      goto 999
    endif

    ddum = max(abs(fold), abs(f), one)
    if ((fold - f) <= tol*ddum) then
      ! terminate the algorithm.
      task = 'CONVERGENCE: REL_REDUCTION_OF_F <= FACTR*EPSMCH'
      if (iback >= 10) info = -5
      ! i.e., to issue a warning if iback>10 in the line search.
      goto 999
    endif

    ! Compute d=newx-oldx, r=newg-oldg, rr=y'y and dr=y's.

    do 42 i = 1, n
      r(i) = g(i) - r(i)
42  ENDDO
    rr = ddot(n,r,1,r,1)
    if (stp == one) then
      dr = gd - gdold
      ddum = -gdold
    else
      dr = (gd - gdold)*stp
      call dscal(n,stp,d,1)
      ddum = -gdold*stp
    endif

    if (dr <= epsmch*ddum) then
      ! skip the L-BFGS update.
      nskip = nskip + 1
      updatd = .false.
      goto 888
    endif

    ! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

    ! Update the L-BFGS matrix.

    ! cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

    updatd = .true.
    iupdat = iupdat + 1

    ! Update matrices WS and WY and form the middle matrix in B.

    call matupd(n,m,ws,wy,sy,ss,d,r,itail, &
       iupdat,col,head,theta,rr,dr,stp,dtd)

    ! Form the upper half of the pds T = theta*SS + L*D^(-1)*L';
    ! Store T in the upper triangular of the array wt;
    ! Cholesky factorize T to J*J' with
    ! J' stored in the upper triangular of wt.

    call formt(m,wt,sy,ss,col,theta,info)

    if (info /= 0) then
      ! nonpositive definiteness in Cholesky factorization;
      ! refresh the lbfgs memory and restart the iteration.
      info = 0
      col = 0
      head = 1
      theta = one
      iupdat = 0
      updatd = .false.
      goto 222
    endif

    ! Now the inverse of the middle matrix in B is

    ! [  D^(1/2)      O ] [ -D^(1/2)  D^(-1/2)*L' ]
    ! [ -L*D^(-1/2)   J ] [  0        J'          ]

888 continue

    ! -------------------- the end of the loop -----------------------------

    goto 222
999 continue
1000 continue

    ! Save local variables.

    lsave(1)  = prjctd
    lsave(2)  = cnstnd
    lsave(3)  = boxed
    lsave(4)  = updatd

    isave(1)  = nintol
    isave(3)  = itfile
    isave(4)  = iback
    isave(5)  = nskip
    isave(6)  = head
    isave(7)  = col
    isave(8)  = itail
    isave(9)  = iter
    isave(10) = iupdat
    isave(12) = nint
    isave(13) = nfgv
    isave(14) = info
    isave(15) = ifun
    isave(16) = iword
    isave(17) = nfree
    isave(18) = nact
    isave(19) = ileave
    isave(20) = nenter

    dsave(1)  = theta
    dsave(2)  = fold
    dsave(3)  = tol
    dsave(4)  = dnorm
    dsave(5)  = epsmch
    dsave(11) = gd
    dsave(12) = stpmx
    dsave(13) = sbgnrm
    dsave(14) = stp
    dsave(15) = gdold
    dsave(16) = dtd

    return

  end subroutine mainlb

  !======================= The end of mainlb =============================

  subroutine active(n, l, u, nbd, x, iwhere, iprint, &
     prjctd, cnstnd, boxed)

    logical ::          prjctd, cnstnd, boxed
    integer ::          n, iprint, nbd(n), iwhere(n)
    double precision :: x(n), l(n), u(n)

    ! ************

    ! Subroutine active

    ! This subroutine initializes iwhere and projects the initial x to
    ! the feasible set if necessary.

    ! iwhere is an integer array of dimension n.
    ! On entry iwhere is unspecified.
    ! On exit iwhere(i)=-1  if x(i) has no bounds
    ! 3   if l(i)=u(i)
    ! 0   otherwise.
    ! In cauchy, iwhere is given finer gradations.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer ::          nbdd,i
    double precision :: zero
    parameter        (zero=0.0d0)

    ! Initialize nbdd, prjctd, cnstnd and boxed.

    nbdd = 0
    prjctd = .false.
    cnstnd = .false.
    boxed = .true.

    ! Project the initial x to the easible set if necessary.

    do 10 i = 1, n
      if (nbd(i) > 0) then
        if (nbd(i) <= 2 .AND. x(i) <= l(i)) then
          if (x(i) < l(i)) then
            prjctd = .true.
            x(i) = l(i)
          endif
          nbdd = nbdd + 1
        else if (nbd(i) >= 2 .AND. x(i) >= u(i)) then
          if (x(i) > u(i)) then
            prjctd = .true.
            x(i) = u(i)
          endif
          nbdd = nbdd + 1
        endif
      endif
10  ENDDO

    ! Initialize iwhere and assign values to cnstnd and boxed.

    do 20 i = 1, n
      if (nbd(i) /= 2) boxed = .FALSE. 
      if (nbd(i) == 0) then
        ! this variable is always free
        iwhere(i) = -1

        ! otherwise set x(i)=mid(x(i), u(i), l(i)).
      else
        cnstnd = .true.
        if (nbd(i) == 2 .AND. u(i) - l(i) <= zero) then
          ! this variable is always fixed
          iwhere(i) = 3
        else
          iwhere(i) = 0
        endif
      endif
20  ENDDO

    return

  end subroutine active

  !======================= The end of active =============================

  subroutine bmv(m, sy, wt, col, v, p, info)

    integer :: m, col, info
    double precision :: sy(m, m), wt(m, m), v(2*col), p(2*col)

    ! ************

    ! Subroutine bmv

    ! This subroutine computes the product of the 2m x 2m middle matrix
    !	in the compact L-BFGS formula of B and a 2m vector v;
    !	it returns the product in p.

    ! m is an integer variable.
    ! On entry m is the maximum number of variable metric corrections
    ! used to define the limited memory matrix.
    ! On exit m is unchanged.

    ! sy is a double precision array of dimension m x m.
    ! On entry sy specifies the matrix S'Y.
    ! On exit sy is unchanged.

    ! wt is a double precision array of dimension m x m.
    ! On entry wt specifies the upper triangular matrix J' which is
    ! the Cholesky factor of (thetaS'S+LD^(-1)L').
    ! On exit wt is unchanged.

    ! col is an integer variable.
    ! On entry col specifies the number of s-vectors (or y-vectors)
    ! stored in the compact L-BFGS formula.
    ! On exit col is unchanged.

    ! v is a double precision array of dimension 2col.
    ! On entry v specifies vector v.
    ! On exit v is unchanged.

    ! p is a double precision array of dimension 2col.
    ! On entry p is unspecified.
    ! On exit p is the product Mv.

    ! info is an integer variable.
    ! On entry info is unspecified.
    ! On exit info = 0       for normal return,
    ! = nonzero for abnormal return when the system
    ! to be solved by dtrsl is singular.

    ! Subprograms called:

    ! Linpack ... dtrsl.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer ::          i,k,i2
    double precision :: sum

    if (col == 0) return

    ! PART I: solve [  D^(1/2)      O ] [ p1 ] = [ v1 ]
    ! [ -L*D^(-1/2)   J ] [ p2 ]   [ v2 ].

    ! solve Jp2=v2+LD^(-1)v1.
    p(col + 1) = v(col + 1)
    do 20 i = 2, col
      i2 = col + i
      sum = 0.0d0
      do 10 k = 1, i - 1
        sum = sum + sy(i,k)*v(k)/sy(k,k)
10    ENDDO
      p(i2) = v(i2) + sum
20  ENDDO
    ! Solve the triangular system
    call dtrsl(wt,m,col,p(col+1),11,info)
    if (info /= 0) return

    ! solve D^(1/2)p1=v1.
    do 30 i = 1, col
      p(i) = v(i)/sqrt(sy(i,i))
30  ENDDO

    ! PART II: solve [ -D^(1/2)   D^(-1/2)*L'  ] [ p1 ] = [ p1 ]
    ! [  0         J'           ] [ p2 ]   [ p2 ].

    ! solve J^Tp2=p2.
    call dtrsl(wt,m,col,p(col+1),01,info)
    if (info /= 0) return

    ! compute p1=-D^(-1/2)(p1-D^(-1/2)L'p2)
    ! =-D^(-1/2)p1+D^(-1)L'p2.
    do 40 i = 1, col
      p(i) = -p(i)/sqrt(sy(i,i))
40  ENDDO
    do 60 i = 1, col
      sum = 0.d0
      do 50 k = i + 1, col
        sum = sum + sy(k,i)*p(col+k)/sy(i,i)
50    ENDDO
      p(i) = p(i) + sum
60  ENDDO

    return

  end subroutine bmv

  !======================== The end of bmv ===============================

  subroutine cauchy(n, x, l, u, nbd, g, iorder, iwhere, t, d, xcp, &
     m, wy, ws, sy, wt, theta, col, head, p, c, wbp, &
     v, nint, sg, yg, iprint, sbgnrm, info, epsmch)

    integer ::          n, m, head, col, nint, iprint, info, &
       nbd(n), iorder(n), iwhere(n)
    double precision :: theta, epsmch, &
       x(n), l(n), u(n), g(n), t(n), d(n), xcp(n), &
       sg(m), yg(m), wy(n, col), ws(n, col), sy(m, m), &
       wt(m, m), p(2*m), c(2*m), wbp(2*m), v(2*m)

    ! ************

    ! Subroutine cauchy

    ! For given x, l, u, g (with sbgnrm > 0), and a limited memory
    ! BFGS matrix B defined in terms of matrices WY, WS, WT, and
    ! scalars head, col, and theta, this subroutine computes the
    ! generalized Cauchy point (GCP), defined as the first local
    ! minimizer of the quadratic

    ! Q(x + s) = g's + 1/2 s'Bs

    ! along the projected gradient direction P(x-tg,l,u).
    ! The routine returns the GCP in xcp.

    ! n is an integer variable.
    ! On entry n is the dimension of the problem.
    ! On exit n is unchanged.

    ! x is a double precision array of dimension n.
    ! On entry x is the starting point for the GCP computation.
    ! On exit x is unchanged.

    ! l is a double precision array of dimension n.
    ! On entry l is the lower bound of x.
    ! On exit l is unchanged.

    ! u is a double precision array of dimension n.
    ! On entry u is the upper bound of x.
    ! On exit u is unchanged.

    ! nbd is an integer array of dimension n.
    ! On entry nbd represents the type of bounds imposed on the
    ! variables, and must be specified as follows:
    ! nbd(i)=0 if x(i) is unbounded,
    ! 1 if x(i) has only a lower bound,
    ! 2 if x(i) has both lower and upper bounds, and
    ! 3 if x(i) has only an upper bound.
    ! On exit nbd is unchanged.

    ! g is a double precision array of dimension n.
    ! On entry g is the gradient of f(x).  g must be a nonzero vector.
    ! On exit g is unchanged.

    ! iorder is an integer working array of dimension n.
    ! iorder will be used to store the breakpoints in the piecewise
    ! linear path and free variables encountered. On exit,
    ! iorder(1),...,iorder(nleft) are indices of breakpoints
    ! which have not been encountered;
    ! iorder(nleft+1),...,iorder(nbreak) are indices of
    ! encountered breakpoints; and
    ! iorder(nfree),...,iorder(n) are indices of variables which
    ! have no bound constraits along the search direction.

    ! iwhere is an integer array of dimension n.
    ! On entry iwhere indicates only the permanently fixed (iwhere=3)
    ! or free (iwhere= -1) components of x.
    ! On exit iwhere records the status of the current x variables.
    ! iwhere(i)=-3  if x(i) is free and has bounds, but is not moved
    ! 0   if x(i) is free and has bounds, and is moved
    ! 1   if x(i) is fixed at l(i), and l(i) .ne. u(i)
    ! 2   if x(i) is fixed at u(i), and u(i) .ne. l(i)
    ! 3   if x(i) is always fixed, i.e.,  u(i)=x(i)=l(i)
    ! -1  if x(i) is always free, i.e., it has no bounds.

    ! t is a double precision working array of dimension n.
    ! t will be used to store the break points.

    ! d is a double precision array of dimension n used to store
    ! the Cauchy direction P(x-tg)-x.

    ! xcp is a double precision array of dimension n used to return the
    ! GCP on exit.

    ! m is an integer variable.
    ! On entry m is the maximum number of variable metric corrections
    ! used to define the limited memory matrix.
    ! On exit m is unchanged.

    ! ws, wy, sy, and wt are double precision arrays.
    ! On entry they store information that defines the
    ! limited memory BFGS matrix:
    ! ws(n,m) stores S, a set of s-vectors;
    ! wy(n,m) stores Y, a set of y-vectors;
    ! sy(m,m) stores S'Y;
    ! wt(m,m) stores the
    ! Cholesky factorization of (theta*S'S+LD^(-1)L').
    ! On exit these arrays are unchanged.

    ! theta is a double precision variable.
    ! On entry theta is the scaling factor specifying B_0 = theta I.
    ! On exit theta is unchanged.

    ! col is an integer variable.
    ! On entry col is the actual number of variable metric
    ! corrections stored so far.
    ! On exit col is unchanged.

    ! head is an integer variable.
    ! On entry head is the location of the first s-vector (or y-vector)
    ! in S (or Y).
    ! On exit col is unchanged.

    ! p is a double precision working array of dimension 2m.
    ! p will be used to store the vector p = W^(T)d.

    ! c is a double precision working array of dimension 2m.
    ! c will be used to store the vector c = W^(T)(xcp-x).

    ! wbp is a double precision working array of dimension 2m.
    ! wbp will be used to store the row of W corresponding
    ! to a breakpoint.

    ! v is a double precision working array of dimension 2m.

    ! nint is an integer variable.
    ! On exit nint records the number of quadratic segments explored
    ! in searching for the GCP.

    ! sg and yg are double precision arrays of dimension m.
    ! On entry sg  and yg store S'g and Y'g correspondingly.
    ! On exit they are unchanged.

    ! iprint is an INTEGER variable that must be set by the user.
    ! It controls the frequency and type of output generated:
    ! iprint<0    no output is generated;
    ! iprint=0    print only one line at the last iteration;
    ! 0<iprint<99 print also f and |proj g| every iprint iterations;
    ! iprint=99   print details of every iteration except n-vectors;
    ! iprint=100  print also the changes of active set and final x;
    ! iprint>100  print details of every iteration including x and g;
    ! When iprint > 0, the file iterate.dat will be created to
    ! summarize the iteration.

    ! sbgnrm is a double precision variable.
    ! On entry sbgnrm is the norm of the projected gradient at x.
    ! On exit sbgnrm is unchanged.

    ! info is an integer variable.
    ! On entry info is 0.
    ! On exit info = 0       for normal return,
    ! = nonzero for abnormal return when the the system
    ! used in routine bmv is singular.

    ! Subprograms called:

    ! L-BFGS-B Library ... hpsolb, bmv.

    ! Linpack ... dscal dcopy, daxpy.


    ! References:

    ! [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
    ! memory algorithm for bound constrained optimization'',
    ! SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.

    ! [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: FORTRAN
    ! Subroutines for Large Scale Bound Constrained Optimization''
    ! Tech. Report, NAM-11, EECS Department, Northwestern University,
    ! 1994.

    ! (Postscript files of these papers are available via anonymous
    ! ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)

    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    logical ::          xlower,xupper,bnded
    integer ::          i,j,col2,nfree,nbreak,pointr, &
       ibp,nleft,ibkmin,iter
    double precision :: f1,f2,dt,dtm,tsum,dibp,zibp,dibp2,bkmin, &
       tu,tl,wmc,wmp,wmw,ddot,tj,tj0,neggi,sbgnrm, &
       f2_org
    double precision :: one,zero
    parameter        (one=1.0d0,zero=0.0d0)

    ! Check the status of the variables, reset iwhere(i) if necessary;
    ! compute the Cauchy direction d and the breakpoints t; initialize
    ! the derivative f1 and the vector p = W'd (for theta = 1).

    if (sbgnrm <= zero) then
      call dcopy(n,x,1,xcp,1)
      return
    endif
    bnded = .true.
    nfree = n + 1
    nbreak = 0
    ibkmin = 0
    bkmin = zero
    col2 = 2*col
    f1 = zero

    ! We set p to zero and build it up as we determine d.

    do 20 i = 1, col2
      p(i) = zero
20  ENDDO

    ! In the following loop we determine for each variable its bound
    ! status and its breakpoint, and update p accordingly.
    ! Smallest breakpoint is identified.

    do 50 i = 1, n
      neggi = -g(i)
      if (iwhere(i) /= 3 .AND. iwhere(i) /= -1) then
        ! if x(i) is not a constant and has bounds,
        ! compute the difference between x(i) and its bounds.
        if (nbd(i) <= 2) tl = x(i) - l(i)
        if (nbd(i) >= 2) tu = u(i) - x(i)

        ! If a variable is close enough to a bound
        ! we treat it as at bound.
        xlower = nbd(i) .le. 2 .and. tl .le. zero
        xupper = nbd(i) .ge. 2 .and. tu .le. zero

        ! reset iwhere(i).
        iwhere(i) = 0
        if (xlower) then
          if (neggi <= zero) iwhere(i) = 1
        else if (xupper) then
          if (neggi >= zero) iwhere(i) = 2
        else
          if (abs(neggi) <= zero) iwhere(i) = -3
        endif
      endif
      pointr = head
      if (iwhere(i) /= 0 .AND. iwhere(i) /= -1) then
        d(i) = zero
      else
        d(i) = neggi
        f1 = f1 - neggi*neggi
        ! calculate p := p - W'e_i* (g_i).
        do 40 j = 1, col
          p(j) = p(j) +  wy(i,pointr)* neggi
          p(col + j) = p(col + j) + ws(i,pointr)*neggi
          pointr = mod(pointr,m) + 1
40      ENDDO
        if (nbd(i) <= 2 .AND. nbd(i) /= 0 &
           .AND. neggi < zero) then
          ! x(i) + d(i) is bounded; compute t(i).
          nbreak = nbreak + 1
          iorder(nbreak) = i
          t(nbreak) = tl/(-neggi)
          if (nbreak == 1 .OR. t(nbreak) < bkmin) then
            bkmin = t(nbreak)
            ibkmin = nbreak
          endif
        else if (nbd(i) >= 2 .AND. neggi > zero) then
          ! x(i) + d(i) is bounded; compute t(i).
          nbreak = nbreak + 1
          iorder(nbreak) = i
          t(nbreak) = tu/neggi
          if (nbreak == 1 .OR. t(nbreak) < bkmin) then
            bkmin = t(nbreak)
            ibkmin = nbreak
          endif
        else
          ! x(i) + d(i) is not bounded.
          nfree = nfree - 1
          iorder(nfree) = i
          if (abs(neggi) > zero) bnded = .FALSE. 
        endif
      endif
50  ENDDO

    ! The indices of the nonzero components of d are now stored
    ! in iorder(1),...,iorder(nbreak) and iorder(nfree),...,iorder(n).
    ! The smallest of the nbreak breakpoints is in t(ibkmin)=bkmin.

    if (theta /= one) then
      ! complete the initialization of p for theta not= one.
      call dscal(col,theta,p(col+1),1)
    endif

    ! Initialize GCP xcp = x.

    call dcopy(n,x,1,xcp,1)

    if (nbreak == 0 .AND. nfree == n + 1) then
      ! is a zero vector, return with the initial xcp as GCP.
      return
    endif

    ! Initialize c = W'(xcp - x) = 0.

    do 60 j = 1, col2
      c(j) = zero
60  ENDDO

    ! Initialize derivative f2.

    f2 =  -theta*f1
    f2_org  =  f2
    if (col > 0) then
      call bmv(m,sy,wt,col,p,v,info)
      if (info /= 0) return
      f2 = f2 - ddot(col2,v,1,p,1)
    endif
    dtm = -f1/f2
    tsum = zero
    nint = 1

    ! If there are no breakpoints, locate the GCP and return.

    if (nbreak == 0) goto 888

    nleft = nbreak
    iter = 1


    tj = zero

    !------------------- the beginning of the loop -------------------------

777 continue

    ! Find the next smallest breakpoint;
    ! compute dt = t(nleft) - t(nleft + 1).

    tj0 = tj
    if (iter == 1) then
      ! Since we already have the smallest breakpoint we need not do
      ! heapsort yet. Often only one breakpoint is used and the
      ! cost of heapsort is avoided.
      tj = bkmin
      ibp = iorder(ibkmin)
    else
      if (iter == 2) then
        ! Replace the already used smallest breakpoint with the
        ! breakpoint numbered nbreak > nlast, before heapsort call.
        if (ibkmin /= nbreak) then
          t(ibkmin) = t(nbreak)
          iorder(ibkmin) = iorder(nbreak)
        endif
        ! Update heap structure of breakpoints
        ! (if iter=2, initialize heap).
      endif
      call hpsolb(nleft,t,iorder,iter-2)
      tj = t(nleft)
      ibp = iorder(nleft)
    endif

    dt = tj - tj0


    ! If a minimizer is within this interval, locate the GCP and return.

    if (dtm < dt) goto 888

    ! Otherwise fix one variable and
    ! reset the corresponding component of d to zero.

    tsum = tsum + dt
    nleft = nleft - 1
    iter = iter + 1
    dibp = d(ibp)
    d(ibp) = zero
    if (dibp > zero) then
      zibp = u(ibp) - x(ibp)
      xcp(ibp) = u(ibp)
      iwhere(ibp) = 2
    else
      zibp = l(ibp) - x(ibp)
      xcp(ibp) = l(ibp)
      iwhere(ibp) = 1
    endif
    if (nleft == 0 .AND. nbreak == n) then
      ! all n variables are fixed,
      ! return with xcp as GCP.
      dtm = dt
      goto 999
    endif

    ! Update the derivative information.

    nint = nint + 1
    dibp2 = dibp**2

    ! Update f1 and f2.

    ! temporarily set f1 and f2 for col=0.
    f1 = f1 + dt*f2 + dibp2 - theta*dibp*zibp
    f2 = f2 - theta*dibp2

    if (col > 0) then
      ! update c = c + dt*p.
      call daxpy(col2,dt,p,1,c,1)

      ! choose wbp,
      ! the row of W corresponding to the breakpoint encountered.
      pointr = head
      do 70 j = 1,col
        wbp(j) = wy(ibp,pointr)
        wbp(col + j) = theta*ws(ibp,pointr)
        pointr = mod(pointr,m) + 1
70    ENDDO

      ! compute (wbp)Mc, (wbp)Mp, and (wbp)M(wbp)'.
      call bmv(m,sy,wt,col,wbp,v,info)
      if (info /= 0) return
      wmc = ddot(col2,c,1,v,1)
      wmp = ddot(col2,p,1,v,1)
      wmw = ddot(col2,wbp,1,v,1)

      ! update p = p - dibp*wbp.
      call daxpy(col2,-dibp,wbp,1,p,1)

      ! complete updating f1 and f2 while col > 0.
      f1 = f1 + dibp*wmc
      f2 = f2 + 2.0d0*dibp*wmp - dibp2*wmw
    endif

    f2 = max(epsmch*f2_org,f2)
    if (nleft > 0) then
      dtm = -f1/f2
      goto 777
      ! to repeat the loop for unsearched intervals.
    else if(bnded) then
      f1 = zero
      f2 = zero
      dtm = zero
    else
      dtm = -f1/f2
    endif

    !------------------- the end of the loop -------------------------------

888 continue
    if (dtm <= zero) dtm = zero
    tsum = tsum + dtm

    ! Move free variables (i.e., the ones w/o breakpoints) and
    ! the variables whose breakpoints haven't been reached.

    call daxpy(n,tsum,d,1,xcp,1)

999 continue

    ! Update c = c + dtm*p = W'(x^c - x)
    ! which will be used in computing r = Z'(B(x^c - x) + g).

    if (col > 0) call daxpy(col2,dtm,p,1,c,1)

    return

  end subroutine cauchy

  !====================== The end of cauchy ==============================

  subroutine cmprlb(n, m, x, g, ws, wy, sy, wt, z, r, wa, index, &
     theta, col, head, nfree, cnstnd, info)

    logical ::          cnstnd
    integer ::          n, m, col, head, nfree, info, index(n)
    double precision :: theta, &
       x(n), g(n), z(n), r(n), wa(4*m), &
       ws(n, m), wy(n, m), sy(m, m), wt(m, m)

    ! ************

    ! Subroutine cmprlb

    ! This subroutine computes r=-Z'B(xcp-xk)-Z'g by using
    ! wa(2m+1)=W'(xcp-x) from subroutine cauchy.

    ! Subprograms called:

    ! L-BFGS-B Library ... bmv.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer ::          i,j,k,pointr
    double precision :: a1,a2

    if ( .NOT. cnstnd .AND. col > 0) then
      do 26 i = 1, n
        r(i) = -g(i)
26    ENDDO
    else
      do 30 i = 1, nfree
        k = index(i)
        r(i) = -theta*(z(k) - x(k)) - g(k)
30    ENDDO
      call bmv(m,sy,wt,col,wa(2*m+1),wa(1),info)
      if (info /= 0) then
        info = -8
        return
      endif
      pointr = head
      do 34 j = 1, col
        a1 = wa(j)
        a2 = theta*wa(col + j)
        do 32 i = 1, nfree
          k = index(i)
          r(i) = r(i) + wy(k,pointr)*a1 + ws(k,pointr)*a2
32      ENDDO
        pointr = mod(pointr,m) + 1
34    ENDDO
    endif

    return

  end subroutine cmprlb

  !======================= The end of cmprlb =============================

  subroutine errclb(n, m, factr, l, u, nbd, task, info, k)

    character(60) ::     task
    integer ::          n, m, info, k, nbd(n)
    double precision :: factr, l(n), u(n)

    ! ************

    ! Subroutine errclb

    ! This subroutine checks the validity of the input data.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer ::          i
    double precision :: one,zero
    parameter        (one=1.0d0,zero=0.0d0)

    ! Check the input arguments for errors.

    if (n <= 0) task = 'ERROR: N <= 0'
    if (m <= 0) task = 'ERROR: M <= 0'
    if (factr < zero) task = 'ERROR: FACTR < 0'

    ! Check the validity of the arrays nbd(i), u(i), and l(i).

    do 10 i = 1, n
      if (nbd(i) < 0 .OR. nbd(i) > 3) then
        ! return
        task = 'ERROR: INVALID NBD'
        info = -6
        k = i
      endif
      if (nbd(i) == 2) then
        if (l(i) > u(i)) then
          ! return
          task = 'ERROR: NO FEASIBLE SOLUTION'
          info = -7
          k = i
        endif
      endif
10  ENDDO

    return

  end subroutine errclb

  !======================= The end of errclb =============================

  subroutine formk(n, nsub, ind, nenter, ileave, indx2, iupdat, &
     updatd, wn, wn1, m, ws, wy, sy, theta, col, &
     head, info)

    integer ::          n, nsub, m, col, head, nenter, ileave, iupdat, &
       info, ind(n), indx2(n)
    double precision :: theta, wn(2*m, 2*m), wn1(2*m, 2*m), &
       ws(n, m), wy(n, m), sy(m, m)
    logical ::          updatd

    ! ************

    ! Subroutine formk

    ! This subroutine forms  the LEL^T factorization of the indefinite

    ! matrix    K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
    ! [L_a -R_z           theta*S'AA'S ]
    ! where E = [-I  0]
    ! [ 0  I]
    ! The matrix K can be shown to be equal to the matrix M^[-1]N
    ! occurring in section 5.1 of [1], as well as to the matrix
    ! Mbar^[-1] Nbar in section 5.3.

    ! n is an integer variable.
    ! On entry n is the dimension of the problem.
    ! On exit n is unchanged.

    ! nsub is an integer variable
    ! On entry nsub is the number of subspace variables in free set.
    ! On exit nsub is not changed.

    ! ind is an integer array of dimension nsub.
    ! On entry ind specifies the indices of subspace variables.
    ! On exit ind is unchanged.

    ! nenter is an integer variable.
    ! On entry nenter is the number of variables entering the
    ! free set.
    ! On exit nenter is unchanged.

    ! ileave is an integer variable.
    ! On entry indx2(ileave),...,indx2(n) are the variables leaving
    ! the free set.
    ! On exit ileave is unchanged.

    ! indx2 is an integer array of dimension n.
    ! On entry indx2(1),...,indx2(nenter) are the variables entering
    ! the free set, while indx2(ileave),...,indx2(n) are the
    ! variables leaving the free set.
    ! On exit indx2 is unchanged.

    ! iupdat is an integer variable.
    ! On entry iupdat is the total number of BFGS updates made so far.
    ! On exit iupdat is unchanged.

    ! updatd is a logical variable.
    ! On entry 'updatd' is true if the L-BFGS matrix is updatd.
    ! On exit 'updatd' is unchanged.

    ! wn is a double precision array of dimension 2m x 2m.
    ! On entry wn is unspecified.
    ! On exit the upper triangle of wn stores the LEL^T factorization
    ! of the 2*col x 2*col indefinite matrix
    ! [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
    ! [L_a -R_z           theta*S'AA'S ]

    ! wn1 is a double precision array of dimension 2m x 2m.
    ! On entry wn1 stores the lower triangular part of
    ! [Y' ZZ'Y   L_a'+R_z']
    ! [L_a+R_z   S'AA'S   ]
    ! in the previous iteration.
    ! On exit wn1 stores the corresponding updated matrices.
    ! The purpose of wn1 is just to store these inner products
    ! so they can be easily updated and inserted into wn.

    ! m is an integer variable.
    ! On entry m is the maximum number of variable metric corrections
    ! used to define the limited memory matrix.
    ! On exit m is unchanged.

    ! ws, wy, sy, and wtyy are double precision arrays;
    ! theta is a double precision variable;
    ! col is an integer variable;
    ! head is an integer variable.
    ! On entry they store the information defining the
    ! limited memory BFGS matrix:
    ! ws(n,m) stores S, a set of s-vectors;
    ! wy(n,m) stores Y, a set of y-vectors;
    ! sy(m,m) stores S'Y;
    ! wtyy(m,m) stores the Cholesky factorization
    ! of (theta*S'S+LD^(-1)L')
    ! theta is the scaling factor specifying B_0 = theta I;
    ! col is the number of variable metric corrections stored;
    ! head is the location of the 1st s- (or y-) vector in S (or Y).
    ! On exit they are unchanged.

    ! info is an integer variable.
    ! On entry info is unspecified.
    ! On exit info =  0 for normal return;
    ! = -1 when the 1st Cholesky factorization failed;
    ! = -2 when the 2st Cholesky factorization failed.

    ! Subprograms called:

    ! Linpack ... dcopy, dpofa, dtrsl.


    ! References:
    ! [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
    ! memory algorithm for bound constrained optimization'',
    ! SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.

    ! [2] C. Zhu, R.H. Byrd, P. Lu, J. Nocedal, ``L-BFGS-B: a
    ! limited memory FORTRAN code for solving bound constrained
    ! optimization problems'', Tech. Report, NAM-11, EECS Department,
    ! Northwestern University, 1994.

    ! (Postscript files of these papers are available via anonymous
    ! ftp to eecs.nwu.edu in the directory pub/lbfgs/lbfgs_bcm.)

    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer ::          m2,ipntr,jpntr,iy,is,jy,js,is1,js1,k1,i,k, &
       col2,pbegin,pend,dbegin,dend,upcl
    double precision :: ddot,temp1,temp2,temp3,temp4
    double precision :: one,zero
    parameter        (one=1.0d0,zero=0.0d0)

    ! Form the lower triangular part of
    ! WN1 = [Y' ZZ'Y   L_a'+R_z']
    ! [L_a+R_z   S'AA'S   ]
    ! where L_a is the strictly lower triangular part of S'AA'Y
    ! R_z is the upper triangular part of S'ZZ'Y.

    if (updatd) then
      if (iupdat > m) then
        ! shift old part of WN1.
        do 10 jy = 1, m - 1
          js = m + jy
          call dcopy(m-jy,wn1(jy+1,jy+1),1,wn1(jy,jy),1)
          call dcopy(m-jy,wn1(js+1,js+1),1,wn1(js,js),1)
          call dcopy(m-1,wn1(m+2,jy+1),1,wn1(m+1,jy),1)
10      ENDDO
      endif

      ! put new rows in blocks (1,1), (2,1) and (2,2).
      pbegin = 1
      pend = nsub
      dbegin = nsub + 1
      dend = n
      iy = col
      is = m + col
      ipntr = head + col - 1
      if (ipntr > m) ipntr = ipntr - m
      jpntr = head
      do 20 jy = 1, col
        js = m + jy
        temp1 = zero
        temp2 = zero
        temp3 = zero
        ! compute element jy of row 'col' of Y'ZZ'Y
        do 15 k = pbegin, pend
          k1 = ind(k)
          temp1 = temp1 + wy(k1,ipntr)*wy(k1,jpntr)
15      ENDDO
        ! compute elements jy of row 'col' of L_a and S'AA'S
        do 16 k = dbegin, dend
          k1 = ind(k)
          temp2 = temp2 + ws(k1,ipntr)*ws(k1,jpntr)
          temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
16      ENDDO
        wn1(iy,jy) = temp1
        wn1(is,js) = temp2
        wn1(is,jy) = temp3
        jpntr = mod(jpntr,m) + 1
20    ENDDO

      ! put new column in block (2,1).
      jy = col
      jpntr = head + col - 1
      if (jpntr > m) jpntr = jpntr - m
      ipntr = head
      do 30 i = 1, col
        is = m + i
        temp3 = zero
        ! compute element i of column 'col' of R_z
        do 25 k = pbegin, pend
          k1 = ind(k)
          temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
25      ENDDO
        ipntr = mod(ipntr,m) + 1
        wn1(is,jy) = temp3
30    ENDDO
      upcl = col - 1
    else
      upcl = col
    endif

    ! modify the old parts in blocks (1,1) and (2,2) due to changes
    ! in the set of free variables.
    ipntr = head
    do 45 iy = 1, upcl
      is = m + iy
      jpntr = head
      do 40 jy = 1, iy
        js = m + jy
        temp1 = zero
        temp2 = zero
        temp3 = zero
        temp4 = zero
        do 35 k = 1, nenter
          k1 = indx2(k)
          temp1 = temp1 + wy(k1,ipntr)*wy(k1,jpntr)
          temp2 = temp2 + ws(k1,ipntr)*ws(k1,jpntr)
35      ENDDO
        do 36 k = ileave, n
          k1 = indx2(k)
          temp3 = temp3 + wy(k1,ipntr)*wy(k1,jpntr)
          temp4 = temp4 + ws(k1,ipntr)*ws(k1,jpntr)
36      ENDDO
        wn1(iy,jy) = wn1(iy,jy) + temp1 - temp3
        wn1(is,js) = wn1(is,js) - temp2 + temp4
        jpntr = mod(jpntr,m) + 1
40    ENDDO
      ipntr = mod(ipntr,m) + 1
45  ENDDO

    ! modify the old parts in block (2,1).
    ipntr = head
    do 60 is = m + 1, m + upcl
      jpntr = head
      do 55 jy = 1, upcl
        temp1 = zero
        temp3 = zero
        do 50 k = 1, nenter
          k1 = indx2(k)
          temp1 = temp1 + ws(k1,ipntr)*wy(k1,jpntr)
50      ENDDO
        do 51 k = ileave, n
          k1 = indx2(k)
          temp3 = temp3 + ws(k1,ipntr)*wy(k1,jpntr)
51      ENDDO
        if (is <= jy + m) then
          wn1(is,jy) = wn1(is,jy) + temp1 - temp3
        else
          wn1(is,jy) = wn1(is,jy) - temp1 + temp3
        endif
        jpntr = mod(jpntr,m) + 1
55    ENDDO
      ipntr = mod(ipntr,m) + 1
60  ENDDO

    ! Form the upper triangle of WN = [D+Y' ZZ'Y/theta   -L_a'+R_z' ]
    ! [-L_a +R_z        S'AA'S*theta]

    m2 = 2*m
    do 70 iy = 1, col
      is = col + iy
      is1 = m + iy
      do 65 jy = 1, iy
        js = col + jy
        js1 = m + jy
        wn(jy,iy) = wn1(iy,jy)/theta
        wn(js,is) = wn1(is1,js1)*theta
65    ENDDO
      do 66 jy = 1, iy - 1
        wn(jy,is) = -wn1(is1,jy)
66    ENDDO
      do 67 jy = iy, col
        wn(jy,is) = wn1(is1,jy)
67    ENDDO
      wn(iy,iy) = wn(iy,iy) + sy(iy,iy)
70  ENDDO

    ! Form the upper triangle of WN= [  LL'            L^-1(-L_a'+R_z')]
    ! [(-L_a +R_z)L'^-1   S'AA'S*theta  ]

    ! first Cholesky factor (1,1) block of wn to get LL'
    ! with L' stored in the upper triangle of wn.
    call dpofa(wn,m2,col,info)
    if (info /= 0) then
      info = -1
      return
    endif
    ! then form L^-1(-L_a'+R_z') in the (1,2) block.
    col2 = 2*col
    do 71 js = col+1 ,col2
      call dtrsl(wn,m2,col,wn(1,js),11,info)
71  ENDDO

    ! Form S'AA'S*theta + (L^-1(-L_a'+R_z'))'L^-1(-L_a'+R_z') in the
    ! upper triangle of (2,2) block of wn.


    do 72 is = col+1, col2
      do 74 js = is, col2
        wn(is,js) = wn(is,js) + ddot(col,wn(1,is),1,wn(1,js),1)
74    ENDDO
72  ENDDO

    ! Cholesky factorization of (2,2) block of wn.

    call dpofa(wn(col+1,col+1),m2,col,info)
    if (info /= 0) then
      info = -2
      return
    endif

    return

  end subroutine formk

  !======================= The end of formk ==============================

  subroutine formt(m, wt, sy, ss, col, theta, info)

    integer ::          m, col, info
    double precision :: theta, wt(m, m), sy(m, m), ss(m, m)

    ! ************

    ! Subroutine formt

    ! This subroutine forms the upper half of the pos. def. and symm.
    ! T = theta*SS + L*D^(-1)*L', stores T in the upper triangle
    ! of the array wt, and performs the Cholesky factorization of T
    ! to produce J*J', with J' stored in the upper triangle of wt.

    ! Subprograms called:

    ! Linpack ... dpofa.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer ::          i,j,k,k1
    double precision :: ddum
    double precision :: zero
    parameter        (zero=0.0d0)


    ! Form the upper half of  T = theta*SS + L*D^(-1)*L',
    ! store T in the upper triangle of the array wt.

    do 52 j = 1, col
      wt(1,j) = theta*ss(1,j)
52  ENDDO
    do 55 i = 2, col
      do 54 j = i, col
        k1 = min(i,j) - 1
        ddum  = zero
        do 53 k = 1, k1
          ddum  = ddum + sy(i,k)*sy(j,k)/sy(k,k)
53      ENDDO
        wt(i,j) = ddum + theta*ss(i,j)
54    ENDDO
55  ENDDO

    ! Cholesky factorize T to J*J' with
    ! J' stored in the upper triangle of wt.

    call dpofa(wt,m,col,info)
    if (info /= 0) then
      info = -3
    endif

    return

  end subroutine formt

  !======================= The end of formt ==============================

  subroutine freev(n, nfree, index, nenter, ileave, indx2, &
     iwhere, wrk, updatd, cnstnd, iprint, iter)

    integer :: n, nfree, nenter, ileave, iprint, iter, &
       index(n), indx2(n), iwhere(n)
    logical :: wrk, updatd, cnstnd

    ! ************

    ! Subroutine freev

    ! This subroutine counts the entering and leaving variables when
    ! iter > 0, and finds the index set of free and active variables
    ! at the GCP.

    ! cnstnd is a logical variable indicating whether bounds are present

    ! index is an integer array of dimension n
    ! for i=1,...,nfree, index(i) are the indices of free variables
    ! for i=nfree+1,...,n, index(i) are the indices of bound variables
    ! On entry after the first iteration, index gives
    ! the free variables at the previous iteration.
    ! On exit it gives the free variables based on the determination
    ! in cauchy using the array iwhere.

    ! indx2 is an integer array of dimension n
    ! On entry indx2 is unspecified.
    ! On exit with iter>0, indx2 indicates which variables
    ! have changed status since the previous iteration.
    ! For i= 1,...,nenter, indx2(i) have changed from bound to free.
    ! For i= ileave+1,...,n, indx2(i) have changed from free to bound.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer :: iact,i,k

    nenter = 0
    ileave = n + 1
    if (iter > 0 .AND. cnstnd) then
      ! count the entering and leaving variables.
      do 20 i = 1, nfree
        k = index(i)
        if (iwhere(k) > 0) then
          ileave = ileave - 1
          indx2(ileave) = k
        endif
20    ENDDO
      do 22 i = 1 + nfree, n
        k = index(i)
        if (iwhere(k) <= 0) then
          nenter = nenter + 1
          indx2(nenter) = k
        endif
22    ENDDO
    endif
    wrk = (ileave .lt. n+1) .or. (nenter .gt. 0) .or. updatd

    ! Find the index set of free and active variables at the GCP.

    nfree = 0
    iact = n + 1
    do 24 i = 1, n
      if (iwhere(i) <= 0) then
        nfree = nfree + 1
        index(nfree) = i
      else
        iact = iact - 1
        index(iact) = i
      endif
24  ENDDO

    return

  end subroutine freev

  !======================= The end of freev ==============================

  subroutine hpsolb(n, t, iorder, iheap)
    integer ::          iheap, n, iorder(n)
    double precision :: t(n)

    ! ************

    ! Subroutine hpsolb

    ! This subroutine sorts out the least element of t, and puts the
    ! remaining elements of t in a heap.

    ! n is an integer variable.
    ! On entry n is the dimension of the arrays t and iorder.
    ! On exit n is unchanged.

    ! t is a double precision array of dimension n.
    ! On entry t stores the elements to be sorted,
    ! On exit t(n) stores the least elements of t, and t(1) to t(n-1)
    ! stores the remaining elements in the form of a heap.

    ! iorder is an integer array of dimension n.
    ! On entry iorder(i) is the index of t(i).
    ! On exit iorder(i) is still the index of t(i), but iorder may be
    ! permuted in accordance with t.

    ! iheap is an integer variable specifying the task.
    ! On entry iheap should be set as follows:
    ! iheap .eq. 0 if t(1) to t(n) is not in the form of a heap,
    ! iheap .ne. 0 if otherwise.
    ! On exit iheap is unchanged.


    ! References:
    ! Algorithm 232 of CACM (J. W. J. Williams): HEAPSORT.

    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.

    ! ************

    integer ::          i,j,k,indxin,indxou
    double precision :: ddum,out

    if (iheap == 0) then

      ! Rearrange the elements t(1) to t(n) to form a heap.

      do 20 k = 2, n
        ddum  = t(k)
        indxin = iorder(k)

        ! Add ddum to the heap.
        i = k
10      continue
        if (i > 1) then
          j = i/2
          if (ddum < t(j)) then
            t(i) = t(j)
            iorder(i) = iorder(j)
            i = j
            goto 10
          endif
        endif
        t(i) = ddum
        iorder(i) = indxin
20    ENDDO
    endif

    ! Assign to 'out' the value of t(1), the least member of the heap,
    ! and rearrange the remaining members to form a heap as
    ! elements 1 to n-1 of t.

    if (n > 1) then
      i = 1
      out = t(1)
      indxou = iorder(1)
      ddum  = t(n)
      indxin  = iorder(n)

      ! Restore the heap
30    continue
      j = i+i
      if (j <= n-1) then
        if (t(j+1) < t(j)) j = j+1
        if (t(j) < ddum ) then
          t(i) = t(j)
          iorder(i) = iorder(j)
          i = j
          goto 30
        endif
      endif
      t(i) = ddum
      iorder(i) = indxin

      ! Put the least member in t(n).

      t(n) = out
      iorder(n) = indxou
    endif

    return

  end subroutine hpsolb

  !====================== The end of hpsolb ==============================

  subroutine lnsrlb(n, l, u, nbd, x, f, fold, gd, gdold, g, d, r, t, &
     z, stp, dnorm, dtd, xstep, stpmx, iter, ifun, &
     iback, nfgv, info, task, boxed, cnstnd, csave, &
     isave, dsave)

    character(60) ::     task, csave
    logical ::          boxed, cnstnd
    integer ::          n, iter, ifun, iback, nfgv, info, &
       nbd(n), isave(2)
    double precision :: f, fold, gd, gdold, stp, dnorm, dtd, xstep, &
       stpmx, x(n), l(n), u(n), g(n), d(n), r(n), t(n), &
       z(n), dsave(13)
    ! **********

    ! Subroutine lnsrlb

    ! This subroutine calls subroutine dcsrch from the Minpack2 library
    ! to perform the line search.  Subroutine dscrch is safeguarded so
    ! that all trial points lie within the feasible region.

    ! Subprograms called:

    ! Minpack2 Library ... dcsrch.

    ! Linpack ... dtrsl, ddot.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! **********

    integer ::          i
    double           precision ddot,a1,a2
    double precision :: one,zero,big
    parameter        (one=1.0d0,zero=0.0d0,big=1.0d+10)
    double precision :: ftol,gtol,xtol
    parameter        (ftol=1.0d-3,gtol=0.9d0,xtol=0.1d0)

    if (task(1:5) == 'FG_LN') goto 556

    dtd = ddot(n,d,1,d,1)
    dnorm = sqrt(dtd)

    ! Determine the maximum step length.

    stpmx = big
    if (cnstnd) then
      if (iter == 0) then
        stpmx = one
      else
        do 43 i = 1, n
          a1 = d(i)
          if (nbd(i) /= 0) then
            if (a1 < zero .AND. nbd(i) <= 2) then
              a2 = l(i) - x(i)
              if (a2 >= zero) then
                stpmx = zero
              else if (a1*stpmx < a2) then
                stpmx = a2/a1
              endif
            else if (a1 > zero .AND. nbd(i) >= 2) then
              a2 = u(i) - x(i)
              if (a2 <= zero) then
                stpmx = zero
              else if (a1*stpmx > a2) then
                stpmx = a2/a1
              endif
            endif
          endif
43      ENDDO
      endif
    endif

    if (iter == 0 .AND. .NOT. boxed) then
      stp = min(one/dnorm, stpmx)
    else
      stp = one
    endif

    call dcopy(n,x,1,t,1)
    call dcopy(n,g,1,r,1)
    fold = f
    ifun = 0
    iback = 0
    csave = 'START'
556 continue
    gd = ddot(n,g,1,d,1)
    if (ifun == 0) then
      gdold=gd
      if (gd >= zero) then
        ! the directional derivative >=0.
        ! Line search is impossible.
        info = -4
        return
      endif
    endif

    call dcsrch(f,gd,stp,ftol,gtol,xtol,zero,stpmx,csave,isave,dsave)

    xstep = stp*dnorm
    if (csave(1:4) /= 'CONV' .AND. csave(1:4) /= 'WARN') then
      task = 'FG_LNSRCH'
      ifun = ifun + 1
      nfgv = nfgv + 1
      iback = ifun - 1
      if (stp == one) then
        call dcopy(n,z,1,x,1)
      else
        do 41 i = 1, n
          x(i) = stp*d(i) + t(i)
41      ENDDO
      endif
    else
      task = 'NEW_X'
    endif

    return

  end subroutine lnsrlb

  !======================= The end of lnsrlb =============================

  subroutine matupd(n, m, ws, wy, sy, ss, d, r, itail, &
     iupdat, col, head, theta, rr, dr, stp, dtd)

    integer ::          n, m, itail, iupdat, col, head
    double precision :: theta, rr, dr, stp, dtd, d(n), r(n), &
       ws(n, m), wy(n, m), sy(m, m), ss(m, m)

    ! ************

    ! Subroutine matupd

    ! This subroutine updates matrices WS and WY, and forms the
    ! middle matrix in B.

    ! Subprograms called:

    ! Linpack ... dcopy, ddot.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer ::          j,pointr
    double precision :: ddot
    double precision :: one
    parameter        (one=1.0d0)

    ! Set pointers for matrices WS and WY.

    if (iupdat <= m) then
      col = iupdat
      itail = mod(head+iupdat-2,m) + 1
    else
      itail = mod(itail,m) + 1
      head = mod(head,m) + 1
    endif

    ! Update matrices WS and WY.

    call dcopy(n,d,1,ws(1,itail),1)
    call dcopy(n,r,1,wy(1,itail),1)

    ! Set theta=yy/ys.

    theta = rr/dr

    ! Form the middle matrix in B.

    ! update the upper triangle of SS,
    ! and the lower triangle of SY:
    if (iupdat > m) then
      ! move old information
      do 50 j = 1, col - 1
        call dcopy(j,ss(2,j+1),1,ss(1,j),1)
        call dcopy(col-j,sy(j+1,j+1),1,sy(j,j),1)
50    ENDDO
    endif
    ! add new information: the last row of SY
    ! and the last column of SS:
    pointr = head
    do 51 j = 1, col - 1
      sy(col,j) = ddot(n,d,1,wy(1,pointr),1)
      ss(j,col) = ddot(n,ws(1,pointr),1,d,1)
      pointr = mod(pointr,m) + 1
51  ENDDO
    if (stp == one) then
      ss(col,col) = dtd
    else
      ss(col,col) = stp*stp*dtd
    endif
    sy(col,col) = dr

    return

  end subroutine matupd

  !======================= The end of matupd =============================


  subroutine projgr(n, l, u, nbd, x, g, sbgnrm)

    integer ::          n, nbd(n)
    double precision :: sbgnrm, x(n), l(n), u(n), g(n)

    ! ************

    ! Subroutine projgr

    ! This subroutine computes the infinity norm of the projected
    ! gradient.


    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer :: i
    double precision :: gi
    double precision :: one,zero
    parameter        (one=1.0d0,zero=0.0d0)

    sbgnrm = zero
    do 15 i = 1, n
      gi = g(i)
      if (nbd(i) /= 0) then
        if (gi < zero) then
          if (nbd(i) >= 2) gi = max((x(i)-u(i)),gi)
        else
          if (nbd(i) <= 2) gi = min((x(i)-l(i)),gi)
        endif
      endif
      sbgnrm = max(sbgnrm,abs(gi))
15  ENDDO

    return

  end subroutine projgr

  !======================= The end of projgr =============================

  subroutine subsm(n, m, nsub, ind, l, u, nbd, x, d, ws, wy, theta, &
     col, head, iword, wv, wn, iprint, info)

    integer ::          n, m, nsub, col, head, iword, iprint, info, &
       ind(nsub), nbd(n)
    double precision :: theta, &
       l(n), u(n), x(n), d(n), &
       ws(n, m), wy(n, m), &
       wv(2*m), wn(2*m, 2*m)

    ! ************

    ! Subroutine subsm

    ! Given xcp, l, u, r, an index set that specifies
    !	the active set at xcp, and an l-BFGS matrix B
    !	(in terms of WY, WS, SY, WT, head, col, and theta),
    !	this subroutine computes an approximate solution
    !	of the subspace problem

    ! (P)   min Q(x) = r'(x-xcp) + 1/2 (x-xcp)' B (x-xcp)

    ! subject to l<=x<=u
    ! x_i=xcp_i for all i in A(xcp)

    !	along the subspace unconstrained Newton direction

    ! d = -(Z'BZ)^(-1) r.

    ! The formula for the Newton direction, given the L-BFGS matrix
    ! and the Sherman-Morrison formula, is

    ! d = (1/theta)r + (1/theta*2) Z'WK^(-1)W'Z r.

    ! where
    ! K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
    ! [L_a -R_z           theta*S'AA'S ]

    ! Note that this procedure for computing d differs
    ! from that described in [1]. One can show that the matrix K is
    ! equal to the matrix M^[-1]N in that paper.

    ! n is an integer variable.
    ! On entry n is the dimension of the problem.
    ! On exit n is unchanged.

    ! m is an integer variable.
    ! On entry m is the maximum number of variable metric corrections
    ! used to define the limited memory matrix.
    ! On exit m is unchanged.

    ! nsub is an integer variable.
    ! On entry nsub is the number of free variables.
    ! On exit nsub is unchanged.

    ! ind is an integer array of dimension nsub.
    ! On entry ind specifies the coordinate indices of free variables.
    ! On exit ind is unchanged.

    ! l is a double precision array of dimension n.
    ! On entry l is the lower bound of x.
    ! On exit l is unchanged.

    ! u is a double precision array of dimension n.
    ! On entry u is the upper bound of x.
    ! On exit u is unchanged.

    ! nbd is a integer array of dimension n.
    ! On entry nbd represents the type of bounds imposed on the
    ! variables, and must be specified as follows:
    ! nbd(i)=0 if x(i) is unbounded,
    ! 1 if x(i) has only a lower bound,
    ! 2 if x(i) has both lower and upper bounds, and
    ! 3 if x(i) has only an upper bound.
    ! On exit nbd is unchanged.

    ! x is a double precision array of dimension n.
    ! On entry x specifies the Cauchy point xcp.
    ! On exit x(i) is the minimizer of Q over the subspace of
    ! free variables.

    ! d is a double precision array of dimension n.
    ! On entry d is the reduced gradient of Q at xcp.
    ! On exit d is the Newton direction of Q.

    ! ws and wy are double precision arrays;
    ! theta is a double precision variable;
    ! col is an integer variable;
    ! head is an integer variable.
    ! On entry they store the information defining the
    ! limited memory BFGS matrix:
    ! ws(n,m) stores S, a set of s-vectors;
    ! wy(n,m) stores Y, a set of y-vectors;
    ! theta is the scaling factor specifying B_0 = theta I;
    ! col is the number of variable metric corrections stored;
    ! head is the location of the 1st s- (or y-) vector in S (or Y).
    ! On exit they are unchanged.

    ! iword is an integer variable.
    ! On entry iword is unspecified.
    ! On exit iword specifies the status of the subspace solution.
    ! iword = 0 if the solution is in the box,
    ! 1 if some bound is encountered.

    ! wv is a double precision working array of dimension 2m.

    ! wn is a double precision array of dimension 2m x 2m.
    ! On entry the upper triangle of wn stores the LEL^T factorization
    ! of the indefinite matrix

    ! K = [-D -Y'ZZ'Y/theta     L_a'-R_z'  ]
    ! [L_a -R_z           theta*S'AA'S ]
    ! where E = [-I  0]
    ! [ 0  I]
    ! On exit wn is unchanged.

    ! iprint is an INTEGER variable that must be set by the user.
    ! It controls the frequency and type of output generated:
    ! iprint<0    no output is generated;
    ! iprint=0    print only one line at the last iteration;
    ! 0<iprint<99 print also f and |proj g| every iprint iterations;
    ! iprint=99   print details of every iteration except n-vectors;
    ! iprint=100  print also the changes of active set and final x;
    ! iprint>100  print details of every iteration including x and g;
    ! When iprint > 0, the file iterate.dat will be created to
    ! summarize the iteration.

    ! info is an integer variable.
    ! On entry info is unspecified.
    ! On exit info = 0       for normal return,
    ! = nonzero for abnormal return
    ! when the matrix K is ill-conditioned.

    ! Subprograms called:

    ! Linpack dtrsl.


    ! References:

    ! [1] R. H. Byrd, P. Lu, J. Nocedal and C. Zhu, ``A limited
    ! memory algorithm for bound constrained optimization'',
    ! SIAM J. Scientific Computing 16 (1995), no. 5, pp. 1190--1208.



    ! *  *  *

    ! NEOS, November 1994. (Latest revision June 1996.)
    ! Optimization Technology Center.
    ! Argonne National Laboratory and Northwestern University.
    ! Written by
    ! Ciyou Zhu
    ! in collaboration with R.H. Byrd, P. Lu-Chen and J. Nocedal.


    ! ************

    integer ::          pointr,m2,col2,ibd,jy,js,i,j,k
    double precision :: alpha,dk,temp1,temp2
    double precision :: one,zero
    parameter        (one=1.0d0,zero=0.0d0)

    if (nsub <= 0) return

    ! Compute wv = W'Zd.

    pointr = head
    do 20 i = 1, col
      temp1 = zero
      temp2 = zero
      do 10 j = 1, nsub
        k = ind(j)
        temp1 = temp1 + wy(k,pointr)*d(j)
        temp2 = temp2 + ws(k,pointr)*d(j)
10    ENDDO
      wv(i) = temp1
      wv(col + i) = theta*temp2
      pointr = mod(pointr,m) + 1
20  ENDDO

    ! Compute wv:=K^(-1)wv.

    m2 = 2*m
    col2 = 2*col
    call dtrsl(wn,m2,col2,wv,11,info)
    if (info /= 0) return
    do 25 i = 1, col
      wv(i) = -wv(i)
25  ENDDO
    call dtrsl(wn,m2,col2,wv,01,info)
    if (info /= 0) return

    ! Compute d = (1/theta)d + (1/theta**2)Z'W wv.

    pointr = head
    do 40 jy = 1, col
      js = col + jy
      do 30 i = 1, nsub
        k = ind(i)
        d(i) = d(i) + wy(k,pointr)*wv(jy)/theta &
           + ws(k,pointr)*wv(js)
30    ENDDO
      pointr = mod(pointr,m) + 1
40  ENDDO
    do 50 i = 1, nsub
      d(i) = d(i)/theta
50  ENDDO

    ! Backtrack to the feasible region.

    alpha = one
    temp1 = alpha
    do 60 i = 1, nsub
      k = ind(i)
      dk = d(i)
      if (nbd(k) /= 0) then
        if (dk < zero .AND. nbd(k) <= 2) then
          temp2 = l(k) - x(k)
          if (temp2 >= zero) then
            temp1 = zero
          else if (dk*alpha < temp2) then
            temp1 = temp2/dk
          endif
        else if (dk > zero .AND. nbd(k) >= 2) then
          temp2 = u(k) - x(k)
          if (temp2 <= zero) then
            temp1 = zero
          else if (dk*alpha > temp2) then
            temp1 = temp2/dk
          endif
        endif
        if (temp1 < alpha) then
          alpha = temp1
          ibd = i
        endif
      endif
60  ENDDO

    if (alpha < one) then
      dk = d(ibd)
      k = ind(ibd)
      if (dk > zero) then
        x(k) = u(k)
        d(ibd) = zero
      else if (dk < zero) then
        x(k) = l(k)
        d(ibd) = zero
      endif
    endif
    do 70 i = 1, nsub
      k = ind(i)
      x(k) = x(k) + alpha*d(i)
70  ENDDO

    if (alpha < one) then
      iword = 1
    else
      iword = 0
    endif

    return

  end subroutine subsm

  !====================== The end of subsm ===============================

  subroutine dcsrch(f,g,stp,ftol,gtol,xtol,stpmin,stpmax, &
     task,isave,dsave)
    character(60) :: task
    integer :: isave(2)
    double precision :: f,g,stp,ftol,gtol,xtol,stpmin,stpmax
    double precision :: dsave(13)
    ! **********

    ! Subroutine dcsrch

    ! This subroutine finds a step that satisfies a sufficient
    ! decrease condition and a curvature condition.

    ! Each call of the subroutine updates an interval with
    ! endpoints stx and sty. The interval is initially chosen
    ! so that it contains a minimizer of the modified function

    ! psi(stp) = f(stp) - f(0) - ftol*stp*f'(0).

    ! If psi(stp) <= 0 and f'(stp) >= 0 for some step, then the
    ! interval is chosen so that it contains a minimizer of f.

    ! The algorithm is designed to find a step that satisfies
    ! the sufficient decrease condition

    ! f(stp) <= f(0) + ftol*stp*f'(0),

    ! and the curvature condition

    ! abs(f'(stp)) <= gtol*abs(f'(0)).

    ! If ftol is less than gtol and if, for example, the function
    ! is bounded below, then there is always a step which satisfies
    ! both conditions.

    ! If no step can be found that satisfies both conditions, then
    ! the algorithm stops with a warning. In this case stp only
    ! satisfies the sufficient decrease condition.

    ! A typical invocation of dcsrch has the following outline:

    ! task = 'START'
    ! 10 continue
    ! call dcsrch( ... )
    ! if (task .eq. 'FG') then
    ! Evaluate the function and the gradient at stp
    ! goto 10
    ! end if

    ! NOTE: The user must no alter work arrays between calls.

    ! The subroutine statement is

    ! subroutine dcsrch(f,g,stp,ftol,gtol,xtol,stpmin,stpmax,
    ! task,isave,dsave)
    ! where

    ! f is a double precision variable.
    ! On initial entry f is the value of the function at 0.
    ! On subsequent entries f is the value of the
    ! function at stp.
    ! On exit f is the value of the function at stp.

    !	g is a double precision variable.
    ! On initial entry g is the derivative of the function at 0.
    ! On subsequent entries g is the derivative of the
    ! function at stp.
    ! On exit g is the derivative of the function at stp.

    !	stp is a double precision variable.
    ! On entry stp is the current estimate of a satisfactory
    ! step. On initial entry, a positive initial estimate
    ! must be provided.
    ! On exit stp is the current estimate of a satisfactory step
    ! if task = 'FG'. If task = 'CONV' then stp satisfies
    ! the sufficient decrease and curvature condition.

    ! ftol is a double precision variable.
    ! On entry ftol specifies a nonnegative tolerance for the
    ! sufficient decrease condition.
    ! On exit ftol is unchanged.

    ! gtol is a double precision variable.
    ! On entry gtol specifies a nonnegative tolerance for the
    ! curvature condition.
    ! On exit gtol is unchanged.

    !	xtol is a double precision variable.
    ! On entry xtol specifies a nonnegative relative tolerance
    ! for an acceptable step. The subroutine exits with a
    ! warning if the relative difference between sty and stx
    ! is less than xtol.
    ! On exit xtol is unchanged.

    !	stpmin is a double precision variable.
    ! On entry stpmin is a nonnegative lower bound for the step.
    ! On exit stpmin is unchanged.

    !	stpmax is a double precision variable.
    ! On entry stpmax is a nonnegative upper bound for the step.
    ! On exit stpmax is unchanged.

    ! task is a character variable of length at least 60.
    ! On initial entry task must be set to 'START'.
    ! On exit task indicates the required action:

    ! If task(1:2) = 'FG' then evaluate the function and
    ! derivative at stp and call dcsrch again.

    ! If task(1:4) = 'CONV' then the search is successful.

    ! If task(1:4) = 'WARN' then the subroutine is not able
    ! to satisfy the convergence conditions. The exit value of
    ! stp contains the best point found during the search.

    ! If task(1:5) = 'ERROR' then there is an error in the
    ! input arguments.

    ! On exit with convergence, a warning or an error, the
    ! variable task contains additional information.

    ! isave is an integer work array of dimension 2.

    ! dsave is a double precision work array of dimension 13.

    ! Subprograms called

    !	MINPACK-2 ... dcstep

    ! MINPACK-1 Project. June 1983.
    ! Argonne National Laboratory.
    ! Jorge J. More' and David J. Thuente.

    ! MINPACK-2 Project. October 1993.
    ! Argonne National Laboratory and University of Minnesota.
    ! Brett M. Averick, Richard G. Carter, and Jorge J. More'.

    ! **********
    double precision :: zero,p5,p66
    parameter(zero=0.0d0,p5=0.5d0,p66=0.66d0)
    double precision :: xtrapl,xtrapu
    parameter(xtrapl=1.1d0,xtrapu=4.0d0)

    logical :: brackt
    integer :: stage
    double precision :: finit,ftest,fm,fx,fxm,fy,fym,ginit,gtest, &
       gm,gx,gxm,gy,gym,stx,sty,stmin,stmax,width,width1

    ! Initialization block.

    if (task(1:5) == 'START') then

      ! Check the input arguments for errors.

      if (stp < stpmin) task = 'ERROR: STP < STPMIN'
      if (stp > stpmax) task = 'ERROR: STP > STPMAX'
      if (g >= zero) task = 'ERROR: INITIAL G >= ZERO'
      if (ftol < zero) task = 'ERROR: FTOL < ZERO'
      if (gtol < zero) task = 'ERROR: GTOL < ZERO'
      if (xtol < zero) task = 'ERROR: XTOL < ZERO'
      if (stpmin < zero) task = 'ERROR: STPMIN < ZERO'
      if (stpmax < stpmin) task = 'ERROR: STPMAX < STPMIN'

      ! Exit if there are errors on input.

      if (task(1:5) == 'ERROR') return

      ! Initialize local variables.

      brackt = .false.
      stage = 1
      finit = f
      ginit = g
      gtest = ftol*ginit
      width = stpmax - stpmin
      width1 = width/p5

      ! The variables stx, fx, gx contain the values of the step,
      ! function, and derivative at the best step.
      ! The variables sty, fy, gy contain the value of the step,
      ! function, and derivative at sty.
      ! The variables stp, f, g contain the values of the step,
      ! function, and derivative at stp.

      stx = zero
      fx = finit
      gx = ginit
      sty = zero
      fy = finit
      gy = ginit
      stmin = zero
      stmax = stp + xtrapu*stp
      task = 'FG'

      goto 1000

    else

      ! Restore local variables.

      if (isave(1) == 1) then
        brackt = .true.
      else
        brackt = .false.
      endif
      stage = isave(2)
      ginit = dsave(1)
      gtest = dsave(2)
      gx = dsave(3)
      gy = dsave(4)
      finit = dsave(5)
      fx = dsave(6)
      fy = dsave(7)
      stx = dsave(8)
      sty = dsave(9)
      stmin = dsave(10)
      stmax = dsave(11)
      width = dsave(12)
      width1 = dsave(13)

    endif

    ! If psi(stp) <= 0 and f'(stp) >= 0 for some step, then the
    ! algorithm enters the second stage.

    ftest = finit + stp*gtest
    if (stage == 1 .AND. f <= ftest .AND. g >= zero) &
       stage = 2

    ! Test for warnings.

    if (brackt .AND. (stp <= stmin .OR. stp >= stmax)) &
       task = 'WARNING: ROUNDING ERRORS PREVENT PROGRESS'
    if (brackt .AND. stmax - stmin <= xtol*stmax) &
       task = 'WARNING: XTOL TEST SATISFIED'
    if (stp == stpmax .AND. f <= ftest .AND. g <= gtest) &
       task = 'WARNING: STP = STPMAX'
    if (stp == stpmin .AND. (f > ftest .OR. g >= gtest)) &
       task = 'WARNING: STP = STPMIN'

    ! Test for convergence.

    if (f <= ftest .AND. abs(g) <= gtol*(-ginit)) &
       task = 'CONVERGENCE'

    ! Test for termination.

    if (task(1:4) == 'WARN' .OR. task(1:4) == 'CONV') goto 1000

    ! A modified function is used to predict the step during the
    ! first stage if a lower function value has been obtained but
    ! the decrease is not sufficient.

    if (stage == 1 .AND. f <= fx .AND. f > ftest) then

      ! Define the modified function and derivative values.

      fm = f - stp*gtest
      fxm = fx - stx*gtest
      fym = fy - sty*gtest
      gm = g - gtest
      gxm = gx - gtest
      gym = gy - gtest

      ! Call dcstep to update stx, sty, and to compute the new step.

      call dcstep(stx,fxm,gxm,sty,fym,gym,stp,fm,gm, &
         brackt,stmin,stmax)

      ! Reset the function and derivative values for f.

      fx = fxm + stx*gtest
      fy = fym + sty*gtest
      gx = gxm + gtest
      gy = gym + gtest

    else

      ! Call dcstep to update stx, sty, and to compute the new step.

      call dcstep(stx,fx,gx,sty,fy,gy,stp,f,g, &
         brackt,stmin,stmax)

    endif

    ! Decide if a bisection step is needed.

    if (brackt) then
      if (abs(sty-stx) >= p66*width1) stp = stx + p5*(sty - stx)
      width1 = width
      width = abs(sty-stx)
    endif

    ! Set the minimum and maximum steps allowed for stp.

    if (brackt) then
      stmin = min(stx,sty)
      stmax = max(stx,sty)
    else
      stmin = stp + xtrapl*(stp - stx)
      stmax = stp + xtrapu*(stp - stx)
    endif

    ! Force the step to be within the bounds stpmax and stpmin.

    stp = max(stp,stpmin)
    stp = min(stp,stpmax)

    ! If further progress is not possible, let stp be the best
    ! point obtained during the search.

    if (brackt .AND. (stp <= stmin .OR. stp >= stmax) &
       .OR. (brackt .AND. stmax-stmin <= xtol*stmax)) stp = stx

    ! Obtain another function and derivative.

    task = 'FG'

1000 continue

    ! Save local variables.

    if (brackt) then
      isave(1) = 1
    else
      isave(1) = 0
    endif
    isave(2) = stage
    dsave(1) =  ginit
    dsave(2) =  gtest
    dsave(3) =  gx
    dsave(4) =  gy
    dsave(5) =  finit
    dsave(6) =  fx
    dsave(7) =  fy
    dsave(8) =  stx
    dsave(9) =  sty
    dsave(10) = stmin
    dsave(11) = stmax
    dsave(12) = width
    dsave(13) = width1

  end subroutine dcsrch

  !====================== The end of dcsrch ==============================

  subroutine dcstep(stx,fx,dx,sty,fy,dy,stp,fp,dp,brackt, &
     stpmin,stpmax)
    logical :: brackt
    double precision :: stx,fx,dx,sty,fy,dy,stp,fp,dp,stpmin,stpmax
    ! **********

    ! Subroutine dcstep

    ! This subroutine computes a safeguarded step for a search
    ! procedure and updates an interval that contains a step that
    ! satisfies a sufficient decrease and a curvature condition.

    ! The parameter stx contains the step with the least function
    ! value. If brackt is set to .true. then a minimizer has
    ! been bracketed in an interval with endpoints stx and sty.
    ! The parameter stp contains the current step.
    ! The subroutine assumes that if brackt is set to .true. then

    ! min(stx,sty) < stp < max(stx,sty),

    ! and that the derivative at stx is negative in the direction
    ! of the step.

    ! The subroutine statement is

    ! subroutine dcstep(stx,fx,dx,sty,fy,dy,stp,fp,dp,brackt,
    ! stpmin,stpmax)

    ! where

    ! stx is a double precision variable.
    ! On entry stx is the best step obtained so far and is an
    ! endpoint of the interval that contains the minimizer.
    ! On exit stx is the updated best step.

    ! fx is a double precision variable.
    ! On entry fx is the function at stx.
    ! On exit fx is the function at stx.

    ! dx is a double precision variable.
    ! On entry dx is the derivative of the function at
    ! stx. The derivative must be negative in the direction of
    ! the step, that is, dx and stp - stx must have opposite
    ! signs.
    ! On exit dx is the derivative of the function at stx.

    ! sty is a double precision variable.
    ! On entry sty is the second endpoint of the interval that
    ! contains the minimizer.
    ! On exit sty is the updated endpoint of the interval that
    ! contains the minimizer.

    ! fy is a double precision variable.
    ! On entry fy is the function at sty.
    ! On exit fy is the function at sty.

    ! dy is a double precision variable.
    ! On entry dy is the derivative of the function at sty.
    ! On exit dy is the derivative of the function at the exit sty.

    ! stp is a double precision variable.
    ! On entry stp is the current step. If brackt is set to .true.
    ! then on input stp must be between stx and sty.
    ! On exit stp is a new trial step.

    ! fp is a double precision variable.
    ! On entry fp is the function at stp
    ! On exit fp is unchanged.

    ! dp is a double precision variable.
    ! On entry dp is the the derivative of the function at stp.
    ! On exit dp is unchanged.

    ! brackt is an logical variable.
    ! On entry brackt specifies if a minimizer has been bracketed.
    ! Initially brackt must be set to .false.
    ! On exit brackt specifies if a minimizer has been bracketed.
    ! When a minimizer is bracketed brackt is set to .true.

    ! stpmin is a double precision variable.
    ! On entry stpmin is a lower bound for the step.
    ! On exit stpmin is unchanged.

    ! stpmax is a double precision variable.
    ! On entry stpmax is an upper bound for the step.
    ! On exit stpmax is unchanged.

    ! MINPACK-1 Project. June 1983
    ! Argonne National Laboratory.
    ! Jorge J. More' and David J. Thuente.

    ! MINPACK-2 Project. October 1993.
    ! Argonne National Laboratory and University of Minnesota.
    ! Brett M. Averick and Jorge J. More'.

    ! **********
    double precision :: zero,p66,two,three
    parameter(zero=0.0d0,p66=0.66d0,two=2.0d0,three=3.0d0)

    double precision :: gamma,p,q,r,s,sgnd,stpc,stpf,stpq,theta

    sgnd = dp*(dx/abs(dx))

    ! First case: A higher function value. The minimum is bracketed.
    ! If the cubic step is closer to stx than the quadratic step, the
    ! cubic step is taken, otherwise the average of the cubic and
    ! quadratic steps is taken.

    if (fp > fx) then
      theta = three*(fx - fp)/(stp - stx) + dx + dp
      s = max(abs(theta),abs(dx),abs(dp))
      gamma = s*sqrt((theta/s)**2 - (dx/s)*(dp/s))
      if (stp < stx) gamma = -gamma
      p = (gamma - dx) + theta
      q = ((gamma - dx) + gamma) + dp
      r = p/q
      stpc = stx + r*(stp - stx)
      stpq = stx + ((dx/((fx - fp)/(stp - stx) + dx))/two)* &
         (stp - stx)
      if (abs(stpc-stx) < abs(stpq-stx)) then
        stpf = stpc
      else
        stpf = stpc + (stpq - stpc)/two
      endif
      brackt = .true.

      ! Second case: A lower function value and derivatives of opposite
      ! sign. The minimum is bracketed. If the cubic step is farther from
      ! stp than the secant step, the cubic step is taken, otherwise the
      ! secant step is taken.

    else if (sgnd < zero) then
      theta = three*(fx - fp)/(stp - stx) + dx + dp
      s = max(abs(theta),abs(dx),abs(dp))
      gamma = s*sqrt((theta/s)**2 - (dx/s)*(dp/s))
      if (stp > stx) gamma = -gamma
      p = (gamma - dp) + theta
      q = ((gamma - dp) + gamma) + dx
      r = p/q
      stpc = stp + r*(stx - stp)
      stpq = stp + (dp/(dp - dx))*(stx - stp)
      if (abs(stpc-stp) > abs(stpq-stp)) then
        stpf = stpc
      else
        stpf = stpq
      endif
      brackt = .true.

      ! Third case: A lower function value, derivatives of the same sign,
      ! and the magnitude of the derivative decreases.

    else if (abs(dp) < abs(dx)) then

      ! The cubic step is computed only if the cubic tends to infinity
      ! in the direction of the step or if the minimum of the cubic
      ! is beyond stp. Otherwise the cubic step is defined to be the
      ! secant step.

      theta = three*(fx - fp)/(stp - stx) + dx + dp
      s = max(abs(theta),abs(dx),abs(dp))

      ! The case gamma = 0 only arises if the cubic does not tend
      ! to infinity in the direction of the step.

      gamma = s*sqrt(max(zero,(theta/s)**2-(dx/s)*(dp/s)))
      if (stp > stx) gamma = -gamma
      p = (gamma - dp) + theta
      q = (gamma + (dx - dp)) + gamma
      r = p/q
      if (r < zero .AND. gamma /= zero) then
        stpc = stp + r*(stx - stp)
      else if (stp > stx) then
        stpc = stpmax
      else
        stpc = stpmin
      endif
      stpq = stp + (dp/(dp - dx))*(stx - stp)

      if (brackt) then

        ! A minimizer has been bracketed. If the cubic step is
        ! closer to stp than the secant step, the cubic step is
        ! taken, otherwise the secant step is taken.

        if (abs(stpc-stp) < abs(stpq-stp)) then
          stpf = stpc
        else
          stpf = stpq
        endif
        if (stp > stx) then
          stpf = min(stp+p66*(sty-stp),stpf)
        else
          stpf = max(stp+p66*(sty-stp),stpf)
        endif
      else

        ! A minimizer has not been bracketed. If the cubic step is
        ! farther from stp than the secant step, the cubic step is
        ! taken, otherwise the secant step is taken.

        if (abs(stpc-stp) > abs(stpq-stp)) then
          stpf = stpc
        else
          stpf = stpq
        endif
        stpf = min(stpmax,stpf)
        stpf = max(stpmin,stpf)
      endif

      ! Fourth case: A lower function value, derivatives of the same sign,
      ! and the magnitude of the derivative does not decrease. If the
      ! minimum is not bracketed, the step is either stpmin or stpmax,
      ! otherwise the cubic step is taken.

    else
      if (brackt) then
        theta = three*(fp - fy)/(sty - stp) + dy + dp
        s = max(abs(theta),abs(dy),abs(dp))
        gamma = s*sqrt((theta/s)**2 - (dy/s)*(dp/s))
        if (stp > sty) gamma = -gamma
        p = (gamma - dp) + theta
        q = ((gamma - dp) + gamma) + dy
        r = p/q
        stpc = stp + r*(sty - stp)
        stpf = stpc
      else if (stp > stx) then
        stpf = stpmax
      else
        stpf = stpmin
      endif
    endif

    ! Update the interval which contains a minimizer.

    if (fp > fx) then
      sty = stp
      fy = fp
      dy = dp
    else
      if (sgnd < zero) then
        sty = stx
        fy = fx
        dy = dx
      endif
      stx = stp
      fx = fp
      dx = dp
    endif

    ! Compute the new step.

    stp = stpf

  end subroutine dcstep

  !====================== The end of dcstep ==============================

  double precision function dnrm2(n,x,incx)
    integer :: n,incx
    double precision :: x(n)
    ! **********

    ! Function dnrm2

    ! Given a vector x of length n, this function calculates the
    ! Euclidean norm of x with stride incx.

    ! The function statement is

    ! double precision function dnrm2(n,x,incx)

    ! where

    ! n is a positive integer input variable.

    ! x is an input array of length n.

    ! incx is a positive integer variable that specifies the
    ! stride of the vector.

    ! Subprograms called

    ! FORTRAN-supplied ... abs, max, sqrt

    ! MINPACK-2 Project. February 1991.
    ! Argonne National Laboratory.
    ! Brett M. Averick.

    ! **********
    integer :: i
    double precision :: scale

    dnrm2 = 0.0d0
    scale = 0.0d0

    do 10 i = 1, n, incx
      scale = max(scale, abs(x(i)))
10  ENDDO

    if (scale == 0.0d0) return

    do 20 i = 1, n, incx
      dnrm2 = dnrm2 + (x(i)/scale)**2
20  ENDDO

    dnrm2 = scale*sqrt(dnrm2)


    return

  end function dnrm2

  !====================== The end of dnrm2 ===============================


  subroutine daxpy(n,da,dx,incx,dy,incy)

    ! constant times a vector plus a vector.
    ! uses unrolled loops for increments equal to one.
    ! jack dongarra, linpack, 3/11/78.

    double precision :: dx(*),dy(*),da
    integer :: i,incx,incy,ix,iy,m,mp1,n

    if(n <= 0)return
    if (da == 0.0d0) return
    if(incx == 1 .AND. incy == 1)go to 20

    ! code for unequal increments or equal increments
    ! not equal to 1

    ix = 1
    iy = 1
    if(incx < 0)ix = (-n+1)*incx + 1
    if(incy < 0)iy = (-n+1)*incy + 1
    do 10 i = 1,n
      dy(iy) = dy(iy) + da*dx(ix)
      ix = ix + incx
      iy = iy + incy
10  ENDDO
    return

    ! code for both increments equal to 1


    ! clean-up loop

20  m = mod(n,4)
    if( m == 0 ) go to 40
    do 30 i = 1,m
      dy(i) = dy(i) + da*dx(i)
30  ENDDO
    if( n < 4 ) return
40  mp1 = m + 1
    do 50 i = mp1,n,4
      dy(i) = dy(i) + da*dx(i)
      dy(i + 1) = dy(i + 1) + da*dx(i + 1)
      dy(i + 2) = dy(i + 2) + da*dx(i + 2)
      dy(i + 3) = dy(i + 3) + da*dx(i + 3)
50  ENDDO
    return
  end subroutine daxpy

  !====================== The end of daxpy ===============================

  subroutine dcopy(n,dx,incx,dy,incy)

    ! copies a vector, x, to a vector, y.
    ! uses unrolled loops for increments equal to one.
    ! jack dongarra, linpack, 3/11/78.

    double precision :: dx(*),dy(*)
    integer :: i,incx,incy,ix,iy,m,mp1,n

    if(n <= 0)return
    if(incx == 1 .AND. incy == 1)go to 20

    ! code for unequal increments or equal increments
    ! not equal to 1

    ix = 1
    iy = 1
    if(incx < 0)ix = (-n+1)*incx + 1
    if(incy < 0)iy = (-n+1)*incy + 1
    do 10 i = 1,n
      dy(iy) = dx(ix)
      ix = ix + incx
      iy = iy + incy
10  ENDDO
    return

    ! code for both increments equal to 1


    ! clean-up loop

20  m = mod(n,7)
    if( m == 0 ) go to 40
    do 30 i = 1,m
      dy(i) = dx(i)
30  ENDDO
    if( n < 7 ) return
40  mp1 = m + 1
    do 50 i = mp1,n,7
      dy(i) = dx(i)
      dy(i + 1) = dx(i + 1)
      dy(i + 2) = dx(i + 2)
      dy(i + 3) = dx(i + 3)
      dy(i + 4) = dx(i + 4)
      dy(i + 5) = dx(i + 5)
      dy(i + 6) = dx(i + 6)
50  ENDDO
    return
  end subroutine dcopy

  !====================== The end of dcopy ===============================

  double precision function ddot(n,dx,incx,dy,incy)

    ! forms the dot product of two vectors.
    ! uses unrolled loops for increments equal to one.
    ! jack dongarra, linpack, 3/11/78.

    double precision :: dx(*),dy(*),dtemp
    integer :: i,incx,incy,ix,iy,m,mp1,n

    ddot = 0.0d0
    dtemp = 0.0d0
    if(n <= 0)return
    if(incx == 1 .AND. incy == 1)go to 20

    ! code for unequal increments or equal increments
    ! not equal to 1

    ix = 1
    iy = 1
    if(incx < 0)ix = (-n+1)*incx + 1
    if(incy < 0)iy = (-n+1)*incy + 1
    do 10 i = 1,n
      dtemp = dtemp + dx(ix)*dy(iy)
      ix = ix + incx
      iy = iy + incy
10  ENDDO
    ddot = dtemp
    return

    ! code for both increments equal to 1


    ! clean-up loop

20  m = mod(n,5)
    if( m == 0 ) go to 40
    do 30 i = 1,m
      dtemp = dtemp + dx(i)*dy(i)
30  ENDDO
    if( n < 5 ) go to 60
40  mp1 = m + 1
    do 50 i = mp1,n,5
      dtemp = dtemp + dx(i)*dy(i) + dx(i + 1)*dy(i + 1) + &
         dx(i + 2)*dy(i + 2) + dx(i + 3)*dy(i + 3) + dx(i + 4)*dy(i + 4)
50  ENDDO
60  ddot = dtemp
    return
  end function ddot

  !====================== The end of ddot ================================

  subroutine dpofa(a,lda,n,info)
    integer :: lda,n,info
    double precision :: a(lda,*)

    ! dpofa factors a double precision symmetric positive definite
    ! matrix.

    ! dpofa is usually called by dpoco, but it can be called
    ! directly with a saving in time if  rcond  is not needed.
    ! (time for dpoco) = (1 + 18/n)*(time for dpofa) .

    ! on entry

    ! a       double precision(lda, n)
    ! the symmetric matrix to be factored.  only the
    ! diagonal and upper triangle are used.

    ! lda     integer
    ! the leading dimension of the array  a .

    ! n       integer
    ! the order of the matrix  a .

    ! on return

    ! a       an upper triangular matrix  r  so that  a = trans(r)*r
    ! where  trans(r)  is the transpose.
    ! the strict lower triangle is unaltered.
    ! if  info .ne. 0 , the factorization is not complete.

    ! info    integer
    ! = 0  for normal return.
    ! = k  signals an error condition.  the leading minor
    ! of order  k  is not positive definite.

    ! linpack.  this version dated 08/14/78 .
    ! cleve moler, university of new mexico, argonne national lab.

    ! subroutines and functions

    ! blas ddot
    ! fortran sqrt

    ! internal variables

    double precision :: ddot,t
    double precision :: s
    integer :: j,jm1,k
    ! begin block with ...exits to 40


    do 30 j = 1, n
      info = j
      s = 0.0d0
      jm1 = j - 1
      if (jm1 < 1) go to 20
      do 10 k = 1, jm1
        t = a(k,j) - ddot(k-1,a(1,k),1,a(1,j),1)
        t = t/a(k,k)
        a(k,j) = t
        s = s + t*t
10    ENDDO
20    continue
      s = a(j,j) - s
      ! ......exit
      if (s <= 0.0d0) go to 40
      a(j,j) = sqrt(s)
30  ENDDO
    info = 0
40  continue
    return
  end subroutine dpofa

  !====================== The end of dpofa ===============================

  subroutine  dscal(n,da,dx,incx)

    ! scales a vector by a constant.
    ! uses unrolled loops for increment equal to one.
    ! jack dongarra, linpack, 3/11/78.
    ! modified 3/93 to return if incx .le. 0.

    double precision :: da,dx(*)
    integer :: i,incx,m,mp1,n,nincx

    if( n <= 0 .OR. incx <= 0 )return
    if(incx == 1)go to 20

    ! code for increment not equal to 1

    nincx = n*incx
    do 10 i = 1,nincx,incx
      dx(i) = da*dx(i)
10  ENDDO
    return

    ! code for increment equal to 1


    ! clean-up loop

20  m = mod(n,5)
    if( m == 0 ) go to 40
    do 30 i = 1,m
      dx(i) = da*dx(i)
30  ENDDO
    if( n < 5 ) return
40  mp1 = m + 1
    do 50 i = mp1,n,5
      dx(i) = da*dx(i)
      dx(i + 1) = da*dx(i + 1)
      dx(i + 2) = da*dx(i + 2)
      dx(i + 3) = da*dx(i + 3)
      dx(i + 4) = da*dx(i + 4)
50  ENDDO
    return
  end subroutine dscal

  !====================== The end of dscal ===============================

  subroutine dtrsl(t,ldt,n,b,job,info)
    integer :: ldt,n,job,info
    double precision :: t(ldt,*),b(*)


    ! dtrsl solves systems of the form

    ! t * x = b
    ! or
    ! trans(t) * x = b

    ! where t is a triangular matrix of order n. here trans(t)
    ! denotes the transpose of the matrix t.

    ! on entry

    ! t         double precision(ldt,n)
    ! t contains the matrix of the system. the zero
    ! elements of the matrix are not referenced, and
    ! the corresponding elements of the array can be
    ! used to store other information.

    ! ldt       integer
    ! ldt is the leading dimension of the array t.

    ! n         integer
    ! n is the order of the system.

    ! b         double precision(n).
    ! b contains the right hand side of the system.

    ! job       integer
    ! job specifies what kind of system is to be solved.
    ! if job is

    ! 00   solve t*x=b, t lower triangular,
    ! 01   solve t*x=b, t upper triangular,
    ! 10   solve trans(t)*x=b, t lower triangular,
    ! 11   solve trans(t)*x=b, t upper triangular.

    ! on return

    ! b         b contains the solution, if info .eq. 0.
    ! otherwise b is unaltered.

    ! info      integer
    ! info contains zero if the system is nonsingular.
    ! otherwise info contains the index of
    ! the first zero diagonal element of t.

    ! linpack. this version dated 08/14/78 .
    ! g. w. stewart, university of maryland, argonne national lab.

    ! subroutines and functions

    ! blas daxpy,ddot
    ! fortran mod

    ! internal variables

    double precision :: ddot,temp
    integer :: kase,j,jj

    ! begin block permitting ...exits to 150

    ! check for zero diagonal elements.

    do 10 info = 1, n
      ! ......exit
      if (t(info,info) == 0.0d0) go to 150
10  ENDDO
    info = 0

    ! determine the task and go to it.

    kase = 1
    if (mod(job,10) /= 0) kase = 2
    if (mod(job,100)/10 /= 0) kase = kase + 2
    ! go to (20,50,80,110), kase
    select case (kase)
    case (1)
      goto 20
    case (2)
      goto 50
    case (3)
      goto 80
    case (4)
      goto 110
    end select

    ! solve t*x=b for t lower triangular

20  continue
    b(1) = b(1)/t(1,1)
    if (n < 2) go to 40
    do 30 j = 2, n
      temp = -b(j-1)
      call daxpy(n-j+1,temp,t(j,j-1),1,b(j),1)
      b(j) = b(j)/t(j,j)
30  ENDDO
40  continue
    go to 140

    ! solve t*x=b for t upper triangular.

50  continue
    b(n) = b(n)/t(n,n)
    if (n < 2) go to 70
    do 60 jj = 2, n
      j = n - jj + 1
      temp = -b(j+1)
      call daxpy(j,temp,t(1,j+1),1,b(1),1)
      b(j) = b(j)/t(j,j)
60  ENDDO
70  continue
    go to 140

    ! solve trans(t)*x=b for t lower triangular.

80  continue
    b(n) = b(n)/t(n,n)
    if (n < 2) go to 100
    do 90 jj = 2, n
      j = n - jj + 1
      b(j) = b(j) - ddot(jj-1,t(j+1,j),1,b(j+1),1)
      b(j) = b(j)/t(j,j)
90  ENDDO
100 continue
    go to 140

    ! solve trans(t)*x=b for t upper triangular.

110 continue
    b(1) = b(1)/t(1,1)
    if (n < 2) go to 130
    do 120 j = 2, n
      b(j) = b(j) - ddot(j-1,t(1,j),1,b(1),1)
      b(j) = b(j)/t(j,j)
120 ENDDO
130 continue
140 continue
150 continue
    return
  end subroutine dtrsl

  !====================== The end of dtrsl ===============================
end module lbfgsbmod

subroutine transformz (sample, nu, n, ifam)
  use modelfcns
  integer, intent(in) :: n, ifam
  double precision, intent(in) :: nu
  double precision, intent(inout) :: sample(n)
  integer i
  call create_model (ifam)
  do i = 1, n
    sample(i) = invtrw(sample(i),nu)
  end do
end subroutine transformz

subroutine llikfcn_dh_tr (dlglk, hlglk, phi, omg, nu, kappa, &
   sample, Ntot, y, l, F, offset, dm, betm0, betQ0, ssqdf, ssqsc, tsqdf, tsq, &
   icf, n, p, ifam, itr)
!! Log-likelihood function 1st and 2nd derivatives w.r.t link and
!! covariance parameters. The order is (link,phi,omg,kappa).
!! itr is the type of transformation used: 0 = no, 1 = mu, 2 = wo.

  use modelfcns, logpdfydmu => logpdfydlnk, logpdfyhmu => logpdfyhlnk
  use covfun
  use betaprior
  implicit none
  integer, intent(in) :: n, p, ifam, Ntot, icf, itr(n)
  double precision, intent(in) :: phi, omg, nu, &
     sample(n, Ntot), y(n), l(n), F(n,p), offset(n), &
     dm(n, n), betm0(p), betQ0(p, p), ssqdf, ssqsc, tsqdf, tsq, kappa
  double precision, intent(out) :: dlglk(4,Ntot), hlglk(4,4,Ntot)
  logical lmxi
  double precision T(n,n), TiF(n,p), FTF(p,p), Ups(n,n), &
     ldh_Ups, ssqdfsc, modeldfh, tsqval, respdfh, xi(n)
  integer i, j, k
  double precision zsam(n), msam(n), alpha, UzzU(n,n)
  double precision djsam(n), hjsam(n)
  double precision zU(n), zUz, zUDTUz(3), zUHTUz, zUDTUDTUz
  double precision dlogpy(n), dmudnu(n), dy, dj, dz, dlogpz(n), dzdnu(n)
  double precision hlogpy(n), hmuhnu(n), hy, hj, hz, hzhnu(n), hlogpz(n,n)
  double precision DT(n,n,3), HT(n,n,3,3)
  double precision trUpsDT(3), trUpsHT(3,3), trUpsDTUpsDT(3,3)
  double precision UpsDT(n,n), DTUpsDT(n,n)
  double precision DzU(n), DzUDTUz(3), DzUz(3), DTzU(n)

  call create_model (ifam)
  call create_spcor (icf,n)

  ssqdfsc = ssqdf*ssqsc
  select case (ifam)
  case (0)
    tsqval = tsqdf*tsq
    respdfh = .5d0*(n + tsqdf)
  case default
    tsqval = tsq
  end select

  ! Determine flat or normal prior
  call betapriorz (modeldfh, xi, lmxi, betm0, betQ0, F, n, p, ssqdf, offset)

  call calc_cov (phi,omg,dm,F,betQ0,kappa,n,p,T,TiF,FTF,Ups,ldh_Ups)
  call fill_symmetric_matrix(Ups,n)

  do i = 1, 3
    DT(:,:,i) = cor_dcov(n, dm, phi, omg, kappa, i)
    trUpsDT(i) = traceAB(Ups,DT(:,:,i),n)
  end do

  do j = 1, 3
    do i = 1, j
      HT(:,:,i,j) = cor_hcov(n, dm, phi, omg, kappa, i, j)
      trUpsHT(i,j) = traceAB(Ups,HT(:,:,i,j),n)
      trUpsDTUpsDT(i,j) = traceABAC(Ups,DT(:,:,i),DT(:,:,j),n)
    end do
  end do

  do k = 1, Ntot
    call rchkusr
    where (itr .eq. 0)
      zsam = sample(:,k)
      msam = invlink(zsam,nu)
      dlogpy = logpdfydmu(y,l,msam)
      hlogpy = logpdfyhmu(y,l,msam)
      dmudnu = invlinkdn(zsam,nu)
      hmuhnu = invlinkhn(zsam,nu)
      djsam = 0d0
      dzdnu = 0d0
      hzhnu = 0d0
      hjsam = 0d0
    elsewhere (itr .eq. 1)
      msam = sample(:,k)
      zsam = flink(msam,nu)
      dlogpy = 0d0
      hlogpy = 0d0
      dzdnu = -invlinkdn(zsam,nu)/invlinkdz(zsam,nu)
      djsam = (invlinkdzdn(zsam,nu) + invlinkhz(zsam,nu)*dzdnu) &
         /invlinkdz(zsam,nu)
      hzhnu = -(invlinkhn(zsam,nu) + 2*invlinkdzdn(zsam,nu)*dzdnu &
         + invlinkhz(zsam,nu)*dzdnu*dzdnu)/invlinkdz(zsam,nu)
      hjsam = (invlink3z(zsam,nu)*dzdnu*dzdnu + invlinkhz(zsam,nu)*hzhnu &
         + 2*invlinkhzdn(zsam,nu)*dzdnu + invlinkdzhn(zsam,nu))&
         /invlinkdz(zsam,nu) - djsam*djsam
      dmudnu = 0d0
      hmuhnu = 0d0
    elsewhere (itr .eq. 2)
      zsam = transfw(sample(:,k),nu)
      msam = invlink(zsam,nu)
      dlogpy = logpdfydmu(y,l,msam)
      hlogpy = logpdfyhmu(y,l,msam)
      dzdnu = -invtrwdn(zsam,nu)/invtrwdz(zsam,nu)
      hzhnu = -(invtrwhn(zsam,nu) + 2*invtrwdzdn(zsam,nu)*dzdnu &
         + invtrwhz(zsam,nu)*dzdnu*dzdnu)/invtrwdz(zsam,nu)
      dmudnu = invlinkdn(zsam,nu) + invlinkdz(zsam,nu)*dzdnu
      hmuhnu = invlinkhn(zsam,nu) + invlinkdz(zsam,nu)*hzhnu &
         + 2*invlinkdzdn(zsam,nu)*dzdnu + invlinkhz(zsam,nu)*dzdnu*dzdnu
      djsam = (invtrwdzdn(zsam,nu) + invtrwhz(zsam,nu)*dzdnu)/invtrwdz(zsam,nu)
      hjsam = (invtrw3z(zsam,nu)*dzdnu*dzdnu + invtrwhz(zsam,nu)*hzhnu &
         + 2*invtrwhzdn(zsam,nu)*dzdnu + invtrwdzhn(zsam,nu))&
         /invtrwdz(zsam,nu) - djsam*djsam
    end where
    if (lmxi) zsam = zsam - xi
    call dsymv ('u',n,1d0,Ups,n,zsam,1,0d0,zU,1)
    zUz = dot_product(zsam,zU) + ssqdfsc
    alpha = -2d0/zUz
    where (upper_tri())
      UzzU = Ups
    end where
    call dsyr ('u',n,alpha,zU,1,UzzU,n)
    ! Link parameter
    dy = dot_product(dlogpy,dmudnu)
    dlogpz = (alpha*modeldfh)*zU
    where (upper_tri())
      hlogpz = (alpha*modeldfh)*UzzU
    end where
    dz = dot_product(dlogpz,dzdnu)
    dj = -sum(djsam)
    hy = dot_product(hlogpy,dmudnu*dmudnu) + dot_product(dlogpy,hmuhnu)
    hz = qform(dzdnu,hlogpz,n) + dot_product(dlogpz,hzhnu)
    hj = -sum(hjsam)
    dlglk(1,k) = dy + dz + dj
    hlglk(1,1,k) = hy + hz + hj
    ! Covariance parameters
    do j = 1, 3
      call dsymv ('u',n,1d0,DT(:,:,j),n,zU,1,0d0,DTzU,1)
      zUDTUz(j) = dot_product(zU,DTzU)/zUz
      call dsymv ('u',n,1d0,Ups,n,dzdnu,1,0d0,DzU,1)
      DzUDTUz(j) = dot_product(DzU,DTzU)
      DzUz(j) = dot_product(dzdnu,zU)
    end do
    do j = 1, 3
      dlglk(j+1,k) = -.5d0*trUpsDT(j) + modeldfh*zUDTUz(j)
    end do
    do j = 1, 3
      call dsymm ('r','u',n,n,1d0,DT(:,:,j),n,Ups,n,0d0,UpsDT,n)
      do i = 1, j
        zUHTUz = qform(zU,HT(:,:,i,j),n)
        call dsymm ('l','u',n,n,1d0,DT(:,:,i),n,UpsDT,n,0d0,DTUpsDT,n)
        DTUpsDT = .5d0*(DTUpsDT + transpose(DTUpsDT))
        zUDTUDTUz = qform(zU,DTUpsDT,n)
        hlglk(i+1,j+1,k) = -.5d0*trUpsHT(i,j) + .5d0*trUpsDTUpsDT(i,j) &
           + modeldfh*zUDTUz(i)*zUDTUz(j) &
           - modeldfh*(zUDTUDTUz + zUDTUDTUz - zUHTUz)/zUz
      end do
    end do
    ! Link-covariance 2nd derivative
    do j = 1, 3
      hlglk(1,j+1,k) = (alpha*modeldfh)*(zUDTUz(j)*DzUz(j) - DzUDTUz(j))
    end do
  end do

contains

  double precision function qform(v,A,n)
    ! Compute v'*A*v where A is symmetric with only the upper triangle
    ! specified.
    integer, intent(in) :: n
    double precision, intent(in) :: v(n), A(n,n)
    double precision Av(n)
    call dsymv ('u',n,1d0,A,n,v,1,0d0,Av,1)
    qform = dot_product(Av,v)
  end function qform

  double precision function traceAB(A,B,n)
    ! Compute the trace(matmul(A,B)) where A,B are symmetric matrices with
    ! only their upper triangular elements specified.
    integer, intent(in) :: n
    double precision, intent(in) :: A(n,n), B(n,n)
    integer j
    double precision c
    traceAB = A(1,1)*B(1,1)
    do j = 2, n
      c = dot_product(A(:j-1,j),B(:j-1,j))
      traceAB = traceAB + c + c + A(j,j)*B(j,j)
    end do
  end function traceAB

  double precision function traceABAC(A,B,C,n)
    ! Compute the trace(matmul(A,B,A,C)) where A is some matrix and B,C are
    ! symmetric matrices with only their upper triangular elements
    ! specified.
    integer, intent(in) :: n
    double precision, intent(in) :: A(n,n), B(n,n), C(n,n)
    !integer i,j
    double precision AB(n,n), AC(n,n)
    call dsymm('r','u',n,n,1d0,B,n,A,n,0d0,AB,n)
    call dsymm('r','u',n,n,1d0,C,n,A,n,0d0,AC,n)
    traceABAC = sum(AB*transpose(AC))
  end function traceABAC

  double precision function cor_dcov (n, dm, phi, omg, kappa, id) result (DT)
    ! id = which derivative do you want? phi, omg, kappa
    ! Return dT where T = omg*I + R(phi,kappa)
    use covfun, only: spcor_dh, spcor_dk, fill_symmetric_matrix, upper_tri
    implicit none
    integer, intent(in) :: n, id
    double precision, intent(in) :: dm(n,n), phi, omg, kappa
    !double precision, intent(out) :: DT(n,n)
    dimension :: DT(n,n)
    integer i
    double precision H(n,n)
    select case (id)
    case (1)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = -(H/phi)*spcor_dh(H,kappa)
        end where
      end if
    case (2)
      DT = 0d0
      do i = 1, n
        DT(i,i) = 1d0
      end do
    case (3)
      if (phi .eq. 0d0) then
        DT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          DT = spcor_dk(H,kappa)
        end where
      end if
    end select
  end function cor_dcov

  double precision function cor_hcov (n, dm, phi, omg, kappa, id, jd) result (HT)
    ! id,jd = which derivative do you want? phi, omg, kappa
    ! Return d^2 T where T = omg*I + R(phi,kappa)
    use covfun, only: spcor_dh, spcor_dk, spcor_hh, spcor_hk, spcor_dhdk, &
       fill_symmetric_matrix, upper_tri
    implicit none
    integer, intent(in) :: n, id, jd
    double precision, intent(in) :: dm(n,n), phi, omg, kappa
    dimension :: HT(n,n)
    integer ij
    double precision H(n,n), H1(n,n)
    ij = id + 3*(jd-1)
    select case (ij)
    case (1) ! (phi,phi)
      if (phi .eq. 0d0) then
        HT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          H1 = -H/phi
          HT = H1*H1*spcor_hh(H,kappa) - (2/phi)*H1*spcor_dh(H,kappa)
        end where
      end if
    case (2,4) ! (phi,omg)
      HT = 0d0
    case (3,7) ! (phi,kappa)
      H = dm/phi
      H1 = -H/phi
      HT = H1*spcor_dhdk(H,kappa)
    case (5) ! (omg,omg)
      HT = 0d0
    case (6,8) ! (omg,kappa)
      HT = 0d0
    case (9) ! (kappa,kappa)
      if (phi .eq. 0d0) then
        HT = 0d0
        return
      else
        where (upper_tri())
          H = dm/phi
          HT = spcor_hk(H,kappa)
        end where
      end if
    end select
  end function cor_hcov
end subroutine llikfcn_dh_tr

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revlogreg.R
\name{revlogreg}
\alias{revlogreg}
\title{Reverse logistic regression estimation}
\usage{
revlogreg(lglk, N)
}
\arguments{
\item{lglk}{The value of the loglikelihood at different samples and
different parameters. This should be entered as a matrix where the
rows are the values of the samples and the columns correspond to
the parameters. The
[i,j] element of the matrix is the value of the loglikelihood at
the ith sample when all samples are put together evaluated
at the jth parameter value.}

\item{N}{A vector of length ncol(lglk) or a scalar corresponding to
the sample sizes from each model. Must sum(N) == nrow(lglk). The
first N[1] samples come from model corresponding to the first set
of parameters, then (N[1]+1):N[2] are from the model corresponding
to the second set of parameters, and so on.}
}
\value{
A vector containing the reverse logistic regression estimates
  of the logarithm of the Bayes factors.
  The first set of parameters is taken as the reference model so its
  estimate is always 0.
}
\description{
Perform the reverse logistic regression estimation
}
\details{
Estimation is done by maximising the reverse logistic log likelihood.
}
\references{
Geyer, C. J. (1994). Estimating normalizing constants
  and reweighting mixtures in Markov chain Monte Carlo. Technical
  Report 568, School of Statistics, University of Minnesota.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calcbf.R
\name{bf2optim}
\alias{bf2optim}
\title{Empirical Bayes estimator}
\usage{
bf2optim(bf1obj, paroptim, useCV = TRUE, control = list())
}
\arguments{
\item{bf1obj}{Output from the function \code{\link{bf1skel}} which
contains the Bayes factors and importance sampling weights.}

\item{paroptim}{A named list with the components "linkp", "phi",
"omg", "kappa". Each component must be numeric with length 1, 2,
or 3 with elements in increasing order but for the binomial family
linkp is also allowed to be the character "logit" and "probit". If
the compontent's length is 1, then the corresponding parameter is
considered to be fixed at that value. If 2, then the two numbers
denote the lower and upper bounds for the optimisation of that
parameter (infinities are allowed). If 3, these correspond to
lower bound, starting value, upper bound for the estimation of
that parameter.}

\item{useCV}{Whether to use control variates for finer
corrections.}

\item{control}{A list of control parameters for the optimisation.
See \code{\link[stats]{optim}}.}
}
\value{
The output from the function \code{\link[stats]{optim}}.
  The \code{"value"} element is the log-Bayes factor, not the
  negative log-Bayes factor.
}
\description{
Estimation by empirical Bayes.
}
\details{
This function is a wrap around \code{\link{bf2new}} using the
"L-BFGS-B" method of the function \code{\link[stats]{optim}} to
estimate the parameters.
}
\examples{
\dontrun{
data(rhizoctonia)
### Define the model
corrf <- "spherical"
kappa <- 0
ssqdf <- 1
ssqsc <- 1
betm0 <- 0
betQ0 <- .01
family <- "binomial.probit"
### Skeleton points
philist <- c(100, 140, 180)
omglist <- c(.5, 1)
parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
### MCMC sizes
Nout <- 100
Nthin <- 1
Nbi <- 0
### Take MCMC samples
runs <- list()
for (i in 1:NROW(parlist)) {
  runs[[i]] <- mcsglmm(Infected ~ 1, family, rhizoctonia, weights = Total,
                       atsample = ~ Xcoord + Ycoord,
                       Nout = Nout*c(.8,.2), Nthin = Nthin, Nbi = Nbi,
                       betm0 = betm0, betQ0 = betQ0,
                       ssqdf = ssqdf, ssqsc = ssqsc,
                       phi = parlist$phi[i], omg = parlist$omg[i],
                       linkp = parlist$linkp[i], kappa = parlist$kappa[i],
                       corrfcn = corrf,
                       corrtuning=list(phi = 0, omg = 0, kappa = 0))
}
bf <- bf1skel(runs)
est <- bf2optim(bf, list(phi = c(100, 200), omg = c(0, 2)))
est
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mcsp2.R
\name{subset.geomcmc}
\alias{subset.geomcmc}
\title{Subset MCMC chain}
\usage{
\method{subset}{geomcmc}(x, subset, ...)
}
\arguments{
\item{x}{Output from the functions \code{\link{mcsglmm}} or
\code{\link{mcstrga}}.}

\item{subset}{Logical or integer vector.}

\item{...}{Further arguments to be passed to or from other methods.}
}
\value{
A similar object as \code{x} with the subsetted chain.
}
\description{
Return subset of MCMC chain.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bfse.R
\name{bmbfse}
\alias{bmbfse}
\title{Batch means, Bayes factors standard errors}
\usage{
bmbfse(
  pargrid,
  runs,
  bfsize1 = 0.8,
  nbatch1 = 0.5,
  nbatch2 = 0.5,
  S1method = c("RL", "MW"),
  bvmethod = c("Standard", "TukeyHanning", "Bartlett"),
  reference = 1,
  transf = c("no", "mu", "wo")
)
}
\arguments{
\item{pargrid}{A data frame with components "linkp", "phi", "omg",
"kappa". Each row gives a combination of the parameters to
compute the new standard errors.}

\item{runs}{A list with outputs from the function
\code{\link{mcsglmm}} or \code{\link{mcstrga}}.}

\item{bfsize1}{A scalar or vector of the same length as
\code{runs} with all integer values or all values in (0, 1]. How
many samples (or what proportion of the sample) to use for
estimating the Bayes factors at the first stage. The remaining
sample will be used for estimating the standard errors in the
second stage. Setting it to 1 will perform only the first stage.}

\item{nbatch1}{A scalar or vector of the same length as
\code{runs}. All values must be integers or less than 1. This is
used for calculating how many batches to split each of the
sample in runs for the calculation of the Bayes factors standard
errors for the parameters corresponding to \code{runs}.}

\item{nbatch2}{A scalar or vector of the same length as
\code{runs}. All values must be integers or less than 1. This is
used for calculating how many batches to split each of the
sample in runs for the calculation of the Bayes factors standard
errors for the parameters corresponding to \code{pargrid}.}

\item{S1method}{Which method to use to calculate the Bayes factors
in stage 1: Reverse logistic or Meng-Wong.}

\item{bvmethod}{Which method to use for the calculation of the
batch variance. The standard method splits to disjoint batches.
The second and third method use the spectral variance method
with different lag windows.}

\item{reference}{Which model goes in the denominator.}

\item{transf}{Whether to use a transformed sample for the
computations. If \code{"no"} or \code{FALSE}, it doesn't. If
\code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
\code{"wo"} it uses an alternative transformation. The latter
can be used only for the families indicated by
\code{.geoBayes_models$haswo}.}
}
\value{
A list with components
\itemize{
\item \code{pargrid} The inputted pargrid augmented with the
  computed log Bayes factors and standard errors.
\item \code{bfEstimate} The estimates of the Bayes factors
\item \code{bfSigma} The covariance matrix for the Bayes factors
estimates.
}
}
\description{
Compute the standard errors for the Bayes factors estimates.
}
\details{
Uses the batch means method to compute the standard errors for
Bayes factors.
}
\references{
Roy, V., Tan, A. and Flegal, J. (2018). Estimating
  standard errors for importance sampling estimators with multiple
  Markov chains, Statistica Sinica, 28 1079-1101.

Roy, V., & Evangelou, E. (2018). Selection of proposal
  distributions for generalized importance sampling estimators.
  arXiv preprint arXiv:1805.00829.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geobayes-correlation.R
\docType{data}
\name{geoBayes_correlation}
\alias{geoBayes_correlation}
\alias{.geoBayes_corrfcn}
\title{Spatial correlation used in the geoBayes package}
\format{
An object of class \code{data.frame} with 5 rows and 4 columns.
}
\usage{
.geoBayes_corrfcn
}
\description{
This hidden variable contains a choice of correlation functions
that can be fit with this package. The function can be chosen in
the \code{corrfcn} input of the relevant function. This variable
cannot be overwritten.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bf.R
\name{bf1skel}
\alias{bf1skel}
\title{Computation of Bayes factors at the skeleton points}
\usage{
bf1skel(
  runs,
  bfsize1 = 0.8,
  method = c("RL", "MW"),
  reference = 1,
  transf = c("no", "mu", "wo")
)
}
\arguments{
\item{runs}{A list with outputs from the function
\code{\link{mcsglmm}} or \code{\link{mcstrga}}.}

\item{bfsize1}{A scalar or vector of the same length as
\code{runs} with all integer values or all values in (0, 1]. How
many samples (or what proportion of the sample) to use for
estimating the Bayes factors at the first stage. The remaining
sample will be used for estimating the Bayes factors in the
second stage. Setting it to 1 will perform only the first stage.}

\item{method}{Which method to use to calculate the Bayes factors:
Reverse logistic or Meng-Wong.}

\item{reference}{Which model goes in the denominator.}

\item{transf}{Whether to use a transformed sample for the
computations. If \code{"no"} or \code{FALSE}, it doesn't. If
\code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
\code{"wo"} it uses an alternative transformation. The latter
can be used only for the families indicated by
\code{.geoBayes_models$haswo}.}
}
\value{
A list with components
\itemize{
\item \code{logbf} A vector containing logarithm of the Bayes factors.
\item \code{logLik1} \code{logLik2} Matrices with the values of
the log-likelihood computed from the samples for each model at the
first and second stages.
\item \code{isweights} A vector with the importance sampling
weights for computing the Bayes factors at new points that will be
used at the second stage. Used internally in
\code{\link{bf2new}} and \code{\link{bf2optim}}.
\item \code{controlvar} A matrix with the control variates
computed at the samples that will be used in the second stage.
\item \code{sample2} The MCMC sample for mu or z that will be
used in the second stage. Used internally in
\code{\link{bf2new}} and \code{\link{bf2optim}}.
\item \code{N1}, \code{N2} Vectors containing the sample sizes
used in the first and second stages.
\item \code{distmat} Matrix of distances between locations.
\item \code{betm0}, \code{betQ0}, \code{ssqdf}, \code{ssqsc},
\code{tsqdf}, \code{tsqsc}, \code{dispersion}, \code{response},
\code{weights}, \code{modelmatrix}, \code{locations},
\code{family}, \code{corrfcn}, \code{transf} Model parameters used
internally in.
\code{\link{bf2new}} and \code{\link{bf2optim}}.
\item \code{pnts} A list containing the skeleton points. Used
internally in \code{\link{bf2new}} and \code{\link{bf2optim}}.
}
}
\description{
Function to compute the Bayes factors from MCMC samples.
}
\details{
Computes the Bayes factors using \code{method} with respect to
\code{reference}.
}
\examples{
\dontrun{
data(rhizoctonia)
### Define the model
corrf <- "spherical"
kappa <- 0
ssqdf <- 1
ssqsc <- 1
betm0 <- 0
betQ0 <- .01
family <- "binomial.probit"
### Skeleton points
philist <- c(100, 140, 180)
omglist <- c(.5, 1)
parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
### MCMC sizes
Nout <- 100
Nthin <- 1
Nbi <- 0
### Take MCMC samples
runs <- list()
for (i in 1:NROW(parlist)) {
  runs[[i]] <- mcsglmm(Infected ~ 1, family, rhizoctonia, weights = Total,
                       atsample = ~ Xcoord + Ycoord,
                       Nout = Nout, Nthin = Nthin, Nbi = Nbi,
                       betm0 = betm0, betQ0 = betQ0,
                       ssqdf = ssqdf, ssqsc = ssqsc,
                       phi = parlist$phi[i], omg = parlist$omg[i],
                       linkp = parlist$linkp[i], kappa = parlist$kappa[i],
                       corrfcn = corrf,
                       corrtuning=list(phi = 0, omg = 0, kappa = 0))
}
bf <- bf1skel(runs)
bf$logbf
}
}
\references{
Geyer, C. J. (1994). Estimating normalizing constants
and reweighting mixtures. Technical report, University of
Minnesota.

Meng, X. L., & Wong, W. H. (1996). Simulating ratios of
normalizing constants via a simple identity: A theoretical
exploration. \emph{Statistica Sinica}, 6, 831-860.

Roy, V., Evangelou, E., and Zhu, Z. (2015). Efficient estimation
and prediction for the Bayesian spatial generalized linear mixed
model with flexible link functions. \emph{Biometrics}, 72(1), 289-298.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spcovariance.R
\name{spcovariance}
\alias{spcovariance}
\alias{spcovariance.formula}
\alias{spcovariance.numeric}
\alias{spcovariance.dist}
\title{Spatial variance-covariance matrix}
\usage{
spcovariance(...)

\method{spcovariance}{formula}(
  formula,
  data,
  subset,
  corrfcn,
  ssq,
  phi,
  omg,
  kappa,
  longlat = FALSE,
  ...
)

\method{spcovariance}{numeric}(x, corrfcn, ssq, phi, omg, kappa, ...)

\method{spcovariance}{dist}(x, corrfcn, ssq, phi, omg, kappa, ...)
}
\arguments{
\item{...}{Further arguments. Not currently in use.}

\item{formula}{A formula of the form \code{~ Xcoord + Ycoord}
specifying the sampled locations.}

\item{data}{An optional data frame containing the variables in the
model.}

\item{subset}{An optional set of indices. The covariance will be
calculated for those coordinates only.}

\item{corrfcn}{The correlation function to use.}

\item{ssq}{The partial sill parameter.}

\item{phi}{The spatial range parameter.}

\item{omg}{The relative nugget parameter.}

\item{kappa}{The spatial smoothness parameter.}

\item{longlat}{How to compute the distance between locations. If
\code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
distance. See \code{\link[sp]{spDists}}.}

\item{x}{A numerical object of distances.}
}
\value{
For a formula input, a variance-covariance matrix. For a
  numeric input, an object of the the same dimensions as its first
  input.
}
\description{
Calculates the spatial variance-covariance matrix for a selection
of correlation functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mcsp2.R
\name{mcstrga}
\alias{mcstrga}
\title{MCMC samples from the transformed Gaussian model}
\usage{
mcstrga(
  formula,
  data,
  weights,
  subset,
  offset,
  atsample,
  corrfcn = "matern",
  linkp,
  phi,
  omg,
  kappa,
  Nout,
  Nthin = 1,
  Nbi = 0,
  betm0,
  betQ0,
  ssqdf,
  ssqsc,
  tsqdf,
  tsqsc,
  corrpriors,
  corrtuning,
  longlat = FALSE,
  test = FALSE
)
}
\arguments{
\item{formula}{A representation of the model in the form
\code{response ~ terms}. The response must be set to \code{NA}'s
at the prediction locations (see the example in
\code{\link{mcsglmm}} for how to do this using
\code{\link{stackdata}}). At the observed locations the response
is assumed to be a total of replicated measurements. The number of
replications is inputted using the argument \code{weights}.}

\item{data}{An optional data frame containing the variables in the
model.}

\item{weights}{An optional vector of weights. Number of replicated
samples.}

\item{subset}{An optional vector specifying a subset of
observations to be used in the fitting process.}

\item{offset}{See \code{\link[stats]{lm}}.}

\item{atsample}{A formula in the form \code{~ x1 + x2 + ... + xd}
with the coordinates of the sampled locations.}

\item{corrfcn}{Spatial correlation function. See
\code{\link{geoBayes_correlation}} for details.}

\item{linkp}{Parameter of the link function. A scalar value.}

\item{phi}{Optional starting value for the MCMC for the
spatial range parameter \code{phi}. Defaults to the mean of its
prior. If \code{corrtuning[["phi"]]} is 0, then this argument is required and
it corresponds to the fixed value of \code{phi}. This can be a
vector of the same length as Nout.}

\item{omg}{Optional starting value for the MCMC for the
relative nugget parameter \code{omg}. Defaults to the mean of
its prior. If \code{corrtuning[["omg"]]} is 0, then this argument is required
and it corresponds to the fixed value of \code{omg}. This can be
a vector of the same length as Nout.}

\item{kappa}{Optional starting value for the MCMC for the
spatial correlation parameter \code{kappa} (Matern smoothness or
exponential power). Defaults to the mean of
its prior. If \code{corrtuning[["kappa"]]} is 0 and it is needed for
the chosen correlation function, then this argument is required
and it corresponds to the fixed value of \code{kappa}. This can be
a vector of the same length as Nout.}

\item{Nout}{Number of MCMC samples to return. This can be a vector
for running independent chains.}

\item{Nthin}{The thinning of the MCMC algorithm.}

\item{Nbi}{The burn-in of the MCMC algorithm.}

\item{betm0}{Prior mean for beta (a vector or scalar).}

\item{betQ0}{Prior standardised precision (inverse variance)
matrix. Can be a scalar, vector or matrix. The first two imply a
diagonal with those elements. Set this to 0 to indicate a flat
improper prior.}

\item{ssqdf}{Degrees of freedom for the scaled inverse chi-square
prior for the partial sill parameter.}

\item{ssqsc}{Scale for the scaled inverse chi-square prior for the
partial sill parameter.}

\item{tsqdf}{Degrees of freedom for the scaled inverse chi-square
prior for the measurement error parameter.}

\item{tsqsc}{Scale for the scaled inverse chi-square prior for the
measurement error parameter.}

\item{corrpriors}{A list with the components \code{phi},
\code{omg} and \code{kappa} as needed. These correspond to the
prior distribution parameters. For \code{phi} and \code{omg} it
must be a vector of length 4. The generalized inverse gamma
prior is assumed and the input corresponds to the parameters
scale, shape, exponent, location in that order (see Details).
For \code{kappa} it must be a vector of length 2. A uniform
prior is assumed and the input corresponds to the lower and
upper bounds in that order.}

\item{corrtuning}{A vector or list with the components \code{phi},
\code{omg} and \code{kappa} as needed. These correspond to the
random walk parameter for the Metropolis-Hastings step. Smaller values
increase the acceptance ratio. Set this to 0 for fixed
parameter value.}

\item{longlat}{How to compute the distance between locations. If
\code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
distance. See \code{\link[sp]{spDists}}.}

\item{test}{Whether this is a trial run to monitor the acceptance
ratio of the random walk for \code{phi} and \code{omg}. If set to
\code{TRUE}, the acceptance ratio will be printed on the screen
every 100 iterations of the MCMC. Tune the \code{phisc} and
\code{omgsc} parameters in order to achive 20 to 30\% acceptance.
Set this to a positive number to change the default 100. No
thinning or burn-in are done when testing.}
}
\value{
A list containing the objects \code{MODEL}, \code{DATA},
  \code{FIXED}, \code{MCMC} and \code{call}. The MCMC samples are
  stored in the object \code{MCMC} as follows:
\itemize{
 \item \code{z} A matrix containing the MCMC samples for the
spatial random field. Each column is one sample.
 \item \code{mu} A matrix containing the MCMC samples for the
mean response (a transformation of z). Each column is one sample.
 \item \code{beta} A matrix containing the MCMC samples for the
regressor coefficients. Each column is one sample.
 \item \code{ssq} A vector with the MCMC samples for the partial
 \item \code{tsq} A vector with the MCMC samples for the
measurement error variance.
 \item \code{phi} A vector with the MCMC samples for the spatial
range parameter, if sampled.
 \item \code{omg} A vector with the MCMC samples for the relative
nugget parameter, if sampled.
 \item \code{logLik} A vector containing the value of the
log-likelihood evaluated at each sample.
 \item \code{acc_ratio} The acceptance ratio for the joint update
of the parameters \code{phi} and \code{omg}, if sampled.
 \item \code{sys_time} The total computing time for the MCMC sampling.
 \item \code{Nout}, \code{Nbi},  \code{Nthin} As in input. Used
internally in other functions.
}
The other objects contain input variables. The object \code{call}
  contains the function call.
}
\description{
Draw MCMC samples from the transformed Gaussian model with known
link function
}
\details{
Simulates from the posterior distribution of this model.
}
\examples{
\dontrun{
### Load the data
data(rhizoctonia)
rhiz <- na.omit(rhizoctonia)
rhiz$IR <- rhiz$Infected/rhiz$Total # Incidence rate of the
                              # rhizoctonia disease

### Define the model
corrf <- "spherical"
ssqdf <- 1
ssqsc <- 1
tsqdf <- 1
tsqsc <- 1
betm0 <- 0
betQ0 <- diag(.01, 2, 2)
phiprior <- c(200, 1, 1000, 100) # U(100, 300)
phisc <- 1
omgprior <- c(3, 1, 1000, 0) # U(0, 3)
omgsc <- 1
linkp <- 1

## MCMC parameters
Nout <- 100
Nbi <- 0
Nthin <- 1

samplt <- mcstrga(Yield ~ IR, data = rhiz,
                  atsample = ~ Xcoord + Ycoord, corrf = corrf,
                  Nout = Nout, Nthin = Nthin,
                  Nbi = Nbi, betm0 = betm0, betQ0 = betQ0,
                  ssqdf = ssqdf, ssqsc = ssqsc,
                  tsqdf = tsqdf, tsqsc = tsqsc,
                  corrprior = list(phi = phiprior, omg = omgprior),
                  linkp = linkp,
                  corrtuning = list(phi = phisc, omg = omgsc, kappa = 0),
                  test=10)

sample <- update(samplt, test = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{stackdata}
\alias{stackdata}
\title{Combine \code{data.frame}s}
\usage{
stackdata(..., fillwith = NA, keepclass = FALSE)
}
\arguments{
\item{...}{\code{data.frame}s or objects that can be coerced to
\code{data.frame}s}

\item{fillwith}{Which value to use for missing variables. This
could be a scalar, a named vector, or a named list with one value
in each component; see Details.}

\item{keepclass}{Whether to preserve the \code{\link[base]{class}}
of each variable. The elements in \code{fillwith} are coerced to
the corresponding variable's class.}
}
\value{
A stacked \code{data.frame}.
}
\description{
Combine \code{data.frame}s
}
\details{
This function combines \code{data.frame}s by filling in missing
variables. This is useful for combining data from
sampled locations with prediction locations.

If \code{fillwith} is a named object, its names must correspond to
the names of variables in the data frames. If a variable is
missing, then it is filled with the corresponding value in
\code{fillwith}. \code{fillwith} can contain only one unnamed
component which corresponds to the default filling.
}
\examples{
\dontrun{
d1 <- data.frame(w = 1:3, z = 4:6 + 0.1)
d2 <- data.frame(w = 3:7, x = 1:5, y = 6:10)
(d12a <- stackdata(d1, d2))
lapply(d12a, class)
(d12b <- stackdata(d1, d2, fillwith = c(x = NA, y = 0, z = -99)))
lapply(d12b, class)
(d12c <- stackdata(d1, d2, fillwith = c(x = NA, y = 0, z = -99),
                   keepclass = TRUE))
lapply(d12c, class)
(d12d <- stackdata(d1, d2, fillwith = c(x = NA, 0)))

data(rhizoctonia)
predgrid <- mkpredgrid2d(rhizoctonia[c("Xcoord", "Ycoord")],
                         par.x = 100, chull = TRUE, exf = 1.2)
rhizdata <- stackdata(rhizoctonia, predgrid$grid)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skelpnts.R
\name{alik_cutoff}
\alias{alik_cutoff}
\alias{alik_plot}
\title{Approximate log-likelihood calculation}
\usage{
alik_cutoff(likopt, par_vals, likthreshold)

alik_plot(alikobj)
}
\arguments{
\item{likopt}{Output from the function \code{\link{alik_optim}}.}

\item{par_vals}{A named list with some of the components "linkp",
"phi", "omg", "kappa".}

\item{likthreshold}{A threshold value proportion to calculate the
cutoff. The cutoff will be calculated as that proportion
relative to the maximum value of the log-likelihood.}

\item{alikobj}{Output from \code{\link{alik_cutoff}}.}
}
\value{
A list with the log-likelihood approximation and cutoff values.

Draws a plot.
}
\description{
Calculate the likelihood approximation at different parameter
values. This function is useful for choosing the skeleton set.

Plot likelihood approximation.
}
\details{
The input \code{par_vals} is meant to contain vector of parameter
values for each parameter. For each element in \code{par_vals},
the other parameters are set equal to the maximisers given in
\code{likopt} and the approximate likelihood is computed. The
cuttoff is calculated using linear interpolation provided by
\code{\link[stats]{approx}}.

The plot can be used to visualise the Laplace approximation to the
likelihood provided by the function \code{\link{alik_cutoff}}.
}
\references{
Evangelou, E., & Roy, V. (2019). Estimation and prediction for
  spatial generalized linear mixed models with parametric links
  via reparameterized importance sampling. Spatial Statistics, 29,
  289-315.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate.R
\name{rsglmm}
\alias{rsglmm}
\alias{rstrga}
\alias{rsgrf}
\title{Simulation from a spatial model}
\usage{
rsglmm(
  n,
  formula,
  family = "gaussian",
  data,
  weights,
  subset,
  offset,
  atsample,
  beta,
  linkp,
  phi,
  omg,
  kappa,
  ssq,
  corrfcn = "matern",
  longlat = FALSE,
  dispersion = 1,
  returnGRF = FALSE,
  warndisp = TRUE
)

rstrga(
  n,
  formula,
  data,
  weights,
  subset,
  offset,
  atsample,
  beta,
  linkp,
  phi,
  omg,
  kappa,
  ssq,
  corrfcn = "matern",
  longlat = FALSE,
  dispersion = 1,
  returnGRF = FALSE
)

rsgrf(
  n,
  formula,
  data,
  subset,
  offset,
  atsample,
  beta,
  phi,
  omg,
  kappa,
  ssq,
  corrfcn = "matern",
  longlat = FALSE
)
}
\arguments{
\item{n}{The number of instances to simulate}

\item{formula}{A representation of the model in the form
\code{response ~ terms}. The LHS can be omitted. If the LHS
exists, it can be of the form \code{y}, \code{y|z}, or sums of
terms at either side of the \code{|} to specify the names of the
variables to include in the data frame.}

\item{family}{The distribution of the data to simulate from.}

\item{data}{An optional data frame containing the variables in the
model.}

\item{weights}{An optional vector of weights. Number of replicated
samples for Gaussian and gamma, number of trials for binomial,
time length for Poisson.}

\item{subset}{An optional set of indices. Simulations will be
provided for those locations only.}

\item{offset}{See \code{\link[stats]{lm}}.}

\item{atsample}{A formula of the form \code{~ Xcoord + Ycoord}
specifying the sampled locations.}

\item{beta}{A vector of the regressor coefficents to use.}

\item{linkp}{The link function parameter.}

\item{phi}{The spatial range parameter.}

\item{omg}{The relative nugget parameter.}

\item{kappa}{The spatial smoothness parameter.}

\item{ssq}{The partial sill parameter.}

\item{corrfcn}{The correlation function to use.}

\item{longlat}{How to compute the distance between locations. If
\code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
distance. See \code{\link[sp]{spDists}}.}

\item{dispersion}{The fixed dispersion parameter. When this is not
1 and the sample is from a binomial or a Poisson distribution,
no such distribution exists so an approximate sample is
returned. Use with caution.}

\item{returnGRF}{Whether to return the simulate Gaussian random
field as well.}

\item{warndisp}{Whether to warn when sampling from a quasi
distribution. This is the case for binomial and Poisson when the
dispersion is not one.}
}
\value{
A data frame containing the predictors, sampling
  locations, optional weights, and samples.
}
\description{
Simulate from a variety of spatial models.
}
\details{
The spatial Gaussian random field is simulated using the Cholesky
decomposition of the covariance matrix.

The sample from a quasi distribution uses a hack which matches the
mean and the variance of the distribution. See the source code for
details.
}
\examples{
\dontrun{
n <- 100
beta <- c(-2, 1)
phi <- .2
omg <- .3
linkp <- 0
ssq <- 1
l <- rep(10, n)
corrf <- "matern"
kappa <- .5
family <- "poisson"
Xcoord <- runif(n)
Ycoord <- runif(n)
f <- Xcoord + Ycoord
formula <- y|z ~ f
mydata <- rsglmm(1, formula, family, weights = l,
                 atsample = ~ Xcoord + Ycoord, beta = beta, linkp = linkp,
                 phi = phi, omg = omg, kappa = kappa, ssq = ssq,
                 corrfcn = corrf, returnGRF = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{mkpredgrid2d}
\alias{mkpredgrid2d}
\title{Make prediction grid}
\usage{
mkpredgrid2d(
  pnts.x,
  pnts.y,
  par.x,
  par.y,
  isby = FALSE,
  chull = FALSE,
  exf = 1
)
}
\arguments{
\item{pnts.x}{x coordinate of the domain. Could also be a
two-column matrix containing the x and y coordinates}

\item{pnts.y}{y coordinate of the domain. Should be
omitted or set to \code{NULL} if the argument \code{pnts.x} is a
two-column matrix.}

\item{par.x}{A scalar parameter for the x component of the new
grid. This parameter corresponds to either the \code{by} or the
\code{length.out} arguments of the function
\code{\link[base]{seq}}. Could also be a vector of two elements
containing the parameter for x and y.}

\item{par.y}{As in \code{par.x} for the y component of the new
grid. Should be omitted or set to \code{NULL} if the argument
\code{par.x} is a two-dimensional vector.}

\item{isby}{If \code{TRUE}, the arguments \code{par.x} and
\code{par.y} correspond to the \code{by} argument of the function
\code{\link[base]{seq}}, otherwise they correspond to
\code{length.out}.}

\item{chull}{Whether to calculate the convex hull of the points.
Set this to \code{TRUE} if \code{pnts.x} and \code{pnts.y} denote
the sampled locations. If they correspond to the borders of the
domain, it is recommended to set this to \code{FALSE}.}

\item{exf}{An expansion factor of the convex hull of
\code{cbind(pnts.x, pnts.y)}. Must be positive. If larger or
smaller than 1, the convex hull is respectively expanded or
contracted.}
}
\value{
A list with components
\itemize{
\item \code{grid} A two-column matrix with the prediction grid.
\item \code{xycoord} A list with components "x" and "y"
containing the sequence of points used to create the grid.
\item \code{xygrid} A matrix with the full square grid derived
from \code{xycoord}.
\item \code{borders} The (expanded) borders of the domain.
\item \code{inxygrid} A logical vector indicating which rows of
  \code{xycoord} fall inside \code{borders}, and therefore
  correspond to the \code{grid}.
}
}
\description{
This function creates a grid for prediction.
}
\details{
If \code{chull} this function first calculates the convex hull of
the points. If \code{exf} is not 1 the borders are expanded. Then
the function calls \code{\link[sp]{point.in.polygon}} to select
points that fall inside the borders.
}
\examples{
\dontrun{
data(rhizoctonia)
predgrid <- mkpredgrid2d(rhizoctonia[c("Xcoord", "Ycoord")],
                         par.x = 100, chull = TRUE, exf = 1.2)
plot(predgrid$borders, type = "l")         # Domain for prediction
points(predgrid$grid, pch = 20, cex = .3)  # Prediction locations
points(rhizoctonia[c("Xcoord", "Ycoord")]) # Observed locations
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seeb.R
\name{bf2se}
\alias{bf2se}
\title{Empirical Bayes standard errors}
\usage{
bf2se(mcrun, transf = c("no", "mu", "wo"))
}
\arguments{
\item{mcrun}{The output from the function \code{mcsglmm} where the
parameters linkp, phi, omg, kappa are set at their empirical
Bayes estimates (output of \code{\link{bf2optim}}).}

\item{transf}{The type of transformation to use.}
}
\value{
The precision (inverse covariance) matrix.
}
\description{
Standard errors for the empirical Bayes estimates of the parameters.
}
\references{
Casella, G. (2001). Empirical Bayes Gibbs sampling.
  Biostatistics, 2(4), 485-500.

Evangelou, E., & Roy, V. (2019). Estimation and prediction for
  spatial generalized linear mixed models with parametric links
  via reparameterized importance sampling. Spatial Statistics, 29,
  289-315.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geobayes-models.R
\docType{data}
\name{geoBayes_models}
\alias{geoBayes_models}
\alias{.geoBayes_models}
\title{Models used in the geoBayes package}
\format{
An object of class \code{data.frame} with 12 rows and 7 columns.
}
\usage{
.geoBayes_models
}
\description{
This hidden variable contains a choice of models that can be fit
with this package. The model can be chosen in the \code{family}
input of the relevant function. This variable cannot be
overwritten.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linkfcn.R
\name{linkfcn}
\alias{linkfcn}
\alias{linkinv}
\title{Calculate the link function for exponential families}
\usage{
linkfcn(mu, linkp, family = "gaussian")

linkinv(z, linkp, family = "gaussian")
}
\arguments{
\item{mu}{Numeric. The mean of the response variable.}

\item{linkp}{The link function parameter. A scalar.}

\item{family}{The distribution of the response variable from
\code{\link{.geoBayes_models}}. Either an integer or the
family name.}

\item{z}{Numeric. The linear predictor.}
}
\value{
A numeric array of the same dimension as the function's
first argument.
}
\description{
Link function for the exponential family.
}
\details{
\code{linkfcn} maps the mean of the response variable \code{mu} to
the linear predictor \code{z}. \code{linkinv} is its inverse.

For the Gaussian family, if the link parameter is positive, then
the extended link is used, defined by \deqn{z =
\frac{sign(\mu)|\mu|^\nu - 1}{\nu}}{z = (sign(mu)*abs(mu)^nu -
1)/nu} In the other case, the usual Box-Cox link is used.


For the Poisson and gamma families, if the link parameter is
positive, then the link is defined by \deqn{z = \frac{sign(w)
(e^{\nu |w|}-1)}{\nu}}{z = sign(w)*expm1(nu*w)/nu} where
\eqn{w = \log(\mu)}{w = log(mu)}. In the other case, the usual
Box-Cox link is used.

For the GEV binomial family, the link function is defined by
\deqn{\mu = 1 - \exp\{-\max(0, 1 + \nu z)^{\frac{1}{\nu}}\}}{mu =
1 - exp[-max(0, 1 + nu z)^(1/nu)]} for any real \eqn{\nu}{nu}. At
\eqn{\nu = 0}{nu = 0} it reduces to the complementary log-log
link.

The Wallace binomial family is a fast approximation to the robit
family. It is defined as \deqn{\mu =
\Phi(\mbox{sign}(z) c(\nu) \sqrt{\nu \log(1 + z^2/\nu)})}{mu =
Phi(sign(z) c(nu) sqrt{nu log(1 + z^2/nu)})}
where \eqn{c(\nu) = (8\nu+1)/(8\nu+3)}{c(nu) = (8*nu+1)/(8*nu+3)}
}
\examples{
\dontrun{
mu <- seq(0.1, 0.9, 0.1)
linkfcn(mu, 7, "binomial")       # robit(7) link function
linkfcn(mu, , "binomial.logit")  # logit link function

mu <- seq(-3, 3, 1)
linkfcn(mu, 0.5, "gaussian")     # sqrt transformation
linkinv(linkfcn(mu, 0.5, "gaussian"), 0.5, "gaussian")
curve(linkfcn(x, 0.5, "gaussian"), -3, 3)
}
}
\references{
Evangelou, E., & Roy, V. (2019). Estimation and prediction for
  spatial generalized linear mixed models with parametric links
  via reparameterized importance sampling. Spatial Statistics, 29,
  289-315.
}
\concept{linkfcn}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selpropdist.R
\name{select_proposals}
\alias{select_proposals}
\alias{select_proposals_SEQ}
\alias{select_proposals_MNX}
\alias{select_proposals_ENT}
\title{Selection of multiple importance sampling distributions}
\usage{
select_proposals_SEQ(
  pargrid,
  K,
  istart,
  relativeSE = FALSE,
  N1,
  N2,
  Nthin,
  Nbi,
  formula,
  family = "gaussian",
  data,
  weights,
  subset,
  offset,
  atsample,
  corrfcn = "matern",
  betm0,
  betQ0,
  ssqdf,
  ssqsc,
  dispersion = 1,
  longlat = FALSE,
  nbatch1 = 0.5,
  nbatch2 = 0.5,
  S1method = c("RL", "MW"),
  bvmethod = c("Standard", "TukeyHanning", "Bartlett"),
  transf = c("no", "mu", "wo")
)

select_proposals_MNX(
  pargrid,
  istart,
  nfix,
  relativeSE = FALSE,
  N1,
  N2,
  Nthin,
  Nbi,
  cooling,
  formula,
  family = "gaussian",
  data,
  weights,
  subset,
  offset,
  atsample,
  corrfcn = "matern",
  betm0,
  betQ0,
  ssqdf,
  ssqsc,
  dispersion = 1,
  longlat = FALSE,
  nbatch1 = 0.5,
  nbatch2 = 0.5,
  S1method = c("RL", "MW"),
  bvmethod = c("Standard", "TukeyHanning", "Bartlett"),
  transf = c("no", "mu", "wo"),
  verbose = FALSE
)

select_proposals_ENT(
  pargrid,
  istart,
  nfix,
  relativeSE = FALSE,
  N1,
  Nthin,
  Nbi,
  cooling,
  formula,
  family = "gaussian",
  data,
  weights,
  subset,
  offset,
  atsample,
  corrfcn = "matern",
  betm0,
  betQ0,
  ssqdf,
  ssqsc,
  dispersion = 1,
  longlat = FALSE,
  nbatch1 = 0.5,
  nbatch2 = 0.5,
  S1method = c("RL", "MW"),
  bvmethod = c("Standard", "TukeyHanning", "Bartlett"),
  transf = c("no", "mu", "wo"),
  verbose = FALSE
)
}
\arguments{
\item{pargrid}{A data frame with components "linkp", "phi", "omg",
"kappa". Each row gives a combination of the parameters to
compute the new standard errors.}

\item{K}{How many proposal densities in total to choose among the
rows of \code{pargrid}. Needed for SEQ only. For MNX and ENT
this is determined by the length of \code{istart}.}

\item{istart}{Start with these rows of \code{pargrid}. A vector of
indices.}

\item{relativeSE}{Logical. Whether the choice is based on the
standard error (FALSE), or relative standard error (TRUE).}

\item{N1}{The sample size for stage 1.}

\item{N2}{The sample sie for stage 2.}

\item{Nthin}{Thinning}

\item{Nbi}{Burn-in}

\item{formula}{A representation of the model in the form
\code{response ~ terms}. The response must be set to \code{NA}'s
at the prediction locations (see the examples on how to do this
using the function \code{\link{stackdata}}). At the observed
locations the response is assumed to be a total of replicated
measurements. The number of replications is inputted using the
argument \code{weights}.}

\item{family}{The distribution of the data. The
\code{"GEVbinomial"} family is the binomial family with link the
GEV link (see Details).}

\item{data}{An optional data frame containing the variables in the
model.}

\item{weights}{An optional vector of weights. Number of replicated
samples for Gaussian and gamma, number of trials for binomial,
time length for Poisson.}

\item{subset}{An optional vector specifying a subset of
observations to be used in the fitting process.}

\item{offset}{See \code{\link[stats]{lm}}.}

\item{atsample}{A formula in the form \code{~ x1 + x2 + ... + xd}
with the coordinates of the sampled locations.}

\item{corrfcn}{Spatial correlation function. See
\code{\link{geoBayes_correlation}} for details.}

\item{betm0}{Prior mean for beta (a vector or scalar).}

\item{betQ0}{Prior standardised precision (inverse variance)
matrix. Can be a scalar, vector or matrix. The first two imply a
diagonal with those elements. Set this to 0 to indicate a flat
improper prior.}

\item{ssqdf}{Degrees of freedom for the scaled inverse chi-square
prior for the partial sill parameter.}

\item{ssqsc}{Scale for the scaled inverse chi-square prior for the
partial sill parameter.}

\item{dispersion}{The fixed dispersion parameter.}

\item{longlat}{How to compute the distance between locations. If
\code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
distance. See \code{\link[sp]{spDists}}.}

\item{nbatch1}{A scalar or vector of the same length as
\code{runs}. All values must be integers or less than 1. This is
used for calculating how many batches to split each of the
sample in runs for the calculation of the Bayes factors standard
errors for the parameters corresponding to \code{runs}.}

\item{nbatch2}{A scalar or vector of the same length as
\code{runs}. All values must be integers or less than 1. This is
used for calculating how many batches to split each of the
sample in runs for the calculation of the Bayes factors standard
errors for the parameters corresponding to \code{pargrid}.}

\item{S1method}{Which method to use to calculate the Bayes
factors: Reverse logistic or Meng-Wong.}

\item{bvmethod}{Which method to use for the calculation of the
batch variance. The standard method splits to disjoint batches.
The second and third method use the spectral variance method
with different lag windows.}

\item{transf}{Whether to use a transformed sample for the
computations. If \code{"no"} or \code{FALSE}, it doesn't. If
\code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
\code{"wo"} it uses an alternative transformation. The latter
can be used only for the families indicated by
\code{.geoBayes_models$haswo}.}

\item{nfix}{In the case of MNX and ENT, the first \code{nfix}
elements of \code{istart} will always be included.}

\item{cooling}{A decreasing sequence of temperature values for the
simulated annealing. All elements must be positive. A suggested
value is \code{Tinit / log(((0:N) \%/\% Tstp)*Tstp + exp(1))} for
\code{N+1} iterations, where \code{Tinit} is the initial
temperature and \code{Tstp} is the number of iterations before
the temperature is reduced.}

\item{verbose}{Logical. Prints information about the simulated
annealing.}
}
\value{
A list with components
\describe{
\item{selected}{The rows of \code{pargrid} selected.}
\item{isel}{The indices of the rows of \code{pargrid} selected.}
\item{se}{The standard error corresponding to the selected
  parameters.}
\item{samples}{A list containing the samples from the selected
  parameters.} 
}
}
\description{
Selection of multiple importance sampling distributions
}
\details{
\describe{
\item{SEQ}{is a sequential method starting with \code{istart} and
additng to it until \code{K} proposals have been selected. At each
iteration, the point with the highest (relative?) standard error
is added} 
\item{MNX}{is the minimax method. The chosen proposal corresponds
to the lowest maximum (relative?) standard error.}
\item{ENT}{is the entropy method. The chosen proposal corresponds
to the highest determinant of the (relative?) covariance matrix at
the first stage.}
}
}
\examples{
\dontrun{
data(rhizoctonia)
### Define the model
corrf <- "spherical"
kappa <- 0
ssqdf <- 1
ssqsc <- 1
betm0 <- 0
betQ0 <- .01
family <- "binomial.probit"
formula <- Infected ~ 1
atsample <- ~ Xcoord + Ycoord
### Skeleton points
philist <- seq(100, 200, 10)
omglist <- 0
parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
### MCMC sizes
Nout <- 100
Nthin <- 1
Nbi <- 10
## Select proposals
K <- 3                        # Choose 3 proposals
istart_SEQ <- 6               # Start with middle
istart_MNX <- istart_ENT <- c(6, 2, 10)
cooling_MNX <- .05/log((0:24 \%/\% 5)*5 + exp(1))
cooling_ENT <- .3/log((0:49 \%/\% 10)*10 + exp(1))
prop_SEQ <- select_proposals_SEQ(pargrid = parlist, K = K,
                                 istart = istart_SEQ,
                                 relativeSE = TRUE, 
                                 N1 = Nout, N2 = Nout,
                                 Nthin = Nthin, Nbi = Nbi,
                                 formula = formula, family = family,
                                 data = rhizoctonia, weights = Total, 
                                 atsample = atsample, corrfcn = corrf,
                                 betm0 = betm0, betQ0 = betQ0,
                                 ssqdf = ssqdf, ssqsc = ssqsc,
                                 dispersion = 1, longlat = FALSE,
                                 nbatch1 = 0.5, nbatch2 = 0.5,
                                 bvmethod = "TukeyHanning",
                                 transf = "mu")
prop_MNX <- select_proposals_MNX(pargrid = parlist,
                                 istart = istart_MNX, nfix = 1L,
                                 cooling = cooling_MNX, 
                                 relativeSE = TRUE, 
                                 N1 = Nout, N2 = Nout,
                                 Nthin = Nthin, Nbi = Nbi,
                                 formula = formula, family = family,
                                 data = rhizoctonia, weights = Total, 
                                 atsample = atsample, corrfcn = corrf,
                                 betm0 = betm0, betQ0 = betQ0,
                                 ssqdf = ssqdf, ssqsc = ssqsc,
                                 dispersion = 1, longlat = FALSE,
                                 nbatch1 = 0.5, nbatch2 = 0.5,
                                 bvmethod = "TukeyHanning",
                                 transf = "mu",
                                 verbose = TRUE)
prop_ENT <- select_proposals_ENT(pargrid = parlist,
                                 istart = istart_ENT, nfix = 1L,
                                 cooling = cooling_ENT, 
                                 relativeSE = TRUE, 
                                 N1 = Nout, 
                                 Nthin = Nthin, Nbi = Nbi,
                                 formula = formula, family = family,
                                 data = rhizoctonia, weights = Total, 
                                 atsample = atsample, corrfcn = corrf,
                                 betm0 = betm0, betQ0 = betQ0,
                                 ssqdf = ssqdf, ssqsc = ssqsc,
                                 dispersion = 1, longlat = FALSE,
                                 nbatch1 = 0.5, nbatch2 = 0.5,
                                 bvmethod = "TukeyHanning",
                                 transf = "mu",
                                 verbose = TRUE)
}
}
\references{
Roy, V., & Evangelou, E. (2018). Selection of proposal
  distributions for generalized importance sampling estimators.
  arXiv preprint arXiv:1805.00829.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mcsp2.R
\name{mcmcmake}
\alias{mcmcmake}
\title{Convert to an \code{\link[coda]{mcmc}} object}
\usage{
mcmcmake(...)
}
\arguments{
\item{...}{Output(s) from the functions mentioned in the Details.}
}
\value{
An mcmc object.
}
\description{
Convert to an \code{\link[coda]{mcmc}} object.
}
\details{
This function takes as input the one or more output(s) from
function \code{\link{mcsglmm}} or \code{\link{mcstrga}} and
returns an \code{\link[coda]{mcmc}} object or an
\code{\link[coda]{mcmc.list}} object for coda. The function
requires the \code{coda} package to be installed.

The spatial random field components are assigned the names
\code{z_*} where \code{*} is a number beginning at 1. Similarly,
the regressor coefficients are assigned the names \code{beta_*} if
not unique, or simply \code{beta} if there is only one regressor.
The names \code{ssq}, \code{tsq}, \code{phi}, \code{omg}
correspond to the partial sill, measurement error variance,
spatial range, and relative nugget parameters respectively.
}
\examples{
\dontrun{
 ### Load the data
data(rhizoctonia)
rhiz <- na.omit(rhizoctonia)
rhiz$IR <- rhiz$Infected/rhiz$Total # Incidence rate of the
                              # rhizoctonia disease
 ### Define the model
corrf <- "spherical"
ssqdf <- 1
ssqsc <- 1
tsqdf <- 1
tsqsc <- 1
betm0 <- 0
betQ0 <- diag(.01, 2, 2)
phiprior <- c(200, 1, 1000, 100) # U(100, 300)
phisc <- 1
omgprior <- c(3, 1, 1000, 0) # U(0, 3)
omgsc <- 1.3
linkp <- 1
## MCMC parameters
Nout <- 100
Nbi <- 0
Nthin <- 1
 ### Run MCMC
sample <- mcstrga(Yield ~ IR, data = rhiz,
                  atsample = ~ Xcoord + Ycoord, corrf = corrf,
                  Nout = Nout, Nthin = Nthin,
                  Nbi = Nbi, betm0 = betm0, betQ0 = betQ0,
                  ssqdf = ssqdf, ssqsc = ssqsc,
                  tsqdf = tsqdf, tsqsc = tsqsc,
                  linkp = linkp,
                  corrprior = list(phi = phiprior, omg = omgprior), 
                  corrtuning = list(phi = phisc, omg = omgsc, kappa = 0),
                  test = FALSE)
mcsample <- mcmcmake(sample)
plot(mcsample[, c("phi", "omg", "beta_1", "beta_2", "ssq", "tsq")],
     density = FALSE)
summary(mcsample[, c("phi", "omg", "beta_1", "beta_2", "ssq", "tsq")])
}
}
\seealso{
Functions such as \code{\link[coda]{plot.mcmc}} and
\code{\link[coda]{summary.mcmc}} in the \code{coda} package. The
function \code{\link[base]{do.call}} can be used to pass arguments
stored in a list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calcbf.R
\name{plotbf2}
\alias{plotbf2}
\title{Plot the estimated Bayes factors}
\usage{
plotbf2(
  bf2obj,
  pars = c("linkp", "phi", "omg", "kappa"),
  profile = length(pars) > 2,
  ...
)
}
\arguments{
\item{bf2obj}{Output from the function \code{\link{bf2new}}.}

\item{pars}{A vector with the names of the parameters to plot.}

\item{profile}{Whether it should produce a profile plot or a
contour plot if the length of pars is 2.}

\item{...}{Other input to be passed to either \code{plot} or
\code{contour}.}
}
\value{
This function returns nothing.
}
\description{
This function plots the estimated logarithm Bayes factors from the
function \code{\link{bf2new}}.
}
\details{
Depending on whether \code{pars} has length 1 or 2, this function
creates a line or a contour plot of the estimated Bayes factors.
If its length is 3 or 4, then it produces multiple profile plots.
In this case the variable is fixed at different values and the
maximum Bayes factor corresponding to the fixed value is plotted
against that value.
}
\examples{
\dontrun{
data(rhizoctonia)
### Define the model
corrf <- "spherical"
kappa <- 0
ssqdf <- 1
ssqsc <- 1
betm0 <- 0
betQ0 <- .01
family <- "binomial.probit"
### Skeleton points
philist <- c(100, 140, 180)
omglist <- c(.5, 1)
parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
### MCMC sizes
Nout <- 100
Nthin <- 1
Nbi <- 0
### Take MCMC samples
runs <- list()
for (i in 1:NROW(parlist)) {
  runs[[i]] <- mcsglmm(Infected ~ 1, family, rhizoctonia, weights = Total,
                       atsample = ~ Xcoord + Ycoord,
                       Nout = Nout, Nthin = Nthin, Nbi = Nbi,
                       betm0 = betm0, betQ0 = betQ0,
                       ssqdf = ssqdf, ssqsc = ssqsc,
                       phi = parlist$phi[i], omg = parlist$omg[i],
                       linkp = parlist$linkp[i], kappa = parlist$kappa[i],
                       corrfcn = corrf,
                       corrtuning=list(phi = 0, omg = 0, kappa = 0))
}
bf <- bf1skel(runs)
bfall <- bf2new(bf, phi = seq(100, 200, 10), omg = seq(0, 2, .2))
plotbf2(bfall, c("phi", "omg"))
plotbf2(bfall, c("phi", "omg"), profile = TRUE, type = "b", ylab="log(BF)")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skelpnts.R
\name{alik_optim}
\alias{alik_optim}
\title{Log-likelihood maximisation}
\usage{
alik_optim(
  paroptim,
  formula,
  family = "gaussian",
  data,
  weights,
  subset,
  offset,
  atsample,
  corrfcn = "matern",
  np,
  betm0,
  betQ0,
  ssqdf,
  ssqsc,
  dispersion = 1,
  longlat = FALSE,
  control = list()
)
}
\arguments{
\item{paroptim}{A named list with the components "linkp", "phi",
"omg", "kappa". Each component must be numeric with length 1, 2,
or 3 with elements in increasing order. If the compontent's
length is 1, then the corresponding parameter is considered to
be fixed at that value. If 2, then the two numbers denote the
lower and upper bounds for the optimisation of that parameter
(infinities are allowed). If 3, these correspond to lower bound,
starting value, upper bound for the estimation of that
parameter.}

\item{formula}{A representation of the model in the form
\code{response ~ terms}.}

\item{family}{The distribution of the response.}

\item{data}{An optional data frame containing the variables in the
model.}

\item{weights}{An optional vector of weights. Number of replicated
samples for Gaussian and gamma, number of trials for binomial,
time length for Poisson.}

\item{subset}{An optional vector specifying a subset of
observations to be used in the fitting process.}

\item{offset}{See \code{\link[stats]{lm}}.}

\item{atsample}{A formula in the form \code{~ x1 + x2 + ... + xd}
with the coordinates of the sampled locations.}

\item{corrfcn}{Spatial correlation function. See
\code{\link{geoBayes_correlation}} for details.}

\item{np}{The number of integration points for the spatial
variance parameter sigma^2. The total number of points will be
\code{2*np + 1}.}

\item{betm0}{Prior mean for beta (a vector or scalar).}

\item{betQ0}{Prior standardised precision (inverse variance)
matrix. Can be a scalar, vector or matrix. The first two imply a
diagonal with those elements. Set this to 0 to indicate a flat
improper prior.}

\item{ssqdf}{Degrees of freedom for the scaled inverse chi-square
prior for the partial sill parameter.}

\item{ssqsc}{Scale for the scaled inverse chi-square prior for the
partial sill parameter.}

\item{dispersion}{The fixed dispersion parameter.}

\item{longlat}{How to compute the distance between locations. If
\code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
distance. See \code{\link[sp]{spDists}}.}

\item{control}{A list of control parameters for the optimisation.
See \code{\link[stats]{optim}}.}
}
\value{
The output from the function \code{\link[stats]{optim}}.
  The \code{"value"} element is the log-likelihood, not the
  negative log-likelihood.
}
\description{
Approximate log-likelihood maximisation
}
\details{
Uses the "L-BFGS-B" method of the function
\code{\link[stats]{optim}} to maximise the log-likelihood for the
parameters \code{linkp}, \code{phi}, \code{omg}, \code{kappa}.
}
\references{
Evangelou, E., & Roy, V. (2019). Estimation and prediction for
  spatial generalized linear mixed models with parametric links
  via reparameterized importance sampling. Spatial Statistics, 29,
  289-315.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geobayes-package.R
\docType{package}
\name{geoBayes}
\alias{geoBayes}
\alias{geoBayes-package}
\title{The \code{geoBayes} package}
\description{
Analysis of geostatistical data using Bayes and Empirical Bayes
methods.
}
\details{
This package provides functions to fit geostatistical data. The
data can be continuous, binary or count data and the models
implemented are flexible. Conjugate priors are assumed on some
parameters while inference on the other parameters can be done
through a full Bayesian analysis of by empirical Bayes methods.

Some demonstration examples are provided. Type \code{demo(package
= "geoBayes")} to examine them.
}
\examples{
\dontrun{
demo(package = "geoBayes")
demo(rhizoctonia1, package = "geoBayes")
demo(rhizoctonia1, package = "geoBayes")
}
}
\references{
Roy, V., Evangelou, E. and Zhu, Z. (2014). Empirical
Bayes methods for the transformed Gaussian random fields model
with additive measurement errors. In Upadhyay, S. K., Singh, U.,
Dey, D. K., and Loganathan, A., editors, \emph{Current Trends in
Bayesian Methodology with Applications}, Boca Raton, FL, USA, CRC
Press.

Roy, V., Evangelou, E., and Zhu, Z. (2015). Efficient estimation
and prediction for the Bayesian spatial generalized linear mixed
model with flexible link functions. \emph{Biometrics}, 72(1),
  289-298.

Evangelou, E., & Roy, V. (2019). Estimation and prediction for
  spatial generalized linear mixed models with parametric links
  via reparameterized importance sampling. \emph{Spatial Statistics}, 29,
  289-315.

Roy, V., & Evangelou, E. (2024). Selection of proposal
  distributions for multiple importance sampling.
  \emph{Statistica Sinica}, 34, 27-46.
}
\author{
Evangelos Evangelou <e.evangelou@maths.bath.ac.uk> and
Vivekananda Roy <vroy@iastate.edu>
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rhizoctonia-data.R
\docType{data}
\name{rhizoctonia}
\alias{rhizoctonia}
\title{Rhizoctonia root rot infections}
\format{
A data frame with 100 rows and 5 variables.
}
\source{
See reference below.
}
\usage{
data(rhizoctonia)
}
\description{
Rhizoctonia root rot infections.
}
\details{
A dataset containing the number of infected roots and the sample
coordinate. The data were collected by Dr Jim Cook at Washington
State University. 

\itemize{
\item Xcoord Longitude of the sampling location.
\item Ycoord Latitude of the sampling location.
\item Total Total number of roots sampled at that location.
\item Infected Number of infected roots found at that location.
\item Yield Barley yield at that location. These data were
obtained by hand-harvesting a 4-square-meter area in the sampling
location. One observation is missing.
}
}
\note{
We acknowledge Hao Zhang for providing these data.
}
\references{
Zhang, H. (2002). On estimation and prediction for
spatial generalized linear mixed models. \emph{Biometrics}, 58(1),
129-136.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sploglik.R
\name{sploglik}
\alias{sploglik}
\title{Spatial log likelihood}
\usage{
sploglik(pargrid, runs, transf = c("no", "mu", "wo"))
}
\arguments{
\item{pargrid}{A data frame with components "linkp", "phi", "omg",
"kappa". Each row gives a combination of the parameters to compute
the log-likelihood.}

\item{runs}{A list with outputs from the function
\code{\link{mcsglmm}} or \code{\link{mcstrga}}.}

\item{transf}{Whether to use a transformed sample for the
computations. If \code{"no"} or \code{FALSE}, it doesn't. If
\code{"mu"} or \code{TRUE}, it uses the samples for the mean. If
\code{"wo"} it uses an alternative transformation. The latter
can be used only for the families indicated by
\code{.geoBayes_models$haswo}.}
}
\value{
A matrix with number of rows the total number of samples
  in runs and number of columns the number of rows in pargrid. The
  [i,j] element of the matrix is the value of the loglikelihood at
  the ith sample when all samples in runs are put together evaluated
  at the jth parameter value.
}
\description{
Spatial joint log likelihood
}
\details{
Computes the joint log likelihood log f(y,T(z)|parameters) where
T(z) is the transformation, for each (y,z) in runs and for parameters
in pargrid up to a constant which does not depend on the
parameters. The parameters beta and sigma^2 are integrated out.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skelpnts.R
\name{alik_inla}
\alias{alik_inla}
\title{Log-likelihood approximation}
\usage{
alik_inla(
  par_vals,
  formula,
  family = "gaussian",
  data,
  weights,
  subset,
  offset,
  atsample,
  corrfcn = "matern",
  np,
  betm0,
  betQ0,
  ssqdf,
  ssqsc,
  tsqdf,
  tsqsc,
  dispersion = 1,
  longlat = FALSE
)
}
\arguments{
\item{par_vals}{A data frame with the components "linkp", "phi",
"omg", "kappa". The approximation will be computed at each row
of the data frame.}

\item{formula}{A representation of the model in the form
\code{response ~ terms}.}

\item{family}{The distribution of the response. Can be one of the
options in \code{\link{.geoBayes_models}} or
\code{"transformed.gaussian"}.}

\item{data}{An optional data frame containing the variables in the
model.}

\item{weights}{An optional vector of weights. Number of replicated
samples for Gaussian and gamma, number of trials for binomial,
time length for Poisson.}

\item{subset}{An optional vector specifying a subset of
observations to be used in the fitting process.}

\item{offset}{See \code{\link[stats]{lm}}.}

\item{atsample}{A formula in the form \code{~ x1 + x2 + ... + xd}
with the coordinates of the sampled locations.}

\item{corrfcn}{Spatial correlation function. Can be one of the
choices in \code{\link{.geoBayes_corrfcn}}.}

\item{np}{The number of integration points for the spatial
variance parameter sigma^2. The total number of points will be
\code{2*np + 1}.}

\item{betm0}{Prior mean for beta (a vector or scalar).}

\item{betQ0}{Prior standardised precision (inverse variance)
matrix. Can be a scalar, vector or matrix. The first two imply a
diagonal with those elements. Set this to 0 to indicate a flat
improper prior.}

\item{ssqdf}{Degrees of freedom for the scaled inverse chi-square
prior for the partial sill parameter.}

\item{ssqsc}{Scale for the scaled inverse chi-square prior for the
partial sill parameter.}

\item{tsqdf}{Degrees of freedom for the scaled inverse chi-square
prior for the measurement error parameter.}

\item{tsqsc}{Scale for the scaled inverse chi-square prior for the
measurement error parameter.}

\item{dispersion}{The fixed dispersion parameter.}

\item{longlat}{How to compute the distance between locations. If
\code{FALSE}, Euclidean distance, if \code{TRUE} Great Circle
distance. See \code{\link[sp]{spDists}}.}
}
\value{
A list with components
\itemize{
\item \code{par_vals} A data frame of the parameter values.
\item \code{aloglik} The approximate log-likelihood at thos
  parameter values.
}
}
\description{
Log-likelihood approximation.
}
\details{
Computes and approximation to the log-likelihood for the given
parameters using integrated nested Laplace approximations.
}
\references{
Evangelou, E., & Roy, V. (2019). Estimation and prediction for
  spatial generalized linear mixed models with parametric links
  via reparameterized importance sampling. Spatial Statistics, 29,
  289-315.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calcbf.R
\name{bf2new}
\alias{bf2new}
\title{Compute the Bayes factors at new points}
\usage{
bf2new(bf1obj, linkp, phi, omg, kappa, useCV = TRUE)
}
\arguments{
\item{bf1obj}{Output from the function \code{\link{bf1skel}} which
contains the Bayes factors and importance sampling weights.}

\item{linkp, phi, omg, kappa}{Optional scalar or vector or
\code{NULL}. If scalar or vector, the Bayes factors are calculated
at those values with respect to the reference model used in
\code{\link{bf1skel}}. If missing or \code{NULL} then the unique
values from the MCMC chains that were inputted in
\code{\link{bf1skel}} will be used.}

\item{useCV}{Whether to use control variates for finer
corrections.}
}
\value{
An array of size \code{length(linkp) * length(phi) *
length(omg) * length(kappa)} containing the Bayes factors for each
combination of the parameters.
}
\description{
Compute the Bayes factors.
}
\details{
Computes the Bayes factors using the importance weights at the new
points. The new points are taken from the grid derived by
expanding the parameter values inputted. The arguments
\code{linkp} \code{phi} \code{omg} \code{kappa} correspond to the
link function, spatial range, relative nugget, and correlation
function parameters respectively.
}
\examples{
\dontrun{
data(rhizoctonia)
### Define the model
corrf <- "spherical"
kappa <- 0
ssqdf <- 1
ssqsc <- 1
betm0 <- 0
betQ0 <- .01
family <- "binomial.probit"
### Skeleton points
philist <- c(100, 140, 180)
omglist <- c(.5, 1)
parlist <- expand.grid(linkp=0, phi=philist, omg=omglist, kappa = kappa)
### MCMC sizes
Nout <- 100
Nthin <- 1
Nbi <- 0
### Take MCMC samples
runs <- list()
for (i in 1:NROW(parlist)) {
  runs[[i]] <- mcsglmm(Infected ~ 1, family, rhizoctonia, weights = Total,
                       atsample = ~ Xcoord + Ycoord,
                       Nout = Nout, Nthin = Nthin, Nbi = Nbi,
                       betm0 = betm0, betQ0 = betQ0,
                       ssqdf = ssqdf, ssqsc = ssqsc,
                       phi = parlist$phi[i], omg = parlist$omg[i],
                       linkp = parlist$linkp[i], kappa = parlist$kappa[i],
                       corrfcn = corrf,
                       corrtuning=list(phi = 0, omg = 0, kappa = 0))
}
bf <- bf1skel(runs)
bfall <- bf2new(bf, phi = seq(100, 200, 10), omg = seq(0, 2, .2))
plotbf2(bfall, c("phi", "omg"))
}
}
\references{
Doss, H. (2010). Estimation of large families of Bayes
factors from Markov chain output. \emph{Statistica Sinica}, 20(2),
537.

Roy, V., Evangelou, E., and Zhu, Z. (2015). Efficient estimation
and prediction for the Bayesian spatial generalized linear mixed
model with flexible link functions. \emph{Biometrics}, 72(1),
  289-298.
}
